@@ -307,6 +307,8 @@ order for the change to take effect."
   (".+"        . ("∔"))
   (".-"        . ("∸"))
   (":"         . ,(agda-input-to-string-list "∶⦂ː꞉˸፥፦：﹕︓"))
+  (","         . ,(agda-input-to-string-list "ʻ،⸲⸴⹁⹉、︐︑﹐﹑，､"))
+  (";"         . ,(agda-input-to-string-list "؛⁏፤꛶；︔﹔⍮⸵;"))
   ("::"        . ("∷"))
   ("::-"       . ("∺"))
   ("-:"        . ("∹"))
@@ -945,7 +947,6 @@ order for the change to take effect."
   ("_s" . ("ₛ"))
   ("_t" . ("ₜ"))
   ("_u" . ("ᵤ"))
-  ("_v" . ("ᵥ"))
   ("_x" . ("ₓ"))
 
   ("^a" . ("ᵃ"))
