@@ -16,7 +16,7 @@ while (<>) {
      my $arg = $3;
      my $tag = "$tag_prefix-$3" =~ s/\\_/$underscore/gr;
 
-     $_ = "\n%<*$tag>\n$cmd\{$arg\}\n%</$tag>\n";
+     $_ = "%\n%<*$tag>\n$cmd\{$arg\}%\n%</$tag>\n";
    |gxe;
 
   print;
