@@ -24,6 +24,7 @@ import Agda.TypeChecking.Telescope
 import Agda.Syntax.Common (Hiding(..))
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Abstract.Pretty (prettyA)
+import qualified Agda.Syntax.Concrete.Name as C
 import qualified Text.PrettyPrint as PP
 import qualified Agda.TypeChecking.Pretty as TCM
 import Agda.Syntax.Position
@@ -35,6 +36,7 @@ import Agda.TypeChecking.Reduce (normalise)
 import Agda.Syntax.Common
 import qualified Agda.Syntax.Scope.Base as Scope
 import Agda.Syntax.Scope.Monad (withCurrentModule)
+import Agda.Syntax.Concrete.Name (NameInScope(..), LensInScope(..))
 import qualified Agda.Syntax.Abstract.Name as AN
 import qualified Agda.TypeChecking.Monad.Base as TCM
 import Agda.TypeChecking.EtaContract (etaContract)
@@ -55,6 +57,7 @@ import Agda.Utils.Impossible
 import Agda.Utils.Lens
 import Agda.Utils.Maybe
 import Agda.Utils.Null
+import Agda.Utils.Pretty ( prettyShow )
 import Agda.Utils.Tuple
 
 #include "undefined.h"
@@ -104,12 +107,14 @@ stopWithMsg msg = return $ AutoResult (Solutions []) (Just msg)
 --   If the @autoMessage@ part of the result is set to @Just msg@, the
 --   message @msg@ produced by Agsy should be displayed to the user.
 
+{-# SPECIALIZE auto :: InteractionId -> Range -> String -> TCM AutoResult #-}
 auto
-  :: InteractionId
+  :: MonadTCM tcm
+  => InteractionId
   -> Range
   -> String
-  -> TCM AutoResult
-auto ii rng argstr = do
+  -> tcm AutoResult
+auto ii rng argstr = liftTCM $ do
 
   -- Parse hints and other configuration.
   let autoOptions = parseArgs argstr
@@ -175,7 +180,7 @@ auto ii rng argstr = do
         ticks <- liftIO $ newIORef 0
 
         let exsearch initprop recinfo defdfv =
-             liftIO $ System.Timeout.timeout (getTimeOut timeout * 1000000)
+             liftIO $ System.Timeout.timeout (getTimeOut timeout * 1000)
                     $ loop 0
              where
                loop d = do
@@ -233,7 +238,7 @@ auto ii rng argstr = do
                     mv <- lookupMeta mi
                     withMetaInfo (getMetaInfo mv) $ do
                       (mi,) <$> abstractToConcrete_ e
-                  let ss = dropWhile (== ' ') . dropWhile (/= ' ') . show
+                  let ss = dropWhile (== ' ') . dropWhile (/= ' ') . prettyShow
                       disp [(_, cexpr)] = ss cexpr
                       disp cexprs = concat $ map (\ (mi, cexpr) -> ss cexpr ++ " ") cexprs
                   ticks <- liftIO $ readIORef ticks
@@ -274,7 +279,7 @@ auto ii rng argstr = do
                  ) eqcons
           res <- exsearch initprop recinfo defdfv
           riis <- map swap <$> getInteractionIdsAndMetas
-          let timeoutString | isNothing res = " after timeout (" ++ show timeout ++ "s)"
+          let timeoutString | isNothing res = " after timeout (" ++ show timeout ++ "ms)"
                             | otherwise     = ""
           if listmode then do
             rsols <- liftM reverse $ liftIO $ readIORef sols
@@ -290,10 +295,10 @@ auto ii rng argstr = do
                   withMetaInfo (getMetaInfo mv) $ do
                     e' <- abstractToConcrete_ e
                     return (mi, e')
-              let disp [(_, cexpr)] = show cexpr
+              let disp [(_, cexpr)] = prettyShow cexpr
                   disp cexprs = concat $ for cexprs $ \ (mi, cexpr) ->
                     maybe (show mi) show (lookup mi riis)
-                      ++ " := " ++ show cexpr ++ " "
+                      ++ " := " ++ prettyShow cexpr ++ " "
               ticks <- liftIO $ readIORef ticks
               stopWithMsg $ "Listing solution(s) " ++ show pick ++ "-" ++ show (pick + length rsols - 1) ++ timeoutString ++
                         "\n" ++ unlines (map (\(x, y) -> show y ++ "  " ++ disp x) $ zip cexprss [pick..])
@@ -322,7 +327,7 @@ auto ii rng argstr = do
                     -- On exception, try next solution
                     flip catchError (const $ loop terms') $ do
                       exprs <- getsols term
-                      reportSDoc "auto" 20 $ TCM.text "Trying solution " TCM.<+> TCM.prettyTCM exprs
+                      reportSDoc "auto" 20 $ "Trying solution " TCM.<+> TCM.prettyTCM exprs
                       giveress <- forM exprs $ \ (mi, expr0) -> do
                         let expr = killRange expr0
                         case lookup mi riis of
@@ -336,7 +341,7 @@ auto ii rng argstr = do
                                         let scope = getMetaScope mv
                                         ce <- abstractToConcreteScope scope ae
                                         let cmnt = if ii' == ii then agsyinfo ticks else ""
-                                        return (Just (ii', show ce ++ cmnt), Nothing)
+                                        return (Just (ii', prettyShow ce ++ cmnt), Nothing)
                            -- Andreas, 2015-05-17, Issue 1504
                            -- When Agsy produces an ill-typed solution, return nothing.
                            -- TODO: try other solution.
@@ -363,7 +368,7 @@ auto ii rng argstr = do
           let [rectyp'] = mymrectyp
           defdfv <- getdfv mi def
           myrecdef <- liftIO $ newIORef $ ConstDef {cdname = "", cdorigin = (Nothing, def), cdtype = rectyp', cdcont = Postulate, cddeffreevars = defdfv}
-          sols <- liftIO $ System.Timeout.timeout (getTimeOut timeout * 1000000) (
+          sols <- liftIO $ System.Timeout.timeout (getTimeOut timeout * 1000) (
              let r d = do
                   sols <- liftIO $ caseSplitSearch ticks __IMPOSSIBLE__ myhints meqr __IMPOSSIBLE__ d myrecdef ctx mytype pats
                   case sols of
@@ -414,15 +419,15 @@ auto ii rng argstr = do
            Just (def, _, _) | def == n -> return Nothing
            _ -> do
             cn <- withMetaInfo minfo $ runAbsToCon $ toConcrete n
-            if head (show cn) == '.' then -- not in scope
+            if C.isInScope cn == C.NotInScope then
               return Nothing
-             else do
+            else do
               c <- getConstInfo n
               ctyp <- normalise $ defType c
               cdfv <- withMetaInfo minfo $ getDefFreeVars n
               return $ case matchType cdfv tctx ctyp targettyp of
                Nothing -> Nothing
-               Just score -> Just (show cn, score)
+               Just score -> Just (prettyShow cn, score)
          ) alldefs
        else do
         let scopeinfo = clScope (getMetaInfo mv)
@@ -438,7 +443,7 @@ auto ii rng argstr = do
           cdfv <- withMetaInfo minfo $ getDefFreeVars n
           return $ case matchType cdfv tctx ctyp targettyp of
            Nothing -> Nothing
-           Just score -> Just (show cn, score)
+           Just score -> Just (prettyShow cn, score)
          ) modnames
 
       let sorthits = List.sortBy (\(_, (pa1, pb1)) (_, (pa2, pb2)) -> case compare pa2 pa1 of {EQ -> compare pb1 pb2; o -> o}) hits
