@@ -15,6 +15,7 @@ import Agda.Syntax.Concrete (exprFieldA)
 import qualified Agda.Syntax.Internal as I
 import qualified Agda.Syntax.Internal.Pattern as IP
 import qualified Agda.Syntax.Common as Cm
+import Agda.Syntax.Common (Dom(..),domInfo,unDom)
 import qualified Agda.Syntax.Abstract.Name as AN
 import qualified Agda.Syntax.Abstract as A
 import qualified Agda.Syntax.Position as SP
@@ -35,7 +36,7 @@ import Agda.TypeChecking.Monad.Builtin (constructorForm)
 import Agda.TypeChecking.Free as Free (freeIn)
 import Agda.TypeChecking.Errors ( stringTCErr )
 
-import Agda.Interaction.MakeCase (getClauseForIP)
+import Agda.Interaction.MakeCase (getClauseZipperForIP)
 
 import Agda.Auto.NarrowingSearch
 import Agda.Auto.Syntax hiding (getConst)
@@ -115,6 +116,8 @@ tomy imi icns typs = do
            return (Def narg clauses' Nothing Nothing, [])
      (cont, projfcns2) <- case defn of
       MB.Axiom {} -> return (Postulate, [])
+      MB.DataOrRecSig{} -> return (Postulate, [])
+      MB.GeneralizableVar{} -> __IMPOSSIBLE__
       MB.AbstractDefn{} -> return (Postulate, [])
       MB.Function {MB.funClauses = clauses} -> clausesToDef clauses
       -- MB.Primitive {MB.primClauses = []} -> throwError $ strMsg "Auto: Primitive functions are not supported" -- Andreas, 2013-06-17 breaks interaction/AutoMisc
@@ -298,6 +301,10 @@ literalsNotImplemented :: MB.TCM a
 literalsNotImplemented = MB.typeError $ MB.NotImplemented $
   "The Agda synthesizer (Agsy) does not support literals yet"
 
+hitsNotImplemented :: MB.TCM a
+hitsNotImplemented = MB.typeError $ MB.NotImplemented $
+  "The Agda synthesizer (Agsy) does not support HITs yet"
+
 class Conversion m a b where
   convert :: a -> m b
 
@@ -322,6 +329,7 @@ instance Conversion TOM I.Clause (Maybe ([Pat O], MExp O)) where
 
 instance Conversion TOM (Cm.Arg I.Pattern) (Pat O) where
   convert p = case Cm.unArg p of
+    I.IApplyP _ _ _ n  -> return $ PatVar (show n)
     I.VarP _ n  -> return $ PatVar (show n)
     I.DotP _ _  -> return $ PatVar "_"
       -- because Agda includes these when referring to variables in the body
@@ -337,6 +345,7 @@ instance Conversion TOM (Cm.Arg I.Pattern) (Pat O) where
     -- UNSUPPORTED CASES
     I.ProjP{}   -> lift copatternsNotImplemented
     I.LitP _    -> lift literalsNotImplemented
+    I.DefP{}    -> lift hitsNotImplemented
 
 instance Conversion TOM I.Type (MExp O) where
   convert (I.El _ t) = convert t -- sort info is thrown away
@@ -379,6 +388,7 @@ instance Conversion TOM I.Term (MExp O) where
         return $ NotM $ Pi Nothing (getHiding info) (Free.freeIn 0 y) x' (Abs (Id name) y')
       I.Sort (I.Type (I.Max [I.ClosedLevel l])) -> return $ NotM $ Sort $ Set $ fromIntegral l
       I.Sort _ -> return $ NotM $ Sort UnknownSort
+      I.Dummy{}-> return $ NotM $ Sort UnknownSort
       t@I.MetaV{} -> do
         t <- lift $ instantiate t
         case t of
@@ -422,6 +432,7 @@ fmExp m (I.Pi x y)  = fmType m (Cm.unDom x) || fmType m (I.unAbs y)
 fmExp m (I.Sort _) = False
 fmExp m (I.MetaV mid _) = mid == m
 fmExp m (I.DontCare _) = False
+fmExp _ I.Dummy{} = False
 
 fmExps :: I.MetaId -> I.Args -> Bool
 fmExps m [] = False
@@ -488,7 +499,10 @@ instance Conversion MOT (Exp O) I.Term where
                          Nothing -> (0, \ f vs -> I.Def f vs)
       frommyExps ndrop as (h name [])
     Lam hid t -> I.Lam (icnvh hid) <$> convert t
-    Pi _ hid _ x y -> I.Pi . Cm.Dom (icnvh hid) <$> convert x <*> convert y
+    Pi _ hid _ x y -> do
+      x' <- convert x
+      let dom = (Cm.defaultDom x') {domInfo = icnvh hid}
+      I.Pi dom <$> convert y
    -- maybe have case for Pi where possdep is False which produces Fun (and has to unweaken y), return $ I.Fun (Cm.Arg (icnvh hid) x') y'
     Sort (Set l) -> return $ I.Sort (I.mkType (fromIntegral l))
     Sort Type -> __IMPOSSIBLE__
@@ -540,9 +554,10 @@ modifyAbstractExpr = f
  where
   f (A.App i e1 (Cm.Arg info (Cm.Named n e2))) =
         A.App i (f e1) (Cm.Arg info (Cm.Named n (f e2)))
-  f (A.Lam i (A.DomainFree info (A.BindName n)) _)
-     | prettyShow (A.nameConcrete n) == abslamvarname =
-        A.AbsurdLam i $ Cm.argInfoHiding info
+  f (A.Lam i (A.DomainFree x) _)
+     | A.BindName n <- Cm.namedArg x
+     , prettyShow (A.nameConcrete n) == abslamvarname =
+        A.AbsurdLam i $ Cm.getHiding x
   f (A.Lam i b e) = A.Lam i b (f e)
   f (A.Rec i xs) = A.Rec i (map (mapLeft (over exprFieldA f)) xs)
   f (A.RecUpdate i e xs) = A.RecUpdate i (f e) (map (over exprFieldA f) xs)
@@ -568,6 +583,7 @@ constructPats cmap mainm clause = do
       let hid = getHiding $ Cm.argInfo p
       in case Cm.namedArg p of
        I.VarP _ n -> return ((hid, Id n) : ns, HI hid (CSPatVar $ length ns))
+       I.IApplyP _ _ _ n -> return ((hid, Id n) : ns, HI hid (CSPatVar $ length ns))
        I.ConP con _ ps -> do
         let c = I.conName con
         (c2, _) <- runStateT (getConst True c TMAll) (S {sConsts = (cmap, []), sMetas = initMapS, sEqs = initMapS, sCurMeta = Nothing, sMainMeta = mainm})
@@ -580,6 +596,8 @@ constructPats cmap mainm clause = do
         return (ns, HI hid (CSPatExp t2))
        I.ProjP{} -> copatternsNotImplemented
        I.LitP{} -> literalsNotImplemented
+       I.DefP{} -> hitsNotImplemented
+
  (names, pats) <- cnvps [] (IP.unnumberPatVars $ I.namedClausePats clause)
  return (reverse names, pats)
 
@@ -591,7 +609,8 @@ frommyClause (ids, pats, mrhs) = do
       let Id id = mid
       tel <- ctel ctx
       t' <- convert t
-      return $ I.ExtendTel (Cm.Dom (icnvh hid) t') (I.Abs id tel)
+      let dom = (Cm.defaultDom t') {domInfo = icnvh hid}
+      return $ I.ExtendTel dom (I.Abs id tel)
  tel <- ctel $ reverse ids
  let getperms 0 [] perm nv = return (perm, nv)
      getperms n [] _ _ = __IMPOSSIBLE__
@@ -706,7 +725,7 @@ findClauseDeep ii = ignoreAbstractMode $ do  -- Andreas, 2016-09-04, issue #2162
   case ipCl of
     MB.IPNoClause -> return Nothing
     MB.IPClause f clauseNo _ -> do
-      (_, c, _) <- getClauseForIP f clauseNo
+      (_, (_, c, _)) <- getClauseZipperForIP f clauseNo
       return $ Just (f, c, maybe __IMPOSSIBLE__ toplevel $ I.clauseBody c)
   where
     toplevel e =
