@@ -37,7 +37,7 @@ import Agda.TypeChecking.Pretty as P
 import Agda.Interaction.Options
 import Agda.Interaction.FindFile
 import Agda.Interaction.Highlighting.HTML
-import Agda.Interaction.Imports (getAllWarnings')
+import Agda.Interaction.Imports (getAllWarnings)
 import Agda.TypeChecking.Warnings
 
 import Agda.Utils.FileName
@@ -82,7 +82,7 @@ data Backend' opts env menv mod def = Backend'
       --   required.
   , postModule       :: env -> menv -> IsMain -> ModuleName -> [def] -> TCM mod
       -- ^ Called after all definitions of a module has been compiled.
-  , compileDef       :: env -> menv -> Definition -> TCM def
+  , compileDef       :: env -> menv -> IsMain -> Definition -> TCM def
       -- ^ Compile a single definition.
   , scopeCheckingSuffices :: Bool
     -- ^ True if the backend works if @--only-scope-checking@ is used.
@@ -92,7 +92,7 @@ data Recompile menv mod = Recompile menv | Skip mod
 
 callBackend :: String -> IsMain -> Interface -> TCM ()
 callBackend name iMain i = do
-  backends <- use stBackends
+  backends <- useTC stBackends
   case [ b | b@(Backend b') <- backends, backendName b' == name ] of
     Backend b : _ -> compilerMain b iMain i
     []            -> genericError $
@@ -129,18 +129,19 @@ embedFlag l flag = l flag
 embedOpt :: Lens' a b -> OptDescr (Flag a) -> OptDescr (Flag b)
 embedOpt l = fmap (embedFlag l)
 
-parseBackendOptions :: [Backend] -> [String] -> OptM ([Backend], CommandLineOptions)
-parseBackendOptions backends argv =
+parseBackendOptions :: [Backend] -> [String] -> CommandLineOptions -> OptM ([Backend], CommandLineOptions)
+parseBackendOptions backends argv opts0 =
   case makeAll backendWithOpts backends of
     Some bs -> do
-      let agdaFlags    = map (embedOpt lSnd) standardOptions
+      let agdaFlags    = map (embedOpt lSnd) (deadStandardOptions ++ standardOptions)
           backendFlags = do
             Some i            <- forgetAll Some $ allIndices bs
             BackendWithOpts b <- [lookupIndex bs i]
             opt               <- commandLineFlags b
             return $ embedOpt (lFst . lIndex i . bOptions) opt
-      (backends, opts) <- getOptSimple argv (agdaFlags ++ backendFlags) (embedFlag lSnd . inputFlag)
-                                            (bs, defaultOptions)
+      (backends, opts) <- getOptSimple (stripRTS argv)
+                                       (agdaFlags ++ backendFlags) (embedFlag lSnd . inputFlag)
+                                       (bs, opts0)
       opts <- checkOpts opts
       return (forgetAll forgetOpts backends, opts)
 
@@ -152,10 +153,11 @@ backendInteraction backends _ check = do
       err flag = genericError $ "Cannot mix --" ++ flag ++ " and backends (" ++ List.intercalate ", " backendNames ++ ")"
   when (optInteractive     opts) $ err "interactive"
   when (optGHCiInteraction opts) $ err "interaction"
+  when (optJSONInteraction opts) $ err "interaction-json"
   mi     <- check
 
   -- reset warnings
-  stTCWarnings .= []
+  stTCWarnings `setTCLens` []
 
   noMain <- optCompileNoMain <$> pragmaOptions
   let isMain | noMain    = NotMain
@@ -165,7 +167,7 @@ backendInteraction backends _ check = do
     Just i  -> sequence_ [ compilerMain backend isMain i | Backend backend <- backends ]
 
   -- print warnings that might have accumulated during compilation
-  ws <- filter (not . isUnsolvedWarning . tcWarning) <$> getAllWarnings' AllWarnings RespectFlags
+  ws <- filter (not . isUnsolvedWarning . tcWarning) <$> getAllWarnings AllWarnings
   unless (null ws) $ reportSDoc "warning" 1 $ P.vcat $ P.prettyTCM <$> ws
 
 
@@ -199,8 +201,8 @@ compileModule backend env isMain i = do
     Skip m         -> return m
     Recompile menv -> do
       defs <- map snd . sortDefs <$> curDefs
-      res  <- mapM (compileDef' backend env menv <=< instantiateFull) defs
+      res  <- mapM (compileDef' backend env menv isMain <=< instantiateFull) defs
       postModule backend env menv isMain (iModuleName i) res
 
-compileDef' :: Backend' opts env menv mod def -> env -> menv -> Definition -> TCM def
-compileDef' backend env menv def = setCurrentRange (defName def) $ compileDef backend env menv def
+compileDef' :: Backend' opts env menv mod def -> env -> menv -> IsMain -> Definition -> TCM def
+compileDef' backend env menv isMain def = setCurrentRange (defName def) $ compileDef backend env menv isMain def
