@@ -166,7 +166,7 @@ ghcPreModule _ m ifile = ifM uptodate noComp yesComp
       m   <- show . A.mnameToConcrete <$> curMName
       out <- outFile_
       reportSLn "compile.ghc" 1 $ repl [m, ifile, out] "Compiling <<0>> in <<1>> to <<2>>"
-      stImportedModules .= Set.empty  -- we use stImportedModules to accumulate the required Haskell imports
+      stImportedModules `setTCLens` Set.empty  -- we use stImportedModules to accumulate the required Haskell imports
       return (Recompile ())
 
 ghcPostModule :: GHCOptions -> GHCModuleEnv -> IsMain -> ModuleName -> [[HS.Decl]] -> TCM IsMain
@@ -181,7 +181,7 @@ ghcPostModule _ _ _ _ defs = do
     (map fakeDecl (hsImps ++ code) ++ concat defs)
   hasMainFunction <$> curIF
 
-ghcCompileDef :: GHCOptions -> GHCModuleEnv -> Definition -> TCM [HS.Decl]
+ghcCompileDef :: GHCOptions -> GHCModuleEnv -> IsMain -> Definition -> TCM [HS.Decl]
 ghcCompileDef _ = definition
 
 -- Compilation ------------------------------------------------------------
@@ -215,7 +215,7 @@ imports = (hsImps ++) <$> imps where
   decl m = HS.ImportDecl m True Nothing
 
   mnames :: TCM [ModuleName]
-  mnames = Set.elems <$> use stImportedModules
+  mnames = Set.elems <$> useTC stImportedModules
 
   uniq :: [HS.ModuleName] -> [HS.ModuleName]
   uniq = List.map head . List.group . List.sort
@@ -224,18 +224,18 @@ imports = (hsImps ++) <$> imps where
 -- Main compiling clauses
 --------------------------------------------------
 
-definition :: GHCModuleEnv -> Definition -> TCM [HS.Decl]
+definition :: GHCModuleEnv -> IsMain -> Definition -> TCM [HS.Decl]
 -- ignore irrelevant definitions
 {- Andreas, 2012-10-02: Invariant no longer holds
 definition kit (Defn NonStrict _ _  _ _ _ _ _ _) = __IMPOSSIBLE__
 -}
-definition env Defn{defArgInfo = info, defName = q} | isIrrelevant info = do
+definition _env _isMain Defn{defArgInfo = info, defName = q} | not $ usableModality info = do
   reportSDoc "compile.ghc.definition" 10 $
-    text "Not compiling" <+> prettyTCM q <> text "."
+    "Not compiling" <+> prettyTCM q <> "."
   return []
-definition env Defn{defName = q, defType = ty, theDef = d} = do
+definition env isMain def@Defn{defName = q, defType = ty, theDef = d} = do
   reportSDoc "compile.ghc.definition" 10 $ vcat
-    [ text "Compiling" <+> prettyTCM q <> text ":"
+    [ "Compiling" <+> prettyTCM q <> ":"
     , nest 2 $ text (show d)
     ]
   pragma <- getHaskellPragma q
@@ -243,7 +243,7 @@ definition env Defn{defName = q, defType = ty, theDef = d} = do
   mlist  <- getBuiltinName builtinList
   minf   <- getBuiltinName builtinInf
   mflat  <- getBuiltinName builtinFlat
-  checkTypeOfMain q ty $ do
+  checkTypeOfMain isMain q def $ do
     infodecl q <$> case d of
 
       _ | Just HsDefn{} <- pragma, Just q == mflat ->
@@ -303,6 +303,8 @@ definition env Defn{defName = q, defType = ty, theDef = d} = do
                  , sharpC
                  ]
 
+      DataOrRecSig{} -> __IMPOSSIBLE__
+
       Axiom{} -> do
         ar <- typeArity ty
         return $ [ compiledTypeSynonym q ty ar | Just (HsType r ty) <- [pragma] ] ++
@@ -312,18 +314,26 @@ definition env Defn{defName = q, defType = ty, theDef = d} = do
       Function{} -> function pragma $ functionViaTreeless q
 
       Datatype{ dataPars = np, dataIxs = ni, dataClause = cl, dataCons = cs }
-        | Just (HsData r ty hsCons) <- pragma -> setCurrentRange r $ do
+        | Just hsdata@(HsData r ty hsCons) <- pragma -> setCurrentRange r $ do
+        reportSDoc "compile.ghc.definition" 40 $ hsep $
+          [ "Compiling data type with COMPILE pragma ...", pretty hsdata ]
         computeErasedConstructorArgs q
         ccscov <- constructorCoverageCode q (np + ni) cs ty hsCons
         cds <- mapM compiledcondecl cs
-        return $ tvaldecl q (dataInduction d) (np + ni) [] (Just __IMPOSSIBLE__) ++
-                 [compiledTypeSynonym q ty np] ++ cds ++ ccscov
+        let result = concat $
+              [ tvaldecl q (dataInduction d) (np + ni) [] (Just __IMPOSSIBLE__)
+              , [ compiledTypeSynonym q ty np ]
+              , cds
+              , ccscov
+              ]
+        return result
       Datatype{ dataPars = np, dataIxs = ni, dataClause = cl,
                 dataCons = cs, dataInduction = ind } -> do
         computeErasedConstructorArgs q
         cds <- mapM (flip condecl ind) cs
         return $ tvaldecl q ind (np + ni) cds cl
       Constructor{} -> return []
+      GeneralizableVar{} -> return []
       Record{ recPars = np, recClause = cl, recConHead = con,
               recInduction = ind } ->
         let -- Non-recursive record types are treated as being
@@ -363,7 +373,7 @@ definition env Defn{defName = q, defType = ty, theDef = d} = do
       _ -> return ccls
 
   functionViaTreeless :: QName -> TCM [HS.Decl]
-  functionViaTreeless q = caseMaybeM (toTreeless q) (pure []) $ \ treeless -> do
+  functionViaTreeless q = caseMaybeM (toTreeless LazyEvaluation q) (pure []) $ \ treeless -> do
 
     used <- getCompiledArgUse q
     let dostrip = any not used
@@ -485,9 +495,7 @@ checkCover q ty n cs hsCons = do
         return $ HS.Alt pat (HS.UnGuardedRhs $ HS.unit_con) emptyBinds
 
   cs <- zipWithM makeClause cs hsCons
-  let rhs = case cs of
-              [] -> fakeExp "()" -- There is no empty case statement in Haskell
-              _  -> HS.Case (HS.Var $ HS.UnQual $ HS.Ident "x") cs
+  let rhs = HS.Case (HS.Var $ HS.UnQual $ HS.Ident "x") cs
 
   return [ HS.TypeSig [unqhname "cover" q] $ fakeType $ unwords (ty : tvs) ++ " -> ()"
          , HS.FunBind [HS.Match (unqhname "cover" q) [HS.PVar $ HS.Ident "x"]
@@ -783,6 +791,7 @@ writeModule (HS.Module m ps imp ds) = do
   where
   p = HS.LanguagePragma $ List.map HS.Ident $
         [ "EmptyDataDecls"
+        , "EmptyCase"
         , "ExistentialQuantification"
         , "ScopedTypeVariables"
         , "NoMonomorphismRestriction"
@@ -825,7 +834,7 @@ callGHC opts modIsMain mods = do
 
   -- Warn if no main function and not --no-main
   when (modIsMain /= isMain) $
-    genericWarning =<< fsep (pwords "No main function defined in" ++ [prettyTCM agdaMod <> text "."] ++
+    genericWarning =<< fsep (pwords "No main function defined in" ++ [prettyTCM agdaMod <> "."] ++
                              pwords "Use --no-main to suppress this warning.")
 
   let overridableArgs =
