@@ -26,6 +26,9 @@ import Agda.Compiler.Common
 import Agda.Utils.Impossible
 #include "undefined.h"
 
+type HaskellCode = String
+type HaskellType = String
+
 -- | GHC backend translation pragmas.
 data HaskellPragma
   = HsDefn Range HaskellCode
@@ -44,6 +47,16 @@ instance HasRange HaskellPragma where
   getRange (HsData   r _ _) = r
   getRange (HsExport r _)   = r
 
+instance Pretty HaskellPragma where
+  pretty = \case
+    HsDefn   _r hsCode        -> equals <+> text hsCode
+    HsType   _r hsType        -> equals <+> text hsType
+    HsData   _r hsType hsCons -> hsep $
+      [ equals, "data", text hsType
+      , parens $ hsep $ map text $ List.intersperse "|" hsCons
+      ]
+    HsExport _r hsCode        -> "as" <+> text hsCode
+
 -- Syntax for Haskell pragmas:
 --  HsDefn CODE       "= CODE"
 --  HsType TYPE       "= type TYPE"
@@ -67,8 +80,8 @@ parsePragma (CompilerPragma r s) =
     barP = skipSpaces *> char '|'
 
     -- quite liberal
-    isIdent c = isAlphaNum c || elem c "_.':[]"
-    isOp c    = not $ isSpace c || elem c "()"
+    isIdent c = isAlphaNum c || elem c ("_.':[]" :: String)
+    isOp c    = not $ isSpace c || elem c ("()" :: String)
     hsIdent = fst <$> gather (choice
                 [ string "()"
                 , many1 (satisfy isIdent)
@@ -114,7 +127,7 @@ sanityCheckPragma def (Just HsDefn{}) =
       recOrDataErr which =
         typeError $ GenericDocError $
           sep [ text $ "Bad COMPILE GHC pragma for " ++ which ++ " type. Use"
-              , text "{-# COMPILE GHC <Name> = data <HsData> (<HsCon1> | .. | <HsConN>) #-}" ]
+              , "{-# COMPILE GHC <Name> = data <HsData> (<HsCon1> | .. | <HsConN>) #-}" ]
 sanityCheckPragma def (Just HsData{}) =
   case theDef def of
     Datatype{} -> return ()
