@@ -5,7 +5,7 @@ module Agda.Compiler.MAlonzo.Primitives where
 import Control.Monad.State
 import Data.Char
 import qualified Data.List as List
-import Data.Map as M
+import qualified Data.Map as Map
 import Data.Maybe
 
 import Agda.Compiler.Common
@@ -26,29 +26,46 @@ import Agda.Utils.Either
 import Agda.Utils.Except
 import Agda.Utils.Lens
 import Agda.Utils.Monad
+import Agda.Utils.Pretty (prettyShow)
 import qualified Agda.Utils.HashMap as HMap
 import qualified Agda.Utils.Haskell.Syntax as HS
 
 #include "undefined.h"
 import Agda.Utils.Impossible
 
-isMainFunction :: QName -> Bool
-isMainFunction q = "main" == show (nameConcrete $ qnameName q)
+-- Andreas, 2019-04-29, issue #3731: exclude certain kinds of names, like constructors.
+-- TODO: Also only consider top-level definition (not buried inside a module).
+isMainFunction :: QName -> Defn -> Bool
+isMainFunction q = \case
+    Axiom{}                             -> perhaps
+    Function{ funProjection = Nothing } -> perhaps
+    Function{ funProjection = Just{}  } -> no
+    AbstractDefn{}                      -> no
+    GeneralizableVar{}                  -> no
+    DataOrRecSig{}                      -> no
+    Datatype{}                          -> no
+    Record{}                            -> no
+    Constructor{}                       -> no
+    Primitive{}                         -> no
+  where
+  perhaps = "main" == prettyShow (nameConcrete $ qnameName q)  -- ignores the qualification!?
+  no      = False
 
 hasMainFunction :: Interface -> IsMain
 hasMainFunction i
-  | List.any isMainFunction names = IsMain
-  | otherwise                     = NotMain
+  | List.any (\ (x, def) -> isMainFunction x $ theDef def) names = IsMain
+  | otherwise = NotMain
   where
-    names = HMap.keys $ iSignature i ^. sigDefinitions
+    names = HMap.toList $ iSignature i ^. sigDefinitions
 
 -- | Check that the main function has type IO a, for some a.
-checkTypeOfMain :: QName -> Type -> TCM [HS.Decl] -> TCM [HS.Decl]
-checkTypeOfMain q ty ret
-  | not (isMainFunction q) = ret
+checkTypeOfMain :: IsMain -> QName -> Definition -> TCM [HS.Decl] -> TCM [HS.Decl]
+checkTypeOfMain NotMain q def ret = ret
+checkTypeOfMain  IsMain q def ret
+  | not (isMainFunction q $ theDef def) = ret
   | otherwise = do
     Def io _ <- primIO
-    ty <- normalise ty
+    ty <- normalise $ defType def
     case unEl ty of
       Def d _ | d == io -> (mainAlias :) <$> ret
       _                 -> do
@@ -115,7 +132,7 @@ importsForPrim =
 xForPrim :: [(String, TCM [a])] -> TCM [a]
 xForPrim table = do
   qs <- HMap.keys <$> curDefs
-  bs <- toList <$> gets stBuiltinThings
+  bs <- Map.toList <$> getsTC stBuiltinThings
   let getName (Builtin (Def q _))    = q
       getName (Builtin (Con q _ _))  = conName q
       getName (Builtin (Lam _ b))    = getName (Builtin $ unAbs b)
@@ -147,6 +164,9 @@ primBody s = maybe unimplemented (fromRight (hsVarUQ . HS.Ident) <$>) $
   , "primLevelSuc"    |-> return "(\\ _ -> ())"
   , "primLevelMax"    |-> return "(\\ _ _ -> ())"
 
+  -- Sorts
+  , "primSetOmega"    |-> return "()"
+
   -- Natural number functions
   , "primNatPlus"      |-> binNat "(+)"
   , "primNatMinus"     |-> binNat "(\\ x y -> max 0 (x - y))"
@@ -159,6 +179,7 @@ primBody s = maybe unimplemented (fromRight (hsVarUQ . HS.Ident) <$>) $
   -- Machine word functions
   , "primWord64ToNat"   |-> return "MAlonzo.RTE.word64ToNat"
   , "primWord64FromNat" |-> return "MAlonzo.RTE.word64FromNat"
+  , "primWord64ToNatInjective" |-> return "erased"
 
   -- Floating point functions
   , "primNatToFloat"        |-> return "(fromIntegral :: Integer -> Double)"
@@ -204,6 +225,7 @@ primBody s = maybe unimplemented (fromRight (hsVarUQ . HS.Ident) <$>) $
   , "primCharToNat" |-> return "(fromIntegral . fromEnum :: Char -> Integer)"
   , "primNatToChar" |-> return "(toEnum . fromIntegral :: Integer -> Char)"
   , "primShowChar"  |-> return "(Data.Text.pack . show :: Char -> Data.Text.Text)"
+  , "primCharToNatInjective" |-> return "erased"
 
   -- String functions
   , "primStringToList"   |-> return "Data.Text.unpack"
@@ -211,6 +233,7 @@ primBody s = maybe unimplemented (fromRight (hsVarUQ . HS.Ident) <$>) $
   , "primStringAppend"   |-> binAsis "Data.Text.append" "Data.Text.Text"
   , "primStringEquality" |-> rel "(==)" "Data.Text.Text"
   , "primShowString"     |-> return "(Data.Text.pack . show :: Data.Text.Text -> Data.Text.Text)"
+  , "primStringToListInjective" |-> return "erased"
 
   -- Reflection
   , "primQNameEquality"   |-> rel "(==)" "MAlonzo.RTE.QName"
@@ -225,10 +248,12 @@ primBody s = maybe unimplemented (fromRight (hsVarUQ . HS.Ident) <$>) $
   , "primForce"      |-> return "\\ _ _ _ _ x f -> f $! x"
   , "primForceLemma" |-> return "erased"
 
-  -- Trust me
-  , ("primTrustMe"       , Right <$> do
+  -- Erase
+  , ("primEraseEquality", Right <$> do
        refl <- primRefl
-       closedTerm =<< (closedTermToTreeless $ lam "a" (lam "A" (lam "x" (lam "y" refl)))))
+       let erase = hLam "a" $ hLam "A" $ hLam "x" $ hLam "y" $ nLam "eq" refl
+       closedTerm =<< closedTermToTreeless LazyEvaluation erase
+    )
   ]
   where
   x |-> s = (x, Left <$> s)
@@ -248,9 +273,12 @@ primBody s = maybe unimplemented (fromRight (hsVarUQ . HS.Ident) <$>) $
       "(<<0>> :: Integer -> Integer -> Bool)"
   rel op ty  = rel' "" op ty
   opty t = t ++ "->" ++ t ++ "->" ++ t
-  unimplemented = typeError $ NotImplemented s
+  axiom_prims = ["primIMin","primIMax","primINeg","primPartial","primPartialP","primPFrom1","primPOr","primComp"]
+  unimplemented | s `List.elem` axiom_prims = return $ rtmError $ "primitive with no body evaluated: " ++ s
+                | otherwise = typeError $ NotImplemented s
 
-  lam x t = Lam (setHiding Hidden defaultArgInfo) (Abs x t)
+  hLam x t = Lam (setHiding Hidden defaultArgInfo) (Abs x t)
+  nLam x t = Lam (setHiding NotHidden defaultArgInfo) (Abs x t)
 
 noCheckCover :: QName -> TCM Bool
 noCheckCover q = (||) <$> isBuiltin q builtinNat <*> isBuiltin q builtinInteger
