@@ -17,8 +17,6 @@
 {-# LANGUAGE CPP #-}
 module Agda.Compiler.Treeless.Builtin (translateBuiltins) where
 
-import Control.Applicative
-
 import qualified Agda.Syntax.Internal as I
 import Agda.Syntax.Abstract.Name (QName)
 import Agda.Syntax.Position
@@ -47,6 +45,8 @@ data BuiltinKit = BuiltinKit
   , isLess   :: QName -> Bool
   , isEqual  :: QName -> Bool
   , isForce  :: QName -> Bool
+  , isWord64FromNat :: QName -> Bool
+  , isWord64ToNat   :: QName -> Bool
   }
 
 builtinKit :: TCM BuiltinKit
@@ -60,6 +60,8 @@ builtinKit =
              <*> isB def builtinNatLess
              <*> isB def builtinNatEquals
              <*> isP pf  "primForce"
+             <*> isP pf  "primWord64FromNat"
+             <*> isP pf  "primWord64ToNat"
   where
     con (I.Con c _ _) = pure $ I.conName c
     con _           = Nothing
@@ -91,12 +93,14 @@ transform BuiltinKit{..} = tr
              | isTimes f  -> TPrim PMul
              | isLess f   -> TPrim PLt
              | isEqual f  -> TPrim PEqI
+             | isWord64ToNat f   -> TPrim P64ToI
+             | isWord64FromNat f -> TPrim PITo64
         -- Note: Don't do this for builtinNatMinus! PSub is integer minus and
         --       builtin minus is monus. The simplifier will do it if it can see
         --       that it won't underflow.
 
-      TApp (TDef q) [_, _, _, _, e, f]
-        | isForce q -> tr $ TLet e $ tOp PSeq (TVar 0) $ mkTApp (raise 1 f) [TVar 0]
+      TApp (TDef q) (_ : _ : _ : _ : e : f : es)
+        | isForce q -> tr $ TLet e $ mkTApp (tOp PSeq (TVar 0) $ mkTApp (raise 1 f) [TVar 0]) es
 
       TApp (TCon s) [e] | isSuc s ->
         case tr e of
@@ -112,7 +116,7 @@ transform BuiltinKit{..} = tr
           e | Just (i, e) <- plusKView e -> tNegPlusK (i + 1) e
           e -> tNegPlusK 1 e
 
-      TCase e t d bs -> TCase e (caseType t bs) (tr d) $ concatMap trAlt bs
+      TCase e t d bs -> TCase e (inferCaseType t bs) (tr d) $ concatMap trAlt bs
         where
           trAlt b = case b of
             TACon c 0 b | isZero c -> [TALit (LitNat noRange 0) (tr b)]
@@ -175,16 +179,18 @@ transform BuiltinKit{..} = tr
       TErased{} -> t
       TError{}  -> t
 
+      TCoerce a -> TCoerce (tr a)
+
       TLam b                  -> TLam (tr b)
       TApp a bs               -> TApp (tr a) (map tr bs)
       TLet e b                -> TLet (tr e) (tr b)
 
-    caseType t (TACon c _ _ : _)
-      | isZero c   = CTNat
-      | isSuc c    = CTNat
-      | isPos c    = CTInt
-      | isNegSuc c = CTInt
-    caseType t _ = t
+    inferCaseType t (TACon c _ _ : _)
+      | isZero c   = t { caseType = CTNat }
+      | isSuc c    = t { caseType = CTNat }
+      | isPos c    = t { caseType = CTInt }
+      | isNegSuc c = t { caseType = CTInt }
+    inferCaseType t _ = t
 
     nPlusKView (TAGuard (TApp (TPrim PGeq) [TVar 0, (TLit (LitNat _ k))])
                         (TLet (TApp (TPrim PSub) [TVar 0, (TLit (LitNat _ j))]) b))
