@@ -63,8 +63,8 @@ computeErasedConstructorArgs d = do
   cs <- getConstructors d
   runE $ mapM_ getFunInfo cs
 
-eraseTerms :: QName -> TTerm -> TCM TTerm
-eraseTerms q t = usedArguments q t *> runE (eraseTop q t)
+eraseTerms :: QName -> EvaluationStrategy -> TTerm -> TCM TTerm
+eraseTerms q eval t = usedArguments q t *> runE (eraseTop q t)
   where
     eraseTop q t = do
       (_, h) <- getFunInfo q
@@ -88,7 +88,7 @@ eraseTerms q t = usedArguments q t *> runE (eraseTop q t)
         case h of
           Erasable -> pure TErased
           Empty    -> pure TErased
-          _        -> tApp (TDef f) <$> zipWithM eraseRel (rs ++ repeat Relevant) vs
+          _        -> tApp (TDef f) <$> zipWithM eraseRel (rs ++ repeat NotErasable) vs
 
       _ -> case t of
         TVar{}         -> pure t
@@ -117,8 +117,9 @@ eraseTerms q t = usedArguments q t *> runE (eraseTop q t)
         TError{}       -> pure t
         TCoerce e      -> TCoerce <$> erase e
 
-    tLam TErased = TErased
-    tLam t       = TLam t
+    -- #3380: this is not safe for strict backends
+    tLam TErased | eval == LazyEvaluation = TErased
+    tLam t                                = TLam t
 
     tLet e b
       | freeIn 0 b = TLet e b
@@ -145,13 +146,13 @@ eraseTerms q t = usedArguments q t *> runE (eraseTop q t)
 
     isErased t = t == TErased || isUnreachable t
 
-    eraseRel r t | erasableR r = pure TErased
-                 | otherwise   = erase t
+    eraseRel r t | erasable r = pure TErased
+                 | otherwise  = erase t
 
     eraseAlt a = case a of
       TALit l b   -> TALit l   <$> erase b
       TACon c a b -> do
-        rs <- map erasableR . fst <$> getFunInfo c
+        rs <- map erasable . fst <$> getFunInfo c
         let sub = foldr (\ e -> if e then (TErased :#) . wkS 1 else liftS 1) idS $ reverse rs
         TACon c a <$> erase (applySubst sub b)
       TAGuard g b -> TAGuard   <$> erase g <*> erase b
@@ -212,17 +213,12 @@ sumTypeInfo is = foldr plus Empty is
     plus r           Erasable    = r
     plus NotErasable NotErasable = NotErasable
 
-erasableR :: Relevance -> Bool
-erasableR Relevant   = False
-erasableR NonStrict  = True
-erasableR Irrelevant = True
-
 erasable :: TypeInfo -> Bool
 erasable Erasable    = True
 erasable Empty       = True
 erasable NotErasable = False
 
-type FunInfo = ([Relevance], TypeInfo)
+type FunInfo = ([TypeInfo], TypeInfo)
 
 getFunInfo :: QName -> E FunInfo
 getFunInfo q = memo (funMap . key q) $ getInfo q
@@ -233,20 +229,19 @@ getFunInfo q = memo (funMap . key q) $ getInfo q
         is     <- mapM (getTypeInfo . snd . dget) tel
         used   <- lift $ (++ repeat True) <$> getCompiledArgUse q
         forced <- lift $ (++ repeat NotForced) <$> getForcedArgs q
-        return (zipWith3 (uncurry . mkR . getRelevance) tel (zip forced used) is, t)
+        return (zipWith3 (uncurry . mkR . getModality) tel (zip forced used) is, t)
       h <- if isAbsurdLambdaName q then pure Erasable else getTypeInfo t
       lift $ reportSLn "treeless.opt.erase.info" 50 $ "type info for " ++ prettyShow q ++ ": " ++ show rs ++ " -> " ++ show h
-      lift $ setErasedConArgs q $ map erasableR rs
+      lift $ setErasedConArgs q $ map erasable rs
       return (rs, h)
 
-    -- Treat empty, erasable, or unused arguments as NonStrict (and thus erasable)
-    mkR :: Relevance -> IsForced -> Bool -> TypeInfo -> Relevance
-    mkR Irrelevant _ _ _  = Irrelevant
-    mkR _ _ False _       = NonStrict
-    mkR _ Forced _ _      = NonStrict
-    mkR r _ _ NotErasable = r
-    mkR _ _ _ Empty       = NonStrict
-    mkR _ _ _ Erasable    = NonStrict
+    -- Treat empty, erasable, or unused arguments as Erasable
+    mkR :: Modality -> IsForced -> Bool -> TypeInfo -> TypeInfo
+    mkR m f b i
+      | not (usableModality m) = Erasable
+      | not b                  = Erasable
+      | Forced <- f            = Erasable
+      | otherwise              = i
 
 telListView :: Type -> TCM (ListTel, Type)
 telListView t = do
@@ -296,14 +291,14 @@ getTypeInfo t0 = do
         _ | Just q `elem` msizes -> return Erasable
         Just [c] -> do
           (ts, _) <- lift $ typeWithoutParams c
-          let rs = map getRelevance ts
+          let rs = map getModality ts
           is <- mapM (getTypeInfo . snd . dget) ts
-          let er = and [ erasable i || erasableR r | (i, r) <- zip is rs ]
+          let er = and [ erasable i || not (usableModality r) | (i, r) <- zip is rs ]
           return $ if er then Erasable else NotErasable
         Just []      -> return Empty
         Just (_:_:_) -> return NotErasable
         Nothing ->
           case I.theDef def of
             I.Function{ funClauses = cs } ->
-              sumTypeInfo <$> mapM (maybe (return Empty) (getTypeInfo . El Prop) . clauseBody) cs
+              sumTypeInfo <$> mapM (maybe (return Empty) (getTypeInfo . El __DUMMY_SORT__) . clauseBody) cs
             _ -> return NotErasable
