@@ -1,5 +1,7 @@
+{-# LANGUAGE BangPatterns          #-}
 {-# LANGUAGE CPP                   #-}
 {-# LANGUAGE UndecidableInstances  #-}
+{-# LANGUAGE NondecreasingIndentation #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
@@ -13,6 +15,7 @@ import Control.Monad.Reader
 import Control.Monad.State
 import Control.Monad.Identity
 
+import qualified Data.IntMap as IntMap
 import qualified Data.Map as Map
 import qualified Data.Set as Set
 import qualified Data.List as List
@@ -20,6 +23,7 @@ import Data.Maybe
 import Data.Traversable hiding (mapM, forM, for)
 import Data.Monoid
 
+import Agda.Interaction.Options
 import qualified Agda.Syntax.Concrete as C -- ToDo: Remove with instance of ToConcrete
 import Agda.Syntax.Position
 import Agda.Syntax.Abstract as A hiding (Open, Apply, Assign)
@@ -49,14 +53,17 @@ import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.With
 import Agda.TypeChecking.Coverage
+import Agda.TypeChecking.Coverage.Match ( SplitPattern )
 import Agda.TypeChecking.Records
 import Agda.TypeChecking.Irrelevance (wakeIrrelevantVars)
-import Agda.TypeChecking.Pretty (prettyTCM)
+import Agda.TypeChecking.Pretty ( PrettyTCM, prettyTCM )
+import Agda.TypeChecking.Primitive
+import Agda.TypeChecking.Names
 import Agda.TypeChecking.Free
 import Agda.TypeChecking.CheckInternal
 import Agda.TypeChecking.SizedTypes.Solve
 import qualified Agda.TypeChecking.Pretty as TP
-import Agda.TypeChecking.Warnings (runPM)
+import Agda.TypeChecking.Warnings ( runPM, warning )
 
 import Agda.Termination.TermCheck (termMutual)
 
@@ -103,25 +110,25 @@ giveExpr force mii mi e = do
       metaTypeCheck mv (mvJudgement mv)
   where
     metaTypeCheck mv IsSort{}      = __IMPOSSIBLE__
-    metaTypeCheck mv (HasType _ t) = disableDestructiveUpdate $ do
+    metaTypeCheck mv (HasType _ t) = do
       reportSDoc "interaction.give" 20 $
-        TP.text "give: meta type =" TP.<+> prettyTCM t
+        "give: meta type =" TP.<+> prettyTCM t
       -- Here, we must be in the same context where the meta was created.
       -- Thus, we can safely apply its type to the context variables.
       ctx <- getContextArgs
       t' <- t `piApplyM` permute (takeP (length ctx) $ mvPermutation mv) ctx
-      traceCall (CheckExprCall e t') $ do
+      traceCall (CheckExprCall CmpLeq e t') $ do
         reportSDoc "interaction.give" 20 $
-          TP.text "give: instantiated meta type =" TP.<+> prettyTCM t'
+          "give: instantiated meta type =" TP.<+> prettyTCM t'
         v <- checkExpr e t'
         case mvInstantiation mv of
 
-          InstV xs v' -> unlessM ((Irrelevant ==) <$> asks envRelevance) $ do
+          InstV xs v' -> unlessM ((Irrelevant ==) <$> asksTC envRelevance) $ do
             reportSDoc "interaction.give" 20 $ TP.sep
-              [ TP.text "meta was already set to value v' = " TP.<+> prettyTCM v'
-                TP.<+> TP.text " with free variables " TP.<+> return (fsep $ map pretty xs)
-              , TP.text "now comparing it to given value v = " TP.<+> prettyTCM v
-              , TP.text "in context " TP.<+> inTopContext (prettyTCM ctx)
+              [ "meta was already set to value v' = " TP.<+> prettyTCM v'
+                TP.<+> " with free variables " TP.<+> return (fsep $ map pretty xs)
+              , "now comparing it to given value v = " TP.<+> prettyTCM v
+              , "in context " TP.<+> inTopContext (prettyTCM ctx)
               ]
             -- The number of free variables should be at least the size of the context
             -- (Ideally, if we implemented contextual type theory, it should be the same.)
@@ -131,7 +138,7 @@ giveExpr force mii mi e = do
             let (_xs1, xs2) = splitAt (size ctx) xs
             v' <- return $ foldr mkLam v' xs2
             reportSDoc "interaction.give" 20 $ TP.sep
-              [ TP.text "in meta context, v' = " TP.<+> prettyTCM v'
+              [ "in meta context, v' = " TP.<+> prettyTCM v'
               ]
             equalTerm t' v v'  -- Note: v' now lives in context of meta
 
@@ -140,13 +147,15 @@ giveExpr force mii mi e = do
             args <- getContextArgs
             nowSolvingConstraints $ assign DirEq mi args v
 
-        reportSDoc "interaction.give" 20 $ TP.text "give: meta variable updated!"
+        reportSDoc "interaction.give" 20 $ "give: meta variable updated!"
         unless (force == WithForce) $ redoChecks mii
         wakeupConstraints mi
         solveSizeConstraints DontDefaultToInfty
-        unless (force == WithForce) $ do
+        cubical <- optCubical <$> pragmaOptions
+        -- don't double check with cubical, because it gets in the way too often.
+        unless (cubical || force == WithForce) $ do
           -- Double check.
-          reportSDoc "interaction.give" 20 $ TP.text "give: double checking"
+          reportSDoc "interaction.give" 20 $ "give: double checking"
           vfull <- instantiateFull v
           checkInternal vfull t'
 
@@ -161,8 +170,8 @@ redoChecks (Just ii) = do
     IPNoClause -> return ()
     IPClause f _ _ -> do
       mb <- mutualBlockOf f
-      terErrs <- local (\ e -> e { envMutualBlock = Just mb }) $ termMutual []
-      unless (null terErrs) $ typeError $ TerminationCheckFailed terErrs
+      terErrs <- localTC (\ e -> e { envMutualBlock = Just mb }) $ termMutual []
+      unless (null terErrs) $ warning $ TerminationIssue terErrs
   -- TODO redo positivity check!
 
 -- | Try to fill hole by expression.
@@ -179,7 +188,7 @@ give force ii mr e = liftTCM $ do
   -- if Range is given, update the range of the interaction meta
   mi  <- lookupInteractionId ii
   whenJust mr $ updateMetaVarRange mi
-  reportSDoc "interaction.give" 10 $ TP.text "giving expression" TP.<+> prettyTCM e
+  reportSDoc "interaction.give" 10 $ "giving expression" TP.<+> prettyTCM e
   reportSDoc "interaction.give" 50 $ TP.text $ show $ deepUnscope e
   -- Try to give mi := e
   do setMetaOccursCheck mi DontRunMetaOccursCheck -- #589, #2710: Allow giving recursive solutions.
@@ -187,7 +196,7 @@ give force ii mr e = liftTCM $ do
     `catchError` \ case
       -- Turn PatternErr into proper error:
       PatternErr -> typeError . GenericDocError =<< do
-        withInteractionId ii $ TP.text "Failed to give" TP.<+> prettyTCM e
+        withInteractionId ii $ "Failed to give" TP.<+> prettyTCM e
       err -> throwError err
   removeInteractionPoint ii
   return e
@@ -209,7 +218,7 @@ refine force ii mr e = do
   let range = fromMaybe (getRange mv) mr
       scope = M.getMetaScope mv
   reportSDoc "interaction.refine" 10 $
-    TP.text "refining with expression" TP.<+> prettyTCM e
+    "refining with expression" TP.<+> prettyTCM e
   reportSDoc "interaction.refine" 50 $
     TP.text $ show $ deepUnscope e
   -- We try to append up to 10 meta variables
@@ -243,10 +252,10 @@ refine force ii mr e = do
                 where isX (A.Var y) | x == y = Sum 1
                       isX _                  = mempty
 
-              lamView (A.Lam _ (DomainFree _ x) e) = Just (x, e)
-              lamView (A.Lam i (DomainFull (TypedBindings r (Arg ai (TBind br (x : xs) a)))) e)
-                | null xs   = Just (dget x, e)
-                | otherwise = Just (dget x, A.Lam i (DomainFull $ TypedBindings r $ Arg ai $ TBind br xs a) e)
+              lamView (A.Lam _ (DomainFree x) e) = Just (namedArg x, e)
+              lamView (A.Lam i (DomainFull (TBind r (x : xs) a)) e)
+                | null xs   = Just (namedArg x, e)
+                | otherwise = Just (namedArg x, A.Lam i (DomainFull $ TBind r xs a) e)
               lamView _ = Nothing
 
               -- reduce beta-redexes where the argument is used at most once
@@ -321,7 +330,7 @@ showComputed :: ComputeMode -> Expr -> TCM Doc
 showComputed UseShowInstance e =
   case e of
     A.Lit (LitString _ s) -> pure (text s)
-    _                     -> (text "Not a string:" $$) <$> prettyATop e
+    _                     -> ("Not a string:" $$) <$> prettyATop e
 showComputed _ e = prettyATop e
 
 -- | Modifier for interactive commands,
@@ -345,8 +354,9 @@ data OutputConstraint a b
       | Assign b a | TypedAssign b a a | PostponedCheckArgs b [a] a a
       | IsEmptyType a
       | SizeLtSat a
-      | FindInScopeOF b a [(a,a)]
+      | FindInstanceOF b a [(a,a)]
       | PTSInstance b b
+      | PostponedCheckFunDef QName a
   deriving (Functor)
 
 -- | A subset of 'OutputConstraint'.
@@ -375,14 +385,16 @@ outputFormId (OutputForm _ _ o) = out o
       PostponedCheckArgs i _ _ _ -> i
       IsEmptyType _              -> __IMPOSSIBLE__   -- Should never be used on IsEmpty constraints
       SizeLtSat{}                -> __IMPOSSIBLE__
-      FindInScopeOF _ _ _        -> __IMPOSSIBLE__
+      FindInstanceOF _ _ _        -> __IMPOSSIBLE__
       PTSInstance i _            -> i
+      PostponedCheckFunDef{}     -> __IMPOSSIBLE__
 
 instance Reify ProblemConstraint (Closure (OutputForm Expr Expr)) where
   reify (PConstr pids cl) = enterClosure cl $ \c -> buildClosure =<< (OutputForm (getRange c) (Set.toList pids) <$> reify c)
 
 reifyElimToExpr :: I.Elim -> TCM Expr
 reifyElimToExpr e = case e of
+    I.IApply _ _ v -> appl "iapply" <$> reify (defaultArg $ v) -- TODO Andrea: endpoints?
     I.Apply v -> appl "apply" <$> reify v
     I.Proj _o f -> appl "proj" <$> reify ((defaultArg $ I.Def f []) :: Arg Term)
   where
@@ -391,6 +403,13 @@ reifyElimToExpr e = case e of
 
 instance Reify Constraint (OutputConstraint Expr Expr) where
     reify (ValueCmp cmp t u v)   = CmpInType cmp <$> reify t <*> reify u <*> reify v
+    reify (ValueCmpOnFace cmp p t u v) = CmpInType cmp <$> (reify =<< ty) <*> reify (lam_o u) <*> reify (lam_o v)
+      where
+        lam_o = I.Lam (setRelevance Irrelevant defaultArgInfo) . NoAbs "_"
+        ty = runNamesT [] $ do
+          p <- open p
+          t <- open t
+          pPi' "o" p (\ o -> t)
     reify (ElimCmp cmp _ t v es1 es2) =
       CmpElim cmp <$> reify t <*> mapM reifyElimToExpr es1
                               <*> mapM reifyElimToExpr es2
@@ -401,6 +420,9 @@ instance Reify Constraint (OutputConstraint Expr Expr) where
     reify (Guarded c pid) = do
         o  <- reify c
         return $ Guard o pid
+    reify (UnquoteTactic _ tac _ goal) = do
+        tac <- A.App defaultAppInfo_ (A.Unquote exprNoRange) . defaultNamedArg <$> reify tac
+        OfType tac <$> reify goal
     reify (UnBlock m) = do
         mi <- mvInstantiation <$> lookupMeta m
         m' <- reify (MetaV m [])
@@ -409,80 +431,82 @@ instance Reify Constraint (OutputConstraint Expr Expr) where
             e  <- reify t
             return $ Assign m' e
           PostponedTypeCheckingProblem cl _ -> enterClosure cl $ \p -> case p of
-            CheckExpr e a -> do
+            CheckExpr cmp e a -> do
                 a  <- reify a
                 return $ TypedAssign m' e a
-            CheckLambda (Arg ai (xs, mt)) body target -> do
+            CheckLambda cmp (Arg ai (xs, mt)) body target -> do
               domType <- maybe (return underscore) reify mt
               target  <- reify target
-              let bs = TypedBindings noRange $ Arg ai $
-                       TBind noRange (map (fmap A.BindName) xs) domType
+              let mkN (WithHiding h x) = setHiding h $ defaultNamedArg $ A.BindName x
+                  bs = TBind noRange (map mkN xs) domType
                   e  = A.Lam Info.exprNoRange (DomainFull bs) body
               return $ TypedAssign m' e target
             CheckArgs _ _ args t0 t1 _ -> do
               t0 <- reify t0
               t1 <- reify t1
               return $ PostponedCheckArgs m' (map (namedThing . unArg) args) t0 t1
-            UnquoteTactic tac _ goal -> do
-              tac <- A.App defaultAppInfo_ (A.Unquote exprNoRange) . defaultNamedArg <$> reify tac
-              OfType tac <$> reify goal
+            CheckProjAppToKnownPrincipalArg cmp e _ _ _ t _ _ _ -> TypedAssign m' e <$> reify t
+            DoQuoteTerm cmp v t -> do
+              tm <- A.App defaultAppInfo_ (A.QuoteTerm exprNoRange) . defaultNamedArg <$> reify v
+              OfType tm <$> reify t
           Open{}  -> __IMPOSSIBLE__
-          OpenIFS{}  -> __IMPOSSIBLE__
+          OpenInstance{}  -> __IMPOSSIBLE__
           InstV{} -> __IMPOSSIBLE__
-    reify (FindInScope m _b mcands) = FindInScopeOF
+    reify (FindInstance m _b mcands) = FindInstanceOF
       <$> (reify $ MetaV m [])
       <*> (reify =<< getMetaType m)
-      <*> (forM (fromMaybe [] mcands) $ \ (Candidate tm ty eti _) -> do
+      <*> (forM (fromMaybe [] mcands) $ \ (Candidate tm ty _) -> do
             (,) <$> reify tm <*> reify ty)
     reify (IsEmpty r a) = IsEmptyType <$> reify a
     reify (CheckSizeLtSat a) = SizeLtSat  <$> reify a
-    reify (CheckFunDef d i q cs) = __IMPOSSIBLE__
+    reify (CheckFunDef d i q cs) = do
+      a <- reify =<< defType <$> getConstInfo q
+      return $ PostponedCheckFunDef q a
     reify (HasBiggerSort a) = OfType <$> reify a <*> reify (UnivSort a)
     reify (HasPTSRule a b) = do
       (a,(x,b)) <- reify (a,b)
       return $ PTSInstance a b
 
--- ASR TODO (28 December 2014): This function will be unnecessary when
--- using a Pretty instance for OutputConstraint instead of the Show
--- instance.
-showComparison :: Comparison -> String
-showComparison cmp = " " ++ prettyShow cmp ++ " "
-
-instance (Show a,Show b) => Show (OutputForm a b) where
-  show o =
-    case o of
-      OutputForm r []   c -> show c ++ range r
-      OutputForm r pids c -> show pids ++ " " ++ show c ++ range r
+instance (Pretty a, Pretty b) => Pretty (OutputForm a b) where
+  pretty (OutputForm r pids c)
+    | null pids = sep [pretty c, prange r]
+    | otherwise = sep [pretty pids, nest 2 $ sep [pretty c, prange r]]
     where
-      range r | null s    = ""
-              | otherwise = " [ at " ++ s ++ " ]"
-        where s = show r
+      prange r | null s = empty
+               | otherwise = text $ " [ at " ++ s ++ " ]"
+        where s = prettyShow r
+
+instance (Pretty a, Pretty b) => Pretty (OutputConstraint a b) where
+  pretty oc =
+    case oc of
+      OfType e t           -> pretty e .: t
+      JustType e           -> "Type" <+> pretty e
+      JustSort e           -> "Sort" <+> pretty e
+      CmpInType cmp t e e' -> pcmp cmp e e' .: t
+      CmpElim cmp t e e'   -> pcmp cmp e e' .: t
+      CmpTypes  cmp t t'   -> pcmp cmp t t'
+      CmpLevels cmp t t'   -> pcmp cmp t t'
+      CmpTeles  cmp t t'   -> pcmp cmp t t'
+      CmpSorts cmp s s'    -> pcmp cmp s s'
+      Guard o pid          -> pretty o <?> brackets ("blocked by problem" <+> pretty pid)
+      Assign m e           -> bin (pretty m) ":=" (pretty e)
+      TypedAssign m e a    -> bin (pretty m) ":=" $ bin (pretty e) ":?" (pretty a)
+      PostponedCheckArgs m es t0 t1 ->
+        bin (pretty m) ":=" $ (parens ("_" .: t0) <+> fsep (map (paren . pretty) es)) .: t1
+        where paren d = mparens (any (`elem` [' ', '\n']) $ show d) d
+      IsEmptyType a        -> "Is empty:" <+> pretty a
+      SizeLtSat a          -> "Not empty type of sizes:" <+> pretty a
+      FindInstanceOF s t cs -> vcat
+        [ "Resolve instance argument" <?> (pretty s .: t)
+        , nest 2 $ "Candidate:"
+        , nest 4 $ vcat [ pretty v .: t | (v, t) <- cs ] ]
+      PTSInstance a b      -> "PTS instance for" <+> pretty (a, b)
+      PostponedCheckFunDef q a -> "Check definition of" <+> pretty q <+> ":" <+> pretty a
+    where
+      bin a op b = sep [a, nest 2 $ op <+> b]
+      pcmp cmp a b = bin (pretty a) (pretty cmp) (pretty b)
+      val .: ty = bin val ":" (pretty ty)
 
-instance (Show a,Show b) => Show (OutputConstraint a b) where
-    show (OfType e t)           = show e ++ " : " ++ show t
-    show (JustType e)           = "Type " ++ show e
-    show (JustSort e)           = "Sort " ++ show e
-    show (CmpInType cmp t e e') = show e ++ showComparison cmp ++ show e' ++ " : " ++ show t
-    show (CmpElim cmp t e e')   = show e ++ " == " ++ show e' ++ " : " ++ show t
-    show (CmpTypes  cmp t t')   = show t ++ showComparison cmp ++ show t'
-    show (CmpLevels cmp t t')   = show t ++ showComparison cmp ++ show t'
-    show (CmpTeles  cmp t t')   = show t ++ showComparison cmp ++ show t'
-    show (CmpSorts cmp s s')    = show s ++ showComparison cmp ++ show s'
-    show (Guard o pid)          = show o ++ " [blocked by problem " ++ prettyShow pid ++ "]"
-    show (Assign m e)           = show m ++ " := " ++ show e
-    show (TypedAssign m e a)    = show m ++ " := " ++ show e ++ " :? " ++ show a
-    show (PostponedCheckArgs m es t0 t1) = show m ++ " := (_ : " ++ show t0 ++ ") " ++ unwords (map (paren . show) es)
-                                                  ++ " : " ++ show t1
-      where paren s | elem ' ' s = "(" ++ s ++ ")"
-                    | otherwise  = s
-    show (IsEmptyType a)        = "Is empty: " ++ show a
-    show (SizeLtSat a)          = "Not empty type of sizes: " ++ show a
-    show (FindInScopeOF s t cs) = "Resolve instance argument " ++ showCand (s,t) ++ ".\n  Candidates:\n    [ " ++
-                                    List.intercalate "\n    , " (map showCand cs) ++ " ]"
-      where
-      showCand (tm,ty) = indent 6 $ show tm ++ " : " ++ show ty
-      indent n s = List.intercalate ("\n" ++ replicate n ' ') $ lines s
-    show (PTSInstance a b)      = "PTS instance for " ++ show (a,b)
 
 instance (ToConcrete a c, ToConcrete b d) =>
          ToConcrete (OutputForm a b) (OutputForm c d) where
@@ -513,21 +537,25 @@ instance (ToConcrete a c, ToConcrete b d) =>
       PostponedCheckArgs <$> toConcrete m <*> toConcrete args <*> toConcrete t0 <*> toConcrete t1
     toConcrete (IsEmptyType a) = IsEmptyType <$> toConcreteCtx TopCtx a
     toConcrete (SizeLtSat a) = SizeLtSat <$> toConcreteCtx TopCtx a
-    toConcrete (FindInScopeOF s t cs) =
-      FindInScopeOF <$> toConcrete s <*> toConcrete t
-                    <*> mapM (\(tm,ty) -> (,) <$> toConcrete tm <*> toConcrete ty) cs
+    toConcrete (FindInstanceOF s t cs) =
+      FindInstanceOF <$> toConcrete s <*> toConcrete t
+                     <*> mapM (\(tm,ty) -> (,) <$> toConcrete tm <*> toConcrete ty) cs
     toConcrete (PTSInstance a b) = PTSInstance <$> toConcrete a <*> toConcrete b
+    toConcrete (PostponedCheckFunDef q a) = PostponedCheckFunDef q <$> toConcrete a
 
 instance (Pretty a, Pretty b) => Pretty (OutputConstraint' a b) where
-  pretty (OfType' e t) = pretty e <+> text ":" <+> pretty t
+  pretty (OfType' e t) = pretty e <+> ":" <+> pretty t
 
 instance (ToConcrete a c, ToConcrete b d) =>
             ToConcrete (OutputConstraint' a b) (OutputConstraint' c d) where
   toConcrete (OfType' e t) = OfType' <$> toConcrete e <*> toConcreteCtx TopCtx t
 
 getConstraints :: TCM [OutputForm C.Expr C.Expr]
-getConstraints = liftTCM $ do
-    cs <- M.getAllConstraints
+getConstraints = getConstraints' $ const True
+
+getConstraints' :: (ProblemConstraint -> Bool) -> TCM [OutputForm C.Expr C.Expr]
+getConstraints' f = liftTCM $ do
+    cs <- filter f <$> M.getAllConstraints
     cs <- forM cs $ \c -> do
             cl <- reify c
             enterClosure cl abstractToConcrete_
@@ -566,7 +594,7 @@ getSolvedInteractionPoints all norm = concat <$> do
         case mvInstantiation mv of
           InstV{}                        -> sol (MetaV m $ map Apply args)
           Open{}                         -> unsol
-          OpenIFS{}                      -> unsol
+          OpenInstance{}                 -> unsol
           BlockedConst{}                 -> unsol
           PostponedTypeCheckingProblem{} -> unsol
 
@@ -580,21 +608,25 @@ typeOfMetaMI norm mi =
                    TCM (OutputConstraint Expr NamedMeta)
     rewriteJudg mv (HasType i t) = do
       ms <- getMetaNameSuggestion i
-      t <- normalForm norm t
+      -- Andreas, 2019-03-17, issue #3638:
+      -- Need to put meta type into correct context _before_ normalizing,
+      -- otherwise rewrite rules in parametrized modules will not fire.
       vs <- getContextArgs
+      t <- t `piApplyM` permute (takeP (size vs) $ mvPermutation mv) vs
+      t <- normalForm norm t
       let x = NamedMeta ms i
       reportSDoc "interactive.meta" 10 $ TP.vcat
         [ TP.text $ unwords ["permuting", show i, "with", show $ mvPermutation mv]
         , TP.nest 2 $ TP.vcat
-          [ TP.text "len  =" TP.<+> TP.text (show $ length vs)
-          , TP.text "args =" TP.<+> prettyTCM vs
-          , TP.text "t    =" TP.<+> prettyTCM t
-          , TP.text "x    =" TP.<+> TP.pretty x
+          [ "len  =" TP.<+> TP.text (show $ length vs)
+          , "args =" TP.<+> prettyTCM vs
+          , "t    =" TP.<+> prettyTCM t
+          , "x    =" TP.<+> TP.pretty x
           ]
         ]
       reportSDoc "interactive.meta.scope" 20 $ TP.text $ show $ getMetaScope mv
       -- Andreas, 2016-01-19, issue #1783: need piApplyM instead of just piApply
-      OfType x <$> do reify =<< t `piApplyM` permute (takeP (size vs) $ mvPermutation mv) vs
+      OfType x <$> reify t
     rewriteJudg mv (IsSort i t) = do
       ms <- getMetaNameSuggestion i
       return $ JustSort $ NamedMeta ms i
@@ -613,14 +645,14 @@ typesOfVisibleMetas norm =
 typesOfHiddenMetas :: Rewrite -> TCM [OutputConstraint Expr NamedMeta]
 typesOfHiddenMetas norm = liftTCM $ do
   is    <- getInteractionMetas
-  store <- Map.filterWithKey (openAndImplicit is) <$> getMetaStore
-  mapM (typeOfMetaMI norm) $ Map.keys store
+  store <- IntMap.filterWithKey (openAndImplicit is . MetaId) <$> getMetaStore
+  mapM (typeOfMetaMI norm . MetaId) $ IntMap.keys store
   where
   openAndImplicit is x m =
     case mvInstantiation m of
       M.InstV{} -> False
       M.Open    -> x `notElem` is
-      M.OpenIFS -> x `notElem` is  -- OR: True !?
+      M.OpenInstance -> x `notElem` is  -- OR: True !?
       M.BlockedConst{} -> True
       M.PostponedTypeCheckingProblem{} -> False
 
@@ -641,19 +673,19 @@ metaHelperType norm ii rng s = case words s of
       TelV atel _ <- telView a
       let arity = size atel
           (delta1, delta2, _, a', as', vs') = splitTelForWith tel a (map OtherType as) vs
-      a <- local (\e -> e { envPrintDomainFreePi = True }) $ do
+      a <- localTC (\e -> e { envPrintDomainFreePi = True }) $ do
         reify =<< cleanupType arity args =<< normalForm norm =<< fst <$> withFunctionType delta1 vs' as' delta2 a'
       reportSDoc "interaction.helper" 10 $ TP.vcat
-        [ TP.text "generating helper function"
-        , TP.nest 2 $ TP.text "tel    = " TP.<+> inTopContext (prettyTCM tel)
-        , TP.nest 2 $ TP.text "a      = " TP.<+> prettyTCM a
-        , TP.nest 2 $ TP.text "vs     = " TP.<+> prettyTCM vs
-        , TP.nest 2 $ TP.text "as     = " TP.<+> prettyTCM as
-        , TP.nest 2 $ TP.text "delta1 = " TP.<+> inTopContext (prettyTCM delta1)
-        , TP.nest 2 $ TP.text "delta2 = " TP.<+> inTopContext (addContext delta1 $ prettyTCM delta2)
-        , TP.nest 2 $ TP.text "a'     = " TP.<+> inTopContext (addContext delta1 $ addContext delta2 $ prettyTCM a')
-        , TP.nest 2 $ TP.text "as'    = " TP.<+> inTopContext (addContext delta1 $ prettyTCM as')
-        , TP.nest 2 $ TP.text "vs'    = " TP.<+> inTopContext (addContext delta1 $ prettyTCM vs')
+        [ "generating helper function"
+        , TP.nest 2 $ "tel    = " TP.<+> inTopContext (prettyTCM tel)
+        , TP.nest 2 $ "a      = " TP.<+> prettyTCM a
+        , TP.nest 2 $ "vs     = " TP.<+> prettyTCM vs
+        , TP.nest 2 $ "as     = " TP.<+> prettyTCM as
+        , TP.nest 2 $ "delta1 = " TP.<+> inTopContext (prettyTCM delta1)
+        , TP.nest 2 $ "delta2 = " TP.<+> inTopContext (addContext delta1 $ prettyTCM delta2)
+        , TP.nest 2 $ "a'     = " TP.<+> inTopContext (addContext delta1 $ addContext delta2 $ prettyTCM a')
+        , TP.nest 2 $ "as'    = " TP.<+> inTopContext (addContext delta1 $ prettyTCM as')
+        , TP.nest 2 $ "vs'    = " TP.<+> inTopContext (addContext delta1 $ prettyTCM vs')
         ]
       return (OfType' h a)
   where
@@ -719,6 +751,7 @@ metaHelperType norm ii rng s = case words s of
       I.Sort{}     -> pure v
       I.Level{}    -> pure v
       I.MetaV{}    -> pure v
+      I.Dummy{}    -> pure v
     onNamesElims f = traverse $ traverse $ onNamesTm f
     onNamesArgs f  = traverse $ traverse $ onNamesTm f
     onNamesAbs f   = onNamesAbs' f (stringToArgName <.> f . argNameToString)
@@ -728,9 +761,8 @@ metaHelperType norm ii rng s = case words s of
     unW "w" = return ".w"
     unW s   = return s
 
-    renameVar ('.':s) = pure s
-    renameVar "w"     = betterName
-    renameVar s       = pure s
+    renameVar "w" = betterName
+    renameVar s   = pure s
 
     betterName = do
       arg : args <- get
@@ -751,15 +783,14 @@ contextOfMeta ii norm = do
     let n         = length cxt
         localVars = zipWith raise [1..] cxt
         mkLet (x, lb) = do
-          (tm, Dom c ty) <- getOpen lb
-          return $ Dom c (x, ty)
-    letVars <- mapM mkLet . Map.toDescList =<< asks envLetBindings
-    gfilter visible . reverse <$> mapM out (letVars ++ localVars)
-  where gfilter p = catMaybes . map p
-        visible (OfType x y) | not (isNoName x) = Just (OfType' x y)
+          (tm, !dom) <- getOpen lb
+          return $ (,) x <$> dom
+    letVars <- mapM mkLet . Map.toDescList =<< asksTC envLetBindings
+    mapMaybe visible . reverse <$> mapM out (letVars ++ localVars)
+  where visible (OfType x y) | not (isNoName x) = Just (OfType' x y)
                              | otherwise        = Nothing
         visible _            = __IMPOSSIBLE__
-        out (Dom _ (x, t)) = do
+        out (Dom{unDom = (x, t)}) = do
           t' <- reify =<< normalForm norm t
           return $ OfType x t'
 
@@ -792,11 +823,12 @@ withMetaId m ret = do
   mv <- lookupMeta m
   withMetaInfo' mv ret
 
--- The intro tactic
-
+-- | The intro tactic.
+--
 -- Returns the terms (as strings) that can be
 -- used to refine the goal. Uses the coverage checker
 -- to find out which constructors are possible.
+--
 introTactic :: Bool -> InteractionId -> TCM [String]
 introTactic pmLambda ii = do
   mi <- lookupInteractionId ii
@@ -809,11 +841,12 @@ introTactic pmLambda ii = do
         TelV tel' t <- telViewUpTo' (-1) notVisible t
         -- if we cannot introduce a constructor, we try a lambda
         let fallback = do
-              TelV tel _ <- telView t
+              cubical <- optCubical <$> pragmaOptions
+              TelV tel _ <- (if cubical then telViewPath else telView) t
               reportSDoc "interaction.intro" 20 $ TP.sep
-                [ TP.text "introTactic/fallback"
-                , TP.text "tel' = " TP.<+> prettyTCM tel'
-                , TP.text "tel  = " TP.<+> prettyTCM tel
+                [ "introTactic/fallback"
+                , "tel' = " TP.<+> prettyTCM tel'
+                , "tel  = " TP.<+> prettyTCM tel
                 ]
               case (tel', tel) of
                 (EmptyTel, EmptyTel) -> return []
@@ -832,13 +865,16 @@ introTactic pmLambda ii = do
      `catchError` \_ -> return []
     _ -> __IMPOSSIBLE__
   where
+    conName :: [NamedArg SplitPattern] -> [I.ConHead]
     conName [p] = [ c | I.ConP c _ _ <- [namedArg p] ]
     conName _   = __IMPOSSIBLE__
 
-    showTCM v = show <$> prettyTCM v
+    showTCM :: PrettyTCM a => a -> TCM String
+    showTCM v = render <$> prettyTCM v
 
+    introFun :: ListTel -> TCM [String]
     introFun tel = addContext tel' $ do
-        reportSDoc "interaction.intro" 10 $ do TP.text "introFun" TP.<+> prettyTCM (telFromList tel)
+        reportSDoc "interaction.intro" 10 $ do "introFun" TP.<+> prettyTCM (telFromList tel)
         imp <- showImplicitArguments
         let okHiding0 h = imp || h == NotHidden
             -- if none of the vars were displayed, we would get a parse error
@@ -861,6 +897,7 @@ introTactic pmLambda ii = do
         makeName ("_", t) = ("x", t)
         makeName (x, t)   = (x, t)
 
+    introData :: I.Type -> TCM [String]
     introData t = do
       let tel  = telFromList [defaultDom ("_", t)]
           pat  = [defaultArg $ unnamed $ debruijnNamedVar "c" 0]
@@ -878,6 +915,10 @@ introTactic pmLambda ii = do
       let e = C.Rec noRange $ for fs $ \ f ->
             Left $ C.FieldAssignment f $ C.QuestionMark noRange Nothing
       return [ prettyShow e ]
+      -- Andreas, 2019-02-25, remark:
+      -- prettyShow is ok here since we are just printing something like
+      -- record { f1 = ? ; ... ; fn = ?}
+      -- which does not involve any qualified names, and the fi are C.Name.
 
 -- | Runs the given computation as if in an anonymous goal at the end
 --   of the top-level module.
@@ -886,7 +927,7 @@ introTactic pmLambda ii = do
 atTopLevel :: TCM a -> TCM a
 atTopLevel m = inConcreteMode $ do
   let err = typeError $ GenericError "The file has not been loaded yet."
-  caseMaybeM (use stCurrentModule) err $ \ current -> do
+  caseMaybeM (useTC stCurrentModule) err $ \ current -> do
     caseMaybeM (getVisitedModule $ toTopLevelModuleName current) __IMPOSSIBLE__ $ \ mi -> do
       let scope = iInsideScope $ miInterface mi
       tel <- lookupSection current
@@ -918,17 +959,17 @@ atTopLevel m = inConcreteMode $ do
           gamma = fromMaybe __IMPOSSIBLE__ $
                     zipWith' (\ x dom -> (x,) <$> dom) names types
       reportSDoc "interaction.top" 20 $ TP.vcat
-        [ TP.text "BasicOps.atTopLevel"
-        , TP.text "  names = " TP.<+> TP.sep (map prettyA   names)
-        , TP.text "  types = " TP.<+> TP.sep (map prettyTCM types)
+        [ "BasicOps.atTopLevel"
+        , "  names = " TP.<+> TP.sep (map prettyA   names)
+        , "  types = " TP.<+> TP.sep (map prettyTCM types)
         ]
       M.withCurrentModule current $
         withScope_ scope $
           addContext gamma $ do
             -- We're going inside the top-level module, so we have to set the
             -- checkpoint for it and all its submodules to the new checkpoint.
-            cp <- view eCurrentCheckpoint
-            stModuleCheckpoints %= fmap (const cp)
+            cp <- viewTC eCurrentCheckpoint
+            stModuleCheckpoints `modifyTCLens` fmap (const cp)
             m
 
 -- | Parse a name.
@@ -958,8 +999,10 @@ moduleContents
      -- ^ The range of the next argument.
   -> String
      -- ^ The module name.
-  -> TCM ([C.Name], [(C.Name, Type)])
-     -- ^ Module names, names paired up with corresponding types.
+  -> TCM ([C.Name], I.Telescope, [(C.Name, Type)])
+     -- ^ Module names,
+     --   context extension needed to print types,
+     --   names paired up with corresponding types.
 
 moduleContents norm rng s = traceCall ModuleContents $ do
   e <- parseExpr rng s
@@ -978,19 +1021,31 @@ moduleContents norm rng s = traceCall ModuleContents $ do
 getRecordContents
   :: Rewrite  -- ^ Amount of normalization in types.
   -> C.Expr   -- ^ Expression presumably of record type.
-  -> TCM ([C.Name], [(C.Name, Type)])
-              -- ^ Module names, names paired up with corresponding types.
+  -> TCM ([C.Name], I.Telescope, [(C.Name, Type)])
+              -- ^ Module names,
+              --   context extension,
+              --   names paired up with corresponding types.
 getRecordContents norm ce = do
   e <- toAbstract ce
   (_, t) <- inferExpr e
   let notRecordType = typeError $ ShouldBeRecordType t
   (q, vs, defn) <- fromMaybeM notRecordType $ isRecordType t
   case defn of
-    Record{ recFields = fs, recTel = tel } -> do
+    Record{ recFields = fs, recTel = rtel } -> do
       let xs   = map (nameConcrete . qnameName . unArg) fs
-          doms = telToList $ apply tel vs
-      ts <- mapM (normalForm norm) $ map (snd . unDom) doms
-      return ([], zip xs ts)
+          tel  = apply rtel vs
+          doms = flattenTel tel
+      -- Andreas, 2019-04-10, issue #3687: use flattenTel
+      -- to bring types into correct scope.
+      reportSDoc "interaction.contents.record" 20 $ TP.vcat
+        [ "getRecordContents"
+        , "  cxt  = " TP.<+> (prettyTCM =<< getContextTelescope)
+        , "  tel  = " TP.<+> prettyTCM tel
+        , "  doms = " TP.<+> prettyTCM doms
+        , "  doms'= " TP.<+> (addContext tel $ prettyTCM doms)
+        ]
+      ts <- mapM (normalForm norm . unDom) doms
+      return ([], tel, zip xs ts)
     _ -> __IMPOSSIBLE__
 
 -- | Returns the contents of the given module.
@@ -998,8 +1053,10 @@ getRecordContents norm ce = do
 getModuleContents
   :: Rewrite  -- ^ Amount of normalization in types.
   -> C.QName  -- ^ Module name.
-  -> TCM ([C.Name], [(C.Name, Type)])
-              -- ^ Module names, names paired up with corresponding types.
+  -> TCM ([C.Name], I.Telescope, [(C.Name, Type)])
+              -- ^ Module names,
+              --   context extension,
+              --   names paired up with corresponding types.
 getModuleContents norm m = do
   modScope <- getNamedScope . amodName =<< resolveModule m
   let modules :: ThingsInScope AbstractModule
@@ -1011,7 +1068,7 @@ getModuleContents norm m = do
     d <- getConstInfo $ anameName n
     t <- normalForm norm =<< (defType <$> instantiateDef d)
     return (x, t)
-  return (Map.keys modules, types)
+  return (Map.keys modules, EmptyTel, types)
 
 
 whyInScope :: String -> TCM (Maybe LocalVar, [AbstractName], [AbstractModule])
