@@ -20,8 +20,6 @@ import Agda.Syntax.Scope.Base
 import Agda.Syntax.Translation.ConcreteToAbstract
 import Agda.Syntax.Abstract.Pretty
 
-import Text.PrettyPrint
-
 import Agda.TheTypeChecker
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Monad
@@ -33,6 +31,7 @@ import Agda.TypeChecking.Warnings (runPM)
 
 import Agda.Utils.Except ( MonadError(catchError) )
 import Agda.Utils.Monad
+import Agda.Utils.Pretty
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -52,7 +51,7 @@ interaction prompt cmds eval = loop
     where
         go (Return x)       = return x
         go Continue         = loop
-        go (ContinueIn env) = local (const env) loop
+        go (ContinueIn env) = localTC (const env) loop
 
         loop =
             do  ms <- readline prompt
@@ -101,7 +100,7 @@ interactionLoop doTypeCheck =
             [ "quit"        |>  \_ -> return $ Return ()
             , "?"           |>  \_ -> continueAfter $ liftIO $ help commands
             , "reload"      |>  \_ -> do reload
-                                         ContinueIn <$> ask
+                                         ContinueIn <$> askTC
             , "constraints" |> \args -> continueAfter $ showConstraints args
             , "Context"     |> \args -> continueAfter $ showContext args
             , "give"        |> \args -> continueAfter $ giveMeta args
@@ -125,7 +124,7 @@ continueAfter m = withCurrentFile $ do
 withCurrentFile :: TCM a -> TCM a
 withCurrentFile cont = do
   mpath <- getInputFile'
-  local (\ e -> e { envCurrentPath = mpath }) cont
+  localTC (\ e -> e { envCurrentPath = mpath }) cont
 
 loadFile :: TCM () -> [String] -> TCM ()
 loadFile reload [file] = do
@@ -136,7 +135,7 @@ loadFile _ _ = liftIO $ putStrLn ":load file"
 showConstraints :: [String] -> TCM ()
 showConstraints [] =
     do  cs <- BasicOps.getConstraints
-        liftIO $ putStrLn $ unlines (List.map show cs)
+        liftIO $ putStrLn $ unlines (List.map prettyShow cs)
 showConstraints _ = liftIO $ putStrLn ":constraints [cid]"
 
 
@@ -146,22 +145,22 @@ showMetas [m] =
         withInteractionId i $ do
           s <- typeOfMeta AsIs i
           r <- getInteractionRange i
-          d <- showA s
-          liftIO $ putStrLn $ d ++ " " ++ show r
+          d <- prettyA s
+          liftIO $ putStrLn $ show d ++ " " ++ show r
 showMetas [m,"normal"] =
     do  i <- InteractionId <$> readM m
         withInteractionId i $ do
-          s <- showA =<< typeOfMeta Normalised i
+          s <- prettyA =<< typeOfMeta Normalised i
           r <- getInteractionRange i
-          liftIO $ putStrLn $ s ++ " " ++ show r
+          liftIO $ putStrLn $ show s ++ " " ++ show r
 showMetas [] =
     do  interactionMetas <- typesOfVisibleMetas AsIs
         hiddenMetas      <- typesOfHiddenMetas  AsIs
-        mapM_ (liftIO . putStrLn) =<< mapM showII interactionMetas
+        mapM_ (liftIO . print) =<< mapM showII interactionMetas
         mapM_ print' hiddenMetas
     where
-        showII o = withInteractionId (outputFormId $ OutputForm noRange [] o) $ showA o
-        showM  o = withMetaId (nmid $ outputFormId $ OutputForm noRange [] o) $ showA o
+        showII o = withInteractionId (outputFormId $ OutputForm noRange [] o) $ prettyA o
+        showM  o = withMetaId (nmid $ outputFormId $ OutputForm noRange [] o) $ prettyA o
 
         metaId (OfType i _) = i
         metaId (JustType i) = i
@@ -171,7 +170,7 @@ showMetas [] =
         print' x = do
             r <- getMetaRange $ nmid $ metaId x
             d <- showM x
-            liftIO $ putStrLn $ d ++ "  [ at " ++ show r ++ " ]"
+            liftIO $ putStrLn $ show d ++ "  [ at " ++ show r ++ " ]"
 showMetas _ = liftIO $ putStrLn $ ":meta [metaid]"
 
 
@@ -274,7 +273,7 @@ showContext (meta:args) = do
                     ["normal"] -> normalise $ raise n t
                     _          -> return $ raise n t
             d <- prettyTCM t
-            liftIO $ print $ text (I.argNameToString x) <+> text ":" <+> d
+            liftIO $ print $ text (I.argNameToString x) <+> ":" <+> d
 showContext _ = liftIO $ putStrLn ":Context meta"
 
 -- | The logo that prints when Agda is started in interactive mode.
