@@ -9,7 +9,6 @@ module Agda.Interaction.EmacsCommand
   , response
   , putResponse
   , display_info'
-  , display_warning
   , clearRunningInfo
   , clearWarning
   , displayRunningInfo
@@ -39,9 +38,9 @@ data Lisp a
 
 instance Pretty a => Pretty (Lisp a) where
   pretty (A a )     = pretty a
-  pretty (Cons a b) = parens (pretty a <+> text "." <+> pretty b)
+  pretty (Cons a b) = parens (pretty a <+> "." <+> pretty b)
   pretty (L xs)     = parens (hsep (map pretty xs))
-  pretty (Q x)      = text "'" <> pretty x
+  pretty (Q x)      = "'" <> pretty x
 
 instance Show (Lisp String) where
   showsPrec _ (A a)      = showString a
@@ -83,9 +82,6 @@ displayInBuffer buffername append header content =
 display_info' :: Bool -> String -> String -> Lisp String
 display_info' = displayInBuffer "agda2-info-action"
 
-display_warning :: String -> String -> Lisp String
-display_warning = displayInBuffer "agda2-warning-action" False
-
 ------------------------------------------------------------------------
 -- Running info
 
