@@ -3,30 +3,22 @@
 module Agda.Interaction.EmacsTop
     ( mimicGHCi
     ) where
+
 import Control.Monad.State
-
-import Data.Char
 import qualified Data.List as List
-import Data.Maybe
 
-import System.IO
-
-import Agda.Utils.Monad
 import Agda.Utils.Maybe
 import Agda.Utils.Pretty
 import Agda.Utils.String
 
 import Agda.Syntax.Common
-
 import Agda.TypeChecking.Monad
-import qualified Agda.TypeChecking.Monad.Benchmark as Bench
 
+import Agda.Interaction.AgdaTop
 import Agda.Interaction.Response as R
-import Agda.Interaction.InteractionTop
 import Agda.Interaction.EmacsCommand hiding (putResponse)
 import Agda.Interaction.Highlighting.Emacs
 import Agda.Interaction.Highlighting.Precise (TokenBased(..))
-import Agda.Interaction.Options
 
 import Agda.VersionCommit
 
@@ -37,64 +29,8 @@ import Agda.VersionCommit
 --
 --   'mimicGHCi' reads the Emacs frontend commands from stdin,
 --   interprets them and print the result into stdout.
-
 mimicGHCi :: TCM () -> TCM ()
-mimicGHCi setup = do
-    liftIO $ do
-      hSetBuffering stdout LineBuffering
-      hSetBuffering stdin  LineBuffering
-      hSetEncoding  stdout utf8
-      hSetEncoding  stdin  utf8
-
-    setInteractionOutputCallback $
-        mapM_ print <=< lispifyResponse
-
-    commands <- liftIO $ initialiseCommandQueue readCommand
-
-    handleCommand_ (lift setup) `evalStateT` initCommandState commands
-
-    opts <- commandLineOptions
-    _ <- interact' `runStateT`
-           (initCommandState commands)
-             { optionsOnReload = opts{ optAbsoluteIncludePaths = [] } }
-    return ()
-  where
-  interact' :: CommandM ()
-  interact' = do
-    Bench.reset
-    done <- Bench.billTo [] $ do
-
-      liftIO $ do
-        putStr "Agda2> "
-        hFlush stdout
-      c <- nextCommand
-      case c of
-        Done      -> return True -- Done.
-        Error s   -> liftIO (putStrLn s) >> return False
-        Command c -> do
-          maybeAbort (runInteraction c)
-          return False
-
-    lift Bench.print
-    unless done interact'
-
-  -- Reads the next command from stdin.
-
-  readCommand :: IO Command
-  readCommand = do
-    done <- isEOF
-    if done then
-      return Done
-    else do
-      r <- getLine
-      _ <- return $! length r     -- force to read the full input line
-      case dropWhile isSpace r of
-        ""          -> readCommand
-        ('-':'-':_) -> readCommand
-        _           -> case listToMaybe $ reads r of
-          Just (x, "")  -> return $ Command x
-          Just (_, rem) -> return $ Error $ "not consumed: " ++ rem
-          _             -> return $ Error $ "cannot read: " ++ r
+mimicGHCi = repl (liftIO . mapM_ print <=< liftIO . lispifyResponse) "Agda2> "
 
 -- | Given strings of goals, warnings and errors, return a pair of the
 --   body and the title for the info buffer
@@ -106,17 +42,17 @@ formatWarningsAndErrors g w e = (body, title)
     isE = not $ null e
     title = List.intercalate "," $ catMaybes
               [ " Goals"    <$ guard isG
-              , " Warnings" <$ guard isW
               , " Errors"   <$ guard isE
+              , " Warnings" <$ guard isW
               , " Done"     <$ guard (not (isG || isW || isE))
               ]
 
     body = List.intercalate "\n" $ catMaybes
              [ g                    <$ guard isG
-             , delimiter "Warnings" <$ guard (isW && (isG || isE))
-             , w                    <$ guard isW
              , delimiter "Errors"   <$ guard (isE && (isG || isW))
              , e                    <$ guard isE
+             , delimiter "Warnings" <$ guard (isW && (isG || isE))
+             , w                    <$ guard isW
              ]
 
 -- | Convert Response to an elisp value for the interactive emacs frontend.
@@ -132,9 +68,6 @@ lispifyResponse (Resp_DisplayInfo info) = return $ case info of
       where (body, title) = formatWarningsAndErrors g w e
     Info_Auto s -> f s "*Auto*"
     Info_Error s -> f s "*Error*"
-    -- FNF: if Info_Warning comes back into use, the above should be
-    -- clearWarning : f s "*Error*"
-    --Info_Warning s -> [ display_warning "*Errors*" s ] -- FNF: currently unused
     Info_Time s -> f (render s) "*Time*"
     Info_NormalForm s -> f (render s) "*Normal Form*"   -- show?
     Info_InferredType s -> f (render s) "*Inferred Type*"
@@ -162,8 +95,6 @@ lispifyResponse (Resp_ClearHighlighting tokenBased) =
          ]
 lispifyResponse Resp_DoneAborting = return [ L [ A "agda2-abort-done" ] ]
 lispifyResponse Resp_ClearRunningInfo = return [ clearRunningInfo ]
--- FNF: if Info_Warning comes back into use, the above should be
--- return [ clearRunningInfo, clearWarning ]
 lispifyResponse (Resp_RunningInfo n s)
   | n <= 1    = return [ displayRunningInfo s ]
   | otherwise = return [ L [A "agda2-verbose", A (quote s)] ]
@@ -202,7 +133,7 @@ lispifyResponse (Resp_SolveAll ps) = return
       L [ A "agda2-solveAll-action", Q . L $ concatMap prn ps ]
   ]
   where
-    prn (ii,e)= [showNumIId ii, A $ quote $ show e]
+    prn (ii,e)= [showNumIId ii, A $ quote $ prettyShow e]
 
 -- | Adds a \"last\" tag to a response.
 
@@ -212,4 +143,4 @@ lastTag n r = Cons (Cons (A "last") (A $ show n)) r
 -- | Show an iteraction point identifier as an elisp expression.
 
 showNumIId :: InteractionId -> Lisp String
-showNumIId = A . tail . show
+showNumIId = A . show . interactionId
