@@ -5,7 +5,8 @@
 module Agda.Interaction.Highlighting.Generate
   ( Level(..)
   , generateAndPrintSyntaxInfo
-  , generateTokenInfo, generateTokenInfoFromString
+  , generateTokenInfo, generateTokenInfoFromSource
+  , generateTokenInfoFromString
   , printSyntaxInfo
   , printErrorInfo, errorHighlighting
   , printUnsolvedInfo
@@ -14,7 +15,7 @@ module Agda.Interaction.Highlighting.Generate
   , warningHighlighting
   , computeUnsolvedMetaWarnings
   , computeUnsolvedConstraints
-  , storeDisambiguatedName
+  , storeDisambiguatedName, disambiguateRecordFields
   ) where
 
 import Prelude hiding (null)
@@ -29,8 +30,12 @@ import Data.Generics.Geniplate
 import qualified Data.Map as Map
 import Data.Maybe
 import Data.List ((\\), isPrefixOf)
+import qualified Data.List as List
 import qualified Data.Foldable as Fold (fold, foldMap, toList)
 import qualified Data.IntMap as IntMap
+import Data.Sequence (Seq)
+import qualified Data.Set as Set
+import qualified Data.Text.Lazy as T
 import Data.Void
 
 import Agda.Interaction.Response (Response(Resp_HighlightingInfo))
@@ -50,10 +55,12 @@ import Agda.TypeChecking.Warnings (runPM)
 import Agda.Syntax.Abstract (IsProjP(..))
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Concrete (FieldAssignment'(..))
+import Agda.Syntax.Concrete.Definitions ( DeclarationWarning(..) )
 import qualified Agda.Syntax.Common as Common
 import qualified Agda.Syntax.Concrete.Name as C
 import qualified Agda.Syntax.Concrete as C
 import Agda.Syntax.Fixity
+import Agda.Syntax.Notation
 import qualified Agda.Syntax.Info as SI
 import qualified Agda.Syntax.Internal as I
 import qualified Agda.Syntax.Literal as L
@@ -103,7 +110,7 @@ highlightAsTypeChecked rPre r m
   r'        = rToR (P.continuousPerLine r)
   delta     = rPre' `minus` r'
   clear     = mempty
-  highlight = parserBased { otherAspects = [TypeChecks] }
+  highlight = parserBased { otherAspects = Set.singleton TypeChecks }
 
   wrap rs x y = do
     p rs x
@@ -121,8 +128,8 @@ printHighlightingInfo ::
   HighlightingInfo ->
   tcm ()
 printHighlightingInfo remove info = do
-  modToSrc <- use stModuleToSource
-  method   <- view eHighlightingMethod
+  modToSrc <- useTC stModuleToSource
+  method   <- viewTC eHighlightingMethod
   liftTCM $ reportSLn "highlighting" 50 $ unlines
     [ "Printing highlighting info:"
     , show info
@@ -161,7 +168,7 @@ generateAndPrintSyntaxInfo
   -> TCM ()
 generateAndPrintSyntaxInfo decl _ _ | null $ P.getRange decl = return ()
 generateAndPrintSyntaxInfo decl hlLevel updateState = do
-  file <- fromMaybe __IMPOSSIBLE__ <$> asks envCurrentPath
+  file <- getCurrentPath
 
   reportSLn "import.iface.create" 15 $
       "Generating syntax info for " ++ filePath file ++ ' ' :
@@ -184,18 +191,18 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
       Full{} -> generateConstructorInfo modMap file kinds decl
       _      -> return mempty
 
-    cm <- P.rangeFile <$> view eRange
+    cm <- P.rangeFile <$> viewTC eRange
     reportSLn "highlighting.warning" 60 $ "current path = " ++ show cm
 
     warnInfo <- Fold.foldMap warningHighlighting
-                 . filter ((cm ==) . tcWarningOrigin) <$> use stTCWarnings
+                 . filter ((cm ==) . tcWarningOrigin) <$> useTC stTCWarnings
 
     let (from, to) = case P.rangeToInterval (P.getRange decl) of
           Nothing -> __IMPOSSIBLE__
           Just i  -> ( fromIntegral $ P.posPos $ P.iStart i
                      , fromIntegral $ P.posPos $ P.iEnd i)
     (prevTokens, (curTokens, postTokens)) <-
-      second (splitAtC to) . splitAtC from <$> use stTokens
+      second (splitAtC to) . splitAtC from <$> useTC stTokens
 
     -- theRest needs to be placed before nameInfo here since record
     -- field declarations contain QNames. constructorInfo also needs
@@ -212,8 +219,8 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
                      curTokens
 
     when updateState $ do
-      stSyntaxInfo %= mappend syntaxInfo
-      stTokens     .= prevTokens `mappend` postTokens
+      stSyntaxInfo `modifyTCLens` mappend syntaxInfo
+      stTokens     `setTCLens` (prevTokens `mappend` postTokens)
 
     ifTopLevelAndHighlightingLevelIs NonInteractive $
       printHighlightingInfo KeepHighlighting syntaxInfo
@@ -222,12 +229,9 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
   names :: [A.AmbiguousQName]
   names =
     (map I.unambiguous $
-     filter (not . extendedLambda) $
+     filter (not . isExtendedLambdaName) $
      universeBi decl) ++
     universeBi decl
-    where
-    extendedLambda :: A.QName -> Bool
-    extendedLambda = (extendedLambdaName `isPrefixOf`) . show . A.nameConcrete . A.qnameName
 
   -- Bound variables, dotted patterns, record fields, module names,
   -- the "as" and "to" symbols.
@@ -243,13 +247,17 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
     , Fold.foldMap getPatSynArgs  $ universeBi decl
     , Fold.foldMap getModuleName  $ universeBi decl
     , Fold.foldMap getModuleInfo  $ universeBi decl
-    , Fold.foldMap getNamedArg    $ universeBi decl
+    , Fold.foldMap getNamedArgE   $ universeBi decl
+    , Fold.foldMap getNamedArgP   $ universeBi decl
+    , Fold.foldMap getNamedArgB   $ universeBi decl
+    , Fold.foldMap getNamedArgL   $ universeBi decl
     ]
     where
-    bound (A.BindName n) = nameToFile modMap file [] (A.nameConcrete n) P.noRange
-                         (\isOp -> parserBased { aspect =
-                                     Just $ Name (Just Bound) isOp })
-                         (Just $ A.nameBindingSite n)
+    bound (A.BindName n) =
+      nameToFile modMap file [] (A.nameConcrete n) P.noRange
+                 (\isOp -> parserBased { aspect =
+                             Just $ Name (Just Bound) isOp })
+                 (Just $ A.nameBindingSite n)
 
     patsyn n =               -- TODO: resolve overloading
               nameToFileA modMap file (I.headAmbQ n) True $ \isOp ->
@@ -259,6 +267,7 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
     macro n = nameToFileA modMap file n True $ \isOp ->
                   parserBased { aspect = Just $ Name (Just Macro) isOp }
 
+    field :: [C.Name] -> C.Name -> File
     field m n = nameToFile modMap file m n P.noRange
                            (\isOp -> parserBased { aspect =
                                        Just $ Name (Just Field) isOp })
@@ -282,16 +291,31 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
 
     getVarAndField :: A.Expr -> File
     getVarAndField (A.Var x)            = bound $ A.BindName x
+    -- Andreas, 2018-06-09, issue #3120
+    -- The highlighting for record field tags is now created by the type checker in
+    -- function disambiguateRecordFields.
+    -- Andreas, Nisse, 2018-10-26, issue #3322
+    -- Still, we extract the highlighting info here for uses such as QuickLatex.
+    -- The aspects from the disambiguation will be merged in.
     getVarAndField (A.Rec       _ fs)   = mconcat [ field [] x | Left (FieldAssignment x _) <- fs ]
     getVarAndField (A.RecUpdate _ _ fs) = mconcat [ field [] x |      (FieldAssignment x _) <- fs ]
     getVarAndField _                    = mempty
 
-    -- Ulf, 2014-04-09: It would be nicer to have it on Named_ a, but
+    -- Ulf, 2019-01-30: It would be nicer to not have to specialize it, but
     -- you can't have polymorphic functions in universeBi.
-    getNamedArg :: Common.RString -> File
-    getNamedArg x = singleton (rToR $ P.getRange x) $
-                       parserBased { aspect =
-                         Just $ Name (Just Argument) False }
+    getNamedArgE :: Common.NamedArg A.Expr -> File
+    getNamedArgE = getNamedArg
+    getNamedArgP :: Common.NamedArg A.Pattern -> File
+    getNamedArgP = getNamedArg
+    getNamedArgB :: Common.NamedArg A.BindName -> File
+    getNamedArgB = getNamedArg
+    getNamedArgL :: Common.NamedArg A.LHSCore -> File
+    getNamedArgL = getNamedArg
+
+    getNamedArg :: Common.NamedArg a -> File
+    getNamedArg x = caseMaybe (Common.nameOf $ Common.unArg x) mempty $ \ s ->
+      singleton (rToR $ P.getRange s) $
+        parserBased { aspect = Just $ Name (Just Argument) False }
 
     getLet :: A.LetBinding -> File
     getLet (A.LetBind _ _ x _ _)     = bound x
@@ -301,11 +325,11 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
     getLet (A.LetDeclaredVariable x) = bound x
 
     getLam :: A.LamBinding -> File
-    getLam (A.DomainFree _ x) = bound x
-    getLam (A.DomainFull {})  = mempty
+    getLam (A.DomainFree x)  = bound $ Common.namedArg x
+    getLam (A.DomainFull {}) = mempty
 
     getTyped :: A.TypedBinding -> File
-    getTyped (A.TBind _ xs _) = mconcat $ map (bound . dget) xs
+    getTyped (A.TBind _ xs _) = mconcat $ map (bound . Common.namedArg) xs
     getTyped A.TLet{}         = mempty
 
     getPatSynArgs :: A.Declaration -> File
@@ -319,8 +343,14 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
       | Just _ <- isProjP e = mempty
       | otherwise =
           singleton (rToR $ P.getRange pi)
-                (parserBased { otherAspects = [DottedPattern] })
+                (parserBased { otherAspects = Set.singleton DottedPattern })
     getPattern' (A.PatternSynP _ q _) = patsyn q
+    -- Andreas, 2018-06-09, issue #3120
+    -- The highlighting for record field tags is now created by the type checker in
+    -- function disambiguateRecordFields.
+    -- Andreas, Nisse, 2018-10-26, issue #3322
+    -- Still, we extract the highlighting info here for uses such as QuickLatex.
+    -- The aspects from the disambiguation will be merged in.
     getPattern' (A.RecP _ fs) = mconcat [ field [] x | FieldAssignment x _ <- fs ]
     getPattern' _             = mempty
 
@@ -336,7 +366,7 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
     getExpr _                = mempty
 
     getFieldDecl :: A.Declaration -> File
-    getFieldDecl (A.RecDef _ _ _ _ _ _ _ fs) = Fold.foldMap extractField fs
+    getFieldDecl (A.RecDef _ _ _ _ _ _ _ _ fs) = Fold.foldMap extractField fs
       where
       extractField (A.ScopedDecl _ ds) = Fold.foldMap extractField ds
       extractField (A.Field _ x _)     = field (concreteQualifier x)
@@ -349,8 +379,7 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
       mconcat $ map (mod isTopLevelModule) xs
       where
       isTopLevelModule =
-        case catMaybes $
-             map (join .
+        case mapMaybe (join .
                   fmap (Strict.toLazy . P.srcFile) .
                   P.rStart .
                   A.nameBindingSite) xs of
@@ -368,13 +397,28 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
 -- | Generate and return the syntax highlighting information for the
 -- tokens in the given file.
 
-generateTokenInfo
-  :: AbsolutePath          -- ^ The module to highlight.
-  -> TCM CompressedFile
+generateTokenInfo :: AbsolutePath -> TCM CompressedFile
 generateTokenInfo file =
-  runPM $ tokenHighlighting <$> Pa.parseFile' Pa.tokensParser file
+  generateTokenInfoFromSource file . T.unpack =<<
+    runPM (Pa.readFilePM file)
+
+-- | Generate and return the syntax highlighting information for the
+-- tokens in the given file.
+
+generateTokenInfoFromSource
+  :: AbsolutePath
+     -- ^ The module to highlight.
+  -> String
+     -- ^ The file contents. Note that the file is /not/ read from
+     -- disk.
+  -> TCM CompressedFile
+generateTokenInfoFromSource file input =
+  runPM $ tokenHighlighting <$> fst <$> Pa.parseFile Pa.tokensParser file input
+
+-- | Generate and return the syntax highlighting information for the
+-- tokens in the given string, which is assumed to correspond to the
+-- given range.
 
--- | Same as 'generateTokenInfo' but takes a string instead of a filename.
 generateTokenInfoFromString :: P.Range -> String -> TCM CompressedFile
 generateTokenInfoFromString r _ | r == P.noRange = return mempty
 generateTokenInfoFromString r s = do
@@ -394,6 +438,7 @@ tokenHighlighting = merge . map tokenToCFile
 
   tokenToCFile :: T.Token -> CompressedFile
   tokenToCFile (T.TokSetN (i, _))               = aToF PrimitiveType (P.getRange i)
+  tokenToCFile (T.TokPropN (i, _))              = aToF PrimitiveType (P.getRange i)
   tokenToCFile (T.TokKeyword T.KwSet  i)        = aToF PrimitiveType (P.getRange i)
   tokenToCFile (T.TokKeyword T.KwProp i)        = aToF PrimitiveType (P.getRange i)
   tokenToCFile (T.TokKeyword T.KwForall i)      = aToF Symbol (P.getRange i)
@@ -407,12 +452,11 @@ tokenHighlighting = merge . map tokenToCFile
   tokenToCFile (T.TokLiteral (L.LitQName  r _)) = aToF String r
   tokenToCFile (T.TokLiteral (L.LitMeta r _ _)) = aToF String r
   tokenToCFile (T.TokComment (i, _))            = aToF Comment (P.getRange i)
-  tokenToCFile (T.TokTeX (i, _))                = aToF Comment (P.getRange i)
+  tokenToCFile (T.TokTeX (i, _))                = aToF Background (P.getRange i)
+  tokenToCFile (T.TokMarkup (i, _))             = aToF Markup (P.getRange i)
   tokenToCFile (T.TokId {})                     = mempty
   tokenToCFile (T.TokQId {})                    = mempty
-  tokenToCFile (T.TokString (i,s))
-    | "--" `isPrefixOf` s                       = aToF Option (P.getRange i)
-    | otherwise                                 = mempty
+  tokenToCFile (T.TokString (i,s))              = aToF Pragma (P.getRange i)
   tokenToCFile (T.TokDummy {})                  = mempty
   tokenToCFile (T.TokEOF {})                    = mempty
 
@@ -429,9 +473,9 @@ nameKinds :: Level
           -> A.Declaration
           -> TCM NameKinds
 nameKinds hlLevel decl = do
-  imported <- fix <$> use stImports
+  imported <- fix <$> useTC stImports
   local    <- case hlLevel of
-    Full{} -> fix <$> use stSignature
+    Full{} -> fix <$> useTC stSignature
     _      -> return HMap.empty
       -- Traverses the syntax tree and constructs a map from qualified
       -- names to name kinds. TODO: Handle open public.
@@ -455,6 +499,8 @@ nameKinds hlLevel decl = do
 
   defnToKind :: Defn -> NameKind
   defnToKind   M.Axiom{}                           = Postulate
+  defnToKind   M.DataOrRecSig{}                    = Postulate
+  defnToKind   M.GeneralizableVar{}                = Generalizable
   defnToKind d@M.Function{} | isProperProjection d = Field
                             | otherwise            = Function
   defnToKind   M.Datatype{}                        = Datatype
@@ -481,17 +527,18 @@ nameKinds hlLevel decl = do
   declToKind (A.ScopedDecl {})      = id
   declToKind (A.Open {})            = id
   declToKind (A.PatternSynDef q _ _) = insert q (Constructor Common.Inductive)
+  declToKind (A.Generalize _ _ _ q _) = insert q Generalizable
   declToKind (A.FunDef  _ q _ _)     = insert q Function
   declToKind (A.UnquoteDecl _ _ qs _) = foldr (\ q f -> insert q Function . f) id qs
   declToKind (A.UnquoteDef _ qs _)    = foldr (\ q f -> insert q Function . f) id qs
-  declToKind (A.DataSig _ q _ _)    = insert q Datatype
-  declToKind (A.DataDef _ q _ cs)   = \m ->
+  declToKind (A.DataSig _ q _ _)      = insert q Datatype
+  declToKind (A.DataDef _ q _ _ cs)   = \m ->
                                       insert q Datatype $
                                       foldr (\d -> insert (A.axiomName d)
                                                           (Constructor Common.Inductive))
                                             m cs
-  declToKind (A.RecSig _ q _ _)     = insert q Record
-  declToKind (A.RecDef _ q _ _ c _ _ _) = insert q Record .
+  declToKind (A.RecSig _ q _ _)       = insert q Record
+  declToKind (A.RecDef _ q _ _ _ c _ _ _) = insert q Record .
                                       case c of
                                         Nothing -> id
                                         Just q  -> insert q (Constructor Common.Inductive)
@@ -520,7 +567,7 @@ generateConstructorInfo modMap file kinds decl = do
         end   = fromIntegral $ P.posPos $ P.iEnd   $ last is
 
     -- Get all disambiguated names that fall within the range of decl.
-    m0 <- use stDisambiguatedNames
+    m0 <- useTC stDisambiguatedNames
     let (_, m1) = IntMap.split (pred start) m0
         (m2, _) = IntMap.split end m1
         constrs = IntMap.elems m2
@@ -531,7 +578,7 @@ generateConstructorInfo modMap file kinds decl = do
 
 printSyntaxInfo :: P.Range -> TCM ()
 printSyntaxInfo r = do
-  syntaxInfo <- use stSyntaxInfo
+  syntaxInfo <- useTC stSyntaxInfo
   ifTopLevelAndHighlightingLevelIs NonInteractive $
       printHighlightingInfo KeepHighlighting (selectC r syntaxInfo)
 
@@ -546,12 +593,6 @@ printErrorInfo e =
 --   Does something special for termination errors.
 
 errorHighlighting :: TCErr -> TCM File
-
-errorHighlighting (TypeError s cl@Closure{ clValue = TerminationCheckFailed termErrs }) =
-  -- For termination errors, we keep the previous highlighting,
-  -- just additionally mark the bad calls.
-  return $ terminationErrorHighlighting termErrs
-
 errorHighlighting e = do
 
   -- Erase previous highlighting.
@@ -561,7 +602,7 @@ errorHighlighting e = do
   -- Print new highlighting.
   s <- E.prettyError e
   let error = singleton (rToR r)
-         $ parserBased { otherAspects = [Error]
+         $ parserBased { otherAspects = Set.singleton Error
                        , note         = Just s
                        }
   return $ mconcat [ erase, error ]
@@ -572,17 +613,25 @@ warningHighlighting :: TCWarning -> File
 warningHighlighting w = case tcWarning w of
   TerminationIssue terrs     -> terminationErrorHighlighting terrs
   NotStrictlyPositive d ocs  -> positivityErrorHighlighting d ocs
-  UnreachableClauses{}       -> unreachableErrorHighlighting $ P.getRange w
+  UnreachableClauses{}       -> deadcodeHighlighting $ P.getRange w
   CoverageIssue{}            -> coverageErrorHighlighting $ P.getRange w
   CoverageNoExactSplit{}     -> catchallHighlighting $ P.getRange w
   UnsolvedConstraints cs     -> constraintsHighlighting cs
   UnsolvedMetaVariables rs   -> metasHighlighting rs
+  AbsurdPatternRequiresNoRHS{} -> deadcodeHighlighting $ P.getRange w
+  ModuleDoesntExport{}         -> deadcodeHighlighting $ P.getRange w
   -- expanded catch-all case to get a warning for new constructors
+  CantGeneralizeOverSorts{}  -> mempty
   UnsolvedInteractionMetas{} -> mempty
   OldBuiltin{}               -> mempty
-  EmptyRewritePragma{}       -> mempty
+  EmptyRewritePragma{}       -> deadcodeHighlighting $ P.getRange w
+  IllformedAsClause{}        -> deadcodeHighlighting $ P.getRange w
   UselessPublic{}            -> mempty
   UselessInline{}            -> mempty
+  WrongInstanceDeclaration{} -> mempty
+  InstanceWithExplicitArg{}  -> deadcodeHighlighting $ P.getRange w
+  InstanceNoOutputTypeName{} -> mempty
+  InstanceArgWithExplicitArg{} -> mempty
   ParseWarning{}             -> mempty
   InversionDepthReached{}    -> mempty
   GenericWarning{}           -> mempty
@@ -591,19 +640,52 @@ warningHighlighting w = case tcWarning w of
   SafeFlagPragma{}           -> mempty
   SafeFlagNonTerminating     -> mempty
   SafeFlagTerminating        -> mempty
-  SafeFlagPrimTrustMe        -> mempty
+  SafeFlagWithoutKFlagPrimEraseEquality -> mempty
+  WithoutKFlagPrimEraseEquality -> mempty
   SafeFlagNoPositivityCheck  -> mempty
   SafeFlagPolarity           -> mempty
+  SafeFlagNoUniverseCheck    -> mempty
   DeprecationWarning{}       -> mempty
-  NicifierIssue{}            -> mempty
   UserWarning{}              -> mempty
+  LibraryWarning{}           -> mempty
+  InfectiveImport{}          -> mempty
+  CoInfectiveImport{}        -> mempty
+  NicifierIssue w           -> case w of
+    -- we intentionally override the binding of `w` here so that our pattern of
+    -- using `P.getRange w` still yields the most precise range information we
+    -- can get.
+    NotAllowedInMutual{} -> deadcodeHighlighting $ P.getRange w
+    EmptyAbstract{}      -> deadcodeHighlighting $ P.getRange w
+    EmptyInstance{}      -> deadcodeHighlighting $ P.getRange w
+    EmptyMacro{}         -> deadcodeHighlighting $ P.getRange w
+    EmptyMutual{}        -> deadcodeHighlighting $ P.getRange w
+    EmptyPostulate{}     -> deadcodeHighlighting $ P.getRange w
+    EmptyPrivate{}       -> deadcodeHighlighting $ P.getRange w
+    EmptyGeneralize{}    -> deadcodeHighlighting $ P.getRange w
+    UselessAbstract{}    -> deadcodeHighlighting $ P.getRange w
+    UselessInstance{}    -> deadcodeHighlighting $ P.getRange w
+    UselessPrivate{}     -> deadcodeHighlighting $ P.getRange w
+    -- TODO: explore highlighting opportunities here!
+    EmptyPrimitive{} -> mempty
+    InvalidCatchallPragma{} -> mempty
+    InvalidNoPositivityCheckPragma{} -> mempty
+    InvalidNoUniverseCheckPragma{} -> mempty
+    InvalidTerminationCheckPragma{} -> mempty
+    MissingDefinitions{} -> mempty
+    PolarityPragmasButNotPostulates{} -> mempty
+    PragmaNoTerminationCheck{} -> mempty
+    PragmaCompiled{} -> mempty
+    UnknownFixityInMixfixDecl{} -> mempty
+    UnknownNamesInFixityDecl{} -> mempty
+    UnknownNamesInPolarityPragmas{} -> mempty
+
 
 -- | Generate syntax highlighting for termination errors.
 
 terminationErrorHighlighting :: [TerminationError] -> File
 terminationErrorHighlighting termErrs = functionDefs `mappend` callSites
   where
-    m            = parserBased { otherAspects = [TerminationProblem] }
+    m            = parserBased { otherAspects = Set.singleton TerminationProblem }
     functionDefs = Fold.foldMap (\x -> singleton (rToR $ bindingSite x) m) $
                    concatMap M.termErrFunctions termErrs
     callSites    = Fold.foldMap (\r -> singleton (rToR r) m) $
@@ -613,24 +695,25 @@ terminationErrorHighlighting termErrs = functionDefs `mappend` callSites
 -- definitions.
 
 -- TODO: highlight also the problematic occurrences
-positivityErrorHighlighting :: I.QName -> OccursWhere -> File
-positivityErrorHighlighting q o = several (rToR <$> P.getRange q : rs) m
+positivityErrorHighlighting :: I.QName -> Seq OccursWhere -> File
+positivityErrorHighlighting q os =
+  several (rToR <$> P.getRange q : rs) m
   where
-    rs = case o of Unknown -> []; Known r _ -> [r]
-    m  = parserBased { otherAspects = [PositivityProblem] }
+    rs = map (\(OccursWhere r _ _) -> r) (Fold.toList os)
+    m  = parserBased { otherAspects = Set.singleton PositivityProblem }
 
-unreachableErrorHighlighting :: P.Range -> File
-unreachableErrorHighlighting r = singleton (rToR $ P.continuousPerLine r) m
-  where m = parserBased { otherAspects = [ReachabilityProblem] }
+deadcodeHighlighting :: P.Range -> File
+deadcodeHighlighting r = singleton (rToR $ P.continuousPerLine r) m
+  where m = parserBased { otherAspects = Set.singleton Deadcode }
 
 coverageErrorHighlighting :: P.Range -> File
 coverageErrorHighlighting r = singleton (rToR $ P.continuousPerLine r) m
-  where m = parserBased { otherAspects = [CoverageProblem] }
+  where m = parserBased { otherAspects = Set.singleton CoverageProblem }
 
 
 catchallHighlighting :: P.Range -> File
 catchallHighlighting r = singleton (rToR $ P.continuousPerLine r) m
-  where m = parserBased { otherAspects = [CatchallClause] }
+  where m = parserBased { otherAspects = Set.singleton CatchallClause }
 
 
 -- | Generates and prints syntax highlighting information for unsolved
@@ -662,7 +745,7 @@ computeUnsolvedMetaWarnings = do
 
 metasHighlighting :: [P.Range] -> File
 metasHighlighting rs = several (map (rToR . P.continuousPerLine) rs)
-                     $ parserBased { otherAspects = [UnsolvedMeta] }
+                     $ parserBased { otherAspects = Set.singleton UnsolvedMeta }
 
 -- | Generates syntax highlighting information for unsolved constraints
 --   (ideally: that are not connected to a meta variable).
@@ -673,7 +756,7 @@ computeUnsolvedConstraints = constraintsHighlighting <$> getAllConstraints
 constraintsHighlighting :: Constraints -> File
 constraintsHighlighting cs =
   several (map (rToR . P.continuousPerLine) rs)
-          (parserBased { otherAspects = [UnsolvedConstraint] })
+          (parserBased { otherAspects = Set.singleton UnsolvedConstraint })
   where
   -- get ranges of interesting unsolved constraints
   rs = (`mapMaybe` (map theConstraint cs)) $ \case
@@ -746,7 +829,7 @@ nameToFile modMap file xs x fr m mR =
     mempty
   where
   aspects    = m $ C.isOperator x
-  fileNames  = catMaybes $ map (fmap P.srcFile . P.rStart . P.getRange) (x : xs)
+  fileNames  = mapMaybe (fmap P.srcFile . P.rStart . P.getRange) (x : xs)
   frFile     = singleton (rToR fr) (aspects { definitionSite = notHere <$> mFilePos })
   rs         = map P.getRange (x : xs)
 
@@ -787,6 +870,7 @@ nameToFile modMap file xs x fr m mR =
   isLocal :: NameKind -> Bool
   isLocal = \case
     Bound         -> True
+    Generalizable -> True
     Argument      -> True
     Constructor{} -> False
     Datatype      -> False
@@ -798,7 +882,6 @@ nameToFile modMap file xs x fr m mR =
     Record        -> False
     Macro         -> False
 
-
 -- | A variant of 'nameToFile' for qualified abstract names.
 
 nameToFileA
@@ -823,6 +906,7 @@ nameToFileA modMap file x include m =
              r
              m
              (if include then Just $ bindingSite x else Nothing)
+    `mappend` notationFile
   where
     -- Andreas, 2016-09-08, for issue #2140:
     -- Range of name from fixity declaration:
@@ -831,6 +915,13 @@ nameToFileA modMap file x include m =
     -- (I do not understand how we get them as they should not be serialized...)
     r = if P.rangeFile fr == Strict.Just file then fr else P.noRange
 
+    notationFile = mconcat $ map genPartFile $ theNotation $ A.nameFixity $ A.qnameName x
+    boundAspect = parserBased{ aspect = Just $ Name (Just Bound) False }
+    genPartFile (BindHole r i)   = several [rToR r, rToR $ getRange i] boundAspect
+    genPartFile (NormalHole r i) = several [rToR r, rToR $ getRange i] boundAspect
+    genPartFile WildHole{}       = mempty
+    genPartFile (IdPart x)       = singleton (rToR $ P.getRange x) (m False)
+
 concreteBase :: I.QName -> C.Name
 concreteBase = A.nameConcrete . A.qnameName
 
@@ -844,6 +935,15 @@ bindingSite = A.nameBindingSite . A.qnameName
 --   To be used later during syntax highlighting.
 storeDisambiguatedName :: A.QName -> TCM ()
 storeDisambiguatedName q = whenJust (start $ P.getRange q) $ \ i ->
-  stDisambiguatedNames %= IntMap.insert i q
+  stDisambiguatedNames `modifyTCLens` IntMap.insert i q
   where
   start r = fromIntegral . P.posPos <$> P.rStart' r
+
+-- | Store a disambiguation of record field tags for the purpose of highlighting.
+disambiguateRecordFields
+  :: [C.Name]   -- ^ Record field names in a record expression.
+  -> [A.QName]  -- ^ Record field names in the corresponding record type definition
+  -> TCM ()
+disambiguateRecordFields cxs axs = forM_ cxs $ \ cx -> do
+  caseMaybe (List.find ((cx ==) . A.nameConcrete . A.qnameName) axs) (return ()) $ \ ax -> do
+    storeDisambiguatedName ax { A.qnameName = (A.qnameName ax) { A.nameConcrete = cx } }
