@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP             #-}
+{-# LANGUAGE NondecreasingIndentation #-}
 
 module Agda.Interaction.MakeCase where
 
@@ -6,6 +7,7 @@ import Prelude hiding (mapM, mapM_, null)
 
 import Control.Applicative hiding (empty)
 import Control.Monad hiding (mapM, mapM_, forM)
+import Control.Monad.Reader (asks)
 
 import qualified Data.Map as Map
 import qualified Data.List as List
@@ -14,20 +16,22 @@ import Data.Traversable
 
 import Agda.Syntax.Common
 import Agda.Syntax.Position
+import Agda.Syntax.Concrete (NameInScope(..), LensInScope(..))
 import qualified Agda.Syntax.Concrete as C
 import qualified Agda.Syntax.Abstract as A
 import qualified Agda.Syntax.Abstract.Views as A
 import qualified Agda.Syntax.Info as A
 import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Pattern
-import Agda.Syntax.Scope.Base  ( ResolvedName(..), Binder(..) )
-import Agda.Syntax.Scope.Monad ( resolveName )
+import Agda.Syntax.Scope.Base  ( ResolvedName(..), Binder(..), KindOfName(..), allKindsOfNames )
+import Agda.Syntax.Scope.Monad ( resolveName, resolveName' )
 import Agda.Syntax.Translation.ConcreteToAbstract
 import Agda.Syntax.Translation.InternalToAbstract
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Coverage
 import Agda.TypeChecking.Coverage.Match ( SplitPatVar(..) , SplitPattern , applySplitPSubst , fromSplitPatterns )
+import Agda.TypeChecking.Empty ( isEmptyTel )
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.RecordPatterns
 import Agda.TypeChecking.Reduce
@@ -64,7 +68,8 @@ parseVariables
   -> InteractionId   -- ^ The hole of this function we are working on.
   -> Range           -- ^ The range of this hole.
   -> [String]        -- ^ The words the user entered in this hole (variable names).
-  -> TCM [Int]       -- ^ The computed de Bruijn indices of the variables to split on.
+  -> TCM [(Int,NameInScope)] -- ^ The computed de Bruijn indices of the variables to split on,
+                             --   with information about whether each variable is in scope.
 parseVariables f tel ii rng ss = do
 
   -- Get into the context of the meta.
@@ -83,16 +88,19 @@ parseVariables f tel ii rng ss = do
     let nlocals = n - size tel
     unless (nlocals >= 0) __IMPOSSIBLE__
 
+    fv <- getDefFreeVars f
     reportSDoc "interaction.case" 20 $ do
       m   <- currentModule
       tel <- lookupSection m
-      fv  <- getDefFreeVars f
+      cxt <- getContextTelescope
       vcat
-       [ text "parseVariables:"
-       , text "current module  =" <+> prettyTCM m
-       , text "current section =" <+> inTopContext (prettyTCM tel)
+       [ "parseVariables:"
+       , "current module  =" <+> prettyTCM m
+       , "current section =" <+> inTopContext (prettyTCM tel)
        , text $ "function's fvs  = " ++ show fv
        , text $ "number of locals= " ++ show nlocals
+       , "context         =" <+> do inTopContext $ prettyTCM cxt
+       , "checkpoints     =" <+> do (text . show) =<< asksTC envCheckpoints
        ]
 
     -- Resolve each string to a variable.
@@ -105,14 +113,21 @@ parseVariables f tel ii rng ss = do
           failModuleBound = typeError $ GenericError $
             "Cannot split on module parameter " ++ s
           failLetBound v  = typeError . GenericError $
-            "cannot split on let-bound variable " ++ s
+            "Cannot split on let-bound variable " ++ s
           failInstantiatedVar v = typeError . GenericDocError =<< sep
               [ text $ "Cannot split on variable " ++ s ++ ", because it is bound to"
               , prettyTCM v
               ]
+          failCaseLet     = typeError $ GenericError $
+            "Cannot split on variable " ++ s ++
+            ", because let-declarations may not be defined by pattern-matching"
 
+      -- Jesper, 2018-12-19: Don't consider generalizable names since
+      -- they can be shadowed by hidden variables.
+      let kinds = List.delete GeneralizeName allKindsOfNames
+          cname = C.QName $ C.Name r C.InScope $ C.stringNameParts s
       -- Note: the range in the concrete name is only approximate.
-      resName <- resolveName $ C.QName $ C.Name r $ C.stringNameParts s
+      resName <- resolveName' kinds Nothing cname
       case resName of
 
         -- Fail if s is a name, but not of a variable.
@@ -131,8 +146,9 @@ parseVariables f tel ii rng ss = do
             -- case the instantiation could as well be the other way
             -- around, so the new clauses will still make sense.
             (Var i [] , PatternBound) -> do
-              when (i < nlocals) __IMPOSSIBLE__
-              return $ i - nlocals
+              reportSLn "interaction.case" 30 $ "resolved variable " ++ show x ++ " = " ++ show i
+              when (i < nlocals) failCaseLet
+              return (i - nlocals , C.InScope)
             (Var i [] , LambdaBound)
               | i < nlocals -> failLocal
               | otherwise   -> failModuleBound
@@ -142,59 +158,73 @@ parseVariables f tel ii rng ss = do
         -- If s is not a name, compare it to the printed variable representation.
         -- This fallback is to enable splitting on hidden variables.
         UnknownName -> do
-          case filter ((s ==) . fst) xs of
-            []                    -> failUnbound
-            [(_,i)] | i < nlocals -> failLocal
-                    | otherwise   -> return $ i - nlocals
+          let xs' = filter ((s ==) . fst) xs
+          when (null xs') $ failUnbound
+          reportSLn "interaction.case" 20 $ "matching names corresponding to indices " ++ show xs'
+          -- Andreas, 2018-05-28, issue #3095
+          -- We want to act on an ambiguous name if it corresponds to only one local index.
+          let xs'' = mapMaybe (\ (_,i) -> if i < nlocals then Nothing else Just $ i - nlocals) xs'
+          when (null xs'') $ failLocal
+          -- Filter out variable bound by parent function or module.
+          let xs''' = mapMaybe (\ i -> if i < fv then Nothing else Just i) xs''
+          case xs''' of
+            []  -> failModuleBound
+            [i] -> return (i , C.NotInScope)
             -- Issue 1325: Variable names in context can be ambiguous.
-            _                     -> failAmbiguous
+            _   -> failAmbiguous
 
 -- | Lookup the clause for an interaction point in the signature.
---   Returns the CaseContext, the clause itself, and a list of previous clauses
+--   Returns the CaseContext, the previous clauses, the clause itself,
+--   and a list of the remaining ones.
 
--- Andreas, 2016-06-08, issue #289 and #2006.
--- This replace the old findClause hack (shutter with disgust).
-getClauseForIP :: QName -> Int -> TCM (CaseContext, Clause, [Clause])
-getClauseForIP f clauseNo = do
+type ClauseZipper =
+   ( [Clause] -- previous clauses
+   , Clause   -- clause of interest
+   , [Clause] -- other clauses
+   )
+
+getClauseZipperForIP :: QName -> Int -> TCM (CaseContext, ClauseZipper)
+getClauseZipperForIP f clauseNo = do
   (theDef <$> getConstInfo f) >>= \case
     Function{funClauses = cs, funExtLam = extlam} -> do
-      let (cs1,cs2) = fromMaybe __IMPOSSIBLE__ $ splitExactlyAt clauseNo cs
-          c         = fromMaybe __IMPOSSIBLE__ $ headMaybe cs2
-      return (extlam, c, cs1)
+      let (cs1,ccs2) = fromMaybe __IMPOSSIBLE__ $ splitExactlyAt clauseNo cs
+          (c,cs2)    = fromMaybe __IMPOSSIBLE__ $ uncons ccs2
+      return (extlam, (cs1, c, cs2))
     d -> do
       reportSDoc "impossible" 10 $ vcat
-        [ text "getClauseForIP" <+> prettyTCM f <+> text (show clauseNo)
-          <+> text "received"
+        [ "getClauseZipperForIP" <+> prettyTCM f <+> text (show clauseNo)
+          <+> "received"
         , text (show d)
         ]
       __IMPOSSIBLE__
 
-
 -- | Entry point for case splitting tactic.
 
 makeCase :: InteractionId -> Range -> String -> TCM (QName, CaseContext, [A.Clause])
 makeCase hole rng s = withInteractionId hole $ do
 
-  -- Get function clause which contains the interaction point.
+  -- Jesper, 2018-12-10: print unsolved metas in dot patterns as _
+  localTC (\ e -> e { envPrintMetasBare = True }) $ do
 
+  -- Get function clause which contains the interaction point.
   InteractionPoint { ipMeta = mm, ipClause = ipCl} <- lookupInteractionPoint hole
   let meta = fromMaybe __IMPOSSIBLE__ mm
   (f, clauseNo, rhs) <- case ipCl of
-    IPClause f clauseNo rhs-> return (f, clauseNo, rhs)
+    IPClause f clauseNo rhs -> return (f, clauseNo, rhs)
     IPNoClause -> typeError $ GenericError $
       "Cannot split here, as we are not in a function definition"
-  (casectxt, clause, prevClauses) <- getClauseForIP f clauseNo
+  (casectxt, (prevClauses, clause, follClauses)) <- getClauseZipperForIP f clauseNo
   let perm = fromMaybe __IMPOSSIBLE__ $ clausePerm clause
       tel  = clauseTel  clause
       ps   = namedClausePats clause
   reportSDoc "interaction.case" 10 $ vcat
-    [ text "splitting clause:"
+    [ "splitting clause:"
     , nest 2 $ vcat
-      [ text "f       =" <+> prettyTCM f
-      , text "context =" <+> ((inTopContext . prettyTCM) =<< getContextTelescope)
-      , text "tel     =" <+> (inTopContext . prettyTCM) tel
-      , text "perm    =" <+> text (show perm)
-      , text "ps      =" <+> text (show ps)
+      [ "f       =" <+> prettyTCM f
+      , "context =" <+> ((inTopContext . prettyTCM) =<< getContextTelescope)
+      , "tel     =" <+> (inTopContext . prettyTCM) tel
+      , "perm    =" <+> text (show perm)
+      , "ps      =" <+> prettyTCMPatternList ps
       ]
     ]
 
@@ -244,8 +274,7 @@ makeCase hole rng s = withInteractionId hole $ do
               isVarP _ = Nothing
           -- If all are already coming from the user, there is really nothing todo!
           when (all ((UserWritten ==) . getOrigin) trailingPatVars) $ do
-            typeError . GenericDocError =<< do
-              text "Cannot split on result here, because" <+> prettyTCM err
+            typeError $ SplitError err
           -- Otherwise, we make these user-written
           let xs = map (dbPatVarIndex . namedArg) trailingPatVars
           return [makePatternVarsVisible xs sc]
@@ -254,25 +283,45 @@ makeCase hole rng s = withInteractionId hole $ do
           -- Andreas, 2016-05-03: do not introduce function arguments after projection.
           -- This is sometimes annoying and can anyway be done by another C-c C-c.
           -- mapM (snd <.> fixTarget) $ splitClauses cov
-          return $ splitClauses cov
+          return cov
     checkClauseIsClean ipCl
     (f, casectxt,) <$> mapM (makeAbstractClause f rhs) scs
   else do
     -- split on variables
     xs <- parseVariables f tel hole rng vars
+    reportSLn "interaction.case" 30 $ "parsedVariables: " ++ show (zip xs vars)
     -- Variables that are not in scope yet are brought into scope (@toShow@)
     -- The other variables are split on (@toSplit@).
-    let (toShow, toSplit) = flip mapEither (zip xs vars) $ \ (x, s) ->
-          if take 1 s == "." then Left x else Right x
+    let (toShow, toSplit) = flip mapEither (zip xs vars) $ \ ((x,nis), s) ->
+          if (nis == C.NotInScope) then Left x else Right x
     let sc = makePatternVarsVisible toShow $ clauseToSplitClause clause
     scs <- split f toSplit sc
-    -- filter out clauses that are already covered
+    reportSLn "interaction.case" 70 $ "makeCase: survived the splitting"
+
+    -- CLEAN UP OF THE GENERATED CLAUSES
+    -- 1. filter out clauses that are already covered
     scs <- filterM (not <.> isCovered f prevClauses . fst) scs
-    cs <- forM scs $ \(sc, isAbsurd) -> do
-            if isAbsurd then makeAbsurdClause f sc else makeAbstractClause f rhs sc
+    reportSLn "interaction.case" 70 $ "makeCase: survived filtering out already covered clauses"
+    -- 2. filter out trivially impossible clauses not asked for by the user
+    cs <- catMaybes <$> do
+     forM scs $ \ (sc, isAbsurd) -> if isAbsurd
+      -- absurd clause coming from a split asked for by the user
+      then Just <$> makeAbsurdClause f sc
+      -- trivially empty clause due to the refined patterns
+      else
+        ifM (liftTCM $ isEmptyTel (scTel sc))
+          {- then -} (pure Nothing)
+          {- else -} (Just <$> makeAbstractClause f rhs sc)
+    reportSLn "interaction.case" 70 $ "makeCase: survived filtering out impossible clauses"
+    -- 3. If the cleanup removed everything then we know that none of the clauses where
+    --    absurd but that all of them were trivially empty. In this case we rewind and
+    --    insert all the clauses (garbage in, garbage out!)
+    cs <- if not (null cs) then pure cs
+          else mapM (makeAbstractClause f rhs . fst) scs
+
     reportSDoc "interaction.case" 65 $ vcat
-      [ text "split result:"
-      , nest 2 $ vcat $ map (text . show . A.deepUnscope) cs
+      [ "split result:"
+      , nest 2 $ vcat $ map prettyA cs
       ]
     checkClauseIsClean ipCl
     return (f, casectxt, cs)
@@ -304,7 +353,7 @@ makeCase hole rng s = withInteractionId hole $ do
   -- In this case, we refuse to split, as this might lose the refinements.
   checkClauseIsClean :: IPClause -> TCM ()
   checkClauseIsClean ipCl = do
-    sips <- filter ipSolved . Map.elems <$> use stInteractionPoints
+    sips <- filter ipSolved . Map.elems <$> useTC stInteractionPoints
     when (List.any ((== ipCl) . ipClause) sips) $
       typeError $ GenericError $ "Cannot split as clause rhs has been refined.  Please reload"
 
@@ -330,11 +379,11 @@ makeAbsurdClause :: QName -> SplitClause -> TCM A.Clause
 makeAbsurdClause f (SClause tel sps _ _ t) = do
   let ps = fromSplitPatterns sps
   reportSDoc "interaction.case" 10 $ vcat
-    [ text "Interaction.MakeCase.makeAbsurdClause: split clause:"
+    [ "Interaction.MakeCase.makeAbsurdClause: split clause:"
     , nest 2 $ vcat
-      [ text "context =" <+> do (inTopContext . prettyTCM) =<< getContextTelescope
-      , text "tel     =" <+> do inTopContext $ prettyTCM tel
-      , text "ps      =" <+> do inTopContext $ addContext tel $ prettyTCMPatternList ps -- P.sep <$> prettyTCMPatterns ps
+      [ "context =" <+> do (inTopContext . prettyTCM) =<< getContextTelescope
+      , "tel     =" <+> do inTopContext $ prettyTCM tel
+      , "ps      =" <+> do inTopContext $ addContext tel $ prettyTCMPatternList ps -- P.sep <$> prettyTCMPatterns ps
       ]
     ]
   withCurrentModule (qnameModule f) $ do
@@ -345,7 +394,7 @@ makeAbsurdClause f (SClause tel sps _ _ t) = do
     let c = Clause noRange noRange tel ps Nothing t False Nothing
     -- Normalise the dot patterns
     ps <- addContext tel $ normalise $ namedClausePats c
-    reportSDoc "interaction.case" 60 $ text "normalized patterns: " <+> text (show ps)
+    reportSDoc "interaction.case" 60 $ "normalized patterns: " <+> prettyTCMPatternList ps
     inTopContext $ reify $ QNamed f $ c { namedClausePats = ps }
 
 
@@ -355,7 +404,7 @@ makeAbstractClause :: QName -> A.RHS -> SplitClause -> TCM A.Clause
 makeAbstractClause f rhs cl = do
 
   lhs <- A.clauseLHS <$> makeAbsurdClause f cl
-  reportSDoc "interaction.case" 60 $ text "reified lhs: " <+> text (show lhs)
+  reportSDoc "interaction.case" 60 $ "reified lhs: " <+> prettyA lhs
   return $ A.Clause lhs [] rhs A.noWhereDecls False
   -- let ii = InteractionId (-1)  -- Dummy interaction point since we never type check this.
   --                              -- Can end up in verbose output though (#1842), hence not __IMPOSSIBLE__.
