@@ -1,17 +1,18 @@
-{-# LANGUAGE CPP #-}
+{-# LANGUAGE CPP       #-}
+{-# LANGUAGE DataKinds #-}
 
 module Agda.Interaction.Options
     ( CommandLineOptions(..)
-    , IgnoreFlags(..)
     , PragmaOptions(..)
     , OptionsPragma
     , Flag, OptM, runOptM, OptDescr(..), ArgDescr(..)
     , Verbosity
+    , HtmlHighlight(..)
     , WarningMode(..)
     , checkOpts
-    , parseStandardOptions, parseStandardOptions'
     , parsePragmaOptions
     , parsePluginOptions
+    , stripRTS
     , defaultOptions
     , defaultInteractionOptions
     , defaultVerbosity
@@ -19,13 +20,16 @@ module Agda.Interaction.Options
     , defaultPragmaOptions
     , standardOptions_
     , unsafePragmaOptions
-    , isLiterate
+    , restartOptions
+    , infectiveOptions
+    , coinfectiveOptions
+    , safeFlag
     , mapFlag
     , usage
     , defaultLibDir
     -- Reused by PandocAgda
     , inputFlag
-    , standardOptions
+    , standardOptions, deadStandardOptions
     , getOptSimple
     ) where
 
@@ -34,6 +38,7 @@ import Control.Monad.Trans
 
 import Data.IORef
 import Data.Either
+import Data.Function
 import Data.Maybe
 import Data.List                ( isSuffixOf , intercalate )
 import Data.Set                 ( Set )
@@ -60,12 +65,14 @@ import Agda.Utils.Except
   )
 
 import Agda.Utils.FileName      ( absolute, AbsolutePath, filePath )
-import Agda.Utils.Monad         ( ifM, readM )
+import Agda.Utils.Functor       ( (<&>) )
+import Agda.Utils.Lens          ( Lens', over )
 import Agda.Utils.List          ( groupOn, wordsBy )
+import Agda.Utils.Monad         ( ifM, readM )
 import Agda.Utils.String        ( indent )
 import Agda.Utils.Trie          ( Trie )
-import Agda.Syntax.Parser.Literate ( literateExts )
 import qualified Agda.Utils.Trie as Trie
+import Agda.Utils.WithDefault
 
 import Agda.Version
 -- Paths_Agda.hs is in $(BUILD_DIR)/build/autogen/.
@@ -73,18 +80,12 @@ import Paths_Agda ( getDataFileName )
 
 import qualified System.IO.Unsafe as UNSAFE (unsafePerformIO)
 
--- | This should probably go somewhere else.
-isLiterate :: FilePath -> Bool
-isLiterate file = any (`isSuffixOf` file) literateExts
-
 -- OptDescr is a Functor --------------------------------------------------
 
 type Verbosity = Trie String Int
 
--- ignore or respect the flags --allow-unsolved-metas,
--- --no-termination-check, --no-positivity-check?
-data IgnoreFlags = IgnoreFlags | RespectFlags
-  deriving Eq
+data HtmlHighlight = HighlightAll | HighlightCode | HighlightAuto
+  deriving (Show, Eq)
 
 -- Don't forget to update
 --   doc/user-manual/tools/command-line-options.rst
@@ -106,17 +107,20 @@ data CommandLineOptions = Options
   , optShowHelp         :: Maybe Help
   , optInteractive      :: Bool
   , optGHCiInteraction  :: Bool
+  , optJSONInteraction  :: Bool
   , optOptimSmashing    :: Bool
   , optCompileDir       :: Maybe FilePath
   -- ^ In the absence of a path the project root is used.
   , optGenerateVimFile  :: Bool
   , optGenerateLaTeX    :: Bool
   , optGenerateHTML     :: Bool
+  , optHTMLHighlight    :: HtmlHighlight
   , optDependencyGraph  :: Maybe FilePath
   , optLaTeXDir         :: FilePath
   , optHTMLDir          :: FilePath
   , optCSSFile          :: Maybe FilePath
   , optIgnoreInterfaces :: Bool
+  , optIgnoreAllInterfaces :: Bool
   , optForcing          :: Bool
   , optPragmaOptions    :: PragmaOptions
   , optOnlyScopeChecking :: Bool
@@ -132,7 +136,7 @@ data PragmaOptions = PragmaOptions
   , optShowIrrelevant            :: Bool
   , optUseUnicode                :: Bool
   , optVerbose                   :: Verbosity
-  , optProofIrrelevance          :: Bool
+  , optProp                      :: Bool
   , optAllowUnsolved             :: Bool
   , optDisablePositivity         :: Bool
   , optTerminationCheck          :: Bool
@@ -140,24 +144,29 @@ data PragmaOptions = PragmaOptions
     -- ^ Cut off structural order comparison at some depth in termination checker?
   , optCompletenessCheck         :: Bool
   , optUniverseCheck             :: Bool
-  , optSizedTypes                :: Bool
+  , optOmegaInOmega              :: Bool
+  , optSizedTypes                :: WithDefault 'True
+  , optGuardedness               :: WithDefault 'True
   , optInjectiveTypeConstructors :: Bool
-  , optGuardingTypeConstructors  :: Bool
   , optUniversePolymorphism      :: Bool
   , optIrrelevantProjections     :: Bool
   , optExperimentalIrrelevance   :: Bool  -- ^ irrelevant levels, irrelevant data matching
-  , optWithoutK                  :: Bool
+  , optWithoutK                  :: WithDefault 'False
   , optCopatterns                :: Bool  -- ^ Allow definitions by copattern matching?
   , optPatternMatching           :: Bool  -- ^ Is pattern matching allowed in the current file?
   , optExactSplit                :: Bool
   , optEta                       :: Bool
   , optRewriting                 :: Bool  -- ^ Can rewrite rules be added and used?
+  , optCubical                   :: Bool
   , optPostfixProjections        :: Bool
       -- ^ Should system generated projections 'ProjSystem' be printed
       --   postfix (True) or prefix (False).
   , optInstanceSearchDepth       :: Int
+  , optOverlappingInstances      :: Bool
   , optInversionMaxDepth         :: Int
   , optSafe                      :: Bool
+  , optDoubleCheck               :: Bool
+  , optSyntacticEquality         :: Bool  -- ^ Should conversion checker use syntactic equality shortcut?
   , optWarningMode               :: WarningMode
   , optCompileNoMain             :: Bool
   , optCaching                   :: Bool
@@ -168,6 +177,8 @@ data PragmaOptions = PragmaOptions
     -- ^ Automatic compile-time inlining for simple definitions (unless marked
     --   NOINLINE).
   , optPrintPatternSynonyms      :: Bool
+  , optFastReduce                :: Bool
+    -- ^ Use the Agda abstract machine (fastReduce)?
   }
   deriving (Show, Eq)
 
@@ -203,16 +214,19 @@ defaultOptions = Options
   , optShowHelp         = Nothing
   , optInteractive      = False
   , optGHCiInteraction  = False
+  , optJSONInteraction  = False
   , optOptimSmashing    = True
   , optCompileDir       = Nothing
   , optGenerateVimFile  = False
   , optGenerateLaTeX    = False
   , optGenerateHTML     = False
+  , optHTMLHighlight    = HighlightAll
   , optDependencyGraph  = Nothing
   , optLaTeXDir         = defaultLaTeXDir
   , optHTMLDir          = defaultHTMLDir
   , optCSSFile          = Nothing
   , optIgnoreInterfaces = False
+  , optIgnoreAllInterfaces = False
   , optForcing          = True
   , optPragmaOptions    = defaultPragmaOptions
   , optOnlyScopeChecking = False
@@ -224,35 +238,41 @@ defaultPragmaOptions = PragmaOptions
   , optShowIrrelevant            = False
   , optUseUnicode                = True
   , optVerbose                   = defaultVerbosity
-  , optProofIrrelevance          = False
+  , optProp                      = False
   , optExperimentalIrrelevance   = False
-  , optIrrelevantProjections     = True
+  , optIrrelevantProjections     = False -- off by default in > 2.5.4, see issue #2170
   , optAllowUnsolved             = False
   , optDisablePositivity         = False
   , optTerminationCheck          = True
   , optTerminationDepth          = defaultCutOff
   , optCompletenessCheck         = True
   , optUniverseCheck             = True
-  , optSizedTypes                = True
+  , optOmegaInOmega              = False
+  , optSizedTypes                = Default
+  , optGuardedness               = Default
   , optInjectiveTypeConstructors = False
-  , optGuardingTypeConstructors  = False
   , optUniversePolymorphism      = True
-  , optWithoutK                  = False
+  , optWithoutK                  = Default
   , optCopatterns                = True
   , optPatternMatching           = True
   , optExactSplit                = False
   , optEta                       = True
   , optRewriting                 = False
+  , optCubical                   = False
   , optPostfixProjections        = False
   , optInstanceSearchDepth       = 500
+  , optOverlappingInstances      = False
   , optInversionMaxDepth         = 50
   , optSafe                      = False
+  , optDoubleCheck               = False
+  , optSyntacticEquality         = True
   , optWarningMode               = defaultWarningMode
   , optCompileNoMain             = False
   , optCaching                   = True
   , optCountClusters             = False
   , optAutoInline                = True
   , optPrintPatternSynonyms      = True
+  , optFastReduce                = True
   }
 
 -- | The default termination depth.
@@ -285,13 +305,15 @@ type Flag opts = opts -> OptM opts
 
 checkOpts :: Flag CommandLineOptions
 checkOpts opts
-  | not (matches [optGHCiInteraction, isJust . optInputFile] <= 1) =
-      throwError "Choose at most one: input file or --interaction.\n"
+  | htmlRelated = throwError htmlRelatedMessage
+  | not (matches [optGHCiInteraction, optJSONInteraction, isJust . optInputFile] <= 1) =
+      throwError "Choose at most one: input file, --interactive, or --interaction-json.\n"
   | or [ p opts && matches ps > 1 | (p, ps) <- exclusive ] =
       throwError exclusiveMessage
   | otherwise = return opts
   where
   matches = length . filter ($ opts)
+  optionChanged opt = ((/=) `on` opt) opts defaultOptions
 
   atMostOne =
     [ optGenerateHTML
@@ -301,8 +323,7 @@ checkOpts opts
 
   exclusive =
     [ ( optOnlyScopeChecking
-      , optSafe . optPragmaOptions :
-        optGenerateVimFile :
+      , optGenerateVimFile :
         atMostOne
       )
     , ( optInteractive
@@ -311,18 +332,32 @@ checkOpts opts
     , ( optGHCiInteraction
       , optGenerateLaTeX : atMostOne
       )
+    , ( optJSONInteraction
+      , optGenerateLaTeX : atMostOne
+      )
     ]
 
   exclusiveMessage = unlines $
-    [ "The options --interactive, --interaction and"
+    [ "The options --interactive, --interaction, --interaction-json and"
     , "--only-scope-checking cannot be combined with each other or"
     , "with --html or --dependency-graph. Furthermore"
     , "--interactive and --interaction cannot be combined with"
     , "--latex, and --only-scope-checking cannot be combined with"
-    , "--safe or --vim."
+    , "--vim."
     ]
 
--- | Check for unsafe pramas. Gives a list of used unsafe flags.
+  htmlRelated = not (optGenerateHTML opts) &&
+    (  optionChanged optHTMLDir
+    || optionChanged optHTMLHighlight
+    || optionChanged optCSSFile
+    )
+
+  htmlRelatedMessage = unlines $
+    [ "The options --html-highlight, --css-dir and --html-dir"
+    , "only be used along with --html flag."
+    ]
+
+-- | Check for unsafe pragmas. Gives a list of used unsafe flags.
 
 unsafePragmaOptions :: PragmaOptions -> [String]
 unsafePragmaOptions opts =
@@ -330,13 +365,79 @@ unsafePragmaOptions opts =
   [ "--no-positivity-check"                      | optDisablePositivity opts         ] ++
   [ "--no-termination-check"                     | not (optTerminationCheck opts)    ] ++
   [ "--type-in-type"                             | not (optUniverseCheck opts)       ] ++
-  -- [ "--sized-types"                              | optSizedTypes opts                ] ++
+  [ "--omega-in-omega"                           | optOmegaInOmega opts              ] ++
+-- [ "--sized-types"                              | optSizedTypes opts                ] ++
+  [ "--sized-types and --guardedness"            | collapseDefault (optSizedTypes opts)
+                                                 , collapseDefault (optGuardedness opts) ] ++
   [ "--injective-type-constructors"              | optInjectiveTypeConstructors opts ] ++
-  [ "--guardedness-preserving-type-constructors" | optGuardingTypeConstructors opts  ] ++
+  [ "--irrelevant-projections"                   | optIrrelevantProjections opts     ] ++
   [ "--experimental-irrelevance"                 | optExperimentalIrrelevance opts   ] ++
   [ "--rewriting"                                | optRewriting opts                 ] ++
+  [ "--cubical and --with-K"                     | optCubical opts
+                                                 , not (collapseDefault $ optWithoutK opts) ] ++
   []
 
+-- | If any these options have changed, then the file will be
+--   rechecked. Boolean options are negated to mention non-default
+--   options, where possible.
+
+restartOptions :: [(PragmaOptions -> RestartCodomain, String)]
+restartOptions =
+  [ (C . optTerminationDepth, "--termination-depth")
+  , (B . not . optUseUnicode, "--no-unicode")
+  , (B . optAllowUnsolved, "--allow-unsolved-metas")
+  , (B . optDisablePositivity, "--no-positivity-check")
+  , (B . optTerminationCheck,  "--no-termination-check")
+  , (B . not . optUniverseCheck, "--type-in-type")
+  , (B . optOmegaInOmega, "--omega-in-omega")
+  , (B . not . collapseDefault . optSizedTypes, "--no-sized-types")
+  , (B . not . collapseDefault . optGuardedness, "--no-guardedness")
+  , (B . optInjectiveTypeConstructors, "--injective-type-constructors")
+  , (B . optProp, "--prop")
+  , (B . not . optUniversePolymorphism, "--no-universe-polymorphism")
+  , (B . optIrrelevantProjections, "--irrelevant-projections")
+  , (B . optExperimentalIrrelevance, "--experimental-irrelevance")
+  , (B . collapseDefault . optWithoutK, "--without-K")
+  , (B . optExactSplit, "--exact-split")
+  , (B . not . optEta, "--no-eta-equality")
+  , (B . optRewriting, "--rewriting")
+  , (B . optCubical, "--cubical")
+  , (B . optOverlappingInstances, "--overlapping-instances")
+  , (B . optSafe, "--safe")
+  , (B . optDoubleCheck, "--double-check")
+  , (B . not . optSyntacticEquality, "--no-syntactic-equality")
+  , (B . not . optAutoInline, "--no-auto-inline")
+  , (B . not . optFastReduce, "--no-fast-reduce")
+  , (I . optInstanceSearchDepth, "--instance-search-depth")
+  , (I . optInversionMaxDepth, "--inversion-max-depth")
+  , (W . optWarningMode, "--warning")
+  ]
+
+-- to make all restart options have the same type
+data RestartCodomain = C CutOff | B Bool | I Int | W WarningMode
+  deriving Eq
+
+-- | An infective option is an option that if used in one module, must
+--   be used in all modules that depend on this module.
+
+infectiveOptions :: [(PragmaOptions -> Bool, String)]
+infectiveOptions =
+  [ (optCubical, "--cubical")
+  , (optProp, "--prop")
+  ]
+
+-- | A coinfective option is an option that if used in one module, must
+--   be used in all modules that this module depends on.
+
+coinfectiveOptions :: [(PragmaOptions -> Bool, String)]
+coinfectiveOptions =
+  [ (optSafe, "--safe")
+  , (collapseDefault . optWithoutK, "--without-K")
+  , (not . optUniversePolymorphism, "--no-universe-polymorphism")
+  , (not . collapseDefault . optSizedTypes, "--no-sized-types")
+  , (not . collapseDefault . optGuardedness, "--no-guardedness")
+  ]
+
 inputFlag :: FilePath -> Flag CommandLineOptions
 inputFlag f o =
     case optInputFile o of
@@ -354,28 +455,54 @@ helpFlag (Just str) o = case string2HelpTopic str of
                            intercalate ", " (map fst allHelpTopics) ++ ")"
 
 safeFlag :: Flag PragmaOptions
-safeFlag o = return $ o { optSafe = True }
+safeFlag o = do
+  let guardedness = optGuardedness o
+  let sizedTypes  = optSizedTypes o
+  return $ o { optSafe        = True
+             , optGuardedness = setDefault False guardedness
+             , optSizedTypes  = setDefault False sizedTypes
+             }
+
+doubleCheckFlag :: Flag PragmaOptions
+doubleCheckFlag o = return $ o { optDoubleCheck = True }
+
+noSyntacticEqualityFlag :: Flag PragmaOptions
+noSyntacticEqualityFlag o = return $ o { optSyntacticEquality = False }
 
 sharingFlag :: Bool -> Flag CommandLineOptions
-sharingFlag _ o = return o
+sharingFlag _ _ = throwError $
+  "Feature --sharing has been removed (in favor of the Agda abstract machine)."
 
 cachingFlag :: Bool -> Flag PragmaOptions
 cachingFlag b o = return $ o { optCaching = b }
 
-proofIrrelevanceFlag :: Flag PragmaOptions
-proofIrrelevanceFlag o = return $ o { optProofIrrelevance = True }
+propFlag :: Flag PragmaOptions
+propFlag o = return $ o { optProp = True }
+
+noPropFlag :: Flag PragmaOptions
+noPropFlag o = return $ o { optProp = False }
 
 experimentalIrrelevanceFlag :: Flag PragmaOptions
 experimentalIrrelevanceFlag o = return $ o { optExperimentalIrrelevance = True }
 
+irrelevantProjectionsFlag :: Flag PragmaOptions
+irrelevantProjectionsFlag o = return $ o { optIrrelevantProjections = True }
+
 noIrrelevantProjectionsFlag :: Flag PragmaOptions
 noIrrelevantProjectionsFlag o = return $ o { optIrrelevantProjections = False }
 
 ignoreInterfacesFlag :: Flag CommandLineOptions
 ignoreInterfacesFlag o = return $ o { optIgnoreInterfaces = True }
 
+ignoreAllInterfacesFlag :: Flag CommandLineOptions
+ignoreAllInterfacesFlag o = return $ o { optIgnoreAllInterfaces = True }
+
 allowUnsolvedFlag :: Flag PragmaOptions
-allowUnsolvedFlag o = return $ o { optAllowUnsolved = True }
+allowUnsolvedFlag o = do
+  let upd = over warningSet (Set.\\ unsolvedWarnings)
+  return $ o { optAllowUnsolved = True
+             , optWarningMode   = upd (optWarningMode o)
+             }
 
 showImplicitFlag :: Flag PragmaOptions
 showImplicitFlag o = return $ o { optShowImplicit = True }
@@ -391,6 +518,9 @@ asciiOnlyFlag o = do
 ghciInteractionFlag :: Flag CommandLineOptions
 ghciInteractionFlag o = return $ o { optGHCiInteraction = True }
 
+jsonInteractionFlag :: Flag CommandLineOptions
+jsonInteractionFlag o = return $ o { optJSONInteraction = True }
+
 vimFlag :: Flag CommandLineOptions
 vimFlag o = return $ o { optGenerateVimFile = True }
 
@@ -415,23 +545,37 @@ noAutoInlineFlag o = return $ o { optAutoInline = False }
 noPrintPatSynFlag :: Flag PragmaOptions
 noPrintPatSynFlag o = return $ o { optPrintPatternSynonyms = False }
 
+noFastReduceFlag :: Flag PragmaOptions
+noFastReduceFlag o = return $ o { optFastReduce = False }
+
 latexDirFlag :: FilePath -> Flag CommandLineOptions
 latexDirFlag d o = return $ o { optLaTeXDir = d }
 
 noPositivityFlag :: Flag PragmaOptions
-noPositivityFlag o = return $ o { optDisablePositivity = True }
+noPositivityFlag o = do
+  let upd = over warningSet (Set.delete NotStrictlyPositive_)
+  return $ o { optDisablePositivity = True
+             , optWarningMode   = upd (optWarningMode o)
+             }
 
 dontTerminationCheckFlag :: Flag PragmaOptions
-dontTerminationCheckFlag o = return $ o { optTerminationCheck = False }
+dontTerminationCheckFlag o = do
+  let upd = over warningSet (Set.delete TerminationIssue_)
+  return $ o { optTerminationCheck = False
+             , optWarningMode   = upd (optWarningMode o)
+             }
 
 -- The option was removed. See Issue 1918.
 dontCompletenessCheckFlag :: Flag PragmaOptions
 dontCompletenessCheckFlag _ =
-  throwError "the --no-coverage-check option has been removed"
+  throwError "The --no-coverage-check option has been removed."
 
 dontUniverseCheckFlag :: Flag PragmaOptions
-dontUniverseCheckFlag o = return $ o { optUniverseCheck        = False
-                                     }
+dontUniverseCheckFlag o = return $ o { optUniverseCheck = False }
+
+omegaInOmegaFlag :: Flag PragmaOptions
+omegaInOmegaFlag o = return $ o { optOmegaInOmega = True }
+
 etaFlag :: Flag PragmaOptions
 etaFlag o = return $ o { optEta = True }
 
@@ -439,16 +583,23 @@ noEtaFlag :: Flag PragmaOptions
 noEtaFlag o = return $ o { optEta = False }
 
 sizedTypes :: Flag PragmaOptions
-sizedTypes o = return $ o { optSizedTypes = True }
+sizedTypes o = return $ o { optSizedTypes = Value True }
 
 noSizedTypes :: Flag PragmaOptions
-noSizedTypes o = return $ o { optSizedTypes = False }
+noSizedTypes o = return $ o { optSizedTypes = Value False }
+
+guardedness :: Flag PragmaOptions
+guardedness o = return $ o { optGuardedness = Value True }
+
+noGuardedness :: Flag PragmaOptions
+noGuardedness o = return $ o { optGuardedness = Value False }
 
 injectiveTypeConstructorFlag :: Flag PragmaOptions
 injectiveTypeConstructorFlag o = return $ o { optInjectiveTypeConstructors = True }
 
 guardingTypeConstructorFlag :: Flag PragmaOptions
-guardingTypeConstructorFlag o = return $ o { optGuardingTypeConstructors = True }
+guardingTypeConstructorFlag _ = throwError $
+  "Experimental feature --guardedness-preserving-type-constructors has been removed."
 
 universePolymorphismFlag :: Flag PragmaOptions
 universePolymorphismFlag o = return $ o { optUniversePolymorphism = True }
@@ -460,10 +611,10 @@ noForcingFlag :: Flag CommandLineOptions
 noForcingFlag o = return $ o { optForcing = False }
 
 withKFlag :: Flag PragmaOptions
-withKFlag o = return $ o { optWithoutK = False }
+withKFlag o = return $ o { optWithoutK = Value False }
 
 withoutKFlag :: Flag PragmaOptions
-withoutKFlag o = return $ o { optWithoutK = True }
+withoutKFlag o = return $ o { optWithoutK = Value True }
 
 copatternsFlag :: Flag PragmaOptions
 copatternsFlag o = return $ o { optCopatterns = True }
@@ -475,14 +626,29 @@ noPatternMatchingFlag :: Flag PragmaOptions
 noPatternMatchingFlag o = return $ o { optPatternMatching = False }
 
 exactSplitFlag :: Flag PragmaOptions
-exactSplitFlag o = return $ o { optExactSplit = True }
+exactSplitFlag o = do
+  let upd = over warningSet (Set.insert CoverageNoExactSplit_)
+  return $ o { optExactSplit = True
+             , optWarningMode   = upd (optWarningMode o)
+             }
 
 noExactSplitFlag :: Flag PragmaOptions
-noExactSplitFlag o = return $ o { optExactSplit = False }
+noExactSplitFlag o = do
+  let upd = over warningSet (Set.delete CoverageNoExactSplit_)
+  return $ o { optExactSplit = False
+             , optWarningMode   = upd (optWarningMode o)
+             }
 
 rewritingFlag :: Flag PragmaOptions
 rewritingFlag o = return $ o { optRewriting = True }
 
+cubicalFlag :: Flag PragmaOptions
+cubicalFlag o = do
+  let withoutK = optWithoutK o
+  return $ o { optCubical  = True
+             , optWithoutK = setDefault True withoutK
+             }
+
 postfixProjectionsFlag :: Flag PragmaOptions
 postfixProjectionsFlag o = return $ o { optPostfixProjections = True }
 
@@ -491,16 +657,23 @@ instanceDepthFlag s o = do
   d <- integerArgument "--instance-search-depth" s
   return $ o { optInstanceSearchDepth = d }
 
+overlappingInstancesFlag :: Flag PragmaOptions
+overlappingInstancesFlag o = return $ o { optOverlappingInstances = True }
+
+noOverlappingInstancesFlag :: Flag PragmaOptions
+noOverlappingInstancesFlag o = return $ o { optOverlappingInstances = False }
+
 inversionMaxDepthFlag :: String -> Flag PragmaOptions
 inversionMaxDepthFlag s o = do
   d <- integerArgument "--inversion-max-depth" s
   return $ o { optInversionMaxDepth = d }
 
 interactiveFlag :: Flag CommandLineOptions
-interactiveFlag  o = return $ o { optInteractive    = True
-                                , optPragmaOptions  = (optPragmaOptions o)
-                                                      { optAllowUnsolved = True }
-                                }
+interactiveFlag  o = do
+  prag <- allowUnsolvedFlag (optPragmaOptions o)
+  return $ o { optInteractive    = True
+             , optPragmaOptions = prag
+             }
 
 compileFlagNoMain :: Flag PragmaOptions
 compileFlagNoMain o = return $ o { optCompileNoMain = True }
@@ -511,6 +684,13 @@ compileDirFlag f o = return $ o { optCompileDir = Just f }
 htmlFlag :: Flag CommandLineOptions
 htmlFlag o = return $ o { optGenerateHTML = True }
 
+htmlHighlightFlag :: String -> Flag CommandLineOptions
+htmlHighlightFlag "code" o = return $ o { optHTMLHighlight = HighlightCode }
+htmlHighlightFlag "all"  o = return $ o { optHTMLHighlight = HighlightAll  }
+htmlHighlightFlag "auto" o = return $ o { optHTMLHighlight = HighlightAuto  }
+htmlHighlightFlag opt    o = throwError $ "Invalid option <" ++ opt
+  ++ ">, expected <all>, <auto> or <code>"
+
 dependencyGraphFlag :: FilePath -> Flag CommandLineOptions
 dependencyGraphFlag f o = return $ o { optDependencyGraph = Just f }
 
@@ -543,7 +723,7 @@ verboseFlag s o =
     do  (k,n) <- parseVerbose s
         return $ o { optVerbose = Trie.insert k n $ optVerbose o }
   where
-    parseVerbose s = case wordsBy (`elem` ":.") s of
+    parseVerbose s = case wordsBy (`elem` (":." :: String)) s of
       []  -> usage
       ss  -> do
         n <- readM (last ss) `catchError` \_ -> usage
@@ -578,6 +758,8 @@ standardOptions =
                     "start in interactive mode"
     , Option []     ["interaction"] (NoArg ghciInteractionFlag)
                     "for use with the Emacs mode"
+    , Option []     ["interaction-json"] (NoArg jsonInteractionFlag)
+                    "for use with other editors such as Atom"
 
     , Option []     ["compile-dir"] (ReqArg compileDirFlag "DIR")
                     ("directory for compiler output (default: the project root)")
@@ -596,6 +778,10 @@ standardOptions =
                      defaultHTMLDir ++ ")")
     , Option []     ["css"] (ReqArg cssFlag "URL")
                     "the CSS file used by the HTML files (can be relative)"
+    , Option []     ["html-highlight"] (ReqArg htmlHighlightFlag "[code,all,auto]")
+                    ("whether to highlight only the code parts (code) or " ++
+                    "the file as a whole (all) or " ++
+                    "decide by source file type (auto)")
     , Option []     ["dependency-graph"] (ReqArg dependencyGraphFlag "FILE")
                     "generate a Dot file with a module dependency graph"
     , Option []     ["ignore-interfaces"] (NoArg ignoreInterfacesFlag)
@@ -612,18 +798,26 @@ standardOptions =
                     "don't use default libraries"
     , Option []     ["no-forcing"] (NoArg noForcingFlag)
                     "disable the forcing optimisation"
-    , Option []     ["sharing"] (NoArg $ sharingFlag True)
+    , Option []     ["only-scope-checking"] (NoArg onlyScopeCheckingFlag)
+                    "only scope-check the top-level module, do not type-check it"
+    ] ++ map (fmap lensPragmaOptions) pragmaOptions
+
+-- | Defined locally here since module ''Agda.Interaction.Options.Lenses''
+--   has cyclic dependency.
+lensPragmaOptions :: Lens' PragmaOptions CommandLineOptions
+lensPragmaOptions f st = f (optPragmaOptions st) <&> \ opts -> st { optPragmaOptions = opts }
+
+-- | Command line options of previous versions of Agda.
+--   Should not be listed in the usage info, put parsed by GetOpt for good error messaging.
+deadStandardOptions :: [OptDescr (Flag CommandLineOptions)]
+deadStandardOptions =
+    [ Option []     ["sharing"] (NoArg $ sharingFlag True)
                     "DEPRECATED: does nothing"
     , Option []     ["no-sharing"] (NoArg $ sharingFlag False)
                     "DEPRECATED: does nothing"
-    , Option []     ["only-scope-checking"] (NoArg onlyScopeCheckingFlag)
-                    "only scope-check the top-level module, do not type-check it"
-    ] ++ map (fmap lift) pragmaOptions
-  where
-  lift :: Flag PragmaOptions -> Flag CommandLineOptions
-  lift f = \opts -> do
-    ps <- f (optPragmaOptions opts)
-    return (opts { optPragmaOptions = ps })
+    , Option []     ["ignore-all-interfaces"] (NoArg ignoreAllInterfacesFlag) -- not deprecated! Just hidden
+                    "ignore all interface files (re-type check everything, including builtin files)"
+    ] ++ map (fmap lensPragmaOptions) deadPragmaOptions
 
 pragmaOptions :: [OptDescr (Flag PragmaOptions)]
 pragmaOptions =
@@ -635,8 +829,6 @@ pragmaOptions =
                     "don't use unicode characters when printing terms"
     , Option ['v']  ["verbose"] (ReqArg verboseFlag "N")
                     "set verbosity level to N"
-    -- , Option []          ["proof-irrelevance"] (NoArg proofIrrelevanceFlag)
-    --              "enable proof irrelevance (experimental feature)"
     , Option []     ["allow-unsolved-metas"] (NoArg allowUnsolvedFlag)
                     "succeed and create interface file regardless of unsolved meta variables"
     , Option []     ["no-positivity-check"] (NoArg noPositivityFlag)
@@ -645,28 +837,36 @@ pragmaOptions =
                     "do not warn about possibly nonterminating code"
     , Option []     ["termination-depth"] (ReqArg terminationDepthFlag "N")
                     "allow termination checker to count decrease/increase upto N (default N=1)"
-    , Option []     ["no-coverage-check"] (NoArg dontCompletenessCheckFlag)
-                    "the option has been removed"
     , Option []     ["type-in-type"] (NoArg dontUniverseCheckFlag)
                     "ignore universe levels (this makes Agda inconsistent)"
+    , Option []     ["omega-in-omega"] (NoArg omegaInOmegaFlag)
+                    "enable typing rule Setω : Setω (this makes Agda inconsistent)"
+    , Option []     ["prop"] (NoArg propFlag)
+                    "enable the use of the Prop universe"
+    , Option []     ["no-prop"] (NoArg noPropFlag)
+                    "disable the use of the Prop universe (default)"
     , Option []     ["sized-types"] (NoArg sizedTypes)
-                    "use sized types (default, inconsistent with `musical' coinduction)"
+                    "enable sized types (default, inconsistent with --guardedness)"
     , Option []     ["no-sized-types"] (NoArg noSizedTypes)
                     "disable sized types"
+    , Option []     ["guardedness"] (NoArg guardedness)
+                    "enable constructor-based guarded corecursion (default, inconsistent with --sized-types)"
+    , Option []     ["no-guardedness"] (NoArg noGuardedness)
+                    "disable constructor-based guarded corecursion"
     , Option []     ["injective-type-constructors"] (NoArg injectiveTypeConstructorFlag)
                     "enable injective type constructors (makes Agda anti-classical and possibly inconsistent)"
-    , Option []     ["guardedness-preserving-type-constructors"] (NoArg guardingTypeConstructorFlag)
-                    "treat type constructors as inductive constructors when checking productivity"
     , Option []     ["no-universe-polymorphism"] (NoArg noUniversePolymorphismFlag)
                     "disable universe polymorphism"
     , Option []     ["universe-polymorphism"] (NoArg universePolymorphismFlag)
                     "enable universe polymorphism (default)"
+    , Option []     ["irrelevant-projections"] (NoArg irrelevantProjectionsFlag)
+                    "enable projection of irrelevant record fields and similar irrelevant definitions (inconsistent)"
     , Option []     ["no-irrelevant-projections"] (NoArg noIrrelevantProjectionsFlag)
-                    "disable projection of irrelevant record fields"
+                    "disable projection of irrelevant record fields and similar irrelevant definitions (default)"
     , Option []     ["experimental-irrelevance"] (NoArg experimentalIrrelevanceFlag)
                     "enable potentially unsound irrelevance features (irrelevant levels, irrelevant data matching)"
     , Option []     ["with-K"] (NoArg withKFlag)
-                    "enable the K rule in pattern matching"
+                    "enable the K rule in pattern matching (default)"
     , Option []     ["without-K"] (NoArg withoutKFlag)
                     "disable the K rule in pattern matching"
     , Option []     ["copatterns"] (NoArg copatternsFlag)
@@ -683,14 +883,24 @@ pragmaOptions =
                     "default records to no-eta-equality"
     , Option []     ["rewriting"] (NoArg rewritingFlag)
                     "enable declaration and use of REWRITE rules"
+    , Option []     ["cubical"] (NoArg cubicalFlag)
+                    "enable cubical features (e.g. overloads lambdas for paths), implies --without-K"
     , Option []     ["postfix-projections"] (NoArg postfixProjectionsFlag)
                     "make postfix projection notation the default"
     , Option []     ["instance-search-depth"] (ReqArg instanceDepthFlag "N")
                     "set instance search depth to N (default: 500)"
+    , Option []     ["overlapping-instances"] (NoArg overlappingInstancesFlag)
+                    "consider recursive instance arguments during pruning of instance candidates"
+    , Option []     ["no-overlapping-instances"] (NoArg noOverlappingInstancesFlag)
+                    "don't consider recursive instance arguments during pruning of instance candidates (default)"
     , Option []     ["inversion-max-depth"] (ReqArg inversionMaxDepthFlag "N")
                     "set maximum depth for pattern match inversion to N (default: 50)"
     , Option []     ["safe"] (NoArg safeFlag)
-                    "disable postulates, unsafe OPTION pragmas and primTrustMe"
+                    "disable postulates, unsafe OPTION pragmas and primEraseEquality, implies --no-sized-types and --no-guardedness "
+    , Option []     ["double-check"] (NoArg doubleCheckFlag)
+                    "enable double-checking of all terms using the internal typechecker"
+    , Option []     ["no-syntactic-equality"] (NoArg noSyntacticEqualityFlag)
+                    "disable the syntactic equality shortcut in the conversion checker"
     , Option ['W']  ["warning"] (ReqArg warningModeFlag "FLAG")
                     ("set warning flags. See --help=warning.")
     , Option []     ["no-main"] (NoArg compileFlagNoMain)
@@ -713,9 +923,22 @@ pragmaOptions =
                      "(only definitions marked INLINE will be inlined)")
     , Option []     ["no-print-pattern-synonyms"] (NoArg noPrintPatSynFlag)
                     "expand pattern synonyms when printing terms"
+    , Option []     ["no-fast-reduce"] (NoArg noFastReduceFlag)
+                    "disable reduction using the Agda Abstract Machine"
+    ]
+
+-- | Pragma options of previous versions of Agda.
+--   Should not be listed in the usage info, put parsed by GetOpt for good error messaging.
+deadPragmaOptions :: [OptDescr (Flag PragmaOptions)]
+deadPragmaOptions =
+    [ Option []     ["guardedness-preserving-type-constructors"] (NoArg guardingTypeConstructorFlag)
+                    "treat type constructors as inductive constructors when checking productivity"
+    , Option []     ["no-coverage-check"] (NoArg dontCompletenessCheckFlag)
+                    "the option has been removed"
     ]
 
 -- | Used for printing usage info.
+--   Does not include the dead options.
 standardOptions_ :: [OptDescr ()]
 standardOptions_ = map (fmap $ const ()) standardOptions
 
@@ -767,14 +990,16 @@ getOptSimple argv opts fileArg = \ defaults ->
       sugs [a] = a
       sugs as  = "any of " ++ intercalate " " as
 
+{- No longer used in favour of parseBackendOptions in Agda.Compiler.Backend
 -- | Parse the standard options.
 parseStandardOptions :: [String] -> OptM CommandLineOptions
 parseStandardOptions argv = parseStandardOptions' argv defaultOptions
 
 parseStandardOptions' :: [String] -> Flag CommandLineOptions
 parseStandardOptions' argv opts = do
-  opts <- getOptSimple (stripRTS argv) standardOptions inputFlag opts
+  opts <- getOptSimple (stripRTS argv) (deadStandardOptions ++ standardOptions) inputFlag opts
   checkOpts opts
+-}
 
 -- | Parse options from an options pragma.
 parsePragmaOptions
@@ -784,7 +1009,7 @@ parsePragmaOptions
      -- ^ Command-line options which should be updated.
   -> OptM PragmaOptions
 parsePragmaOptions argv opts = do
-  ps <- getOptSimple argv pragmaOptions
+  ps <- getOptSimple argv (deadPragmaOptions ++ pragmaOptions)
           (\s _ -> throwError $ "Bad option in pragma: " ++ s)
           (optPragmaOptions opts)
   _ <- checkOpts (opts { optPragmaOptions = ps })
@@ -807,7 +1032,8 @@ usage options progName GeneralHelp = usageInfo (header progName) options
 
 usage options progName (HelpFor topic) = helpTopicUsage topic
 
--- Remove +RTS .. -RTS from arguments
+-- | Removes RTS options from a list of options.
+
 stripRTS :: [String] -> [String]
 stripRTS [] = []
 stripRTS ("--RTS" : argv) = argv
