@@ -8,6 +8,11 @@ module Agda.Interaction.Options.Lenses where
 
 import Control.Monad.State
 
+import Data.Set (Set)
+import qualified Data.Set as Set
+
+import System.FilePath ((</>))
+
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.State
 import Agda.Interaction.Options
@@ -20,9 +25,11 @@ import Agda.Utils.FileName
 ---------------------------------------------------------------------------
 
 class LensPragmaOptions a where
-  getPragmaOptions :: a -> PragmaOptions
-  setPragmaOptions :: PragmaOptions -> a -> a
-  mapPragmaOptions :: (PragmaOptions -> PragmaOptions) -> a -> a
+  getPragmaOptions  :: a -> PragmaOptions
+  setPragmaOptions  :: PragmaOptions -> a -> a
+  mapPragmaOptions  :: (PragmaOptions -> PragmaOptions) -> a -> a
+  lensPragmaOptions :: Lens' PragmaOptions a
+  -- lensPragmaOptions :: forall f. Functor f => (PragmaOptions -> f PragmaOptions) -> a -> f a
 
   -- default implementations
   setPragmaOptions     = mapPragmaOptions . const
@@ -31,13 +38,15 @@ class LensPragmaOptions a where
 instance LensPragmaOptions CommandLineOptions where
   getPragmaOptions = optPragmaOptions
   setPragmaOptions opts st = st { optPragmaOptions = opts }
+  lensPragmaOptions f st = f (optPragmaOptions st) <&> \ opts -> st { optPragmaOptions = opts }
 
 instance LensPragmaOptions TCState where
   getPragmaOptions = (^.stPragmaOptions)
   setPragmaOptions = set stPragmaOptions
+  lensPragmaOptions = stPragmaOptions
 
 modifyPragmaOptions :: (PragmaOptions -> PragmaOptions) -> TCM ()
-modifyPragmaOptions = modify . mapPragmaOptions
+modifyPragmaOptions = modifyTC . mapPragmaOptions
 
 ---------------------------------------------------------------------------
 -- ** Verbosity in the local pragma options
@@ -61,10 +70,10 @@ instance LensVerbosity TCState where
   mapVerbosity = mapPragmaOptions . mapVerbosity
 
 modifyVerbosity :: (Verbosity -> Verbosity) -> TCM ()
-modifyVerbosity = modify . mapVerbosity
+modifyVerbosity = modifyTC . mapVerbosity
 
 putVerbosity :: Verbosity -> TCM ()
-putVerbosity = modify . setVerbosity
+putVerbosity = modifyTC . setVerbosity
 
 ---------------------------------------------------------------------------
 -- * Command line options
@@ -88,7 +97,7 @@ instance LensCommandLineOptions TCState where
   mapCommandLineOptions = updatePersistentState . mapCommandLineOptions
 
 modifyCommandLineOptions :: (CommandLineOptions -> CommandLineOptions) -> TCM ()
-modifyCommandLineOptions = modify . mapCommandLineOptions
+modifyCommandLineOptions = modifyTC . mapCommandLineOptions
 
 ---------------------------------------------------------------------------
 -- ** Safe mode
@@ -107,7 +116,7 @@ class LensSafeMode a where
 
 instance LensSafeMode PragmaOptions where
   getSafeMode = optSafe
-  setSafeMode is opts = opts { optSafe = is }
+  setSafeMode is opts = opts { optSafe = is } -- setSafeOption
 
 instance LensSafeMode CommandLineOptions where
   getSafeMode = getSafeMode . getPragmaOptions
@@ -122,11 +131,74 @@ instance LensSafeMode TCState where
   mapSafeMode = mapCommandLineOptions . mapSafeMode
 
 modifySafeMode :: (SafeMode -> SafeMode) -> TCM ()
-modifySafeMode = modify . mapSafeMode
+modifySafeMode = modifyTC . mapSafeMode
 
 putSafeMode :: SafeMode -> TCM ()
-putSafeMode = modify . setSafeMode
+putSafeMode = modifyTC . setSafeMode
 
+-- | These builtins may use postulates, and are still considered --safe
+
+builtinModulesWithSafePostulates :: Set FilePath
+builtinModulesWithSafePostulates =
+  primitiveModules `Set.union` (Set.fromList
+  [ "Agda" </> "Builtin" </> "Bool.agda"
+  , "Agda" </> "Builtin" </> "Char.agda"
+  , "Agda" </> "Builtin" </> "Char" </> "Properties.agda"
+  , "Agda" </> "Builtin" </> "Coinduction.agda"
+  , "Agda" </> "Builtin" </> "Cubical" </> "Glue.agda"
+  , "Agda" </> "Builtin" </> "Cubical" </> "Id.agda"
+  , "Agda" </> "Builtin" </> "Cubical" </> "Path.agda"
+  , "Agda" </> "Builtin" </> "Cubical" </> "Sub.agda"
+  , "Agda" </> "Builtin" </> "Equality" </> "Erase.agda"
+  , "Agda" </> "Builtin" </> "Equality.agda"
+  , "Agda" </> "Builtin" </> "Float.agda"
+  , "Agda" </> "Builtin" </> "FromNat.agda"
+  , "Agda" </> "Builtin" </> "FromNeg.agda"
+  , "Agda" </> "Builtin" </> "FromString.agda"
+  , "Agda" </> "Builtin" </> "Int.agda"
+  , "Agda" </> "Builtin" </> "IO.agda"
+  , "Agda" </> "Builtin" </> "List.agda"
+  , "Agda" </> "Builtin" </> "Nat.agda"
+  , "Agda" </> "Builtin" </> "Reflection.agda"
+  , "Agda" </> "Builtin" </> "Sigma.agda"
+  , "Agda" </> "Builtin" </> "Size.agda"
+  , "Agda" </> "Builtin" </> "Strict.agda"
+  , "Agda" </> "Builtin" </> "String.agda"
+  , "Agda" </> "Builtin" </> "String" </> "Properties.agda"
+  , "Agda" </> "Builtin" </> "Unit.agda"
+  , "Agda" </> "Builtin" </> "Word.agda"
+  , "Agda" </> "Builtin" </> "Word" </> "Properties.agda"
+  ])
+
+-- | These builtins may not use postulates under --safe. They are not
+--   automatically unsafe, but will be if they use an unsafe feature.
+
+builtinModulesWithUnsafePostulates :: Set FilePath
+builtinModulesWithUnsafePostulates = Set.fromList
+  [ "Agda" </> "Builtin" </> "TrustMe.agda"
+  , "Agda" </> "Builtin" </> "Equality" </> "Rewrite.agda"
+  ]
+
+primitiveModules :: Set FilePath
+primitiveModules = Set.fromList
+  [ "Agda" </> "Primitive.agda"
+  , "Agda" </> "Primitive" </> "Cubical.agda"
+  ]
+
+builtinModules :: Set FilePath
+builtinModules = builtinModulesWithSafePostulates `Set.union`
+                 builtinModulesWithUnsafePostulates
+
+isBuiltinModule :: FilePath -> TCM Bool
+isBuiltinModule file = do
+  libdirPrim <- (</> "prim") <$> liftIO defaultLibDir
+  return (file `Set.member` Set.map (libdirPrim </>) builtinModules)
+
+isBuiltinModuleWithSafePostulates :: FilePath -> TCM Bool
+isBuiltinModuleWithSafePostulates file = do
+  libdirPrim <- (</> "prim") <$> liftIO defaultLibDir
+  let safeBuiltins   = Set.map (libdirPrim </>) builtinModulesWithSafePostulates
+  return (file `Set.member` safeBuiltins)
 
 ---------------------------------------------------------------------------
 -- ** Include directories
@@ -166,16 +238,16 @@ instance LensIncludePaths TCState where
   mapAbsoluteIncludePaths = mapCommandLineOptions . mapAbsoluteIncludePaths
 
 modifyIncludePaths :: ([FilePath] -> [FilePath]) -> TCM ()
-modifyIncludePaths = modify . mapIncludePaths
+modifyIncludePaths = modifyTC . mapIncludePaths
 
 putIncludePaths :: [FilePath] -> TCM ()
-putIncludePaths = modify . setIncludePaths
+putIncludePaths = modifyTC . setIncludePaths
 
 modifyAbsoluteIncludePaths :: ([AbsolutePath] -> [AbsolutePath]) -> TCM ()
-modifyAbsoluteIncludePaths = modify . mapAbsoluteIncludePaths
+modifyAbsoluteIncludePaths = modifyTC . mapAbsoluteIncludePaths
 
 putAbsoluteIncludePaths :: [AbsolutePath] -> TCM ()
-putAbsoluteIncludePaths = modify . setAbsoluteIncludePaths
+putAbsoluteIncludePaths = modifyTC . setAbsoluteIncludePaths
 
 ---------------------------------------------------------------------------
 -- ** Include directories
@@ -208,7 +280,7 @@ instance LensPersistentVerbosity TCState where
   mapPersistentVerbosity = mapCommandLineOptions . mapPersistentVerbosity
 
 modifyPersistentVerbosity :: (PersistentVerbosity -> PersistentVerbosity) -> TCM ()
-modifyPersistentVerbosity = modify . mapPersistentVerbosity
+modifyPersistentVerbosity = modifyTC . mapPersistentVerbosity
 
 putPersistentVerbosity :: PersistentVerbosity -> TCM ()
-putPersistentVerbosity = modify . setPersistentVerbosity
+putPersistentVerbosity = modifyTC . setPersistentVerbosity
