@@ -82,7 +82,6 @@ data DisplayInfo
     | Info_Error String
         -- ^ When an error message is displayed this constructor should be
         -- used, if appropriate.
---    | Info_Warning String --FNF: currently unused
     | Info_Intro Doc
         -- ^ 'Info_Intro' denotes two different types of errors
         --   TODO: split these into separate constructors
@@ -139,7 +138,7 @@ data GiveResult
 --      closure of the 'InteractionOutputCallback' function.
 --      (suitable for intra-process communication).
 
-type InteractionOutputCallback = Response -> IO ()
+type InteractionOutputCallback = Response -> TCM ()
 
 -- | The default 'InteractionOutputCallback' function prints certain
 -- things to stdout (other things generate internal errors).
