@@ -15,18 +15,20 @@ import Prelude hiding ((<>))
 
 import Control.DeepSeq
 
-import Data.Foldable (Foldable, toList)
+import Data.Foldable (Foldable)
 import Data.Traversable (Traversable)
 import Data.Data (Data)
 import Data.List
 import Data.Function
 import Data.Hashable (Hashable(..))
+import Data.Set (Set)
+import qualified Data.Set as Set
 import Data.Void
 
 import Agda.Syntax.Position
 import Agda.Syntax.Common
 import {-# SOURCE #-} Agda.Syntax.Fixity
-import Agda.Syntax.Concrete.Name (IsNoName(..), NumHoles(..))
+import Agda.Syntax.Concrete.Name (IsNoName(..), NumHoles(..), NameInScope(..), LensInScope(..))
 import qualified Agda.Syntax.Concrete.Name as C
 
 import Agda.Utils.List
@@ -35,7 +37,6 @@ import Agda.Utils.Monad
 import Agda.Utils.NonemptyList
 import Agda.Utils.Pretty
 import Agda.Utils.Size
-import Agda.Utils.Suffix
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -47,9 +48,14 @@ data Name = Name { nameId          :: !NameId
                  , nameConcrete    :: C.Name
                  , nameBindingSite :: Range
                  , nameFixity      :: Fixity'
+                 , nameIsRecordName :: Bool
                  }
     deriving Data
 
+-- | Useful for debugging scoping problems
+uglyShowName :: Name -> String
+uglyShowName (Name i c _ _ _) = show (i,c)
+
 -- | Qualified names are non-empty lists of names. Equality on qualified names
 --   are just equality on the last name, i.e. the module part is just
 --   for show.
@@ -104,7 +110,10 @@ class IsProjP a where
   isProjP :: a -> Maybe (ProjOrigin, AmbiguousQName)
 
 instance IsProjP a => IsProjP (Arg a) where
-  isProjP = isProjP . unArg
+  isProjP p = case isProjP $ unArg p of
+    Just (ProjPostfix , f)
+     | getHiding p /= NotHidden -> Nothing
+    x -> x
 
 instance IsProjP a => IsProjP (Named n a) where
   isProjP = isProjP . namedThing
@@ -166,7 +175,7 @@ class MkName a where
   mkName_ = mkName noRange
 
 instance MkName String where
-  mkName r i s = Name i (C.Name noRange (C.stringNameParts s)) r noFixity'
+  mkName r i s = Name i (C.Name noRange InScope (C.stringNameParts s)) r noFixity' False
 
 
 qnameToList :: QName -> [Name]
@@ -240,16 +249,10 @@ isInModule q m = mnameToList m `isPrefixOf` qnameToList q
 -- | Get the next version of the concrete name. For instance, @nextName "x" = "x₁"@.
 --   The name must not be a 'NoName'.
 nextName :: Name -> Name
-nextName x = x { nameConcrete = C.Name noRange $ nextSuf ps }
-    where
-        C.Name _ ps = nameConcrete x
-        -- NoName cannot appear here
-        nextSuf [C.Id s]         = [C.Id $ nextStr s]
-        nextSuf [C.Id s, C.Hole] = [C.Id $ nextStr s, C.Hole]
-        nextSuf (p : ps)         = p : nextSuf ps
-        nextSuf []               = __IMPOSSIBLE__
-        nextStr s = case suffixView s of
-            (s0, suf) -> addSuffix s0 (nextSuffix suf)
+nextName x = x { nameConcrete = C.nextName (nameConcrete x) }
+
+sameRoot :: Name -> Name -> Bool
+sameRoot = C.sameRoot `on` nameConcrete
 
 ------------------------------------------------------------------------
 -- * Important instances: Eq, Ord, Hashable
@@ -295,6 +298,18 @@ instance NumHoles QName where
 instance NumHoles AmbiguousQName where
   numHoles = numHoles . headAmbQ
 
+------------------------------------------------------------------------
+-- * LensInScope instances
+------------------------------------------------------------------------
+
+instance LensInScope Name where
+  lensInScope f n@Name{ nameConcrete = x } =
+    (\y -> n { nameConcrete = y }) <$> lensInScope f x
+
+instance LensInScope QName where
+  lensInScope f q@QName{ qnameName = n } =
+    (\n' -> q { qnameName = n' }) <$> lensInScope f n
+
 ------------------------------------------------------------------------
 -- * Show instances (only for debug printing!)
 --
@@ -334,16 +349,16 @@ instance Pretty Name where
   pretty = pretty . nameConcrete
 
 instance Pretty ModuleName where
-  pretty = hcat . punctuate (text ".") . map pretty . mnameToList
+  pretty = hcat . punctuate "." . map pretty . mnameToList
 
 instance Pretty QName where
-  pretty = hcat . punctuate (text ".") . map pretty . qnameToList
+  pretty = hcat . punctuate "." . map pretty . qnameToList
 
 instance Pretty AmbiguousQName where
-  pretty (AmbQ qs) = hcat $ punctuate (text " | ") $ map pretty (toList qs)
+  pretty (AmbQ qs) = hcat $ punctuate " | " $ map pretty (toList qs)
 
 instance Pretty a => Pretty (QNamed a) where
-  pretty (QNamed a b) = pretty a <> text "." <> pretty b
+  pretty (QNamed a b) = pretty a <> "." <> pretty b
 
 ------------------------------------------------------------------------
 -- * Range instances
@@ -387,8 +402,8 @@ instance SetRange ModuleName where
 -- ** KillRange
 
 instance KillRange Name where
-  killRange (Name a b c d) =
-    (killRange4 Name a b c d) { nameBindingSite = c }
+  killRange (Name a b c d e) =
+    (killRange4 Name a b c d e) { nameBindingSite = c }
     -- Andreas, 2017-07-25, issue #2649
     -- Preserve the nameBindingSite for error message.
     --
@@ -431,7 +446,7 @@ instance Sized ModuleName where
 -- | The range is not forced.
 
 instance NFData Name where
-  rnf (Name _ a _ b) = rnf a `seq` rnf b
+  rnf (Name _ a _ b c) = rnf a `seq` rnf b `seq` rnf c
 
 instance NFData QName where
   rnf (QName a b) = rnf a `seq` rnf b
