@@ -57,6 +57,7 @@ instance MapNamedArgPattern NAP where
       VarP{}    -> f p
       WildP{}   -> f p
       DotP{}    -> f p
+      EqualP{}  -> f p
       LitP{}    -> f p
       AbsurdP{} -> f p
       ProjP{}   -> f p
@@ -152,6 +153,7 @@ instance APatternLike a (Pattern' a) where
       DotP _ _           -> mempty
       AbsurdP _          -> mempty
       LitP _             -> mempty
+      EqualP _ _         -> mempty
 
   traverseAPatternM pre post = pre >=> recurse >=> post
     where
@@ -163,6 +165,7 @@ instance APatternLike a (Pattern' a) where
       A.LitP{}             -> return p
       A.AbsurdP{}          -> return p
       A.ProjP{}            -> return p
+      A.EqualP{}           -> return p
       -- Recursive cases:
       A.ConP        i ds ps -> A.ConP        i ds <$> traverseAPatternM pre post ps
       A.DefP        i q  ps -> A.DefP        i q  <$> traverseAPatternM pre post ps
@@ -211,6 +214,7 @@ patternVars p = foldAPattern f p `appEndo` []
     A.WildP       {} -> mempty
     A.DotP        {} -> mempty
     A.AbsurdP     {} -> mempty
+    A.EqualP      {} -> mempty
     A.PatternSynP {} -> mempty
     A.WithP _ _      -> mempty
 
@@ -267,8 +271,9 @@ substPattern'
   -> Pattern' e            -- ^ Input pattern.
   -> Pattern' e
 substPattern' subE s = mapAPattern $ \ p -> case p of
-  VarP (BindName x) -> fromMaybe p $ lookup x s
+  VarP x            -> fromMaybe p $ lookup (A.unBind x) s
   DotP i e          -> DotP i $ subE e
+  EqualP i es       -> EqualP i $ map (subE *** subE) es
   -- No action on the other patterns (besides the recursion):
   ConP _ _ _        -> p
   RecP _ _          -> p
@@ -293,9 +298,7 @@ instance IsWithP (Pattern' e) where
 
 -- | Split patterns into (patterns, trailing with-patterns).
 splitOffTrailingWithPatterns :: A.Patterns -> (A.Patterns, A.Patterns)
-splitOffTrailingWithPatterns ps = (reverse rps, reverse rwps)
-  where
-  (rwps, rps) = span (isJust . isWithP) $ reverse ps
+splitOffTrailingWithPatterns = spanEnd (isJust . isWithP)
 
 -- | Get the tail of with-patterns of a pattern spine.
 trailingWithPatterns :: Patterns -> Patterns
