@@ -2,6 +2,7 @@
 {-# LANGUAGE DeriveDataTypeable         #-}
 {-# LANGUAGE DeriveGeneric              #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE UndecidableInstances       #-} -- for: LensNamed name (Arg a)
 
 {-| Some common syntactic entities are defined in this module.
 -}
@@ -45,6 +46,21 @@ data Delayed = Delayed | NotDelayed
 instance KillRange Delayed where
   killRange = id
 
+---------------------------------------------------------------------------
+-- * File
+---------------------------------------------------------------------------
+
+data FileType = AgdaFileType | MdFileType | RstFileType | TexFileType | OrgFileType
+  deriving (Data, Eq, Ord, Show)
+
+instance Pretty FileType where
+  pretty = \case
+    AgdaFileType -> "Agda"
+    MdFileType   -> "Markdown"
+    RstFileType  -> "ReStructedText"
+    TexFileType  -> "LaTeX"
+    OrgFileType  -> "org-mode"
+
 ---------------------------------------------------------------------------
 -- * Eta-equality
 ---------------------------------------------------------------------------
@@ -67,11 +83,11 @@ instance NFData HasEta where
 ---------------------------------------------------------------------------
 
 data Induction = Inductive | CoInductive
-  deriving (Data, Eq, Ord)
+  deriving (Data, Eq, Ord, Show)
 
-instance Show Induction where
-  show Inductive   = "inductive"
-  show CoInductive = "coinductive"
+instance Pretty Induction where
+  pretty Inductive   = "inductive"
+  pretty CoInductive = "coinductive"
 
 instance HasRange Induction where
   getRange _ = noRange
@@ -93,6 +109,12 @@ data Overlappable = YesOverlap | NoOverlap
 data Hiding  = Hidden | Instance Overlappable | NotHidden
   deriving (Data, Show, Eq, Ord)
 
+instance Pretty Hiding where
+  pretty = \case
+    Hidden     -> "hidden"
+    NotHidden  -> "visible"
+    Instance{} -> "instance"
+
 -- | Just for the 'Hiding' instance. Should never combine different
 --   overlapping.
 instance Semigroup Overlappable where
@@ -241,6 +263,7 @@ data Modality = Modality
   , modQuantity  :: Quantity
       -- ^ Cardinality / runtime erasure.
       --   See Conor McBride, I got plenty o' nutting, Wadlerfest 2016.
+      --   See Bob Atkey, Syntax and Semantics of Quantitative Type Theory, LiCS 2018.
   } deriving (Data, Eq, Ord, Show, Generic)
 
 defaultModality :: Modality
@@ -262,6 +285,44 @@ instance PartialOrd Modality where
 instance POSemigroup Modality where
 instance POMonoid Modality where
 
+instance LeftClosedPOMonoid Modality where
+  inverseCompose = inverseComposeModality
+
+-- | @m `moreUsableModality` m'@ means that an @m@ can be used
+--   where ever an @m'@ is required.
+
+moreUsableModality :: Modality -> Modality -> Bool
+moreUsableModality m m' = related m POLE m'
+
+usableModality :: LensModality a => a -> Bool
+usableModality a = usableRelevance m && usableQuantity m
+  where m = getModality a
+
+composeModality :: Modality -> Modality -> Modality
+composeModality = (<>)
+
+-- | Compose with modality flag from the left.
+--   This function is e.g. used to update the modality information
+--   on pattern variables @a@ after a match against something of modality @q@.
+applyModality :: LensModality a => Modality -> a -> a
+applyModality m = mapModality (m `composeModality`)
+
+-- | @inverseComposeModality r x@ returns the least modality @y@
+--   such that forall @x@, @y@ we have
+--   @x \`moreUsableModality\` (r \`composeModality\` y)@
+--   iff
+--   @(r \`inverseComposeModality\` x) \`moreUsableModality\` y@ (Galois connection).
+inverseComposeModality :: Modality -> Modality -> Modality
+inverseComposeModality (Modality r q) (Modality r' q') =
+  Modality (r `inverseComposeRelevance` r')
+           (q `inverseComposeQuantity`  q')
+
+-- | Left division by a 'Modality'.
+--   Used e.g. to modify context when going into a @m@ argument.
+inverseApplyModality :: LensModality a => Modality -> a -> a
+inverseApplyModality m = mapModality (m `inverseComposeModality`)
+
+
 -- boilerplate instances
 
 instance KillRange Modality where
@@ -269,6 +330,14 @@ instance KillRange Modality where
 
 instance NFData Modality where
 
+-- Lens stuff
+
+lModRelevance :: Lens' Relevance Modality
+lModRelevance f m = f (modRelevance m) <&> \ r -> m { modRelevance = r }
+
+lModQuantity :: Lens' Quantity Modality
+lModQuantity f m = f (modQuantity m) <&> \ q -> m { modQuantity = q }
+
 class LensModality a where
 
   getModality :: a -> Modality
@@ -321,12 +390,18 @@ mapQuantityMod = mapModality . mapQuantity
 ---------------------------------------------------------------------------
 
 -- | Quantity for linearity.
+--
+--   A quantity is a set of natural numbers, indicating possible semantic
+--   uses of a variable.  A singleton set @{n}@ requires that the
+--   corresponding variable is used exactly @n@ times.
+--
 data Quantity
-  = Quantity0  -- ^ Zero uses, erased at runtime.
-  -- TODO: | Quantity1  -- ^ Linear use (could be updated destructively).
-  -- (needs postponable constraints between quantities to compute uses).
-  | Quantityω  -- ^ Unrestricted use.
-  deriving (Data, Show, Generic, Eq, Enum, Bounded)
+  = Quantity0  -- ^ Zero uses @{0}@, erased at runtime.
+  | Quantity1  -- ^ Linear use @{1}@ (could be updated destructively).
+    -- Mostly TODO (needs postponable constraints between quantities to compute uses).
+  | Quantityω  -- ^ Unrestricted use @ℕ@.
+  deriving (Data, Show, Generic, Eq, Enum, Bounded, Ord)
+    -- @Ord@ instance in case @Quantity@ is used in keys for maps etc.
 
 defaultQuantity :: Quantity
 defaultQuantity = Quantityω
@@ -334,31 +409,77 @@ defaultQuantity = Quantityω
 -- | Composition of quantities (multiplication).
 --
 -- 'Quantity0' is dominant.
+-- 'Quantity1' is neutral.
+--
 instance Semigroup Quantity where
+  Quantity1 <> q = q
+  q <> Quantity1 = q
   Quantity0 <> _ = Quantity0
   _ <> Quantity0 = Quantity0
   Quantityω <> _ = Quantityω
   -- _ <> Quantityω = Quantityω  -- redundant
 
 -- | In the absense of finite quantities besides 0, ω is the unit.
+--   Otherwise, 1 is the unit.
 instance Monoid Quantity where
-  mempty = Quantityω
+  mempty  = Quantity1
   mappend = (<>)
 
--- | Note that the order is @ω ≤ 0@, more relevant is smaller.
-instance Ord Quantity where
-  compare = curry $ \case
-    (Quantityω, Quantityω) -> EQ
-    (Quantityω, Quantity0) -> LT
-    (Quantity0, Quantityω) -> GT
-    (Quantity0, Quantity0) -> EQ
-
+-- | Note that the order is @ω ≤ 0,1@, more options is smaller.
 instance PartialOrd Quantity where
-  comparable = comparableOrd
+  comparable = curry $ \case
+    (q, q') | q == q' -> POEQ
+    -- ω is least
+    (Quantityω, _)    -> POLT
+    (_, Quantityω)    -> POGT
+    -- others are uncomparable
+    _ -> POAny
 
 instance POSemigroup Quantity where
 instance POMonoid Quantity where
 
+instance LeftClosedPOMonoid Quantity where
+  inverseCompose = inverseComposeQuantity
+
+-- | @m `moreUsableQuantity` m'@ means that an @m@ can be used
+--   where ever an @m'@ is required.
+
+moreQuantity :: Quantity -> Quantity -> Bool
+moreQuantity m m' = related m POLE m'
+
+-- | A thing of quantity 0 is unusable, all others are usable.
+
+usableQuantity :: LensQuantity a => a -> Bool
+usableQuantity a = getQuantity a /= Quantity0
+
+composeQuantity :: Quantity -> Quantity -> Quantity
+composeQuantity = (<>)
+
+-- | Compose with quantity flag from the left.
+--   This function is e.g. used to update the quantity information
+--   on pattern variables @a@ after a match against something of quantity @q@.
+applyQuantity :: LensQuantity a => Quantity -> a -> a
+applyQuantity q = mapQuantity (q `composeQuantity`)
+
+-- | @inverseComposeQuantity r x@ returns the least quantity @y@
+--   such that forall @x@, @y@ we have
+--   @x \`moreQuantity\` (r \`composeQuantity\` y)@
+--   iff
+--   @(r \`inverseComposeQuantity\` x) \`moreQuantity\` y@ (Galois connection).
+inverseComposeQuantity :: Quantity -> Quantity -> Quantity
+inverseComposeQuantity q x =
+  case (q, x) of
+    (Quantity1 , x)          -> x          -- going to linear arg: nothing changes
+    (Quantity0 , x)          -> Quantityω  -- going to erased arg: every thing usable
+    (Quantityω , Quantityω)  -> Quantityω
+    (Quantityω , _)          -> Quantity0  -- linear resources are unusable as arguments to unrestricted functions
+
+-- | Left division by a 'Quantity'.
+--   Used e.g. to modify context when going into a @q@ argument.
+inverseApplyQuantity :: LensQuantity a => Quantity -> a -> a
+inverseApplyQuantity q = mapQuantity (q `inverseComposeQuantity`)
+
+
 -- boilerplate instances
 
 class LensQuantity a where
@@ -381,6 +502,7 @@ instance KillRange Quantity where
 
 instance NFData Quantity where
   rnf Quantity0 = ()
+  rnf Quantity1 = ()
   rnf Quantityω = ()
 
 ---------------------------------------------------------------------------
@@ -461,12 +583,12 @@ instance Ord Relevance where
 instance PartialOrd Relevance where
   comparable = comparableOrd
 
--- | @unusableRelevance rel == True@ iff we cannot use a variable of @rel@.
-unusableRelevance :: LensRelevance a => a -> Bool
-unusableRelevance a = case getRelevance a of
-  Irrelevant -> True
-  NonStrict  -> True
-  Relevant   -> False
+-- | @usableRelevance rel == False@ iff we cannot use a variable of @rel@.
+usableRelevance :: LensRelevance a => a -> Bool
+usableRelevance a = case getRelevance a of
+  Irrelevant -> False
+  NonStrict  -> False
+  Relevant   -> True
 
 -- | 'Relevance' composition.
 --   'Irrelevant' is dominant, 'Relevant' is neutral.
@@ -479,6 +601,12 @@ composeRelevance r r' =
     (_, NonStrict)  -> NonStrict
     (Relevant, Relevant) -> Relevant
 
+-- | Compose with relevance flag from the left.
+--   This function is e.g. used to update the relevance information
+--   on pattern variables @a@ after a match against something @rel@.
+applyRelevance :: LensRelevance a => Relevance -> a -> a
+applyRelevance rel = mapRelevance (rel `composeRelevance`)
+
 -- | @inverseComposeRelevance r x@ returns the most irrelevant @y@
 --   such that forall @x@, @y@ we have
 --   @x \`moreRelevant\` (r \`composeRelevance\` y)@
@@ -487,11 +615,16 @@ composeRelevance r r' =
 inverseComposeRelevance :: Relevance -> Relevance -> Relevance
 inverseComposeRelevance r x =
   case (r, x) of
-    (Relevant, x)        -> x          -- going to relevant arg.: nothing changes
-    _ | r == x           -> Relevant   -- because Relevant is comp.-neutral
-    (Irrelevant, x)      -> Relevant   -- going irrelevant: every thing usable
-    (_, Irrelevant)      -> Irrelevant -- otherwise: irrelevant things remain unusable
-    (NonStrict, _)       -> Relevant   -- but @NonStrict@s become usable
+    (Relevant  , x)          -> x          -- going to relevant arg.: nothing changes
+                                           -- because Relevant is comp.-neutral
+    (Irrelevant, x)          -> Relevant   -- going irrelevant: every thing usable
+    (NonStrict , Irrelevant) -> Irrelevant -- otherwise: irrelevant things remain unusable
+    (NonStrict , _)          -> Relevant   -- but @NonStrict@s become usable
+
+-- | Left division by a 'Relevance'.
+--   Used e.g. to modify context when going into a @rel@ argument.
+inverseApplyRelevance :: LensRelevance a => Relevance -> a -> a
+inverseApplyRelevance rel = mapRelevance (rel `inverseComposeRelevance`)
 
 -- | 'Relevance' forms a semigroup under composition.
 instance Semigroup Relevance where
@@ -511,7 +644,6 @@ instance LeftClosedPOMonoid Relevance where
 -- | Irrelevant function arguments may appear non-strictly in the codomain type.
 irrToNonStrict :: Relevance -> Relevance
 irrToNonStrict Irrelevant = NonStrict
--- irrToNonStrict NonStrict  = Relevant -- TODO: this is bad if we apply irrToNonStrict several times!
 irrToNonStrict rel        = rel
 
 -- | Applied when working on types (unless --experimental-irrelevance).
@@ -533,6 +665,7 @@ data Origin
   | Inserted     -- ^ E.g. inserted hidden arguments.
   | Reflected    -- ^ Produced by the reflection machinery.
   | CaseSplit    -- ^ Produced by an interactive case split.
+  | Substitution -- ^ Named application produced to represent a substitution. E.g. "?0 (x = n)" instead of "?0 n"
   deriving (Data, Show, Eq, Ord)
 
 instance KillRange Origin where
@@ -543,6 +676,7 @@ instance NFData Origin where
   rnf Inserted = ()
   rnf Reflected = ()
   rnf CaseSplit = ()
+  rnf Substitution = ()
 
 -- | Decorating something with 'Origin' information.
 data WithOrigin a = WithOrigin
@@ -769,7 +903,7 @@ mapFreeVariablesArgInfo = mapArgInfo . mapFreeVariables
 data Arg e  = Arg
   { argInfo :: ArgInfo
   , unArg :: e
-  } deriving (Data, Ord, Functor, Foldable, Traversable)
+  } deriving (Data, Ord, Show, Functor, Foldable, Traversable)
 
 instance Decoration Arg where
   traverseF f (Arg ai a) = Arg ai <$> f a
@@ -794,28 +928,56 @@ instance Eq a => Eq (Arg a) where
     -- and more generally not use it on Syntax.
     -- Andrea: except for caching.
 
-instance Show a => Show (Arg a) where
-    show (Arg (ArgInfo h (Modality r q) o fv) a) = showFVs fv $ showQ q $ showR r $ showO o $ showH h $ show a
-      where
-        showH Hidden       s = "{" ++ s ++ "}"
-        showH NotHidden    s = "(" ++ s ++ ")"
-        showH (Instance o) s = showOv o ++ "{{" ++ s ++ "}}"
-          where showOv YesOverlap = "overlap "
-                showOv NoOverlap  = ""
-        showR r s = case r of
-          Irrelevant   -> "." ++ s
-          NonStrict    -> "?" ++ s
-          Relevant     -> "r" ++ s -- Andreas: I want to see it explicitly
-        showQ q s = case q of
-          Quantity0   -> "0" ++ s
-          Quantityω   -> "ω" ++ s
-        showO o s = case o of
-          UserWritten -> "u" ++ s
-          Inserted    -> "i" ++ s
-          Reflected   -> "g" ++ s -- generated by reflection
-          CaseSplit   -> "c" ++ s -- generated by case split
-        showFVs UnknownFVs    s = s
-        showFVs (KnownFVs fv) s = "fv" ++ show (IntSet.toList fv) ++ s
+-- instance Show a => Show (Arg a) where
+--     show (Arg (ArgInfo h (Modality r q) o fv) a) = showFVs fv $ showQ q $ showR r $ showO o $ showH h $ show a
+--       where
+--         showH Hidden       s = "{" ++ s ++ "}"
+--         showH NotHidden    s = "(" ++ s ++ ")"
+--         showH (Instance o) s = showOv o ++ "{{" ++ s ++ "}}"
+--           where showOv YesOverlap = "overlap "
+--                 showOv NoOverlap  = ""
+--         showR r s = case r of
+--           Irrelevant   -> "." ++ s
+--           NonStrict    -> "?" ++ s
+--           Relevant     -> "r" ++ s -- Andreas: I want to see it explicitly
+--         showQ q s = case q of
+--           Quantity0   -> "0" ++ s
+--           Quantity1   -> "1" ++ s
+--           Quantityω   -> "ω" ++ s
+--         showO o s = case o of
+--           UserWritten -> "u" ++ s
+--           Inserted    -> "i" ++ s
+--           Reflected   -> "g" ++ s -- generated by reflection
+--           CaseSplit   -> "c" ++ s -- generated by case split
+--           Substitution -> "s" ++ s
+--         showFVs UnknownFVs    s = s
+--         showFVs (KnownFVs fv) s = "fv" ++ show (IntSet.toList fv) ++ s
+
+-- -- defined in Concrete.Pretty
+-- instance Pretty a => Pretty (Arg a) where
+--     pretty (Arg (ArgInfo h (Modality r q) o fv) a) = prettyFVs fv $ prettyQ q $ prettyR r $ prettyO o $ prettyH h $ pretty a
+--       where
+--         prettyH Hidden       s = "{" <> s <> "}"
+--         prettyH NotHidden    s = "(" <> s <> ")"
+--         prettyH (Instance o) s = prettyOv o <> "{{" <> s <> "}}"
+--           where prettyOv YesOverlap = "overlap "
+--                 prettyOv NoOverlap  = ""
+--         prettyR r s = case r of
+--           Irrelevant   -> "." <> s
+--           NonStrict    -> "?" <> s
+--           Relevant     -> "r" <> s -- Andreas: I want to see it explicitly
+--         prettyQ q s = case q of
+--           Quantity0   -> "0" <> s
+--           Quantity1   -> "1" <> s
+--           Quantityω   -> "ω" <> s
+--         prettyO o s = case o of
+--           UserWritten -> "u" <> s
+--           Inserted    -> "i" <> s
+--           Reflected   -> "g" <> s -- generated by reflection
+--           CaseSplit   -> "c" <> s -- generated by case split
+--           Substitution -> "s" <> s
+--         prettyFVs UnknownFVs    s = s
+--         prettyFVs (KnownFVs fv) s = "fv" <> pretty (IntSet.toList fv) <> s
 
 instance NFData e => NFData (Arg e) where
   rnf (Arg a b) = rnf a `seq` rnf b
@@ -905,27 +1067,34 @@ instance Underscore Doc where
 --   @Dom@ is used in 'Pi' of internal syntax, in 'Context' and 'Telescope'.
 --   'Arg' is used for actual arguments ('Var', 'Con', 'Def' etc.)
 --   and in 'Abstract' syntax and other situations.
+--
+--   [ cubical ] When @domFinite = True@ for the domain of a 'Pi'
+--   type, the elements should be compared by tabulating the domain type.
+--   Only supported in case the domain type is primIsOne, to obtain
+--   the correct equality for partial elements.
 data Dom e = Dom
   { domInfo   :: ArgInfo
+  , domFinite :: !Bool
+  , domName   :: Maybe RString
   , unDom     :: e
-  } deriving (Data, Ord, Functor, Foldable, Traversable)
+  } deriving (Data, Ord, Show, Functor, Foldable, Traversable)
 
 instance Decoration Dom where
-  traverseF f (Dom ai a) = Dom ai <$> f a
+  traverseF f (Dom ai b x a) = Dom ai b x <$> f a
 
 instance HasRange a => HasRange (Dom a) where
   getRange = getRange . unDom
 
 instance KillRange a => KillRange (Dom a) where
-  killRange (Dom info a) = killRange2 Dom info a
+  killRange (Dom info b x a) = killRange4 Dom info b x a
 
 -- | Ignores 'Origin' and 'FreeVariables'.
 instance Eq a => Eq (Dom a) where
-  Dom (ArgInfo h1 m1 _ _) x1 == Dom (ArgInfo h2 m2 _ _) x2 =
-    (h1, m1, x1) == (h2, m2, x2)
+  Dom (ArgInfo h1 m1 _ _) b1 s1 x1 == Dom (ArgInfo h2 m2 _ _) b2 s2 x2 =
+    (h1, m1, b1, s1, x1) == (h2, m2, b2, s2, x2)
 
-instance Show a => Show (Dom a) where
-  show = show . argFromDom
+-- instance Show a => Show (Dom a) where
+--   show = show . argFromDom
 
 instance LensArgInfo (Dom e) where
   getArgInfo        = domInfo
@@ -967,13 +1136,25 @@ instance LensQuantity (Dom e) where
   mapQuantity = mapQuantityMod
 
 argFromDom :: Dom a -> Arg a
-argFromDom (Dom i a) = Arg i a
+argFromDom (Dom i _ _ a) = Arg i a
+
+namedArgFromDom :: Dom a -> NamedArg a
+namedArgFromDom (Dom i _ s a) = Arg i $ Named s a
 
 domFromArg :: Arg a -> Dom a
-domFromArg (Arg i a) = Dom i a
+domFromArg (Arg i a) = Dom i False Nothing a
+
+domFromNamedArg :: NamedArg a -> Dom a
+domFromNamedArg (Arg i a) = Dom i False (nameOf a) (namedThing a)
 
 defaultDom :: a -> Dom a
-defaultDom = Dom defaultArgInfo
+defaultDom = defaultArgDom defaultArgInfo
+
+defaultArgDom :: ArgInfo -> a -> Dom a
+defaultArgDom info x = Dom info False Nothing x
+
+defaultNamedArgDom :: ArgInfo -> String -> a -> Dom a
+defaultNamedArgDom info s = Dom info False (Just $ unranged s)
 
 ---------------------------------------------------------------------------
 -- * Named arguments
@@ -984,7 +1165,7 @@ data Named name a =
     Named { nameOf     :: Maybe name
           , namedThing :: a
           }
-    deriving (Eq, Ord, Data, Functor, Foldable, Traversable)
+    deriving (Eq, Ord, Show, Data, Functor, Foldable, Traversable)
 
 -- | Standard naming.
 type Named_ = Named RString
@@ -995,6 +1176,30 @@ unnamed = Named Nothing
 named :: name -> a -> Named name a
 named = Named . Just
 
+-- | Accessor/editor for the 'nameOf' component.
+class LensNamed name a | a -> name where
+  lensNamed :: Lens' (Maybe name) a
+
+instance LensNamed name (Named name a) where
+  lensNamed f (Named mn a) = f mn <&> \ mn' -> Named mn' a
+
+getNameOf :: LensNamed name a => a -> Maybe name
+getNameOf a = a ^. lensNamed
+
+setNameOf :: LensNamed name a => Maybe name -> a -> a
+setNameOf = set lensNamed
+
+mapNameOf :: LensNamed name a => (Maybe name -> Maybe name) -> a -> a
+mapNameOf = over lensNamed
+
+-- Lenses lift through decorations:
+-- instance (Decoration f, LensNamed name a) => LensNamed name (f a) where
+
+instance LensNamed name a => LensNamed name (Arg a) where
+  lensNamed = traverseF . lensNamed
+
+-- Standard instances for 'Named':
+
 instance Decoration (Named name) where
   traverseF f (Named n a) = Named n <$> f a
 
@@ -1007,9 +1212,14 @@ instance SetRange a => SetRange (Named name a) where
 instance (KillRange name, KillRange a) => KillRange (Named name a) where
   killRange (Named n a) = Named (killRange n) (killRange a)
 
-instance Show a => Show (Named_ a) where
-    show (Named Nothing a)  = show a
-    show (Named (Just n) a) = rawNameToString (rangedThing n) ++ " = " ++ show a
+-- instance Show a => Show (Named_ a) where
+--     show (Named Nothing a)  = show a
+--     show (Named (Just n) a) = rawNameToString (rangedThing n) ++ " = " ++ show a
+
+-- -- Defined in Concrete.Pretty
+-- instance Pretty a => Pretty (Named_ a) where
+--     pretty (Named Nothing a)  = pretty a
+--     pretty (Named (Just n) a) = text (rawNameToString (rangedThing n)) <+> "=" <+> pretty a
 
 instance (NFData name, NFData a) => NFData (Named name a) where
   rnf (Named a b) = rnf a `seq` rnf b
@@ -1022,7 +1232,10 @@ namedArg :: NamedArg a -> a
 namedArg = namedThing . unArg
 
 defaultNamedArg :: a -> NamedArg a
-defaultNamedArg = defaultArg . unnamed
+defaultNamedArg = unnamedArg defaultArgInfo
+
+unnamedArg :: ArgInfo -> a -> NamedArg a
+unnamedArg info = Arg info . unnamed
 
 -- | The functor instance for 'NamedArg' would be ambiguous,
 --   so we give it another name here.
@@ -1042,14 +1255,17 @@ data Ranged a = Ranged
   { rangeOf     :: Range
   , rangedThing :: a
   }
-  deriving (Data, Functor, Foldable, Traversable)
+  deriving (Data, Show, Functor, Foldable, Traversable)
 
 -- | Thing with no range info.
 unranged :: a -> Ranged a
 unranged = Ranged noRange
 
-instance Show a => Show (Ranged a) where
-  show = show . rangedThing
+instance Pretty a => Pretty (Ranged a) where
+  pretty = pretty . rangedThing
+
+-- instance Show a => Show (Ranged a) where
+--   show = show . rangedThing
 
 instance Eq a => Eq (Ranged a) where
   Ranged _ x == Ranged _ y = x == y
@@ -1192,13 +1408,13 @@ type Arity  = Nat
 -- | The unique identifier of a name. Second argument is the top-level module
 --   identifier.
 data NameId = NameId {-# UNPACK #-} !Word64 {-# UNPACK #-} !Word64
-    deriving (Eq, Ord, Data, Generic)
+    deriving (Eq, Ord, Data, Generic, Show)
 
 instance KillRange NameId where
   killRange = id
 
-instance Show NameId where
-  show (NameId n m) = show n ++ "@" ++ show m
+instance Pretty NameId where
+  pretty (NameId n m) = text $ show n ++ "@" ++ show m
 
 instance Enum NameId where
   succ (NameId n m)     = NameId (n + 1) m
@@ -1286,6 +1502,7 @@ instance NFData a => NFData (MaybePlaceholder a) where
 newtype InteractionId = InteractionId { interactionId :: Nat }
     deriving ( Eq
              , Ord
+             , Show
              , Num
              , Integral
              , Real
@@ -1293,8 +1510,8 @@ newtype InteractionId = InteractionId { interactionId :: Nat }
              , Data
              )
 
-instance Show InteractionId where
-    show (InteractionId i) = "?" ++ show i
+instance Pretty InteractionId where
+    pretty (InteractionId i) = text $ "?" ++ show i
 
 instance KillRange InteractionId where killRange = id
 
@@ -1304,53 +1521,58 @@ instance KillRange InteractionId where killRange = id
 
 -- | The things you are allowed to say when you shuffle names between name
 --   spaces (i.e. in @import@, @namespace@, or @open@ declarations).
-data ImportDirective' a b = ImportDirective
+data ImportDirective' n m = ImportDirective
   { importDirRange :: Range
-  , using          :: Using' a b
-  , hiding         :: [ImportedName' a b]
-  , impRenaming    :: [Renaming' a b]
+  , using          :: Using' n m
+  , hiding         :: [ImportedName' n m]
+  , impRenaming    :: [Renaming' n m]
   , publicOpen     :: Bool -- ^ Only for @open@. Exports the opened names from the current module.
   }
   deriving (Data, Eq)
 
-data Using' a b = UseEverything | Using [ImportedName' a b]
+data Using' n m = UseEverything | Using [ImportedName' n m]
   deriving (Data, Eq)
 
-instance Semigroup (Using' a b) where
+instance Semigroup (Using' n m) where
   UseEverything <> u             = u
   u             <> UseEverything = u
   Using xs      <> Using ys      = Using (xs ++ ys)
 
-instance Monoid (Using' a b) where
+instance Monoid (Using' n m) where
   mempty  = UseEverything
   mappend = (<>)
 
 -- | Default is directive is @private@ (use everything, but do not export).
-defaultImportDir :: ImportDirective' a b
+defaultImportDir :: ImportDirective' n m
 defaultImportDir = ImportDirective noRange UseEverything [] [] False
 
-isDefaultImportDir :: ImportDirective' a b -> Bool
+isDefaultImportDir :: ImportDirective' n m -> Bool
 isDefaultImportDir (ImportDirective _ UseEverything [] [] False) = True
 isDefaultImportDir _                                             = False
 
--- | An imported name can be a module or a defined name
-data ImportedName' a b
-  = ImportedModule  b
-  | ImportedName    a
-  deriving (Data, Eq, Ord)
+-- | An imported name can be a module or a defined name.
+data ImportedName' n m
+  = ImportedModule  m  -- ^ Imported module name of type @m@.
+  | ImportedName    n  -- ^ Imported name of type @n@.
+  deriving (Data, Eq, Ord, Show)
 
 setImportedName :: ImportedName' a a -> a -> ImportedName' a a
 setImportedName (ImportedName   x) y = ImportedName   y
 setImportedName (ImportedModule x) y = ImportedModule y
 
-instance (Show a, Show b) => Show (ImportedName' a b) where
-  show (ImportedModule b) = "module " ++ show b
-  show (ImportedName   a) = show a
+-- -- Defined in Concrete.Pretty
+-- instance (Pretty n, Pretty m) => Pretty (ImportedName' n m) where
+--   pretty (ImportedModule x) = "module" <+> pretty x
+--   pretty (ImportedName   x) = pretty x
 
-data Renaming' a b = Renaming
-  { renFrom    :: ImportedName' a b
+-- instance (Show n, Show m) => Show (ImportedName' n m) where
+--   show (ImportedModule x) = "module " ++ show x
+--   show (ImportedName   x) = show x
+
+data Renaming' n m = Renaming
+  { renFrom    :: ImportedName' n m
     -- ^ Rename from this name.
-  , renTo      :: ImportedName' a b
+  , renTo      :: ImportedName' n m
     -- ^ To this one.  Must be same kind as 'renFrom'.
   , renToRange :: Range
     -- ^ The range of the \"to\" keyword.  Retained for highlighting purposes.
@@ -1445,3 +1667,14 @@ instance NFData a => NFData (TerminationCheck a) where
 
 -- | Positivity check? (Default = True).
 type PositivityCheck = Bool
+
+-----------------------------------------------------------------------------
+-- * Universe checking
+-----------------------------------------------------------------------------
+
+-- | Universe check? (Default is yes).
+data UniverseCheck = YesUniverseCheck | NoUniverseCheck
+  deriving (Eq, Ord, Show, Bounded, Enum, Data)
+
+instance KillRange UniverseCheck where
+  killRange = id
