@@ -3,6 +3,7 @@
 {-# LANGUAGE DeriveFunctor      #-}
 {-# LANGUAGE DeriveFoldable     #-}
 {-# LANGUAGE DeriveTraversable  #-}
+{-# LANGUAGE LambdaCase         #-}
 
 {-| The concrete syntax is a raw representation of the program text
     without any desugaring at all.  This is what the parser produces.
@@ -15,11 +16,10 @@ module Agda.Syntax.Concrete
   , OpApp(..), fromOrdinary
   , module Agda.Syntax.Concrete.Name
   , appView, AppView(..)
+  , isSingleIdentifierP, removeSingletonRawAppP
     -- * Bindings
   , LamBinding
   , LamBinding'(..)
-  , TypedBindings
-  , TypedBindings'(..)
   , TypedBinding
   , TypedBinding'(..)
   , RecordAssignment
@@ -29,6 +29,8 @@ module Agda.Syntax.Concrete
   , BoundName(..), mkBoundName_, mkBoundName
   , Telescope -- (..)
   , countTelVars
+  , lamBindingsToTelescope
+  , makePi
     -- * Declarations
   , Declaration(..)
   , ModuleApplication(..)
@@ -36,7 +38,7 @@ module Agda.Syntax.Concrete
   , TypeSignatureOrInstanceBlock
   , ImportDirective, Using, ImportedName
   , Renaming
-  , AsName(..)
+  , AsName'(..), AsName
   , OpenShortHand(..), RewriteEqn, WithExpr
   , LHS(..), Pattern(..), LHSCore(..)
   , LamClause(..)
@@ -142,11 +144,12 @@ data Expr
   | Lam Range [LamBinding] Expr                -- ^ ex: @\\x {y} -> e@ or @\\(x:A){y:B} -> e@
   | AbsurdLam Range Hiding                     -- ^ ex: @\\ ()@
   | ExtendedLam Range [LamClause]              -- ^ ex: @\\ { p11 .. p1a -> e1 ; .. ; pn1 .. pnz -> en }@
-  | Fun Range Expr Expr                        -- ^ ex: @e -> e@ or @.e -> e@ (NYI: @{e} -> e@)
+  | Fun Range (Arg Expr) Expr                  -- ^ ex: @e -> e@ or @.e -> e@ (NYI: @{e} -> e@)
   | Pi Telescope Expr                          -- ^ ex: @(xs:e) -> e@ or @{xs:e} -> e@
   | Set Range                                  -- ^ ex: @Set@
   | Prop Range                                 -- ^ ex: @Prop@
   | SetN Range Integer                         -- ^ ex: @Set0, Set1, ..@
+  | PropN Range Integer                        -- ^ ex: @Prop0, Prop1, ..@
   | Rec Range RecordAssignments                -- ^ ex: @record {x = a; y = b}@, or @record { x = a; M1; M2 }@
   | RecUpdate Range Expr [FieldAssignment]     -- ^ ex: @record e {x = a; y = b}@
   | Let Range [Declaration] (Maybe Expr)       -- ^ ex: @let Ds in e@, missing body when parsing do-notation let
@@ -166,6 +169,7 @@ data Expr
   | DontCare Expr                              -- ^ to print irrelevant things
   | Equal Range Expr Expr                      -- ^ ex: @a = b@, used internally in the parser
   | Ellipsis Range                             -- ^ @...@, used internally to parse patterns.
+  | Generalized Expr
   deriving Data
 
 -- | Concrete patterns. No literals in patterns at the moment.
@@ -189,6 +193,7 @@ data Pattern
   | DotP Range Expr                        -- ^ @.e@
   | LitP Literal                           -- ^ @0@, @1@, etc.
   | RecP Range [FieldAssignment' Pattern]  -- ^ @record {x = p; y = q}@
+  | EqualP Range [(Expr,Expr)]             -- ^ @i = i1@ i.e. cubical face lattice generator
   | EllipsisP Range                        -- ^ @...@, only as left-most pattern.
   | WithP Range Pattern                    -- ^ @| p@, for with-patterns.
   deriving Data
@@ -200,28 +205,14 @@ data DoStmt
   deriving Data
 
 -- | A lambda binding is either domain free or typed.
-type LamBinding = LamBinding' TypedBindings
+type LamBinding = LamBinding' TypedBinding
 data LamBinding' a
-  = DomainFree ArgInfo BoundName  -- ^ . @x@ or @{x}@ or @.x@ or @.{x}@ or @{.x}@
-  | DomainFull a                  -- ^ . @(xs : e)@ or @{xs : e}@
-  deriving (Data, Functor, Foldable, Traversable)
-
-
--- | A sequence of typed bindings with hiding information. Appears in dependent
---   function spaces, typed lambdas, and telescopes.
---
---   If the individual binding contains hiding information as well, the
---   'Hiding' in @TypedBindings@ must be the unit 'NotHidden'.
-
-type TypedBindings = TypedBindings' TypedBinding
-
-data TypedBindings' a = TypedBindings Range (Arg a)
-     -- ^ . @(xs : e)@ or @{xs : e}@ or something like @(x {y} _ : e)@.
+  = DomainFree (NamedArg BoundName) -- ^ . @x@ or @{x}@ or @.x@ or @.{x}@ or @{.x}@
+  | DomainFull a                    -- ^ . @(xs : e)@ or @{xs : e}@
   deriving (Data, Functor, Foldable, Traversable)
 
 data BoundName = BName
   { boundName   :: Name
-  , boundLabel  :: Name    -- ^ for implicit function types the label matters and can't be alpha-renamed
   , bnameFixity :: Fixity'
   }
   deriving (Data, Eq, Show)
@@ -230,27 +221,41 @@ mkBoundName_ :: Name -> BoundName
 mkBoundName_ x = mkBoundName x noFixity'
 
 mkBoundName :: Name -> Fixity' -> BoundName
-mkBoundName x f = BName x x f
+mkBoundName x f = BName x f
 
 -- | A typed binding.
 
 type TypedBinding = TypedBinding' Expr
 
 data TypedBinding' e
-  = TBind Range [WithHiding BoundName] e  -- ^ Binding @(x1 ... xn : A)@.
-  | TLet  Range [Declaration]  -- ^ Let binding @(let Ds)@ or @(open M args)@.
+  = TBind Range [NamedArg BoundName] e  -- ^ Binding @(x1 ... xn : A)@.
+  | TLet  Range [Declaration]           -- ^ Let binding @(let Ds)@ or @(open M args)@.
   deriving (Data, Functor, Foldable, Traversable)
 
 -- | A telescope is a sequence of typed bindings. Bound variables are in scope
 --   in later types.
-type Telescope = [TypedBindings]
+type Telescope = [TypedBinding]
 
 countTelVars :: Telescope -> Nat
 countTelVars tel =
-  sum [ case unArg b of
+  sum [ case b of
           TBind _ xs _ -> genericLength xs
           TLet{}       -> 0
-      | TypedBindings _ b <- tel ]
+      | b <- tel ]
+
+-- | We can try to get a @Telescope@ from a @[LamBinding]@.
+--   If we have a type annotation already, we're happy.
+--   Otherwise we manufacture a binder with an underscore for the type.
+lamBindingsToTelescope :: Range -> [LamBinding] -> Telescope
+lamBindingsToTelescope r = map $ \case
+  DomainFull ty -> ty
+  DomainFree nm -> TBind r [nm] $ Underscore r Nothing
+
+-- | Smart constructor for @Pi@: check whether the @Telescope@ is empty
+
+makePi :: Telescope -> Expr -> Expr
+makePi [] e = e
+makePi bs e = Pi bs e
 
 {-| Left hand sides can be written in infix style. For example:
 
@@ -321,13 +326,18 @@ type Renaming        = Renaming'        Name Name
 -- | An imported name can be a module or a defined name.
 type ImportedName = ImportedName' Name Name
 
-data AsName = AsName
-  { asName  :: Name
+-- | The content of the @as@-clause of the import statement.
+data AsName' a = AsName
+  { asName  :: a
     -- ^ The \"as\" name.
   , asRange :: Range
     -- ^ The range of the \"as\" keyword.  Retained for highlighting purposes.
   }
-  deriving (Data, Show)
+  deriving (Data, Show, Functor, Foldable, Traversable)
+
+-- | From the parser, we get an expression for the @as@-'Name', which
+--   we have to parse into a 'Name'.
+type AsName = AsName' (Either Expr Name)
 
 {--------------------------------------------------------------------------
     Declarations
@@ -346,17 +356,20 @@ type TypeSignatureOrInstanceBlock = Declaration
 data Declaration
   = TypeSig ArgInfo Name Expr
   -- ^ Axioms and functions can be irrelevant. (Hiding should be NotHidden)
+  | Generalize Range [TypeSignature] -- ^ Variables to be generalized, can be hidden and/or irrelevant.
   | Field IsInstance Name (Arg Expr) -- ^ Record field, can be hidden and/or irrelevant.
   | FunClause LHS RHS WhereClause Bool
   | DataSig     Range Induction Name [LamBinding] Expr -- ^ lone data signature in mutual block
-  | Data        Range Induction Name [LamBinding] (Maybe Expr) [TypeSignatureOrInstanceBlock]
+  | Data        Range Induction Name [LamBinding] Expr [TypeSignatureOrInstanceBlock]
+  | DataDef     Range Induction Name [LamBinding] [TypeSignatureOrInstanceBlock]
   | RecordSig   Range Name [LamBinding] Expr -- ^ lone record signature in mutual block
-  | Record      Range Name (Maybe (Ranged Induction)) (Maybe HasEta) (Maybe (Name, IsInstance)) [LamBinding] (Maybe Expr) [Declaration]
+  | RecordDef   Range Name (Maybe (Ranged Induction)) (Maybe HasEta) (Maybe (Name, IsInstance)) [LamBinding] [Declaration]
+  | Record      Range Name (Maybe (Ranged Induction)) (Maybe HasEta) (Maybe (Name, IsInstance)) [LamBinding] Expr [Declaration]
     -- ^ The optional name is a name for the record constructor.
   | Infix Fixity [Name]
   | Syntax      Name Notation -- ^ notation declaration for a name
   | PatternSyn  Range Name [Arg Name] Pattern
-  | Mutual      Range [Declaration]
+  | Mutual      Range [Declaration]  -- @Range@ of the whole @mutual@ block.
   | Abstract    Range [Declaration]
   | Private     Range Origin [Declaration]
     -- ^ In "Agda.Syntax.Concrete.Definitions" we generate private blocks
@@ -369,16 +382,16 @@ data Declaration
   | Open        Range QName ImportDirective
   | Import      Range QName (Maybe AsName) !OpenShortHand ImportDirective
   | ModuleMacro Range  Name ModuleApplication !OpenShortHand ImportDirective
-  | Module      Range QName [TypedBindings] [Declaration]
+  | Module      Range QName Telescope [Declaration]
   | UnquoteDecl Range [Name] Expr
   | UnquoteDef  Range [Name] Expr
   | Pragma      Pragma
   deriving Data
 
 data ModuleApplication
-  = SectionApp Range [TypedBindings] Expr
+  = SectionApp Range Telescope Expr
     -- ^ @tel. M args@
-  | RecordModuleIFS Range QName
+  | RecordModuleInstance Range QName
     -- ^ @M {{...}}@
   deriving Data
 
@@ -391,39 +404,34 @@ data Pragma
   = OptionsPragma             Range [String]
   | BuiltinPragma             Range String QName
   | RewritePragma             Range [QName]
-  | CompiledDataPragma        Range QName String [String]
-  | CompiledTypePragma        Range QName String
-  | CompiledPragma            Range QName String
-  | CompiledExportPragma      Range QName String
-  | CompiledJSPragma          Range QName String
-  | CompiledUHCPragma         Range QName String
-  | CompiledDataUHCPragma     Range QName String [String]
-  | HaskellCodePragma         Range String
   | ForeignPragma             Range String String       -- ^ first string is backend name
   | CompilePragma             Range String QName String -- ^ first string is backend name
   | StaticPragma              Range QName
-  | InjectivePragma           Range QName
   | InlinePragma              Range Bool QName  -- ^ INLINE or NOINLINE
-  | ImportPragma              Range String
-    -- ^ Invariant: The string must be a valid Haskell module name.
-  | ImportUHCPragma           Range String
-    -- ^ same as above, but for the UHC backend
+
   | ImpossiblePragma          Range
     -- ^ Throws an internal error in the scope checker.
   | EtaPragma                 Range QName
     -- ^ For coinductive records, use pragma instead of regular
     --   @eta-equality@ definition (as it is might make Agda loop).
+  | WarningOnUsage            Range QName String
+    -- ^ Applies to the named function
+  | InjectivePragma           Range QName
+    -- ^ Mark a definition as injective for the pattern matching unifier.
+  | DisplayPragma             Range Pattern Expr
+    -- ^ Display lhs as rhs (modifies the printer).
+
+  -- Attached (more or less) pragmas handled in the nicifier (Concrete.Definitions):
+  | CatchallPragma            Range
+    -- ^ Applies to the following function clause.
   | TerminationCheckPragma    Range (TerminationCheck Name)
     -- ^ Applies to the following function (and all that are mutually recursive with it)
     --   or to the functions in the following mutual block.
-  | WarningOnUsage            Range QName String
-    -- ^ Applies to the named function
-  | CatchallPragma            Range
-    -- ^ Applies to the following function clause.
-  | DisplayPragma             Range Pattern Expr
   | NoPositivityCheckPragma   Range
     -- ^ Applies to the following data/record type or mutual block.
   | PolarityPragma            Range Name [Occurrence]
+  | NoUniverseCheckPragma     Range
+    -- ^ Applies to the following data/record type.
   deriving Data
 
 ---------------------------------------------------------------------------
@@ -492,6 +500,18 @@ appView e =
     arg (InstanceArg _ e) = makeInstance $ defaultArg e
     arg e                 = defaultArg (unnamed e)
 
+isSingleIdentifierP :: Pattern -> Maybe Name
+isSingleIdentifierP p = case removeSingletonRawAppP p of
+  IdentP (QName x) -> Just x
+  WildP r          -> Just $ noName r
+  _                -> Nothing
+
+removeSingletonRawAppP :: Pattern -> Pattern
+removeSingletonRawAppP p = case p of
+    RawAppP _ [p'] -> removeSingletonRawAppP p'
+    ParenP _ p'    -> removeSingletonRawAppP p'
+    _ -> p
+
 {--------------------------------------------------------------------------
     Instances
  --------------------------------------------------------------------------}
@@ -511,19 +531,25 @@ instance Null (WhereClause' a) where
 -- Lenses
 ------------------------------------------------------------------------
 
-instance LensRelevance TypedBindings where
-  getRelevance   (TypedBindings _ b) = getRelevance b
-  mapRelevance f (TypedBindings r b) = TypedBindings r $ mapRelevance f b
-
-instance LensHiding TypedBindings where
-  getHiding   (TypedBindings _ b) = getHiding b
-  mapHiding f (TypedBindings r b) = TypedBindings r $ mapHiding f b
-
 instance LensHiding LamBinding where
-  getHiding   (DomainFree ai _) = getHiding ai
-  getHiding   (DomainFull a)    = getHiding a
-  mapHiding f (DomainFree ai x) = DomainFree (mapHiding f ai) x
-  mapHiding f (DomainFull a)    = DomainFull $ mapHiding f a
+  getHiding   (DomainFree x) = getHiding x
+  getHiding   (DomainFull a) = getHiding a
+  mapHiding f (DomainFree x) = DomainFree $ mapHiding f x
+  mapHiding f (DomainFull a) = DomainFull $ mapHiding f a
+
+instance LensHiding TypedBinding where
+  getHiding (TBind _ (x : _) _) = getHiding x   -- Slightly dubious
+  getHiding (TBind _ [] _)      = __IMPOSSIBLE__
+  getHiding TLet{}              = mempty
+  mapHiding f (TBind r xs e) = TBind r ((map . mapHiding) f xs) e
+  mapHiding f b@TLet{}       = b
+
+instance LensRelevance TypedBinding where
+  getRelevance (TBind _ (x : _) _) = getRelevance x   -- Slightly dubious
+  getRelevance (TBind _ [] _)      = __IMPOSSIBLE__
+  getRelevance TLet{}              = mempty
+  mapRelevance f (TBind r xs e) = TBind r ((map . mapRelevance) f xs) e
+  mapRelevance f b@TLet{}       = b
 
 -- HasRange instances
 ------------------------------------------------------------------------
@@ -552,6 +578,7 @@ instance HasRange Expr where
       Set r              -> r
       Prop r             -> r
       SetN r _           -> r
+      PropN r _          -> r
       Let r _ _          -> r
       Paren r _          -> r
       IdiomBrackets r _  -> r
@@ -573,21 +600,19 @@ instance HasRange Expr where
       DontCare{}         -> noRange
       Equal r _ _        -> r
       Ellipsis r         -> r
+      Generalized e      -> getRange e
 
 -- instance HasRange Telescope where
 --     getRange (TeleBind bs) = getRange bs
 --     getRange (TeleFun x y) = fuseRange x y
 
-instance HasRange TypedBindings where
-  getRange (TypedBindings r _) = r
-
 instance HasRange TypedBinding where
   getRange (TBind r _ _) = r
   getRange (TLet r _)    = r
 
 instance HasRange LamBinding where
-  getRange (DomainFree _ x) = getRange x
-  getRange (DomainFull b)   = getRange b
+  getRange (DomainFree x) = getRange x
+  getRange (DomainFull b) = getRange b
 
 instance HasRange BoundName where
   getRange = getRange . boundName
@@ -599,7 +624,7 @@ instance HasRange WhereClause where
 
 instance HasRange ModuleApplication where
   getRange (SectionApp r _ _) = r
-  getRange (RecordModuleIFS r _) = r
+  getRange (RecordModuleInstance r _) = r
 
 instance HasRange a => HasRange (FieldAssignment' a) where
   getRange (FieldAssignment a b) = fuseRange a b
@@ -613,10 +638,13 @@ instance HasRange Declaration where
   getRange (FunClause lhs rhs wh _) = fuseRange lhs rhs `fuseRange` wh
   getRange (DataSig r _ _ _ _)     = r
   getRange (Data r _ _ _ _ _)      = r
+  getRange (DataDef r _ _ _ _)     = r
   getRange (RecordSig r _ _ _)     = r
+  getRange (RecordDef r _ _ _ _ _ _) = r
   getRange (Record r _ _ _ _ _ _ _)  = r
   getRange (Mutual r _)            = r
   getRange (Abstract r _)          = r
+  getRange (Generalize r _)        = r
   getRange (Open r _ _)            = r
   getRange (ModuleMacro r _ _ _ _) = r
   getRange (Import r _ _ _ _)      = r
@@ -657,21 +685,11 @@ instance HasRange Pragma where
   getRange (OptionsPragma r _)               = r
   getRange (BuiltinPragma r _ _)             = r
   getRange (RewritePragma r _)               = r
-  getRange (CompiledDataPragma r _ _ _)      = r
-  getRange (CompiledTypePragma r _ _)        = r
-  getRange (CompiledPragma r _ _)            = r
-  getRange (CompiledExportPragma r _ _)      = r
-  getRange (CompiledJSPragma r _ _)          = r
-  getRange (CompiledUHCPragma r _ _)         = r
-  getRange (CompiledDataUHCPragma r _ _ _)   = r
-  getRange (HaskellCodePragma r _)           = r
   getRange (CompilePragma r _ _ _)           = r
   getRange (ForeignPragma r _ _)             = r
   getRange (StaticPragma r _)                = r
   getRange (InjectivePragma r _)             = r
   getRange (InlinePragma r _ _)              = r
-  getRange (ImportPragma r _)                = r
-  getRange (ImportUHCPragma r _)             = r
   getRange (ImpossiblePragma r)              = r
   getRange (EtaPragma r _)                   = r
   getRange (TerminationCheckPragma r _)      = r
@@ -680,6 +698,7 @@ instance HasRange Pragma where
   getRange (DisplayPragma r _ _)             = r
   getRange (NoPositivityCheckPragma r)       = r
   getRange (PolarityPragma r _ _)            = r
+  getRange (NoUniverseCheckPragma r)         = r
 
 instance HasRange AsName where
   getRange a = getRange (asRange a, asName a)
@@ -699,15 +718,13 @@ instance HasRange Pattern where
   getRange (InstanceP r _)    = r
   getRange (DotP r _)         = r
   getRange (RecP r _)         = r
+  getRange (EqualP r _)       = r
   getRange (EllipsisP r)      = r
   getRange (WithP r _)        = r
 
 -- SetRange instances
 ------------------------------------------------------------------------
 
-instance SetRange TypedBindings where
-  setRange r (TypedBindings _ b) = TypedBindings r b
-
 instance SetRange Pattern where
   setRange r (IdentP x)         = IdentP (setRange r x)
   setRange r (AppP p q)         = AppP (setRange r p) (setRange r q)
@@ -723,9 +740,14 @@ instance SetRange Pattern where
   setRange r (InstanceP _ p)    = InstanceP r p
   setRange r (DotP _ e)         = DotP r e
   setRange r (RecP _ fs)        = RecP r fs
+  setRange r (EqualP _ es)      = EqualP r es
   setRange r (EllipsisP _)      = EllipsisP r
   setRange r (WithP _ p)        = WithP r p
 
+instance SetRange TypedBinding where
+  setRange r (TBind _ xs e) = TBind r xs e
+  setRange r (TLet _ ds)    = TLet r ds
+
 -- KillRange instances
 ------------------------------------------------------------------------
 
@@ -739,16 +761,19 @@ instance KillRange AsName where
   killRange (AsName n _) = killRange1 (flip AsName noRange) n
 
 instance KillRange BoundName where
-  killRange (BName n l f) = killRange3 BName n l f
+  killRange (BName n f) = killRange2 BName n f
 
 instance KillRange Declaration where
   killRange (TypeSig i n e)         = killRange2 (TypeSig i) n e
+  killRange (Generalize r ds )      = killRange1 (Generalize noRange) ds
   killRange (Field i n a)           = killRange2 (Field i) n a
   killRange (FunClause l r w ca)    = killRange4 FunClause l r w ca
   killRange (DataSig _ i n l e)     = killRange4 (DataSig noRange) i n l e
   killRange (Data _ i n l e c)      = killRange4 (Data noRange i) n l e c
+  killRange (DataDef _ i n l c)     = killRange3 (DataDef noRange i) n l c
   killRange (RecordSig _ n l e)     = killRange3 (RecordSig noRange) n l e
-  killRange (Record _ n mi mb mn k e d)= killRange7 (Record noRange) n mi mb mn k e d
+  killRange (RecordDef _ n mi mb mn k d) = killRange6 (RecordDef noRange) n mi mb mn k d
+  killRange (Record _ n mi mb mn k e d)  = killRange7 (Record noRange) n mi mb mn k e d
   killRange (Infix f n)             = killRange2 Infix f n
   killRange (Syntax n no)           = killRange1 (\n -> Syntax n no) n
   killRange (PatternSyn _ n ns p)   = killRange3 (PatternSyn noRange) n ns p
@@ -786,6 +811,7 @@ instance KillRange Expr where
   killRange (Set _)              = Set noRange
   killRange (Prop _)             = Prop noRange
   killRange (SetN _ n)           = SetN noRange n
+  killRange (PropN _ n)          = PropN noRange n
   killRange (Rec _ ne)           = killRange1 (Rec noRange) ne
   killRange (RecUpdate _ e ne)   = killRange2 (RecUpdate noRange) e ne
   killRange (Let _ d e)          = killRange2 (Let noRange) d e
@@ -805,10 +831,11 @@ instance KillRange Expr where
   killRange (DontCare e)         = killRange1 DontCare e
   killRange (Equal _ x y)        = Equal noRange x y
   killRange (Ellipsis _)         = Ellipsis noRange
+  killRange (Generalized e)      = killRange1 Generalized e
 
 instance KillRange LamBinding where
-  killRange (DomainFree i b) = killRange2 DomainFree i b
-  killRange (DomainFull t)   = killRange1 DomainFull t
+  killRange (DomainFree b) = killRange1 DomainFree b
+  killRange (DomainFull t) = killRange1 DomainFull t
 
 instance KillRange LHS where
   killRange (LHS p r w)     = killRange3 LHS p r w
@@ -823,7 +850,7 @@ instance KillRange DoStmt where
 
 instance KillRange ModuleApplication where
   killRange (SectionApp _ t e)    = killRange2 (SectionApp noRange) t e
-  killRange (RecordModuleIFS _ q) = killRange1 (RecordModuleIFS noRange) q
+  killRange (RecordModuleInstance _ q) = killRange1 (RecordModuleInstance noRange) q
 
 instance KillRange e => KillRange (OpApp e) where
   killRange (SyntaxBindingLambda _ l e) = killRange2 (SyntaxBindingLambda noRange) l e
@@ -844,6 +871,7 @@ instance KillRange Pattern where
   killRange (LitP l)          = killRange1 LitP l
   killRange (QuoteP _)        = QuoteP noRange
   killRange (RecP _ fs)       = killRange1 (RecP noRange) fs
+  killRange (EqualP _ es)     = killRange1 (EqualP noRange) es
   killRange (EllipsisP _)     = EllipsisP noRange
   killRange (WithP _ p)       = killRange1 (WithP noRange) p
 
@@ -851,29 +879,20 @@ instance KillRange Pragma where
   killRange (OptionsPragma _ s)               = OptionsPragma noRange s
   killRange (BuiltinPragma _ s e)             = killRange1 (BuiltinPragma noRange s) e
   killRange (RewritePragma _ qs)              = killRange1 (RewritePragma noRange) qs
-  killRange (CompiledDataPragma _ q s ss)     = killRange1 (\q -> CompiledDataPragma noRange q s ss) q
-  killRange (CompiledTypePragma _ q s)        = killRange1 (\q -> CompiledTypePragma noRange q s) q
-  killRange (CompiledPragma _ q s)            = killRange1 (\q -> CompiledPragma noRange q s) q
-  killRange (CompiledExportPragma _ q s)      = killRange1 (\q -> CompiledExportPragma noRange q s) q
-  killRange (CompiledJSPragma _ q s)          = killRange1 (\q -> CompiledJSPragma noRange q s) q
-  killRange (CompiledUHCPragma _ q s)         = killRange1 (\q -> CompiledUHCPragma noRange q s) q
-  killRange (CompiledDataUHCPragma _ q s ss)  = killRange1 (\q -> CompiledDataUHCPragma noRange q s ss) q
-  killRange (HaskellCodePragma _ s)           = HaskellCodePragma noRange s
   killRange (StaticPragma _ q)                = killRange1 (StaticPragma noRange) q
   killRange (InjectivePragma _ q)             = killRange1 (InjectivePragma noRange) q
   killRange (InlinePragma _ b q)              = killRange1 (InlinePragma noRange b) q
-  killRange (ImportPragma _ s)                = ImportPragma noRange s
-  killRange (ImportUHCPragma _ s)             = ImportUHCPragma noRange s
   killRange (CompilePragma _ b q s)           = killRange1 (\ q -> CompilePragma noRange b q s) q
   killRange (ForeignPragma _ b s)             = ForeignPragma noRange b s
   killRange (ImpossiblePragma _)              = ImpossiblePragma noRange
   killRange (TerminationCheckPragma _ t)      = TerminationCheckPragma noRange (killRange t)
-  killRange (WarningOnUsage _ nm str)         = WarningOnUsage noRange nm str
+  killRange (WarningOnUsage _ nm str)         = WarningOnUsage noRange (killRange nm) str
   killRange (CatchallPragma _)                = CatchallPragma noRange
   killRange (DisplayPragma _ lhs rhs)         = killRange2 (DisplayPragma noRange) lhs rhs
   killRange (EtaPragma _ q)                   = killRange1 (EtaPragma noRange) q
   killRange (NoPositivityCheckPragma _)       = NoPositivityCheckPragma noRange
   killRange (PolarityPragma _ q occs)         = killRange1 (\q -> PolarityPragma noRange q occs) q
+  killRange (NoUniverseCheckPragma _)         = NoUniverseCheckPragma noRange
 
 instance KillRange RHS where
   killRange AbsurdRHS = AbsurdRHS
@@ -883,9 +902,6 @@ instance KillRange TypedBinding where
   killRange (TBind _ b e) = killRange2 (TBind noRange) b e
   killRange (TLet r ds)   = killRange2 TLet r ds
 
-instance KillRange TypedBindings where
-  killRange (TypedBindings _ t) = killRange1 (TypedBindings noRange) t
-
 instance KillRange WhereClause where
   killRange NoWhere         = NoWhere
   killRange (AnyWhere d)    = killRange1 AnyWhere d
@@ -915,6 +931,7 @@ instance NFData Expr where
   rnf (Set _)            = ()
   rnf (Prop _)           = ()
   rnf (SetN _ a)         = rnf a
+  rnf (PropN _ a)        = rnf a
   rnf (Rec _ a)          = rnf a
   rnf (RecUpdate _ a b)  = rnf a `seq` rnf b
   rnf (Let _ a b)        = rnf a `seq` rnf b
@@ -934,6 +951,7 @@ instance NFData Expr where
   rnf (DontCare a)       = rnf a
   rnf (Equal _ a b)      = rnf a `seq` rnf b
   rnf (Ellipsis _)       = ()
+  rnf (Generalized e)    = rnf e
 
 -- | Ranges are not forced.
 
@@ -952,6 +970,7 @@ instance NFData Pattern where
   rnf (DotP _ a) = rnf a
   rnf (LitP a) = rnf a
   rnf (RecP _ a) = rnf a
+  rnf (EqualP _ es) = rnf es
   rnf (EllipsisP _) = ()
   rnf (WithP _ a) = rnf a
 
@@ -959,11 +978,14 @@ instance NFData Pattern where
 
 instance NFData Declaration where
   rnf (TypeSig a b c)         = rnf a `seq` rnf b `seq` rnf c
+  rnf (Generalize _ a)        = rnf a
   rnf (Field a b c)           = rnf a `seq` rnf b `seq` rnf c
   rnf (FunClause a b c d)     = rnf a `seq` rnf b `seq` rnf c `seq` rnf d
   rnf (DataSig _ a b c d)     = rnf a `seq` rnf b `seq` rnf c `seq` rnf d
   rnf (Data _ a b c d e)      = rnf a `seq` rnf b `seq` rnf c `seq` rnf d `seq` rnf e
+  rnf (DataDef _ a b c d)     = rnf a `seq` rnf b `seq` rnf c `seq` rnf d
   rnf (RecordSig _ a b c)     = rnf a `seq` rnf b `seq` rnf c
+  rnf (RecordDef _ a b c d e f) = rnf a `seq` rnf b `seq` rnf c `seq` rnf d `seq` rnf e `seq` rnf f
   rnf (Record _ a b c d e f g)  = rnf a `seq` rnf b `seq` rnf c `seq` rnf d `seq` rnf e `seq` rnf f `seq` rnf g
   rnf (Infix a b)             = rnf a `seq` rnf b
   rnf (Syntax a b)            = rnf a `seq` rnf b
@@ -989,21 +1011,11 @@ instance NFData Pragma where
   rnf (OptionsPragma _ a)               = rnf a
   rnf (BuiltinPragma _ a b)             = rnf a `seq` rnf b
   rnf (RewritePragma _ a)               = rnf a
-  rnf (CompiledDataPragma _ a b c)      = rnf a `seq` rnf b `seq` rnf c
-  rnf (CompiledTypePragma _ a b)        = rnf a `seq` rnf b
-  rnf (CompiledPragma _ a b)            = rnf a `seq` rnf b
-  rnf (CompiledExportPragma _ a b)      = rnf a `seq` rnf b
-  rnf (CompiledJSPragma _ a b)          = rnf a `seq` rnf b
-  rnf (CompiledUHCPragma _ a b)         = rnf a `seq` rnf b
-  rnf (CompiledDataUHCPragma _ a b c)   = rnf a `seq` rnf b `seq` rnf c
-  rnf (HaskellCodePragma _ s)           = rnf s
   rnf (CompilePragma _ a b c)           = rnf a `seq` rnf b `seq` rnf c
   rnf (ForeignPragma _ b s)             = rnf b `seq` rnf s
   rnf (StaticPragma _ a)                = rnf a
   rnf (InjectivePragma _ a)             = rnf a
   rnf (InlinePragma _ _ a)              = rnf a
-  rnf (ImportPragma _ a)                = rnf a
-  rnf (ImportUHCPragma _ a)             = rnf a
   rnf (ImpossiblePragma _)              = ()
   rnf (EtaPragma _ a)                   = rnf a
   rnf (TerminationCheckPragma _ a)      = rnf a
@@ -1012,11 +1024,7 @@ instance NFData Pragma where
   rnf (DisplayPragma _ a b)             = rnf a `seq` rnf b
   rnf (NoPositivityCheckPragma _)       = ()
   rnf (PolarityPragma _ a b)            = rnf a `seq` rnf b
-
--- | Ranges are not forced.
-
-instance NFData a => NFData (TypedBindings' a) where
-  rnf (TypedBindings _ a) = rnf a
+  rnf (NoUniverseCheckPragma _)         = ()
 
 -- | Ranges are not forced.
 
@@ -1033,7 +1041,7 @@ instance NFData a => NFData (TypedBinding' a) where
 
 instance NFData ModuleApplication where
   rnf (SectionApp _ a b)    = rnf a `seq` rnf b
-  rnf (RecordModuleIFS _ a) = rnf a
+  rnf (RecordModuleInstance _ a) = rnf a
 
 -- | Ranges are not forced.
 
@@ -1061,11 +1069,11 @@ instance NFData LamClause where
   rnf (LamClause a b c d) = rnf (a, b, c, d)
 
 instance NFData a => NFData (LamBinding' a) where
-  rnf (DomainFree a b) = rnf a `seq` rnf b
-  rnf (DomainFull a)   = rnf a
+  rnf (DomainFree a) = rnf a
+  rnf (DomainFull a) = rnf a
 
 instance NFData BoundName where
-  rnf (BName a b c) = rnf a `seq` rnf b `seq` rnf c
+  rnf (BName a b) = rnf a `seq` rnf b
 
 instance NFData a => NFData (RHS' a) where
   rnf AbsurdRHS = ()
