@@ -210,8 +210,8 @@ buildParsers r flat kind exprNames = do
           Trie.member (addHole withHole p) partListsInExpr
           where
           p = case n of
-            NormalHole{} : IdPart p : _ -> p
-            IdPart p : _                -> p
+            NormalHole{} : IdPart p : _ -> rangedThing p
+            IdPart p : _                -> rangedThing p
             _                           -> __IMPOSSIBLE__
 
         -- Is the last identifier part present in n present in the
@@ -221,8 +221,8 @@ buildParsers r flat kind exprNames = do
           Trie.member (addHole withHole p) reversedPartListsInExpr
           where
           p = case reverse n of
-            NormalHole{} : IdPart p : _ -> p
-            IdPart p : _                -> p
+            NormalHole{} : IdPart p : _ -> rangedThing p
+            IdPart p : _                -> rangedThing p
             _                           -> __IMPOSSIBLE__
 
         -- | Are the initial and final identifier parts present with
@@ -263,7 +263,8 @@ buildParsers r flat kind exprNames = do
 
         (non, fix) = List.partition nonfix (filter (and . partsPresent) ops)
 
-        cons       = getDefinedNames [ConName, PatternSynName] flat
+        cons       = getDefinedNames
+                       [ConName, FldName, PatternSynName] flat
         conNames   = Set.fromList $
                        filter (flip Set.member namesInExpr) $
                        map (notaName . head) cons
@@ -390,7 +391,7 @@ buildParsers r flat kind exprNames = do
               }
 
     reportSDoc "scope.grammar" 10 $ return $
-      text "Operator grammar:" $$ nest 2 (grammar (pTop g))
+      "Operator grammar:" $$ nest 2 (grammar (pTop g))
 
     return (parseSections, everything, g)
     where
@@ -542,6 +543,7 @@ parsePat prs p = case p of
     QuoteP _         -> return p
     IdentP _         -> return p
     RecP r fs        -> RecP r <$> mapM (traverse (parsePat prs)) fs
+    EqualP{}         -> return p -- Andrea: cargo culted from DotP
     EllipsisP _      -> fail "bad ellipsis"
     WithP r p        -> WithP r <$> parsePat prs p
 
@@ -605,7 +607,7 @@ parseLHS' lhsOrPatSyn top p = do
                , p' <- foldr seq () result `seq` result
                , res <- validPattern (PatternCheckConfig top cons flds) p' ] of
         [(p,lhs)] -> do reportSDoc "scope.operators" 50 $ return $
-                          text "Parsed lhs:" <+> pretty lhs
+                          "Parsed lhs:" <+> pretty lhs
                         return (lhs, ops)
         []        -> typeError $ OperatorInformation ops
                                $ NoParseForLHS lhsOrPatSyn p
@@ -737,7 +739,7 @@ parseApplication es  = billToParser IsExpr $ do
     case foldr seq () result `seq` result of
         [e] -> do
           reportSDoc "scope.operators" 50 $ return $
-            text "Parsed an operator application:" <+> pretty e
+            "Parsed an operator application:" <+> pretty e
           return e
         []  -> typeError $ OperatorInformation ops
                          $ NoParseForApplication es
