@@ -40,38 +40,38 @@ import qualified System.IO.Unsafe as UNSAFE (unsafePerformIO)
 
 -- Andreas, 2017-10-02, TODO: restore Show to its original purpose
 --
--- deriving instance Show Expr
--- deriving instance (Show a) => Show (OpApp a)
--- deriving instance Show Declaration
--- deriving instance Show Pattern
--- deriving instance Show TypedBinding
--- deriving instance Show TypedBindings
--- deriving instance Show LamBinding
--- deriving instance Show ModuleAssignment
--- deriving instance (Show a, Show b) => Show (ImportDirective' a b)
--- deriving instance (Show a, Show b) => Show (Using' a b)
--- deriving instance (Show a, Show b) => Show (Renaming' a b)
--- deriving instance Show Pragma
--- deriving instance Show RHS
--- deriving instance Show LHS
--- deriving instance Show LHSCore
--- deriving instance Show WhereClause
--- deriving instance Show ModuleApplication
+deriving instance Show Expr
+deriving instance (Show a) => Show (OpApp a)
+deriving instance Show Declaration
+deriving instance Show Pattern
+deriving instance Show TypedBinding
+deriving instance Show LamBinding
+deriving instance Show ModuleAssignment
+deriving instance (Show a, Show b) => Show (ImportDirective' a b)
+deriving instance (Show a, Show b) => Show (Using' a b)
+deriving instance (Show a, Show b) => Show (Renaming' a b)
+deriving instance Show Pragma
+deriving instance Show RHS
+deriving instance Show LHS
+deriving instance Show LHSCore
+deriving instance Show LamClause
+deriving instance Show WhereClause
+deriving instance Show ModuleApplication
+deriving instance Show DoStmt
 
-instance Show Expr            where show = show . pretty
-instance Show Declaration     where show = show . pretty
-instance Show Pattern         where show = show . pretty
-instance Show TypedBinding    where show = show . pretty
-instance Show TypedBindings   where show = show . pretty
-instance Show LamBinding      where show = show . pretty
-instance (Pretty a, Pretty b) => Show (ImportDirective' a b)
-                              where show = show . pretty
-instance Show Pragma          where show = show . pretty
-instance Show RHS             where show = show . pretty
-instance Show LHS where show = show . pretty
-instance Show LHSCore where show = show . pretty
-instance Show WhereClause where show = show . pretty
-instance Show ModuleApplication where show = show . pretty
+-- instance Show Expr            where show = show . pretty
+-- instance Show Declaration     where show = show . pretty
+-- instance Show Pattern         where show = show . pretty
+-- instance Show TypedBinding    where show = show . pretty
+-- instance Show LamBinding      where show = show . pretty
+-- instance (Pretty a, Pretty b) => Show (ImportDirective' a b)
+--                               where show = show . pretty
+-- instance Show Pragma          where show = show . pretty
+-- instance Show RHS             where show = show . pretty
+-- instance Show LHS where show = show . pretty
+-- instance Show LHSCore where show = show . pretty
+-- instance Show WhereClause where show = show . pretty
+-- instance Show ModuleApplication where show = show . pretty
 
 
 -- | Picking the appropriate set of special characters depending on
@@ -90,22 +90,25 @@ specialCharacters :: SpecialCharacters
 specialCharacters =
   let opt = UNSAFE.unsafePerformIO (readIORef unicodeOrAscii) in
   case opt of
-    UnicodeOk -> SpecialCharacters { _dbraces = ((text "\x2983 " <>) . (<> text " \x2984"))
-                                   , _lambda  = text "\x03bb"
-                                   , _arrow   = text "\x2192"
-                                   , _forallQ = text "\x2200"
+    UnicodeOk -> SpecialCharacters { _dbraces = (("\x2983 " <>) . (<> " \x2984"))
+                                   , _lambda  = "\x03bb"
+                                   , _arrow   = "\x2192"
+                                   , _forallQ = "\x2200"
                                    }
     AsciiOnly -> SpecialCharacters { _dbraces = braces . braces'
-                                   , _lambda  = text "\\"
-                                   , _arrow   = text "->"
-                                   , _forallQ = text "forall"
+                                   , _lambda  = "\\"
+                                   , _arrow   = "->"
+                                   , _forallQ = "forall"
                                    }
 
 braces' :: Doc -> Doc
-braces' d = case render d of
-  -- Add space to avoid starting a comment
-  '-':_ -> braces (text " " <> d)
-  _     -> braces d
+braces' d = ifNull (render d) (braces d) {-else-} $ \ s ->
+  braces (spaceIfDash (head s) <> d <> spaceIfDash (last s))
+  -- Add space to avoid starting a comment (Ulf, 2010-09-13, #269)
+  -- Andreas, 2018-07-21, #3161: Also avoid ending a comment
+  where
+  spaceIfDash '-' = " "
+  spaceIfDash _   = empty
 
 -- double braces...
 dbraces :: Doc -> Doc
@@ -125,9 +128,9 @@ forallQ = _forallQ specialCharacters
 -- If the list is empty, then the notation @{}@ is used.
 
 bracesAndSemicolons :: [Doc] -> Doc
-bracesAndSemicolons []       = text "{}"
+bracesAndSemicolons []       = "{}"
 bracesAndSemicolons (d : ds) =
-  sep ([text "{" <+> d] ++ map (text ";" <+>) ds ++ [text "}"])
+  sep (["{" <+> d] ++ map (";" <+>) ds ++ ["}"])
 
 arrow, lambda :: Doc
 arrow  = _arrow specialCharacters
@@ -158,19 +161,15 @@ instance Pretty a => Pretty (WithHiding a) where
 
 instance Pretty Relevance where
   pretty Relevant   = empty
-  pretty Irrelevant = text "."
-  pretty NonStrict  = text ".."
-
-instance Pretty Induction where
-  pretty Inductive = text "data"
-  pretty CoInductive = text "codata"
+  pretty Irrelevant = "."
+  pretty NonStrict  = ".."
 
 instance Pretty (OpApp Expr) where
   pretty (Ordinary e) = pretty e
   pretty (SyntaxBindingLambda r bs e) = pretty (Lam r bs e)
 
 instance Pretty a => Pretty (MaybePlaceholder a) where
-  pretty Placeholder{}       = text "_"
+  pretty Placeholder{}       = "_"
   pretty (NoPlaceholder _ e) = pretty e
 
 instance Pretty Expr where
@@ -178,7 +177,7 @@ instance Pretty Expr where
         case e of
             Ident x          -> pretty x
             Lit l            -> pretty l
-            QuestionMark _ n -> text "?" <> maybe empty (text . show) n
+            QuestionMark _ n -> "?" <> maybe empty (text . show) n
             Underscore _ n   -> maybe underscore text n
 --          Underscore _ n   -> underscore <> maybe empty (text . show) n
             App _ _ _        ->
@@ -192,7 +191,7 @@ instance Pretty Expr where
             OpApp _ q _ es -> fsep $ prettyOpApp q es
 
             WithApp _ e es -> fsep $
-              pretty e : map ((text "|" <+>) . pretty) es
+              pretty e : map (("|" <+>) . pretty) es
 
             HiddenArg _ e -> braces' $ pretty e
             InstanceArg _ e -> dbraces $ pretty e
@@ -211,48 +210,50 @@ instance Pretty Expr where
                 sep [ pretty (Tel $ smashTel tel) <+> arrow
                     , pretty e
                     ]
-            Set _   -> text "Set"
-            Prop _  -> text "Prop"
-            SetN _ n    -> text "Set" <> text (showIndex n)
+            Set _   -> "Set"
+            Prop _  -> "Prop"
+            SetN _ n    -> "Set" <> text (showIndex n)
+            PropN _ n   -> "Prop" <> text (showIndex n)
             Let _ ds me  ->
-                sep [ text "let" <+> vcat (map pretty ds)
-                    , maybe empty (\ e -> text "in" <+> pretty e) me
+                sep [ "let" <+> vcat (map pretty ds)
+                    , maybe empty (\ e -> "in" <+> pretty e) me
                     ]
             Paren _ e -> parens $ pretty e
-            IdiomBrackets _ e -> text "(|" <+> pretty e <+> text "|)"
-            DoBlock _ ss -> text "do" <+> vcat (map pretty ss)
-            As _ x e  -> pretty x <> text "@" <> pretty e
-            Dot _ e   -> text "." <> pretty e
-            Absurd _  -> text "()"
-            Rec _ xs  -> sep [text "record", bracesAndSemicolons (map pretty xs)]
+            IdiomBrackets _ e -> "(|" <+> pretty e <+> "|)"
+            DoBlock _ ss -> "do" <+> vcat (map pretty ss)
+            As _ x e  -> pretty x <> "@" <> pretty e
+            Dot _ e   -> "." <> pretty e
+            Absurd _  -> "()"
+            Rec _ xs  -> sep ["record", bracesAndSemicolons (map pretty xs)]
             RecUpdate _ e xs ->
-              sep [text "record" <+> pretty e, bracesAndSemicolons (map pretty xs)]
-            ETel []  -> text "()"
+              sep ["record" <+> pretty e, bracesAndSemicolons (map pretty xs)]
+            ETel []  -> "()"
             ETel tel -> fsep $ map pretty tel
-            QuoteGoal _ x e -> sep [text "quoteGoal" <+> pretty x <+> text "in",
+            QuoteGoal _ x e -> sep ["quoteGoal" <+> pretty x <+> "in",
                                     nest 2 $ pretty e]
-            QuoteContext _ -> text "quoteContext"
-            Quote _ -> text "quote"
-            QuoteTerm _ -> text "quoteTerm"
-            Unquote _ -> text "unquote"
+            QuoteContext _ -> "quoteContext"
+            Quote _ -> "quote"
+            QuoteTerm _ -> "quoteTerm"
+            Unquote _ -> "unquote"
             Tactic _ t es ->
-              sep [ text "tactic" <+> pretty t
-                  , fsep [ text "|" <+> pretty e | e <- es ]
+              sep [ "tactic" <+> pretty t
+                  , fsep [ "|" <+> pretty e | e <- es ]
                   ]
             -- Andreas, 2011-10-03 print irrelevant things as .(e)
-            DontCare e -> text "." <> parens (pretty e)
-            Equal _ a b -> pretty a <+> text "=" <+> pretty b
-            Ellipsis _  -> text "..."
+            DontCare e -> "." <> parens (pretty e)
+            Equal _ a b -> pretty a <+> "=" <+> pretty b
+            Ellipsis _  -> "..."
+            Generalized e -> pretty e
         where
-          absurd NotHidden  = text "()"
-          absurd Instance{} = text "{{}}"
-          absurd Hidden     = text "{}"
+          absurd NotHidden  = "()"
+          absurd Instance{} = "{{}}"
+          absurd Hidden     = "{}"
 
 instance (Pretty a, Pretty b) => Pretty (Either a b) where
   pretty = either pretty pretty
 
 instance Pretty a => Pretty (FieldAssignment' a) where
-  pretty (FieldAssignment x e) = sep [ pretty x <+> text "=" , nest 2 $ pretty e ]
+  pretty (FieldAssignment x e) = sep [ pretty x <+> "=" , nest 2 $ pretty e ]
 
 instance Pretty ModuleAssignment where
   pretty (ModuleAssignment m es i) = (fsep $ pretty m : map pretty es) <+> pretty i
@@ -267,24 +268,48 @@ instance Pretty LamClause where
       pretty' AbsurdRHS = empty
 
 instance Pretty BoundName where
-  pretty BName{ boundName = x, boundLabel = l }
-    | x == l    = pretty x
-    | otherwise = pretty l <+> text "=" <+> pretty x
+  pretty BName{ boundName = x } = pretty x
+
+data NamedBinding = NamedBinding { withHiding   :: Bool
+                                 , namedBinding :: NamedArg BoundName }
+
+getLabel :: NamedArg a -> Maybe String
+getLabel = fmap rangedThing . nameOf . unArg
+
+isLabeled :: NamedArg BoundName -> Bool
+isLabeled x
+  | visible x            = False  -- Ignore labels on visible arguments
+  | Just l <- getLabel x = l /= nameToRawName (boundName $ namedArg x)
+  | otherwise            = False
+
+instance Pretty NamedBinding where
+  pretty (NamedBinding withH x) =
+    prH $ if | isLabeled x -> text (fromMaybe __IMPOSSIBLE__ $ getLabel x) <+> "=" <+> pretty (namedArg x)
+             | otherwise   -> pretty (namedArg x)
+    where
+      prH | withH     = prettyRelevance x . prettyHiding x id
+          | otherwise = id
 
 instance Pretty LamBinding where
-    -- TODO guilhem: colors are unused (colored syntax disallowed)
-    pretty (DomainFree i x) = prettyRelevance i $ prettyHiding i id $ pretty x
-    pretty (DomainFull b)   = pretty b
+    pretty (DomainFree x) = pretty (NamedBinding True x)
+    pretty (DomainFull b) = pretty b
 
-instance Pretty TypedBindings where
-  pretty (TypedBindings _ a) = prettyRelevance a $ prettyHiding a p $
-    pretty $ unArg a
+instance Pretty TypedBinding where
+    pretty (TLet _ ds) = parens $ "let" <+> vcat (map pretty ds)
+    pretty (TBind _ xs (Underscore _ Nothing)) =
+      fsep (map (pretty . NamedBinding True) xs)
+    pretty (TBind _ xs e) = fsep
+      [ prettyRelevance y $ prettyHiding y parens $
+        sep [ fsep (map (pretty . NamedBinding False) ys)
+            , ":" <+> pretty e ]
+      | ys@(y : _) <- groupBinds xs ]
       where
-        p | isUnderscore (unArg a) = id
-          | otherwise        = parens
-
-        isUnderscore (TBind _ _ (Underscore _ Nothing)) = True
-        isUnderscore _ = False
+        groupBinds [] = []
+        groupBinds (x : xs)
+          | isLabeled x = [x] : groupBinds xs
+          | otherwise   = (x : ys) : groupBinds zs
+          where (ys, zs) = span (same x) xs
+                same x y = getArgInfo x == getArgInfo y && not (isLabeled y)
 
 newtype Tel = Tel Telescope
 
@@ -293,43 +318,29 @@ instance Pretty Tel where
       | any isMeta tel = forallQ <+> fsep (map pretty tel)
       | otherwise      = fsep (map pretty tel)
       where
-        isMeta (TypedBindings _ (Arg _ (TBind _ _ (Underscore _ Nothing)))) = True
+        isMeta (TBind _ _ (Underscore _ Nothing)) = True
         isMeta _ = False
 
-
-instance Pretty TypedBinding where
-    pretty (TBind _ xs (Underscore _ Nothing)) =
-      fsep (map pretty xs)
-    pretty (TBind _ xs e) =
-        sep [ fsep (map pretty xs)
-            , text ":" <+> pretty e
-            ]
-    pretty (TLet _ ds) =
-        text "let" <+> vcat (map pretty ds)
-
 smashTel :: Telescope -> Telescope
-smashTel (TypedBindings r (Arg i  (TBind r' xs e)) :
-          TypedBindings _ (Arg i' (TBind _  ys e')) : tel)
-  | show i == show i' && show e == show e' && all (isUnnamed . dget) (xs ++ ys) =
-    smashTel (TypedBindings r (Arg i (TBind r' (xs ++ ys) e)) : tel)
-  where
-    isUnnamed x = boundLabel x == boundName x
+smashTel (TBind r xs e  :
+          TBind _ ys e' : tel)
+  | show e == show e' = smashTel (TBind r (xs ++ ys) e : tel)
 smashTel (b : tel) = b : smashTel tel
 smashTel [] = []
 
 
 instance Pretty RHS where
-    pretty (RHS e)   = text "=" <+> pretty e
+    pretty (RHS e)   = "=" <+> pretty e
     pretty AbsurdRHS = empty
 
 instance Pretty WhereClause where
   pretty  NoWhere = empty
   pretty (AnyWhere [Module _ x [] ds]) | isNoName (unqualify x)
-                       = vcat [ text "where", nest 2 (vcat $ map pretty ds) ]
-  pretty (AnyWhere ds) = vcat [ text "where", nest 2 (vcat $ map pretty ds) ]
+                       = vcat [ "where", nest 2 (vcat $ map pretty ds) ]
+  pretty (AnyWhere ds) = vcat [ "where", nest 2 (vcat $ map pretty ds) ]
   pretty (SomeWhere m a ds) =
-    vcat [ hsep $ applyWhen (a == PrivateAccess UserWritten) (text "private" :)
-             [ text "module", pretty m, text "where" ]
+    vcat [ hsep $ applyWhen (a == PrivateAccess UserWritten) ("private" :)
+             [ "module", pretty m, "where" ]
          , nest 2 (vcat $ map pretty ds)
          ]
 
@@ -344,7 +355,7 @@ instance Pretty LHS where
             ]
       pThing thing []       = empty
       pThing thing (e : es) = fsep $ (text thing <+> pretty e)
-                                   : map ((text "|" <+>) . pretty) es
+                                   : map (("|" <+>) . pretty) es
 
 instance Pretty LHSCore where
   pretty (LHSHead f ps) = sep $ pretty f : map (parens . pretty) ps
@@ -354,110 +365,92 @@ instance Pretty LHSCore where
   pretty (LHSWith h wps ps) = if null ps then doc else
       sep $ parens doc : map (parens . pretty) ps
     where
-    doc = sep $ pretty h : map ((text "|" <+>) . pretty) wps
+    doc = sep $ pretty h : map (("|" <+>) . pretty) wps
 
 instance Pretty ModuleApplication where
-  pretty (SectionApp _ bs e) = fsep (map pretty bs) <+> text "=" <+> pretty e
-  pretty (RecordModuleIFS _ rec) = text "=" <+> pretty rec <+> text "{{...}}"
+  pretty (SectionApp _ bs e) = fsep (map pretty bs) <+> "=" <+> pretty e
+  pretty (RecordModuleInstance _ rec) = "=" <+> pretty rec <+> "{{...}}"
 
 instance Pretty DoStmt where
   pretty (DoBind _ p e cs) =
-    ((pretty p <+> text "←") <?> pretty e) <?> prCs cs
+    ((pretty p <+> "←") <?> pretty e) <?> prCs cs
     where
       prCs [] = empty
-      prCs cs = text "where" <?> vcat (map pretty cs)
+      prCs cs = "where" <?> vcat (map pretty cs)
   pretty (DoThen e) = pretty e
-  pretty (DoLet _ ds) = text "let" <+> vcat (map pretty ds)
+  pretty (DoLet _ ds) = "let" <+> vcat (map pretty ds)
 
 instance Pretty Declaration where
     prettyList = vcat . map pretty
     pretty d =
         case d of
             TypeSig i x e ->
-                sep [ prettyRelevance i $ pretty x <+> text ":"
+                sep [ prettyRelevance i $ pretty x <+> ":"
                     , nest 2 $ pretty e
                     ]
             Field inst x (Arg i e) ->
-                sep [ text "field"
+                sep [ "field"
                     , nest 2 $ mkInst inst $ mkOverlap i $
                       prettyRelevance i $ prettyHiding i id $
                         pretty $ TypeSig (setRelevance Relevant i) x e
                     ]
                 where
-                  mkInst InstanceDef    d = sep [ text "instance", nest 2 d ]
+                  mkInst InstanceDef    d = sep [ "instance", nest 2 d ]
                   mkInst NotInstanceDef d = d
 
-                  mkOverlap i d | isOverlappable i = text "overlap" <+> d
+                  mkOverlap i d | isOverlappable i = "overlap" <+> d
                                 | otherwise        = d
             FunClause lhs rhs wh _ ->
                 sep [ pretty lhs
                     , nest 2 $ pretty rhs
                     ] $$ nest 2 (pretty wh)
             DataSig _ ind x tel e ->
-                sep [ hsep  [ pretty ind
+                sep [ hsep  [ "data"
                             , pretty x
                             , fcat (map pretty tel)
                             ]
                     , nest 2 $ hsep
-                            [ text ":"
+                            [ ":"
                             , pretty e
                             ]
                     ]
-            Data _ ind x tel (Just e) cs ->
-                sep [ hsep  [ pretty ind
+            Data _ ind x tel e cs ->
+                sep [ hsep  [ "data"
                             , pretty x
                             , fcat (map pretty tel)
                             ]
                     , nest 2 $ hsep
-                            [ text ":"
+                            [ ":"
                             , pretty e
-                            , text "where"
+                            , "where"
                             ]
                     ] $$ nest 2 (vcat $ map pretty cs)
-            Data _ ind x tel Nothing cs ->
-                sep [ hsep  [ pretty ind
+            DataDef _ ind x tel cs ->
+                sep [ hsep  [ "data"
                             , pretty x
                             , fcat (map pretty tel)
                             ]
-                    , nest 2 $ text "where"
+                    , nest 2 $ "where"
                     ] $$ nest 2 (vcat $ map pretty cs)
             RecordSig _ x tel e ->
-                sep [ hsep  [ text "record"
+                sep [ hsep  [ "record"
                             , pretty x
                             , fcat (map pretty tel)
                             ]
                     , nest 2 $ hsep
-                            [ text ":"
+                            [ ":"
                             , pretty e
                             ]
                     ]
-            Record _ x ind eta con tel me cs ->
-                sep [ hsep  [ text "record"
-                            , pretty x
-                            , fcat (map pretty tel)
-                            ]
-                    , nest 2 $ pType me
-                    ] $$ nest 2 (vcat $ pInd ++
-                                        pEta ++
-                                        pCon ++
-                                        map pretty cs)
-              where pType (Just e) = hsep
-                            [ text ":"
-                            , pretty e
-                            , text "where"
-                            ]
-                    pType Nothing  =
-                              text "where"
-                    pInd = maybeToList $ text . show . rangedThing <$> ind
-                    pEta = maybeToList $ eta <&> \case
-                      YesEta -> text "eta-equality"
-                      NoEta  -> text "no-eta-equality"
-                    pCon = maybeToList $ (text "constructor" <+>) . pretty <$> fst <$> con
+            Record _ x ind eta con tel e cs ->
+              pRecord x ind eta con tel (Just e) cs
+            RecordDef _ x ind eta con tel cs ->
+              pRecord x ind eta con tel Nothing cs
             Infix f xs  ->
                 pretty f <+> (fsep $ punctuate comma $ map pretty xs)
-            Syntax n xs -> text "syntax" <+> pretty n <+> text "..."
-            PatternSyn _ n as p -> text "pattern" <+> pretty n <+> fsep (map pretty as)
-                                     <+> text "=" <+> pretty p
+            Syntax n xs -> "syntax" <+> pretty n <+> "..."
+            PatternSyn _ n as p -> "pattern" <+> pretty n <+> fsep (map pretty as)
+                                     <+> "=" <+> pretty p
             Mutual _ ds     -> namedBlock "mutual" ds
             Abstract _ ds   -> namedBlock "abstract" ds
             Private _ _ ds  -> namedBlock "private" ds
@@ -465,11 +458,12 @@ instance Pretty Declaration where
             Macro _ ds      -> namedBlock "macro" ds
             Postulate _ ds  -> namedBlock "postulate" ds
             Primitive _ ds  -> namedBlock "primitive" ds
+            Generalize _ ds -> namedBlock "variable" ds
             Module _ x tel ds ->
-                hsep [ text "module"
+                hsep [ "module"
                      , pretty x
                      , fcat (map pretty tel)
-                     , text "where"
+                     , "where"
                      ] $$ nest 2 (vcat $ map pretty ds)
             ModuleMacro _ x (SectionApp _ [] e) DoOpen i | isNoName x ->
                 sep [ pretty DoOpen
@@ -477,92 +471,97 @@ instance Pretty Declaration where
                     , nest 4 $ pretty i
                     ]
             ModuleMacro _ x (SectionApp _ tel e) open i ->
-                sep [ pretty open <+> text "module" <+> pretty x <+> fcat (map pretty tel)
-                    , nest 2 $ text "=" <+> pretty e <+> pretty i
+                sep [ pretty open <+> "module" <+> pretty x <+> fcat (map pretty tel)
+                    , nest 2 $ "=" <+> pretty e <+> pretty i
                     ]
-            ModuleMacro _ x (RecordModuleIFS _ rec) open i ->
-                sep [ pretty open <+> text "module" <+> pretty x
-                    , nest 2 $ text "=" <+> pretty rec <+> text "{{...}}"
+            ModuleMacro _ x (RecordModuleInstance _ rec) open i ->
+                sep [ pretty open <+> "module" <+> pretty x
+                    , nest 2 $ "=" <+> pretty rec <+> "{{...}}"
                     ]
-            Open _ x i  -> hsep [ text "open", pretty x, pretty i ]
+            Open _ x i  -> hsep [ "open", pretty x, pretty i ]
             Import _ x rn open i   ->
-                hsep [ pretty open, text "import", pretty x, as rn, pretty i ]
+                hsep [ pretty open, "import", pretty x, as rn, pretty i ]
                 where
                     as Nothing  = empty
-                    as (Just x) = text "as" <+> pretty (asName x)
+                    as (Just x) = "as" <+> pretty (asName x)
             UnquoteDecl _ xs t ->
-              sep [ text "unquoteDecl" <+> fsep (map pretty xs) <+> text "=", nest 2 $ pretty t ]
+              sep [ "unquoteDecl" <+> fsep (map pretty xs) <+> "=", nest 2 $ pretty t ]
             UnquoteDef _ xs t ->
-              sep [ text "unquoteDef" <+> fsep (map pretty xs) <+> text "=", nest 2 $ pretty t ]
-            Pragma pr   -> sep [ text "{-#" <+> pretty pr, text "#-}" ]
+              sep [ "unquoteDef" <+> fsep (map pretty xs) <+> "=", nest 2 $ pretty t ]
+            Pragma pr   -> sep [ "{-#" <+> pretty pr, "#-}" ]
         where
             namedBlock s ds =
                 sep [ text s
                     , nest 2 $ vcat $ map pretty ds
                     ]
 
+pRecord :: Name -> Maybe (Ranged Induction) -> Maybe HasEta -> Maybe (Name, IsInstance) -> [LamBinding] -> Maybe Expr -> [Declaration] -> Doc
+pRecord x ind eta con tel me cs =
+  sep [ hsep  [ "record"
+              , pretty x
+              , fcat (map pretty tel)
+              ]
+      , nest 2 $ pType me
+      ] $$ nest 2 (vcat $ pInd ++
+                          pEta ++
+                          pCon ++
+                          map pretty cs)
+  where pType (Just e) = hsep
+                [ ":"
+                , pretty e
+                , "where"
+                ]
+        pType Nothing  =
+                  "where"
+        pInd = maybeToList $ text . show . rangedThing <$> ind
+        pEta = maybeToList $ eta <&> \case
+          YesEta -> "eta-equality"
+          NoEta  -> "no-eta-equality"
+        pCon = maybeToList $ ("constructor" <+>) . pretty <$> fst <$> con
+
 instance Pretty OpenShortHand where
-    pretty DoOpen = text "open"
+    pretty DoOpen = "open"
     pretty DontOpen = empty
 
 instance Pretty Pragma where
-    pretty (OptionsPragma _ opts) = fsep $ map text $ "OPTIONS" : opts
-    pretty (BuiltinPragma _ b x)  = hsep [ text "BUILTIN", text b, pretty x ]
+    pretty (OptionsPragma _ opts)  = fsep $ map text $ "OPTIONS" : opts
+    pretty (BuiltinPragma _ b x)   = hsep [ "BUILTIN", text b, pretty x ]
     pretty (RewritePragma _ xs)    =
-      hsep [ text "REWRITE", hsep $ map pretty xs ]
-    pretty (CompiledPragma _ x hs) =
-      hsep [ text "COMPILED", pretty x, text hs ]
-    pretty (CompiledExportPragma _ x hs) =
-      hsep [ text "COMPILED_EXPORT", pretty x, text hs ]
-    pretty (CompiledTypePragma _ x hs) =
-      hsep [ text "COMPILED_TYPE", pretty x, text hs ]
-    pretty (CompiledDataPragma _ x hs hcs) =
-      hsep $ [text "COMPILED_DATA", pretty x] ++ map text (hs : hcs)
-    pretty (CompiledJSPragma _ x e) =
-      hsep [ text "COMPILED_JS", pretty x, text e ]
-    pretty (CompiledUHCPragma _ x e) =
-      hsep [ text "COMPILED_UHC", pretty x, text e ]
-    pretty (CompiledDataUHCPragma _ x crd crcs) =
-      hsep $ [ text "COMPILED_DATA_UHC", pretty x] ++ map text (crd : crcs)
-    pretty (HaskellCodePragma _ s) =
-      vcat (text "HASKELL" : map text (lines s))
+      hsep [ "REWRITE", hsep $ map pretty xs ]
     pretty (CompilePragma _ b x e) =
-      hsep [ text "COMPILE", text b, pretty x, text e ]
+      hsep [ "COMPILE", text b, pretty x, text e ]
     pretty (ForeignPragma _ b s) =
       vcat $ text ("FOREIGN " ++ b) : map text (lines s)
     pretty (StaticPragma _ i) =
-      hsep $ [text "STATIC", pretty i]
+      hsep $ ["STATIC", pretty i]
     pretty (InjectivePragma _ i) =
-      hsep $ [text "INJECTIVE", pretty i]
+      hsep $ ["INJECTIVE", pretty i]
     pretty (InlinePragma _ True i) =
-      hsep $ [text "INLINE", pretty i]
+      hsep $ ["INLINE", pretty i]
     pretty (InlinePragma _ False i) =
-      hsep $ [text "NOINLINE", pretty i]
-    pretty (ImportPragma _ i) =
-      hsep $ [text "IMPORT", text i]
-    pretty (ImportUHCPragma _ i) =
-      hsep $ [text "IMPORT_UHC", text i]
+      hsep $ ["NOINLINE", pretty i]
     pretty (ImpossiblePragma _) =
-      hsep $ [text "IMPOSSIBLE"]
+      hsep $ ["IMPOSSIBLE"]
     pretty (EtaPragma _ x) =
-      hsep $ [text "ETA", pretty x]
+      hsep $ ["ETA", pretty x]
     pretty (TerminationCheckPragma _ tc) =
       case tc of
         TerminationCheck       -> __IMPOSSIBLE__
-        NoTerminationCheck     -> text "NO_TERMINATION_CHECK"
-        NonTerminating         -> text "NON_TERMINATING"
-        Terminating            -> text "TERMINATING"
-        TerminationMeasure _ x -> hsep $ [text "MEASURE", pretty x]
-    pretty (WarningOnUsage _ nm str) = hsep [ text "WARNING_ON_USAGE", pretty nm, text str ]
-    pretty (CatchallPragma _) = text "CATCHALL"
-    pretty (DisplayPragma _ lhs rhs) = text "DISPLAY" <+> sep [ pretty lhs <+> text "=", nest 2 $ pretty rhs ]
-    pretty (NoPositivityCheckPragma _) = text "NO_POSITIVITY_CHECK"
+        NoTerminationCheck     -> "NO_TERMINATION_CHECK"
+        NonTerminating         -> "NON_TERMINATING"
+        Terminating            -> "TERMINATING"
+        TerminationMeasure _ x -> hsep $ ["MEASURE", pretty x]
+    pretty (WarningOnUsage _ nm str) = hsep [ "WARNING_ON_USAGE", pretty nm, text str ]
+    pretty (CatchallPragma _) = "CATCHALL"
+    pretty (DisplayPragma _ lhs rhs) = "DISPLAY" <+> sep [ pretty lhs <+> "=", nest 2 $ pretty rhs ]
+    pretty (NoPositivityCheckPragma _) = "NO_POSITIVITY_CHECK"
     pretty (PolarityPragma _ q occs) =
-      hsep (text "POLARITY" : pretty q : map pretty occs)
+      hsep ("POLARITY" : pretty q : map pretty occs)
+    pretty (NoUniverseCheckPragma _) = "NO_UNIVERSE_CHECK"
 
 instance Pretty Fixity where
     pretty (Fixity _ Unrelated   _)   = __IMPOSSIBLE__
-    pretty (Fixity _ (Related n) ass) = text s <+> text (show n)
+    pretty (Fixity _ (Related n) ass) = s <+> text (show n)
       where
       s = case ass of
             LeftAssoc  -> "infixl"
@@ -570,7 +569,7 @@ instance Pretty Fixity where
             NonAssoc   -> "infix"
 
 instance Pretty GenPart where
-    pretty (IdPart x)   = text x
+    pretty (IdPart x)   = text $ rangedThing x
     pretty BindHole{}   = underscore
     pretty NormalHole{} = underscore
     pretty WildHole{}   = underscore
@@ -580,18 +579,23 @@ instance Pretty GenPart where
 instance Pretty Fixity' where
     pretty (Fixity' fix nota _)
       | nota == noNotation = pretty fix
-      | otherwise          = text "syntax" <+> pretty nota
+      | otherwise          = "syntax" <+> pretty nota
 
  -- Andreas 2010-09-21: do not print relevance in general, only in function types!
  -- Andreas 2010-09-24: and in record fields
 instance Pretty a => Pretty (Arg a) where
-  prettyPrec p (Arg ai e) = prettyHiding ai id $ prettyPrec p' e
+  prettyPrec p (Arg ai e) = prettyHiding ai localParens $ prettyPrec p' e
       where p' | visible ai = p
                | otherwise  = 0
+            localParens | getOrigin ai == Substitution = parens
+                        | otherwise = id
+
+instance Pretty a => Pretty (Dom a) where
+  pretty = pretty . argFromDom
 
 instance Pretty e => Pretty (Named_ e) where
     prettyPrec p (Named Nothing e) = prettyPrec p e
-    prettyPrec p (Named (Just s) e) = mparens (p > 0) $ sep [ text (rawNameToString $ rangedThing s) <+> text "=", pretty e ]
+    prettyPrec p (Named (Just s) e) = mparens (p > 0) $ sep [ text (rawNameToString $ rangedThing s) <+> "=", pretty e ]
 
 instance Pretty Pattern where
     prettyList = fsep . map pretty
@@ -604,14 +608,15 @@ instance Pretty Pattern where
             InstanceP _ p   -> dbraces $ pretty p
             ParenP _ p      -> parens $ pretty p
             WildP _         -> underscore
-            AsP _ x p       -> pretty x <> text "@" <> pretty p
-            DotP _ p        -> text "." <> pretty p
-            AbsurdP _       -> text "()"
+            AsP _ x p       -> pretty x <> "@" <> pretty p
+            DotP _ p        -> "." <> pretty p
+            AbsurdP _       -> "()"
             LitP l          -> pretty l
-            QuoteP _        -> text "quote"
-            RecP _ fs       -> sep [ text "record", bracesAndSemicolons (map pretty fs) ]
-            EllipsisP _     -> text "..."
-            WithP _ p       -> text "|" <+> pretty p
+            QuoteP _        -> "quote"
+            RecP _ fs       -> sep [ "record", bracesAndSemicolons (map pretty fs) ]
+            EqualP _ es     -> sep $ [ parens (sep [pretty e1, "=", pretty e2]) | (e1,e2) <- es ]
+            EllipsisP _     -> "..."
+            WithP _ p       -> "|" <+> pretty p
 
 prettyOpApp :: forall a .
   Pretty a => QName -> [NamedArg (MaybePlaceholder a)] -> [Doc]
@@ -621,8 +626,8 @@ prettyOpApp q es = merge [] $ prOp ms xs es
     ms = init (qnameParts q)
     -- xs: the concrete name (alternation of @Id@ and @Hole@)
     xs = case unqualify q of
-           Name _ xs -> xs
-           NoName{}  -> __IMPOSSIBLE__
+           Name _ _ xs    -> xs
+           NoName{}       -> __IMPOSSIBLE__
 
     prOp :: [Name] -> [NamePart] -> [NamedArg (MaybePlaceholder a)] -> [(Doc, Maybe PositionInName)]
     prOp ms (Hole : xs) (e : es) =
@@ -631,7 +636,7 @@ prettyOpApp q es = merge [] $ prOp ms xs es
         NoPlaceholder{} -> (pretty e, Nothing) : prOp ms xs es
           -- Module qualifier needs to go on section holes (#3072)
     prOp _  (Hole : _)  []       = __IMPOSSIBLE__
-    prOp ms (Id x : xs) es       = ( qual ms $ pretty $ Name noRange $ [Id x]
+    prOp ms (Id x : xs) es       = ( qual ms $ pretty $ Name noRange InScope $ [Id x]
                                    , Nothing
                                    ) : prOp [] xs es
       -- Qualify the name part with the module.
@@ -639,7 +644,7 @@ prettyOpApp q es = merge [] $ prOp ms xs es
 
     prOp _  []       es          = map (\e -> (pretty e, Nothing)) es
 
-    qual ms doc = hcat $ punctuate (text ".") $ map pretty ms ++ [doc]
+    qual ms doc = hcat $ punctuate "." $ map pretty ms ++ [doc]
 
     -- Section underscores should be printed without surrounding
     -- whitespace. This function takes care of that.
@@ -670,24 +675,24 @@ instance (Pretty a, Pretty b) => Pretty (ImportDirective' a b) where
             , rename $ impRenaming i
             ]
         where
-            public True  = text "public"
+            public True  = "public"
             public False = empty
 
             prettyHiding [] = empty
-            prettyHiding xs = text "hiding" <+> parens (fsep $ punctuate (text ";") $ map pretty xs)
+            prettyHiding xs = "hiding" <+> parens (fsep $ punctuate ";" $ map pretty xs)
 
             rename [] = empty
-            rename xs = hsep [ text "renaming"
-                             , parens $ fsep $ punctuate (text ";") $ map pr xs
+            rename xs = hsep [ "renaming"
+                             , parens $ fsep $ punctuate ";" $ map pr xs
                              ]
 
-            pr r = hsep [ pretty (renFrom r), text "to", pretty (renTo r) ]
+            pr r = hsep [ pretty (renFrom r), "to", pretty (renTo r) ]
 
 instance (Pretty a, Pretty b) => Pretty (Using' a b) where
     pretty UseEverything = empty
     pretty (Using xs)    =
-        text "using" <+> parens (fsep $ punctuate (text ";") $ map pretty xs)
+        "using" <+> parens (fsep $ punctuate ";" $ map pretty xs)
 
 instance (Pretty a, Pretty b) => Pretty (ImportedName' a b) where
     pretty (ImportedName x)     = pretty x
-    pretty (ImportedModule x)   = text "module" <+> pretty x
+    pretty (ImportedModule x)   = "module" <+> pretty x
