@@ -62,10 +62,10 @@ type NamedArgs  = [NamedArg Term]
 data ConHead = ConHead
   { conName      :: QName     -- ^ The name of the constructor.
   , conInductive :: Induction -- ^ Record constructors can be coinductive.
-  , conFields    :: [QName]   -- ^ The name of the record fields.
-                              --   Empty list for data constructors.
-                              --   'Arg' is not needed here since it
-                              --   is stored in the constructor args.
+  , conFields    :: [Arg QName]   -- ^ The name of the record fields.
+      --   Empty list for data constructors.
+      --   'Arg' is stored since the info in the constructor args
+      --   might not be accurate because of subtyping (issue #2170).
   } deriving (Data, Show)
 
 instance Eq ConHead where
@@ -118,6 +118,11 @@ data Term = Var {-# UNPACK #-} !Int Elims -- ^ @x es@ neutral
             -- ^ Irrelevant stuff in relevant position, but created
             --   in an irrelevant context.  Basically, an internal
             --   version of the irrelevance axiom @.irrAx : .A -> A@.
+          | Dummy String
+            -- ^ A (part of a) term or type which is only used for internal purposes.
+            --   Replaces the @Sort Prop@ hack.
+            --   The @String@ typically describes the location where we create this dummy,
+            --   but can contain other information as well.
   deriving (Data, Show)
 
 type ConInfo = ConOrigin
@@ -127,6 +132,7 @@ type ConInfo = ConOrigin
 data Elim' a
   = Apply (Arg a)         -- ^ Application.
   | Proj ProjOrigin QName -- ^ Projection.  'QName' is name of a record projection.
+  | IApply a a a -- ^ IApply x y r, x and y are the endpoints
   deriving (Data, Show, Functor, Foldable, Traversable)
 
 type Elim = Elim' Term
@@ -134,11 +140,14 @@ type Elims = [Elim]  -- ^ eliminations ordered left-to-right.
 
 -- | This instance cheats on 'Proj', use with care.
 --   'Proj's are always assumed to be 'UserWritten', since they have no 'ArgInfo'.
+--   Same for IApply
 instance LensOrigin (Elim' a) where
   getOrigin (Apply a)   = getOrigin a
   getOrigin Proj{}      = UserWritten
+  getOrigin IApply{}    = UserWritten
   mapOrigin f (Apply a) = Apply $ mapOrigin f a
   mapOrigin f e@Proj{}  = e
+  mapOrigin f e@IApply{} = e
 
 -- | Names in binders and arguments.
 type ArgName = String
@@ -155,6 +164,9 @@ appendArgNames = (++)
 nameToArgName :: Name -> ArgName
 nameToArgName = stringToArgName . prettyShow
 
+namedArgName :: NamedArg Name -> ArgName
+namedArgName x = maybe (nameToArgName $ namedArg x) rangedThing $ nameOf $ unArg x
+
 -- | Binder.
 --   'Abs': The bound variable might appear in the body.
 --   'NoAbs' is pseudo-binder, it does not introduce a fresh variable,
@@ -192,6 +204,9 @@ instance LensSort (Type' a) where
 instance LensSort a => LensSort (Dom a) where
   lensSort = traverseF . lensSort
 
+instance LensSort a => LensSort (Arg a) where
+  lensSort = traverseF . lensSort
+
 instance LensSort a => LensSort (Abs a) where
   lensSort = traverseF . lensSort
 
@@ -207,12 +222,18 @@ type Telescope = Tele (Dom Type)
 --
 data Sort
   = Type Level  -- ^ @Set ℓ@.
-  | Prop        -- ^ Dummy sort.
+  | Prop Level  -- ^ @Prop ℓ@.
   | Inf         -- ^ @Setω@.
   | SizeUniv    -- ^ @SizeUniv@, a sort inhabited by type @Size@.
   | PiSort Sort (Abs Sort) -- ^ Sort of the pi type.
   | UnivSort Sort -- ^ Sort of another sort.
   | MetaS {-# UNPACK #-} !MetaId Elims
+  | DefS QName Elims -- ^ A postulated sort.
+  | DummyS String
+    -- ^ A (part of a) term or type which is only used for internal purposes.
+    --   Replaces the abuse of @Prop@ for a dummy sort.
+    --   The @String@ typically describes the location where we create this dummy,
+    --   but can contain other information as well.
   deriving (Data, Show)
 
 -- | A level is a maximum expression of 0..n 'PlusLevel' expressions
@@ -351,6 +372,9 @@ stuckOn e r =
 -- * Definitions
 ---------------------------------------------------------------------------
 
+-- | Named pattern arguments.
+type NAPs = [NamedArg DeBruijnPattern]
+
 -- | A clause is a list of patterns and the clause body.
 --
 --  The telescope contains the types of the pattern variables and the
@@ -369,7 +393,7 @@ data Clause = Clause
     , clauseFullRange :: Range
     , clauseTel       :: Telescope
       -- ^ @Δ@: The types of the pattern variables in dependency order.
-    , namedClausePats :: [NamedArg DeBruijnPattern]
+    , namedClausePats :: NAPs
       -- ^ @Δ ⊢ ps@.  The de Bruijn indices refer to @Δ@.
     , clauseBody      :: Maybe Term
       -- ^ @Just v@ with @Δ ⊢ v@ for a regular clause, or @Nothing@ for an
@@ -435,6 +459,10 @@ data Pattern' x
     -- ^ E.g. @5@, @"hello"@.
   | ProjP ProjOrigin QName
     -- ^ Projection copattern.  Can only appear by itself.
+  | IApplyP PatOrigin Term Term x
+    -- ^ Path elimination pattern, like @VarP@ but keeps track of endpoints.
+  | DefP PatOrigin QName [NamedArg (Pattern' x)]
+    -- ^ Used for HITs, the QName should be the one from primHComp.
   deriving (Data, Show, Functor, Foldable, Traversable)
 
 type Pattern = Pattern' PatVarName
@@ -461,6 +489,10 @@ namedVarP x = Named named $ varP x
 namedDBVarP :: Int -> PatVarName -> Named_ DeBruijnPattern
 namedDBVarP m = (fmap . fmap) (\x -> DBPatVar x m) . namedVarP
 
+-- | Make an absurd pattern with the given de Bruijn index.
+absurdP :: Int -> DeBruijnPattern
+absurdP = VarP PatOAbsurd . DBPatVar absurdPatternName
+
 -- | The @ConPatternInfo@ states whether the constructor belongs to
 --   a record type (@Just@) or data type (@Nothing@).
 --   In the former case, the @PatOrigin@ says whether the record pattern
@@ -472,6 +504,9 @@ data ConPatternInfo = ConPatternInfo
   { conPRecord :: Maybe PatOrigin
     -- ^ @Nothing@ if data constructor.
     --   @Just@ if record constructor.
+  , conPFallThrough :: Bool
+    -- ^ Should the match block on non-canonical terms or can it
+    --   proceed to the catch-all clause?
   , conPType   :: Maybe (Arg Type)
     -- ^ The type of the whole constructor pattern.
     --   Should be present (@Just@) if constructor pattern is
@@ -490,7 +525,7 @@ data ConPatternInfo = ConPatternInfo
   deriving (Data, Show)
 
 noConPatternInfo :: ConPatternInfo
-noConPatternInfo = ConPatternInfo Nothing Nothing False
+noConPatternInfo = ConPatternInfo Nothing False Nothing False
 
 -- | Build partial 'ConPatternInfo' from 'ConInfo'
 toConPatternInfo :: ConInfo -> ConPatternInfo
@@ -523,8 +558,11 @@ instance PatternVars a (Arg (Pattern' a)) where
   patternVars (Arg i (VarP _ x)   ) = [Arg i $ Left x]
   patternVars (Arg i (DotP _ t)   ) = [Arg i $ Right t]
   patternVars (Arg _ (ConP _ _ ps)) = patternVars ps
+  patternVars (Arg _ (DefP _ _ ps)) = patternVars ps
   patternVars (Arg _ (LitP _)     ) = []
   patternVars (Arg _ ProjP{}      ) = []
+  patternVars (Arg i (IApplyP _ _ _ x)) = [Arg i $ Left x]
+
 
 instance PatternVars a (NamedArg (Pattern' a)) where
   patternVars = patternVars . fmap namedThing
@@ -540,6 +578,8 @@ patternOrigin (DotP o _) = Just o
 patternOrigin LitP{}     = Nothing
 patternOrigin (ConP _ ci _) = conPRecord ci
 patternOrigin ProjP{}    = Nothing
+patternOrigin (IApplyP o _ _ _) = Just o
+patternOrigin (DefP o _ _) = Just o
 
 -- | Does the pattern perform a match that could fail?
 properlyMatching :: DeBruijnPattern -> Bool
@@ -551,6 +591,8 @@ properlyMatching LitP{} = True
 properlyMatching (ConP _ ci ps) = isNothing (conPRecord ci) || -- not a record cons
   List.any (properlyMatching . namedArg) ps  -- or one of subpatterns is a proper m
 properlyMatching ProjP{} = True
+properlyMatching IApplyP{} = False
+properlyMatching DefP{}  = True
 
 instance IsProjP (Pattern' a) where
   isProjP (ProjP o d) = Just (o, unambiguous d)
@@ -648,6 +690,32 @@ isEqualityType :: EqualityView -> Bool
 isEqualityType EqualityType{} = True
 isEqualityType OtherType{}    = False
 
+-- | View type as path type.
+
+data PathView
+  = PathType
+    { pathSort  :: Sort     -- ^ Sort of this type.
+    , pathName  :: QName    -- ^ Builtin PATH.
+    , pathLevel :: Arg Term -- ^ Hidden
+    , pathType  :: Arg Term -- ^ Hidden
+    , pathLhs   :: Arg Term -- ^ NotHidden
+    , pathRhs   :: Arg Term -- ^ NotHidden
+    }
+  | OType Type -- ^ reduced
+
+isPathType :: PathView -> Bool
+isPathType PathType{} = True
+isPathType OType{}    = False
+
+data IntervalView
+      = IZero
+      | IOne
+      | IMin (Arg Term) (Arg Term)
+      | IMax (Arg Term) (Arg Term)
+      | INeg (Arg Term)
+      | OTerm Term
+      deriving Show
+
 ---------------------------------------------------------------------------
 -- * Absurd Lambda
 ---------------------------------------------------------------------------
@@ -683,9 +751,41 @@ dontCare v =
     DontCare{} -> v
     _          -> DontCare v
 
--- | A dummy type.
-typeDontCare :: Type
-typeDontCare = sort Prop
+-- | Aux: A dummy term to constitute a dummy term/level/sort/type.
+dummyTerm' :: String -> Int -> Term
+dummyTerm' file line = Dummy $ file ++ ":" ++ show line
+
+-- | Aux: A dummy level to constitute a level/sort.
+dummyLevel' :: String -> Int -> Level
+dummyLevel' file line = unreducedLevel $ dummyTerm' file line
+
+-- | A dummy term created at location.
+--   Note: use macro __DUMMY_TERM__ !
+dummyTerm :: String -> Int -> Term
+dummyTerm file line = dummyTerm' ("dummyTerm: " ++ file) line
+
+-- | A dummy level created at location.
+--   Note: use macro __DUMMY_LEVEL__ !
+dummyLevel :: String -> Int -> Level
+dummyLevel file line = dummyLevel' ("dummyLevel: " ++ file) line
+
+-- | A dummy sort created at location.
+--   Note: use macro __DUMMY_SORT__ !
+dummySort :: String -> Int -> Sort
+dummySort file line = DummyS $ file ++ ":" ++ show line
+
+-- | A dummy type created at location.
+--   Note: use macro __DUMMY_TYPE__ !
+dummyType :: String -> Int -> Type
+dummyType file line = El (DummyS "") $ dummyTerm' ("dummyType: " ++ file) line
+
+-- | Context entries without a type have this dummy type.
+--   Note: use macro __DUMMY_DOM__ !
+dummyDom :: String -> Int -> Dom Type
+dummyDom file line = defaultDom $ dummyType file line
+
+unreducedLevel :: Term -> Level
+unreducedLevel v = Max [ Plus 0 $ UnreducedLevel v ]
 
 -- | Top sort (Set\omega).
 topSort :: Type
@@ -697,6 +797,9 @@ sort s = El (UnivSort s) $ Sort s
 varSort :: Int -> Sort
 varSort n = Type $ Max [Plus 0 $ NeutralLevel mempty $ var n]
 
+tmSort :: Term -> Sort
+tmSort t = Type $ Max [Plus 0 $ UnreducedLevel t]
+
 levelSuc :: Level -> Level
 levelSuc (Max []) = Max [ClosedLevel 1]
 levelSuc (Max as) = Max $ map inc as
@@ -706,24 +809,20 @@ levelSuc (Max as) = Max $ map inc as
 mkType :: Integer -> Sort
 mkType n = Type $ Max [ClosedLevel n | n > 0]
 
+mkProp :: Integer -> Sort
+mkProp n = Prop $ Max [ClosedLevel n | n > 0]
+
 isSort :: Term -> Maybe Sort
 isSort v = case v of
   Sort s -> Just s
   _      -> Nothing
 
 impossibleTerm :: String -> Int -> Term
-impossibleTerm file line = Lit $ LitString noRange $ unlines
+impossibleTerm file line = Dummy $ unlines
   [ "An internal error has occurred. Please report this as a bug."
   , "Location of the error: " ++ file ++ ":" ++ show line
   ]
 
-hackReifyToMeta :: Term
-hackReifyToMeta = DontCare $ Lit $ LitNat noRange (-42)
-
-isHackReifyToMeta :: Term -> Bool
-isHackReifyToMeta (DontCare (Lit (LitNat r (-42)))) = r == noRange
-isHackReifyToMeta _ = False
-
 ---------------------------------------------------------------------------
 -- * Telescopes.
 ---------------------------------------------------------------------------
@@ -757,18 +856,22 @@ type ListTel = ListTel' ArgName
 telFromList' :: (a -> ArgName) -> ListTel' a -> Telescope
 telFromList' f = List.foldr extTel EmptyTel
   where
-    extTel (Dom info (x, a)) = ExtendTel (Dom info a) . Abs (f x)
+    extTel dom@(Dom{unDom = (x, a)}) = ExtendTel (dom{unDom = a}) . Abs (f x)
 
 -- | Convert a list telescope to a telescope.
 telFromList :: ListTel -> Telescope
 telFromList = telFromList' id
 
 -- | Convert a telescope to its list form.
-telToList :: Telescope -> ListTel
+telToList :: Tele (Dom t) -> [Dom (ArgName,t)]
 telToList EmptyTel                    = []
 telToList (ExtendTel arg (Abs x tel)) = fmap (x,) arg : telToList tel
 telToList (ExtendTel _    NoAbs{}   ) = __IMPOSSIBLE__
 
+-- | Lens to edit a 'Telescope' as a list.
+listTel :: Lens' ListTel Telescope
+listTel f = fmap telFromList . f . telToList
+
 -- | Drop the types from a telescope.
 class TelToArgs a where
   telToArgs :: a -> [Arg ArgName]
@@ -784,10 +887,10 @@ class SgTel a where
   sgTel :: a -> Telescope
 
 instance SgTel (ArgName, Dom Type) where
-  sgTel (x, dom) = ExtendTel dom $ Abs x EmptyTel
+  sgTel (x, !dom) = ExtendTel dom $ Abs x EmptyTel
 
 instance SgTel (Dom (ArgName, Type)) where
-  sgTel (Dom ai (x, t)) = ExtendTel (Dom ai t) $ Abs x EmptyTel
+  sgTel dom = ExtendTel (snd <$> dom) $ Abs (fst $ unDom dom) EmptyTel
 
 instance SgTel (Dom Type) where
   sgTel dom = sgTel (stringToArgName "_", dom)
@@ -822,17 +925,6 @@ arity t = case unEl t of
   Pi  _ b -> 1 + arity (unAbs b)
   _       -> 0
 
--- | Make a name that is not in scope.
-notInScopeName :: ArgName -> ArgName
-notInScopeName = stringToArgName . ("." ++) . argNameToString
-
--- | Remove the dot from a notInScopeName. This is used when printing function
---   types that have abstracted over not-in-scope names.
-unNotInScopeName :: ArgName -> ArgName
-unNotInScopeName = stringToArgName . undot . argNameToString
-  where undot ('.':s) = s
-        undot s       = s
-
 -- | Pick the better name suggestion, i.e., the one that is not just underscore.
 class Suggest a b where
   suggest :: a -> b -> String
@@ -894,16 +986,13 @@ hasElims v =
     Sort{}     -> Nothing
     Level{}    -> Nothing
     DontCare{} -> Nothing
-
--- | Drop 'Apply' constructor. (Unsafe!)
-argFromElim :: Elim' a -> Arg a
-argFromElim (Apply u) = u
-argFromElim Proj{}    = __IMPOSSIBLE__
+    Dummy{}    -> Nothing
 
 -- | Drop 'Apply' constructor. (Safe)
 isApplyElim :: Elim' a -> Maybe (Arg a)
 isApplyElim (Apply u) = Just u
 isApplyElim Proj{}    = Nothing
+isApplyElim (IApply _ _ r) = Just (defaultArg r)
 
 isApplyElim' :: Empty -> Elim' a -> Arg a
 isApplyElim' e = fromMaybe (absurd e) . isApplyElim
@@ -923,14 +1012,11 @@ class IsProjElim e where
 instance IsProjElim Elim where
   isProjElim (Proj o d) = Just (o, d)
   isProjElim Apply{}    = Nothing
-
--- | Discard @Proj f@ entries.
-dropProjElims :: IsProjElim e => [e] -> [e]
-dropProjElims = filter (isNothing . isProjElim)
+  isProjElim IApply{} = Nothing
 
 -- | Discards @Proj f@ entries.
 argsFromElims :: Elims -> Args
-argsFromElims = map argFromElim . dropProjElims
+argsFromElims = mapMaybe isApplyElim
 
 -- | Drop 'Proj' constructors. (Safe)
 allProjElims :: Elims -> Maybe [(ProjOrigin, QName)]
@@ -1013,16 +1099,19 @@ instance TermSize Term where
     Pi a b      -> 1 + tsize a + tsize b
     Sort s      -> tsize s
     DontCare mv -> tsize mv
+    Dummy{}     -> 1
 
 instance TermSize Sort where
   tsize s = case s of
     Type l    -> 1 + tsize l
-    Prop      -> 1
+    Prop l    -> 1 + tsize l
     Inf       -> 1
     SizeUniv  -> 1
     PiSort s s' -> 1 + tsize s + tsize s'
     UnivSort s -> 1 + tsize s
     MetaS _ es -> 1 + tsize es
+    DefS _ es  -> 1 + tsize es
+    DummyS{}   -> 1
 
 instance TermSize Level where
   tsize (Max as) = 1 + tsize as
@@ -1064,6 +1153,7 @@ instance KillRange Term where
     Pi a b      -> killRange2 Pi a b
     Sort s      -> killRange1 Sort s
     DontCare mv -> killRange1 DontCare mv
+    Dummy{}     -> v
 
 instance KillRange Level where
   killRange (Max as) = killRange1 Max as
@@ -1083,13 +1173,15 @@ instance (KillRange a) => KillRange (Type' a) where
 
 instance KillRange Sort where
   killRange s = case s of
-    Prop       -> Prop
     Inf        -> Inf
     SizeUniv   -> SizeUniv
     Type a     -> killRange1 Type a
+    Prop a     -> killRange1 Prop a
     PiSort s1 s2 -> killRange2 PiSort s1 s2
     UnivSort s -> killRange1 UnivSort s
     MetaS x es -> killRange1 (MetaS x) es
+    DefS d es  -> killRange2 DefS d es
+    DummyS{}   -> s
 
 instance KillRange Substitution where
   killRange IdS                  = IdS
@@ -1103,7 +1195,7 @@ instance KillRange PatOrigin where
   killRange = id
 
 instance KillRange ConPatternInfo where
-  killRange (ConPatternInfo mr mt lz) = killRange1 (ConPatternInfo mr) mt lz
+  killRange (ConPatternInfo mr b mt lz) = killRange1 (ConPatternInfo mr b) mt lz
 
 instance KillRange DBPatVar where
   killRange (DBPatVar x i) = killRange2 DBPatVar x i
@@ -1116,6 +1208,8 @@ instance KillRange a => KillRange (Pattern' a) where
       ConP con info ps -> killRange3 ConP con info ps
       LitP l           -> killRange1 LitP l
       ProjP o q        -> killRange1 (ProjP o) q
+      IApplyP o u t x  -> killRange3 (IApplyP o) u t x
+      DefP o q ps      -> killRange2 (DefP o) q ps
 
 instance KillRange Clause where
   killRange (Clause rl rf tel ps body t catchall unreachable) =
@@ -1153,10 +1247,10 @@ instance Pretty a => Pretty (Substitution' a) where
   prettyPrec p rho = pr p rho
     where
     pr p rho = case rho of
-      IdS              -> text "idS"
-      EmptyS err       -> text "emptyS"
-      t :# rho         -> mparens (p > 2) $ sep [ pr 2 rho P.<> text ",", prettyPrec 3 t ]
-      Strengthen _ rho -> mparens (p > 9) $ text "strS" <+> pr 10 rho
+      IdS              -> "idS"
+      EmptyS err       -> "emptyS"
+      t :# rho         -> mparens (p > 2) $ sep [ pr 2 rho P.<> ",", prettyPrec 3 t ]
+      Strengthen _ rho -> mparens (p > 9) $ "strS" <+> pr 10 rho
       Wk n rho         -> mparens (p > 9) $ text ("wkS " ++ show n) <+> pr 10 rho
       Lift n rho       -> mparens (p > 9) $ text ("liftS " ++ show n) <+> pr 10 rho
 
@@ -1166,21 +1260,22 @@ instance Pretty Term where
       Var x els -> text ("@" ++ show x) `pApp` els
       Lam ai b   ->
         mparens (p > 0) $
-        sep [ text "λ" <+> prettyHiding ai id (text . absName $ b) <+> text "->"
+        sep [ "λ" <+> prettyHiding ai id (text . absName $ b) <+> "->"
             , nest 2 $ pretty (unAbs b) ]
       Lit l                -> pretty l
       Def q els            -> pretty q `pApp` els
       Con c ci vs          -> pretty (conName c) `pApp` vs
       Pi a (NoAbs _ b)     -> mparens (p > 0) $
-        sep [ prettyPrec 1 (unDom a) <+> text "->"
+        sep [ prettyPrec 1 (unDom a) <+> "->"
             , nest 2 $ pretty b ]
       Pi a b               -> mparens (p > 0) $
-        sep [ pDom (domInfo a) (text (absName b) <+> text ":" <+> pretty (unDom a)) <+> text "->"
+        sep [ pDom (domInfo a) (text (absName b) <+> ":" <+> pretty (unDom a)) <+> "->"
             , nest 2 $ pretty (unAbs b) ]
       Sort s      -> prettyPrec p s
       Level l     -> prettyPrec p l
       MetaV x els -> pretty x `pApp` els
       DontCare v  -> prettyPrec p v
+      Dummy s     -> parens $ text s
     where
       pApp d els = mparens (not (null els) && p > 9) $
                    sep [d, nest 2 $ fsep (map (prettyPrec 10) els)]
@@ -1194,16 +1289,16 @@ pDom i =
 
 instance Pretty Clause where
   pretty Clause{clauseTel = tel, namedClausePats = ps, clauseBody = b, clauseType = t} =
-    sep [ pretty tel <+> text "|-"
-        , nest 2 $ sep [ fsep (map (prettyPrec 10) ps) <+> text "="
+    sep [ pretty tel <+> "|-"
+        , nest 2 $ sep [ fsep (map (prettyPrec 10) ps) <+> "="
                        , nest 2 $ pBody b t ] ]
     where
-      pBody Nothing _ = text "(absurd)"
+      pBody Nothing _ = "(absurd)"
       pBody (Just b) Nothing  = pretty b
-      pBody (Just b) (Just t) = sep [ pretty b <+> text ":", nest 2 $ pretty t ]
+      pBody (Just b) (Just t) = sep [ pretty b <+> ":", nest 2 $ pretty t ]
 
 instance Pretty a => Pretty (Tele (Dom a)) where
-  pretty tel = fsep [ pDom a (text x <+> text ":" <+> pretty (unDom a)) | (x, a) <- telToList tel ]
+  pretty tel = fsep [ pDom a (text x <+> ":" <+> pretty (unDom a)) | (x, a) <- telToList tel ]
     where
       telToList EmptyTel = []
       telToList (ExtendTel a tel) = (absName tel, a) : telToList (unAbs tel)
@@ -1213,16 +1308,16 @@ instance Pretty Level where
     case as of
       []  -> prettyPrec p (ClosedLevel 0)
       [a] -> prettyPrec p a
-      _   -> mparens (p > 9) $ List.foldr1 (\a b -> text "lub" <+> a <+> b) $ map (prettyPrec 10) as
+      _   -> mparens (p > 9) $ List.foldr1 (\a b -> "lub" <+> a <+> b) $ map (prettyPrec 10) as
 
 instance Pretty PlusLevel where
   prettyPrec p l =
     case l of
-      ClosedLevel n -> sucs p n $ \_ -> text "lzero"
+      ClosedLevel n -> sucs p n $ \_ -> "lzero"
       Plus n a      -> sucs p n $ \p -> prettyPrec p a
     where
       sucs p 0 d = d p
-      sucs p n d = mparens (p > 9) $ text "lsuc" <+> sucs 10 (n - 1) d
+      sucs p n d = mparens (p > 9) $ "lsuc" <+> sucs 10 (n - 1) d
 
 instance Pretty LevelAtom where
   prettyPrec p a =
@@ -1235,18 +1330,22 @@ instance Pretty LevelAtom where
 instance Pretty Sort where
   prettyPrec p s =
     case s of
-      Type (Max []) -> text "Set"
+      Type (Max []) -> "Set"
       Type (Max [ClosedLevel n]) -> text $ "Set" ++ show n
-      Type l -> mparens (p > 9) $ text "Set" <+> prettyPrec 10 l
-      Prop -> text "Prop"
-      Inf -> text "Setω"
-      SizeUniv -> text "SizeUniv"
+      Type l -> mparens (p > 9) $ "Set" <+> prettyPrec 10 l
+      Prop (Max []) -> "Prop"
+      Prop (Max [ClosedLevel n]) -> text $ "Prop" ++ show n
+      Prop l -> mparens (p > 9) $ "Prop" <+> prettyPrec 10 l
+      Inf -> "Setω"
+      SizeUniv -> "SizeUniv"
       PiSort s b -> mparens (p > 9) $
-        text "piSort" <+> prettyPrec 10 s
+        "piSort" <+> prettyPrec 10 s
                       <+> parens (sep [ text ("λ " ++ absName b ++ " ->")
                                       , nest 2 $ pretty (unAbs b) ])
-      UnivSort s -> mparens (p > 9) $ text "univSort" <+> prettyPrec 10 s
+      UnivSort s -> mparens (p > 9) $ "univSort" <+> prettyPrec 10 s
       MetaS x es -> prettyPrec p $ MetaV x es
+      DefS d es  -> prettyPrec p $ Def d es
+      DummyS s   -> parens $ text s
 
 instance Pretty Type where
   prettyPrec p (El _ a) = prettyPrec p a
@@ -1254,25 +1353,29 @@ instance Pretty Type where
 instance Pretty tm => Pretty (Elim' tm) where
   prettyPrec p (Apply v)    = prettyPrec p v
   prettyPrec _ (Proj _o x)  = text ("." ++ prettyShow x)
+  prettyPrec p (IApply x y r) = prettyPrec p r
 
 instance Pretty DBPatVar where
   prettyPrec _ x = text $ patVarNameToString (dbPatVarName x) ++ "@" ++ show (dbPatVarIndex x)
 
 instance Pretty a => Pretty (Pattern' a) where
   prettyPrec n (VarP _o x)   = prettyPrec n x
-  prettyPrec _ (DotP _o t)   = text "." P.<> prettyPrec 10 t
+  prettyPrec _ (DotP _o t)   = "." P.<> prettyPrec 10 t
   prettyPrec n (ConP c i nps)= mparens (n > 0 && not (null nps)) $
     pretty (conName c) <+> fsep (map (prettyPrec 10) ps)
     where ps = map (fmap namedThing) nps
+  prettyPrec n (DefP o q nps)= mparens (n > 0 && not (null nps)) $
+    pretty q <+> fsep (map (prettyPrec 10) ps)
+    where ps = map (fmap namedThing) nps
   -- -- Version with printing record type:
   -- prettyPrec _ (ConP c i ps) = (if b then braces else parens) $ prTy $
   --   text (show $ conName c) <+> fsep (map (pretty . namedArg) ps)
   --   where
   --     b = maybe False (== ConOSystem) $ conPRecord i
-  --     prTy d = caseMaybe (conPType i) d $ \ t -> d  <+> text ":" <+> pretty t
+  --     prTy d = caseMaybe (conPType i) d $ \ t -> d  <+> ":" <+> pretty t
   prettyPrec _ (LitP l)      = pretty l
   prettyPrec _ (ProjP _o q)  = text ("." ++ prettyShow q)
-
+  prettyPrec n (IApplyP _o _ _ x) = prettyPrec n x
 -----------------------------------------------------------------------------
 -- * NFData instances
 -----------------------------------------------------------------------------
@@ -1291,6 +1394,7 @@ instance NFData Term where
     Level l    -> rnf l
     MetaV _ es -> rnf es
     DontCare v -> rnf v
+    Dummy _    -> ()
 
 instance NFData Type where
   rnf (El s v) = rnf (s, v)
@@ -1298,12 +1402,14 @@ instance NFData Type where
 instance NFData Sort where
   rnf s = case s of
     Type l   -> rnf l
-    Prop     -> ()
+    Prop l   -> rnf l
     Inf      -> ()
     SizeUniv -> ()
     PiSort a b -> rnf (a, unAbs b)
     UnivSort a -> rnf a
     MetaS _ es -> rnf es
+    DefS _ es  -> rnf es
+    DummyS _   -> ()
 
 instance NFData Level where
   rnf (Max as) = rnf as
@@ -1321,3 +1427,4 @@ instance NFData LevelAtom where
 instance NFData a => NFData (Elim' a) where
   rnf (Apply x) = rnf x
   rnf Proj{}    = ()
+  rnf (IApply x y r) = rnf x `seq` rnf y `seq` rnf r
