@@ -67,6 +67,7 @@ instance GetDefs Term where
     Level l    -> getDefs l
     MetaV x vs -> getDefs x >> getDefs vs
     DontCare v -> getDefs v
+    Dummy{}    -> return ()
 
 instance GetDefs MetaId where
   getDefs x = doMeta x
@@ -77,12 +78,14 @@ instance GetDefs Type where
 instance GetDefs Sort where
   getDefs s = case s of
     Type l    -> getDefs l
-    Prop      -> return ()
+    Prop l    -> getDefs l
     Inf       -> return ()
     SizeUniv  -> return ()
     PiSort s s' -> getDefs s >> getDefs s'
     UnivSort s  -> getDefs s
     MetaS x es  -> getDefs x >> getDefs es
+    DefS d es   -> doDef d >> getDefs es
+    DummyS{}    -> return ()
 
 instance GetDefs Level where
   getDefs (Max ls) = getDefs ls
