@@ -8,14 +8,17 @@ module Agda.Syntax.Parser
       -- * Parse functions
     , Agda.Syntax.Parser.parse
     , Agda.Syntax.Parser.parsePosString
-    , parseFile'
+    , parseFile
       -- * Parsers
     , moduleParser
     , moduleNameParser
+    , acceptableFileExts
     , exprParser
     , exprWhereParser
     , holeContentParser
     , tokensParser
+      -- * Reading files.
+    , readFilePM
       -- * Parse errors
     , ParseError(..)
     , ParseWarning(..)
@@ -31,7 +34,10 @@ import Control.Monad.Reader
 import Control.Monad.Writer hiding ((<>))
 
 import qualified Data.List as List
+import Data.Text.Lazy (Text)
+import qualified Data.Text.Lazy as T
 
+import Agda.Syntax.Common
 import Agda.Syntax.Position
 import Agda.Syntax.Parser.Monad as M hiding (Parser, parseFlags)
 import qualified Agda.Syntax.Parser.Monad as M
@@ -99,12 +105,6 @@ type LiterateParser a = Parser a -> [Layer] -> PM a
 parse :: Parser a -> String -> PM a
 parse p = wrapM . return . M.parse (parseFlags p) [normal] (parser p)
 
-parseFile :: Parser a -> AbsolutePath -> PM a
-parseFile p = wrapM . M.parseFile (parseFlags p) [layout, normal] (parser p)
-
-parseString :: Parser a -> String -> PM a
-parseString = parseStringFromFile Strict.Nothing
-
 parseStringFromFile :: SrcFile -> Parser a -> String -> PM a
 parseStringFromFile src p = wrapM . return . M.parseFromSrc (parseFlags p) [layout, normal] (parser p) src
 
@@ -122,36 +122,58 @@ parseLiterateWithComments p layers = do
   return$ concat [ case m of
                        Left t -> [t]
                        Right (Layer Comment interval s) -> [TokTeX (interval, s)]
-                       Right (Layer Markup _ _) -> []
+                       Right (Layer Markup interval s) -> [TokMarkup (interval, s)]
                        Right (Layer Code _ _) -> []
                    | m <- terms ]
 
-readFilePM :: AbsolutePath -> PM String
-readFilePM path = wrapIOM (ReadFileError path) (readTextFile (filePath path))
+-- | Returns the contents of the given file.
 
-parseLiterateFile :: Processor -> Parser a -> AbsolutePath -> PM a
-parseLiterateFile po p path = readFilePM path >>= parseLiterate p p . po (startPos (Just path))
+readFilePM :: AbsolutePath -> PM Text
+readFilePM path = wrapIOM (ReadFileError path) (readTextFile $ filePath path)
+
+parseLiterateFile
+  :: Processor
+  -> Parser a
+  -> AbsolutePath
+     -- ^ The path to the file.
+  -> String
+     -- ^ The file contents. Note that the file is /not/ read from
+     -- disk.
+  -> PM a
+parseLiterateFile po p path = parseLiterate p p . po (startPos (Just path))
 
 parsePosString :: Parser a -> Position -> String -> PM a
 parsePosString p pos = wrapM . return . M.parsePosString pos (parseFlags p) [normal] (parser p)
 
--- | Extensions supported by `parseFile'`
-parseFileExts :: [String]
-parseFileExts = ".agda":literateExts
+-- | Extensions supported by `parseFile`.
 
-parseFile' :: (Show a) => Parser a -> AbsolutePath -> PM a
-parseFile' p file =
+acceptableFileExts :: [String]
+acceptableFileExts = ".agda" : (fst <$> literateProcessors)
+
+parseFile
+  :: Show a
+  => Parser a
+  -> AbsolutePath
+     -- ^ The path to the file.
+  -> String
+     -- ^ The file contents. Note that the file is /not/ read from
+     -- disk.
+  -> PM (a, FileType)
+parseFile p file input =
   if ".agda" `List.isSuffixOf` filePath file then
-    Agda.Syntax.Parser.parseFile p file
+    (, AgdaFileType) <$> Agda.Syntax.Parser.parseStringFromFile
+                         (Strict.Just file) p input
   else
     go literateProcessors
   where
-    go [] = throwError InvalidExtensionError {
-                     errPath = file
-                   , errValidExts = parseFileExts
+    go [] = throwError InvalidExtensionError
+                   { errPath = file
+                   , errValidExts = acceptableFileExts
                    }
-    go ((ext, po):pos) | ext `List.isSuffixOf` filePath file = parseLiterateFile po p file
-    go (_:pos) = go pos
+    go ((ext, (po, ft)) : pos)
+      | ext `List.isSuffixOf` filePath file =
+          (, ft) <$> parseLiterateFile po p file input
+      | otherwise = go pos
 
 ------------------------------------------------------------------------
 -- Specific parsers
