@@ -19,10 +19,8 @@ data Keyword
         | KwEta | KwNoEta
         | KwHiding | KwUsing | KwRenaming | KwTo | KwPublic
         | KwOPTIONS | KwBUILTIN | KwLINE
-        | KwCOMPILED_DATA | KwCOMPILED_TYPE | KwCOMPILED | KwCOMPILED_EXPORT
-        | KwHASKELL | KwFOREIGN | KwCOMPILE
-        | KwCOMPILED_JS | KwCOMPILED_UHC | KwCOMPILED_DATA_UHC
-        | KwIMPORT | KwIMPORT_UHC | KwIMPOSSIBLE | KwSTATIC | KwINJECTIVE | KwINLINE | KwNOINLINE
+        | KwFOREIGN | KwCOMPILE
+        | KwIMPOSSIBLE | KwSTATIC | KwINJECTIVE | KwINLINE | KwNOINLINE
         | KwETA
         | KwNO_TERMINATION_CHECK | KwTERMINATING | KwNON_TERMINATING
         | KwWARNING_ON_USAGE
@@ -32,12 +30,14 @@ data Keyword
         | KwUnquote | KwUnquoteDecl | KwUnquoteDef
         | KwSyntax
         | KwPatternSyn | KwTactic | KwCATCHALL
+        | KwVariable
         | KwNO_POSITIVITY_CHECK | KwPOLARITY
+        | KwNO_UNIVERSE_CHECK
     deriving (Eq, Show)
 
 layoutKeywords :: [Keyword]
 layoutKeywords =
-    [ KwLet, KwWhere, KwDo, KwPostulate, KwMutual, KwAbstract, KwPrivate, KwInstance, KwMacro, KwPrimitive, KwField ]
+    [ KwLet, KwWhere, KwDo, KwPostulate, KwMutual, KwAbstract, KwPrivate, KwInstance, KwMacro, KwPrimitive, KwField, KwVariable ]
 
 data Symbol
         = SymDot | SymSemi | SymVirtualSemi | SymBar
@@ -67,11 +67,13 @@ data Token
           -- Other tokens
         | TokString (Interval, String)  -- arbitrary string, used in pragmas
         | TokSetN (Interval, Integer)
+        | TokPropN (Interval, Integer)
         | TokTeX (Interval, String)
+        | TokMarkup (Interval, String)
         | TokComment (Interval, String)
         | TokDummy      -- Dummy token to make Happy not complain
                         -- about overlapping cases.
-        | TokEOF
+        | TokEOF Interval
     deriving (Eq, Show)
 
 instance HasRange Token where
@@ -82,7 +84,9 @@ instance HasRange Token where
   getRange (TokSymbol _ i)     = getRange i
   getRange (TokString (i, _))  = getRange i
   getRange (TokSetN (i, _))    = getRange i
+  getRange (TokPropN (i, _))   = getRange i
   getRange (TokTeX (i, _))     = getRange i
+  getRange (TokMarkup (i, _))  = getRange i
   getRange (TokComment (i, _)) = getRange i
   getRange TokDummy            = noRange
-  getRange TokEOF              = noRange
+  getRange (TokEOF i)          = getRange i
