@@ -25,6 +25,7 @@ import qualified Data.Map as Map
 import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Maybe
+import qualified Data.Semigroup as Sgrp
 
 import Data.Data (Data)
 
@@ -36,6 +37,7 @@ import Agda.Syntax.Fixity
 import Agda.Syntax.Abstract.Name as A
 import Agda.Syntax.Concrete.Name as C
 import qualified Agda.Syntax.Concrete as C
+import Agda.Syntax.Concrete.Fixity as C
 
 import Agda.Utils.AssocList (AssocList)
 import qualified Agda.Utils.AssocList as AssocList
@@ -45,6 +47,7 @@ import Agda.Utils.List
 import Agda.Utils.NonemptyList
 import Agda.Utils.Null
 import Agda.Utils.Pretty
+import Agda.Utils.Singleton
 import qualified Agda.Utils.Map as Map
 
 #include "undefined.h"
@@ -106,14 +109,20 @@ data ScopeInfo = ScopeInfo
       , scopeVarsToBind    :: LocalVars
       , scopeLocals        :: LocalVars
       , scopePrecedence    :: PrecedenceStack
-      , scopeInverseName   :: Map A.QName [C.QName]
-      , scopeInverseModule :: Map A.ModuleName [C.QName]
+      , scopeInverseName   :: NameMap
+      , scopeInverseModule :: ModuleMap
       , scopeInScope       :: InScopeSet
+      , scopeFixities      :: C.Fixities    -- ^ Maps concrete names to fixities
+      , scopePolarities    :: C.Polarities  -- ^ Maps concrete names to polarities
       }
   deriving (Data, Show)
 
+type NameMap   = Map A.QName      (NonemptyList C.QName)
+type ModuleMap = Map A.ModuleName [C.QName]
+-- type ModuleMap = Map A.ModuleName (NonemptyList C.QName)
+
 instance Eq ScopeInfo where
-  ScopeInfo c1 m1 v1 l1 p1 _ _ _ == ScopeInfo c2 m2 v2 l2 p2 _ _ _ =
+  ScopeInfo c1 m1 v1 l1 p1 _ _ _ _ _ == ScopeInfo c2 m2 v2 l2 p2 _ _ _ _ _ =
     c1 == c2 && m1 == m2 && v1 == v2 && l1 == l2 && p1 == p2
 
 -- | Local variables.
@@ -150,13 +159,19 @@ instance Ord LocalVar where
 -- | We show shadowed variables as prefixed by a ".", as not in scope.
 instance Pretty LocalVar where
   pretty (LocalVar x _ []) = pretty x
-  pretty (LocalVar x _ xs) = text "." <> pretty x
+  pretty (LocalVar x _ xs) = "." <> pretty x
 
 -- | Shadow a local name by a non-empty list of imports.
 shadowLocal :: [AbstractName] -> LocalVar -> LocalVar
 shadowLocal [] _ = __IMPOSSIBLE__
 shadowLocal ys (LocalVar x b zs) = LocalVar x b (ys ++ zs)
 
+-- | Treat patternBound variable as a module parameter
+patternToModuleBound :: LocalVar -> LocalVar
+patternToModuleBound x
+ | localBinder x == PatternBound = x { localBinder = LambdaBound }
+ | otherwise                     = x
+
 -- | Project name of unshadowed local variable.
 notShadowedLocal :: LocalVar -> Maybe A.Name
 notShadowedLocal (LocalVar x _ []) = Just x
@@ -245,6 +260,8 @@ data KindOfName
   | FldName        -- ^ Record field name.
   | DefName        -- ^ Ordinary defined name.
   | PatternSynName -- ^ Name of a pattern synonym.
+  | GeneralizeName -- ^ Name to be generalized
+  | DisallowedGeneralizeName -- ^ Generalizable variable from a let open
   | MacroName      -- ^ Name of a macro
   | QuotableName   -- ^ A name that can only be quoted.
   deriving (Eq, Show, Data, Enum, Bounded)
@@ -274,9 +291,16 @@ data AbstractName = AbsName
     -- ^ The kind (definition, constructor, record field etc.).
   , anameLineage :: WhyInScope
     -- ^ Explanation where this name came from.
+  , anameMetadata :: NameMetadata
+    -- ^ Additional information needed during scope checking. Currently used
+    --   for generalized data/record params.
   }
   deriving (Data, Show)
 
+data NameMetadata = NoMetadata
+                  | GeneralizedVarsMetadata (Map A.QName A.Name)
+  deriving (Data, Show)
+
 -- | A decoration of abstract syntax module names.
 data AbstractModule = AbsModule
   { amodName    :: A.ModuleName
@@ -332,12 +356,12 @@ data ResolvedName
 
 instance Pretty ResolvedName where
   pretty = \case
-    VarName x _         -> text "variable"    <+> pretty x
+    VarName x _         -> "variable"    <+> pretty x
     DefinedName a x     -> pretty a           <+> pretty x
-    FieldName xs        -> text "field"       <+> pretty xs
-    ConstructorName xs  -> text "constructor" <+> pretty xs
-    PatternSynResName x -> text "pattern"     <+> pretty x
-    UnknownName         -> text "<unknown name>"
+    FieldName xs        -> "field"       <+> pretty xs
+    ConstructorName xs  -> "constructor" <+> pretty xs
+    PatternSynResName x -> "pattern"     <+> pretty x
+    UnknownName         -> "<unknown name>"
 
 -- * Operations on name and module maps.
 
@@ -410,6 +434,8 @@ emptyScopeInfo = ScopeInfo
   , scopeInverseName   = Map.empty
   , scopeInverseModule = Map.empty
   , scopeInScope       = Set.empty
+  , scopeFixities      = Map.empty
+  , scopePolarities    = Map.empty
   }
 
 -- | Map functions over the names and modules in a scope.
@@ -671,6 +697,14 @@ restrictLocalPrivate m = mapScope' PrivateNS (Map.mapMaybe rName) (Map.mapMaybe
 removeOnlyQualified :: Scope -> Scope
 removeOnlyQualified s = setNameSpace OnlyQualifiedNS emptyNameSpace s
 
+-- | Disallow using generalized variables from the scope
+disallowGeneralizedVars :: Scope -> Scope
+disallowGeneralizedVars = mapScope_ ((fmap . map) disallow) id id
+  where
+    disallow a = a { anameKind = disallowGen (anameKind a) }
+    disallowGen GeneralizeName = DisallowedGeneralizeName
+    disallowGen k              = k
+
 -- | Add an explanation to why things are in scope.
 inScopeBecause :: (WhyInScope -> WhyInScope) -> Scope -> Scope
 inScopeBecause f = mapScope_ mapName mapMod id
@@ -705,6 +739,27 @@ everythingInScope scope = allThingsInScope $ mergeScopes $
     look m = fromMaybe __IMPOSSIBLE__ $ Map.lookup m $ scopeModules scope
     s0     = look $ scopeCurrent scope
 
+everythingInScopeQualified :: ScopeInfo -> NameSpace
+everythingInScopeQualified scope =
+  allThingsInScope $ mergeScopes $
+    chase Set.empty scopes
+  where
+    s0      = look $ scopeCurrent scope
+    scopes  = s0 : map look (scopeParents s0)
+    look m  = fromMaybe __IMPOSSIBLE__ $ Map.lookup m $ scopeModules scope
+    lookP   = restrictPrivate . look
+
+    -- We start with the current module and all its parents and look through
+    -- all their imports and submodules.
+    chase seen [] = []
+    chase seen (s : ss)
+      | Set.member name seen = chase seen ss
+      | otherwise = s : chase (Set.insert name seen) (imports ++ submods ++ ss)
+      where
+        name    = scopeName s
+        imports = map lookP $ Map.elems $ scopeImports s
+        submods = map (lookP . amodName) $ concat $ Map.elems $ allNamesInScope s
+
 -- | Compute a flattened scope. Only include unqualified names or names
 -- qualified by modules in the first argument.
 flattenScope :: [[C.Name]] -> ScopeInfo -> Map C.QName [AbstractName]
@@ -877,8 +932,8 @@ inverseScopeLookup' amb name scope = billToPure [ Scoping , InverseScopeLookup ]
     Left  m -> best $ filter unambiguousModule $ findModule m
     Right q -> best $ filter unambiguousName   $ findName q
   where
-    findName   x = fromMaybe [] $ Map.lookup x (scopeInverseName scope)
-    findModule x = fromMaybe [] $ Map.lookup x (scopeInverseModule scope)
+    findName   x = maybe [] toList $ Map.lookup x (scopeInverseName scope)
+    findModule x = fromMaybe []    $ Map.lookup x (scopeInverseModule scope)
 
     len :: C.QName -> Int
     len (C.QName _)  = 1
@@ -903,7 +958,7 @@ recomputeInverseScopeMaps :: ScopeInfo -> ScopeInfo
 recomputeInverseScopeMaps scope = billToPure [ Scoping , InverseScopeLookup ] $
   scope { scopeInverseName   = nameMap
         , scopeInverseModule = Map.fromList [ (x, findModule x) | x <- Map.keys moduleMap ++ Map.keys importMap ]
-        , scopeInScope       = Set.unions $ map (scopeSet . snd) scopes
+        , scopeInScope       = nsInScope $ everythingInScopeQualified scope
         }
   where
     this = scopeCurrent scope
@@ -916,15 +971,12 @@ recomputeInverseScopeMaps scope = billToPure [ Scoping , InverseScopeLookup ] $
     restrict m s | m `elem` current = s
                  | otherwise = restrictPrivate s
 
-    scopeSet     s = Set.unions $ map (namespaceSet . snd) $ scopeNameSpaces s
-    namespaceSet s = nsInScope s
-
     internalName :: C.QName -> Bool
     internalName C.QName{} = False
     internalName (C.Qual m n) = intern m || internalName n
       where
       -- Recognize fresh names created Parser.y
-      intern (C.Name _ [ C.Id ('.' : '#' : _) ]) = True
+      intern (C.Name _ _ [ C.Id ('.' : '#' : _) ]) = True
       intern _ = False
 
     findName :: Ord a => Map a [(A.ModuleName, C.Name)] -> a -> [C.QName]
@@ -945,25 +997,26 @@ recomputeInverseScopeMaps scope = billToPure [ Scoping , InverseScopeLookup ] $
     importMap = Map.fromListWith (++) $ do
       (m, s) <- scopes
       (x, y) <- Map.toList $ scopeImports s
-      return (y, [x])
+      return (y, singleton x)
 
     moduleMap = Map.fromListWith (++) $ do
       (m, s)  <- scopes
       (x, ms) <- Map.toList (allNamesInScope s)
       q       <- amodName <$> ms
-      return (q, [(m, x)])
+      return (q, singleton (m, x))
 
-    nameMap = Map.fromListWith (++) $ do
+    nameMap :: NameMap
+    nameMap = Map.fromListWith (Sgrp.<>) $ do
       (m, s)  <- scopes
       (x, ms) <- Map.toList (allNamesInScope s)
       q       <- anameName <$> ms
       if elem m current
-        then return (q, [C.QName x])
+        then return (q, singleton (C.QName x))
         else do
           y <- findModule m
           let z = C.qualify y x
           guard $ not $ internalName z
-          return (q, [z])
+          return (q, singleton z)
 
 -- | Find the concrete names that map (uniquely) to a given abstract qualified name.
 --   Sort by length, shortest first.
@@ -1000,19 +1053,19 @@ instance Pretty NameSpace where
 
 prettyNameSpace :: NameSpace -> [Doc]
 prettyNameSpace (NameSpace names mods _) =
-    blockOfLines (text "names")   (map pr $ Map.toList names) ++
-    blockOfLines (text "modules") (map pr $ Map.toList mods)
+    blockOfLines "names"   (map pr $ Map.toList names) ++
+    blockOfLines "modules" (map pr $ Map.toList mods)
   where
     pr :: (Pretty a, Pretty b) => (a,b) -> Doc
-    pr (x, y) = pretty x <+> text "-->" <+> pretty y
+    pr (x, y) = pretty x <+> "-->" <+> pretty y
 
 instance Pretty Scope where
   pretty (scope@Scope{ scopeName = name, scopeParents = parents, scopeImports = imps }) =
     vcat $
-      [ text "scope" <+> pretty name ] ++ ind (
+      [ "scope" <+> pretty name ] ++ ind (
         concat [ blockOfLines (pretty nsid) $ prettyNameSpace $ scopeNameSpace nsid scope
                | nsid <- [minBound..maxBound] ]
-      ++ blockOfLines (text "imports")
+      ++ blockOfLines "imports"
            (case Map.keys imps of [] -> []; ks -> [ prettyList ks ])
       )
     where ind = map $ nest 2
@@ -1023,14 +1076,14 @@ blockOfLines _  [] = []
 blockOfLines hd ss = hd : map (nest 2) ss
 
 instance Pretty ScopeInfo where
-  pretty (ScopeInfo this mods toBind locals ctx _ _ _) = vcat $
-    [ text "ScopeInfo"
-    , text "  current = " <> pretty this
+  pretty (ScopeInfo this mods toBind locals ctx _ _ _ _ _) = vcat $
+    [ "ScopeInfo"
+    , "  current = " <> pretty this
     ] ++
-    (if null toBind then [] else [ text "  toBind  = " <> pretty locals ]) ++
-    (if null locals then [] else [ text "  locals  = " <> pretty locals ]) ++
-    [ text "  context = " <> pretty ctx
-    , text "  modules"
+    (if null toBind then [] else [ "  toBind  = " <> pretty locals ]) ++
+    (if null locals then [] else [ "  locals  = " <> pretty locals ]) ++
+    [ "  context = " <> pretty ctx
+    , "  modules"
     ] ++
     map (nest 4) (List.filter (not . null) $ map pretty $ Map.elems mods)
 
