@@ -99,6 +99,8 @@ apps e = elims e . map I.Apply
 -- | Drops hidden arguments unless --show-implicit.
 nelims :: Expr -> [I.Elim' (Named_ Expr)] -> TCM Expr
 nelims e [] = return e
+nelims e (I.IApply x y r : es) =
+  nelims (A.App defaultAppInfo_ e $ defaultArg r) es
 nelims e (I.Apply arg : es) = do
   arg <- reify arg  -- This replaces the arg by _ if irrelevant
   dontShowImp <- not <$> showImplicitArguments
@@ -106,7 +108,7 @@ nelims e (I.Apply arg : es) = do
          | otherwise                     = A.App defaultAppInfo_ e arg
   nelims hd es
 nelims e (I.Proj ProjPrefix d : es)             = nelimsProjPrefix e d es
-nelims e (I.Proj o          d : es) | isSelf e  = nelimsProjPrefix e d es
+nelims e (I.Proj o          d : es) | isSelf e  = nelims (A.Proj ProjPrefix $ unambiguous d) es
                                     | otherwise =
   nelims (A.App defaultAppInfo_ e (defaultNamedArg $ A.Proj o $ unambiguous d)) es
 
@@ -114,15 +116,11 @@ nelimsProjPrefix :: Expr -> QName -> [I.Elim' (Named_ Expr)] -> TCM Expr
 nelimsProjPrefix e d es =
   nelims (A.App defaultAppInfo_ (A.Proj ProjPrefix $ unambiguous d) $ defaultNamedArg e) es
 
--- | If we are inside a record definition, the record value (self)
---   is a variable with name "".
---   In this case, we need to use a prefix-projection.  (Issue #2868.)
---   Prefix projections magically print correctly since the thing
---   we are projecting from has a null name, so vanishes in the visible world.
---   We love hacks, don't we?  Sigh.
+-- | If we are referencing the record from inside the record definition, we don't insert an
+-- | A.App
 isSelf :: Expr -> Bool
 isSelf = \case
-  A.Var x -> null $ prettyShow x
+  A.Var n -> nameIsRecordName n
   _ -> False
 
 -- | Drops hidden arguments unless --show-implicit.
@@ -160,7 +158,7 @@ instance Reify Expr Expr where
 instance Reify MetaId Expr where
     reifyWhen = reifyWhenE
     reify x@(MetaId n) = liftTCM $ do
-      b <- asks envPrintMetasBare
+      b <- asksTC envPrintMetasBare
       mi  <- mvInfo <$> lookupMeta x
       let mi' = Info.MetaInfo
                  { metaRange          = getRange $ miClosRange mi
@@ -169,8 +167,17 @@ instance Reify MetaId Expr where
                  , metaNameSuggestion = if b then "" else miNameSuggestion mi
                  }
           underscore = return $ A.Underscore mi'
-      caseMaybeM (isInteractionMeta x) underscore $ \ ii@InteractionId{} ->
+      -- If we are printing a term that will be pasted into the user
+      -- source, we turn all unsolved (non-interaction) metas into
+      -- interaction points
+      isInteractionMeta x >>= \case
+        Nothing | b -> do
+          ii <- registerInteractionPoint False noRange Nothing
+          connectInteractionPoint ii x
           return $ A.QuestionMark mi' ii
+        Just ii | b -> underscore
+        Nothing     -> underscore
+        Just ii     -> return $ A.QuestionMark mi' ii
 
 -- Does not print with-applications correctly:
 -- instance Reify DisplayTerm Expr where
@@ -230,8 +237,8 @@ reifyDisplayFormP f ps wps = do
         reportSDoc "reify.display" 70 $ do
           doc <- prettyA $ SpineLHS empty f' (ps' ++ wps' ++ wps)
           return $ vcat
-            [ text "rewritten lhs to"
-            , text "  lhs' = " <+> doc
+            [ "rewritten lhs to"
+            , "  lhs' = " <+> doc
             ]
         reifyDisplayFormP f' ps' (wps' ++ wps)
       _ -> do
@@ -263,12 +270,14 @@ reifyDisplayFormP f ps wps = do
     okDisplayTerm _         = False
 
     okDElim :: Elim' DisplayTerm -> Bool
+    okDElim (I.IApply x y r) = okDisplayTerm r
     okDElim (I.Apply v) = okDisplayTerm $ unArg v
     okDElim I.Proj{}    = True
 
     okToDropE :: Elim' Term -> Bool
     okToDropE (I.Apply v) = okToDrop v
     okToDropE I.Proj{}    = False
+    okToDropE (I.IApply x y r) = False
 
     okToDrop :: Arg I.Term -> Bool
     okToDrop arg = notVisible arg && case unArg arg of
@@ -281,6 +290,7 @@ reifyDisplayFormP f ps wps = do
     okArg = okTerm . unArg
 
     okElim :: Elim' I.Term -> Bool
+    okElim (I.IApply x y r) = okTerm r
     okElim (I.Apply a) = okArg a
     okElim (I.Proj{})  = True
 
@@ -313,6 +323,7 @@ reifyDisplayFormP f ps wps = do
         argToPat arg = traverse termToPat arg
 
         elimToPat :: I.Elim' DisplayTerm -> TCM (NamedArg A.Pattern)
+        elimToPat (I.IApply _ _ r) = argToPat (Arg defaultArgInfo r)
         elimToPat (I.Apply arg) = argToPat arg
         elimToPat (I.Proj o d)  = return $ defaultNamedArg $ A.ProjP patNoRange o $ unambiguous d
 
@@ -323,14 +334,16 @@ reifyDisplayFormP f ps wps = do
         termToPat (DTerm (I.Var n [])) = return $ unArg $ fromMaybe __IMPOSSIBLE__ $ ps !!! n
 
         termToPat (DCon c ci vs)          = fmap unnamed <$> tryRecPFromConP =<< do
-           A.ConP (ConPatInfo ci patNoRange False) (unambiguous (conName c)) <$> mapM argToPat vs
+           A.ConP (ConPatInfo ci patNoRange ConPatEager) (unambiguous (conName c)) <$> mapM argToPat vs
 
         termToPat (DTerm (I.Con c ci vs)) = fmap unnamed <$> tryRecPFromConP =<< do
-           A.ConP (ConPatInfo ci patNoRange False) (unambiguous (conName c)) <$> mapM (elimToPat . fmap DTerm) vs
+           A.ConP (ConPatInfo ci patNoRange ConPatEager) (unambiguous (conName c)) <$> mapM (elimToPat . fmap DTerm) vs
 
         termToPat (DTerm (I.Def _ [])) = return $ unnamed $ A.WildP patNoRange
         termToPat (DDef _ [])          = return $ unnamed $ A.WildP patNoRange
 
+        termToPat (DTerm (I.Lit l))    = return $ unnamed $ A.LitP l
+
         termToPat (DDot v)             = unnamed . A.DotP patNoRange <$> termToExpr v
         termToPat v                    = unnamed . A.DotP patNoRange <$> reify v
 
@@ -358,7 +371,7 @@ reifyDisplayFormP f ps wps = do
               -- even the pattern variables @n < len@ can be
               -- applied to some args @vs@.
               e <- if n < len
-                   then return $ A.patternToExpr $ namedArg $ ps !! n
+                   then return $ A.patternToExpr $ namedArg $ indexWithDefault __IMPOSSIBLE__ ps n
                    else reify (I.var (n - len))
               apps e =<< argsToExpr vs
             _ -> return underscore
@@ -371,9 +384,32 @@ instance Reify Term Expr where
   reifyWhen = reifyWhenE
   reify v = reifyTerm True v
 
+reifyPathPConstAsPath :: QName -> Elims -> TCM (QName, Elims)
+reifyPathPConstAsPath x es@[I.Apply l, I.Apply t, I.Apply lhs, I.Apply rhs] = do
+   reportSLn "reify.def" 100 $ "reifying def path " ++ show (x,es)
+   mpath  <- getBuiltinName' builtinPath
+   mpathp <- getBuiltinName' builtinPathP
+   let fallback = return (x,es)
+   case (,) <$> mpath <*> mpathp of
+     Just (path,pathp) | x == pathp -> do
+       let a = case unArg t of
+                I.Lam _ (NoAbs _ b)    -> Just b
+                I.Lam _ (Abs   _ b)
+                  | not $ 0 `freeIn` b -> Just (strengthen __IMPOSSIBLE__ b)
+                _                      -> Nothing
+       case a of
+         Just a -> return (path, [I.Apply l, I.Apply (setHiding Hidden $ defaultArg a), I.Apply lhs, I.Apply rhs])
+         Nothing -> fallback
+     _ -> fallback
+reifyPathPConstAsPath x es = return (x,es)
+
 reifyTerm :: Bool -> Term -> TCM Expr
-reifyTerm expandAnonDefs0 v = do
-  metasBare <- asks envPrintMetasBare
+reifyTerm expandAnonDefs0 v0 = do
+  -- Jesper 2018-11-02: If 'PrintMetasBare', drop all meta eliminations.
+  metasBare <- asksTC envPrintMetasBare
+  v <- instantiate v0 >>= \case
+    I.MetaV x _ | metasBare -> return $ I.MetaV x []
+    v -> return v
   -- Ulf 2014-07-10: Don't expand anonymous when display forms are disabled
   -- (i.e. when we don't care about nice printing)
   expandAnonDefs <- return expandAnonDefs0 `and2M` displayFormsEnabled
@@ -381,13 +417,19 @@ reifyTerm expandAnonDefs0 v = do
   -- then we print system-generated projections as postfix, else prefix.
   havePfp <- optPostfixProjections <$> pragmaOptions
   let pred = if havePfp then (== ProjPrefix) else (/= ProjPostfix)
-  v <- instantiate v
-  case applyUnless metasBare (unSpine' pred) v of
+  case unSpine' pred v of
+    -- Hack to print generalized field projections with nicer names. Should
+    -- only show up in errors. Check the spined form!
+    _ | I.Var n (I.Proj _ p : es) <- v,
+        Just name <- getGeneralizedFieldName p -> do
+      let fakeName = (qnameName p) { nameConcrete = C.Name noRange C.InScope [C.Id name] } -- TODO: infix names!?
+      elims (A.Var fakeName) =<< reify es
     I.Var n es   -> do
         x  <- liftTCM $ nameOfBV n `catchError` \_ -> freshName_ ("@" ++ show n)
         elims (A.Var x) =<< reify es
     I.Def x es   -> do
       reportSLn "reify.def" 100 $ "reifying def " ++ prettyShow x
+      (x,es) <- reifyPathPConstAsPath x es
       reifyDisplayForm x es $ reifyDef expandAnonDefs x es
     I.Con c ci vs -> do
       let x = conName c
@@ -411,7 +453,7 @@ reifyTerm expandAnonDefs0 v = do
           when (n > np) __IMPOSSIBLE__
           let h = A.Con (unambiguous x)
           if null vs then return h else do
-            es <- reify (fromMaybe __IMPOSSIBLE__ $ allApplyElims vs)
+            es <- reify (map (fromMaybe __IMPOSSIBLE__ . isApplyElim) vs)
             -- Andreas, 2012-04-20: do not reify parameter arguments of constructor
             -- if the first regular constructor argument is hidden
             -- we turn it into a named argument, in order to avoid confusion
@@ -433,8 +475,8 @@ reifyTerm expandAnonDefs0 v = do
                 -- Andreas, 2012-09-18
                 -- If the first regular constructor argument is hidden,
                 -- we keep the parameters to avoid confusion.
-                (Dom info _ : _) | notVisible info -> do
-                  let us = for (drop n pars) $ \ (Dom ai _) ->
+                (Dom {domInfo = info} : _) | notVisible info -> do
+                  let us = for (drop n pars) $ \ (Dom {domInfo = ai}) ->
                              -- setRelevance Relevant $
                              hideOrKeepInstance $ Arg ai underscore
                   apps h $ us ++ es  -- Note: unless --show-implicit, @apps@ will drop @us@.
@@ -444,7 +486,7 @@ reifyTerm expandAnonDefs0 v = do
 --    I.Lam info b | isAbsurdBody b -> return $ A. AbsurdLam noExprInfo $ getHiding info
     I.Lam info b    -> do
       (x,e) <- reify b
-      return $ A.Lam exprNoRange (DomainFree info $ BindName x) e
+      return $ A.Lam exprNoRange (DomainFree $ unnamedArg info $ BindName x) e
       -- Andreas, 2011-04-07 we do not need relevance information at internal Lambda
     I.Lit l        -> reify l
     I.Level l      -> reify l
@@ -461,22 +503,67 @@ reifyTerm expandAnonDefs0 v = do
             {- then -} (pure $ Arg (domInfo a) underscore)
             {- else -} (reify a)
       where
-        mkPi b (Arg info a) = do
-          -- #2776: Out-of-scope dots are not helpful at this point.
-          (x, b) <- reify b{ absName = unNotInScopeName $ absName b }
-          return $ A.Pi noExprInfo [TypedBindings noRange $ Arg info (TBind noRange [pure $ BindName x] a)] b
+        mkPi b (Arg info a') = do
+          (x, b) <- reify b
+          return $ A.Pi noExprInfo [TBind noRange [Arg info $ Named (domName a) $ BindName x] a'] b
         -- We can omit the domain type if it doesn't have any free variables
         -- and it's mentioned in the target type.
         domainFree a b = do
-          df <- asks envPrintDomainFreePi
+          df <- asksTC envPrintDomainFreePi
           return $ and [df, freeIn 0 b, closed a]
 
     I.Sort s     -> reify s
     I.MetaV x es -> do
-      x' <- reify x
-      ifM (asks envPrintMetasBare) {-then-} (return x') {-else-} $
-        elims x' =<< reify es
+          x' <- reify x
+
+          es' <- reify es
+
+          mv <- lookupMeta x
+          (msub1,meta_tel,msub2) <- do
+            local_chkpt <- viewTC eCurrentCheckpoint
+            (chkpt, tel, msub2) <- enterClosure (getMetaInfo mv) $ \ _ ->
+                               (,,) <$> viewTC eCurrentCheckpoint
+                                    <*> getContextTelescope
+                                    <*> viewTC (eCheckpoints . key local_chkpt)
+            (,,) <$> viewTC (eCheckpoints . key chkpt) <*> pure tel <*> pure msub2
+          let
+              addNames []    es = map (fmap unnamed) es
+              addNames _     [] = []
+              addNames xs     (I.Proj{} : _) = __IMPOSSIBLE__
+              addNames xs     (I.IApply x y r : es) =
+                -- Needs to be I.Apply so it can have an Origin field.
+                addNames xs (I.Apply (defaultArg r) : es)
+              addNames (x:xs) (I.Apply arg : es) =
+                I.Apply (Named (Just x) <$> (setOrigin Substitution arg)) : addNames xs es
+
+              p = mvPermutation mv
+              applyPerm p vs = permute (takeP (size vs) p) vs
+
+              names = map unranged $ p `applyPerm` teleNames meta_tel
+              named_es' = addNames names es'
+
+              dropIdentitySubs sub_local2G sub_tel2G =
+                 let
+                     args_G = applySubst sub_tel2G $ p `applyPerm` (teleArgs meta_tel :: [Arg Term])
+                     es_G = sub_local2G `applySubst` es
+                     sameVar x (I.Apply y) = isJust xv && xv == deBruijnView (unArg y)
+                      where
+                       xv = deBruijnView $ unArg x
+                     sameVar _ _ = False
+                     dropArg = take (size names) $ zipWith sameVar args_G es_G
+                     doDrop (b : xs)  (e : es) = (if b then id else (e :)) $ doDrop xs es
+                     doDrop []        es = es
+                     doDrop _         [] = []
+                 in doDrop dropArg $ named_es'
+
+              simpl_named_es' | Just sub_mtel2local <- msub1 = dropIdentitySubs IdS           sub_mtel2local
+                              | Just sub_local2mtel <- msub2 = dropIdentitySubs sub_local2mtel IdS
+                              | otherwise                    = named_es'
+
+          nelims x' simpl_named_es'
+
     I.DontCare v -> A.DontCare <$> reifyTerm expandAnonDefs v
+    I.Dummy s -> return $ A.Lit $ LitString noRange s
   where
     -- Andreas, 2012-10-20  expand a copy if not in scope
     -- to improve error messages.
@@ -529,7 +616,7 @@ reifyTerm expandAnonDefs0 v = do
                             vars = map (getArgInfo &&& name . namedArg) $ drop (length es) $ init $ namedClausePats cl
                             lam (i, s) = do
                               x <- freshName_ s
-                              return $ A.Lam exprNoRange (A.DomainFree i $ A.BindName x)
+                              return $ A.Lam exprNoRange (A.DomainFree $ unnamedArg i $ A.BindName x)
                         foldr ($) absLam <$> mapM lam vars
                       | otherwise -> elims absLam =<< reify (drop n es)
 
@@ -545,16 +632,16 @@ reifyTerm expandAnonDefs0 v = do
         df <- displayFormsEnabled
 
         -- #3004: give up if we have to print a pattern lambda inside its own body!
-        alreadyPrinting <- view ePrintingPatternLambdas
+        alreadyPrinting <- viewTC ePrintingPatternLambdas
 
         extLam <- case def of
           Function{ funExtLam = Just{}, funProjection = Just{} } -> __IMPOSSIBLE__
-          Function{ funExtLam = Just (ExtLamInfo m) } -> Just . size <$> lookupSection m
+          Function{ funExtLam = Just (ExtLamInfo m sys) } -> Just . (,sys) . size <$> lookupSection m
           _ -> return Nothing
         case extLam of
-          Just pars | df, notElem x alreadyPrinting ->
-            locally ePrintingPatternLambdas (x :) $
-            reifyExtLam x pars (defClauses defn) es
+          Just (pars, sys) | df, notElem x alreadyPrinting ->
+            locallyTC ePrintingPatternLambdas (x :) $
+            reifyExtLam x pars sys (defClauses defn) es
 
         -- Otherwise (ordinary function call):
           _ -> do
@@ -578,7 +665,7 @@ reifyTerm expandAnonDefs0 v = do
               -- These are the dropped projection arguments
               scope <- getScope
               let underscore = A.Underscore $ Info.emptyMetaInfo { metaScope = scope }
-              let pad = for as $ \ (Dom ai (x, _)) ->
+              let pad = for as $ \ (Dom{domInfo = ai, unDom = (x, _)}) ->
                     Arg ai $ Named (Just $ unranged x) underscore
 
               -- Now pad' ++ es' = drop n (pad ++ es)
@@ -622,7 +709,9 @@ reifyTerm expandAnonDefs0 v = do
              [ "  pad = " ++ show pad
              , "  nes = " ++ show nes
              ]
-           let hd = List.foldl' (A.App defaultAppInfo_) (A.Def x) pad
+           let hd0 | isProperProjection def = A.Proj ProjPrefix $ AmbQ $ singleton x
+                   | otherwise = A.Def x
+           let hd = List.foldl' (A.App defaultAppInfo_) hd0 pad
            nelims hd =<< reify nes
 
     -- Andreas, 2016-07-06 Issue #2047
@@ -644,20 +733,24 @@ reifyTerm expandAnonDefs0 v = do
     -- patterns, we fall back to printing the internal function created for the
     -- extended lambda, instead trying to construct the nice syntax.
 
-    reifyExtLam :: QName -> Int -> [I.Clause] -> I.Elims -> TCM Expr
-    reifyExtLam x npars cls es = do
+    reifyExtLam :: QName -> Int -> Maybe System -> [I.Clause] -> I.Elims -> TCM Expr
+    reifyExtLam x npars msys cls es = do
       reportSLn "reify.def" 10 $ "reifying extended lambda " ++ prettyShow x
       reportSLn "reify.def" 50 $ render $ nest 2 $ vcat
-        [ text "npars =" <+> pretty npars
-        , text "es    =" <+> fsep (map (prettyPrec 10) es)
-        , text "def   =" <+> vcat (map pretty cls) ]
+        [ "npars =" <+> pretty npars
+        , "es    =" <+> fsep (map (prettyPrec 10) es)
+        , "def   =" <+> vcat (map pretty cls) ]
       -- As extended lambda clauses live in the top level, we add the whole
       -- section telescope to the number of parameters.
-      let (pars, rest) = splitAt npars es
+      let (pares, rest) = splitAt npars es
+      let pars = fromMaybe __IMPOSSIBLE__ $ allApplyElims pares
+
       -- Since we applying the clauses to the parameters,
       -- we do not need to drop their initial "parameter" patterns
       -- (this is taken care of by @apply@).
-      cls <- mapM (reify . NamedClause x False . (`applyE` pars)) cls
+      cls <- caseMaybe msys
+               (mapM (reify . NamedClause x False . (`apply` pars)) cls)
+               (reify . QNamed x . (`apply` pars))
       let cx    = nameConcrete $ qnameName x
           dInfo = mkDefInfo cx noFixity' PublicAccess ConcreteDef (getRange x)
       elims (A.ExtendedLam exprNoRange dInfo x cls) =<< reify rest
@@ -681,6 +774,15 @@ instance (Reify i a) => Reify (Arg i) (Arg a) where
               `and2M` (return (getRelevance info /= Irrelevant) `or2M` showIrrelevantArguments)
   reifyWhen b i = traverse (reifyWhen b) i
 
+-- instance Reify Elim Expr where
+--   reifyWhen = reifyWhenE
+--   reify e = case e of
+--     I.IApply x y r -> appl "iapply" <$> reify (defaultArg r :: Arg Term)
+--     I.Apply v -> appl "apply" <$> reify v
+--     I.Proj f  -> appl "proj"  <$> reify ((defaultArg $ I.Def f []) :: Arg Term)
+--     where
+--       appl :: String -> Arg Expr -> Expr
+--       appl s v = A.App exprInfo (A.Lit (LitString noRange s)) $ fmap unnamed v
 
 data NamedClause = NamedClause QName Bool I.Clause
   -- ^ Also tracks whether module parameters should be dropped from the patterns.
@@ -699,8 +801,8 @@ instance (Ord k, Monoid v) => Monoid (MonoidMap k v) where
 -- | Removes implicit arguments that are not needed, that is, that don't bind
 --   any variables that are actually used and doesn't do pattern matching.
 --   Doesn't strip any arguments that were written explicitly by the user.
-stripImplicits :: A.Patterns -> TCM A.Patterns
-stripImplicits ps = do
+stripImplicits :: A.Patterns -> A.Patterns -> TCM A.Patterns
+stripImplicits params ps = do
   -- if --show-implicit we don't need the names
   ifM showImplicitArguments (return $ map (unnamed . namedThing <$>) ps) $ do
     reportSLn "reify.implicit" 30 $ unlines
@@ -715,7 +817,7 @@ stripImplicits ps = do
     where
       -- Replace variables in dot patterns by an underscore _ if they are hidden
       -- in the pattern. This is slightly nicer than making the implicts explicit.
-      blankDots ps = blank (varsBoundIn ps) ps
+      blankDots ps = blank (varsBoundIn $ params ++ ps) ps
 
       strip ps = stripArgs True ps
         where
@@ -736,6 +838,7 @@ stripImplicits ps = do
           stripName True  = fmap (unnamed . namedThing)
           stripName False = id
 
+          -- TODO: vars appearing in EqualPs shouldn't be stripped.
           canStrip a = and
             [ notVisible a
             , getOrigin a `notElem` [ UserWritten , CaseSplit ]
@@ -758,6 +861,7 @@ stripImplicits ps = do
             A.AsP i x p   -> A.AsP i x $ stripPat p
             A.PatternSynP _ _ _ -> __IMPOSSIBLE__ -- p
             A.RecP i fs   -> A.RecP i $ map (fmap stripPat) fs  -- TODO Andreas: is this right?
+            A.EqualP{}    -> p -- EqualP cannot be blanked.
             A.WithP i p   -> A.WithP i $ stripPat p -- TODO #2822: right?
 
           varOrDot A.VarP{}      = True
@@ -781,7 +885,7 @@ class BlankVars a where
 instance BlankVars a => BlankVars (Arg a)              where
 instance BlankVars a => BlankVars (Named s a)          where
 instance BlankVars a => BlankVars [a]                  where
-instance BlankVars a => BlankVars (A.Pattern' a)       where
+-- instance BlankVars a => BlankVars (A.Pattern' a)       where  -- see case EqualP !
 instance BlankVars a => BlankVars (FieldAssignment' a) where
 
 instance (BlankVars a, BlankVars b) => BlankVars (a, b) where
@@ -810,6 +914,22 @@ instance BlankVars A.LHSCore where
   blank bound (A.LHSProj p b ps) = uncurry (A.LHSProj p) $ blank bound (b, ps)
   blank bound (A.LHSWith h wps ps) = uncurry (uncurry A.LHSWith) $ blank bound ((h, wps), ps)
 
+instance BlankVars A.Pattern where
+  blank bound p = case p of
+    A.VarP _      -> p   -- do not blank pattern vars
+    A.ConP c i ps -> A.ConP c i $ blank bound ps
+    A.ProjP{}     -> p
+    A.DefP i f ps -> A.DefP i f $ blank bound ps
+    A.DotP i e    -> A.DotP i $ blank bound e
+    A.WildP _     -> p
+    A.AbsurdP _   -> p
+    A.LitP _      -> p
+    A.AsP i n p   -> A.AsP i n $ blank bound p
+    A.PatternSynP _ _ _ -> __IMPOSSIBLE__
+    A.RecP i fs   -> A.RecP i $ blank bound fs
+    A.EqualP{}    -> p
+    A.WithP i p   -> A.WithP i (blank bound p)
+
 instance BlankVars A.Expr where
   blank bound e = case e of
     A.ScopedExpr i e       -> A.ScopedExpr i $ blank bound e
@@ -830,9 +950,10 @@ instance BlankVars A.Expr where
     A.ExtendedLam i d f cs -> A.ExtendedLam i d f $ blank bound cs
     A.Pi i tel e           -> let bound' = varsBoundIn tel `Set.union` bound
                               in  uncurry (A.Pi i) $ blank bound' (tel, e)
+    A.Generalized {}       -> __IMPOSSIBLE__
     A.Fun i a b            -> uncurry (A.Fun i) $ blank bound (a, b)
     A.Set _ _              -> e
-    A.Prop _               -> e
+    A.Prop _ _             -> e
     A.Let _ _ _            -> __IMPOSSIBLE__
     A.Rec i es             -> A.Rec i $ blank bound es
     A.RecUpdate i e es     -> uncurry (A.RecUpdate i) $ blank bound (e, es)
@@ -860,9 +981,6 @@ instance BlankVars A.LamBinding where
   blank bound b@A.DomainFree{} = b
   blank bound (A.DomainFull bs) = A.DomainFull $ blank bound bs
 
-instance BlankVars TypedBindings where
-  blank bound (TypedBindings r bs) = TypedBindings r $ blank bound bs
-
 instance BlankVars TypedBinding where
   blank bound (TBind r n e) = TBind r n $ blank bound e
   blank bound (TLet _ _)    = __IMPOSSIBLE__ -- Since the internal syntax has no let bindings left
@@ -886,8 +1004,8 @@ instance Binder A.LHSCore where
 
 instance Binder A.Pattern where
   varsBoundIn = foldAPattern $ \case
-    A.VarP x            -> singleton $ unBind x
-    A.AsP _ x _         -> empty
+    A.VarP x            -> varsBoundIn x
+    A.AsP _ x _         -> empty    -- Not x because of #2414 (?)
     A.ConP _ _ _        -> empty
     A.ProjP{}           -> empty
     A.DefP _ _ _        -> empty
@@ -897,32 +1015,27 @@ instance Binder A.Pattern where
     A.LitP{}            -> empty
     A.PatternSynP _ _ _ -> empty
     A.RecP _ _          -> empty
+    A.EqualP{}          -> empty
     A.WithP _ _         -> empty
 
 instance Binder A.LamBinding where
-  varsBoundIn (A.DomainFree _ x) = singleton $ unBind x
-  varsBoundIn (A.DomainFull b)   = varsBoundIn b
-
-instance Binder TypedBindings where
-  varsBoundIn (TypedBindings _ b) = varsBoundIn b
+  varsBoundIn (A.DomainFree x) = varsBoundIn x
+  varsBoundIn (A.DomainFull b) = varsBoundIn b
 
 instance Binder TypedBinding where
   varsBoundIn (TBind _ xs _) = varsBoundIn xs
   varsBoundIn (TLet _ bs)    = varsBoundIn bs
 
+instance Binder BindName where
+  varsBoundIn x = singleton (unBind x)
+
 instance Binder LetBinding where
-  varsBoundIn (LetBind _ _ x _ _) = singleton $ unBind x
+  varsBoundIn (LetBind _ _ x _ _) = varsBoundIn x
   varsBoundIn (LetPatBind _ p _)  = varsBoundIn p
   varsBoundIn LetApply{}          = empty
   varsBoundIn LetOpen{}           = empty
   varsBoundIn LetDeclaredVariable{} = empty
 
-instance Binder (WithHiding Name) where
-  varsBoundIn (WithHiding _ x) = singleton x
-
-instance Binder (WithHiding BindName) where
-  varsBoundIn (WithHiding _ x) = singleton $ unBind x
-
 instance Binder a => Binder (FieldAssignment' a) where
 instance Binder a => Binder (Arg a)              where
 instance Binder a => Binder (Named x a)          where
@@ -935,13 +1048,19 @@ instance (Binder a, Binder b) => Binder (a, b) where
 -- | Assumes that pattern variables have been added to the context already.
 --   Picks pattern variable names from context.
 reifyPatterns :: MonadTCM tcm => [NamedArg I.DeBruijnPattern] -> tcm [NamedArg A.Pattern]
-reifyPatterns = mapM $ stripNameFromExplicit <.> traverse (traverse reifyPat)
+reifyPatterns = mapM $ (stripNameFromExplicit . stripHidingFromPostfixProj) <.>
+                       traverse (traverse reifyPat)
   where
     stripNameFromExplicit :: NamedArg p -> NamedArg p
     stripNameFromExplicit a
       | visible a = fmap (unnamed . namedThing) a
       | otherwise = a
 
+    stripHidingFromPostfixProj :: IsProjP p => NamedArg p -> NamedArg p
+    stripHidingFromPostfixProj a = case isProjP a of
+      Just (o, _) | o /= ProjPrefix -> setHiding NotHidden a
+      _                             -> a
+
     reifyPat :: MonadTCM tcm => I.DeBruijnPattern -> tcm A.Pattern
     reifyPat p = do
      liftTCM $ reportSLn "reify.pat" 80 $ "reifying pattern " ++ show p
@@ -956,7 +1075,7 @@ reifyPatterns = mapM $ stripNameFromExplicit <.> traverse (traverse reifyPat)
       I.DotP (PatOVar x) v@(I.Var i []) -> do
         x' <- nameOfBV i
         if nameConcrete x == nameConcrete x' then
-          return $ A.VarP (BindName x')
+          return $ A.VarP $ BindName x'
         else
           reifyDotP v
       I.DotP o v -> reifyDotP v
@@ -966,19 +1085,27 @@ reifyPatterns = mapM $ stripNameFromExplicit <.> traverse (traverse reifyPat)
         Just PatOWild   -> return $ A.WildP patNoRange
         Just PatOAbsurd -> return $ A.AbsurdP patNoRange
         _               -> reifyConP c cpi ps
+      I.DefP o f ps  -> case o of
+        PatOWild   -> return $ A.WildP patNoRange
+        PatOAbsurd -> return $ A.AbsurdP patNoRange
+        _ -> A.DefP patNoRange (unambiguous f) <$> reifyPatterns ps
+      I.IApplyP PatODot _ _ x -> reifyDotP $ var $ dbPatVarIndex x
+      I.IApplyP PatOWild _ _ x -> return $ A.WildP patNoRange
+      I.IApplyP PatOAbsurd _ _ x -> return $ A.AbsurdP patNoRange
+      I.IApplyP _ _ _ x -> reifyVarP x
 
     reifyVarP :: MonadTCM tcm => DBPatVar -> tcm A.Pattern
     reifyVarP x = do
       n <- liftTCM $ nameOfBV $ dbPatVarIndex x
       case dbPatVarName x of
-        "_"  -> return $ A.VarP $ A.BindName n
+        "_"  -> return $ A.VarP $ BindName n
         -- Andreas, 2017-09-03: TODO for #2580
         -- Patterns @VarP "()"@ should have been replaced by @AbsurdP@, but the
         -- case splitter still produces them.
-        y    -> if prettyShow (nameConcrete n) == "()" then return $ A.VarP $ A.BindName n else
+        y    -> if prettyShow (nameConcrete n) == "()" then return $ A.VarP (BindName n) else
           -- Andreas, 2017-09-03, issue #2729
           -- Restore original pattern name.  AbstractToConcrete picks unique names.
-          return $ A.VarP $ A.BindName n { nameConcrete = C.Name noRange [ C.Id y ] }
+          return $ A.VarP $ BindName n { nameConcrete = C.Name noRange C.InScope [ C.Id y ] }
 
     reifyDotP :: MonadTCM tcm => Term -> tcm A.Pattern
     reifyDotP v = do
@@ -991,7 +1118,7 @@ reifyPatterns = mapM $ stripNameFromExplicit <.> traverse (traverse reifyPat)
     reifyConP c cpi ps = do
       tryRecPFromConP =<< do A.ConP ci (unambiguous (conName c)) <$> reifyPatterns ps
       where
-        ci = ConPatInfo origin patNoRange False
+        ci = ConPatInfo origin patNoRange ConPatEager
         origin = fromConPatternInfo cpi
 
 
@@ -1030,10 +1157,10 @@ instance Reify NamedClause A.Clause where
     -- (e.g. for extended lambdas). We still get here with toDrop = True and
     -- pattern lambdas when doing make-case, so take care to drop the right
     -- number of parameters.
-    lhs <- if not toDrop then return lhs else do
+    (params , lhs) <- if not toDrop then return ([] , lhs) else do
       nfv <- (size <.> lookupSection =<< getDefModule f) `catchError` \_ -> return 0
-      return $ dropParams nfv lhs
-    lhs <- stripImps lhs
+      return $ splitParams nfv lhs
+    lhs <- stripImps params lhs
     reportSLn "reify.clause" 60 $ "reifying NamedClause, lhs = " ++ show lhs
     rhs <- caseMaybe (clauseBody cl) (return AbsurdRHS) $ \ e -> do
        RHS <$> reify e <*> pure Nothing
@@ -1042,9 +1169,30 @@ instance Reify NamedClause A.Clause where
     reportSLn "reify.clause" 60 $ "reified NamedClause, result = " ++ show result
     return result
     where
-      dropParams n (SpineLHS i f ps) = SpineLHS i f $ drop n ps
-      stripImps :: SpineLHS -> TCM SpineLHS
-      stripImps (SpineLHS i f ps) =  SpineLHS i f <$> stripImplicits ps
+      splitParams n (SpineLHS i f ps) =
+        let (params , pats) = splitAt n ps
+        in  (params , SpineLHS i f pats)
+      stripImps :: [NamedArg A.Pattern] -> SpineLHS -> TCM SpineLHS
+      stripImps params (SpineLHS i f ps) =  SpineLHS i f <$> stripImplicits params ps
+
+instance Reify (QNamed System) [A.Clause] where
+  reify (QNamed f (System tel sys)) = addContext tel $ do
+    reportSLn "reify.system" 40 $ unlines $ show tel : map show sys
+    unview <- intervalUnview'
+    forM sys $ \ (alpha,u) -> do
+      rhs <- RHS <$> reify u <*> pure Nothing
+      ep <- fmap (A.EqualP patNoRange) . forM alpha $ \ (phi,b) -> do
+        let
+            d True = unview IOne
+            d False = unview IZero
+        reify (phi, d b)
+
+      ps <- reifyPatterns $ teleNamedArgs tel
+      ps <- stripImplicits [] $ ps ++ [defaultNamedArg ep]
+      let
+        lhs = SpineLHS (LHSRange noRange) f ps
+        result = A.Clause (spineToLhs lhs) [] rhs A.noWhereDecls False
+      return result
 
 instance Reify Type Expr where
     reifyWhen = reifyWhenE
@@ -1060,8 +1208,14 @@ instance Reify Sort Expr where
         I.Type a -> do
           a <- reify a
           return $ A.App defaultAppInfo_ (A.Set noExprInfo 0) (defaultNamedArg a)
-        I.Prop       -> return $ A.Prop noExprInfo
-        I.Inf       -> A.Var <$> freshName_ ("Setω" :: String)
+        I.Prop (I.Max [])                -> return $ A.Prop noExprInfo 0
+        I.Prop (I.Max [I.ClosedLevel n]) -> return $ A.Prop noExprInfo n
+        I.Prop a -> do
+          a <- reify a
+          return $ A.App defaultAppInfo_ (A.Prop noExprInfo 0) (defaultNamedArg a)
+        I.Inf       -> do
+          I.Def inf [] <- primSetOmega
+          return $ A.Def inf
         I.SizeUniv  -> do
           I.Def sizeU [] <- primSizeUniv
           return $ A.Def sizeU
@@ -1075,6 +1229,8 @@ instance Reify Sort Expr where
           e <- reify s
           return $ A.App defaultAppInfo_ (A.Var univs) $ defaultNamedArg e
         I.MetaS x es -> reify $ I.MetaV x es
+        I.DefS d es -> reify $ I.Def d es
+        I.DummyS s -> return $ A.Lit $ LitString noRange s
 
 instance Reify Level Expr where
   reifyWhen = reifyWhenE
@@ -1101,12 +1257,13 @@ instance Reify I.Telescope A.Telescope where
   reify EmptyTel = return []
   reify (ExtendTel arg tel) = do
     Arg info e <- reify arg
-    (x,bs)  <- reify tel
-    let r = getRange e
-    return $ TypedBindings r (Arg info (TBind r [pure $ BindName x] e)) : bs
+    (x, bs)  <- reify tel
+    let r    = getRange e
+        name = domName arg
+    return $ TBind r [Arg info $ Named name $ BindName x] e : bs
 
 instance Reify i a => Reify (Dom i) (Arg a) where
-    reify (Dom info i) = Arg info <$> reify i
+    reify (Dom{domInfo = info, unDom = i}) = Arg info <$> reify i
 
 instance Reify i a => Reify (I.Elim' i) (I.Elim' a) where
   reify = traverse reify
