@@ -65,10 +65,10 @@ instance ToAbstract r a => ToAbstract (Arg r) (NamedArg a) where
 instance ToAbstract [Arg Term] [NamedArg Expr] where
   toAbstract = traverse toAbstract
 
-instance ToAbstract r Expr => ToAbstract (Dom r, Name) (A.TypedBindings) where
-  toAbstract (Dom i x, name) = do
+instance ToAbstract r Expr => ToAbstract (Dom r, Name) (A.TypedBinding) where
+  toAbstract (Dom{domInfo = i,unDom = x}, name) = do
     dom <- toAbstract x
-    return $ TypedBindings noRange $ Arg i $ TBind noRange [pure $ BindName name] dom
+    return $ TBind noRange [unnamedArg i $ BindName name] dom
 
 instance ToAbstract (Expr, Elim) Expr where
   toAbstract (f, Apply arg) = do
@@ -105,7 +105,7 @@ instance ToAbstract Term Expr where
     R.Lam h t  -> do
       (e, name) <- toAbstract t
       let info  = setHiding h $ setOrigin Reflected defaultArgInfo
-      return $ A.Lam exprNoRange (DomainFree info $ BindName name) e
+      return $ A.Lam exprNoRange (DomainFree $ unnamedArg info $ BindName name) e
     R.ExtLam cs es -> do
       name <- freshName_ extendedLambdaName
       m    <- lift $ getCurrentModule
@@ -142,7 +142,7 @@ instance ToAbstract R.Pattern (Names, A.Pattern) where
   toAbstract pat = case pat of
     R.ConP c args -> do
       (names, args) <- toAbstractPats args
-      return (names, A.ConP (ConPatInfo ConOCon patNoRange False) (unambiguous $ killRange c) args)
+      return (names, A.ConP (ConPatInfo ConOCon patNoRange ConPatEager) (unambiguous $ killRange c) args)
     R.DotP    -> return ([], A.WildP patNoRange)
     R.VarP s | isNoName s -> withName "z" $ \ name -> return ([name], A.VarP $ BindName name)
         -- Ulf, 2016-08-09: Also bind noNames (#2129). This to make the
