@@ -1,5 +1,4 @@
-{-# LANGUAGE CPP                    #-}
-{-# LANGUAGE ScopedTypeVariables    #-}
+{-# LANGUAGE CPP #-}
 
 {- | Sparse matrices.
 
@@ -43,6 +42,7 @@ module Agda.Termination.SparseMatrix
   , Diagonal(..)
   , toSparseRows
   , supSize
+  , zipAssocWith
     -- * Modifying matrices
   , addRow
   , addColumn
@@ -351,16 +351,6 @@ interAssocWith f l@((i,a):l') m@((j,b):m')
     | i > j = interAssocWith f l m'
     | otherwise = (i, f a b) : interAssocWith f l' m'
 
-interAssocWith2 :: Ord i => (a -> a -> a) -> [(i,a)] -> [(i,a)] -> [(i,a)]
-interAssocWith2 f = zipAssocWith (const []) (const []) (const Nothing) (const Nothing) (\ a -> Just . f a)
-
-prop_interAssocWith_correct2 :: [(Int,Int)] -> [(Int,Int)] -> Bool
-prop_interAssocWith_correct2 xs ys =
-  interAssocWith (*) xs ys == interAssocWith2 (*) xs ys
-  where
-    l  = List.sortBy (compare `on` fst) xs
-    l' = List.sortBy (compare `on` fst) ys
-
 -- | @'mul' semiring m1 m2@ multiplies matrices @m1@ and @m2@.
 --   Uses the operations of the semiring @semiring@ to perform the
 --   multiplication.
@@ -449,7 +439,7 @@ instance (Integral i, HasZero b, Pretty b) =>
          . Boxes.render
          . Boxes.hsep 1 Boxes.right
          . map ( Boxes.vcat Boxes.right
-               . map ( Boxes.alignHoriz Boxes.right 2
+               . map ( Boxes.alignHoriz Boxes.right 4
                      . Boxes.text . render . pretty
                      )
                )
