@@ -10,11 +10,14 @@
 
 module Agda.TypeChecking.CheckInternal
   ( checkType
+  , checkType'
+  , checkSort
   , checkInternal
   , checkInternal'
   , Action(..), defaultAction, eraseUnusedAction
   , infer
   , inferSort
+  , shouldBeSort
   ) where
 
 import Control.Arrow ((&&&), (***), first, second)
@@ -29,6 +32,7 @@ import Agda.TypeChecking.Level
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.Primitive
 import Agda.TypeChecking.ProjectionLike (elimView)
 import Agda.TypeChecking.Records (getDefType)
 import Agda.TypeChecking.Reduce
@@ -36,6 +40,7 @@ import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Sort
 import Agda.TypeChecking.Telescope
 
+
 import Agda.Utils.Functor (($>))
 import Agda.Utils.Monad
 import Agda.Utils.Size
@@ -66,7 +71,7 @@ checkType t = void $ checkType' t
 checkType' :: Type -> TCM Sort
 checkType' t = do
   reportSDoc "tc.check.internal" 20 $ sep
-    [ text "checking internal type "
+    [ "checking internal type "
     , prettyTCM t
     ]
   v <- elimView True $ unEl t -- bring projection-like funs in post-fix form
@@ -95,6 +100,7 @@ checkType' t = do
     v@Lit{}    -> typeError $ InvalidType v
     v@Level{}  -> typeError $ InvalidType v
     DontCare v -> checkType' $ t $> v
+    Dummy s    -> __IMPOSSIBLE_VERBOSE__ s
 
 checkTypeSpine :: Type -> Term -> Elims -> TCM Sort
 checkTypeSpine a self es = shouldBeSort =<< do snd <$> inferSpine a self es
@@ -144,8 +150,8 @@ checkInternal v t = void $ checkInternal' defaultAction v t
 checkInternal' :: Action -> Term -> Type -> TCM Term
 checkInternal' action v t = do
   reportSDoc "tc.check.internal" 20 $ sep
-    [ text "checking internal "
-    , nest 2 $ sep [ prettyTCM v <+> text ":"
+    [ "checking internal "
+    , nest 2 $ sep [ prettyTCM v <+> ":"
                    , nest 2 $ prettyTCM t ] ]
   -- Bring projection-like funs in post-fix form,
   -- even lone ones (True).
@@ -169,7 +175,7 @@ checkInternal' action v t = do
           $ Con c ci $ take (length vs) vs2
     Lit l      -> Lit l <$ ((`subtype` t) =<< litType l)
     Lam ai vb  -> do
-      (a, b) <- shouldBePi t
+      (a, b) <- maybe (shouldBePi t) return =<< isPath t
       ai <- checkArgInfo action ai $ domInfo a
       addContext (suggest vb b, a) $ do
         Lam ai . Abs (absName vb) <$> checkInternal' action (absBody vb) (absBody b)
@@ -194,6 +200,7 @@ checkInternal' action v t = do
       l <- checkLevel action l
       Level l <$ ((`subtype` t) =<< levelType)
     DontCare v -> DontCare <$> checkInternal' action v t
+    Dummy s -> __IMPOSSIBLE_VERBOSE__ s
 
 -- | Make sure a constructor is fully applied
 --   and infer the type of the constructor.
@@ -232,10 +239,10 @@ checkSpine
   -> TCM Term  -- ^ The application after modification by the @Action@.
 checkSpine action a self es t = do
   reportSDoc "tc.check.internal" 20 $ sep
-    [ text "checking spine "
-    , nest 2 $ sep [ parens (sep [ prettyTCM self <+> text ":"
+    [ "checking spine "
+    , nest 2 $ sep [ parens (sep [ prettyTCM self <+> ":"
                                  , nest 2 $ prettyTCM a ])
-                   , nest 4 $ prettyTCM es <+> text ":"
+                   , nest 4 $ prettyTCM es <+> ":"
                    , nest 2 $ prettyTCM t ] ]
   ((v, v'), t') <- inferSpine' action a self self es
   t' <- reduce t'
@@ -317,13 +324,21 @@ inferSpine' :: Action -> Type -> Term -> Term -> Elims -> TCM ((Term, Term), Typ
 inferSpine' action t self self' [] = return ((self, self'), t)
 inferSpine' action t self self' (e : es) = do
   reportSDoc "tc.infer.internal" 30 $ sep
-    [ text "inferSpine': "
-    , text "type t = " <+> prettyTCM t
-    , text "self  = " <+> prettyTCM self
-    , text "self' = " <+> prettyTCM self'
-    , text "eliminated by e = " <+> prettyTCM e
+    [ "inferSpine': "
+    , "type t = " <+> prettyTCM t
+    , "self  = " <+> prettyTCM self
+    , "self' = " <+> prettyTCM self'
+    , "eliminated by e = " <+> prettyTCM e
     ]
   case e of
+    IApply x y r -> do
+      (a, b) <- shouldBePath t
+      r' <- checkInternal' action r (unDom a)
+      izero <- primIZero
+      ione  <- primIOne
+      x' <- checkInternal' action x (b `absApp` izero)
+      y' <- checkInternal' action y (b `absApp` ione)
+      inferSpine' action (b `absApp` r) (self `applyE` [e]) (self' `applyE` [IApply x' y' r']) es
     Apply (Arg ai v) -> do
       (a, b) <- shouldBePi t
       ai <- checkArgInfo action ai $ domInfo a
@@ -344,6 +359,13 @@ shouldBeProjectible t f = maybe failure return =<< getDefType f =<< reduce t
   where failure = typeError $ ShouldBeRecordType t
     -- TODO: more accurate error that makes sense also for proj.-like funs.
 
+shouldBePath :: Type -> TCM (Dom Type, Abs Type)
+shouldBePath t = do
+  m <- isPath t
+  case m of
+    Just p  -> return p
+    Nothing -> typeError $ ShouldBePath t
+
 shouldBePi :: Type -> TCM (Dom Type, Abs Type)
 shouldBePi t = ifPiType t (\ a b -> return (a, b)) $ const $ typeError $ ShouldBePi t
 
@@ -363,8 +385,7 @@ checkSort :: Action -> Sort -> TCM Sort
 checkSort action s =
   case s of
     Type l   -> Type <$> checkLevel action l
-    Prop     -> __IMPOSSIBLE__
-      -- the dummy Prop should not be part of a term we check
+    Prop l   -> Prop <$> checkLevel action l
     Inf      -> return Inf
     SizeUniv -> return SizeUniv
     PiSort a b -> do
@@ -379,7 +400,18 @@ checkSort action s =
       case v of
         Sort s     -> return s
         MetaV x es -> return $ MetaS x es
+        Def d es   -> return $ DefS d es
         _          -> __IMPOSSIBLE__
+    DefS d es -> do
+      a <- defType <$> getConstInfo d
+      let self = Sort $ DefS d []
+      ((_,v),_) <- inferSpine' action a self self es
+      case v of
+        Sort s     -> return s
+        MetaV x es -> return $ MetaS x es
+        Def d es   -> return $ DefS d es
+        _          -> __IMPOSSIBLE__
+    DummyS s -> __IMPOSSIBLE_VERBOSE__ s
 
 -- | Check if level is well-formed.
 checkLevel :: Action -> Level -> TCM Level
@@ -406,7 +438,6 @@ subtype t1 t2 = do
   ifIsSort t1 (\ s1 -> (s1 `leqSort`) =<< shouldBeSort t2) $ do
     -- Andreas, 2017-03-09, issue #2493
     -- Only check subtyping, do not solve any metas!
-    -- TODO: NEED? disableDestructiveUpdate
     dontAssignMetas $ leqType t1 t2
 
 -- | Compute the sort of a type.
@@ -432,17 +463,20 @@ inferSort t = case t of
     Lam{}      -> __IMPOSSIBLE__
     Level{}    -> __IMPOSSIBLE__
     DontCare{} -> __IMPOSSIBLE__
+    Dummy s    -> __IMPOSSIBLE_VERBOSE__ s
 
 -- | @eliminate t self es@ eliminates value @self@ of type @t@ by spine @es@
 --   and returns the remaining value and its type.
 eliminate :: Term -> Type -> Elims -> TCM (Term, Type)
 eliminate self t [] = return (self, t)
 eliminate self t (e : es) = case e of
-    Apply (Arg _ v) -> do
-      (_, b) <- shouldBePi t
-      eliminate (self `apply1` v) (b `absApp` v) es
+    Apply (Arg _ v) -> ifNotPiType t __IMPOSSIBLE__ {-else-} $ \ _ b ->
+      eliminate (self `applyE` [e]) (b `absApp` v) es
+    IApply _ _ v -> do
+      (_, b) <- shouldBePath t
+      eliminate (self `applyE` [e]) (b `absApp` v) es
     -- case: projection or projection-like
     Proj o f -> do
-      (Dom ai _, b) <- shouldBePi =<< shouldBeProjectible t f
+      (Dom{domInfo = ai}, b) <- shouldBePi =<< shouldBeProjectible t f
       u  <- applyDef o f $ Arg ai self
       eliminate u (b `absApp` self) es
