@@ -51,6 +51,8 @@ data Case c = Branches
     -- ^ Map from literal to case subtree.
   , catchAllBranch :: Maybe c
     -- ^ (Possibly additional) catch-all clause.
+  , fallThrough :: Maybe Bool
+    -- ^ (if True) In case of non-canonical argument use catchAllBranch.
   , lazyMatch :: Bool
     -- ^ Lazy pattern match. Requires single (non-copattern) branch with no lit
     --   branches and no catch-all.
@@ -77,19 +79,19 @@ data CompiledClauses' a
 type CompiledClauses = CompiledClauses' Term
 
 litCase :: Literal -> c -> Case c
-litCase l x = Branches False Map.empty Nothing (Map.singleton l x) Nothing False
+litCase l x = Branches False Map.empty Nothing (Map.singleton l x) Nothing (Just False) False
 
-conCase :: QName -> WithArity c -> Case c
-conCase c x = Branches False (Map.singleton c x) Nothing Map.empty Nothing False
+conCase :: QName -> Bool -> WithArity c -> Case c
+conCase c b x = Branches False (Map.singleton c x) Nothing Map.empty Nothing (Just b) False
 
 etaCase :: ConHead -> WithArity c -> Case c
-etaCase c x = Branches False Map.empty (Just (c, x)) Map.empty Nothing True
+etaCase c x = Branches False Map.empty (Just (c, x)) Map.empty Nothing (Just False) True
 
 projCase :: QName -> c -> Case c
-projCase c x = Branches True (Map.singleton c $ WithArity 0 x) Nothing Map.empty Nothing False
+projCase c x = Branches True (Map.singleton c $ WithArity 0 x) Nothing Map.empty Nothing (Just False) False
 
 catchAll :: c -> Case c
-catchAll x = Branches False Map.empty Nothing Map.empty (Just x) False
+catchAll x = Branches False Map.empty Nothing Map.empty (Just x) (Just True) False
 
 -- | Check that the requirements on lazy matching (single inductive case) are
 --   met, and set lazy to False otherwise.
@@ -130,24 +132,30 @@ instance (Semigroup c, Monoid c) => Monoid (WithArity c) where
   mappend = (<>)
 
 instance Semigroup m => Semigroup (Case m) where
-  Branches cop cs eta ls m lazy <> Branches cop' cs' eta' ls' m' lazy' = checkLazyMatch $
+  Branches cop cs eta ls m b lazy <> Branches cop' cs' eta' ls' m' b' lazy' = checkLazyMatch $
     Branches (cop || cop') -- for @projCase <> mempty@
              (Map.unionWith (<>) cs cs')
              (unionEta eta eta')
              (Map.unionWith (<>) ls ls')
              (m <> m')
+             (combine b b')
              (lazy && lazy')
-    where unionEta Nothing b = b
-          unionEta b Nothing = b
-          unionEta Just{} Just{} = __IMPOSSIBLE__
+   where
+     combine Nothing  b'        = b
+     combine b        Nothing   = b
+     combine (Just b) (Just b') = Just $ b && b'
+
+     unionEta Nothing b = b
+     unionEta b Nothing = b
+     unionEta Just{} Just{} = __IMPOSSIBLE__
 
 instance (Semigroup m, Monoid m) => Monoid (Case m) where
   mempty  = empty
   mappend = (<>)
 
 instance Null (Case m) where
-  empty = Branches False Map.empty Nothing Map.empty Nothing True
-  null (Branches _cop cs eta ls mcatch _lazy) = null cs && null eta && null ls && null mcatch
+  empty = Branches False Map.empty Nothing Map.empty Nothing Nothing True
+  null (Branches _cop cs eta ls mcatch _b _lazy) = null cs && null eta && null ls && null mcatch
 
 -- * Pretty instances.
 
@@ -155,26 +163,26 @@ instance Pretty a => Pretty (WithArity a) where
   pretty = pretty . content
 
 instance Pretty a => Pretty (Case a) where
-  prettyPrec p (Branches _cop cs eta ls m lazy) =
+  prettyPrec p (Branches _cop cs eta ls m b lazy) =
     mparens (p > 0) $ prLazy lazy <+> vcat (prettyMap cs ++ prEta eta ++ prettyMap ls ++ prC m)
     where
-      prLazy True  = text "~"
+      prLazy True  = "~"
       prLazy False = empty
       prC Nothing = []
-      prC (Just x) = [text "_ ->" <+> pretty x]
+      prC (Just x) = ["_ ->" <+> pretty x]
       prEta Nothing = []
-      prEta (Just (c, cc)) = [(text "eta" <+> pretty c <+> text "->") <?> pretty cc]
+      prEta (Just (c, cc)) = [("eta" <+> pretty c <+> "->") <?> pretty cc]
 
 prettyMap :: (Pretty k, Pretty v) => Map k v -> [Doc]
-prettyMap m = [ sep [ pretty k <+> text "->"
+prettyMap m = [ sep [ pretty k <+> "->"
                     , nest 2 $ pretty v ]
               | (k, v) <- Map.toList m ]
 
 instance Pretty CompiledClauses where
-  pretty (Done hs t) = text ("done" ++ show hs) <?> pretty t
-  pretty Fail        = text "fail"
+  pretty (Done hs t) = ("done" <> pretty hs) <?> pretty t
+  pretty Fail        = "fail"
   pretty (Case n bs) | projPatterns bs =
-    sep [ text "record"
+    sep [ "record"
         , nest 2 $ pretty bs
         ]
   pretty (Case n bs) =
@@ -186,12 +194,12 @@ instance KillRange c => KillRange (WithArity c) where
   killRange = fmap killRange
 
 instance KillRange c => KillRange (Case c) where
-  killRange (Branches cop con eta lit all lazy) = Branches cop
+  killRange (Branches cop con eta lit all b lazy) = Branches cop
     (killRangeMap con)
     (killRange eta)
     (killRangeMap lit)
     (killRange all)
-    lazy
+    b lazy
 
 instance KillRange CompiledClauses where
   killRange (Case i br) = killRange2 Case i br
