@@ -1,6 +1,14 @@
 {-# LANGUAGE CPP           #-}
 
-module Agda.TypeChecking.Coverage.Match where
+module Agda.TypeChecking.Coverage.Match
+  ( Match(..), match
+  , SplitPattern, SplitPatVar(..), fromSplitPatterns, toSplitPatterns
+  , toSplitPSubst, applySplitPSubst
+  , isTrivialPattern
+  , BlockingVar(..), BlockingVars, BlockedOnResult(..)
+  , setBlockingVarOverlap
+  , ApplyOrIApply(..)
+  ) where
 
 import Control.Monad.State
 
@@ -17,10 +25,13 @@ import Agda.Syntax.Common
 import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Pattern ()
 import Agda.Syntax.Literal
+import Agda.Syntax.Position
 
 import Agda.TypeChecking.Monad
+import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Pretty ( PrettyTCM(..) )
 import Agda.TypeChecking.Records
+import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 
 import Agda.Utils.Null
@@ -48,11 +59,14 @@ In these substitutions, we look for a column that has only constructor patterns.
 We try to split on this column first.
 -}
 
+type SplitAssignment = (Nat,SplitPattern)
+
 -- | Match the given patterns against a list of clauses
-match :: [Clause] -> [NamedArg SplitPattern] -> Match (Nat,[SplitPattern])
-match cs ps = foldr choice No $ zipWith matchIt [0..] cs
+match :: (MonadReduce m , HasConstInfo m , HasBuiltins m) => [Clause] -> [NamedArg SplitPattern] -> m (Match (Nat,[SplitAssignment]))
+match cs ps = foldr choice (return No) $ zipWith matchIt [0..] cs
   where
-    matchIt i c = (i,) <$> matchClause ps c
+    matchIt :: (MonadReduce m , HasConstInfo m , HasBuiltins m) => Nat -> Clause -> m (Match (Nat,[SplitAssignment]))
+    matchIt i c = fmap (i,) <$> matchClause ps c
 
 -- | For each variable in the patterns of a split clause, we remember the
 --   de Bruijn-index and the literals excluded by previous matches.
@@ -67,7 +81,7 @@ instance Pretty SplitPatVar where
     (text $ patVarNameToString (splitPatVarName x)) P.<>
     (text $ "@" ++ show (splitPatVarIndex x)) P.<>
     (ifNull (splitExcludedLits x) empty $ \lits ->
-      text "\\{" P.<> prettyList_ lits P.<> text "}")
+      "\\{" P.<> prettyList_ lits P.<> "}")
 
 instance PrettyTCM SplitPatVar where
   prettyTCM = prettyTCM . var . splitPatVarIndex
@@ -80,6 +94,10 @@ toSplitVar x = SplitPatVar (dbPatVarName x) (dbPatVarIndex x) []
 fromSplitVar :: SplitPatVar -> DBPatVar
 fromSplitVar x = DBPatVar (splitPatVarName x) (splitPatVarIndex x)
 
+instance DeBruijn SplitPatVar where
+  deBruijnView x = deBruijnView (fromSplitVar x)
+  debruijnNamedVar n i = toSplitVar (debruijnNamedVar n i)
+
 toSplitPatterns :: [NamedArg DeBruijnPattern] -> [NamedArg SplitPattern]
 toSplitPatterns = (fmap . fmap . fmap . fmap) toSplitVar
 
@@ -113,10 +131,23 @@ instance Subst SplitPattern SplitPattern where
       lookupS rho $ splitPatVarIndex x
     DotP o u     -> DotP o $ applySplitPSubst rho u
     ConP c ci ps -> ConP c ci $ applySubst rho ps
+    DefP o q ps -> DefP o q $ applySubst rho ps
     LitP x       -> p
     ProjP{}      -> p
+    IApplyP o l r x  ->
+      useEndPoints (applySplitPSubst rho l) (applySplitPSubst rho r) $
+      usePatOrigin o $
+      useName (splitPatVarName x) $
+      useExcludedLits (splitExcludedLits x) $
+      lookupS rho $ splitPatVarIndex x
 
     where
+      -- see Subst for DeBruijnPattern
+      useEndPoints :: Term -> Term -> SplitPattern -> SplitPattern
+      useEndPoints l r (VarP o x)        = IApplyP o l r x
+      useEndPoints l r (IApplyP o _ _ x) = IApplyP o l r x
+      useEndPoints l r x                 = __IMPOSSIBLE__
+
       useName :: PatVarName -> SplitPattern -> SplitPattern
       useName n (VarP o x)
         | isUnderscore (splitPatVarName x)
@@ -137,27 +168,43 @@ isTrivialPattern p = case p of
   DotP{}      -> return True
   ConP c i ps -> andM $ (isEtaCon $ conName c)
                       : (map (isTrivialPattern . namedArg) ps)
+  DefP{}      -> return False
   LitP{}      -> return False
   ProjP{}     -> return False
+  IApplyP{}   -> return True
 
 -- | If matching succeeds, we return the instantiation of the clause pattern vector
 --   to obtain the split clause pattern vector.
-type MatchResult = Match [SplitPattern]
+type MatchResult = Match [SplitAssignment]
 
 -- | If matching is inconclusive (@Block@) we want to know which
---   variables are blocking the match.
+--   variables or projections are blocking the match.
 data Match a
   = Yes a   -- ^ Matches unconditionally.
   | No      -- ^ Definitely does not match.
   | Block
-    { blockedOnResult :: Bool
-      -- ^ True if the clause has a result split
+    { blockedOnResult :: BlockedOnResult
+      -- ^ @BlockedOnProj o@ if the clause has a result split
     , blockedOnVars :: BlockingVars
       -- ^ @BlockingVar i cs ls o@ means variable @i@ is blocked on
       -- constructors @cs@ and literals @ls@.
     }
   deriving (Functor)
 
+data ApplyOrIApply = IsApply | IsIApply
+
+data BlockedOnResult
+  = BlockedOnProj      -- ^ Blocked on unsplit projection
+     { blockedOnResultOverlap :: Bool
+       -- ^ True if there are also matching clauses without an unsplit
+       -- copattern.
+     }
+  | BlockedOnApply     -- ^ Blocked on unintroduced argument
+     { blockedOnResultIApply :: ApplyOrIApply
+       -- ^ True if the unintroduced argument was an IApply pattern
+     }
+  | NotBlockedOnResult
+
 -- | Variable blocking a match.
 data BlockingVar = BlockingVar
   { blockingVarNo  :: Nat
@@ -174,21 +221,30 @@ data BlockingVar = BlockingVar
 instance Pretty BlockingVar where
   pretty (BlockingVar i cs ls o) = cat
     [ text ("variable " ++ show i)
-    , if null cs then empty else text " blocked on constructors" <+> pretty cs
-    , if null ls then empty else text " blocked on literals" <+> pretty ls
-    , if o then text " (overlapping)" else empty
+    , if null cs then empty else " blocked on constructors" <+> pretty cs
+    , if null ls then empty else " blocked on literals" <+> pretty ls
+    , if o then " (overlapping)" else empty
     ]
 
 type BlockingVars = [BlockingVar]
 
-blockedOnConstructor :: Nat -> ConHead -> Match a
-blockedOnConstructor i c = Block False [BlockingVar i [c] [] False]
+yes :: Monad m => a -> m (Match a)
+yes = return . Yes
 
-blockedOnLiteral :: Nat -> Literal -> Match a
-blockedOnLiteral i l = Block False [BlockingVar i [] [l] False]
+no :: Monad m => m (Match a)
+no = return No
 
-blockedOnProjection :: Match a
-blockedOnProjection = Block True []
+blockedOnConstructor :: Monad m => Nat -> ConHead -> m (Match a)
+blockedOnConstructor i c = return $ Block NotBlockedOnResult [BlockingVar i [c] [] False]
+
+blockedOnLiteral :: Monad m => Nat -> Literal -> m (Match a)
+blockedOnLiteral i l = return $ Block NotBlockedOnResult [BlockingVar i [] [l] False]
+
+blockedOnProjection :: Monad m => m (Match a)
+blockedOnProjection = return $ Block (BlockedOnProj False) []
+
+blockedOnApplication :: Monad m => ApplyOrIApply -> m (Match a)
+blockedOnApplication b = return $ Block (BlockedOnApply b) []
 
 -- | Lens for 'blockingVarCons'.
 mapBlockingVarCons :: ([ConHead] -> [ConHead]) -> BlockingVar -> BlockingVar
@@ -212,26 +268,50 @@ zipBlockingVars xs ys = map upd xs
       Just (BlockingVar _ cons' lits' o') -> BlockingVar x (cons ++ cons') (lits ++ lits') (o || o')
       Nothing -> BlockingVar x cons lits True
 
+setBlockedOnResultOverlap :: BlockedOnResult -> BlockedOnResult
+setBlockedOnResultOverlap b = case b of
+  BlockedOnProj{}      -> b { blockedOnResultOverlap = True }
+  BlockedOnApply{}     -> b
+  NotBlockedOnResult{} -> b
+
+anyBlockedOnResult :: BlockedOnResult -> BlockedOnResult -> BlockedOnResult
+anyBlockedOnResult b1 b2 = case (b1,b2) of
+  (NotBlockedOnResult , b2                ) -> b2
+  (b1                 , NotBlockedOnResult) -> b1
+  (_                  , _                 ) -> __IMPOSSIBLE__
+
+-- | Left dominant merge of `BlockedOnResult`.
+choiceBlockedOnResult :: BlockedOnResult -> BlockedOnResult -> BlockedOnResult
+choiceBlockedOnResult b1 b2 = case (b1,b2) of
+  (NotBlockedOnResult  , _                 ) -> NotBlockedOnResult
+  (BlockedOnProj _     , NotBlockedOnResult) -> BlockedOnProj True
+  (BlockedOnProj o1    , BlockedOnProj o2  ) -> BlockedOnProj (o1 || o2)
+  (BlockedOnProj o1    , BlockedOnApply{}  ) -> BlockedOnProj True
+  (BlockedOnApply b    , _                 ) -> BlockedOnApply b
+
 -- | @choice m m'@ combines the match results @m@ of a function clause
 --   with the (already combined) match results $m'$ of the later clauses.
 --   It is for skipping clauses that definitely do not match ('No').
 --   It is left-strict, to be used with @foldr@.
 --   If one clause unconditionally matches ('Yes') we do not look further.
-choice :: Match a -> Match a -> Match a
-choice (Yes a)      _            = Yes a
-choice (Block r xs) (Block s ys) = Block (r && s) $ zipBlockingVars xs ys
-choice (Block r xs) (Yes _)      = Block r $ overlapping xs
-choice m@Block{}    No           = m
-choice No           m            = m
+choice :: Monad m => m (Match a) -> m (Match a) -> m (Match a)
+choice m m' = m >>= \case
+  Yes a -> yes a
+  Block r xs -> m' >>= \case
+    Block s ys -> return $ Block (choiceBlockedOnResult r s) $ zipBlockingVars xs ys
+    Yes _      -> return $ Block (setBlockedOnResultOverlap r) $ overlapping xs
+    No         -> return $ Block r xs
+  No    -> m'
 
 -- | @matchClause qs i c@ checks whether clause @c@
 --   covers a split clause with patterns @qs@.
 matchClause
-  :: [NamedArg SplitPattern]
+  :: (MonadReduce m , HasConstInfo m , HasBuiltins m)
+  => [NamedArg SplitPattern]
      -- ^ Split clause patterns @qs@.
   -> Clause
      -- ^ Clause @c@ to cover split clause.
-  -> MatchResult
+  -> m MatchResult
      -- ^ Result.
      --   If 'Yes' the instantiation @rs@ such that @(namedClausePats c)[rs] == qs@.
 matchClause qs c = matchPats (namedClausePats c) qs
@@ -253,38 +333,33 @@ matchClause qs c = matchPats (namedClausePats c) qs
 --   in @mconcat []@ which is @Yes []@.
 
 matchPats
-  :: [NamedArg (Pattern' a)]
+  :: (MonadReduce m , HasConstInfo m , HasBuiltins m, DeBruijn a)
+  => [NamedArg (Pattern' a)]
      -- ^ Clause pattern vector @ps@ (to cover split clause pattern vector).
   -> [NamedArg SplitPattern]
      -- ^ Split clause pattern vector @qs@ (to be covered by clause pattern vector).
-  -> MatchResult
+  -> m MatchResult
      -- ^ Result.
      --   If 'Yes' the instantiation @rs@ such that @ps[rs] == qs@.
+matchPats [] [] = yes []
+matchPats (p:ps) (q:qs) =
+  matchPat (namedArg p) (namedArg q) `combine` matchPats ps qs
 
-matchPats ps qs = mconcat $ [ projPatternsLeftInSplitClause ] ++
-    zipWith matchPat (map namedArg ps) (map namedArg qs) ++
-    [ projPatternsLeftInMatchedClause ]
-  where
-    -- Patterns left in split clause:
-    qsrest = drop (length ps) qs
-    -- Andreas, 2016-06-03, issue #1986:
-    -- catch-all for copatterns is inconsistent as found by Ulf.
-    -- Thus, if the split clause has copatterns left,
-    -- the current (shorter) clause is not considered covering.
-    projPatternsLeftInSplitClause =
-        case mapMaybe isProjP qsrest of
-            [] -> mempty -- no proj. patterns left
-            _  -> No     -- proj. patterns left
-
-    -- Patterns left in candidate clause:
-    psrest = drop (length qs) ps
-    -- If the current clause has additional copatterns in
-    -- comparison to the split clause, we should split on them.
-    projPatternsLeftInMatchedClause =
-        case mapMaybe isProjP psrest of
-            [] -> mempty               -- no proj. patterns left
-            ds -> blockedOnProjection  -- proj. patterns left
+-- Patterns left in split clause:
+-- Andreas, 2016-06-03, issue #1986:
+-- catch-all for copatterns is inconsistent as found by Ulf.
+-- Thus, if the split clause has copatterns left,
+-- the current (shorter) clause is not considered covering.
+matchPats [] qs@(_:_) = case mapMaybe isProjP qs of
+  [] -> yes [] -- no proj. patterns left
+  _  -> no     -- proj. patterns left
 
+-- Patterns left in candidate clause:
+-- If the current clause has additional copatterns in
+-- comparison to the split clause, we should split on them.
+matchPats (p:ps) [] = case isProjP p of
+  Just{}  -> blockedOnProjection
+  Nothing -> blockedOnApplication (case namedArg p of IApplyP{} -> IsIApply; _ -> IsApply)
 
 -- | Combine results of checking whether function clause patterns
 --   covers split clause patterns.
@@ -298,19 +373,16 @@ matchPats ps qs = mconcat $ [ projPatternsLeftInSplitClause ] ++
 --   'Block' accumulates variables of the split clause
 --   that have to be instantiated (an projection names of copattern matches)
 --   to make the split clause an instance of the function clause.
-
-instance Semigroup a => Semigroup (Match a) where
-  Yes a      <> Yes b      = Yes (a <> b)
-  Yes _      <> m          = m
-  No         <> _          = No
-  Block{}    <> No         = No
-  Block r xs <> Block s ys = Block (r || s) (xs ++ ys)
-  m@Block{}  <> Yes{}      = m
-
-instance (Semigroup a, Monoid a) => Monoid (Match a) where
-  mempty  = Yes mempty
-  mappend = (<>)
-
+combine :: (Monad m, Semigroup a) => m (Match a) -> m (Match a) -> m (Match a)
+combine m m' = m >>= \case
+    Yes a -> m' >>= \case
+      Yes b -> yes (a <> b)
+      y     -> return y
+    No    -> no
+    x@(Block r xs) -> m' >>= \case
+      No    -> no
+      Block s ys -> return $ Block (anyBlockedOnResult r s) (xs ++ ys)
+      Yes{} -> return x
 
 -- | @matchPat p q@ checks whether a function clause pattern @p@
 --   covers a split clause pattern @q@.  There are three results:
@@ -323,36 +395,84 @@ instance (Semigroup a, Monoid a) => Monoid (Match a) where
 --   a cover if variable @x@ is instantiated with an appropriate literal.
 
 matchPat
-  :: Pattern' a
+  :: (MonadReduce m , HasConstInfo m , HasBuiltins m, DeBruijn a)
+  => Pattern' a
      -- ^ Clause pattern @p@ (to cover split clause pattern).
   -> SplitPattern
      -- ^ Split clause pattern @q@ (to be covered by clause pattern).
-  -> MatchResult
+  -> m MatchResult
      -- ^ Result.
      --   If 'Yes', also the instantiation @rs@ of the clause pattern variables
      --   to produce the split clause pattern, @p[rs] = q@.
+matchPat p q = case p of
+  VarP _ x   -> yes [(fromMaybe __IMPOSSIBLE__ (deBruijnView x),q)]
+  DotP{}   -> yes []
+  -- Jesper, 2014-11-04: putting 'Yes [q]' here triggers issue 1333.
+  -- Not checking for trivial patterns should be safe here, as dot patterns are
+  -- guaranteed to match if the rest of the pattern does, so some extra splitting
+  -- on them doesn't change the reduction behaviour.
+  p@(LitP l) -> case q of
+    VarP _ x -> if l `elem` splitExcludedLits x
+                then no
+                else blockedOnLiteral (splitPatVarIndex x) l
+    _ -> isLitP q >>= \case
+      Just l' -> if l == l' then yes [] else no
+      Nothing -> no
+  ProjP _ d -> case q of
+    ProjP _ d' -> do
+      d <- getOriginalProjection d
+      if d == d' then yes [] else no
+    _          -> __IMPOSSIBLE__
+  IApplyP _ _ _ x -> yes [(fromMaybe __IMPOSSIBLE__ (deBruijnView x),q)]
+  ConP c _ ps -> unDotP q >>= \case
+    VarP _ x -> blockedOnConstructor (splitPatVarIndex x) c
+    ConP c' i qs
+      | c == c'   -> matchPats ps qs
+      | otherwise -> no
+    DotP o t  -> no
+    LitP l    -> isLitP p >>= \case
+      Just l' -> if l == l' then yes [] else no
+      Nothing -> no
+    DefP{}   -> no
+    ProjP{}   -> __IMPOSSIBLE__  -- excluded by typing
+    IApplyP _ _ _ x -> blockedOnConstructor (splitPatVarIndex x) c
+  (DefP o c ps) -> unDotP q >>= \case
+    VarP _ x -> __IMPOSSIBLE__ -- blockedOnConstructor (splitPatVarIndex x) c
+    ConP c' i qs -> no
+    DotP o t  -> no
+    LitP _    -> no
+    DefP o c' qs
+      | c == c'   -> matchPats ps qs
+      | otherwise -> no
+    ProjP{}   -> __IMPOSSIBLE__  -- excluded by typing
+    IApplyP _ _ _ x -> __IMPOSSIBLE__ --blockedOnConstructor (splitPatVarIndex x) c
 
-matchPat VarP{}   q = Yes [q]
-matchPat DotP{}   q = mempty
--- Jesper, 2014-11-04: putting 'Yes [q]' here triggers issue 1333.
--- Not checking for trivial patterns should be safe here, as dot patterns are
--- guaranteed to match if the rest of the pattern does, so some extra splitting
--- on them doesn't change the reduction behaviour.
-matchPat p@(LitP l) q = case q of
-  VarP _ x -> if l `elem` splitExcludedLits x
-              then No
-              else blockedOnLiteral (splitPatVarIndex x) l
-  ConP{}   -> No
-  DotP{}   -> No
-  LitP l'  -> if l == l' then Yes [] else No
-  ProjP{}  -> __IMPOSSIBLE__  -- excluded by typing
-matchPat (ProjP _ d) (ProjP _ d') = if d == d' then mempty else No
-matchPat ProjP{} _ = __IMPOSSIBLE__
-matchPat p@(ConP c _ ps) q = case q of
-  VarP _ x -> blockedOnConstructor (splitPatVarIndex x) c
-  ConP c' i qs
-    | c == c'   -> matchPats ps qs
-    | otherwise -> No
-  DotP o t  -> No
-  LitP _    -> No
-  ProjP{}   -> __IMPOSSIBLE__  -- excluded by typing
+-- unfold one level of a dot pattern to a proper pattern if possible
+unDotP :: (MonadReduce m, DeBruijn (Pattern' a)) => Pattern' a -> m (Pattern' a)
+unDotP (DotP o v) = reduce v >>= \case
+  Var i [] -> return $ deBruijnVar i
+  Con c _ vs -> do
+    let ps = map (fmap $ unnamed . DotP o) $ fromMaybe __IMPOSSIBLE__ $ allApplyElims vs
+    return $ ConP c noConPatternInfo ps
+  Lit l -> return $ LitP l
+  v     -> return $ dotP v
+unDotP p = return p
+
+isLitP :: (MonadReduce m, HasBuiltins m) => Pattern' a -> m (Maybe Literal)
+isLitP (LitP l) = return $ Just l
+isLitP (DotP _ u) = reduce u >>= \case
+  Lit l -> return $ Just l
+  _     -> return $ Nothing
+isLitP (ConP c ci []) = do
+  Con zero _ [] <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinZero
+  if | c == zero -> return $ Just $ LitNat (getRange c) 0
+     | otherwise -> return Nothing
+isLitP (ConP c ci [a]) | visible a && isRelevant a = do
+  Con suc _ [] <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinSuc
+  if | c == suc  -> fmap inc <$> isLitP (namedArg a)
+     | otherwise -> return Nothing
+  where
+    inc :: Literal -> Literal
+    inc (LitNat r n) = LitNat (fuseRange c r) $ n + 1
+    inc _ = __IMPOSSIBLE__
+isLitP _ = return Nothing
