@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable         #-}
 
 {-| Split tree for transforming pattern clauses into case trees.
 
@@ -14,13 +15,13 @@ each leaf of the split tree.
 module Agda.TypeChecking.Coverage.SplitTree where
 
 import Data.Tree
+import Data.Data (Data, toConstr)
 
 import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Common
 import Agda.Syntax.Internal as I
 import Agda.Syntax.Literal
-
-import Agda.TypeChecking.Pretty ( PrettyTCM(..) )
+import Agda.Syntax.Position
 
 import Agda.Utils.Monad
 import Agda.Utils.Pretty
@@ -43,6 +44,7 @@ data SplitTree' a
     { splitArg   :: Arg Int       -- ^ Arg. no to split at.
     , splitTrees :: SplitTrees' a -- ^ Sub split trees.
     }
+  deriving (Data, Show)
 
 -- | Split tree branching.  A finite map from constructor names to splittrees
 --   A list representation seems appropriate, since we are expecting not
@@ -57,18 +59,13 @@ data SplitTag
   = SplitCon QName
   | SplitLit Literal
   | SplitCatchall
-  deriving (Show, Eq, Ord)
+  deriving (Show, Eq, Ord, Data)
 
 instance Pretty SplitTag where
   pretty (SplitCon c)  = pretty c
   pretty (SplitLit l)  = pretty l
   pretty SplitCatchall = underscore
 
-instance PrettyTCM SplitTag where
-  prettyTCM (SplitCon c)  = prettyTCM c
-  prettyTCM (SplitLit l)  = prettyTCM l
-  prettyTCM SplitCatchall = return underscore
-
 -- * Printing a split tree
 
 data SplitTreeLabel a = SplitTreeLabel
@@ -78,10 +75,10 @@ data SplitTreeLabel a = SplitTreeLabel
   }
 instance Pretty a => Pretty (SplitTreeLabel a) where
   pretty = \case
-    SplitTreeLabel Nothing Nothing (Just n)  -> text $ "done, " ++ show n ++ " bindings"
-    SplitTreeLabel Nothing (Just n) Nothing  -> text $ "split at " ++ show n
-    SplitTreeLabel (Just q) Nothing (Just n) -> pretty q <+> text (" -> done, " ++ show n ++ " bindings")
-    SplitTreeLabel (Just q) (Just n) Nothing -> pretty q <+> text (" -> split at " ++ show n)
+    SplitTreeLabel Nothing Nothing (Just n)  -> text $ "done, " ++ prettyShow n ++ " bindings"
+    SplitTreeLabel Nothing (Just n) Nothing  -> text $ "split at " ++ prettyShow n
+    SplitTreeLabel (Just q) Nothing (Just n) -> pretty q <+> text (" -> done, " ++ prettyShow n ++ " bindings")
+    SplitTreeLabel (Just q) (Just n) Nothing -> pretty q <+> text (" -> split at " ++ prettyShow n)
     _ -> __IMPOSSIBLE__
 
 -- | Convert a split tree into a 'Data.Tree' (for printing).
@@ -98,3 +95,14 @@ toTrees = map (\ (c,t) -> setCons c $ toTree t)
 
 instance Pretty a => Pretty (SplitTree' a) where
   pretty = text . drawTree . fmap prettyShow . toTree
+
+instance KillRange SplitTag where
+  killRange = \case
+    SplitCon c -> killRange1 SplitCon c
+    SplitLit l -> killRange1 SplitLit l
+    SplitCatchall -> SplitCatchall
+
+instance KillRange a => KillRange (SplitTree' a) where
+  killRange = \case
+    SplittingDone n -> SplittingDone n
+    SplitAt i ts    -> killRange1 (SplitAt i) ts
