@@ -2,6 +2,8 @@
 
 module Agda.TypeChecking.Datatypes where
 
+import Control.Monad.Except
+
 import Data.Maybe (fromMaybe)
 import qualified Data.List as List
 
@@ -27,8 +29,13 @@ import Agda.Utils.Impossible
 ---------------------------------------------------------------------------
 
 -- | Get true constructor with record fields.
-getConHead :: QName -> TCM (Either SigError ConHead)
-getConHead c = mapRight (conSrcCon . theDef) <$> getConstInfo' c
+getConHead :: (HasConstInfo m) => QName -> m (Either SigError ConHead)
+getConHead c = runExceptT $ do
+  def <- ExceptT $ getConstInfo' c
+  case theDef def of
+    Constructor { conSrcCon = c' } -> return c'
+    Record     { recConHead = c' } -> return c'
+    _ -> throwError $ SigUnknown $ prettyShow c ++ " is not a constructor"
 
 -- | Get true constructor with fields, expanding literals to constructors
 --   if possible.
@@ -52,6 +59,17 @@ getConstructorData c = do
     Constructor{conData = d} -> return d
     _                        -> __IMPOSSIBLE__
 
+-- | Is the datatype of this constructor a Higher Inductive Type?
+--   Precondition: The argument must refer to a constructor of a datatype or record.
+consOfHIT :: QName -> TCM Bool
+consOfHIT c = do
+  d <- getConstructorData c
+  def <- theDef <$> getConstInfo d
+  case def of
+    Datatype {dataPathCons = xs} -> return $ not $ null xs
+    Record{} -> return False
+    _  -> __IMPOSSIBLE__
+
 -- | @getConType c t@ computes the constructor parameters from type @t@
 --   and returns them plus the instantiated type of constructor @c@.
 --   This works also if @t@ is a function type ending in a data/record type;
@@ -72,9 +90,9 @@ getConType
        --     @ct@   is the type of the constructor instantiated to the parameters.
 getConType c t = do
   reportSDoc "tc.getConType" 30 $ sep $
-    [ text "getConType: constructor "
+    [ "getConType: constructor "
     , prettyTCM c
-    , text " at type "
+    , " at type "
     , prettyTCM t
     ]
   TelV tel t <- telView t
@@ -104,9 +122,10 @@ getConType c t = do
 --
 --   Precondition: @t@ is reduced.
 getFullyAppliedConType
-  :: ConHead  -- ^ Constructor.
+  :: (HasConstInfo m, MonadReduce m, MonadDebug m)
+  => ConHead  -- ^ Constructor.
   -> Type     -- ^ Reduced type of the fully applied constructor.
-  -> TCM (Maybe ((QName, Type, Args), Type))
+  -> m (Maybe ((QName, Type, Args), Type))
        -- ^ @Nothing@ if not data or record type.
        --
        --   @Just ((d, dt, pars), ct)@ otherwise, where
