@@ -61,13 +61,13 @@ displayForm q es = do
   else do
     -- Display debug info about the @Open@s.
     verboseS "tc.display.top" 100 $ unlessDebugPrinting $ do
-      cps <- view eCheckpoints
+      cps <- viewTC eCheckpoints
       cxt <- getContextTelescope
       reportSDoc "tc.display.top" 100 $ return $ vcat
-        [ text "displayForm for" <+> pretty q
-        , nest 2 $ text "cxt =" <+> pretty cxt
-        , nest 2 $ text "cps =" <+> vcat (map pretty (Map.toList cps))
-        , nest 2 $ text "dfs =" <+> vcat (map pshow odfs) ]
+        [ "displayForm for" <+> pretty q
+        , nest 2 $ "cxt =" <+> pretty cxt
+        , nest 2 $ "cps =" <+> vcat (map pretty (Map.toList cps))
+        , nest 2 $ "dfs =" <+> vcat (map pshow odfs) ]
     -- Use only the display forms that can be opened in the current context.
     dfs   <- catMaybes <$> mapM tryGetOpen odfs
     scope <- getScope
@@ -180,9 +180,15 @@ class SubstWithOrigin a where
 instance SubstWithOrigin a => SubstWithOrigin [a] where
   substWithOrigin rho ots = map (substWithOrigin rho ots)
 
-instance SubstWithOrigin (Arg a) => SubstWithOrigin (Elim' a) where
+instance (SubstWithOrigin a, SubstWithOrigin (Arg a)) => SubstWithOrigin (Elim' a) where
   substWithOrigin rho ots (Apply arg) = Apply $ substWithOrigin rho ots arg
   substWithOrigin rho ots e@Proj{}    = e
+  substWithOrigin rho ots (IApply u v w) = IApply
+    (substWithOrigin rho ots u)
+    (substWithOrigin rho ots v)
+    (substWithOrigin rho ots w)
+
+
 
 instance SubstWithOrigin (Arg Term) where
   substWithOrigin rho ots (Arg ai v) =
