@@ -287,6 +287,7 @@ instance Free Term where
     Level l      -> freeVars' l
     MetaV m ts   -> go (Flexible $ singleton m) $ freeVars' ts
     DontCare mt  -> goRel Irrelevant $ freeVars' mt
+    Dummy{}      -> mempty
 
 instance Free a => Free (Type' a) where
   freeVars' (El s t) =
@@ -299,12 +300,14 @@ instance Free Sort where
     ifM ((IgnoreAll ==) <$> asks feIgnoreSorts) mempty $ {- else -}
     case s of
       Type a     -> freeVars' a
-      Prop       -> mempty
+      Prop a     -> freeVars' a
       Inf        -> mempty
       SizeUniv   -> mempty
       PiSort s1 s2 -> go WeaklyRigid $ freeVars' (s1, s2)
       UnivSort s -> go WeaklyRigid $ freeVars' s
       MetaS x es -> go (Flexible $ singleton x) $ freeVars' es
+      DefS _ es  -> go WeaklyRigid $ freeVars' es
+      DummyS{}   -> mempty
 
 instance Free Level where
   freeVars' (Max as) = freeVars' as
@@ -332,6 +335,7 @@ instance (Free a, Free b) => Free (a, b) where
 instance Free a => Free (Elim' a) where
   freeVars' (Apply a) = freeVars' a
   freeVars' (Proj{} ) = mempty
+  freeVars' (IApply x y r) = mconcat $ map freeVars' [x,y,r]
 
 instance Free a => Free (Arg a) where
   freeVars' a = goRel (getRelevance a) $ freeVars' $ unArg a
