@@ -60,10 +60,10 @@ etaExpandClause clause = liftTCM $ do
           ps'   = raise n ps ++ teleNamedArgs tel
           body' = raise n body `apply` teleArgs tel
       reportSDoc "term.clause.expand" 30 $ inTopContext $ vcat
-        [ text "etaExpandClause"
-        , text "  body    = " <+> (addContext ctel' $ prettyTCM body)
-        , text "  xs      = " <+> text (prettyShow xs)
-        , text "  new tel = " <+> prettyTCM ctel'
+        [ "etaExpandClause"
+        , "  body    = " <+> (addContext ctel' $ prettyTCM body)
+        , "  xs      = " <+> text (prettyShow xs)
+        , "  new tel = " <+> prettyTCM ctel'
         ]
       return $ Clause rl rf ctel' ps' (Just body') (Just (t $> t')) catchall unreachable
   where
@@ -81,7 +81,9 @@ etaExpandClause clause = liftTCM $ do
     -- it means we skipped an element of the telescope.
     useNames :: [Arg ArgName] -> ListTel -> ListTel
     useNames []     tel       = map (setOrigin Inserted) tel
-    useNames (_:_)  []        = []  -- Andreas, 2017-03-24: not IMPOSSIBLE when positivity checking comes before termination checking, see examples/tactics/ac/AC.agda
+    -- Andrea: we can have more Lam's than Pi's, because they might be for Path
+    -- Andreas, 2017-03-24: the following case is not IMPOSSIBLE when positivity checking comes before termination checking, see examples/tactics/ac/AC.agda
+    useNames (_:_)  []        = []
     useNames (x:xs) (dom:tel)
       | sameHiding x dom =
           -- set the ArgName of the dom
