@@ -18,6 +18,7 @@ instance MentionsMeta Term where
     Pi a b       -> mm (a, b)
     Sort s       -> mm s
     Level l      -> mm l
+    Dummy{}      -> False
     DontCare v   -> False   -- we don't have to look inside don't cares when deciding to wake constraints
     MetaV y args -> x == y || mm args   -- TODO: we really only have to look one level deep at meta args
     where
@@ -43,12 +44,14 @@ instance MentionsMeta Type where
 instance MentionsMeta Sort where
   mentionsMeta x s = case s of
     Type l     -> mentionsMeta x l
-    Prop       -> False
+    Prop l     -> mentionsMeta x l
     Inf        -> False
     SizeUniv   -> False
     PiSort s1 s2 -> mentionsMeta x (s1, s2)
     UnivSort s -> mentionsMeta x s
     MetaS m es -> x == m || mentionsMeta x es
+    DefS d es  -> mentionsMeta x es
+    DummyS{}   -> False
 
 instance MentionsMeta t => MentionsMeta (Abs t) where
   mentionsMeta x = mentionsMeta x . unAbs
@@ -79,6 +82,7 @@ instance MentionsMeta a => MentionsMeta (Closure a) where
 instance MentionsMeta Elim where
   mentionsMeta x Proj{} = False
   mentionsMeta x (Apply v) = mentionsMeta x v
+  mentionsMeta x (IApply y0 y1 v) = mentionsMeta x (y0,y1,v)
 
 instance MentionsMeta a => MentionsMeta (Tele a) where
   mentionsMeta x EmptyTel = False
@@ -90,6 +94,7 @@ instance MentionsMeta ProblemConstraint where
 instance MentionsMeta Constraint where
   mentionsMeta x c = case c of
     ValueCmp _ t u v    -> mm (t, u, v)
+    ValueCmpOnFace _ p t u v    -> mm ((p,t), u, v)
     ElimCmp _ _ t v as bs -> mm ((t, v), (as, bs))
     LevelCmp _ u v      -> mm (u, v)
     TypeCmp _ a b       -> mm (a, b)
@@ -99,12 +104,13 @@ instance MentionsMeta Constraint where
     UnBlock _           -> True   -- this might be a postponed typechecking
                                   -- problem and we don't have a handle on
                                   -- what metas it depends on
-    FindInScope{}       -> True   -- this needs to be woken up for any meta
+    FindInstance{}      -> True   -- this needs to be woken up for any meta
     IsEmpty r t         -> mm t
     CheckSizeLtSat t    -> mm t
     CheckFunDef{}       -> True   -- not sure what metas this depends on
     HasBiggerSort a     -> mm a
     HasPTSRule a b      -> mm (a, b)
+    UnquoteTactic bl tac hole goal -> Just x == bl
     where
       mm v = mentionsMeta x v
 
