@@ -6,12 +6,10 @@ module Agda.TypeChecking.Monad
     , module Agda.TypeChecking.Monad.Debug
     , module Agda.TypeChecking.Monad.Env
     , module Agda.TypeChecking.Monad.Imports
-    , module Agda.TypeChecking.Monad.Local
     , module Agda.TypeChecking.Monad.MetaVars
     , module Agda.TypeChecking.Monad.Mutual
     , module Agda.TypeChecking.Monad.Open
     , module Agda.TypeChecking.Monad.Options
-    , module Agda.TypeChecking.Monad.Sharing
     , module Agda.TypeChecking.Monad.Signature
     , module Agda.TypeChecking.Monad.SizedTypes
     , module Agda.TypeChecking.Monad.State
@@ -27,12 +25,10 @@ import Agda.TypeChecking.Monad.Context
 import Agda.TypeChecking.Monad.Debug
 import Agda.TypeChecking.Monad.Env
 import Agda.TypeChecking.Monad.Imports
-import Agda.TypeChecking.Monad.Local
 import Agda.TypeChecking.Monad.MetaVars
 import Agda.TypeChecking.Monad.Mutual
 import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Monad.Open
-import Agda.TypeChecking.Monad.Sharing
 import Agda.TypeChecking.Monad.Signature
 import Agda.TypeChecking.Monad.SizedTypes
 import Agda.TypeChecking.Monad.State
