@@ -1,21 +1,30 @@
 {-# LANGUAGE CPP #-}
 
-module Agda.TypeChecking.Monad.Builtin where
+module Agda.TypeChecking.Monad.Builtin
+  ( module Agda.TypeChecking.Monad.Builtin
+  , module Agda.Syntax.Builtin  -- The names are defined here.
+  ) where
+
+#if __GLASGOW_HASKELL__ >= 800
+import qualified Control.Monad.Fail as Fail
+#endif
 
 import Control.Monad.State
 import Control.Monad.Trans.Maybe
 
 import qualified Data.Map as Map
+import Data.Function ( on )
 
 import Agda.Syntax.Common
 import Agda.Syntax.Position
 import Agda.Syntax.Literal
+import Agda.Syntax.Builtin
 import Agda.Syntax.Internal as I
 import Agda.TypeChecking.Monad.Base
 -- import Agda.TypeChecking.Functions  -- LEADS TO IMPORT CYCLE
 import Agda.TypeChecking.Substitute
 
-import Agda.Utils.Except ( MonadError(catchError) )
+import Agda.Utils.Except
 import Agda.Utils.Lens
 import Agda.Utils.Monad
 import Agda.Utils.Maybe
@@ -24,12 +33,25 @@ import Agda.Utils.Tuple
 #include "undefined.h"
 import Agda.Utils.Impossible
 
-class (Functor m, Applicative m, Monad m) => HasBuiltins m where
+class ( Functor m
+      , Applicative m
+#if __GLASGOW_HASKELL__ < 800
+      , Monad m
+#else
+      , Fail.MonadFail m
+#endif
+      ) => HasBuiltins m where
   getBuiltinThing :: String -> m (Maybe (Builtin PrimFun))
 
 instance HasBuiltins m => HasBuiltins (MaybeT m) where
   getBuiltinThing b = lift $ getBuiltinThing b
 
+instance HasBuiltins m => HasBuiltins (ExceptT e m) where
+  getBuiltinThing b = lift $ getBuiltinThing b
+
+instance HasBuiltins m => HasBuiltins (StateT s m) where
+  getBuiltinThing b = lift $ getBuiltinThing b
+
 litType :: Literal -> TCM Type
 litType l = case l of
   LitNat _ n    -> do
@@ -46,24 +68,27 @@ litType l = case l of
     el t = El (mkType 0) t
 
 instance MonadIO m => HasBuiltins (TCMT m) where
-  getBuiltinThing b = liftM2 mplus (Map.lookup b <$> use stLocalBuiltins)
-                      (Map.lookup b <$> use stImportedBuiltins)
+  getBuiltinThing b = liftM2 mplus (Map.lookup b <$> useTC stLocalBuiltins)
+                      (Map.lookup b <$> useTC stImportedBuiltins)
 
 setBuiltinThings :: BuiltinThings PrimFun -> TCM ()
-setBuiltinThings b = stLocalBuiltins .= b
+setBuiltinThings b = stLocalBuiltins `setTCLens` b
 
 bindBuiltinName :: String -> Term -> TCM ()
 bindBuiltinName b x = do
-        builtin <- getBuiltinThing b
-        case builtin of
-            Just (Builtin y) -> typeError $ DuplicateBuiltinBinding b y x
-            Just (Prim _)    -> typeError $ NoSuchBuiltinName b
-            Nothing          -> stLocalBuiltins %= Map.insert b (Builtin x)
+  builtin <- getBuiltinThing b
+  case builtin of
+    Just (Builtin y) -> typeError $ DuplicateBuiltinBinding b y x
+    Just (Prim _)    -> typeError $ NoSuchBuiltinName b
+    Nothing          -> stLocalBuiltins `modifyTCLens` Map.insert b (Builtin x)
 
 bindPrimitive :: String -> PrimFun -> TCM ()
 bindPrimitive b pf = do
-  builtin <- use stLocalBuiltins
-  setBuiltinThings $ Map.insert b (Prim pf) builtin
+  builtin <- getBuiltinThing b
+  case builtin of
+    Just (Builtin _) -> typeError $ NoSuchPrimitiveFunction b
+    Just (Prim x)    -> typeError $ (DuplicatePrimitiveBinding b `on` primFunName) x pf
+    Nothing          -> stLocalBuiltins `modifyTCLens` Map.insert b (Prim pf)
 
 getBuiltin :: String -> TCM Term
 getBuiltin x =
@@ -86,6 +111,27 @@ getPrimitive :: String -> TCM PrimFun
 getPrimitive x =
   fromMaybeM (typeError $ NoSuchPrimitiveFunction x) $ getPrimitive' x
 
+getPrimitiveTerm :: String -> TCM Term
+getPrimitiveTerm x = (`Def` []) <$> primFunName <$> getPrimitive x
+
+getPrimitiveTerm' :: HasBuiltins m => String -> m (Maybe Term)
+getPrimitiveTerm' x = fmap (`Def` []) <$> getPrimitiveName' x
+
+getTerm' :: HasBuiltins m => String -> m (Maybe Term)
+getTerm' x = mplus <$> getBuiltin' x <*> getPrimitiveTerm' x
+
+getName' :: HasBuiltins m => String -> m (Maybe QName)
+getName' x = mplus <$> getBuiltinName' x <*> getPrimitiveName' x
+
+-- | @getTerm use name@ looks up @name@ as a primitive or builtin, and
+-- throws an error otherwise.
+-- The @use@ argument describes how the name is used for the sake of
+-- the error message.
+getTerm :: (HasBuiltins m) => String -> String -> m Term
+getTerm use name = flip fromMaybeM (getTerm' name) $
+  return $! (throwImpossible $ ImpMissingDefinitions [name] use)
+
+
 -- | Rewrite a literal to constructor form if possible.
 constructorForm :: Term -> TCM Term
 constructorForm v = constructorForm' primZero primSuc v
@@ -105,7 +151,17 @@ constructorForm' pZero pSuc v =
 
 primInteger, primIntegerPos, primIntegerNegSuc,
     primFloat, primChar, primString, primUnit, primUnitUnit, primBool, primTrue, primFalse,
+    primSigma,
     primList, primNil, primCons, primIO, primNat, primSuc, primZero,
+    primPath, primPathP, primInterval, primIZero, primIOne, primPartial, primPartialP,
+    primIMin, primIMax, primINeg,
+    primIsOne, primItIsOne, primIsOne1, primIsOne2, primIsOneEmpty,
+    primSub, primSubIn, primSubOut,
+    primTrans, primHComp,
+    primId, primConId, primIdElim,
+    primEquiv, primEquivFun, primEquivProof, primPathToEquiv,
+    primGlue, prim_glue, prim_unglue,
+    primFaceForall,
     primNatPlus, primNatMinus, primNatTimes, primNatDivSucAux, primNatModSucAux,
     primNatEquality, primNatLess,
     -- Machine words
@@ -115,6 +171,7 @@ primInteger, primIntegerPos, primIntegerNegSuc,
     primEquality, primRefl,
     primRewrite, -- Name of rewrite relation
     primLevel, primLevelZero, primLevelSuc, primLevelMax,
+    primSetOmega,
     primFromNat, primFromNeg, primFromString,
     -- builtins for reflection:
     primQName, primArgInfo, primArgArgInfo, primArg, primArgArg, primAbs, primAbsAbs, primAgdaTerm, primAgdaTermVar,
@@ -143,7 +200,9 @@ primInteger, primIntegerPos, primIntegerNegSuc,
     primAgdaTCMGetType, primAgdaTCMGetDefinition,
     primAgdaTCMQuoteTerm, primAgdaTCMUnquoteTerm,
     primAgdaTCMBlockOnMeta, primAgdaTCMCommit, primAgdaTCMIsMacro,
-    primAgdaTCMWithNormalisation, primAgdaTCMDebugPrint
+    primAgdaTCMWithNormalisation, primAgdaTCMDebugPrint,
+    primAgdaTCMNoConstraints,
+    primAgdaTCMRunSpeculative
     :: TCM Term
 
 primInteger      = getBuiltin builtinInteger
@@ -153,6 +212,7 @@ primFloat        = getBuiltin builtinFloat
 primChar         = getBuiltin builtinChar
 primString       = getBuiltin builtinString
 primBool         = getBuiltin builtinBool
+primSigma        = getBuiltin builtinSigma
 primUnit         = getBuiltin builtinUnit
 primUnitUnit     = getBuiltin builtinUnitUnit
 primTrue         = getBuiltin builtinTrue
@@ -161,6 +221,37 @@ primList         = getBuiltin builtinList
 primNil          = getBuiltin builtinNil
 primCons         = getBuiltin builtinCons
 primIO           = getBuiltin builtinIO
+primId           = getBuiltin builtinId
+primConId        = getBuiltin builtinConId
+primIdElim       = getPrimitiveTerm builtinIdElim
+primPath         = getBuiltin builtinPath
+primPathP        = getBuiltin builtinPathP
+primInterval     = getBuiltin builtinInterval
+primIZero        = getBuiltin builtinIZero
+primIOne         = getBuiltin builtinIOne
+primIMin         = getPrimitiveTerm builtinIMin
+primIMax         = getPrimitiveTerm builtinIMax
+primINeg         = getPrimitiveTerm builtinINeg
+primPartial      = getPrimitiveTerm "primPartial"
+primPartialP     = getPrimitiveTerm "primPartialP"
+primIsOne        = getBuiltin builtinIsOne
+primItIsOne      = getBuiltin builtinItIsOne
+primTrans        = getPrimitiveTerm builtinTrans
+primHComp        = getPrimitiveTerm builtinHComp
+primEquiv        = getBuiltin builtinEquiv
+primEquivFun     = getBuiltin builtinEquivFun
+primEquivProof   = getBuiltin builtinEquivProof
+primPathToEquiv  = getBuiltin builtinPathToEquiv
+primGlue         = getPrimitiveTerm builtinGlue
+prim_glue        = getPrimitiveTerm builtin_glue
+prim_unglue      = getPrimitiveTerm builtin_unglue
+primFaceForall   = getPrimitiveTerm builtinFaceForall
+primIsOne1       = getBuiltin builtinIsOne1
+primIsOne2       = getBuiltin builtinIsOne2
+primIsOneEmpty   = getBuiltin builtinIsOneEmpty
+primSub          = getBuiltin builtinSub
+primSubIn        = getBuiltin builtinSubIn
+primSubOut       = getPrimitiveTerm builtinSubOut
 primNat          = getBuiltin builtinNat
 primSuc          = getBuiltin builtinSuc
 primZero         = getBuiltin builtinZero
@@ -188,6 +279,7 @@ primLevel        = getBuiltin builtinLevel
 primLevelZero    = getBuiltin builtinLevelZero
 primLevelSuc     = getBuiltin builtinLevelSuc
 primLevelMax     = getBuiltin builtinLevelMax
+primSetOmega     = getBuiltin builtinSetOmega
 primFromNat      = getBuiltin builtinFromNat
 primFromNeg      = getBuiltin builtinFromNeg
 primFromString   = getBuiltin builtinFromString
@@ -285,215 +377,8 @@ primAgdaTCMCommit             = getBuiltin builtinAgdaTCMCommit
 primAgdaTCMIsMacro            = getBuiltin builtinAgdaTCMIsMacro
 primAgdaTCMWithNormalisation  = getBuiltin builtinAgdaTCMWithNormalisation
 primAgdaTCMDebugPrint         = getBuiltin builtinAgdaTCMDebugPrint
-
-builtinNat, builtinSuc, builtinZero, builtinNatPlus, builtinNatMinus,
-  builtinNatTimes, builtinNatDivSucAux, builtinNatModSucAux, builtinNatEquals,
-  builtinNatLess, builtinInteger, builtinIntegerPos, builtinIntegerNegSuc,
-  builtinWord64,
-  builtinFloat, builtinChar, builtinString, builtinUnit, builtinUnitUnit,
-  builtinBool, builtinTrue, builtinFalse,
-  builtinList, builtinNil, builtinCons, builtinIO,
-  builtinSizeUniv, builtinSize, builtinSizeLt,
-  builtinSizeSuc, builtinSizeInf, builtinSizeMax,
-  builtinInf, builtinSharp, builtinFlat,
-  builtinEquality, builtinRefl, builtinRewrite, builtinLevelMax,
-  builtinLevel, builtinLevelZero, builtinLevelSuc,
-  builtinFromNat, builtinFromNeg, builtinFromString,
-  builtinQName, builtinAgdaSort, builtinAgdaSortSet, builtinAgdaSortLit,
-  builtinAgdaSortUnsupported,
-  builtinHiding, builtinHidden, builtinInstance, builtinVisible,
-  builtinRelevance, builtinRelevant, builtinIrrelevant, builtinArg,
-  builtinAssoc, builtinAssocLeft, builtinAssocRight, builtinAssocNon,
-  builtinPrecedence, builtinPrecRelated, builtinPrecUnrelated,
-  builtinFixity, builtinFixityFixity,
-  builtinArgInfo, builtinArgArgInfo, builtinArgArg,
-  builtinAbs, builtinAbsAbs, builtinAgdaTerm,
-  builtinAgdaTermVar, builtinAgdaTermLam, builtinAgdaTermExtLam,
-  builtinAgdaTermDef, builtinAgdaTermCon, builtinAgdaTermPi,
-  builtinAgdaTermSort, builtinAgdaTermLit, builtinAgdaTermUnsupported, builtinAgdaTermMeta,
-  builtinAgdaErrorPart, builtinAgdaErrorPartString, builtinAgdaErrorPartTerm, builtinAgdaErrorPartName,
-  builtinAgdaLiteral, builtinAgdaLitNat, builtinAgdaLitWord64, builtinAgdaLitFloat,
-  builtinAgdaLitChar, builtinAgdaLitString, builtinAgdaLitQName, builtinAgdaLitMeta,
-  builtinAgdaClause, builtinAgdaClauseClause, builtinAgdaClauseAbsurd, builtinAgdaPattern,
-  builtinAgdaPatVar, builtinAgdaPatCon, builtinAgdaPatDot, builtinAgdaPatLit,
-  builtinAgdaPatProj, builtinAgdaPatAbsurd,
-  builtinAgdaDefinitionFunDef,
-  builtinAgdaDefinitionDataDef, builtinAgdaDefinitionRecordDef,
-  builtinAgdaDefinitionDataConstructor, builtinAgdaDefinitionPostulate,
-  builtinAgdaDefinitionPrimitive, builtinAgdaDefinition,
-  builtinAgdaMeta,
-  builtinAgdaTCM, builtinAgdaTCMReturn, builtinAgdaTCMBind, builtinAgdaTCMUnify,
-  builtinAgdaTCMTypeError, builtinAgdaTCMInferType,
-  builtinAgdaTCMCheckType, builtinAgdaTCMNormalise, builtinAgdaTCMReduce,
-  builtinAgdaTCMCatchError,
-  builtinAgdaTCMGetContext, builtinAgdaTCMExtendContext, builtinAgdaTCMInContext,
-  builtinAgdaTCMFreshName, builtinAgdaTCMDeclareDef, builtinAgdaTCMDeclarePostulate, builtinAgdaTCMDefineFun,
-  builtinAgdaTCMGetType, builtinAgdaTCMGetDefinition,
-  builtinAgdaTCMQuoteTerm, builtinAgdaTCMUnquoteTerm,
-  builtinAgdaTCMBlockOnMeta, builtinAgdaTCMCommit, builtinAgdaTCMIsMacro,
-  builtinAgdaTCMWithNormalisation, builtinAgdaTCMDebugPrint
-  :: String
-
-builtinNat                           = "NATURAL"
-builtinSuc                           = "SUC"
-builtinZero                          = "ZERO"
-builtinNatPlus                       = "NATPLUS"
-builtinNatMinus                      = "NATMINUS"
-builtinNatTimes                      = "NATTIMES"
-builtinNatDivSucAux                  = "NATDIVSUCAUX"
-builtinNatModSucAux                  = "NATMODSUCAUX"
-builtinNatEquals                     = "NATEQUALS"
-builtinNatLess                       = "NATLESS"
-builtinWord64                        = "WORD64"
-builtinInteger                       = "INTEGER"
-builtinIntegerPos                    = "INTEGERPOS"
-builtinIntegerNegSuc                 = "INTEGERNEGSUC"
-builtinFloat                         = "FLOAT"
-builtinChar                          = "CHAR"
-builtinString                        = "STRING"
-builtinUnit                          = "UNIT"
-builtinUnitUnit                      = "UNITUNIT"
-builtinBool                          = "BOOL"
-builtinTrue                          = "TRUE"
-builtinFalse                         = "FALSE"
-builtinList                          = "LIST"
-builtinNil                           = "NIL"
-builtinCons                          = "CONS"
-builtinIO                            = "IO"
-builtinSizeUniv                      = "SIZEUNIV"
-builtinSize                          = "SIZE"
-builtinSizeLt                        = "SIZELT"
-builtinSizeSuc                       = "SIZESUC"
-builtinSizeInf                       = "SIZEINF"
-builtinSizeMax                       = "SIZEMAX"
-builtinInf                           = "INFINITY"
-builtinSharp                         = "SHARP"
-builtinFlat                          = "FLAT"
-builtinEquality                      = "EQUALITY"
-builtinRefl                          = "REFL"
-builtinRewrite                       = "REWRITE"
-builtinLevelMax                      = "LEVELMAX"
-builtinLevel                         = "LEVEL"
-builtinLevelZero                     = "LEVELZERO"
-builtinLevelSuc                      = "LEVELSUC"
-builtinFromNat                       = "FROMNAT"
-builtinFromNeg                       = "FROMNEG"
-builtinFromString                    = "FROMSTRING"
-builtinQName                         = "QNAME"
-builtinAgdaSort                      = "AGDASORT"
-builtinAgdaSortSet                   = "AGDASORTSET"
-builtinAgdaSortLit                   = "AGDASORTLIT"
-builtinAgdaSortUnsupported           = "AGDASORTUNSUPPORTED"
-builtinHiding                        = "HIDING"
-builtinHidden                        = "HIDDEN"
-builtinInstance                      = "INSTANCE"
-builtinVisible                       = "VISIBLE"
-builtinRelevance                     = "RELEVANCE"
-builtinRelevant                      = "RELEVANT"
-builtinIrrelevant                    = "IRRELEVANT"
-builtinAssoc                         = "ASSOC"
-builtinAssocLeft                     = "ASSOCLEFT"
-builtinAssocRight                    = "ASSOCRIGHT"
-builtinAssocNon                      = "ASSOCNON"
-builtinPrecedence                    = "PRECEDENCE"
-builtinPrecRelated                   = "PRECRELATED"
-builtinPrecUnrelated                 = "PRECUNRELATED"
-builtinFixity                        = "FIXITY"
-builtinFixityFixity                  = "FIXITYFIXITY"
-builtinArg                           = "ARG"
-builtinArgInfo                       = "ARGINFO"
-builtinArgArgInfo                    = "ARGARGINFO"
-builtinArgArg                        = "ARGARG"
-builtinAbs                           = "ABS"
-builtinAbsAbs                        = "ABSABS"
-builtinAgdaTerm                      = "AGDATERM"
-builtinAgdaTermVar                   = "AGDATERMVAR"
-builtinAgdaTermLam                   = "AGDATERMLAM"
-builtinAgdaTermExtLam                = "AGDATERMEXTLAM"
-builtinAgdaTermDef                   = "AGDATERMDEF"
-builtinAgdaTermCon                   = "AGDATERMCON"
-builtinAgdaTermPi                    = "AGDATERMPI"
-builtinAgdaTermSort                  = "AGDATERMSORT"
-builtinAgdaTermLit                   = "AGDATERMLIT"
-builtinAgdaTermUnsupported           = "AGDATERMUNSUPPORTED"
-builtinAgdaTermMeta                  = "AGDATERMMETA"
-builtinAgdaErrorPart                 = "AGDAERRORPART"
-builtinAgdaErrorPartString           = "AGDAERRORPARTSTRING"
-builtinAgdaErrorPartTerm             = "AGDAERRORPARTTERM"
-builtinAgdaErrorPartName             = "AGDAERRORPARTNAME"
-builtinAgdaLiteral                   = "AGDALITERAL"
-builtinAgdaLitNat                    = "AGDALITNAT"
-builtinAgdaLitWord64                 = "AGDALITWORD64"
-builtinAgdaLitFloat                  = "AGDALITFLOAT"
-builtinAgdaLitChar                   = "AGDALITCHAR"
-builtinAgdaLitString                 = "AGDALITSTRING"
-builtinAgdaLitQName                  = "AGDALITQNAME"
-builtinAgdaLitMeta                   = "AGDALITMETA"
-builtinAgdaClause                    = "AGDACLAUSE"
-builtinAgdaClauseClause              = "AGDACLAUSECLAUSE"
-builtinAgdaClauseAbsurd              = "AGDACLAUSEABSURD"
-builtinAgdaPattern                   = "AGDAPATTERN"
-builtinAgdaPatVar                    = "AGDAPATVAR"
-builtinAgdaPatCon                    = "AGDAPATCON"
-builtinAgdaPatDot                    = "AGDAPATDOT"
-builtinAgdaPatLit                    = "AGDAPATLIT"
-builtinAgdaPatProj                   = "AGDAPATPROJ"
-builtinAgdaPatAbsurd                 = "AGDAPATABSURD"
-builtinAgdaDefinitionFunDef          = "AGDADEFINITIONFUNDEF"
-builtinAgdaDefinitionDataDef         = "AGDADEFINITIONDATADEF"
-builtinAgdaDefinitionRecordDef       = "AGDADEFINITIONRECORDDEF"
-builtinAgdaDefinitionDataConstructor = "AGDADEFINITIONDATACONSTRUCTOR"
-builtinAgdaDefinitionPostulate       = "AGDADEFINITIONPOSTULATE"
-builtinAgdaDefinitionPrimitive       = "AGDADEFINITIONPRIMITIVE"
-builtinAgdaDefinition                = "AGDADEFINITION"
-builtinAgdaMeta                      = "AGDAMETA"
-builtinAgdaTCM           = "AGDATCM"
-builtinAgdaTCMReturn     = "AGDATCMRETURN"
-builtinAgdaTCMBind       = "AGDATCMBIND"
-builtinAgdaTCMUnify      = "AGDATCMUNIFY"
-builtinAgdaTCMTypeError  = "AGDATCMTYPEERROR"
-builtinAgdaTCMInferType  = "AGDATCMINFERTYPE"
-builtinAgdaTCMCheckType  = "AGDATCMCHECKTYPE"
-builtinAgdaTCMNormalise  = "AGDATCMNORMALISE"
-builtinAgdaTCMReduce     = "AGDATCMREDUCE"
-builtinAgdaTCMCatchError = "AGDATCMCATCHERROR"
-builtinAgdaTCMGetContext = "AGDATCMGETCONTEXT"
-builtinAgdaTCMExtendContext = "AGDATCMEXTENDCONTEXT"
-builtinAgdaTCMInContext     = "AGDATCMINCONTEXT"
-builtinAgdaTCMFreshName     = "AGDATCMFRESHNAME"
-builtinAgdaTCMDeclareDef    = "AGDATCMDECLAREDEF"
-builtinAgdaTCMDeclarePostulate    = "AGDATCMDECLAREPOSTULATE"
-builtinAgdaTCMDefineFun     = "AGDATCMDEFINEFUN"
-builtinAgdaTCMGetType       = "AGDATCMGETTYPE"
-builtinAgdaTCMGetDefinition = "AGDATCMGETDEFINITION"
-builtinAgdaTCMBlockOnMeta   = "AGDATCMBLOCKONMETA"
-builtinAgdaTCMCommit        = "AGDATCMCOMMIT"
-builtinAgdaTCMQuoteTerm     = "AGDATCMQUOTETERM"
-builtinAgdaTCMUnquoteTerm   = "AGDATCMUNQUOTETERM"
-builtinAgdaTCMIsMacro       = "AGDATCMISMACRO"
-builtinAgdaTCMWithNormalisation = "AGDATCMWITHNORMALISATION"
-builtinAgdaTCMDebugPrint = "AGDATCMDEBUGPRINT"
-
--- | Builtins that come without a definition in Agda syntax.
---   These are giving names to Agda internal concepts which
---   cannot be assigned an Agda type.
---
---   An example would be a user-defined name for @Set@.
---
---     {-# BUILTIN TYPE Type #-}
---
---   The type of @Type@ would be @Type : Level → Setω@
---   which is not valid Agda.
-
-builtinsNoDef :: [String]
-builtinsNoDef =
-  [ builtinSizeUniv
-  , builtinSize
-  , builtinSizeLt
-  , builtinSizeSuc
-  , builtinSizeInf
-  , builtinSizeMax
-  ]
+primAgdaTCMNoConstraints      = getBuiltin builtinAgdaTCMNoConstraints
+primAgdaTCMRunSpeculative     = getBuiltin builtinAgdaTCMRunSpeculative
 
 -- | The coinductive primitives.
 
@@ -519,6 +404,127 @@ coinductionKit' = do
 coinductionKit :: TCM (Maybe CoinductionKit)
 coinductionKit = tryMaybe coinductionKit'
 
+
+------------------------------------------------------------------------
+-- * Path equality
+------------------------------------------------------------------------
+
+getPrimName :: Term -> QName
+getPrimName ty = do
+  let lamV (Lam i b)  = mapFst (getHiding i :) $ lamV (unAbs b)
+      lamV (Pi _ b)   = lamV (unEl $ unAbs b)
+      lamV v          = ([], v)
+  case lamV ty of
+            (_, Def path _) -> path
+            (_, Con nm _ _)   -> conName nm
+            (_, _)          -> __IMPOSSIBLE__
+
+getBuiltinName', getPrimitiveName' :: HasBuiltins m => String -> m (Maybe QName)
+getBuiltinName' n = fmap getPrimName <$> getBuiltin' n
+getPrimitiveName' n = fmap primFunName <$> getPrimitive' n
+
+isPrimitive :: HasBuiltins m => String -> QName -> m Bool
+isPrimitive n q = (Just q ==) <$> getPrimitiveName' n
+
+intervalView' :: HasBuiltins m => m (Term -> IntervalView)
+intervalView' = do
+  iz <- getBuiltinName' builtinIZero
+  io <- getBuiltinName' builtinIOne
+  imax <- getPrimitiveName' "primIMax"
+  imin <- getPrimitiveName' "primIMin"
+  ineg <- getPrimitiveName' "primINeg"
+  return $ \ t ->
+    case t of
+      Def q es ->
+        case es of
+          [Apply x,Apply y] | Just q == imin -> IMin x y
+          [Apply x,Apply y] | Just q == imax -> IMax x y
+          [Apply x]         | Just q == ineg -> INeg x
+          _                 -> OTerm t
+      Con q _ [] | Just (conName q) == iz -> IZero
+                 | Just (conName q) == io -> IOne
+      _ -> OTerm t
+
+intervalView :: HasBuiltins m => Term -> m IntervalView
+intervalView t = do
+  f <- intervalView'
+  return (f t)
+
+intervalUnview :: HasBuiltins m => IntervalView -> m Term
+intervalUnview t = do
+  f <- intervalUnview'
+  return (f t)
+
+intervalUnview' :: HasBuiltins m => m (IntervalView -> Term)
+intervalUnview' = do
+  iz <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinIZero -- should it be a type error instead?
+  io <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinIOne
+  imin <- (`Def` []) <$> fromMaybe __IMPOSSIBLE__ <$> getPrimitiveName' "primIMin"
+  imax <- (`Def` []) <$> fromMaybe __IMPOSSIBLE__ <$> getPrimitiveName' "primIMax"
+  ineg <- (`Def` []) <$> fromMaybe __IMPOSSIBLE__ <$> getPrimitiveName' "primINeg"
+  return $ \ v -> case v of
+             IZero -> iz
+             IOne  -> io
+             IMin x y -> apply imin [x,y]
+             IMax x y -> apply imax [x,y]
+             INeg x   -> apply ineg [x]
+             OTerm t -> t
+
+------------------------------------------------------------------------
+-- * Path equality
+------------------------------------------------------------------------
+
+-- | Check whether the type is actually an path (lhs ≡ rhs)
+--   and extract lhs, rhs, and their type.
+--
+--   Precondition: type is reduced.
+
+pathView :: HasBuiltins m => Type -> m PathView
+pathView t0 = do
+  view <- pathView'
+  return $ view t0
+
+pathView' :: HasBuiltins m => m (Type -> PathView)
+pathView' = do
+ mpath  <- getBuiltinName' builtinPath
+ mpathp <- getBuiltinName' builtinPathP
+ return $ \ t0@(El s t) ->
+  case t of
+    Def path' [ Apply level , Apply typ , Apply lhs , Apply rhs ]
+      | Just path' == mpath, Just path <- mpathp -> PathType s path level (lam_i <$> typ) lhs rhs
+      where lam_i = Lam defaultArgInfo . NoAbs "_"
+    Def path' [ Apply level , Apply typ , Apply lhs , Apply rhs ]
+      | Just path' == mpathp, Just path <- mpathp -> PathType s path level typ lhs rhs
+    _ -> OType t0
+
+-- | Non dependent Path
+idViewAsPath :: HasBuiltins m => Type -> m PathView
+idViewAsPath t0@(El s t) = do
+  mid <- fmap getPrimName <$> getBuiltin' builtinId
+  mpath <- fmap getPrimName <$> getBuiltin' builtinPath
+  case mid of
+   Just path | isJust mpath -> case t of
+    Def path' [ Apply level , Apply typ , Apply lhs , Apply rhs ]
+      | path' == path -> return $ PathType s (fromJust mpath) level typ lhs rhs
+    _ -> return $ OType t0
+   _ -> return $ OType t0
+
+boldPathView :: Type -> PathView
+boldPathView t0@(El s t) = do
+  case t of
+    Def path' [ Apply level , Apply typ , Apply lhs , Apply rhs ]
+      -> PathType s path' level typ lhs rhs
+    _ -> OType t0
+
+-- | Revert the 'PathView'.
+--
+--   Postcondition: type is reduced.
+
+pathUnview :: PathView -> Type
+pathUnview (OType t) = t
+pathUnview (PathType s path l t lhs rhs) =
+  El s $ Def path $ map Apply [l, t, lhs, rhs]
+
 ------------------------------------------------------------------------
 -- * Builtin equality
 ------------------------------------------------------------------------
@@ -567,3 +573,19 @@ equalityUnview :: EqualityView -> Type
 equalityUnview (OtherType t) = t
 equalityUnview (EqualityType s equality l t lhs rhs) =
   El s $ Def equality $ map Apply (l ++ [t, lhs, rhs])
+
+-- | Primitives with typechecking constrants.
+constrainedPrims :: [String]
+constrainedPrims =
+  [ builtinConId
+  , builtinPOr
+  , builtinComp
+  , builtinHComp
+  , builtinTrans
+  , builtin_glue
+  ]
+
+getNameOfConstrained :: HasBuiltins m => String -> m (Maybe QName)
+getNameOfConstrained s = do
+  unless (s `elem` constrainedPrims) __IMPOSSIBLE__
+  getName' s
