@@ -12,6 +12,7 @@ import Data.Monoid
 
 import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Common
+import Agda.Syntax.Concrete.Name (LensInScope(..))
 import Agda.Syntax.Internal
 import Agda.Syntax.Scope.Monad (getLocalVars, setLocalVars)
 
@@ -38,7 +39,7 @@ import Agda.Utils.Impossible
 -- | Modify a 'Context' in a computation.
 {-# SPECIALIZE modifyContext :: (Context -> Context) -> TCM a -> TCM a #-}
 modifyContext :: MonadTCM tcm => (Context -> Context) -> tcm a -> tcm a
-modifyContext f = local $ \e -> e { envContext = f $ envContext e }
+modifyContext f = localTC $ \e -> e { envContext = f $ envContext e }
 
 -- | Change to top (=empty) context. Resets the checkpoints.
 {-# SPECIALIZE inTopContext :: TCM a -> TCM a #-}
@@ -47,8 +48,8 @@ safeInTopContext cont = do
   locals <- liftTCM $ getLocalVars
   liftTCM $ setLocalVars []
   a <- modifyContext (const [])
-        $ locally eCurrentCheckpoint (const 0)
-        $ locally eCheckpoints (const $ Map.singleton 0 IdS) cont
+        $ locallyTC eCurrentCheckpoint (const 0)
+        $ locallyTC eCheckpoints (const $ Map.singleton 0 IdS) cont
   liftTCM $ setLocalVars locals
   return a
 
@@ -65,7 +66,7 @@ inTopContext cont = do
 
 -- | Delete the last @n@ bindings from the context.
 --
---   Doesn't update checkpoints!! Use `updateContext rho (drop n)` instead,
+--   Doesn't update checkpoints! Use `updateContext rho (drop n)` instead,
 --   for an appropriate substitution `rho`.
 {-# SPECIALIZE escapeContext :: Int -> TCM a -> TCM a #-}
 escapeContext :: MonadTCM tcm => Int -> tcm a -> tcm a
@@ -77,34 +78,34 @@ escapeContext n = modifyContext $ drop n
 checkpoint :: (MonadDebug tcm, MonadTCM tcm) => Substitution -> tcm a -> tcm a
 checkpoint sub k = do
   unlessDebugPrinting $ reportSLn "tc.cxt.checkpoint" 105 $ "New checkpoint {"
-  old     <- view eCurrentCheckpoint
-  oldMods <- use  stModuleCheckpoints
+  old     <- viewTC eCurrentCheckpoint
+  oldMods <- useTC  stModuleCheckpoints
   chkpt <- fresh
   unlessDebugPrinting $ verboseS "tc.cxt.checkpoint" 105 $ do
     cxt <- getContextTelescope
-    cps <- view eCheckpoints
+    cps <- viewTC eCheckpoints
     let cps' = Map.insert chkpt IdS $ fmap (applySubst sub) cps
-        prCps cps = vcat [ pshow c <+> text ": " <+> pretty s | (c, s) <- Map.toList cps ]
+        prCps cps = vcat [ pshow c <+> ": " <+> pretty s | (c, s) <- Map.toList cps ]
     reportSDoc "tc.cxt.checkpoint" 105 $ return $ nest 2 $ vcat
-      [ text "old =" <+> pshow old
-      , text "new =" <+> pshow chkpt
-      , text "sub =" <+> pretty sub
-      , text "cxt =" <+> pretty cxt
-      , text "old substs =" <+> prCps cps
-      , text "new substs =" <?> prCps cps'
+      [ "old =" <+> pshow old
+      , "new =" <+> pshow chkpt
+      , "sub =" <+> pretty sub
+      , "cxt =" <+> pretty cxt
+      , "old substs =" <+> prCps cps
+      , "new substs =" <?> prCps cps'
       ]
-  x <- flip local k $ \ env -> env
+  x <- flip localTC k $ \ env -> env
     { envCurrentCheckpoint = chkpt
     , envCheckpoints       = Map.insert chkpt IdS $
                               fmap (applySubst sub) (envCheckpoints env)
     }
-  newMods <- use stModuleCheckpoints
+  newMods <- useTC stModuleCheckpoints
   -- Set the checkpoint for introduced modules to the old checkpoint when the
   -- new one goes out of scope. #2897: This isn't actually sound for modules
   -- created under refined parent parameters, but as long as those modules
   -- aren't named we shouldn't look at the checkpoint. The right thing to do
   -- would be to not store these modules in the checkpoint map, but todo..
-  stModuleCheckpoints .= Map.union oldMods (old <$ Map.difference newMods oldMods)
+  stModuleCheckpoints `setTCLens` Map.union oldMods (old <$ Map.difference newMods oldMods)
   unlessDebugPrinting $ reportSLn "tc.cxt.checkpoint" 105 "}"
   return x
 
@@ -114,9 +115,12 @@ updateContext :: (MonadDebug tcm, MonadTCM tcm) => Substitution -> (Context -> C
 updateContext sub f = modifyContext f . checkpoint sub
 
 -- | Get the substitution from the context at a given checkpoint to the current context.
-checkpointSubstitution :: MonadReader TCEnv tcm => CheckpointId -> tcm Substitution
-checkpointSubstitution chkpt =
-  caseMaybeM (view (eCheckpoints . key chkpt)) __IMPOSSIBLE__ return
+checkpointSubstitution :: MonadTCEnv tcm => CheckpointId -> tcm Substitution
+checkpointSubstitution = maybe __IMPOSSIBLE__ return <=< checkpointSubstitution'
+
+-- | Get the substitution from the context at a given checkpoint to the current context.
+checkpointSubstitution' :: MonadTCEnv tcm => CheckpointId -> tcm (Maybe Substitution)
+checkpointSubstitution' chkpt = viewTC (eCheckpoints . key chkpt)
 
 -- | Get substitution @Γ ⊢ ρ : Γm@ where @Γ@ is the current context
 --   and @Γm@ is the module parameter telescope of module @m@.
@@ -126,7 +130,7 @@ checkpointSubstitution chkpt =
 --   This is ok for instance if we are outside module @m@ (in which case we
 --   have to supply all module parameters to any symbol defined within @m@ we
 --   want to refer).
-getModuleParameterSub :: (MonadReader TCEnv m, ReadTCState m) => ModuleName -> m Substitution
+getModuleParameterSub :: (MonadTCEnv m, ReadTCState m) => ModuleName -> m Substitution
 getModuleParameterSub m = do
   mcp <- (^. stModuleCheckpoints . key m) <$> getTCState
   maybe (return IdS) checkpointSubstitution mcp
@@ -142,22 +146,27 @@ getModuleParameterSub m = do
 {-# SPECIALIZE addCtx :: Name -> Dom Type -> TCM a -> TCM a #-}
 addCtx :: (MonadDebug tcm, MonadTCM tcm) => Name -> Dom Type -> tcm a -> tcm a
 addCtx x a ret = do
-  let ce = (x,) <$> a
+  q <- viewTC eQuantity
+  let ce = (x,) <$> inverseApplyQuantity q a
+  when (not $ isNoName x) $ do
+    registerForShadowing x
+    ys <- getContextNames
+    forM_ ys $ \y ->
+      when (not (isNoName y) && sameRoot x y) $ tellShadowing x y
   updateContext (raiseS 1) (ce :) ret
       -- let-bindings keep track of own their context
 
--- | Pick a concrete name that doesn't shadow anything in the given list.
-unshadowedName :: [Name] -> Name -> Name
-unshadowedName xs x = head $ filter (notTaken $ map nameConcrete xs)
-                           $ iterate nextName x
   where
-    notTaken xs x = isNoName x || nameConcrete x `notElem` xs
+    -- add x to the map of possibly shadowed names
+    registerForShadowing x = modifyTCLens stShadowingNames $ Map.insert x []
 
--- | Pick a concrete name that doesn't shadow anything in the context.
-unshadowName :: MonadTCM tcm => Name -> tcm Name
-unshadowName x = do
-  ctx <- map (fst . unDom) <$> getContext
-  return $ unshadowedName ctx x
+    -- register the fact that x possibly shadows the name y
+    tellShadowing x y = modifyTCLens stShadowingNames $ Map.adjust (x:) y
+
+addRecordNameContext :: (MonadDebug m, MonadTCM m) => Dom Type -> m b -> m b
+addRecordNameContext dom ret = do
+  x <- setNotInScope <$> freshRecordName
+  addCtx x dom ret
 
 -- | Various specializations of @addCtx@.
 {-# SPECIALIZE addContext :: b -> TCM a -> TCM a #-}
@@ -165,8 +174,9 @@ class AddContext b where
   addContext :: (MonadTCM tcm, MonadDebug tcm) => b -> tcm a -> tcm a
   contextSize :: b -> Nat
 
--- | Wrapper to tell 'addContext not to 'unshadowName's. Used when adding a
---   user-provided, but already type checked, telescope to the context.
+-- | Wrapper to tell 'addContext' not to mark names as
+--   'NotInScope'. Used when adding a user-provided, but already type
+--   checked, telescope to the context.
 newtype KeepNames a = KeepNames a
 
 instance {-# OVERLAPPABLE #-} AddContext a => AddContext [a] where
@@ -196,9 +206,20 @@ instance AddContext ([WithHiding Name], Dom Type) where
     addContext (xs, raise 1 dom)
   contextSize (xs, _) = length xs
 
+instance AddContext ([Arg Name], Type) where
+  addContext (xs, t) = addContext ((map . fmap) unnamed xs :: [NamedArg Name], t)
+  contextSize (xs, _) = length xs
+
+instance AddContext ([NamedArg Name], Type) where
+  addContext ([], _)     = id
+  addContext (x : xs, t) =
+    addContext (namedArg x, t <$ domFromNamedArgName x) .
+    addContext (xs, raise 1 t)
+  contextSize (xs, _) = length xs
+
 instance AddContext (String, Dom Type) where
   addContext (s, dom) ret = do
-    x <- unshadowName =<< freshName_ s
+    x <- setNotInScope <$> freshName_ s
     addCtx x dom ret
   contextSize _ = 1
 
@@ -209,15 +230,15 @@ instance AddContext (KeepNames String, Dom Type) where
   contextSize _ = 1
 
 instance AddContext (Dom Type) where
-  addContext dom = addContext ("_", dom)
+  addContext dom = addContext ("_" :: String, dom)
   contextSize _ = 1
 
 instance AddContext Name where
-  addContext x = addContext (x, dummyDom)
+  addContext x = addContext (x, __DUMMY_DOM__)
   contextSize _ = 1
 
 instance {-# OVERLAPPING #-} AddContext String where
-  addContext s = addContext (s, dummyDom)
+  addContext s = addContext (s, __DUMMY_DOM__)
   contextSize _ = 1
 
 instance AddContext (KeepNames Telescope) where
@@ -232,11 +253,7 @@ instance AddContext Telescope where
     loop (ExtendTel t tel) = underAbstraction' id t tel loop
   contextSize = size
 
--- | Context entries without a type have this dummy type.
-dummyDom :: Dom Type
-dummyDom = defaultDom typeDontCare
-
--- | Go under an abstraction.
+-- | Go under an abstraction.  Do not extend context in case of 'NoAbs'.
 {-# SPECIALIZE underAbstraction :: Subst t a => Dom Type -> Abs a -> (a -> TCM b) -> TCM b #-}
 underAbstraction :: (Subst t a, MonadTCM tcm, MonadDebug tcm) => Dom Type -> Abs a -> (a -> tcm b) -> tcm b
 underAbstraction = underAbstraction' id
@@ -244,61 +261,79 @@ underAbstraction = underAbstraction' id
 underAbstraction' :: (Subst t a, MonadTCM tcm, MonadDebug tcm, AddContext (name, Dom Type)) =>
                      (String -> name) -> Dom Type -> Abs a -> (a -> tcm b) -> tcm b
 underAbstraction' _ _ (NoAbs _ v) k = k v
-underAbstraction' wrap t a k = addContext (wrap $ realName $ absName a, t) $ k $ absBody a
+underAbstraction' wrap t a k = underAbstractionAbs' wrap t a k
+
+-- | Go under an abstraction, treating 'NoAbs' as 'Abs'.
+underAbstractionAbs :: (Subst t a, MonadTCM tcm, MonadDebug tcm) => Dom Type -> Abs a -> (a -> tcm b) -> tcm b
+underAbstractionAbs = underAbstractionAbs' id
+
+underAbstractionAbs'
+  :: (Subst t a, MonadTCM tcm, MonadDebug tcm, AddContext (name, Dom Type))
+  => (String -> name) -> Dom Type -> Abs a -> (a -> tcm b) -> tcm b
+underAbstractionAbs' wrap t a k = addContext (wrap $ realName $ absName a, t) $ k $ absBody a
   where
     realName s = if isNoName s then "x" else argNameToString s
 
 -- | Go under an abstract without worrying about the type to add to the context.
 {-# SPECIALIZE underAbstraction_ :: Subst t a => Abs a -> (a -> TCM b) -> TCM b #-}
 underAbstraction_ :: (Subst t a, MonadTCM tcm, MonadDebug tcm) => Abs a -> (a -> tcm b) -> tcm b
-underAbstraction_ = underAbstraction dummyDom
+underAbstraction_ = underAbstraction __DUMMY_DOM__
 
--- | Add a let bound variable.
+getLetBindings :: MonadTCM tcm => tcm [(Name,(Term,Dom Type))]
+getLetBindings = do
+  bs <- asksTC envLetBindings
+  forM (Map.toList bs) $ \ (n,o) -> (,) n <$> getOpen o
+
+-- | Add a let bound variable
+{-# SPECIALIZE addLetBinding' :: Name -> Term -> Dom Type -> TCM a -> TCM a #-}
+addLetBinding' :: MonadTCM tcm => Name -> Term -> Dom Type -> tcm a -> tcm a
+addLetBinding' x v t ret = do
+    vt <- liftTCM $ makeOpen (v, t)
+    flip localTC ret $ \e -> e { envLetBindings = Map.insert x vt $ envLetBindings e }
+
+-- | Add a let bound variable
 {-# SPECIALIZE addLetBinding :: ArgInfo -> Name -> Term -> Type -> TCM a -> TCM a #-}
 addLetBinding :: MonadTCM tcm => ArgInfo -> Name -> Term -> Type -> tcm a -> tcm a
-addLetBinding info x v t0 ret = do
-    let t = Dom info t0
-    vt <- liftTCM $ makeOpen (v, t)
-    flip local ret $ \e -> e { envLetBindings = Map.insert x vt $ envLetBindings e }
+addLetBinding info x v t0 ret = addLetBinding' x v (defaultArgDom info t0) ret
 
 
 -- * Querying the context
 
 -- | Get the current context.
 {-# SPECIALIZE getContext :: TCM [Dom (Name, Type)] #-}
-getContext :: MonadReader TCEnv m => m [Dom (Name, Type)]
-getContext = asks envContext
+getContext :: MonadTCEnv m => m [Dom (Name, Type)]
+getContext = asksTC envContext
 
 -- | Get the size of the current context.
 {-# SPECIALIZE getContextSize :: TCM Nat #-}
-getContextSize :: (Applicative m, MonadReader TCEnv m) => m Nat
-getContextSize = length <$> asks envContext
+getContextSize :: (Applicative m, MonadTCEnv m) => m Nat
+getContextSize = length <$> asksTC envContext
 
 -- | Generate @[var (n - 1), ..., var 0]@ for all declarations in the context.
 {-# SPECIALIZE getContextArgs :: TCM Args #-}
-getContextArgs :: (Applicative m, MonadReader TCEnv m) => m Args
+getContextArgs :: (Applicative m, MonadTCEnv m) => m Args
 getContextArgs = reverse . zipWith mkArg [0..] <$> getContext
-  where mkArg i (Dom info _) = Arg info $ var i
+  where mkArg i dom = var i <$ argFromDom dom
 
 -- | Generate @[var (n - 1), ..., var 0]@ for all declarations in the context.
 {-# SPECIALIZE getContextTerms :: TCM [Term] #-}
-getContextTerms :: (Applicative m, MonadReader TCEnv m) => m [Term]
+getContextTerms :: (Applicative m, MonadTCEnv m) => m [Term]
 getContextTerms = map var . downFrom <$> getContextSize
 
 -- | Get the current context as a 'Telescope'.
 {-# SPECIALIZE getContextTelescope :: TCM Telescope #-}
-getContextTelescope :: (Applicative m, MonadReader TCEnv m) => m Telescope
+getContextTelescope :: (Applicative m, MonadTCEnv m) => m Telescope
 getContextTelescope = telFromList' nameToArgName . reverse <$> getContext
 
 -- | Get the names of all declarations in the context.
 {-# SPECIALIZE getContextNames :: TCM [Name] #-}
-getContextNames :: (Applicative m, MonadReader TCEnv m) => m [Name]
+getContextNames :: (Applicative m, MonadTCEnv m) => m [Name]
 getContextNames = map (fst . unDom) <$> getContext
 
 -- | get type of bound variable (i.e. deBruijn index)
 --
 {-# SPECIALIZE lookupBV :: Nat -> TCM (Dom (Name, Type)) #-}
-lookupBV :: MonadReader TCEnv m => Nat -> m (Dom (Name, Type))
+lookupBV :: MonadTCEnv m => Nat -> m (Dom (Name, Type))
 lookupBV n = do
   ctx <- getContext
   let failure = fail $ "de Bruijn index out of scope: " ++ show n ++
@@ -306,25 +341,25 @@ lookupBV n = do
   maybe failure (return . fmap (raise $ n + 1)) $ ctx !!! n
 
 {-# SPECIALIZE typeOfBV' :: Nat -> TCM (Dom Type) #-}
-typeOfBV' :: (Applicative m, MonadReader TCEnv m) => Nat -> m (Dom Type)
+typeOfBV' :: (Applicative m, MonadTCEnv m) => Nat -> m (Dom Type)
 typeOfBV' n = fmap snd <$> lookupBV n
 
 {-# SPECIALIZE typeOfBV :: Nat -> TCM Type #-}
-typeOfBV :: (Applicative m, MonadReader TCEnv m) => Nat -> m Type
+typeOfBV :: (Applicative m, MonadTCEnv m) => Nat -> m Type
 typeOfBV i = unDom <$> typeOfBV' i
 
 {-# SPECIALIZE nameOfBV :: Nat -> TCM Name #-}
-nameOfBV :: (Applicative m, MonadReader TCEnv m) => Nat -> m Name
+nameOfBV :: (Applicative m, MonadTCEnv m) => Nat -> m Name
 nameOfBV n = fst . unDom <$> lookupBV n
 
 -- | Get the term corresponding to a named variable. If it is a lambda bound
 --   variable the deBruijn index is returned and if it is a let bound variable
 --   its definition is returned.
 {-# SPECIALIZE getVarInfo :: Name -> TCM (Term, Dom Type) #-}
-getVarInfo :: MonadReader TCEnv m => Name -> m (Term, Dom Type)
+getVarInfo :: MonadTCEnv m => Name -> m (Term, Dom Type)
 getVarInfo x =
     do  ctx <- getContext
-        def <- asks envLetBindings
+        def <- asksTC envLetBindings
         case List.findIndex ((==x) . fst . unDom) ctx of
             Just n -> do
                 t <- typeOfBV' n
@@ -332,4 +367,5 @@ getVarInfo x =
             _       ->
                 case Map.lookup x def of
                     Just vt -> getOpen vt
-                    _       -> fail $ "unbound variable " ++ prettyShow (nameConcrete x)
+                    _       -> fail $ "unbound variable " ++ prettyShow (nameConcrete x) ++
+                                " (id: " ++ prettyShow (nameId x) ++ ")"
