@@ -10,11 +10,16 @@ import Control.Monad.State
 import Control.Monad.Reader
 import Control.Monad.Writer
 
+import Data.IntMap (IntMap)
+import qualified Data.IntMap as IntMap
+import Data.IntSet (IntSet)
+import qualified Data.IntSet as IntSet
 import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Set (Set)
 import qualified Data.Set as Set
 import qualified Data.Foldable as Fold
+import Data.Monoid
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
@@ -39,6 +44,7 @@ import Agda.Utils.Null
 import Agda.Utils.Permutation
 import Agda.Utils.Pretty (prettyShow)
 import Agda.Utils.Tuple
+import Agda.Utils.Singleton
 import Agda.Utils.Size
 import qualified Agda.Utils.Maybe.Strict as Strict
 
@@ -49,39 +55,42 @@ import Agda.Utils.Impossible
 dontAssignMetas :: TCM a -> TCM a
 dontAssignMetas cont = do
   reportSLn "tc.meta" 45 $ "don't assign metas"
-  local (\ env -> env { envAssignMetas = False }) cont
+  localTC (\ env -> env { envAssignMetas = False }) cont
 
 -- | Get the meta store.
 getMetaStore :: TCM MetaStore
-getMetaStore = use stMetaStore
+getMetaStore = useTC stMetaStore
 
 modifyMetaStore :: (MetaStore -> MetaStore) -> TCM ()
-modifyMetaStore f = stMetaStore %= f
+modifyMetaStore f = stMetaStore `modifyTCLens` f
 
 -- | Run a computation and record which new metas it created.
-metasCreatedBy :: TCM a -> TCM (a, Set MetaId)
+metasCreatedBy :: TCM a -> TCM (a, IntSet)
 metasCreatedBy m = do
-  before <- Map.keysSet <$> use stMetaStore
+  before <- IntMap.keysSet <$> useTC stMetaStore
   a <- m
-  after  <- Map.keysSet <$> use stMetaStore
-  return (a, after Set.\\ before)
+  after  <- IntMap.keysSet <$> useTC stMetaStore
+  return (a, after IntSet.\\ before)
+
+-- | Lookup a meta variable.
+lookupMeta' :: MetaId -> TCM (Maybe MetaVariable)
+lookupMeta' m = IntMap.lookup (metaId m) <$> getMetaStore
 
--- | Lookup a meta variable
 lookupMeta :: MetaId -> TCM MetaVariable
-lookupMeta m = fromMaybeM failure $ Map.lookup m <$> getMetaStore
+lookupMeta m = fromMaybeM failure $ lookupMeta' m
   where failure = fail $ "no such meta variable " ++ prettyShow m
 
+-- | Update the information associated with a meta variable.
 updateMetaVar :: MetaId -> (MetaVariable -> MetaVariable) -> TCM ()
-updateMetaVar m f = modifyMetaStore $ Map.adjust f m
+updateMetaVar m f = modifyMetaStore $ IntMap.adjust f $ metaId m
+
+-- | Insert a new meta variable with associated information into the meta store.
+insertMetaVar :: MetaId -> MetaVariable -> TCM ()
+insertMetaVar m mv = modifyMetaStore $ IntMap.insert (metaId m) mv
 
 getMetaPriority :: MetaId -> TCM MetaPriority
 getMetaPriority = mvPriority <.> lookupMeta
 
-{- UNUSED
-getMetaRelevance :: MetaId -> TCM Relevance
-getMetaRelevance x = miRelevance . mvInfo <$> lookupMeta x
--}
-
 isSortMeta :: MetaId -> TCM Bool
 isSortMeta m = isSortMeta_ <$> lookupMeta m
 
@@ -97,14 +106,19 @@ getMetaType m = do
     HasType{ jMetaType = t } -> t
     IsSort{}  -> __IMPOSSIBLE__
 
+-- | Compute the context variables that a meta should be applied to, accounting
+--   for pruning.
+getMetaContextArgs :: MetaVariable -> TCM Args
+getMetaContextArgs MetaVar{ mvPermutation = p } = do
+  args <- getContextArgs
+  return $ permute (takeP (length args) p) args
+
 -- | Given a meta, return the type applied to the current context.
 getMetaTypeInContext :: MetaId -> TCM Type
 getMetaTypeInContext m = do
-  MetaVar{ mvJudgement = j, mvPermutation = p } <- lookupMeta m
+  mv@MetaVar{ mvJudgement = j } <- lookupMeta m
   case j of
-    HasType{ jMetaType = t } -> do
-      vs <- getContextArgs
-      piApplyM t $ permute (takeP (size vs) p) vs
+    HasType{ jMetaType = t } -> piApplyM t =<< getMetaContextArgs mv
     IsSort{}                 -> __IMPOSSIBLE__
 
 -- | Check whether all metas are instantiated.
@@ -160,20 +174,71 @@ isInstantiatedMeta' m = do
 
 -- | Returns every meta-variable occurrence in the given type, except
 -- for those in 'Sort's.
-allMetas :: TermLike a => a -> [MetaId]
-allMetas = foldTerm metas
+allMetas :: (TermLike a, Monoid m) => (MetaId -> m) -> a -> m
+allMetas singl = foldTerm metas
   where
-  metas (MetaV m _) = [m]
+  metas (MetaV m _) = singl m
   metas (Level l)   = levelMetas l
-  metas _           = []
+  metas _           = mempty
 
-  levelMetas (Max as) = concatMap plusLevelMetas as
+  levelMetas (Max as) = foldMap plusLevelMetas as
 
-  plusLevelMetas ClosedLevel{} = []
+  plusLevelMetas ClosedLevel{} = mempty
   plusLevelMetas (Plus _ l)    = levelAtomMetas l
 
-  levelAtomMetas (MetaLevel m _) = [m]
-  levelAtomMetas _               = []
+  levelAtomMetas (MetaLevel m _) = singl m
+  levelAtomMetas _               = mempty
+
+-- | Returns 'allMetas' in a list.
+--   @allMetasList = allMetas (:[])@.
+--
+--   Note: this resulting list is computed via difference lists.
+--   Thus, use this function if you actually need the whole list of metas.
+--   Otherwise, use 'allMetas' with a suitable monoid.
+allMetasList :: TermLike a => a -> [MetaId]
+allMetasList t = allMetas singleton t `appEndo` []
+
+-- | 'True' if thing contains no metas.
+--   @noMetas = null . allMetasList@.
+noMetas :: TermLike a => a -> Bool
+noMetas = getAll . allMetas (\ _m -> All False)
+
+-- | Returns the first meta it find in the thing, if any.
+--   @firstMeta == headMaybe . allMetasList@.
+firstMeta :: TermLike a => a -> Maybe MetaId
+firstMeta = getFirst . allMetas (First . Just)
+
+-- | Returns all metavariables in a constraint. Slightly complicated by the
+--   fact that blocked terms are represented by two meta variables. To find the
+--   second one we need to look up the meta listeners for the one in the
+--   UnBlock constraint.
+constraintMetas :: Constraint -> TCM (Set MetaId)
+constraintMetas c = metas c
+  where
+    -- We don't use allMetas here since some constraints should not stop us from generalizing. For
+    -- instance CheckSizeLtSat (see #3694). We also have to check meta listeners to get metas of
+    -- UnBlock constraints.
+    metas c = case c of
+      ValueCmp _ t u v         -> return $ allMetas Set.singleton (t, u, v)
+      ValueCmpOnFace _ p t u v -> return $ allMetas Set.singleton (p, t, u, v)
+      ElimCmp _ _ t u es es'   -> return $ allMetas Set.singleton (t, u, es, es')
+      TypeCmp _ t t'           -> return $ allMetas Set.singleton (t, t')
+      LevelCmp _ l l'          -> return $ allMetas Set.singleton (l, l')
+      UnquoteTactic m t h g    -> return $ Set.fromList [x | Just x <- [m]] `Set.union` allMetas Set.singleton (t, h, g)
+      Guarded c _              -> metas c
+      TelCmp _ _ _ tel1 tel2   -> return $ allMetas Set.singleton (tel1, tel2)
+      SortCmp _ s1 s2          -> return $ allMetas Set.singleton (s1, s2)
+      UnBlock x                -> Set.insert x . Set.unions <$> (mapM listenerMetas =<< getMetaListeners x)
+      FindInstance{}           -> return mempty  -- v Ignore these constraints
+      IsEmpty{}                -> return mempty
+      CheckFunDef{}            -> return mempty
+      CheckSizeLtSat{}         -> return mempty
+      HasBiggerSort{}          -> return mempty
+      HasPTSRule{}             -> return mempty
+
+    -- For blocked constant twin variables
+    listenerMetas EtaExpand{}           = return Set.empty
+    listenerMetas (CheckConstraint _ c) = constraintMetas (clValue $ theConstraint c)
 
 -- | Create 'MetaInfo' in the current environment.
 createMetaInfo :: TCM MetaInfo
@@ -183,10 +248,12 @@ createMetaInfo' :: RunMetaOccursCheck -> TCM MetaInfo
 createMetaInfo' b = do
   r   <- getCurrentRange
   cl  <- buildClosure r
+  gen <- viewTC eGeneralizeMetas
   return MetaInfo
     { miClosRange       = cl
     , miMetaOccursCheck = b
     , miNameSuggestion  = ""
+    , miGeneralizable   = hide $ defaultArg gen
     }
 
 setValueMetaName :: Term -> MetaNameSuggestion -> TCM ()
@@ -208,6 +275,11 @@ setMetaNameSuggestion mi s = unless (null s || isUnderscore s) $ do
   updateMetaVar mi $ \ mvar ->
     mvar { mvInfo = (mvInfo mvar) { miNameSuggestion = s }}
 
+setMetaArgInfo :: MetaId -> ArgInfo -> TCM ()
+setMetaArgInfo m i = updateMetaVar m $ \ mv ->
+  mv { mvInfo = (mvInfo mv)
+        { miGeneralizable = setArgInfo i (miGeneralizable (mvInfo mv)) } }
+
 updateMetaVarRange :: MetaId -> Range -> TCM ()
 updateMetaVarRange mi r = updateMetaVar mi (setRange r)
 
@@ -219,13 +291,13 @@ setMetaOccursCheck mi b = updateMetaVar mi $ \ mvar ->
 
 modifyInteractionPoints :: (InteractionPoints -> InteractionPoints) -> TCM ()
 modifyInteractionPoints f =
-  stInteractionPoints %= f
+  stInteractionPoints `modifyTCLens` f
 
 -- | Register an interaction point during scope checking.
 --   If there is no interaction id yet, create one.
 registerInteractionPoint :: Bool -> Range -> Maybe Nat -> TCM InteractionId
 registerInteractionPoint preciseRange r maybeId = do
-  m <- use stInteractionPoints
+  m <- useTC stInteractionPoints
   -- If we're given an interaction id we shouldn't look up by range.
   -- This is important when doing 'refine', since all interaction points
   -- created by the refine gets the same range.
@@ -268,8 +340,8 @@ findInteractionPoint_ r m = do
 -- | Hook up meta variable to interaction point.
 connectInteractionPoint :: InteractionId -> MetaId -> TCM ()
 connectInteractionPoint ii mi = do
-  ipCl <- asks envClause
-  m <- use stInteractionPoints
+  ipCl <- asksTC envClause
+  m <- useTC stInteractionPoints
   let ip = InteractionPoint { ipRange = __IMPOSSIBLE__, ipMeta = Just mi, ipSolved = False, ipClause = ipCl }
   -- The interaction point needs to be present already, we just set the meta.
   case Map.insertLookupWithKey (\ key new old -> new { ipRange = ipRange old }) ii ip m of
@@ -279,21 +351,22 @@ connectInteractionPoint ii mi = do
 -- | Mark an interaction point as solved.
 removeInteractionPoint :: InteractionId -> TCM ()
 removeInteractionPoint ii = do
-  stInteractionPoints %= Map.update (\ ip -> Just ip{ ipSolved = True }) ii
+  stInteractionPoints `modifyTCLens` Map.update (\ ip -> Just ip{ ipSolved = True }) ii
 
 -- | Get a list of interaction ids.
-getInteractionPoints :: TCM [InteractionId]
-getInteractionPoints = Map.keys <$> use stInteractionPoints
+{-# SPECIALIZE getInteractionPoints :: TCM [InteractionId] #-}
+getInteractionPoints :: MonadTCM tcm => tcm [InteractionId]
+getInteractionPoints = Map.keys <$> useTC stInteractionPoints
 
 -- | Get all metas that correspond to unsolved interaction ids.
 getInteractionMetas :: TCM [MetaId]
 getInteractionMetas =
-  mapMaybe ipMeta . filter (not . ipSolved) . Map.elems <$> use stInteractionPoints
+  mapMaybe ipMeta . filter (not . ipSolved) . Map.elems <$> useTC stInteractionPoints
 
 -- | Get all metas that correspond to unsolved interaction ids.
 getInteractionIdsAndMetas :: TCM [(InteractionId,MetaId)]
 getInteractionIdsAndMetas =
-  mapMaybe f . filter (not . ipSolved . snd) . Map.toList <$> use stInteractionPoints
+  mapMaybe f . filter (not . ipSolved . snd) . Map.toList <$> useTC stInteractionPoints
   where f (ii, ip) = (ii,) <$> ipMeta ip
 
 -- | Does the meta variable correspond to an interaction point?
@@ -303,9 +376,10 @@ isInteractionMeta :: MetaId -> TCM (Maybe InteractionId)
 isInteractionMeta x = lookup x . map swap <$> getInteractionIdsAndMetas
 
 -- | Get the information associated to an interaction point.
-lookupInteractionPoint :: InteractionId -> TCM InteractionPoint
+{-# SPECIALIZE lookupInteractionPoint :: InteractionId -> TCM InteractionPoint #-}
+lookupInteractionPoint :: MonadTCM tcm => InteractionId -> tcm InteractionPoint
 lookupInteractionPoint ii =
-  fromMaybeM err $ Map.lookup ii <$> use stInteractionPoints
+  fromMaybeM err $ Map.lookup ii <$> useTC stInteractionPoints
   where
     err  = fail $ "no such interaction point: " ++ show ii
 
@@ -314,24 +388,24 @@ lookupInteractionPoint ii =
 lookupInteractionId :: InteractionId -> TCM MetaId
 lookupInteractionId ii = fromMaybeM err2 $ ipMeta <$> lookupInteractionPoint ii
   where
-    err2 = typeError $ GenericError $ "No type nor action available for hole " ++ show ii ++ ". Possible cause: the hole has not been reached during type checking (do you see yellow?)"
+    err2 = typeError $ GenericError $ "No type nor action available for hole " ++ prettyShow ii ++ ". Possible cause: the hole has not been reached during type checking (do you see yellow?)"
 
 -- | Check whether an interaction id is already associated with a meta variable.
 lookupInteractionMeta :: InteractionId -> TCM (Maybe MetaId)
-lookupInteractionMeta ii = lookupInteractionMeta_ ii <$> use stInteractionPoints
+lookupInteractionMeta ii = lookupInteractionMeta_ ii <$> useTC stInteractionPoints
 
 lookupInteractionMeta_ :: InteractionId -> InteractionPoints -> Maybe MetaId
 lookupInteractionMeta_ ii m = ipMeta =<< Map.lookup ii m
 
 -- | Generate new meta variable.
-newMeta :: MetaInfo -> MetaPriority -> Permutation -> Judgement a -> TCM MetaId
+newMeta :: Frozen -> MetaInfo -> MetaPriority -> Permutation -> Judgement a -> TCM MetaId
 newMeta = newMeta' Open
 
 -- | Generate a new meta variable with some instantiation given.
 --   For instance, the instantiation could be a 'PostponedTypeCheckingProblem'.
-newMeta' :: MetaInstantiation -> MetaInfo -> MetaPriority -> Permutation ->
+newMeta' :: MetaInstantiation -> Frozen -> MetaInfo -> MetaPriority -> Permutation ->
             Judgement a -> TCM MetaId
-newMeta' inst mi p perm j = do
+newMeta' inst frozen mi p perm j = do
   x <- fresh
   let j' = j { jMetaId = x }  -- fill the identifier part of the judgement
       mv = MetaVar{ mvInfo             = mi
@@ -340,14 +414,16 @@ newMeta' inst mi p perm j = do
                   , mvJudgement        = j'
                   , mvInstantiation    = inst
                   , mvListeners        = Set.empty
-                  , mvFrozen           = Instantiable }
+                  , mvFrozen           = frozen
+                  }
   -- printing not available (import cycle)
-  -- reportSDoc "tc.meta.new" 50 $ text "new meta" <+> prettyTCM j'
-  stMetaStore %= Map.insert x mv
+  -- reportSDoc "tc.meta.new" 50 $ "new meta" <+> prettyTCM j'
+  insertMetaVar x mv
   return x
 
 -- | Get the 'Range' for an interaction point.
-getInteractionRange :: InteractionId -> TCM Range
+{-# SPECIALIZE getInteractionRange :: InteractionId -> TCM Range #-}
+getInteractionRange :: MonadTCM tcm => InteractionId -> tcm Range
 getInteractionRange = ipRange <.> lookupInteractionPoint
 
 -- | Get the 'Range' for a meta variable.
@@ -364,25 +440,29 @@ withMetaInfo :: Closure Range -> TCM a -> TCM a
 withMetaInfo mI cont = enterClosure mI $ \ r ->
   setCurrentRange r cont
 
+getMetaVariableSet :: TCM IntSet
+getMetaVariableSet = IntMap.keysSet <$> getMetaStore
+
+getMetaVariables :: (MetaVariable -> Bool) -> TCM [MetaId]
+getMetaVariables p = do
+  store <- getMetaStore
+  return [ MetaId i | (i, mv) <- IntMap.assocs store, p mv ]
+
 getInstantiatedMetas :: TCM [MetaId]
-getInstantiatedMetas = do
-    store <- getMetaStore
-    return [ i | (i, MetaVar{ mvInstantiation = mi }) <- Map.assocs store, isInst mi ]
-    where
-        isInst Open                           = False
-        isInst OpenIFS                        = False
-        isInst BlockedConst{}                 = False
-        isInst PostponedTypeCheckingProblem{} = False
-        isInst InstV{}                        = True
+getInstantiatedMetas = getMetaVariables (isInst . mvInstantiation)
+  where
+    isInst Open                           = False
+    isInst OpenInstance                   = False
+    isInst BlockedConst{}                 = False
+    isInst PostponedTypeCheckingProblem{} = False
+    isInst InstV{}                        = True
 
 getOpenMetas :: TCM [MetaId]
-getOpenMetas = do
-    store <- getMetaStore
-    return [ i | (i, MetaVar{ mvInstantiation = mi }) <- Map.assocs store, isOpenMeta mi ]
+getOpenMetas = getMetaVariables (isOpenMeta . mvInstantiation)
 
 isOpenMeta :: MetaInstantiation -> Bool
 isOpenMeta Open                           = True
-isOpenMeta OpenIFS                        = True
+isOpenMeta OpenInstance                   = True
 isOpenMeta BlockedConst{}                 = True
 isOpenMeta PostponedTypeCheckingProblem{} = True
 isOpenMeta InstV{}                        = False
@@ -424,7 +504,7 @@ freezeMetas = freezeMetas' $ const True
 
 -- | Freeze some meta variables and return the list of metas that got frozen.
 freezeMetas' :: (MetaId -> Bool) -> TCM [MetaId]
-freezeMetas' p = execWriterT $ stMetaStore %== Map.traverseWithKey freeze
+freezeMetas' p = execWriterT $ modifyTCLensM stMetaStore $ IntMap.traverseWithKey (freeze . MetaId)
   where
   freeze :: Monad m => MetaId -> MetaVariable -> WriterT [MetaId] m MetaVariable
   freeze m mvar
@@ -439,7 +519,8 @@ unfreezeMetas = unfreezeMetas' $ const True
 
 -- | Thaw some metas, as indicated by the passed condition.
 unfreezeMetas' :: (MetaId -> Bool) -> TCM ()
-unfreezeMetas' cond = modifyMetaStore $ Map.mapWithKey unfreeze where
+unfreezeMetas' cond = modifyMetaStore $ IntMap.mapWithKey $ unfreeze . MetaId
+  where
   unfreeze :: MetaId -> MetaVariable -> MetaVariable
   unfreeze m mvar
     | cond m    = mvar { mvFrozen = Instantiable }
