@@ -21,17 +21,17 @@ import Agda.Utils.Null
 import Agda.Utils.Pretty ( prettyShow )
 
 noMutualBlock :: TCM a -> TCM a
-noMutualBlock = local $ \e -> e { envMutualBlock = Nothing }
+noMutualBlock = localTC $ \e -> e { envMutualBlock = Nothing }
 
 -- | Pass the current mutual block id
 --   or create a new mutual block if we are not already inside on.
 inMutualBlock :: (MutualId -> TCM a) -> TCM a
 inMutualBlock m = do
-  mi <- asks envMutualBlock
+  mi <- asksTC envMutualBlock
   case mi of
     Nothing -> do
       i <- fresh
-      local (\ e -> e { envMutualBlock = Just i }) $ m i
+      localTC (\ e -> e { envMutualBlock = Just i }) $ m i
     -- Don't create a new mutual block if we're already inside one.
     Just i -> m i
 
@@ -39,7 +39,7 @@ inMutualBlock m = do
 --   possibly overwriting the existing one.
 
 setMutualBlockInfo :: MutualId -> Info.MutualInfo -> TCM ()
-setMutualBlockInfo mi info = stMutualBlocks %= Map.alter f mi
+setMutualBlockInfo mi info = stMutualBlocks `modifyTCLens` Map.alter f mi
   where
   f Nothing                   = Just $ MutualBlock info empty
   f (Just (MutualBlock _ xs)) = Just $ MutualBlock info xs
@@ -47,7 +47,7 @@ setMutualBlockInfo mi info = stMutualBlocks %= Map.alter f mi
 -- | Set the mutual block info for a block if non-existing.
 
 insertMutualBlockInfo :: MutualId -> Info.MutualInfo -> TCM ()
-insertMutualBlockInfo mi info = stMutualBlocks %= Map.alter f mi
+insertMutualBlockInfo mi info = stMutualBlocks `modifyTCLens` Map.alter f mi
   where
   f Nothing = Just $ MutualBlock info empty
   f (Just mb@(MutualBlock info0 xs))
@@ -58,8 +58,8 @@ insertMutualBlockInfo mi info = stMutualBlocks %= Map.alter f mi
 
 setMutualBlock :: MutualId -> QName -> TCM ()
 setMutualBlock i x = do
-  stMutualBlocks %= Map.alter f i
-  stSignature    %= updateDefinition x (\ defn -> defn { defMutual = i })
+  stMutualBlocks `modifyTCLens` Map.alter f i
+  stSignature    `modifyTCLens` updateDefinition x (\ defn -> defn { defMutual = i })
   where
   f Nothing                    = Just $ MutualBlock empty $ Set.singleton x
   f (Just (MutualBlock mi xs)) = Just $ MutualBlock mi $ Set.insert x xs
@@ -67,11 +67,11 @@ setMutualBlock i x = do
 -- | Get the current mutual block, if any, otherwise a fresh mutual
 -- block is returned.
 currentOrFreshMutualBlock :: TCM MutualId
-currentOrFreshMutualBlock = maybe fresh return =<< asks envMutualBlock
+currentOrFreshMutualBlock = maybe fresh return =<< asksTC envMutualBlock
 
 lookupMutualBlock :: MutualId -> TCM MutualBlock
 lookupMutualBlock mi = do
-  mbs <- use stMutualBlocks
+  mbs <- useTC stMutualBlocks
   case Map.lookup mi mbs of
     Just mb -> return mb
     Nothing -> return empty -- can end up here if we ask for the current mutual block and there is none
@@ -79,7 +79,7 @@ lookupMutualBlock mi = do
 -- | Reverse lookup of a mutual block id for a names.
 mutualBlockOf :: QName -> TCM MutualId
 mutualBlockOf x = do
-  mb <- Map.toList <$> use stMutualBlocks
+  mb <- Map.toList <$> useTC stMutualBlocks
   case filter (Set.member x . mutualNames . snd) mb of
     (i, _) : _ -> return i
     _          -> fail $ "No mutual block for " ++ prettyShow x
