@@ -1,3 +1,4 @@
+{-# LANGUAGE BangPatterns      #-}
 {-# LANGUAGE CPP                      #-}
 {-# LANGUAGE NondecreasingIndentation #-}
 
@@ -7,6 +8,11 @@ import Prelude hiding (null)
 
 import Control.Arrow (first, second, (***))
 import Control.Applicative hiding (empty)
+
+#if __GLASGOW_HASKELL__ >= 800
+import qualified Control.Monad.Fail as Fail
+#endif
+
 import Control.Monad.State
 import Control.Monad.Reader
 import Control.Monad.Writer
@@ -29,6 +35,7 @@ import Agda.Syntax.Position
 import Agda.Syntax.Treeless (Compiled(..), TTerm)
 
 import Agda.TypeChecking.Monad.Base
+import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Monad.Debug
 import Agda.TypeChecking.Monad.Context
 import Agda.TypeChecking.Monad.Options
@@ -42,9 +49,12 @@ import Agda.TypeChecking.Warnings
 import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Substitute
 import {-# SOURCE #-} Agda.TypeChecking.Telescope
+import Agda.TypeChecking.CompiledClause
+import Agda.TypeChecking.Coverage.SplitTree
 import {-# SOURCE #-} Agda.TypeChecking.CompiledClause.Compile
 import {-# SOURCE #-} Agda.TypeChecking.Polarity
 import {-# SOURCE #-} Agda.TypeChecking.ProjectionLike
+import Agda.TypeChecking.Monad.Builtin
 
 import Agda.Utils.Except ( ExceptT )
 import Agda.Utils.Functor
@@ -78,7 +88,7 @@ addConstant q d = do
                     Just (doms, dom) -> telFromList $ fmap hideOrKeepInstance doms ++ [dom]
               _ -> tel
   let d' = abstract tel' $ d { defName = q }
-  reportSDoc "tc.signature" 60 $ return $ text "lambda-lifted definition =" <?> pretty d'
+  reportSDoc "tc.signature" 60 $ return $ "lambda-lifted definition =" <?> pretty d'
   modifySignature $ updateDefinitions $ HMap.insertWith (+++) q d'
   i <- currentOrFreshMutualBlock
   setMutualBlock i q
@@ -92,6 +102,20 @@ setTerminates q b = modifySignature $ updateDefinition q $ updateTheDef $ \case
     def@Function{} -> def { funTerminates = Just b }
     def -> def
 
+-- | Set CompiledClauses of a defined function symbol.
+setCompiledClauses :: QName -> CompiledClauses -> TCM ()
+setCompiledClauses q cc = modifySignature $ updateDefinition q $ updateTheDef $ setT
+  where
+    setT def@Function{} = def { funCompiled = Just cc }
+    setT def            = def
+
+-- | Set SplitTree of a defined function symbol.
+setSplitTree :: QName -> SplitTree -> TCM ()
+setSplitTree q st = modifySignature $ updateDefinition q $ updateTheDef $ setT
+  where
+    setT def@Function{} = def { funSplitTree = Just st }
+    setT def            = def
+
 -- | Modify the clauses of a function.
 modifyFunClauses :: QName -> ([Clause] -> [Clause]) -> TCM ()
 modifyFunClauses q f =
@@ -109,62 +133,21 @@ addClauses q cls = do
 mkPragma :: String -> TCM CompilerPragma
 mkPragma s = CompilerPragma <$> getCurrentRange <*> pure s
 
--- | Add a compiler pragma `{-# COMPILE <backend> <name> <text> #-}`
+-- | Add a compiler pragma `{-\# COMPILE <backend> <name> <text> \#-}`
 addPragma :: BackendName -> QName -> String -> TCM ()
 addPragma b q s = modifySignature . updateDefinition q . addCompilerPragma b =<< mkPragma s
 
--- ** Temporary **
---  The functions below are only needed while we still parse the old COMPILED
---  pragmas.
-
-type HaskellCode = String
-type HaskellType = String
-type JSCode = String
-type CoreCode = String
-
-addDeprecatedPragma :: String -> BackendName -> QName -> String -> TCM ()
-addDeprecatedPragma old b q s = do
-  let pq = prettyShow $ nameConcrete $ qnameName q
-  warning $ DeprecationWarning (unwords ["The", old, "pragma"])
-                               (unwords ["{-# COMPILE", b, pq, s, "#-}"]) "2.6"
-  addPragma b q s
-
-dataFormat :: String -> [String] -> String
-dataFormat ty cons = "= data " ++ ty ++ " (" ++ List.intercalate " | " cons ++ ")"
-
-addHaskellCode :: QName -> HaskellCode -> TCM ()
-addHaskellCode q hsCode = addDeprecatedPragma "COMPILED" ghcBackendName q $ "= " ++ hsCode
-
-addHaskellExport :: QName -> String -> TCM ()
-addHaskellExport q hsName = addDeprecatedPragma "COMPILED_EXPORT" ghcBackendName q $ "as " ++ hsName
-
-addHaskellType :: QName -> HaskellType -> TCM ()
-addHaskellType q hsTy = addDeprecatedPragma "COMPILED_TYPE" ghcBackendName q $ "= type " ++ hsTy
-
-addHaskellData :: QName -> HaskellType -> [HaskellCode] -> TCM ()
-addHaskellData q hsTy hsCons = addDeprecatedPragma "COMPILED_DATA" ghcBackendName q $ dataFormat hsTy hsCons
-
-addJSCode :: QName -> JSCode -> TCM ()
-addJSCode q jsDef = addDeprecatedPragma "COMPILED_JS" jsBackendName q ("= " ++ jsDef)
-
-addCoreCode :: QName -> CoreCode -> TCM ()
-addCoreCode q crDef = addDeprecatedPragma "COMPILED_UHC" uhcBackendName q $ "= " ++ crDef
-
-addCoreType :: QName -> CoreCode -> [CoreCode] -> TCM ()
-addCoreType q crTy crCons = addDeprecatedPragma "COMPILED_DATA_UHC" uhcBackendName q $ dataFormat crTy crCons
-
--- ** End of temporary functions **
-
 getUniqueCompilerPragma :: BackendName -> QName -> TCM (Maybe CompilerPragma)
 getUniqueCompilerPragma backend q = do
   ps <- defCompilerPragmas backend <$> getConstInfo q
   case ps of
     []  -> return Nothing
     [p] -> return $ Just p
-    _   -> setCurrentRange (ps !! 1) $
+    (_:p1:_) ->
+      setCurrentRange p1 $
             genericDocError $
-                  hang (text ("Conflicting " ++ backend ++ " pragmas for") <+> pretty q <+> text "at") 2 $
-                       vcat [ text "-" <+> pretty (getRange p) | p <- ps ]
+                  hang (text ("Conflicting " ++ backend ++ " pragmas for") <+> pretty q <+> "at") 2 $
+                       vcat [ "-" <+> pretty (getRange p) | p <- ps ]
 
 setFunctionFlag :: FunctionFlag -> Bool -> QName -> TCM ()
 setFunctionFlag flag val q = modifyGlobalDefinition q $ set (theDefLens . funFlag flag) val
@@ -214,8 +197,8 @@ addSection m = do
 -- | Sets the checkpoint for the given module to the current checkpoint.
 setModuleCheckpoint :: ModuleName -> TCM ()
 setModuleCheckpoint m = do
-  chkpt <- view eCurrentCheckpoint
-  stModuleCheckpoints %= Map.insert m chkpt
+  chkpt <- viewTC eCurrentCheckpoint
+  stModuleCheckpoints `modifyTCLens` Map.insert m chkpt
 
 -- | Get a section.
 --
@@ -343,12 +326,17 @@ applySection new ptel old ts ScopeCopyInfo{ renModules = rm, renNames = rd } = d
 
 applySection' :: ModuleName -> Telescope -> ModuleName -> Args -> ScopeCopyInfo -> TCM ()
 applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } = do
+  do
+    noCopyList <- catMaybes <$> mapM getName' constrainedPrims
+    forM_ (map fst rd) $ \ q -> do
+      when (q `elem` noCopyList) $ typeError (TriedToCopyConstrainedPrim q)
+
   reportSLn "tc.mod.apply" 10 $ render $ vcat
-    [ text "applySection"
-    , text "new  =" <+> pretty new
-    , text "ptel =" <+> pretty ptel
-    , text "old  =" <+> pretty old
-    , text "ts   =" <+> pretty ts
+    [ "applySection"
+    , "new  =" <+> pretty new
+    , "ptel =" <+> pretty ptel
+    , "old  =" <+> pretty old
+    , "ts   =" <+> pretty ts
     ]
   mapM_ (copyDef ts) rd
   mapM_ (copySec ts) rm
@@ -413,6 +401,7 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
             t   = defType d `piApply` ts'
             pol = defPolarity d `apply` ts'
             occ = defArgOccurrences d `apply` ts'
+            gen = defArgGeneralizable d `apply` ts'
             inst = defInstance d
             -- the name is set by the addConstant function
             nd :: QName -> TCM Definition
@@ -422,12 +411,15 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                     , defType           = t
                     , defPolarity       = pol
                     , defArgOccurrences = occ
+                    , defArgGeneralizable = gen
+                    , defGeneralizedParams = [] -- This is only needed for type checking data/record defs so no need to copy it.
                     , defDisplay        = []
                     , defMutual         = -1   -- TODO: mutual block?
                     , defCompiledRep    = noCompiledRep
                     , defInstance       = inst
                     , defCopy           = True
                     , defMatchable      = False
+                    , defNoCompilation  = defNoCompilation d
                     , defInjective      = False
                     , theDef            = df }
             oldDef = theDef d
@@ -464,8 +456,9 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                          , recClause  = Just cl
                          , recTel     = apply tel ts'
                          }
+                GeneralizableVar -> return GeneralizableVar
                 _ -> do
-                  cc <- compileClauses Nothing [cl] -- Andreas, 2012-10-07 non need for record pattern translation
+                  (mst, cc) <- compileClauses Nothing [cl] -- Andreas, 2012-10-07 non need for record pattern translation
                   let newDef =
                         set funMacro  (oldDef ^. funMacro) $
                         set funStatic (oldDef ^. funStatic) $
@@ -473,6 +466,7 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                         emptyFunction
                         { funClauses        = [cl]
                         , funCompiled       = Just cc
+                        , funSplitTree      = mst
                         , funMutual         = mutual
                         , funProjection     = proj
                         , funTerminates     = Just True
@@ -480,7 +474,7 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                         , funWith           = with
                         , funCopatternLHS   = isCopatternLHS [cl]
                         }
-                  reportSDoc "tc.mod.apply" 80 $ return $ (text "new def for" <+> pretty x) <?> pretty newDef
+                  reportSDoc "tc.mod.apply" 80 $ return $ ("new def for" <+> pretty x) <?> pretty newDef
                   return newDef
 
             cl = Clause { clauseLHSRange  = getRange $ defClauses d
@@ -548,18 +542,18 @@ addDisplayForm x df = do
   let add = updateDefinition x $ \ def -> def{ defDisplay = d : defDisplay def }
   ifM (isLocal x)
     {-then-} (modifySignature add)
-    {-else-} (stImportsDisplayForms %= HMap.insertWith (++) x [d])
+    {-else-} (stImportsDisplayForms `modifyTCLens` HMap.insertWith (++) x [d])
   whenM (hasLoopingDisplayForm x) $
-    typeError . GenericDocError $ text "Cannot add recursive display form for" <+> pretty x
+    typeError . GenericDocError $ "Cannot add recursive display form for" <+> pretty x
 
 isLocal :: QName -> TCM Bool
-isLocal x = HMap.member x <$> use (stSignature . sigDefinitions)
+isLocal x = HMap.member x <$> useTC (stSignature . sigDefinitions)
 
 getDisplayForms :: QName -> TCM [LocalDisplayForm]
 getDisplayForms q = do
   ds  <- either (const []) defDisplay <$> getConstInfo' q
-  ds1 <- HMap.lookupDefault [] q <$> use stImportsDisplayForms
-  ds2 <- HMap.lookupDefault [] q <$> use stImportedDisplayForms
+  ds1 <- HMap.lookupDefault [] q <$> useTC stImportsDisplayForms
+  ds2 <- HMap.lookupDefault [] q <$> useTC stImportedDisplayForms
   ifM (isLocal q) (return $ ds ++ ds1 ++ ds2)
                   (return $ ds1 ++ ds ++ ds2)
 
@@ -602,11 +596,15 @@ sameDef d1 d2 = do
 whatInduction :: MonadTCM tcm => QName -> tcm Induction
 whatInduction c = liftTCM $ do
   def <- theDef <$> getConstInfo c
+  mz <- getBuiltinName' builtinIZero
+  mo <- getBuiltinName' builtinIOne
   case def of
     Datatype{ dataInduction = i } -> return i
     Record{} | not (recRecursive def) -> return Inductive
     Record{ recInduction = i    } -> return $ fromMaybe Inductive i
     Constructor{ conInd = i }     -> return i
+    _ | Just c == mz || Just c == mo
+                                  -> return Inductive
     _                             -> __IMPOSSIBLE__
 
 -- | Does the given constructor come from a single-constructor type?
@@ -636,7 +634,17 @@ sigError f a = \case
   SigUnknown s -> f s
   SigAbstract  -> a
 
-class (Functor m, Applicative m, Monad m, HasOptions m, MonadDebug m, MonadReader TCEnv m) => HasConstInfo m where
+class ( Functor m
+      , Applicative m
+#if __GLASGOW_HASKELL__ < 800
+      , Monad m
+#else
+      , Fail.MonadFail m
+#endif
+      , HasOptions m
+      , MonadDebug m
+      , MonadTCEnv m
+      ) => HasConstInfo m where
   -- | Lookup the definition of a name. The result is a closed thing, all free
   --   variables have been abstracted over.
   getConstInfo :: QName -> m Definition
@@ -669,10 +677,10 @@ getOriginalProjection :: HasConstInfo m => QName -> m QName
 getOriginalProjection q = projOrig . fromMaybe __IMPOSSIBLE__ <$> isProjection q
 
 instance HasConstInfo (TCMT IO) where
-  getRewriteRulesFor = defaultGetRewriteRulesFor get
+  getRewriteRulesFor = defaultGetRewriteRulesFor getTC
   getConstInfo' q = do
-    st  <- get
-    env <- ask
+    st  <- getTC
+    env <- askTC
     defaultGetConstInfo st env q
   getConstInfo q = getConstInfo' q >>= \case
       Right d -> return d
@@ -680,7 +688,7 @@ instance HasConstInfo (TCMT IO) where
       Left SigAbstract      -> notInScope $ qnameToConcrete q
 
 defaultGetConstInfo
-  :: (HasOptions m, MonadDebug m, MonadReader TCEnv m)
+  :: (HasOptions m, MonadDebug m, MonadTCEnv m)
   => TCState -> TCEnv -> QName -> m (Either SigError Definition)
 defaultGetConstInfo st env q = do
     let defs  = st^.(stSignature . sigDefinitions)
@@ -716,13 +724,21 @@ instance HasConstInfo m => HasConstInfo (ExceptT err m) where
   getConstInfo' = lift . getConstInfo'
   getRewriteRulesFor = lift . getRewriteRulesFor
 
+instance HasConstInfo m => HasConstInfo (ReaderT r m) where
+  getConstInfo' = lift . getConstInfo'
+  getRewriteRulesFor = lift . getRewriteRulesFor
+
 instance (Monoid w, HasConstInfo m) => HasConstInfo (WriterT w m) where
   getConstInfo' = lift . getConstInfo'
   getRewriteRulesFor = lift . getRewriteRulesFor
 
+instance HasConstInfo m => HasConstInfo (StateT s m) where
+  getConstInfo' = lift . getConstInfo'
+  getRewriteRulesFor = lift . getRewriteRulesFor
+
 {-# INLINE getConInfo #-}
-getConInfo :: MonadTCM tcm => ConHead -> tcm Definition
-getConInfo = liftTCM . getConstInfo . conName
+getConInfo :: HasConstInfo m => ConHead -> m Definition
+getConInfo = getConstInfo . conName
 
 -- | Look up the polarity of a definition.
 getPolarity :: QName -> TCM [Polarity]
@@ -799,6 +815,14 @@ getTreeless q = fmap cTreeless <$> getCompiled q
 getCompiledArgUse :: QName -> TCM [Bool]
 getCompiledArgUse q = maybe [] cArgUsage <$> getCompiled q
 
+-- | add data constructors to a datatype
+addDataCons :: QName -> [QName] -> TCM ()
+addDataCons d cs = modifySignature $ updateDefinition d $ updateTheDef $ \ def ->
+  let !cs' = cs ++ dataCons def in
+  case def of
+    Datatype{} -> def {dataCons = cs' }
+    _          -> __IMPOSSIBLE__
+
 -- | Get the mutually recursive identifiers of a symbol from the signature.
 getMutual :: QName -> TCM (Maybe [QName])
 getMutual d = getMutual_ . theDef <$> getConstInfo d
@@ -839,28 +863,28 @@ getDefModule :: HasConstInfo m => QName -> m ModuleName
 getDefModule f = do
   def <- getConstInfo f
   return $ case theDef def of
-    Function{ funExtLam = Just (ExtLamInfo m) } -> m
-    _                                           -> qnameModule f
+    Function{ funExtLam = Just (ExtLamInfo m _) } -> m
+    _                                             -> qnameModule f
 
 -- | Compute the number of free variables of a defined name. This is the sum of
 --   number of parameters shared with the current module and the number of
 --   anonymous variables (if the name comes from a let-bound module).
-getDefFreeVars :: (Functor m, Applicative m, ReadTCState m, MonadReader TCEnv m) => QName -> m Nat
+getDefFreeVars :: (Functor m, Applicative m, ReadTCState m, MonadTCEnv m) => QName -> m Nat
 getDefFreeVars = getModuleFreeVars . qnameModule
 
 freeVarsToApply :: (Functor m, HasConstInfo m, HasOptions m,
-                    ReadTCState m, MonadReader TCEnv m, MonadDebug m)
+                    ReadTCState m, MonadTCEnv m, MonadDebug m)
                 => QName -> m Args
 freeVarsToApply q = do
   vs <- moduleParamsToApply $ qnameModule q
   t <- defType <$> getConstInfo q
   let TelV tel _ = telView'UpTo (size vs) t
   unless (size tel == size vs) __IMPOSSIBLE__
-  return $ zipWith (\ (Arg _ v) (Dom ai _) -> Arg ai v) vs $ telToList tel
+  return $ zipWith (\ arg dom -> unArg arg <$ argFromDom dom) vs $ telToList tel
 
 {-# SPECIALIZE getModuleFreeVars :: ModuleName -> TCM Nat #-}
 {-# SPECIALIZE getModuleFreeVars :: ModuleName -> ReduceM Nat #-}
-getModuleFreeVars :: (Functor m, Applicative m, MonadReader TCEnv m, ReadTCState m)
+getModuleFreeVars :: (Functor m, Applicative m, MonadTCEnv m, ReadTCState m)
                   => ModuleName -> m Nat
 getModuleFreeVars m = do
   m0   <- commonParentModule m <$> currentModule
@@ -881,7 +905,7 @@ getModuleFreeVars m = do
 --          ... M₁.M₂.f [insert Γ raised by Θ]
 --   @
 moduleParamsToApply :: (Functor m, Applicative m, HasOptions m,
-                        MonadReader TCEnv m, ReadTCState m, MonadDebug m)
+                        MonadTCEnv m, ReadTCState m, MonadDebug m)
                     => ModuleName -> m Args
 moduleParamsToApply m = do
   -- Get the correct number of free variables (correctly raised) of @m@.
@@ -919,7 +943,7 @@ moduleParamsToApply m = do
       -- drop some @args@.
       -- And there are also anonymous modules, thus, the invariant is not trivial.
       when (size stel < size args) __IMPOSSIBLE__
-      return $ zipWith (\ (Dom ai _) (Arg _ v) -> Arg ai v) (telToList stel) args
+      return $ zipWith (\ !dom (Arg _ v) -> v <$ argFromDom dom) (telToList stel) args
 
 -- | Unless all variables in the context are module parameters, create a fresh
 --   module to capture the non-module parameters. Used when unquoting to make
@@ -929,13 +953,18 @@ inFreshModuleIfFreeParams k = do
   sub <- getModuleParameterSub =<< currentModule
   if sub == IdS then k else do
     m  <- currentModule
-    m' <- qualifyM m . mnameFromList . (:[]) <$> freshName_ "_"
+    m' <- qualifyM m . mnameFromList . (:[]) <$>
+            freshName_ ("_" :: String)
     addSection m'
     withCurrentModule m' k
 
 -- | Instantiate a closed definition with the correct part of the current
 --   context.
-instantiateDef :: Definition -> TCM Definition
+{-# SPECIALIZE instantiateDef :: Definition -> TCM Definition #-}
+instantiateDef
+  :: ( Functor m, HasConstInfo m, HasOptions m
+     , ReadTCState m, MonadTCEnv m, MonadDebug m )
+  => Definition -> m Definition
 instantiateDef d = do
   vs  <- freeVarsToApply $ defName d
   verboseS "tc.sig.inst" 30 $ do
@@ -947,18 +976,18 @@ instantiateDef d = do
   return $ d `apply` vs
 
 instantiateRewriteRule :: (Functor m, HasConstInfo m, HasOptions m,
-                           ReadTCState m, MonadReader TCEnv m, MonadDebug m)
+                           ReadTCState m, MonadTCEnv m, MonadDebug m)
                        => RewriteRule -> m RewriteRule
 instantiateRewriteRule rew = do
-  traceSLn "rewriting" 60 ("instantiating rewrite rule " ++ show (rewName rew) ++ " to the local context.") $ do
+  traceSLn "rewriting" 95 ("instantiating rewrite rule " ++ show (rewName rew) ++ " to the local context.") $ do
   vs  <- freeVarsToApply $ rewName rew
   let rew' = rew `apply` vs
-  traceSLn "rewriting" 60 ("instantiated rewrite rule: ") $ do
-  traceSLn "rewriting" 60 (show rew') $ do
+  traceSLn "rewriting" 95 ("instantiated rewrite rule: ") $ do
+  traceSLn "rewriting" 95 (show rew') $ do
   return rew'
 
 instantiateRewriteRules :: (Functor m, HasConstInfo m, HasOptions m,
-                            ReadTCState m, MonadReader TCEnv m, MonadDebug m)
+                            ReadTCState m, MonadTCEnv m, MonadDebug m)
                         => RewriteRules -> m RewriteRules
 instantiateRewriteRules = mapM instantiateRewriteRule
 
@@ -975,6 +1004,8 @@ makeAbstract d =
                }
   where
     makeAbs Axiom         = Just Axiom
+    makeAbs d@DataOrRecSig{}     = Just d
+    makeAbs d@GeneralizableVar{} = Just d
     makeAbs d@Datatype {} = Just $ AbstractDefn d
     makeAbs d@Function {} = Just $ AbstractDefn d
     makeAbs Constructor{} = Nothing
@@ -988,28 +1019,22 @@ makeAbstract d =
 
 -- | Enter abstract mode. Abstract definition in the current module are transparent.
 {-# SPECIALIZE inAbstractMode :: TCM a -> TCM a #-}
-inAbstractMode :: MonadReader TCEnv m => m a -> m a
-inAbstractMode = local $ \e -> e { envAbstractMode = AbstractMode,
-                                   envAllowDestructiveUpdate = False }
-                                    -- Allowing destructive updates when seeing through
-                                    -- abstract may break the abstraction.
+inAbstractMode :: MonadTCEnv m => m a -> m a
+inAbstractMode = localTC $ \e -> e { envAbstractMode = AbstractMode }
 
 -- | Not in abstract mode. All abstract definitions are opaque.
 {-# SPECIALIZE inConcreteMode :: TCM a -> TCM a #-}
-inConcreteMode :: MonadReader TCEnv m => m a -> m a
-inConcreteMode = local $ \e -> e { envAbstractMode = ConcreteMode }
+inConcreteMode :: MonadTCEnv m => m a -> m a
+inConcreteMode = localTC $ \e -> e { envAbstractMode = ConcreteMode }
 
 -- | Ignore abstract mode. All abstract definitions are transparent.
-ignoreAbstractMode :: MonadReader TCEnv m => m a -> m a
-ignoreAbstractMode = local $ \e -> e { envAbstractMode = IgnoreAbstractMode,
-                                       envAllowDestructiveUpdate = False }
-                                       -- Allowing destructive updates when ignoring
-                                       -- abstract may break the abstraction.
+ignoreAbstractMode :: MonadTCEnv m => m a -> m a
+ignoreAbstractMode = localTC $ \e -> e { envAbstractMode = IgnoreAbstractMode }
 
 -- | Enter concrete or abstract mode depending on whether the given identifier
 --   is concrete or abstract.
 {-# SPECIALIZE inConcreteOrAbstractMode :: QName -> (Definition -> TCM a) -> TCM a #-}
-inConcreteOrAbstractMode :: (MonadReader TCEnv m, HasConstInfo m) => QName -> (Definition -> m a) -> m a
+inConcreteOrAbstractMode :: (MonadTCEnv m, HasConstInfo m) => QName -> (Definition -> m a) -> m a
 inConcreteOrAbstractMode q cont = do
   -- Andreas, 2015-07-01: If we do not ignoreAbstractMode here,
   -- we will get ConcreteDef for abstract things, as they are turned into axioms.
@@ -1021,8 +1046,8 @@ inConcreteOrAbstractMode q cont = do
 -- | Check whether a name might have to be treated abstractly (either if we're
 --   'inAbstractMode' or it's not a local name). Returns true for things not
 --   declared abstract as well, but for those 'makeAbstract' will have no effect.
-treatAbstractly :: MonadReader TCEnv m => QName -> m Bool
-treatAbstractly q = asks $ treatAbstractly' q
+treatAbstractly :: MonadTCEnv m => QName -> m Bool
+treatAbstractly q = asksTC $ treatAbstractly' q
 
 -- | Andreas, 2015-07-01:
 --   If the @current@ module is a weak suffix of the identifier module,
@@ -1043,21 +1068,31 @@ treatAbstractly' q env = case envAbstractMode env of
   where
     current = dropAnon $ envCurrentModule env
     m       = dropAnon $ qnameModule q
-    dropAnon (MName ms) = MName $ reverse $ dropWhile isNoName $ reverse ms
+    dropAnon (MName ms) = MName $ List.dropWhileEnd isNoName ms
 
 -- | Get type of a constant, instantiated to the current context.
-typeOfConst :: QName -> TCM Type
+{-# SPECIALIZE typeOfConst :: QName -> TCM Type #-}
+typeOfConst
+  :: ( Functor m, HasConstInfo m, HasOptions m
+     , ReadTCState m, MonadTCEnv m, MonadDebug m )
+  => QName -> m Type
 typeOfConst q = defType <$> (instantiateDef =<< getConstInfo q)
 
 -- | Get relevance of a constant.
 relOfConst :: QName -> TCM Relevance
 relOfConst q = defRelevance <$> getConstInfo q
 
+-- | Get modality of a constant.
+modalityOfConst :: QName -> TCM Modality
+modalityOfConst q = getModality . defArgInfo <$> getConstInfo q
+
 -- | The number of dropped parameters for a definition.
 --   0 except for projection(-like) functions and constructors.
 droppedPars :: Definition -> Int
 droppedPars d = case theDef d of
     Axiom{}                  -> 0
+    DataOrRecSig{}           -> 0
+    GeneralizableVar{}       -> 0
     def@Function{}           -> projectionArgs def
     Datatype  {dataPars = _} -> 0  -- not dropped
     Record     {recPars = _} -> 0  -- not dropped
@@ -1105,7 +1140,8 @@ usesCopatterns q = do
 
 -- | Apply a function @f@ to its first argument, producing the proper
 --   postfix projection if @f@ is a projection.
-applyDef :: ProjOrigin -> QName -> Arg Term -> TCM Term
+applyDef :: (HasConstInfo m)
+         => ProjOrigin -> QName -> Arg Term -> m Term
 applyDef o f a = do
   let fallback = return $ Def f [Apply a]
   caseMaybeM (isProjection f) fallback $ \ isP -> do
