@@ -8,7 +8,7 @@ module Agda.TypeChecking.Monad.State where
 import Control.Arrow (first)
 import qualified Control.Exception as E
 import Control.Monad.Reader (asks)
-import Control.Monad.State (put, get, gets, modify, modify')
+import Control.Monad.State (put, get, gets, modify, modify', void)
 import Control.Monad.Trans (liftIO)
 
 import Data.Maybe
@@ -55,8 +55,8 @@ import Agda.Utils.Impossible
 
 resetState :: TCM ()
 resetState = do
-    pers <- gets stPersistentState
-    put $ initState { stPersistentState = pers }
+    pers <- getsTC stPersistentState
+    putTC $ initState { stPersistentState = pers }
 
 -- | Resets all of the type checking state.
 --
@@ -65,34 +65,47 @@ resetState = do
 resetAllState :: TCM ()
 resetAllState = do
     b <- getBenchmark
-    backends <- use stBackends
-    put $ updatePersistentState (\ s -> s { stBenchmark = b }) initState
-    stBackends .= backends
--- resetAllState = put initState
+    backends <- useTC stBackends
+    putTC $ updatePersistentState (\ s -> s { stBenchmark = b }) initState
+    stBackends `setTCLens` backends
+-- resetAllState = putTC initState
 
 -- | Restore 'TCState' after performing subcomputation.
 --
 --   In contrast to 'Agda.Utils.Monad.localState', the 'Benchmark'
 --   info from the subcomputation is saved.
 localTCState :: TCM a -> TCM a
-localTCState = disableDestructiveUpdate . bracket_ get (\ s -> do
+localTCState = bracket_ getTC (\ s -> do
    b <- getBenchmark
-   put s
+   putTC s
    modifyBenchmark $ const b)
 
 -- | Same as 'localTCState' but also returns the state in which we were just
 --   before reverting it.
 localTCStateSaving :: TCM a -> TCM (a, TCState)
 localTCStateSaving compute = do
-  oldState <- get
+  oldState <- getTC
   result <- compute
-  newState <- get
+  newState <- getTC
   do
     b <- getBenchmark
-    put oldState
+    putTC oldState
     modifyBenchmark $ const b
   return (result, newState)
 
+data SpeculateResult = SpeculateAbort | SpeculateCommit
+
+-- | Allow rolling back the state changes of a TCM computation.
+speculateTCState :: TCM (a, SpeculateResult) -> TCM a
+speculateTCState m = do
+  ((x, res), newState) <- localTCStateSaving m
+  case res of
+    SpeculateAbort  -> return x
+    SpeculateCommit -> x <$ putTC newState
+
+speculateTCState_ :: TCM SpeculateResult -> TCM ()
+speculateTCState_ m = void $ speculateTCState $ ((),) <$> m
+
 -- | A fresh TCM instance.
 --
 -- The computation is run in a fresh state, with the exception that
@@ -103,19 +116,19 @@ localTCStateSaving compute = do
 
 freshTCM :: TCM a -> TCM (Either TCErr a)
 freshTCM m = do
-  ps <- use lensPersistentState
+  ps <- useTC lensPersistentState
   let s = set lensPersistentState ps initState
   r <- liftIO $ (Right <$> runTCM initEnv s m) `E.catch` (return . Left)
   case r of
     Right (a, s) -> do
-      lensPersistentState .= s ^. lensPersistentState
+      setTCLens lensPersistentState $ s ^. lensPersistentState
       return $ Right a
     Left err -> do
       case err of
         TypeError { tcErrState = s } ->
-          lensPersistentState .= s ^. lensPersistentState
+          setTCLens lensPersistentState $ s ^. lensPersistentState
         IOException s _ _ ->
-          lensPersistentState .= s ^. lensPersistentState
+          setTCLens lensPersistentState $ s ^. lensPersistentState
         _ -> return ()
       return $ Left err
 
@@ -132,7 +145,7 @@ updatePersistentState
 updatePersistentState f s = s { stPersistentState = f (stPersistentState s) }
 
 modifyPersistentState :: (PersistentTCState -> PersistentTCState) -> TCM ()
-modifyPersistentState = modify . updatePersistentState
+modifyPersistentState = modifyTC . updatePersistentState
 
 -- | Lens for 'stAccumStatistics'.
 
@@ -149,7 +162,7 @@ lensAccumStatistics =  lensPersistentState . lensAccumStatisticsP
 
 -- | Get the current scope.
 getScope :: TCM ScopeInfo
-getScope = use stScope
+getScope = useTC stScope
 
 -- | Set the current scope.
 setScope :: ScopeInfo -> TCM ()
@@ -157,7 +170,7 @@ setScope scope = modifyScope (const scope)
 
 -- | Modify the current scope without updating the inverse maps.
 modifyScope_ :: (ScopeInfo -> ScopeInfo) -> TCM ()
-modifyScope_ f = stScope %= f
+modifyScope_ f = stScope `modifyTCLens` f
 
 -- | Modify the current scope.
 modifyScope :: (ScopeInfo -> ScopeInfo) -> TCM ()
@@ -204,13 +217,13 @@ printScope tag v s = verboseS ("scope." ++ tag) v $ do
 -- ** Lens for 'stSignature' and 'stImports'
 
 modifySignature :: (Signature -> Signature) -> TCM ()
-modifySignature f = stSignature %= f
+modifySignature f = stSignature `modifyTCLens` f
 
 modifyImportedSignature :: (Signature -> Signature) -> TCM ()
-modifyImportedSignature f = stImports %= f
+modifyImportedSignature f = stImports `modifyTCLens` f
 
 getSignature :: TCM Signature
-getSignature = use stSignature
+getSignature = useTC stSignature
 
 -- | Update a possibly imported definition. Warning: changes made to imported
 --   definitions (during type checking) will not persist outside the current
@@ -278,6 +291,10 @@ updateFunClauses :: ([Clause] -> [Clause]) -> (Defn -> Defn)
 updateFunClauses f def@Function{ funClauses = cs} = def { funClauses = f cs }
 updateFunClauses f _                              = __IMPOSSIBLE__
 
+updateCovering :: ([Closure Clause] -> [Closure Clause]) -> (Defn -> Defn)
+updateCovering f def@Function{ funCovering = cs} = def { funCovering = f cs }
+updateCovering f _                               = __IMPOSSIBLE__
+
 updateCompiledClauses :: (Maybe CompiledClauses -> Maybe CompiledClauses) -> (Defn -> Defn)
 updateCompiledClauses f def@Function{ funCompiled = cc} = def { funCompiled = f cc }
 updateCompiledClauses f _                              = __IMPOSSIBLE__
@@ -297,16 +314,16 @@ updateFunCopatternLHS f _ = __IMPOSSIBLE__
 -- implementation of 'setTopLevelModule' should be changed.
 
 setTopLevelModule :: C.QName -> TCM ()
-setTopLevelModule x = stFreshNameId .= NameId 0 (hashString $ prettyShow x)
+setTopLevelModule x = stFreshNameId `setTCLens` NameId 0 (hashString $ prettyShow x)
 
 -- | Use a different top-level module for a computation. Used when generating
 --   names for imported modules.
 withTopLevelModule :: C.QName -> TCM a -> TCM a
 withTopLevelModule x m = do
-  next <- use stFreshNameId
+  next <- useTC stFreshNameId
   setTopLevelModule x
   y <- m
-  stFreshNameId .= next
+  stFreshNameId `setTCLens` next
   return y
 
 ---------------------------------------------------------------------------
@@ -315,36 +332,8 @@ withTopLevelModule x m = do
 
 addForeignCode :: BackendName -> String -> TCM ()
 addForeignCode backend code = do
-  r <- asks envRange  -- can't use TypeChecking.Monad.Trace.getCurrentRange without cycle
-  stForeignCode . key backend %= Just . (ForeignCode r code :) . fromMaybe []
-
----------------------------------------------------------------------------
--- * Temporary: Haskell imports
---   These will go away when we remove the IMPORT and HASKELL pragmas in
---   favour of the FOREIGN pragma.
----------------------------------------------------------------------------
-
-addDeprecatedForeignCode :: String -> BackendName -> String -> TCM ()
-addDeprecatedForeignCode old backend code = do
-  warning $ DeprecationWarning (unwords ["The", old, "pragma"])
-                               foreignPragma "2.6"
-  addForeignCode backend code
-  where
-    spc | length (lines code) > 1 = "\n"
-        | otherwise               = " "
-    foreignPragma =
-      "{-# FOREIGN " ++ backend ++ spc ++ code ++ spc ++ "#-}"
-
--- | Tell the compiler to import the given Haskell module.
-addHaskellImport :: String -> TCM ()
-addHaskellImport i = addDeprecatedForeignCode "IMPORT" ghcBackendName $ "import qualified " ++ i
-
--- | Tell the compiler to import the given Haskell module.
-addHaskellImportUHC :: String -> TCM ()
-addHaskellImportUHC i = addDeprecatedForeignCode "IMPORT_UHC" ghcBackendName $ "__IMPORT__ " ++ i
-
-addInlineHaskell :: String -> TCM ()
-addInlineHaskell s = addDeprecatedForeignCode "HASKELL" ghcBackendName s
+  r <- asksTC envRange  -- can't use TypeChecking.Monad.Trace.getCurrentRange without cycle
+  modifyTCLens (stForeignCode . key backend) $ Just . (ForeignCode r code :) . fromMaybe []
 
 ---------------------------------------------------------------------------
 -- * Interaction output callback
@@ -352,11 +341,11 @@ addInlineHaskell s = addDeprecatedForeignCode "HASKELL" ghcBackendName s
 
 getInteractionOutputCallback :: TCM InteractionOutputCallback
 getInteractionOutputCallback
-  = gets $ stInteractionOutputCallback . stPersistentState
+  = getsTC $ stInteractionOutputCallback . stPersistentState
 
 appInteractionOutputCallback :: Response -> TCM ()
 appInteractionOutputCallback r
-  = getInteractionOutputCallback >>= \ cb -> liftIO $ cb r
+  = getInteractionOutputCallback >>= \ cb -> cb r
 
 setInteractionOutputCallback :: InteractionOutputCallback -> TCM ()
 setInteractionOutputCallback cb
@@ -367,17 +356,17 @@ setInteractionOutputCallback cb
 ---------------------------------------------------------------------------
 
 getPatternSyns :: TCM PatternSynDefns
-getPatternSyns = use stPatternSyns
+getPatternSyns = useTC stPatternSyns
 
 setPatternSyns :: PatternSynDefns -> TCM ()
 setPatternSyns m = modifyPatternSyns (const m)
 
 -- | Lens for 'stPatternSyns'.
 modifyPatternSyns :: (PatternSynDefns -> PatternSynDefns) -> TCM ()
-modifyPatternSyns f = stPatternSyns %= f
+modifyPatternSyns f = stPatternSyns `modifyTCLens` f
 
 getPatternSynImports :: TCM PatternSynDefns
-getPatternSynImports = use stPatternSynImports
+getPatternSynImports = useTC stPatternSynImports
 
 -- | Get both local and imported pattern synonyms
 getAllPatternSyns :: TCM PatternSynDefns
@@ -415,11 +404,11 @@ updateBenchmark f = updatePersistentState $ \ s -> s { stBenchmark = f (stBenchm
 
 -- | Lens getter for 'Benchmark' from 'TCM'.
 getBenchmark :: TCM Benchmark
-getBenchmark = gets $ theBenchmark
+getBenchmark = getsTC $ theBenchmark
 
 -- | Lens modify for 'Benchmark'.
 modifyBenchmark :: (Benchmark -> Benchmark) -> TCM ()
-modifyBenchmark = modify' . updateBenchmark
+modifyBenchmark = modifyTC' . updateBenchmark
 
 ---------------------------------------------------------------------------
 -- * Instance definitions
@@ -431,19 +420,19 @@ addImportedInstances sig = do
   let itable = Map.fromListWith Set.union
                [ (c, Set.singleton i)
                | (i, Defn{ defInstance = Just c }) <- HMap.toList $ sig ^. sigDefinitions ]
-  stImportedInstanceDefs %= Map.unionWith Set.union itable
+  stImportedInstanceDefs `modifyTCLens` Map.unionWith Set.union itable
 
 -- | Lens for 'stInstanceDefs'.
 updateInstanceDefs :: (TempInstanceTable -> TempInstanceTable) -> (TCState -> TCState)
 updateInstanceDefs = over stInstanceDefs
 
 modifyInstanceDefs :: (TempInstanceTable -> TempInstanceTable) -> TCM ()
-modifyInstanceDefs = modify . updateInstanceDefs
+modifyInstanceDefs = modifyTC . updateInstanceDefs
 
 getAllInstanceDefs :: TCM TempInstanceTable
 getAllInstanceDefs = do
-  (table,xs) <- use stInstanceDefs
-  itable <- use stImportedInstanceDefs
+  (table,xs) <- useTC stInstanceDefs
+  itable <- useTC stImportedInstanceDefs
   let !table' = Map.unionWith Set.union itable table
   return (table', xs)
 
