@@ -13,6 +13,7 @@ import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Pattern
 
 import Agda.TypeChecking.Monad
+import Agda.TypeChecking.Datatypes (getNumberOfParameters)
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.SizedTypes
 import Agda.TypeChecking.Substitute
@@ -142,9 +143,9 @@ computePolarity xs = do
   -- t <- instantiateFull t -- Andreas, 2014-04-11 Issue 1099: needed for
   --                        -- variable occurrence test in  dependentPolarity.
   reportSDoc "tc.polarity.set" 15 $
-    text "Refining polarity with type " <+> prettyTCM t
+    "Refining polarity with type " <+> prettyTCM t
   reportSDoc "tc.polarity.set" 60 $
-    text "Refining polarity with type (raw): " <+> (text .show) t
+    "Refining polarity with type (raw): " <+> (text .show) t
 
   pol <- dependentPolarity t (enablePhantomTypes (theDef def) pol1) pol1
   reportSLn "tc.polarity.set" 10 $ "Polarity of " ++ prettyShow x ++ ": " ++ prettyShow pol
@@ -196,8 +197,8 @@ dependentPolarity t _      []          = return []  -- all remaining are 'Invari
 dependentPolarity t []     (_ : _)     = __IMPOSSIBLE__
 dependentPolarity t (q:qs) pols@(p:ps) = do
   t <- reduce $ unEl t
-  reportSDoc "tc.polarity.dep" 20 $ text "dependentPolarity t = " <+> prettyTCM t
-  reportSDoc "tc.polarity.dep" 70 $ text "dependentPolarity t = " <+> (text . show) t
+  reportSDoc "tc.polarity.dep" 20 $ "dependentPolarity t = " <+> prettyTCM t
+  reportSDoc "tc.polarity.dep" 70 $ "dependentPolarity t = " <+> (text . show) t
   case t of
     Pi dom b -> do
       ps <- underAbstraction dom b $ \ c -> dependentPolarity c qs ps
@@ -235,7 +236,7 @@ relevantInIgnoringNonvariant i t (p:ps) = do
 sizePolarity :: QName -> [Polarity] -> TCM [Polarity]
 sizePolarity d pol0 = do
   let exit = return pol0
-  ifM (not . optSizedTypes <$> pragmaOptions) exit $ do
+  ifNotM sizedTypesOption exit $ {- else -} do
     def <- getConstInfo d
     case theDef def of
       Datatype{ dataPars = np, dataCons = cons } -> do
@@ -243,7 +244,7 @@ sizePolarity d pol0 = do
             (parTel, ixTel) = splitAt np $ telToList tel
         case ixTel of
           []                 -> exit  -- No size index
-          Dom _ (_, a) : _ -> ifM ((/= Just BoundedNo) <$> isSizeType a) exit $ do
+          Dom{unDom = (_,a)} : _ -> ifM ((/= Just BoundedNo) <$> isSizeType a) exit $ do
             -- we assume the size index to be 'Covariant' ...
             let pol   = take np pol0
                 polCo = pol ++ [Covariant]
@@ -271,13 +272,13 @@ sizePolarity d pol0 = do
                           -- check that the size argument appears in the
                           -- right spot in the target type
                           let sizeArg = size tel
-                              isLin = addContext conTel $ checkSizeIndex d np sizeArg target
+                              isLin = addContext conTel $ checkSizeIndex d sizeArg target
 
                           ok <- isPos `and2M` isLin
                           reportSDoc "tc.polarity.size" 15 $
-                            text "constructor" <+> prettyTCM c <+>
+                            "constructor" <+> prettyTCM c <+>
                             text (if ok then "passes" else "fails") <+>
-                            text "size polarity check"
+                            "size polarity check"
                           return ok
 
             ifNotM (andM $ map check cons)
@@ -290,27 +291,27 @@ sizePolarity d pol0 = do
                 return polCo
       _ -> exit
 
--- | @checkSizeIndex d np i a@ checks that constructor target type @a@
---   has form @d ps (↑ⁿ i) idxs@ where @|ps| = np@.
+-- | @checkSizeIndex d i a@ checks that constructor target type @a@
+--   has form @d ps (↑ⁿ i) idxs@ where @|ps| = np(d)@.
 --
 --   Precondition: @a@ is reduced and of form @d ps idxs0@.
-checkSizeIndex :: QName -> Nat -> Nat -> Type -> TCM Bool
-checkSizeIndex d np i a = do
+checkSizeIndex :: QName -> Nat -> Type -> TCM Bool
+checkSizeIndex d i a = do
   reportSDoc "tc.polarity.size" 15 $ withShowAllArguments $ vcat
-    [ text "checking that constructor target type " <+> prettyTCM a
-    , text "  is data type " <+> prettyTCM d
-    , text "  and has size index (successor(s) of) " <+> prettyTCM (var i)
+    [ "checking that constructor target type " <+> prettyTCM a
+    , "  is data type " <+> prettyTCM d
+    , "  and has size index (successor(s) of) " <+> prettyTCM (var i)
     ]
   case unEl a of
     Def d0 es -> do
       whenNothingM (sameDef d d0) __IMPOSSIBLE__
+      np <- fromMaybe __IMPOSSIBLE__ <$> getNumberOfParameters d0
+      let (pars, Apply ix : ixs) = splitAt np es
       s <- deepSizeView $ unArg ix
       case s of
         DSizeVar j _ | i == j
           -> return $ not $ freeIn i (pars ++ ixs)
         _ -> return False
-      where
-        (pars, Apply ix : ixs) = splitAt np es
     _ -> __IMPOSSIBLE__
 
 -- | @polarities i a@ computes the list of polarities of de Bruijn index @i@
@@ -349,6 +350,7 @@ instance HasPolarity Type where
 instance HasPolarity a => HasPolarity (Elim' a) where
   polarities i Proj{}    = return []
   polarities i (Apply a) = polarities i a
+  polarities i (IApply x y a) = polarities i (x,(y,a))
 
 instance HasPolarity Term where
   polarities i v = do
@@ -370,6 +372,7 @@ instance HasPolarity Term where
     Sort s     -> return [] -- polarities i s -- return []
     MetaV _ ts -> map (const Invariant) <$> polarities i ts
     DontCare t -> polarities i t -- return []
+    Dummy{}    -> return []
 
 instance HasPolarity Level where
   polarities i (Max as) = polarities i as
