@@ -47,6 +47,7 @@ import qualified Agda.Syntax.Concrete.Pretty as CP
 import qualified Agda.Syntax.Info as A
 import Agda.Syntax.Scope.Monad (withContextPrecedence)
 
+import Agda.TypeChecking.Coverage.SplitTree
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin (equalityUnview)
 import Agda.TypeChecking.Positivity.Occurrence
@@ -191,9 +192,9 @@ instance PrettyTCM R.Term       where prettyTCM = prettyA <=< toAbstractWithoutI
 
 
 instance (Pretty a, PrettyTCM a, Subst a a) => PrettyTCM (Substitution' a) where
-  prettyTCM IdS        = text "idS"
-  prettyTCM (Wk m IdS) = text "wkS" <+> pretty m
-  prettyTCM (EmptyS _) = text "emptyS"
+  prettyTCM IdS        = "idS"
+  prettyTCM (Wk m IdS) = "wkS" <+> pretty m
+  prettyTCM (EmptyS _) = "emptyS"
   prettyTCM rho = prettyTCM u <+> comma <+> prettyTCM rho1
     where
       (rho1, rho2) = splitS 1 rho
@@ -201,12 +202,12 @@ instance (Pretty a, PrettyTCM a, Subst a a) => PrettyTCM (Substitution' a) where
 
 instance PrettyTCM Clause where
   prettyTCM cl = do
-    x <- qualify_ <$> freshName_ "<unnamedclause>"
+    x <- qualify_ <$> freshName_ ("<unnamedclause>" :: String)
     prettyTCM (QNamed x cl)
 
 instance PrettyTCM a => PrettyTCM (Judgement a) where
-  prettyTCM (HasType a t) = prettyTCM a <+> text ":" <+> prettyTCM t
-  prettyTCM (IsSort  a t) = text "Sort" <+> prettyTCM a <+> text ":" <+> prettyTCM t
+  prettyTCM (HasType a t) = prettyTCM a <+> ":" <+> prettyTCM t
+  prettyTCM (IsSort  a t) = "Sort" <+> prettyTCM a <+> ":" <+> prettyTCM t
 
 instance PrettyTCM MetaId where
   prettyTCM x = do
@@ -214,7 +215,7 @@ instance PrettyTCM MetaId where
     pretty $ NamedMeta mn x
 
 instance PrettyTCM a => PrettyTCM (Blocked a) where
-  prettyTCM (Blocked x a) = text "[" <+> prettyTCM a <+> text "]" <> text (P.prettyShow x)
+  prettyTCM (Blocked x a) = "[" <+> prettyTCM a <+> "]" <> text (P.prettyShow x)
   prettyTCM (NotBlocked _ x) = prettyTCM x
 
 instance (Reify a e, ToConcrete e c, P.Pretty c) => PrettyTCM (Named_ a) where
@@ -227,16 +228,17 @@ instance (Reify a e, ToConcrete e c, P.Pretty c) => PrettyTCM (Dom a) where
   prettyTCM x = prettyA =<< reify x
 
 instance (PrettyTCM k, PrettyTCM v) => PrettyTCM (Map k v) where
-  prettyTCM m = text "Map" <> braces (sep $ punctuate comma
-    [ hang (prettyTCM k <+> text "=") 2 (prettyTCM v) | (k, v) <- Map.toList m ])
+  prettyTCM m = "Map" <> braces (sep $ punctuate comma
+    [ hang (prettyTCM k <+> "=") 2 (prettyTCM v) | (k, v) <- Map.toList m ])
 
 instance {-# OVERLAPPING #-} PrettyTCM ArgName where
   prettyTCM = text . P.prettyShow
 
 -- instance (Reify a e, ToConcrete e c, P.Pretty c, PrettyTCM a) => PrettyTCM (Elim' a) where
 instance PrettyTCM Elim where
-  prettyTCM (Apply v) = text "$" <+> prettyTCM v
-  prettyTCM (Proj _ f)= text "." <> prettyTCM f
+  prettyTCM (IApply x y v) = "I$" <+> prettyTCM v
+  prettyTCM (Apply v) = "$" <+> prettyTCM v
+  prettyTCM (Proj _ f)= "." <> prettyTCM f
 
 instance PrettyTCM a => PrettyTCM (MaybeReduced a) where
   prettyTCM = prettyTCM . ignoreReduced
@@ -251,8 +253,8 @@ instance PrettyTCM A.TypedBinding where
   prettyTCM = prettyA
 
 instance PrettyTCM Relevance where
-  prettyTCM Irrelevant = text "."
-  prettyTCM NonStrict  = text ".."
+  prettyTCM Irrelevant = "."
+  prettyTCM NonStrict  = ".."
   prettyTCM Relevant   = empty
 
 instance PrettyTCM ProblemConstraint where
@@ -262,22 +264,26 @@ instance PrettyTCM ProblemConstraint where
 
 instance PrettyTCM Constraint where
     prettyTCM c = case c of
-        ValueCmp cmp ty s t      -> prettyCmp (prettyTCM cmp) s t <?> (text ":" <+> prettyTCMCtx TopCtx ty)
-        ElimCmp cmps fs t v us vs -> prettyCmp (text "~~") us vs   <?> (text ":" <+> prettyTCMCtx TopCtx t)
+        ValueCmp cmp ty s t      -> prettyCmp (prettyTCM cmp) s t <?> (":" <+> prettyTCMCtx TopCtx ty)
+        ValueCmpOnFace cmp p ty s t ->
+            sep [ prettyTCM p <+> "|"
+                , prettyCmp (prettyTCM cmp) s t ]
+            <?> (":" <+> prettyTCMCtx TopCtx ty)
+        ElimCmp cmps fs t v us vs -> prettyCmp "~~" us vs   <?> (":" <+> prettyTCMCtx TopCtx t)
         LevelCmp cmp a b         -> prettyCmp (prettyTCM cmp) a b
         TypeCmp cmp a b          -> prettyCmp (prettyTCM cmp) a b
         TelCmp a b cmp tela telb -> prettyCmp (prettyTCM cmp) tela telb
         SortCmp cmp s1 s2        -> prettyCmp (prettyTCM cmp) s1 s2
-        Guarded c pid            -> prettyTCM c <?> (brackets $ text "blocked on problem" <+> prettyTCM pid)
+        Guarded c pid            -> prettyTCM c <?> (brackets $ "blocked on problem" <+> prettyTCM pid)
         UnBlock m   -> do
             -- BlockedConst t <- mvInstantiation <$> lookupMeta m
             mi <- mvInstantiation <$> lookupMeta m
             case mi of
-              BlockedConst t -> prettyCmp (text ":=") m t
+              BlockedConst t -> prettyCmp ":=" m t
               PostponedTypeCheckingProblem cl _ -> enterClosure cl $ \p ->
-                prettyCmp (text ":=") m p
+                prettyCmp ":=" m p
               Open{}  -> __IMPOSSIBLE__
-              OpenIFS{}  -> __IMPOSSIBLE__
+              OpenInstance{} -> __IMPOSSIBLE__
               InstV{} -> empty
               -- Andreas, 2017-01-11, issue #2637:
               -- The size solver instantiates some metas with infinity
@@ -290,36 +296,39 @@ instance PrettyTCM Constraint where
               --     , show m ++ show args ++ " := " ++ show t
               --     ]
               --   __IMPOSSIBLE__
-        FindInScope m mb mcands -> do
+        FindInstance m mb mcands -> do
             t <- getMetaType m
-            sep [ text "Resolve instance argument" <+> blk
-                    <?> prettyCmp (text ":") m t
+            sep [ "Resolve instance argument" <+> blk
+                    <?> prettyCmp ":" m t
                 , cands
                 ]
           where
             blk = case mb of
                     Nothing -> empty
-                    Just b  -> parens $ text "blocked on" <+> pretty b
+                    Just b  -> parens $ "blocked on" <+> pretty b
             cands =
               case mcands of
-                Nothing -> text "No candidates yet"
+                Nothing -> "No candidates yet"
                 Just cnds ->
-                  hang (text "Candidates") 2 $
-                    vcat [ hang (overlap c <+> prettyTCM (candidateTerm c) <+> text ":") 2 $
+                  hang "Candidates" 2 $
+                    vcat [ hang (overlap c <+> prettyTCM (candidateTerm c) <+> ":") 2 $
                             prettyTCM (candidateType c) | c <- cnds ]
-              where overlap c | candidateOverlappable c = text "overlap"
+              where overlap c | candidateOverlappable c = "overlap"
                               | otherwise               = empty
         IsEmpty r t ->
-            text "Is empty:" <?> prettyTCMCtx TopCtx t
+            "Is empty:" <?> prettyTCMCtx TopCtx t
         CheckSizeLtSat t ->
-            text "Is not empty type of sizes:" <?> prettyTCMCtx TopCtx t
+            "Is not empty type of sizes:" <?> prettyTCMCtx TopCtx t
         CheckFunDef d i q cs -> do
             t <- defType <$> getConstInfo q
-            prettyTCM q <+> text ":" <+> prettyTCM t
-        HasBiggerSort a -> text "Has bigger sort:" <+> prettyTCM a
-        HasPTSRule a b -> text "Has PTS rule:" <+> case b of
+            "Check definition of" <+> prettyTCM q <+> ":" <+> prettyTCM t
+        HasBiggerSort a -> "Has bigger sort:" <+> prettyTCM a
+        HasPTSRule a b -> "Has PTS rule:" <+> case b of
           NoAbs _ b -> prettyTCM (a,b)
-          Abs x b   -> text "(" <> prettyTCM a <+> text "," <+> addContext x (prettyTCM b) <> text ")"
+          Abs x b   -> "(" <> prettyTCM a <+> "," <+> addContext x (prettyTCM b) <> ")"
+        UnquoteTactic _ v _ _ -> do
+          e <- reify v
+          prettyTCM (A.App A.defaultAppInfo_ (A.Unquote A.exprNoRange) (defaultNamedArg e))
 
       where
         prettyCmp :: (PrettyTCM a, PrettyTCM b) => TCM Doc -> a -> b -> TCM Doc
@@ -327,28 +336,29 @@ instance PrettyTCM Constraint where
 
 
 instance PrettyTCM TypeCheckingProblem where
-  prettyTCM (CheckExpr e a) =
-    sep [ prettyA e <+> text ":?", prettyTCM a ]
+  prettyTCM (CheckExpr cmp e a) =
+    sep [ prettyA e <+> ":?", prettyTCM a ]
   prettyTCM (CheckArgs _ _ es t0 t1 _) =
-    sep [ parens $ text "_ :" <+> prettyTCM t0
+    sep [ parens $ "_ :" <+> prettyTCM t0
         , nest 2 $ prettyList $ map prettyA es
-        , nest 2 $ text ":?" <+> prettyTCM t1 ]
-  prettyTCM (CheckLambda (Arg ai (xs, mt)) e t) =
+        , nest 2 $ ":?" <+> prettyTCM t1 ]
+  prettyTCM (CheckProjAppToKnownPrincipalArg cmp e _ _ _ t _ _ _) = prettyTCM (CheckExpr cmp e t)
+  prettyTCM (CheckLambda cmp (Arg ai (xs, mt)) e t) =
     sep [ return CP.lambda <+>
           (CP.prettyRelevance ai .
            CP.prettyHiding ai (if isNothing mt && length xs == 1 then id
                                else P.parens) <$> do
             fsep $
               map prettyTCM xs ++
-              caseMaybe mt [] (\ a -> [text ":", prettyTCM a])) <+>
+              caseMaybe mt [] (\ a -> [":", prettyTCM a])) <+>
           return CP.arrow <+>
           prettyTCM e <+>
-          text ":?"
+          ":?"
         , prettyTCM t
         ]
-  prettyTCM (UnquoteTactic v _ _) = do
+  prettyTCM (DoQuoteTerm _ v _) = do
     e <- reify v
-    prettyTCM (A.App A.defaultAppInfo_ (A.Unquote A.exprNoRange) (defaultNamedArg e))
+    prettyTCM (A.App A.defaultAppInfo_ (A.QuoteTerm A.exprNoRange) (defaultNamedArg e))
 
 instance PrettyTCM a => PrettyTCM (WithHiding a) where
   prettyTCM (WithHiding h a) = CP.prettyHiding h id <$> prettyTCM a
@@ -368,7 +378,7 @@ instance PrettyTCM ConHead where
 instance PrettyTCM Telescope where
   prettyTCM tel = P.fsep . map P.pretty <$> (do
       tel <- reify tel
-      runAbsToCon $ bindToConcrete tel (return . concat)
+      runAbsToCon $ bindToConcrete tel return
     )
 
 newtype PrettyContext = PrettyContext Context
@@ -380,21 +390,24 @@ instance PrettyTCM DBPatVar where
   prettyTCM = prettyTCM . var . dbPatVarIndex
 
 instance PrettyTCM a => PrettyTCM (Pattern' a) where
+  prettyTCM (IApplyP _ _ _ x)    = prettyTCM x
   prettyTCM (VarP _ x)    = prettyTCM x
-  prettyTCM (DotP _ t)    = text ".(" <> prettyTCM t <> text ")"
+  prettyTCM (DotP _ t)    = ".(" <> prettyTCM t <> ")"
+  prettyTCM (DefP o q ps) = parens $
+        prettyTCM q <+> fsep (map (prettyTCM . namedArg) ps)
   prettyTCM (ConP c i ps) = (if b then braces else parens) $ prTy $
         prettyTCM c <+> fsep (map (prettyTCM . namedArg) ps)
         where
         b = maybe False (/= PatOCon) $ conPRecord i
         showRec :: TCM Doc
         showRec = sep
-          [ text "record"
+          [ "record"
           , bracesAndSemicolons <$> zipWithM showField (conFields c) ps
           ]
-        showField x p =
-          sep [ prettyTCM (A.qnameName x) <+> text "=" , nest 2 $ prettyTCM $ namedArg p ]
+        showField (Arg ai x) p =
+          sep [ prettyTCM (A.qnameName x) <+> "=" , nest 2 $ prettyTCM $ namedArg p ]
         showCon = parens $ prTy $ prettyTCM c <+> fsep (map (prettyTCM . namedArg) ps)
-        prTy d = d -- caseMaybe (conPType i) d $ \ t -> d  <+> text ":" <+> prettyTCM t
+        prTy d = d -- caseMaybe (conPType i) d $ \ t -> d  <+> ":" <+> prettyTCM t
   prettyTCM (LitP l)      = text (P.prettyShow l)
   prettyTCM (ProjP _ q)   = text ("." ++ P.prettyShow q)
 
@@ -406,8 +419,9 @@ prettyTCMPatternList :: [NamedArg DeBruijnPattern] -> TCM Doc
 prettyTCMPatternList = prettyList . map prettyA <=< reifyPatterns
 
 instance PrettyTCM (Elim' DisplayTerm) where
-  prettyTCM (Apply v) = text "$" <+> prettyTCM (unArg v)
-  prettyTCM (Proj _ f)= text "." <> prettyTCM f
+  prettyTCM (IApply x y v) = "$" <+> prettyTCM v
+  prettyTCM (Apply v) = "$" <+> prettyTCM (unArg v)
+  prettyTCM (Proj _ f)= "." <> prettyTCM f
 
 instance PrettyTCM NLPat where
   prettyTCM (PVar x bvs) = prettyTCM (Var x (map (Apply . fmap var) bvs))
@@ -418,19 +432,20 @@ instance PrettyTCM NLPat where
     text ("λ " ++ absName u ++ " →") <+>
     (addContext (absName u) $ prettyTCM $ absBody u)
   prettyTCM (PPi a b)   = parens $
-    text ("(" ++ absName b ++ " :") <+> prettyTCM (unDom a) <> text ") →" <+>
+    text ("(" ++ absName b ++ " :") <+> prettyTCM (unDom a) <> ") →" <+>
     (addContext (absName b) $ prettyTCM $ unAbs b)
   prettyTCM (PBoundVar i []) = prettyTCM (var i)
   prettyTCM (PBoundVar i es) = parens $ prettyTCM (var i) <+> fsep (map prettyTCM es)
-  prettyTCM (PTerm t)   = text "." <> parens (prettyTCM t)
+  prettyTCM (PTerm t)   = "." <> parens (prettyTCM t)
 
 instance PrettyTCM NLPType where
   prettyTCM (NLPType PWild a) = prettyTCM a
-  prettyTCM (NLPType l     a) = text "{" <> prettyTCM l <> text "}" <> prettyTCM a
+  prettyTCM (NLPType l     a) = "{" <> prettyTCM l <> "}" <> prettyTCM a
 
 instance PrettyTCM (Elim' NLPat) where
+  prettyTCM (IApply x y v) = prettyTCM v
   prettyTCM (Apply v) = prettyTCM (unArg v)
-  prettyTCM (Proj _ f)= text "." <> prettyTCM f
+  prettyTCM (Proj _ f)= "." <> prettyTCM f
 
 instance PrettyTCM (Type' NLPat) where
   prettyTCM = prettyTCM . unEl
@@ -438,12 +453,12 @@ instance PrettyTCM (Type' NLPat) where
 instance PrettyTCM RewriteRule where
   prettyTCM (RewriteRule q gamma f ps rhs b) = fsep
     [ prettyTCM q
-    , prettyTCM gamma <+> text " |- "
+    , prettyTCM gamma <+> " |- "
     , addContext gamma $ sep
       [ prettyTCM (PDef f ps)
-      , text " --> "
+      , " --> "
       , prettyTCM rhs
-      , text " : "
+      , " : "
       , prettyTCM b
       ]
     ]
@@ -464,3 +479,8 @@ instance (PrettyTCM n, PrettyTCM (WithNode n e)) => PrettyTCM (Graph n e) where
         [ prettyTCM n
         , nest 2 $ vcat $ map (prettyTCM . uncurry WithNode) $ Map.assocs es
         ]
+
+instance PrettyTCM SplitTag where
+  prettyTCM (SplitCon c)  = prettyTCM c
+  prettyTCM (SplitLit l)  = prettyTCM l
+  prettyTCM SplitCatchall = return underscore
