@@ -124,6 +124,9 @@ projView v = do
           Apply a : es -> return $ ProjectionView f a es
           -- Since a projection is a function, it cannot be projected itself.
           Proj{}  : _  -> __IMPOSSIBLE__
+          -- The principal argument of a projection-like cannot be the interval?
+          IApply{} : _ -> __IMPOSSIBLE__
+
     _ -> fallback
 
 -- | Reduce away top-level projection like functions.
@@ -155,11 +158,11 @@ reduceProjectionLike v = do
 --   on (applications of) projection-like functions.
 elimView :: Bool -> Term -> TCM Term
 elimView loneProjToLambda v = do
-  reportSDoc "tc.conv.elim" 30 $ text "elimView of " <+> prettyTCM v
+  reportSDoc "tc.conv.elim" 30 $ "elimView of " <+> prettyTCM v
   reportSLn  "tc.conv.elim" 50 $ "v = " ++ show v
   v <- reduceProjectionLike v
   reportSDoc "tc.conv.elim" 40 $
-    text "elimView (projections reduced) of " <+> prettyTCM v
+    "elimView (projections reduced) of " <+> prettyTCM v
   pv <- projView v
   case pv of
     NoProjection{}        -> return v
@@ -170,13 +173,15 @@ elimView loneProjToLambda v = do
 
 -- | Which @Def@types are eligible for the principle argument
 --   of a projection-like function?
-eligibleForProjectionLike :: QName -> TCM Bool
+eligibleForProjectionLike :: (HasConstInfo m) => QName -> m Bool
 eligibleForProjectionLike d = eligible . theDef <$> getConstInfo d
   where
   eligible = \case
     Datatype{} -> True
     Record{}   -> True
     Axiom{}    -> True
+    DataOrRecSig{}     -> True
+    GeneralizableVar{} -> False
     Function{}    -> False
     Primitive{}   -> False
     Constructor{} -> __IMPOSSIBLE__
@@ -227,8 +232,8 @@ makeProjection x = -- if True then return () else do
   defn <- getConstInfo x
   let t = defType defn
   reportSDoc "tc.proj.like" 20 $ sep
-    [ text "Checking for projection likeness "
-    , prettyTCM x <+> text " : " <+> prettyTCM t
+    [ "Checking for projection likeness "
+    , prettyTCM x <+> " : " <+> prettyTCM t
     ]
   case theDef defn of
     Function{funClauses = cls}
@@ -239,7 +244,8 @@ makeProjection x = -- if True then return () else do
     -- the dropped arguments.
     -- Nor can abstract definitions be projection-like since they won't reduce
     -- outside the abstract block.
-    def@Function{funProjection = Nothing, funClauses = cls, funCompiled = cc0, funInv = NotInjective,
+    def@Function{funProjection = Nothing, funClauses = cls,
+                 funSplitTree = st0, funCompiled = cc0, funInv = NotInjective,
                  funMutual = Just [], -- Andreas, 2012-09-28: only consider non-mutual funs (or those whose recursion status has not yet been determined)
                  funAbstr = ConcreteDef} -> do
       ps0 <- filterM validProj $ candidateArgs [] t
@@ -252,18 +258,19 @@ makeProjection x = -- if True then return () else do
           {- else -}
           case lastMaybe (filter (checkOccurs cls . snd) ps0) of
             Nothing -> reportSDoc "tc.proj.like" 50 $ nest 2 $ vcat
-              [ text "occurs check failed"
-              , nest 2 $ text "clauses =" <?> vcat (map pretty cls) ]
+              [ "occurs check failed"
+              , nest 2 $ "clauses =" <?> vcat (map pretty cls) ]
             Just (d, n) -> do
               -- Yes, we are projection-like!
               reportSDoc "tc.proj.like" 10 $ sep
-                [ prettyTCM x <+> text " : " <+> prettyTCM t
+                [ prettyTCM x <+> " : " <+> prettyTCM t
                 , text $ " is projection like in argument " ++ show n ++ " for type " ++ show d
                 ]
               __CRASH_WHEN__ "tc.proj.like.crash" 1000
 
               let cls' = map (dropArgs n) cls
                   cc   = dropArgs n cc0
+                  st   = dropArgs n st0
               reportSLn "tc.proj.like" 60 $ "  rewrote clauses to\n    " ++ show cc
 
               -- Andreas, 2013-10-20 build parameter dropping function
@@ -275,11 +282,12 @@ makeProjection x = -- if True then return () else do
                     , projOrig     = x
                     , projFromType = d
                     , projIndex    = pIndex
-                    , projLams     = ProjLams $ map (\ (Dom ai (y, _)) -> Arg ai y) tel
+                    , projLams     = ProjLams $ map (argFromDom . fmap fst) tel
                     }
               let newDef = def
                            { funProjection     = Just projection
                            , funClauses        = cls'
+                           , funSplitTree      = st
                            , funCompiled       = cc
                            , funInv            = dropArgs n $ funInv def
                            }
@@ -299,6 +307,8 @@ makeProjection x = -- if True then return () else do
     Function{funMutual = Nothing} ->
       reportSLn "tc.proj.like" 30 $ "  mutuality check has not run yet"
     Axiom          -> reportSLn "tc.proj.like" 30 $ "  not a function, but Axiom"
+    DataOrRecSig{} -> reportSLn "tc.proj.like" 30 $ "  not a function, but DataOrRecSig"
+    GeneralizableVar{} -> reportSLn "tc.proj.like" 30 $ "  not a function, but GeneralizableVar"
     AbstractDefn{} -> reportSLn "tc.proj.like" 30 $ "  not a function, but AbstractDefn"
     Constructor{}  -> reportSLn "tc.proj.like" 30 $ "  not a function, but Constructor"
     Datatype{}     -> reportSLn "tc.proj.like" 30 $ "  not a function, but Datatype"
@@ -316,10 +326,9 @@ makeProjection x = -- if True then return () else do
     -- and/or positivity checkers.
     recursive = do
       occs <- computeOccurrences x
-      let xocc = Map.lookup (ADef x) occs
-      case xocc of
-        Just (_ : _) -> return True -- recursive occurrence
-        _            -> return False
+      case Map.lookup (ADef x) occs of
+        Just n | n >= 1 -> return True -- recursive occurrence
+        _               -> return False
 
     checkOccurs cls n = all (nonOccur n) cls
 
@@ -345,10 +354,12 @@ makeProjection x = -- if True then return () else do
         shallowMatch _             = True
         noMatches = all (noMatch . namedArg)
         noMatch ConP{} = False
+        noMatch DefP{} = False
         noMatch LitP{} = False
         noMatch ProjP{}= False
         noMatch VarP{} = True
         noMatch DotP{} = True
+        noMatch IApplyP{} = True
 
     -- Make sure non of the parameters occurs in the body of the function.
     checkBody m n b = not . getAny $ runFree badVar IgnoreNot b
