@@ -36,7 +36,6 @@ import Agda.Utils.Except
 import Agda.Utils.Impossible
 import Agda.Utils.Monad ( ifM )
 import Agda.Utils.Permutation ( Permutation(Perm), compactP, reverseP )
-import Agda.Utils.String ( Str(Str), unStr )
 import Agda.Utils.VarSet (VarSet)
 import qualified Agda.Utils.VarSet as Set
 import Agda.Utils.FileName
@@ -55,7 +54,7 @@ data QuotingKit = QuotingKit
 
 quotingKit :: TCM QuotingKit
 quotingKit = do
-  currentFile     <- fromMaybe __IMPOSSIBLE__ <$> asks envCurrentPath
+  currentFile     <- fromMaybe __IMPOSSIBLE__ <$> asksTC envCurrentPath
   hidden          <- primHidden
   instanceH       <- primInstance
   visible         <- primVisible
@@ -149,12 +148,14 @@ quotingKit = do
 
       quoteSort :: Sort -> ReduceM Term
       quoteSort (Type t) = quoteSortLevelTerm t
-      quoteSort Prop     = pure unsupportedSort
+      quoteSort Prop{}   = pure unsupportedSort
       quoteSort Inf      = pure unsupportedSort
       quoteSort SizeUniv = pure unsupportedSort
       quoteSort PiSort{} = pure unsupportedSort
       quoteSort UnivSort{} = pure unsupportedSort
       quoteSort (MetaS x es) = quoteTerm $ MetaV x es
+      quoteSort (DefS d es) = quoteTerm $ Def d es
+      quoteSort (DummyS s) =__IMPOSSIBLE_VERBOSE__ s
 
       quoteType :: Type -> ReduceM Term
       quoteType (El _ t) = quoteTerm t
@@ -173,6 +174,8 @@ quotingKit = do
       quotePat (ConP c _ ps)     = conP !@ quoteQName (conName c) @@ quotePats ps
       quotePat (LitP l)          = litP !@ quoteLit l
       quotePat (ProjP _ x)       = projP !@ quoteQName x
+      quotePat (IApplyP o t u x)  = pure unsupported
+      quotePat DefP{}             = pure unsupported
 
       quoteClause :: Clause -> ReduceM Term
       quoteClause cl@Clause{namedClausePats = ps, clauseBody = body} =
@@ -191,7 +194,7 @@ quotingKit = do
       quoteList q xs = list (map q xs)
 
       quoteDom :: (Type -> ReduceM Term) -> Dom Type -> ReduceM Term
-      quoteDom q (Dom info t) = arg !@ quoteArgInfo info @@ q t
+      quoteDom q (Dom{domInfo = info, unDom = t}) = arg !@ quoteArgInfo info @@ q t
 
       quoteAbs :: Subst t a => (a -> ReduceM Term) -> Abs a -> ReduceM Term
       quoteAbs q (Abs s t)   = abs !@! quoteString s @@ q t
@@ -212,20 +215,27 @@ quotingKit = do
           Lam info t -> lam !@ quoteHiding (getHiding info) @@ quoteAbs quoteTerm t
           Def x es   -> do
             defn <- getConstInfo x
-            n <- getDefFreeVars x
             -- #2220: remember to restore dropped parameters
             let
               conOrProjPars = defParameters defn
               ts = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
-              qx Function{ funExtLam = Just (ExtLamInfo m), funClauses = cs } = do
+              qx Function{ funExtLam = Just (ExtLamInfo m _), funClauses = cs } = do
                     -- An extended lambda should not have any extra parameters!
                     unless (null conOrProjPars) __IMPOSSIBLE__
                     n <- size <$> lookupSection m
-                    extlam !@ list (map (quoteClause . (`apply` (take n ts))) cs)
-              qx Function{ funCompiled = Just Fail, funClauses = [cl] } =
-                    extlam !@ list [quoteClause $ dropArgs (length (namedClausePats cl) - 1) cl]
-              qx _ = def !@! quoteName x
-            qx (theDef defn) @@ list (drop n $ conOrProjPars ++ map (quoteArg quoteTerm) ts)
+                    let (pars, args) = splitAt n ts
+                    extlam !@ list (map (quoteClause . (`apply` pars)) cs)
+                           @@ list (map (quoteArg quoteTerm) args)
+              qx df@Function{ funCompiled = Just Fail, funClauses = [cl] } = do
+                    -- See also corresponding code in InternalToAbstract
+                    let n = length (namedClausePats cl) - 1
+                    extlam !@ list [quoteClause $ dropArgs n cl]
+                           @@ list (drop n $ map (quoteArg quoteTerm) ts)
+              qx _ = do
+                n <- getDefFreeVars x
+                def !@! quoteName x
+                     @@ list (drop n $ conOrProjPars ++ map (quoteArg quoteTerm) ts)
+            qx (theDef defn)
           Con x ci es | Just ts <- allApplyElims es -> do
             cDef <- getConstInfo (conName x)
             n    <- getDefFreeVars (conName x)
@@ -240,6 +250,7 @@ quotingKit = do
           MetaV x es -> meta !@! quoteMeta currentFile x @@ quoteArgs vs
             where vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
           DontCare{} -> pure unsupported -- could be exposed at some point but we have to take care
+          Dummy s    -> __IMPOSSIBLE_VERBOSE__ s
 
       defParameters :: Definition -> [ReduceM Term]
       defParameters def = map par hiding
@@ -264,6 +275,8 @@ quotingKit = do
               !@! quoteName (conName c)
               @@ quoteList (quoteArg (pure . quoteName)) fs
           Axiom{}       -> pure agdaDefinitionPostulate
+          DataOrRecSig{} -> pure agdaDefinitionPostulate
+          GeneralizableVar{} -> pure agdaDefinitionPostulate  -- TODO: reflect generalizable vars
           AbstractDefn{}-> pure agdaDefinitionPostulate
           Primitive{primClauses = cs} | not $ null cs ->
             agdaDefinitionFunDef !@ quoteList quoteClause cs
