@@ -11,6 +11,7 @@ import qualified Data.List as List
 import Data.List ((\\))
 import Data.Maybe
 import Data.Map (Map)
+import Data.Monoid
 import Data.Traversable
 import Data.Hashable
 
@@ -43,6 +44,7 @@ import {-# SOURCE #-} Agda.TypeChecking.Reduce.Fast
 import Agda.Utils.Function
 import Agda.Utils.Functor
 import Agda.Utils.Lens
+import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.HashMap (HashMap)
 import Agda.Utils.Size
@@ -51,30 +53,35 @@ import Agda.Utils.Tuple
 #include "undefined.h"
 import Agda.Utils.Impossible
 
-instantiate :: Instantiate a => a -> TCM a
-instantiate = runReduceM . instantiate'
+instantiate :: (Instantiate a, MonadReduce m) => a -> m a
+instantiate = liftReduce . instantiate'
 
-instantiateFull :: InstantiateFull a => a -> TCM a
-instantiateFull = runReduceM . instantiateFull'
+instantiateFull :: (InstantiateFull a, MonadReduce m) => a -> m a
+instantiateFull = liftReduce . instantiateFull'
 
-reduce :: Reduce a => a -> TCM a
-reduce = runReduceM . reduce'
+reduce :: (Reduce a, MonadReduce m) => a -> m a
+reduce = liftReduce . reduce'
 
-reduceB :: Reduce a => a -> TCM (Blocked a)
-reduceB = runReduceM . reduceB'
+reduceB :: (Reduce a, MonadReduce m) => a -> m (Blocked a)
+reduceB = liftReduce . reduceB'
 
-normalise :: Normalise a => a -> TCM a
-normalise = runReduceM . normalise'
+normalise :: (Normalise a, MonadReduce m) => a -> m a
+normalise = liftReduce . normalise'
 
-simplify :: Simplify a => a -> TCM a
-simplify = runReduceM . simplify'
+-- | Normalise the given term but also preserve blocking tags
+--   TODO: implement a more efficient version of this.
+normaliseB :: (MonadReduce m, Reduce t, Normalise t) => t -> m (Blocked t)
+normaliseB = normalise >=> reduceB
+
+simplify :: (Simplify a, MonadReduce m) => a -> m a
+simplify = liftReduce . simplify'
 
 -- | Meaning no metas left in the instantiation.
 isFullyInstantiatedMeta :: MetaId -> TCM Bool
 isFullyInstantiatedMeta m = do
   mv <- TCM.lookupMeta m
   case mvInstantiation mv of
-    InstV _tel v -> null . allMetas <$> instantiateFull v
+    InstV _tel v -> noMetas <$> instantiateFull v
     _ -> return False
 
 -- | Instantiate something.
@@ -102,7 +109,7 @@ instance Instantiate Term where
           -- specification: inst == foldr mkLam v tel `applyE` es
           inst = applySubst rho (foldr mkLam v $ drop (length es1) tel) `applyE` es2
       Open                             -> return t
-      OpenIFS                          -> return t
+      OpenInstance                     -> return t
       BlockedConst _                   -> return t
       PostponedTypeCheckingProblem _ _ -> return t
   instantiate' (Level l) = levelTm <$> instantiate' l
@@ -133,7 +140,7 @@ instance Instantiate a => Instantiate (Blocked a) where
     case mi of
       InstV{}                        -> notBlocked <$> instantiate' u
       Open                           -> return v
-      OpenIFS                        -> return v
+      OpenInstance                   -> return v
       BlockedConst{}                 -> return v
       PostponedTypeCheckingProblem{} -> return v
 
@@ -145,12 +152,14 @@ instance Instantiate Sort where
     MetaS x es -> instantiate' (MetaV x es) >>= \case
       Sort s'      -> return s'
       MetaV x' es' -> return $ MetaS x' es'
+      Def d es'     -> return $ DefS d es'
       _            -> __IMPOSSIBLE__
     _ -> return s
 
 instance Instantiate Elim where
   instantiate' (Apply v) = Apply <$> instantiate' v
   instantiate' (Proj o f)= pure $ Proj o f
+  instantiate' (IApply x y v) = IApply <$> instantiate' x <*> instantiate' y <*> instantiate' v
 
 instance Instantiate t => Instantiate (Abs t) where
   instantiate' = traverse instantiate'
@@ -161,6 +170,9 @@ instance Instantiate t => Instantiate (Arg t) where
 instance Instantiate t => Instantiate (Dom t) where
     instantiate' = traverse instantiate'
 
+instance Instantiate t => Instantiate (Maybe t) where
+  instantiate' = traverse instantiate'
+
 instance Instantiate t => Instantiate [t] where
     instantiate' = traverse instantiate'
 
@@ -185,6 +197,9 @@ instance Instantiate Constraint where
   instantiate' (ValueCmp cmp t u v) = do
     (t,u,v) <- instantiate' (t,u,v)
     return $ ValueCmp cmp t u v
+  instantiate' (ValueCmpOnFace cmp p t u v) = do
+    ((p,t),u,v) <- instantiate' ((p,t),u,v)
+    return $ ValueCmpOnFace cmp p t u v
   instantiate' (ElimCmp cmp fs t v as bs) =
     ElimCmp cmp fs <$> instantiate' t <*> instantiate' v <*> instantiate' as <*> instantiate' bs
   instantiate' (LevelCmp cmp u v)   = uncurry (LevelCmp cmp) <$> instantiate' (u,v)
@@ -193,18 +208,19 @@ instance Instantiate Constraint where
   instantiate' (SortCmp cmp a b)    = uncurry (SortCmp cmp) <$> instantiate' (a,b)
   instantiate' (Guarded c pid)      = Guarded <$> instantiate' c <*> pure pid
   instantiate' (UnBlock m)          = return $ UnBlock m
-  instantiate' (FindInScope m b args) = FindInScope m b <$> mapM instantiate' args
+  instantiate' (FindInstance m b args) = FindInstance m b <$> mapM instantiate' args
   instantiate' (IsEmpty r t)        = IsEmpty r <$> instantiate' t
   instantiate' (CheckSizeLtSat t)   = CheckSizeLtSat <$> instantiate' t
   instantiate' c@CheckFunDef{}      = return c
   instantiate' (HasBiggerSort a)    = HasBiggerSort <$> instantiate' a
   instantiate' (HasPTSRule a b)     = uncurry HasPTSRule <$> instantiate' (a,b)
+  instantiate' (UnquoteTactic m t h g) = UnquoteTactic m <$> instantiate' t <*> instantiate' h <*> instantiate' g
 
 instance Instantiate e => Instantiate (Map k e) where
     instantiate' = traverse instantiate'
 
 instance Instantiate Candidate where
-  instantiate' (Candidate u t eti ov) = Candidate <$> instantiate' u <*> instantiate' t <*> pure eti <*> pure ov
+  instantiate' (Candidate u t ov) = Candidate <$> instantiate' u <*> instantiate' t <*> pure ov
 
 instance Instantiate EqualityView where
   instantiate' (OtherType t)            = OtherType
@@ -223,23 +239,23 @@ instance Instantiate EqualityView where
 
 -- | Case on whether a term is blocked on a meta (or is a meta).
 --   That means it can change its shape when the meta is instantiated.
-ifBlocked :: MonadTCM tcm =>  Term -> (MetaId -> Term -> tcm a) -> (NotBlocked -> Term -> tcm a) -> tcm a
+ifBlocked :: (MonadReduce m) =>  Term -> (MetaId -> Term -> m a) -> (NotBlocked -> Term -> m a) -> m a
 ifBlocked t blocked unblocked = do
-  t <- liftTCM $ reduceB t
+  t <- reduceB t
   case t of
     Blocked m _              -> blocked m (ignoreBlocking t)
     NotBlocked _ (MetaV m _) -> blocked m (ignoreBlocking t)
     NotBlocked nb _          -> unblocked nb (ignoreBlocking t)
 
-isBlocked :: MonadTCM tcm => Term -> tcm (Maybe MetaId)
+isBlocked :: MonadReduce m => Term -> m (Maybe MetaId)
 isBlocked t = ifBlocked t (\m _ -> return $ Just m) (\_ _ -> return Nothing)
 
 -- | Case on whether a type is blocked on a meta (or is a meta).
-ifBlockedType :: MonadTCM tcm => Type -> (MetaId -> Type -> tcm a) -> (NotBlocked -> Type -> tcm a) -> tcm a
+ifBlockedType :: MonadReduce m => Type -> (MetaId -> Type -> m a) -> (NotBlocked -> Type -> m a) -> m a
 ifBlockedType (El s t) blocked unblocked =
   ifBlocked t (\ m v -> blocked m $ El s v) (\ nb v -> unblocked nb $ El s v)
 
-isBlockedType :: MonadTCM tcm => Type -> tcm (Maybe MetaId)
+isBlockedType :: MonadReduce m => Type -> m (Maybe MetaId)
 isBlockedType t = ifBlockedType t (\m _ -> return $ Just m) (\_ _ -> return Nothing)
 
 class Reduce t where
@@ -254,27 +270,28 @@ instance Reduce Type where
     reduceB' (El s t) = fmap (El s) <$> reduceB' t
 
 instance Reduce Sort where
-    reduce' s = {-# SCC "reduce'<Sort>" #-}
-      ifNotM hasUniversePolymorphism (red s) $ {- else -} red =<< instantiateFull' s
-      where
-        red s = do
-          s <- instantiate' s
-          case s of
-            PiSort s1 s2 -> do
-              (s1,s2) <- reduce' (s1,s2)
-              maybe (return $ PiSort s1 s2) reduce' $ piSort' s1 s2
-            UnivSort s' -> do
-              s' <- reduce' s'
-              maybe (return $ UnivSort s') reduce' $ univSort' s'
-            Prop       -> return s
-            Type s'    -> Type <$> reduce' s'
-            Inf        -> return Inf
-            SizeUniv   -> return SizeUniv
-            MetaS x es -> return s
+    reduce' s = do
+      s <- instantiate' s
+      case s of
+        PiSort s1 s2 -> do
+          (s1,s2) <- reduce' (s1,s2)
+          maybe (return $ PiSort s1 s2) reduce' $ piSort' s1 s2
+        UnivSort s' -> do
+          s' <- reduce' s'
+          ui <- univInf
+          caseMaybe (univSort' ui s') (return $ UnivSort s') reduce'
+        Prop s'    -> Prop <$> reduce' s'
+        Type s'    -> Type <$> reduce' s'
+        Inf        -> return Inf
+        SizeUniv   -> return SizeUniv
+        MetaS x es -> return s
+        DefS d es  -> return s -- postulated sorts do not reduce
+        DummyS{}   -> return s
 
 instance Reduce Elim where
   reduce' (Apply v) = Apply <$> reduce' v
   reduce' (Proj o f)= pure $ Proj o f
+  reduce' (IApply x y v) = IApply <$> reduce' x <*> reduce' y <*> reduce' v
 
 instance Reduce Level where
   reduce'  (Max as) = levelMax <$> mapM reduce' as
@@ -321,19 +338,56 @@ instance Reduce t => Reduce (Dom t) where
     reduce' = traverse reduce'
     reduceB' t = traverse id <$> traverse reduceB' t
 
--- Tuples are never blocked
 instance (Reduce a, Reduce b) => Reduce (a,b) where
     reduce' (x,y)  = (,) <$> reduce' x <*> reduce' y
+    reduceB' (x,y) = do
+      x <- reduceB' x
+      y <- reduceB' y
+      let blk = void x `mappend` void y
+          xy  = (ignoreBlocking x , ignoreBlocking y)
+      return $ blk $> xy
 
 instance (Reduce a, Reduce b,Reduce c) => Reduce (a,b,c) where
     reduce' (x,y,z) = (,,) <$> reduce' x <*> reduce' y <*> reduce' z
+    reduceB' (x,y,z) = do
+      x <- reduceB' x
+      y <- reduceB' y
+      z <- reduceB' z
+      let blk = void x `mappend` void y `mappend` void z
+          xyz = (ignoreBlocking x , ignoreBlocking y , ignoreBlocking z)
+      return $ blk $> xyz
+
+reduceIApply :: ReduceM (Blocked Term) -> [Elim] -> ReduceM (Blocked Term)
+reduceIApply = reduceIApply' reduceB'
+
+blockedOrMeta :: Blocked Term -> Blocked ()
+blockedOrMeta r =
+  case r of
+    Blocked m _              -> Blocked m ()
+    NotBlocked _ (MetaV m _) -> Blocked m ()
+    NotBlocked i _           -> NotBlocked i ()
+
+reduceIApply' :: (Term -> ReduceM (Blocked Term)) -> ReduceM (Blocked Term) -> [Elim] -> ReduceM (Blocked Term)
+reduceIApply' reduceB' d (IApply x y r : es) = do
+  view <- intervalView'
+  r <- reduceB' r
+  -- We need to propagate the blocking information so that e.g.
+  -- we postpone "someNeutralPath ?0 = a" rather than fail.
+  let blockedInfo = blockedOrMeta r
+
+  case view (ignoreBlocking r) of
+   IZero -> reduceB' (applyE x es)
+   IOne  -> reduceB' (applyE y es)
+   _     -> fmap (<* blockedInfo) (reduceIApply d es)
+reduceIApply' reduceB' d (_ : es) = reduceIApply d es
+reduceIApply' reduceB' d [] = d
 
 instance Reduce Term where
   reduceB' = {-# SCC "reduce'<Term>" #-} maybeFastReduceTerm
 
 shouldTryFastReduce :: ReduceM Bool
-shouldTryFastReduce = do
-  allowed <- asks envAllowedReductions
+shouldTryFastReduce = (optFastReduce <$> pragmaOptions) `and2M` do
+  allowed <- asksTC envAllowedReductions
   let optionalReductions = [NonTerminatingReductions, UnconfirmedReductions]
       requiredReductions = allReductions \\ optionalReductions
   return $ (allowed \\ optionalReductions) == requiredReductions
@@ -358,27 +412,30 @@ slowReduceTerm :: Term -> ReduceM (Blocked Term)
 slowReduceTerm v = do
     v <- instantiate' v
     let done = return $ notBlocked v
+        iapp = reduceIApply done
     case v of
 --    Andreas, 2012-11-05 not reducing meta args does not destroy anything
 --    and seems to save 2% sec on the standard library
 --      MetaV x args -> notBlocked . MetaV x <$> reduce' args
-      MetaV x es -> done
-      Def f es   -> unfoldDefinitionE False reduceB' (Def f []) f es
+      MetaV x es -> iapp es
+      Def f es   -> flip reduceIApply es $ unfoldDefinitionE False reduceB' (Def f []) f es
       Con c ci es -> do
-          let args = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
           -- Constructors can reduce' when they come from an
           -- instantiated module.
-          v <- unfoldDefinition False reduceB' (Con c ci []) (conName c) args
+          -- also reduce when they are path constructors
+          v <- flip reduceIApply es
+                 $ unfoldDefinitionE False reduceB' (Con c ci []) (conName c) es
           traverse reduceNat v
       Sort s   -> fmap Sort <$> reduceB' s
-      Level l  -> ifM (elem LevelReductions <$> asks envAllowedReductions)
+      Level l  -> ifM (elem LevelReductions <$> asksTC envAllowedReductions)
                     {- then -} (fmap levelTm <$> reduceB' l)
                     {- else -} done
       Pi _ _   -> done
       Lit _    -> done
-      Var _ _  -> done
+      Var _ es  -> iapp es
       Lam _ _  -> done
       DontCare _ -> done
+      Dummy{}    -> done
     where
       -- NOTE: reduceNat can traverse the entire term.
       reduceNat v@(Con c ci []) = do
@@ -403,6 +460,9 @@ unfoldCorecursionE :: Elim -> ReduceM (Blocked Elim)
 unfoldCorecursionE (Proj o p)           = notBlocked . Proj o <$> getOriginalProjection p
 unfoldCorecursionE (Apply (Arg info v)) = fmap (Apply . Arg info) <$>
   unfoldCorecursion v
+unfoldCorecursionE (IApply x y r) = do -- TODO check if this makes sense
+   [x,y,r] <- mapM unfoldCorecursion [x,y,r]
+   return $ IApply <$> x <*> y <*> r
 
 unfoldCorecursion :: Term -> ReduceM (Blocked Term)
 unfoldCorecursion v = do
@@ -442,7 +502,7 @@ unfoldDefinitionStep unfoldDelayed v0 f es =
   {-# SCC "reduceDef" #-} do
   info <- getConstInfo f
   rewr <- instantiateRewriteRules =<< getRewriteRulesFor f
-  allowed <- asks envAllowedReductions
+  allowed <- asksTC envAllowedReductions
   let def = theDef info
       v   = v0 `applyE` es
       -- Non-terminating functions
@@ -486,11 +546,11 @@ unfoldDefinitionStep unfoldDelayed v0 f es =
       | otherwise = {-# SCC "reducePrimitive" #-} do
           let (es1,es2) = splitAt ar es
               args1     = fromMaybe __IMPOSSIBLE__ $ mapM isApplyElim es1
-          r <- primFunImplementation pf args1
+          r <- primFunImplementation pf args1 (length es2)
           case r of
             NoReduction args1' -> do
               let es1' = map (fmap Apply) args1'
-              if null cls then do
+              if null cls && null rewr then do
                 noReduction $ applyE (Def f []) <$> do
                   traverse id $
                     map mredToBlocked es1' ++ map notBlocked es2
@@ -519,15 +579,15 @@ unfoldDefinitionStep unfoldDelayed v0 f es =
         case ev of
           NoReduction v -> do
             reportSDoc "tc.reduce" 90 $ vcat
-              [ text "*** tried to reduce " <+> prettyTCM f
-              , text "    es =  " <+> sep (map (prettyTCM . ignoreReduced) es)
-              -- , text "*** tried to reduce " <+> prettyTCM vfull
-              , text "    stuck on" <+> prettyTCM (ignoreBlocking v)
+              [ "*** tried to reduce " <+> prettyTCM f
+              , "    es =  " <+> sep (map (prettyTCM . ignoreReduced) es)
+              -- , "*** tried to reduce " <+> prettyTCM vfull
+              , "    stuck on" <+> prettyTCM (ignoreBlocking v)
               ]
           YesReduction _simpl v -> do
-            reportSDoc "tc.reduce"  90 $ text "*** reduced definition: " <+> prettyTCM f
-            reportSDoc "tc.reduce"  95 $ text "    result" <+> prettyTCM v
-            reportSDoc "tc.reduce" 100 $ text "    raw   " <+> text (show v)
+            reportSDoc "tc.reduce"  90 $ "*** reduced definition: " <+> prettyTCM f
+            reportSDoc "tc.reduce"  95 $ "    result" <+> prettyTCM v
+            reportSDoc "tc.reduce" 100 $ "    raw   " <+> text (show v)
 
 -- | Reduce a non-primitive definition if it is a copy linking to another def.
 reduceDefCopy :: QName -> Elims -> TCM (Reduced () Term)
@@ -550,28 +610,26 @@ reduceDefCopy f es = do
             NoReduction{}        -> return $ NoReduction ()
 
 -- | Reduce simple (single clause) definitions.
-reduceHead :: Term -> TCM (Blocked Term)
-reduceHead = runReduceM . reduceHead'
-
-reduceHead' :: Term -> ReduceM (Blocked Term)
-reduceHead' v = do -- ignoreAbstractMode $ do
+reduceHead :: (HasBuiltins m, HasConstInfo m, MonadReduce m, MonadDebug m)
+           => Term -> m (Blocked Term)
+reduceHead v = do -- ignoreAbstractMode $ do
   -- Andreas, 2013-02-18 ignoreAbstractMode leads to information leakage
   -- see Issue 796
 
   -- first, possibly rewrite literal v to constructor form
   v <- constructorForm v
-  traceSDoc "tc.inj.reduce" 30 (text "reduceHead" <+> prettyTCM v) $ do
+  traceSDoc "tc.inj.reduce" 30 ("reduceHead" <+> prettyTCM v) $ do
   case v of
     Def f es -> do
 
-      abstractMode <- envAbstractMode <$> ask
+      abstractMode <- envAbstractMode <$> askTC
       isAbstract <- treatAbstractly f
       traceSLn "tc.inj.reduce" 50 (
         "reduceHead: we are in " ++ show abstractMode++ "; " ++ show f ++
         " is treated " ++ if isAbstract then "abstractly" else "concretely"
         ) $ do
       let v0  = Def f []
-          red = unfoldDefinitionE False reduceHead' v0 f es
+          red = liftReduce $ unfoldDefinitionE False reduceHead v0 f es
       def <- theDef <$> getConstInfo f
       case def of
         -- Andreas, 2012-11-06 unfold aliases (single clause terminating functions)
@@ -588,19 +646,16 @@ reduceHead' v = do -- ignoreAbstractMode $ do
     _ -> return $ notBlocked v
 
 -- | Unfold a single inlined function.
-unfoldInlined :: Term -> TCM Term
-unfoldInlined = runReduceM . unfoldInlined'
-
-unfoldInlined' :: Term -> ReduceM Term
-unfoldInlined' v = do
-  inTypes <- view eWorkingOnTypes
+unfoldInlined :: (HasConstInfo m, MonadReduce m) => Term -> m Term
+unfoldInlined v = do
+  inTypes <- viewTC eWorkingOnTypes
   case v of
     _ | inTypes -> return v -- Don't inline in types (to avoid unfolding of goals)
     Def f es -> do
       def <- theDef <$> getConstInfo f
       case def of   -- Only for simple definitions with no pattern matching (TODO: maybe copatterns?)
         Function{ funCompiled = Just Done{}, funDelayed = NotDelayed }
-          | def ^. funInline ->
+          | def ^. funInline -> liftReduce $
           ignoreBlocking <$> unfoldDefinitionE False (return . notBlocked) (Def f []) f es
         _ -> return v
     _ -> return v
@@ -612,7 +667,7 @@ appDef_ f v0 cls mcc rewr args = appDefE_ f v0 cls mcc rewr $ map (fmap Apply) a
 
 appDefE_ :: QName -> Term -> [Clause] -> Maybe CompiledClauses -> RewriteRules -> MaybeReducedElims -> ReduceM (Reduced (Blocked Term) Term)
 appDefE_ f v0 cls mcc rewr args =
-  local (\ e -> e { envAppDef = Just f }) $
+  localTC (\ e -> e { envAppDef = Just f }) $
   maybe (appDefE' v0 cls rewr args)
         (\cc -> appDefE v0 cc rewr args) mcc
 
@@ -681,6 +736,9 @@ instance Reduce Constraint where
   reduce' (ValueCmp cmp t u v) = do
     (t,u,v) <- reduce' (t,u,v)
     return $ ValueCmp cmp t u v
+  reduce' (ValueCmpOnFace cmp p t u v) = do
+    ((p,t),u,v) <- reduce' ((p,t),u,v)
+    return $ ValueCmpOnFace cmp p t u v
   reduce' (ElimCmp cmp fs t v as bs) =
     ElimCmp cmp fs <$> reduce' t <*> reduce' v <*> reduce' as <*> reduce' bs
   reduce' (LevelCmp cmp u v)    = uncurry (LevelCmp cmp) <$> reduce' (u,v)
@@ -689,18 +747,19 @@ instance Reduce Constraint where
   reduce' (SortCmp cmp a b)     = uncurry (SortCmp cmp) <$> reduce' (a,b)
   reduce' (Guarded c pid)       = Guarded <$> reduce' c <*> pure pid
   reduce' (UnBlock m)           = return $ UnBlock m
-  reduce' (FindInScope m b cands) = FindInScope m b <$> mapM reduce' cands
+  reduce' (FindInstance m b cands) = FindInstance m b <$> mapM reduce' cands
   reduce' (IsEmpty r t)         = IsEmpty r <$> reduce' t
   reduce' (CheckSizeLtSat t)    = CheckSizeLtSat <$> reduce' t
   reduce' c@CheckFunDef{}       = return c
   reduce' (HasBiggerSort a)     = HasBiggerSort <$> reduce' a
   reduce' (HasPTSRule a b)      = uncurry HasPTSRule <$> reduce' (a,b)
+  reduce' (UnquoteTactic m t h g) = UnquoteTactic m <$> reduce' t <*> reduce' h <*> reduce' g
 
 instance Reduce e => Reduce (Map k e) where
     reduce' = traverse reduce'
 
 instance Reduce Candidate where
-  reduce' (Candidate u t eti ov) = Candidate <$> reduce' u <*> reduce' t <*> pure eti <*> pure ov
+  reduce' (Candidate u t ov) = Candidate <$> reduce' u <*> reduce' t <*> pure ov
 
 instance Reduce EqualityView where
   reduce' (OtherType t)            = OtherType
@@ -744,6 +803,7 @@ instance Simplify Term where
       Var i vs   -> Var i    <$> simplify' vs
       Lam h v    -> Lam h    <$> simplify' v
       DontCare v -> dontCare <$> simplify' v
+      Dummy{}    -> return v
 
 simplifyBlocked' :: Simplify t => Blocked t -> ReduceM t
 simplifyBlocked' (Blocked _ t) = return t
@@ -755,17 +815,22 @@ instance Simplify Type where
 instance Simplify Elim where
   simplify' (Apply v) = Apply <$> simplify' v
   simplify' (Proj o f)= pure $ Proj o f
+  simplify' (IApply x y v) = IApply <$> simplify' x <*> simplify' y <*> simplify' v
 
 instance Simplify Sort where
     simplify' s = do
       case s of
         PiSort s1 s2 -> piSort <$> simplify' s1 <*> simplify' s2
-        UnivSort s -> univSort <$> simplify' s
+        UnivSort s -> do
+          ui <- univInf
+          univSort ui <$> simplify' s
         Type s     -> Type <$> simplify' s
-        Prop       -> return s
+        Prop s     -> Prop <$> simplify' s
         Inf        -> return s
         SizeUniv   -> return s
         MetaS x es -> MetaS x <$> simplify' es
+        DefS d es  -> DefS d <$> simplify' es
+        DummyS{}   -> return s
 
 instance Simplify Level where
   simplify' (Max as) = levelMax <$> simplify' as
@@ -829,6 +894,9 @@ instance Simplify Constraint where
   simplify' (ValueCmp cmp t u v) = do
     (t,u,v) <- simplify' (t,u,v)
     return $ ValueCmp cmp t u v
+  simplify' (ValueCmpOnFace cmp p t u v) = do
+    ((p,t),u,v) <- simplify' ((p,t),u,v)
+    return $ ValueCmp cmp t u v
   simplify' (ElimCmp cmp fs t v as bs) =
     ElimCmp cmp fs <$> simplify' t <*> simplify' v <*> simplify' as <*> simplify' bs
   simplify' (LevelCmp cmp u v)    = uncurry (LevelCmp cmp) <$> simplify' (u,v)
@@ -837,12 +905,13 @@ instance Simplify Constraint where
   simplify' (SortCmp cmp a b)     = uncurry (SortCmp cmp) <$> simplify' (a,b)
   simplify' (Guarded c pid)       = Guarded <$> simplify' c <*> pure pid
   simplify' (UnBlock m)           = return $ UnBlock m
-  simplify' (FindInScope m b cands) = FindInScope m b <$> mapM simplify' cands
+  simplify' (FindInstance m b cands) = FindInstance m b <$> mapM simplify' cands
   simplify' (IsEmpty r t)         = IsEmpty r <$> simplify' t
   simplify' (CheckSizeLtSat t)    = CheckSizeLtSat <$> simplify' t
   simplify' c@CheckFunDef{}       = return c
   simplify' (HasBiggerSort a)     = HasBiggerSort <$> simplify' a
   simplify' (HasPTSRule a b)      = uncurry HasPTSRule <$> simplify' (a,b)
+  simplify' (UnquoteTactic m t h g) = UnquoteTactic m <$> simplify' t <*> simplify' h <*> simplify' g
 
 instance Simplify Bool where
   simplify' = return
@@ -864,7 +933,7 @@ instance Simplify DisplayForm where
   simplify' (Display n ps v) = Display n <$> simplify' ps <*> return v
 
 instance Simplify Candidate where
-  simplify' (Candidate u t eti ov) = Candidate <$> simplify' u <*> simplify' t <*> pure eti <*> pure ov
+  simplify' (Candidate u t ov) = Candidate <$> simplify' u <*> simplify' t <*> pure ov
 
 instance Simplify EqualityView where
   simplify' (OtherType t)            = OtherType
@@ -889,12 +958,16 @@ instance Normalise Sort where
       s <- reduce' s
       case s of
         PiSort s1 s2 -> piSort <$> normalise' s1 <*> normalise' s2
-        UnivSort s -> univSort <$> normalise' s
-        Prop       -> return s
+        UnivSort s -> do
+          ui <- univInf
+          univSort ui <$> normalise' s
+        Prop s     -> Prop <$> normalise' s
         Type s     -> Type <$> normalise' s
         Inf        -> return Inf
         SizeUniv   -> return SizeUniv
         MetaS x es -> return s
+        DefS d es  -> return s
+        DummyS{}   -> return s
 
 instance Normalise Type where
     normalise' (El s t) = El <$> normalise' s <*> normalise' t
@@ -914,10 +987,12 @@ slowNormaliseArgs v = case v of
   Sort s      -> Sort       <$> normalise' s
   Pi a b      -> uncurry Pi <$> normalise' (a, b)
   DontCare _  -> return v
+  Dummy{}     -> return v
 
 instance Normalise Elim where
   normalise' (Apply v) = Apply <$> normalise' v
   normalise' (Proj o f)= pure $ Proj o f
+  normalise' (IApply x y v) = IApply <$> normalise' x <*> normalise' y <*> normalise' v
 
 instance Normalise Level where
   normalise' (Max as) = levelMax <$> normalise' as
@@ -976,6 +1051,9 @@ instance Normalise Constraint where
   normalise' (ValueCmp cmp t u v) = do
     (t,u,v) <- normalise' (t,u,v)
     return $ ValueCmp cmp t u v
+  normalise' (ValueCmpOnFace cmp p t u v) = do
+    ((p,t),u,v) <- normalise' ((p,t),u,v)
+    return $ ValueCmpOnFace cmp p t u v
   normalise' (ElimCmp cmp fs t v as bs) =
     ElimCmp cmp fs <$> normalise' t <*> normalise' v <*> normalise' as <*> normalise' bs
   normalise' (LevelCmp cmp u v)    = uncurry (LevelCmp cmp) <$> normalise' (u,v)
@@ -984,12 +1062,13 @@ instance Normalise Constraint where
   normalise' (SortCmp cmp a b)     = uncurry (SortCmp cmp) <$> normalise' (a,b)
   normalise' (Guarded c pid)       = Guarded <$> normalise' c <*> pure pid
   normalise' (UnBlock m)           = return $ UnBlock m
-  normalise' (FindInScope m b cands) = FindInScope m b <$> mapM normalise' cands
+  normalise' (FindInstance m b cands) = FindInstance m b <$> mapM normalise' cands
   normalise' (IsEmpty r t)         = IsEmpty r <$> normalise' t
   normalise' (CheckSizeLtSat t)    = CheckSizeLtSat <$> normalise' t
   normalise' c@CheckFunDef{}       = return c
   normalise' (HasBiggerSort a)     = HasBiggerSort <$> normalise' a
   normalise' (HasPTSRule a b)      = uncurry HasPTSRule <$> normalise' (a,b)
+  normalise' (UnquoteTactic m t h g) = UnquoteTactic m <$> normalise' t <*> normalise' h <*> normalise' g
 
 instance Normalise Bool where
   normalise' = return
@@ -1011,8 +1090,10 @@ instance Normalise a => Normalise (Pattern' a) where
     VarP o x     -> VarP o <$> normalise' x
     LitP _       -> return p
     ConP c mt ps -> ConP c <$> normalise' mt <*> normalise' ps
+    DefP o q ps  -> DefP o q <$> normalise' ps
     DotP o v     -> DotP o <$> normalise' v
     ProjP{}      -> return p
+    IApplyP o t u x -> IApplyP o <$> normalise' t <*> normalise' u <*> normalise' x
 
 instance Normalise DisplayForm where
   normalise' (Display n ps v) = Display n <$> normalise' ps <*> return v
@@ -1024,7 +1105,7 @@ instance Normalise a => Normalise (Maybe a) where
     normalise' = traverse normalise'
 
 instance Normalise Candidate where
-  normalise' (Candidate u t eti ov) = Candidate <$> normalise' u <*> normalise' t <*> pure eti <*> pure ov
+  normalise' (Candidate u t ov) = Candidate <$> normalise' u <*> normalise' t <*> pure ov
 
 instance Normalise EqualityView where
   normalise' (OtherType t)            = OtherType
@@ -1054,19 +1135,23 @@ instance InstantiateFull Sort where
         s <- instantiate' s
         case s of
             Type n     -> Type <$> instantiateFull' n
-            Prop       -> return s
+            Prop n     -> Prop <$> instantiateFull' n
             PiSort s1 s2 -> piSort <$> instantiateFull' s1 <*> instantiateFull' s2
-            UnivSort s -> univSort <$> instantiateFull' s
+            UnivSort s -> do
+              ui <- univInf
+              univSort ui <$> instantiateFull' s
             Inf        -> return s
             SizeUniv   -> return s
             MetaS x es -> MetaS x <$> instantiateFull' es
+            DefS d es  -> DefS d <$> instantiateFull' es
+            DummyS{}   -> return s
 
 instance (InstantiateFull a) => InstantiateFull (Type' a) where
     instantiateFull' (El s t) =
       El <$> instantiateFull' s <*> instantiateFull' t
 
 instance InstantiateFull Term where
-    instantiateFull' v = etaOnce =<< do -- Andreas, 2010-11-12 DONT ETA!! eta-reduction breaks subject reduction
+    instantiateFull' v = etaOnce =<< do -- Andreas, 2010-11-12 DONT ETA!? eta-reduction breaks subject reduction
 -- but removing etaOnce now breaks everything
       v <- instantiate' v
       case v of
@@ -1080,6 +1165,7 @@ instance InstantiateFull Term where
           Sort s      -> Sort <$> instantiateFull' s
           Pi a b      -> uncurry Pi <$> instantiateFull' (a,b)
           DontCare v  -> dontCare <$> instantiateFull' v
+          Dummy{}     -> return v
 
 instance InstantiateFull Level where
   instantiateFull' (Max as) = levelMax <$> instantiateFull' as
@@ -1129,8 +1215,10 @@ instance InstantiateFull a => InstantiateFull (Pattern' a) where
     instantiateFull' (VarP o x)     = VarP o <$> instantiateFull' x
     instantiateFull' (DotP o t)     = DotP o <$> instantiateFull' t
     instantiateFull' (ConP n mt ps) = ConP n <$> instantiateFull' mt <*> instantiateFull' ps
+    instantiateFull' (DefP o q ps) = DefP o q <$> instantiateFull' ps
     instantiateFull' l@LitP{}       = return l
     instantiateFull' p@ProjP{}      = return p
+    instantiateFull' (IApplyP o t u x) = IApplyP o <$> instantiateFull' t <*> instantiateFull' u <*> instantiateFull' x
 
 instance (Subst t a, InstantiateFull a) => InstantiateFull (Abs a) where
     instantiateFull' a@(Abs x _) = Abs x <$> underAbstraction_ a instantiateFull'
@@ -1169,6 +1257,9 @@ instance InstantiateFull Constraint where
     ValueCmp cmp t u v -> do
       (t,u,v) <- instantiateFull' (t,u,v)
       return $ ValueCmp cmp t u v
+    ValueCmpOnFace cmp p t u v -> do
+      ((p,t),u,v) <- instantiateFull' ((p,t),u,v)
+      return $ ValueCmpOnFace cmp p t u v
     ElimCmp cmp fs t v as bs ->
       ElimCmp cmp fs <$> instantiateFull' t <*> instantiateFull' v <*> instantiateFull' as <*> instantiateFull' bs
     LevelCmp cmp u v    -> uncurry (LevelCmp cmp) <$> instantiateFull' (u,v)
@@ -1177,16 +1268,18 @@ instance InstantiateFull Constraint where
     SortCmp cmp a b     -> uncurry (SortCmp cmp) <$> instantiateFull' (a,b)
     Guarded c pid       -> Guarded <$> instantiateFull' c <*> pure pid
     UnBlock m           -> return $ UnBlock m
-    FindInScope m b cands -> FindInScope m b <$> mapM instantiateFull' cands
+    FindInstance m b cands -> FindInstance m b <$> mapM instantiateFull' cands
     IsEmpty r t         -> IsEmpty r <$> instantiateFull' t
     CheckSizeLtSat t    -> CheckSizeLtSat <$> instantiateFull' t
     c@CheckFunDef{}     -> return c
     HasBiggerSort a     -> HasBiggerSort <$> instantiateFull' a
     HasPTSRule a b      -> uncurry HasPTSRule <$> instantiateFull' (a,b)
+    UnquoteTactic m t g h -> UnquoteTactic m <$> instantiateFull' t <*> instantiateFull' g <*> instantiateFull' h
 
 instance (InstantiateFull a) => InstantiateFull (Elim' a) where
   instantiateFull' (Apply v) = Apply <$> instantiateFull' v
   instantiateFull' (Proj o f)= pure $ Proj o f
+  instantiateFull' (IApply x y v) = IApply <$> instantiateFull' x <*> instantiateFull' y <*> instantiateFull' v
 
 instance InstantiateFull e => InstantiateFull (Map k e) where
     instantiateFull' = traverse instantiateFull'
@@ -1214,9 +1307,9 @@ instance InstantiateFull Char where
     instantiateFull' = return
 
 instance InstantiateFull Definition where
-    instantiateFull' (Defn rel x t pol occ df i c inst copy ma inj d) = do
+    instantiateFull' def@Defn{ defType = t ,defDisplay = df, theDef = d } = do
       (t, df, d) <- instantiateFull' (t, df, d)
-      return $ Defn rel x t pol occ df i c inst copy ma inj d
+      return $ def{ defType = t, defDisplay = df, theDef = d }
 
 instance InstantiateFull NLPat where
   instantiateFull' (PVar x y) = return $ PVar x y
@@ -1257,10 +1350,13 @@ instance InstantiateFull DisplayTerm where
 instance InstantiateFull Defn where
     instantiateFull' d = case d of
       Axiom{} -> return d
+      DataOrRecSig{} -> return d
+      GeneralizableVar{} -> return d
       AbstractDefn d -> AbstractDefn <$> instantiateFull' d
-      Function{ funClauses = cs, funCompiled = cc, funInv = inv } -> do
+      Function{ funClauses = cs, funCompiled = cc, funInv = inv, funExtLam = extLam } -> do
         (cs, cc, inv) <- instantiateFull' (cs, cc, inv)
-        return $ d { funClauses = cs, funCompiled = cc, funInv = inv }
+        extLam <- instantiateFull' extLam
+        return $ d { funClauses = cs, funCompiled = cc, funInv = inv, funExtLam = extLam }
       Datatype{ dataSort = s, dataClause = cl } -> do
         s  <- instantiateFull' s
         cl <- instantiateFull' cl
@@ -1274,6 +1370,14 @@ instance InstantiateFull Defn where
         cs <- instantiateFull' cs
         return $ d { primClauses = cs }
 
+instance InstantiateFull ExtLamInfo where
+  instantiateFull' e@(ExtLamInfo { extLamSys = sys}) = do
+    sys <- instantiateFull' sys
+    return $ e { extLamSys = sys}
+
+instance InstantiateFull System where
+  instantiateFull' (System tel sys) = System <$> instantiateFull' tel <*> instantiateFull' sys
+
 instance InstantiateFull FunctionInverse where
   instantiateFull' NotInjective = return NotInjective
   instantiateFull' (Inverse inv) = Inverse <$> instantiateFull' inv
@@ -1282,12 +1386,13 @@ instance InstantiateFull a => InstantiateFull (WithArity a) where
   instantiateFull' (WithArity n a) = WithArity n <$> instantiateFull' a
 
 instance InstantiateFull a => InstantiateFull (Case a) where
-  instantiateFull' (Branches cop cs eta ls m lz) =
+  instantiateFull' (Branches cop cs eta ls m b lz) =
     Branches cop
       <$> instantiateFull' cs
       <*> instantiateFull' eta
       <*> instantiateFull' ls
       <*> instantiateFull' m
+      <*> pure b
       <*> pure lz
 
 instance InstantiateFull CompiledClauses where
@@ -1305,10 +1410,10 @@ instance InstantiateFull Clause where
        <*> return unreachable
 
 instance InstantiateFull Interface where
-    instantiateFull' (Interface h ms mod scope inside
+    instantiateFull' (Interface h s ft ms mod scope inside
                                sig display userwarn b foreignCode
-                               highlighting pragmas patsyns warnings) =
-        Interface h ms mod scope inside
+                               highlighting pragmas usedOpts patsyns warnings) =
+        Interface h s ft ms mod scope inside
             <$> instantiateFull' sig
             <*> instantiateFull' display
             <*> return userwarn
@@ -1316,6 +1421,7 @@ instance InstantiateFull Interface where
             <*> return foreignCode
             <*> return highlighting
             <*> return pragmas
+            <*> return usedOpts
             <*> return patsyns
             <*> return warnings
 
@@ -1333,8 +1439,8 @@ instance InstantiateFull a => InstantiateFull (Maybe a) where
   instantiateFull' = mapM instantiateFull'
 
 instance InstantiateFull Candidate where
-  instantiateFull' (Candidate u t eti ov) =
-    Candidate <$> instantiateFull' u <*> instantiateFull' t <*> pure eti <*> pure ov
+  instantiateFull' (Candidate u t ov) =
+    Candidate <$> instantiateFull' u <*> instantiateFull' t <*> pure ov
 
 instance InstantiateFull EqualityView where
   instantiateFull' (OtherType t)            = OtherType
