@@ -47,6 +47,7 @@ import Control.Monad.ST.Unsafe (unsafeSTToIO, unsafeInterleaveST)
 
 import Data.Map (Map)
 import qualified Data.Map as Map
+import qualified Data.IntMap as IntMap
 import qualified Data.IntSet as IntSet
 import qualified Data.List as List
 import Data.Traversable (traverse)
@@ -111,7 +112,7 @@ data CompactDefn
   = CFun  { cfunCompiled  :: FastCompiledClauses, cfunProjection :: Maybe QName }
   | CCon  { cconSrcCon :: ConHead, cconArity :: Int }
   | CForce   -- ^ primForce
-  | CTrustMe -- ^ primTrustMe
+  | CErase   -- ^ primErase
   | CTyCon   -- ^ Datatype or record type. Need to know this for primForce.
   | CAxiom   -- ^ Axiom or abstract defn
   | CPrimOp Int ([Literal] -> Term) (Maybe FastCompiledClauses)
@@ -120,7 +121,7 @@ data CompactDefn
 
 data BuiltinEnv = BuiltinEnv
   { bZero, bSuc, bTrue, bFalse, bRefl :: Maybe ConHead
-  , bPrimForce, bPrimTrustMe  :: Maybe QName }
+  , bPrimForce, bPrimErase  :: Maybe QName }
 
 -- | Compute a 'CompactDef' from a regular definition.
 compactDef :: BuiltinEnv -> Definition -> RewriteRules -> ReduceM CompactDef
@@ -128,9 +129,9 @@ compactDef bEnv def rewr = do
   cdefn <-
     case theDef def of
       _ | Just (defName def) == bPrimForce bEnv   -> pure CForce
-      _ | Just (defName def) == bPrimTrustMe bEnv ->
+      _ | Just (defName def) == bPrimErase bEnv ->
           case telView' (defType def) of
-            TelV tel _ | size tel == 4 -> pure CTrustMe
+            TelV tel _ | size tel == 5 -> pure CErase
                        | otherwise     -> pure COther
                           -- Non-standard equality. Fall back to slow reduce.
       Constructor{conSrcCon = c, conArity = n} -> pure CCon{cconSrcCon = c, cconArity = n}
@@ -144,7 +145,9 @@ compactDef bEnv def rewr = do
       Datatype{}                     -> pure COther -- TODO
       Record{}                       -> pure COther -- TODO
       Axiom{}                        -> pure CAxiom
+      DataOrRecSig{}                 -> pure CAxiom
       AbstractDefn{}                 -> pure CAxiom
+      GeneralizableVar{}             -> __IMPOSSIBLE__
       Primitive{ primName = name, primCompiled = cc } ->
         case name of
           -- "primShowInteger" -- integers are not literals
@@ -216,7 +219,7 @@ compactDef bEnv def rewr = do
           "primStringEquality"         -> mkPrim 2 $ \ [LitString _ a, LitString _ b] -> bool (b == a)
           "primShowString"             -> mkPrim 1 $ \ a -> string (show $ pretty a)
 
-          -- "primTrustMe"
+          -- "primErase"
           -- "primForce"
           -- "primForceLemma"
           "primQNameEquality"          -> mkPrim 2 $ \ [LitQName _ a, LitQName _ b] -> bool (b == a)
@@ -296,6 +299,8 @@ data FastCase c = FBranches
     -- ^ Map from literal to case subtree.
   , fcatchAllBranch :: Maybe c
     -- ^ (Possibly additional) catch-all clause.
+  , ffallThrough    :: Bool
+    -- ^ (if True) In case of non-canonical argument use catchAllBranch.
   }
 
 noBranches :: FastCase a
@@ -303,7 +308,8 @@ noBranches = FBranches{ fprojPatterns   = False
                       , fconBranches    = Map.empty
                       , fsucBranch      = Nothing
                       , flitBranches    = Map.empty
-                      , fcatchAllBranch = Nothing }
+                      , fcatchAllBranch = Nothing
+                      , ffallThrough    = False }
 
 -- | Case tree with bodies.
 
@@ -313,7 +319,7 @@ data FastCompiledClauses
     -- (counting from zero) with @bs@ as the case branches.
     -- If the @n@-th argument is a projection, we have only 'conBranches'
     -- with arity 0.
-  | FEta Int [QName] FastCompiledClauses (Maybe FastCompiledClauses)
+  | FEta Int [Arg QName] FastCompiledClauses (Maybe FastCompiledClauses)
     -- ^ Match on record constructor. Can still have a catch-all though. Just
     --   contains the fields, not the actual constructor.
   | FDone [Arg ArgName] Term
@@ -334,12 +340,13 @@ fastCompiledClauses bEnv cc =
     Case (Arg _ n) bs -> FCase n (fastCase bEnv bs)
 
 fastCase :: BuiltinEnv -> Case CompiledClauses -> FastCase FastCompiledClauses
-fastCase env (Branches proj con _ lit wild _) =
+fastCase env (Branches proj con _ lit wild fT _) =
   FBranches
     { fprojPatterns   = proj
     , fconBranches    = Map.mapKeysMonotonic (nameId . qnameName) $ fmap (fastCompiledClauses env . content) (stripSuc con)
     , fsucBranch      = fmap (fastCompiledClauses env . content) $ flip Map.lookup con . conName =<< bSuc env
     , flitBranches    = fmap (fastCompiledClauses env) lit
+    , ffallThrough    = fromMaybe False fT
     , fcatchAllBranch = fmap (fastCompiledClauses env) wild }
   where
     stripSuc | Just c <- bSuc env = Map.delete (conName c)
@@ -348,7 +355,7 @@ fastCase env (Branches proj con _ lit wild _) =
 
 {-# INLINE lookupCon #-}
 lookupCon :: QName -> FastCase c -> Maybe c
-lookupCon c (FBranches _ cons _ _ _) = Map.lookup (nameId $ qnameName c) cons
+lookupCon c (FBranches _ cons _ _ _ _) = Map.lookup (nameId $ qnameName c) cons
 
 -- QName memo -------------------------------------------------------------
 
@@ -389,16 +396,16 @@ fastReduce' :: Normalisation -> Term -> ReduceM (Blocked Term)
 fastReduce' norm v = do
   let name (Con c _ _) = c
       name _         = __IMPOSSIBLE__
-  zero    <- fmap name <$> getBuiltin' builtinZero
-  suc     <- fmap name <$> getBuiltin' builtinSuc
-  true    <- fmap name <$> getBuiltin' builtinTrue
-  false   <- fmap name <$> getBuiltin' builtinFalse
-  refl    <- fmap name <$> getBuiltin' builtinRefl
-  force   <- fmap primFunName <$> getPrimitive' "primForce"
-  trustme <- fmap primFunName <$> getPrimitive' "primTrustMe"
+  zero  <- fmap name <$> getBuiltin' builtinZero
+  suc   <- fmap name <$> getBuiltin' builtinSuc
+  true  <- fmap name <$> getBuiltin' builtinTrue
+  false <- fmap name <$> getBuiltin' builtinFalse
+  refl  <- fmap name <$> getBuiltin' builtinRefl
+  force <- fmap primFunName <$> getPrimitive' "primForce"
+  erase <- fmap primFunName <$> getPrimitive' "primErase"
   let bEnv = BuiltinEnv { bZero = zero, bSuc = suc, bTrue = true, bFalse = false, bRefl = refl,
-                          bPrimForce = force, bPrimTrustMe = trustme }
-  allowedReductions <- asks envAllowedReductions
+                          bPrimForce = force, bPrimErase = erase }
+  allowedReductions <- asksTC envAllowedReductions
   rwr <- optRewriting <$> pragmaOptions
   constInfo <- unKleisli $ \f -> do
     info <- getConstInfo f
@@ -522,8 +529,9 @@ extendEnv p (Env xs) = Env (p : xs)
 
 -- | Unsafe.
 lookupEnv_ :: Int -> Env s -> Pointer s
-lookupEnv_ i (Env e) = e !! i
+lookupEnv_ i (Env e) = indexWithDefault __IMPOSSIBLE__ e i
 
+-- Andreas, 2018-11-12, which isn't this just Agda.Utils.List.!!! ?
 lookupEnv :: Int -> Env s -> Maybe (Pointer s)
 lookupEnv i e | i < n     = Just (lookupEnv_ i e)
               | otherwise = Nothing
@@ -567,15 +575,26 @@ data CatchAllFrame s = CatchAll FastCompiledClauses (Spine s)
 
 -- An Elim' with a hole.
 data ElimZipper a = ApplyCxt ArgInfo
+                  | IApplyType a a | IApplyFst a a | IApplySnd a a
   deriving (Eq, Ord, Show, Functor, Foldable, Traversable)
 
 instance Zipper (ElimZipper a) where
   type Carrier (ElimZipper a) = Elim' a
   type Element (ElimZipper a) = a
-  firstHole (Apply arg)   = Just (unArg arg, ApplyCxt (argInfo arg))
-  firstHole Proj{}        = Nothing
-  plugHole x (ApplyCxt i) = Apply (Arg i x)
-  nextHole x c            = Left (plugHole x c)
+
+  firstHole (Apply arg)    = Just (unArg arg, ApplyCxt (argInfo arg))
+  firstHole (IApply a x y) = Just (a, IApplyType x y)
+  firstHole Proj{}         = Nothing
+
+  plugHole x (ApplyCxt i)     = Apply (Arg i x)
+  plugHole a (IApplyType x y) = IApply a x y
+  plugHole x (IApplyFst a y)  = IApply a x y
+  plugHole y (IApplySnd a x)  = IApply a x y
+
+  nextHole a (IApplyType x y) = Right (x, IApplyFst a y)
+  nextHole x (IApplyFst a y)  = Right (y, IApplySnd a x)
+  nextHole y (IApplySnd a x)  = Left (IApply a x y)
+  nextHole x c@ApplyCxt{}     = Left (plugHole x c)
 
 -- | A spine with a single hole for a pointer.
 type SpineContext s = ComposeZipper (ListZipper (Elim' (Pointer s)))
@@ -600,12 +619,14 @@ data ControlFrame s = CaseK QName ArgInfo (FastCase FastCompiledClauses) (Spine
                         --   gets stuck. @spine0@ are the level and type arguments and @spine1@
                         --   contains (if not empty) the continuation and any additional
                         --   eliminations.
-                    | TrustMeK QName (Spine s) (Spine s) (Spine s)
-                        -- ^ @TrustMeK f spine0 spine1 spine2 @. Evaluating @primTrustMe@. The first
-                        --   spine contain the level and type arguments. @spine1@ and @spine2@
+                    | EraseK QName (Spine s) (Spine s) (Spine s) (Spine s)
+                        -- ^ @EraseK f spine0 spine1 spine2 spine3@. Evaluating @primErase@. The
+                        --   first contains the level and type arguments. @spine1@ and @spine2@
                         --   contain at most one argument between them. If in @spine1@ it's the
                         --   value closure of the first argument to be compared and if in @spine2@
                         --   it's the unevaluated closure of the second argument.
+                        --   @spine3@ contains the proof of equality we are erasing. It is passed
+                        --   around but never actually inspected.
                     | NatSucK Integer
                         -- ^ @NatSucK n@. Add @n@ to the focus. If the focus computes to a natural
                         --   number literal this returns a new literal, otherwise it constructs @n@
@@ -691,6 +712,8 @@ elimsToSpine env es = do
 
     thunk (Apply (Arg i t)) = Apply . Arg (unknownFVs i) <$> createThunk (closure (getFreeVariables i) t)
     thunk (Proj o f)        = return (Proj o f)
+    thunk (IApply a x y)    = IApply <$> mkThunk a <*> mkThunk x <*> mkThunk y
+      where mkThunk = createThunk . closure UnknownFVs
 
     -- Going straight for a value for literals is mostly to make debug traces
     -- less verbose and doesn't really buy anything performance-wise.
@@ -728,6 +751,7 @@ buildEnv xs spine = go xs spine emptyEnv
       case sp of
         []           -> (xs0, env, sp)
         Apply c : sp -> go xs sp (unArg c `extendEnv` env)
+        IApply x y r : sp -> go xs sp (r `extendEnv` env)
         _            -> __IMPOSSIBLE__
 
 unusedPointerString :: String
@@ -746,14 +770,15 @@ unusedPointer = Pure (Closure (Value $ notBlocked ())
 --   the term with an attached blocking tag.
 reduceTm :: ReduceEnv -> BuiltinEnv -> (QName -> CompactDef) -> Normalisation -> ReductionFlags -> Term -> Blocked Term
 reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
-    compileAndRun . traceDoc (text "-- fast reduce --")
+    compileAndRun . traceDoc "-- fast reduce --"
   where
     -- Helpers to get information from the ReduceEnv.
     metaStore      = redSt rEnv ^. stMetaStore
-    getMeta m      = maybe __IMPOSSIBLE__ mvInstantiation (Map.lookup m metaStore)
+    getMeta m      = maybe __IMPOSSIBLE__ mvInstantiation (IntMap.lookup (metaId m) metaStore)
     partialDefs    = runReduce getPartialDefs
     rewriteRules f = cdefRewriteRules (constInfo f)
     callByNeed     = envCallByNeed (redEnv rEnv)
+    iview          = runReduce intervalView'
 
     runReduce :: ReduceM a -> a
     runReduce m = unReduceM m rEnv
@@ -777,7 +802,7 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
                Just s  -> (conNameId s ==) . conNameId
 
     -- If there's a non-standard equality (for instance doubly-indexed) we fall back to slow reduce
-    -- for primTrustMe and "unbind" refl.
+    -- for primErase and "unbind" refl.
     refl = refl0 >>= \ c -> if cconArity (cdefDef $ constInfo $ conName c) == 0
                             then Just c else Nothing
 
@@ -808,6 +833,7 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
         -- argument and pushing the appropriate control frame for primitive functions. Fall back to
         -- slow reduce for unsupported definitions.
         Def f [] ->
+          evalIApplyAM spine ctrl $
           let CompactDef{ cdefDelayed        = delayed
                         , cdefNonterminating = nonterm
                         , cdefUnconfirmed    = unconf
@@ -825,9 +851,9 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
             CForce | (spine0, Apply v : spine1) <- splitAt 4 spine ->
               evalPointerAM (unArg v) [] (ForceK f spine0 spine1 : ctrl)
             CForce -> runAM done -- partially applied
-            CTrustMe | (spine0, Apply v : spine1) <- splitAt 2 spine ->
-              evalPointerAM (unArg v) [] (TrustMeK f spine0 [] spine1 : ctrl)
-            CTrustMe -> runAM done -- partially applied
+            CErase | (spine0, Apply v : spine1 : spine2) <- splitAt 2 spine ->
+              evalPointerAM (unArg v) [] (EraseK f spine0 [] [spine1] spine2 : ctrl)
+            CErase -> runAM done -- partially applied
             CPrimOp n op cc | length spine == n,                      -- PrimOps can't be over-applied. They don't
                               Just (v : vs) <- allApplyElims spine -> -- return functions or records.
               evalPointerAM (unArg v) [] (PrimOpK f op [] (map unArg vs) cc : ctrl)
@@ -845,10 +871,11 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
 
         -- Case: constructor. Perform beta reduction if projected from, otherwise return a value.
         Con c i [] ->
+          evalIApplyAM spine ctrl $
           case splitAt ar spine of
-            (args, Proj _ p : spine') -> evalPointerAM (unArg arg) spine' ctrl
+            (args, Proj _ p : spine') -> evalPointerAM (unArg arg) spine' ctrl  -- Andreas #2170: fit argToDontCare here?!
               where
-                fields    = conFields c
+                fields    = map unArg $ conFields c
                 Just n    = List.elemIndex p fields
                 Apply arg = args !! n
             _ -> rewriteAM (evalTrueValue (Con c' i []) env spine ctrl)
@@ -858,6 +885,7 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
         -- pointer. If the variable is not in the environment it's a free variable and we adjust the
         -- deBruijn index appropriately.
         Var x []   ->
+          evalIApplyAM spine ctrl $
           case lookupEnv x env of
             Nothing -> runAM (evalValue (notBlocked ()) (Var (x - envSize env) []) emptyEnv spine ctrl)
             Just p  -> evalPointerAM p spine ctrl
@@ -865,12 +893,15 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
         -- Case: lambda. Perform the beta reduction if applied. Otherwise it's a value.
         Lam h b ->
           case spine of
-            Apply v : spine' ->
-              case b of
-                Abs   _ b -> runAM (evalClosure b (unArg v `extendEnv` env) spine' ctrl)
-                NoAbs _ b -> runAM (evalClosure b env spine' ctrl)
             [] -> runAM done
-            _ -> __IMPOSSIBLE__
+            elim : spine' ->
+              case b of
+                Abs   _ b -> runAM (evalClosure b (getArg elim `extendEnv` env) spine' ctrl)
+                NoAbs _ b -> runAM (evalClosure b env spine' ctrl)
+          where
+            getArg (Apply v)      = unArg v
+            getArg (IApply _ _ v) = v
+            getArg Proj{}         = __IMPOSSIBLE__
 
         -- Case: values. Literals and function types are already in weak-head normal form.
         -- We throw away the environment for literals mostly to make debug printing less verbose.
@@ -888,7 +919,7 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
         -- terms with a specified list of free variables. buildEnv constructs the appropriate
         -- environment for the closure. Avoiding shifting spines for open metas
         -- save a bit of performance.
-        MetaV m es ->
+        MetaV m es -> evalIApplyAM spine ctrl $
           case getMeta m of
             InstV xs t -> do
               spine' <- elimsToSpine env es
@@ -901,6 +932,7 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
         Level{}    -> fallbackAM s
         Sort{}     -> fallbackAM s
         DontCare{} -> fallbackAM s
+        Dummy{}    -> fallbackAM s
 
       where done = Eval (mkValue (notBlocked ()) cl) ctrl
             shiftElims t env0 env es = do
@@ -1004,24 +1036,26 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
           Sort{}     -> True
           Level{}    -> True
           DontCare{} -> True
+          Dummy{}    -> False
           MetaV{}    -> False
           Var{}      -> False
           Def q _  -- Type constructors (data/record) are considered canonical for 'primForce'.
             | CTyCon <- cdefDef (constInfo q) -> True
             | otherwise                       -> False
 
-    -- Case: TrustMeK. We evaluate both arguments to values, then do a simple check for the easy
+    -- Case: EraseK. We evaluate both arguments to values, then do a simple check for the easy
     -- cases and otherwise fall back to slow reduce.
-    runAM' (Eval cl2@(Closure Value{} arg2 _ _) (TrustMeK f spine0 [Apply p1] _ : ctrl)) = do
+    runAM' (Eval cl2@(Closure Value{} arg2 _ _) (EraseK f spine0 [Apply p1] _ spine3 : ctrl)) = do
       cl1@(Closure _ arg1 _ sp1) <- derefPointer_ (unArg p1)
       case (arg1, arg2) of
         (Lit l1, Lit l2) | l1 == l2, isJust refl ->
           runAM (evalTrueValue (Con (fromJust refl) ConOSystem []) emptyEnv [] ctrl)
         _ ->
-          fallbackAM (evalClosure (Def f []) emptyEnv (spine0 ++ map (Apply . hide . defaultArg . pureThunk) [cl1, cl2]) ctrl)
-    runAM' (Eval cl1@(Closure Value{} _ _ _) (TrustMeK f spine0 [] [Apply p2] : ctrl)) =
-      evalPointerAM (unArg p2) [] (TrustMeK f spine0 [Apply $ hide $ defaultArg $ pureThunk cl1] [] : ctrl)
-    runAM' (Eval _ (TrustMeK{} : _)) =
+          let spine = spine0 ++ map (Apply . hide . defaultArg . pureThunk) [cl1, cl2] ++ spine3 in
+          fallbackAM (evalClosure (Def f []) emptyEnv spine ctrl)
+    runAM' (Eval cl1@(Closure Value{} _ _ _) (EraseK f spine0 [] [Apply p2] spine3 : ctrl)) =
+      evalPointerAM (unArg p2) [] (EraseK f spine0 [Apply $ hide $ defaultArg $ pureThunk cl1] [] spine3 : ctrl)
+    runAM' (Eval _ (EraseK{} : _)) =
       __IMPOSSIBLE__
 
     -- Case: UpdateThunk. Write the value to the pointers in the UpdateThunk frame.
@@ -1042,8 +1076,8 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
     runAM' (Eval cl@(Closure (Value blk) t env spine) ctrl0@(CaseK f i bs spine0 spine1 stack : ctrl)) =
       {-# SCC "runAM.CaseK" #-}
       case blk of
-        Blocked{}    -> stuck -- we might as well check the blocking tag first
-        NotBlocked{} -> case t of
+        Blocked{} | null [()|Con{} <- [t]] -> stuck -- we might as well check the blocking tag first
+        _ -> case t of
           -- Case: suc constructor
           Con c ci [] | isSuc c -> matchSuc $ matchCatchall $ failedMatch f stack ctrl
 
@@ -1055,7 +1089,6 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
           Con c ci es -> do
             spine' <- elimsToSpine env es
             runAM (evalValue blk (Con c ci []) emptyEnv (spine' <> spine) ctrl0)
-
           -- Case: natural number literals. Literal natural number patterns are translated to
           -- suc-matches, so there is no need to try matchLit.
           Lit (LitNat _ 0) -> matchLitZero  $ matchCatchall $ failedMatch f stack ctrl
@@ -1064,10 +1097,21 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
           -- Case: literal
           Lit l -> matchLit l $ matchCatchall $ failedMatch f stack ctrl
 
+          -- Case: hcomp
+          Def q [] | isJust $ lookupCon q bs -> matchCon' q (length spine) $ matchCatchall $ failedMatch f stack ctrl
+          Def q es | isJust $ lookupCon q bs -> do
+            spine' <- elimsToSpine env es
+            runAM (evalValue blk (Def q []) emptyEnv (spine' <> spine) ctrl0)
+
           -- Case: not constructor or literal. In this case we are stuck.
           _ -> stuck
       where
-        stuck = do
+        -- If ffallThrough is set we take the catch-all (if any) rather than being stuck. I think
+        -- this happens for partial functions with --cubical (@saizan: is this true?).
+        stuck | ffallThrough bs = matchCatchall reallyStuck
+              | otherwise       = reallyStuck
+
+        reallyStuck = do
             -- Compute new reason for being stuck. See Agda.Syntax.Internal.stuckOn for the logic.
             blk' <- case blk of
                       Blocked{}      -> return blk
@@ -1091,7 +1135,8 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
 
         -- Matching constructor: Switch to the Match state, inserting the constructor arguments in
         -- the spine between spine0 and spine1.
-        matchCon c ci ar = lookupCon (conName c) bs `ifJust` \ cc ->
+        matchCon c ci ar = matchCon' (conName c) ar
+        matchCon' q ar = lookupCon q bs `ifJust` \ cc ->
           runAM (Match f cc (spine0 <> spine <> spine1) catchallStack ctrl)
 
         -- Catch-all: Don't add a CatchAll to the match stack since this _is_ the catch-all.
@@ -1143,7 +1188,7 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
             (spine0, Apply e : spine1) -> do                -- rewriting or 'with'.
               -- Replace e by its projections in the spine. And don't forget a
               -- CatchAll frame if there's a catch-all.
-              let projClosure f = Closure Unevaled (Var 0 []) (extendEnv (unArg e) emptyEnv) [Proj ProjSystem f]
+              let projClosure (Arg ai f) = Closure Unevaled (Var 0 []) (extendEnv (unArg e) emptyEnv) [Proj ProjSystem f]
               projs <- mapM (createThunk . projClosure) fs
               let spine' = spine0 <> map (Apply . defaultArg) projs <> spine1
                   stack' = caseMaybe ca stack $ \ cc -> CatchAll cc spine >: stack
@@ -1172,6 +1217,7 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
                   | otherwise          -> __IMPOSSIBLE__
                 Just cc -> runAM (Match f cc (spine0 <> spine1) stack ctrl)
               where CFun{ cfunProjection = op } = cdefDef (constInfo p)
+            (_, IApply{} : _) -> __IMPOSSIBLE__ -- Paths cannot be defined by pattern matching
       where done why = stuckMatch (NotBlocked why ()) stack ctrl
 
     -- 'evalPointerAM p spine ctrl'. Evaluate the closure pointed to by 'p' applied to 'spine' with
@@ -1187,6 +1233,23 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
         runAM (Eval cl $ updateThunkCtrl p $ [ApplyK spine | not (null spine)] ++ ctrl)
       Thunk cl -> runAM (Eval (clApply cl spine) ctrl)
 
+    -- 'evalIApplyAM spine ctrl fallback' checks if any 'IApply x y r' has a canonical 'r' (i.e. 0 or 1),
+    -- in that case continues evaluating 'x' or 'y' with the rest of 'spine' and same 'ctrl'.
+    -- If no such 'IApply' is found we continue with 'fallback'.
+    evalIApplyAM :: Spine s -> ControlStack s -> ST s (Blocked Term) -> ST s (Blocked Term)
+    evalIApplyAM es ctrl fallback = go es
+      where
+        -- written as a worker/wrapper to possibly trigger some
+        -- specialization wrt fallback
+        go []                  = fallback
+        go (IApply x y r : es) = do
+          br <- evalPointerAM r [] []
+          case iview $ ignoreBlocking br of
+            IZero -> evalPointerAM x es ctrl
+            IOne  -> evalPointerAM y es ctrl
+            _     -> (<* blockedOrMeta br) <$> go es
+        go (e : es) = go es
+
     -- Normalise the spine and apply the closure to the result. The closure must be a value closure.
     normaliseArgsAM :: Closure s -> Spine s -> ControlStack s -> ST s (Blocked Term)
     normaliseArgsAM cl []    ctrl = runAM (Eval cl ctrl)  -- nothing to do
@@ -1225,7 +1288,7 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
     rewriteAM' :: AM s -> ST s (Blocked Term)
     rewriteAM' s@(Eval (Closure (Value blk) t env spine) ctrl)
       | null rewr = runAM s
-      | otherwise = traceDoc (text "R" <+> pretty s) $ do
+      | otherwise = traceDoc ("R" <+> pretty s) $ do
         v0 <- decodeClosure_ (Closure Unevaled t env [])
         es <- decodeSpine spine
         case runReduce (rewrite blk v0 rewr es) of
@@ -1259,7 +1322,7 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
     unfoldDelayed (UpdateThunk{} : ctrl) = unfoldDelayed ctrl
     unfoldDelayed (ApplyK{}      : ctrl) = unfoldDelayed ctrl
     unfoldDelayed (ForceK{}      : ctrl) = unfoldDelayed ctrl
-    unfoldDelayed (TrustMeK{}    : ctrl) = unfoldDelayed ctrl
+    unfoldDelayed (EraseK{}      : ctrl) = unfoldDelayed ctrl
 
     -- When matching is stuck we return the closure from the 'MatchStack' with the appropriate
     -- 'IsValue' set.
@@ -1294,25 +1357,22 @@ reduceTm rEnv bEnv !constInfo normalisation ReductionFlags{..} =
 -- Pretty printing --------------------------------------------------------
 
 instance Pretty a => Pretty (FastCase a) where
-  prettyPrec p (FBranches _cop cs suc ls m) =
+  prettyPrec p (FBranches _cop cs suc ls m _) =
     mparens (p > 0) $ vcat (prettyMap cs ++ prettyMap ls ++ prSuc suc ++ prC m)
     where
       prC Nothing = []
-      prC (Just x) = [text "_ ->" <?> pretty x]
+      prC (Just x) = ["_ ->" <?> pretty x]
 
       prSuc Nothing  = []
-      prSuc (Just x) = [text "suc ->" <?> pretty x]
-
-instance Pretty NameId where
-  pretty = text . show
+      prSuc (Just x) = ["suc ->" <?> pretty x]
 
 instance Pretty FastCompiledClauses where
-  pretty (FDone xs t) = (text "done" <+> prettyList xs) <?> prettyPrec 10 t
-  pretty FFail        = text "fail"
+  pretty (FDone xs t) = ("done" <+> prettyList xs) <?> prettyPrec 10 t
+  pretty FFail        = "fail"
   pretty (FEta n _ cc ca) =
     text ("eta " ++ show n ++ " of") <?>
-      vcat ([ text "{} ->" <?> pretty cc ] ++
-            [ text "_ ->" <?> pretty cc | Just cc <- [ca] ])
+      vcat ([ "{} ->" <?> pretty cc ] ++
+            [ "_ ->" <?> pretty cc | Just cc <- [ca] ])
   pretty (FCase n bs) | fprojPatterns bs =
     sep [ text $ "project " ++ show n
         , nest 2 $ pretty bs
@@ -1321,7 +1381,7 @@ instance Pretty FastCompiledClauses where
     text ("case " ++ show n ++ " of") <?> pretty bs
 
 instance Pretty a => Pretty (Thunk a) where
-  prettyPrec _ BlackHole  = text "<BLACKHOLE>"
+  prettyPrec _ BlackHole  = "<BLACKHOLE>"
   prettyPrec p (Thunk cl) = prettyPrec p cl
 
 instance Pretty (Pointer s) where
@@ -1329,7 +1389,7 @@ instance Pretty (Pointer s) where
 
 instance Pretty (Closure s) where
   prettyPrec _ (Closure Value{} (Lit (LitString _ unused)) _ _)
-    | unused == unusedPointerString = text "_"
+    | unused == unusedPointerString = "_"
   prettyPrec p (Closure isV t env spine) =
     mparens (p > 9) $ fsep [ text tag
                            , nest 2 $ prettyPrec 10 t
@@ -1341,32 +1401,32 @@ instance Pretty (Closure s) where
 instance Pretty (AM s) where
   prettyPrec p (Eval cl ctrl)  = prettyPrec p cl <?> prettyList ctrl
   prettyPrec p (Match f cc sp stack ctrl) =
-    mparens (p > 9) $ sep [ text "M" <+> pretty f
+    mparens (p > 9) $ sep [ "M" <+> pretty f
                           , nest 2 $ prettyList sp
                           , nest 2 $ prettyPrec 10 cc
                           , nest 2 $ pretty stack
                           , nest 2 $ prettyList ctrl ]
 
 instance Pretty (CatchAllFrame s) where
-  pretty CatchAll{} = text "CatchAll"
+  pretty CatchAll{} = "CatchAll"
 
 instance Pretty (MatchStack s) where
   pretty ([] :> _) = empty
   pretty (ca :> _) = prettyList ca
 
 instance Pretty (ControlFrame s) where
-  prettyPrec p (CaseK f _ _ _ _ mc)     = mparens (p > 9) $ (text "CaseK" <+> pretty (qnameName f)) <?> pretty mc
-  prettyPrec p (ForceK _ spine0 spine1) = mparens (p > 9) $ text "ForceK" <?> prettyList (spine0 <> spine1)
-  prettyPrec p (TrustMeK _ sp0 sp1 sp2) = mparens (p > 9) $ sep [ text "TrustMeK"
-                                                                , nest 2 $ prettyList sp0
-                                                                , nest 2 $ prettyList sp1
-                                                                , nest 2 $ prettyList sp2 ]
+  prettyPrec p (CaseK f _ _ _ _ mc)       = mparens (p > 9) $ ("CaseK" <+> pretty (qnameName f)) <?> pretty mc
+  prettyPrec p (ForceK _ spine0 spine1)   = mparens (p > 9) $ "ForceK" <?> prettyList (spine0 <> spine1)
+  prettyPrec p (EraseK _ sp0 sp1 sp2 sp3) = mparens (p > 9) $ sep [ "EraseK"
+                                                                  , nest 2 $ prettyList sp0
+                                                                  , nest 2 $ prettyList sp1
+                                                                  , nest 2 $ prettyList sp2
+                                                                  , nest 2 $ prettyList sp3 ]
   prettyPrec _ (NatSucK n)              = text ("+" ++ show n)
-  prettyPrec p (PrimOpK f _ vs cls _)   = mparens (p > 9) $ sep [ text "PrimOpK" <+> pretty f
+  prettyPrec p (PrimOpK f _ vs cls _)   = mparens (p > 9) $ sep [ "PrimOpK" <+> pretty f
                                                                 , nest 2 $ prettyList vs
                                                                 , nest 2 $ prettyList cls ]
-  prettyPrec p (UpdateThunk ps)         = mparens (p > 9) $ text "UpdateThunk" <+> text (show (length ps))
-  prettyPrec p (ApplyK spine)           = mparens (p > 9) $ text "ApplyK" <?> prettyList spine
-  prettyPrec p NormaliseK               = text "NormaliseK"
-  prettyPrec p (ArgK cl _)              = mparens (p > 9) $ sep [ text "ArgK" <+> prettyPrec 10 cl ]
-
+  prettyPrec p (UpdateThunk ps)         = mparens (p > 9) $ "UpdateThunk" <+> text (show (length ps))
+  prettyPrec p (ApplyK spine)           = mparens (p > 9) $ "ApplyK" <?> prettyList spine
+  prettyPrec p NormaliseK               = "NormaliseK"
+  prettyPrec p (ArgK cl _)              = mparens (p > 9) $ sep [ "ArgK" <+> prettyPrec 10 cl ]
