@@ -4,6 +4,7 @@
 module Agda.TypeChecking.Rules.Builtin
   ( bindBuiltin
   , bindBuiltinNoDef
+  , builtinKindOfName
   , bindPostulatedName
   , isUntypedBuiltin
   , bindUntypedBuiltin
@@ -16,6 +17,8 @@ import Control.Monad.State (get)
 import Data.List (find, sortBy)
 import Data.Function (on)
 
+import Agda.Interaction.Options (optSizedTypes)
+
 import qualified Agda.Syntax.Abstract as A
 import qualified Agda.Syntax.Abstract.Views as A
 import Agda.Syntax.Common
@@ -33,7 +36,9 @@ import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.EtaContract
 import Agda.TypeChecking.Functions
 import Agda.TypeChecking.Irrelevance
+import Agda.TypeChecking.Names
 import Agda.TypeChecking.Primitive
+import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
@@ -50,6 +55,7 @@ import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.NonemptyList
 import Agda.Utils.Null
+import Agda.Utils.Permutation
 import Agda.Utils.Size
 
 #include "undefined.h"
@@ -62,6 +68,9 @@ import Agda.Utils.Impossible
 builtinPostulate :: TCM Type -> BuiltinDescriptor
 builtinPostulate = BuiltinPostulate Relevant
 
+builtinPostulateC :: TCM Type -> BuiltinDescriptor
+builtinPostulateC m = BuiltinPostulate Relevant $ requireCubical >> m
+
 findBuiltinInfo :: String -> Maybe BuiltinInfo
 findBuiltinInfo b = find ((b ==) . builtinName) coreBuiltins
 
@@ -73,6 +82,14 @@ coreBuiltins =
   , (builtinArgInfo            |-> BuiltinData tset [builtinArgArgInfo])
   , (builtinBool               |-> BuiltinData tset [builtinTrue, builtinFalse])
   , (builtinNat                |-> BuiltinData tset [builtinZero, builtinSuc])
+  , (builtinSigma              |-> BuiltinData (runNamesT [] $
+                                                hPi' "la" (el $ cl primLevel) $ \ a ->
+                                                hPi' "lb" (el $ cl primLevel) $ \ b ->
+                                                nPi' "A" (sort . tmSort <$> a) $ \bA ->
+                                                nPi' "B" (el' a bA --> (sort . tmSort <$> b)) $ \bB ->
+                                                ((sort . tmSort) <$> (cl primLevelMax <@> a <@> b))
+                                                )
+                                               ["SIGMACON"])
   , (builtinUnit               |-> BuiltinData tset [builtinUnitUnit])  -- actually record, but they are treated the same
   , (builtinAgdaLiteral        |-> BuiltinData tset [builtinAgdaLitNat, builtinAgdaLitWord64, builtinAgdaLitFloat,
                                                      builtinAgdaLitChar, builtinAgdaLitString,
@@ -96,6 +113,103 @@ coreBuiltins =
   , (builtinQName              |-> builtinPostulate tset)
   , (builtinAgdaMeta           |-> builtinPostulate tset)
   , (builtinIO                 |-> builtinPostulate (tset --> tset))
+  , (builtinPath               |-> BuiltinUnknown
+                                               (Just $ requireCubical >>
+                                               (hPi "a" (el primLevel) $
+                                                hPi "A" (return $ sort $ varSort 0) $
+                                                (El (varSort 1) <$> varM 0) -->
+                                                (El (varSort 1) <$> varM 0) -->
+                                                return (sort $ varSort 1)))
+                                               verifyPath)
+  , (builtinPathP              |-> builtinPostulateC (hPi "a" (el primLevel) $
+                                                nPi "A" (tinterval --> (return $ sort $ varSort 0)) $
+                                                (El (varSort 1) <$> varM 0 <@> primIZero) -->
+                                                (El (varSort 1) <$> varM 0 <@> primIOne) -->
+                                                return (sort $ varSort 1)))
+  , (builtinInterval           |-> BuiltinData (requireCubical >> return (sort Inf)) [builtinIZero,builtinIOne])
+  , (builtinSub                |-> builtinPostulateC (runNamesT [] $ hPi' "a" (el $ cl primLevel) $ \ a ->
+                                                     nPi' "A" (el' (cl primLevelSuc <@> a) (Sort . tmSort <$> a)) $ \ bA ->
+                                                     nPi' "φ" (elInf $ cl primInterval) $ \ phi ->
+                                                     elInf (cl primPartial <#> a <@> phi <@> bA) --> (return $ sort Inf)
+                                                    ))
+  , (builtinSubIn              |-> builtinPostulateC (runNamesT [] $
+                                                     hPi' "a" (el $ cl primLevel) $ \ a ->
+                                                     hPi' "A" (el' (cl primLevelSuc <@> a) (Sort . tmSort <$> a)) $ \ bA ->
+                                                     hPi' "φ" (elInf $ cl primInterval) $ \ phi ->
+                                                     nPi' "x" (el' (Sort . tmSort <$> a) bA) $ \ x ->
+                                                     elInf $ cl primSub <#> a <@> bA <@> phi <@> (lam "o" $ \ _ -> x)))
+  , (builtinIZero              |-> BuiltinDataCons (elInf primInterval))
+  , (builtinIOne               |-> BuiltinDataCons (elInf primInterval))
+  , (builtinPartial            |-> BuiltinPrim "primPartial" (const $ return ()))
+  , (builtinPartialP           |-> BuiltinPrim "primPartialP" (const $ return ()))
+  , (builtinIsOne              |-> builtinPostulateC (tinterval --> return (sort $ Inf)))
+  , (builtinItIsOne            |-> builtinPostulateC (elInf $ primIsOne <@> primIOne))
+  , (builtinIsOne1             |-> builtinPostulateC (runNamesT [] $
+                                                     nPi' "i" (cl tinterval) $ \ i ->
+                                                     nPi' "j" (cl tinterval) $ \ j ->
+                                                     nPi' "i1" (elInf $ cl primIsOne <@> i) $ \ i1 ->
+                                                     (elInf $ cl primIsOne <@> (cl primIMax <@> i <@> j))))
+  , (builtinIsOne2             |-> builtinPostulateC (runNamesT [] $
+                                                     nPi' "i" (cl tinterval) $ \ i ->
+                                                     nPi' "j" (cl tinterval) $ \ j ->
+                                                     nPi' "j1" (elInf $ cl primIsOne <@> j) $ \ j1 ->
+                                                     (elInf $ cl primIsOne <@> (cl primIMax <@> i <@> j))))
+  , (builtinIsOneEmpty         |-> builtinPostulateC (runNamesT [] $
+                                                     hPi' "l" (el $ cl primLevel) $ \ l ->
+                                                     hPi' "A" (pPi' "o" (cl primIZero) $ \ _ ->
+                                                                    el' (cl primLevelSuc <@> l) (Sort . tmSort <$> l)) $ \ bA ->
+                                                     pPi' "o" (cl primIZero) (\ o -> el' l $ gApply' (setRelevance Irrelevant defaultArgInfo) bA o)))
+
+  , (builtinId                 |-> builtinPostulateC (hPi "a" (el primLevel) $
+                                                hPi "A" (return $ sort $ varSort 0) $
+                                                (El (varSort 1) <$> varM 0) -->
+                                                (El (varSort 1) <$> varM 0) -->
+                                                return (sort $ varSort 1)))
+  , (builtinConId                 |-> builtinPostulateC (hPi "a" (el primLevel) $
+                                                hPi "A" (return $ sort $ varSort 0) $
+                                                hPi "x" (El (varSort 1) <$> varM 0) $
+                                                hPi "y" (El (varSort 2) <$> varM 1) $
+                                                tinterval --> (El (varSort 3) <$> primPath <#> varM 3 <#> varM 2 <@> varM 1 <@> varM 0) -->
+                                                (El (varSort 3) <$> primId <#> varM 3 <#> varM 2 <@> varM 1 <@> varM 0)))
+  , (builtinEquiv            |-> BuiltinUnknown (Just $ requireCubical >> runNamesT [] (
+                                                    hPi' "l" (el $ cl primLevel) $ \ a ->
+                                                    hPi' "l'" (el $ cl primLevel) $ \ b ->
+                                                    nPi' "A" (sort . tmSort <$> a) $ \bA ->
+                                                    nPi' "B" (sort . tmSort <$> b) $ \bB ->
+                                                    ((sort . tmSort) <$> (cl primLevelMax <@> a <@> b))
+                                                  ))
+                                                   (const $ const $ return ()))
+  , (builtinEquivFun            |-> BuiltinUnknown (Just $ requireCubical >> runNamesT [] (
+                                                    hPi' "l" (el $ cl primLevel) $ \ a ->
+                                                    hPi' "l'" (el $ cl primLevel) $ \ b ->
+                                                    hPi' "A" (sort . tmSort <$> a) $ \bA ->
+                                                    hPi' "B" (sort . tmSort <$> b) $ \bB ->
+                                                    el' (cl primLevelMax <@> a <@> b) (cl primEquiv <#> a <#> b <@> bA <@> bB) --> (el' a bA --> el' b bB)
+                                                  ))
+                                                   (const $ const $ return ()))
+  , (builtinEquivProof            |-> BuiltinUnknown (Just $ requireCubical >> runNamesT [] (
+                                                    hPi' "l" (el $ cl primLevel) $ \ la ->
+                                                    hPi' "l'" (el $ cl primLevel) $ \ lb ->
+                                                    nPi' "A" (sort . tmSort <$> la) $ \ bA ->
+                                                    nPi' "B" (sort . tmSort <$> lb) $ \ bB ->
+                                                    nPi' "e" (el' (cl primLevelMax <@> la <@> lb) (cl primEquiv <#> la <#> lb <@> bA <@> bB)) $ \ e -> do
+                                                    nPi' "b" (el' lb bB) $ \ b -> do
+                                                     let f = cl primEquivFun <#> la <#> lb <#> bA <#> bB <@> e
+                                                         fiber = el' (cl primLevelMax <@> la <@> lb) (cl primSigma <#> la <#> lb
+                                                                                                       <@> bA
+                                                                                                       <@> (lam "a" $ \ a -> cl primPath <#> lb <#> bB <@> (f <@> a) <@> b))
+                                                     nPi' "φ" (cl tinterval) $ \ phi ->
+                                                       (pPi' "o" phi $ \ o -> fiber) --> fiber
+                                                  ))
+                                                   (const $ const $ return ()))
+  , (builtinPathToEquiv        |-> BuiltinUnknown
+                                                (Just $ requireCubical >> runNamesT [] (
+                                                   hPi' "l" (cl tinterval --> (el $ cl primLevel)) $ \ a ->
+                                                   nPi' "E" (nPi' "i" (cl tinterval) $ \ i -> (sort . tmSort <$> (a <@> i))) $ \bE ->
+                                                   el' (cl primLevelMax <@> (a <@> cl primIZero) <@> (a <@> cl primIOne))
+                                                    (cl primEquiv <#> (a <@> cl primIZero) <#> (a <@> cl primIOne) <@> (bE <@> cl primIZero) <@> (bE <@> cl primIOne)
+                                                         )))
+                                         (const $ const $ return ()))
   , (builtinAgdaSort           |-> BuiltinData tset [builtinAgdaSortSet, builtinAgdaSortLit, builtinAgdaSortUnsupported])
   , (builtinAgdaTerm           |-> BuiltinData tset
                                      [ builtinAgdaTermVar, builtinAgdaTermLam, builtinAgdaTermExtLam
@@ -189,6 +303,7 @@ coreBuiltins =
   , (builtinLevelZero          |-> BuiltinPrim "primLevelZero" (const $ return ()))
   , (builtinLevelSuc           |-> BuiltinPrim "primLevelSuc" (const $ return ()))
   , (builtinLevelMax           |-> BuiltinPrim "primLevelMax" verifyMax)
+  , (builtinSetOmega           |-> BuiltinPrim "primSetOmega" (const $ return ()))
   , (builtinAgdaClause         |-> BuiltinData tset [builtinAgdaClauseClause, builtinAgdaClauseAbsurd])
   , (builtinAgdaClauseClause   |-> BuiltinDataCons (tlist (targ tpat) --> tterm --> tclause))
   , (builtinAgdaClauseAbsurd   |-> BuiltinDataCons (tlist (targ tpat) --> tclause))
@@ -234,10 +349,14 @@ coreBuiltins =
   , builtinAgdaTCMIsMacro            |-> builtinPostulate (tqname --> tTCM_ primBool)
   , builtinAgdaTCMWithNormalisation  |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $ tbool --> tTCM 1 (varM 0) --> tTCM 1 (varM 0))
   , builtinAgdaTCMDebugPrint         |-> builtinPostulate (tstring --> tnat --> tlist terrorpart --> tTCM_ primUnit)
+  , builtinAgdaTCMNoConstraints      |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $ tTCM 1 (varM 0) --> tTCM 1 (varM 0))
+  , builtinAgdaTCMRunSpeculative     |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $
+                                                           tTCM 1 (primSigma <#> varM 1 <#> primLevelZero <@> varM 0 <@> (Lam defaultArgInfo . Abs "_" <$> primBool)) --> tTCM 1 (varM 0))
   ]
   where
         (|->) = BuiltinInfo
 
+        v0,v1,v2,v3 :: TCM Term
         v0 = varM 0
         v1 = varM 1
         v2 = varM 2
@@ -289,6 +408,7 @@ coreBuiltins =
         tclause    = el primAgdaClause
         tTCM l a   = elV l (primAgdaTCM <#> varM l <@> a)
         tTCM_ a    = el (primAgdaTCM <#> primLevelZero <@> a)
+        tinterval  = El Inf <$> primInterval
 
         verifyPlus plus =
             verify ["n","m"] $ \(@@) zero suc (==) (===) choice -> do
@@ -405,6 +525,12 @@ coreBuiltins =
             xs <- mapM freshName_ xs
             addContext (xs, domFromArg $ defaultArg nat) $ f apply1 zero suc (==) (===) choice
 
+        verifyPath :: Term -> Type -> TCM ()
+        verifyPath path t = do
+          let hlam n t = glam (setHiding Hidden defaultArgInfo) n t
+          noConstraints $ equalTerm t path =<< (runNamesT [] $
+            hlam "l" $ \ l -> hlam "A" $ \ bA -> cl primPathP <#> l <@> (lam "i" $ \ _ -> bA))
+
 -- | Checks that builtin with name @b : String@ of type @t : Term@
 --   is a data type or inductive record with @n : Int@ constructors.
 --   Returns the name of the data/record type.
@@ -518,6 +644,15 @@ bindBuiltinUnit t = do
       bindBuiltinName builtinUnitUnit (Con con ConOSystem [])
     _ -> genericError "Builtin UNIT must be a singleton record type"
 
+bindBuiltinSigma :: Term -> TCM ()
+bindBuiltinSigma t = do
+  sigma <- fromMaybe __IMPOSSIBLE__ <$> getDef t
+  def <- theDef <$> getConstInfo sigma
+  case def of
+    Record { recFields = [fst,snd], recConHead = con } -> do
+      bindBuiltinName builtinSigma t
+    _ -> genericError "Builtin SIGMA must be a record type with two fields"
+
 -- | Bind BUILTIN EQUALITY and BUILTIN REFL.
 bindBuiltinEquality :: ResolvedName -> TCM ()
 bindBuiltinEquality x = do
@@ -582,6 +717,7 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
            | s == builtinNat      -> bindBuiltinNat      v
            | s == builtinInteger  -> bindBuiltinInt      v
            | s == builtinUnit     -> bindBuiltinUnit     v
+           | s == builtinSigma    -> bindBuiltinSigma    v
            | s == builtinList     -> bindBuiltinData s   v
            | otherwise            -> bindBuiltinName s   v
 
@@ -641,10 +777,12 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
             case theDef def of
               Axiom {} -> do
                 builtinSizeHook s q t'
+                -- And compilation pragmas for base types
                 when (s == builtinChar)   $ addHaskellPragma q "= type Char"
                 when (s == builtinString) $ addHaskellPragma q "= type Data.Text.Text"
                 when (s == builtinFloat)  $ addHaskellPragma q "= type Double"
                 when (s == builtinWord64) $ addHaskellPragma q "= type MAlonzo.RTE.Word64"
+                when (s == builtinPathP)  $ builtinPathPHook q
                 bindBuiltinName s v
               _        -> err
           _ -> err
@@ -656,6 +794,12 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
         f v t
         bindBuiltinName s v
 
+builtinPathPHook :: QName -> TCM ()
+builtinPathPHook q =
+      modifySignature $ updateDefinition q
+      $ updateDefPolarity       id
+      . updateDefArgOccurrences (const [Unused,StrictPos,Mixed,Mixed])
+
 -- | Bind a builtin thing to an expression.
 bindBuiltin :: String -> ResolvedName -> TCM ()
 bindBuiltin b x = do
@@ -707,11 +851,17 @@ bindUntypedBuiltin b = \case
 -- We simply ignore the parameters.
 bindBuiltinNoDef :: String -> A.QName -> TCM ()
 bindBuiltinNoDef b q = inTopContext $ do
+  when (b `elem` sizeBuiltins) $ unlessM sizedTypesOption $
+    genericError $ "Cannot declare size BUILTIN " ++ b ++ " with option --no-sized-types"
   case builtinDesc <$> findBuiltinInfo b of
     Just (BuiltinPostulate rel mt) -> do
+      -- We start by adding the corresponding postulate
       t <- mt
       addConstant q $ defaultDefn (setRelevance rel defaultArgInfo) q t def
+      -- And we then *modify* the definition based on our needs:
+      -- We add polarity information for SIZE-related definitions
       builtinSizeHook b q t
+      -- Finally, bind the BUILTIN in the environment.
       bindBuiltinName b $ Def q []
       where
         -- Andreas, 2015-02-14
@@ -723,6 +873,57 @@ bindBuiltinNoDef b q = inTopContext $ do
                 , funTerminates = Just True
                 }
             | otherwise = Axiom
+    Just (BuiltinPrim name axioms) -> do
+      PrimImpl t pf <- lookupPrimitiveFunction name
+      bindPrimitive name $ pf { primFunName = q }
+      let v   = Def q []
+          def = Primitive { primAbstr    = ConcreteDef
+                          , primName     = name
+                          , primClauses  = []
+                          , primInv      = NotInjective
+                          , primCompiled = Just (CC.Done [] $ Def q [])
+                          }
+      addConstant q $ defaultDefn defaultArgInfo q t def
+      axioms v
+      bindBuiltinName b v
+    Just (BuiltinDataCons mt) -> do
+      t <- mt
+      d <- return $! getPrimName $ unEl t
+      let
+        ch = ConHead q Inductive []
+        def = Constructor 0 0 ch d ConcreteDef Inductive (emptyCompKit, Nothing) [] [] -- Andrea TODO: fix zeros
 
+      addConstant q $ defaultDefn defaultArgInfo q t def
+      addDataCons d [q]
+      bindBuiltinName b $ Con ch ConOSystem []
+    Just (BuiltinData mt cs) -> do
+      t <- mt
+      addConstant q $ defaultDefn defaultArgInfo q t def
+      bindBuiltinName b $ Def q []
+      where
+        def = Datatype
+              { dataPars       = 0
+              , dataIxs        = 0
+              , dataInduction  = Inductive
+              , dataClause     = Nothing
+              , dataCons       = []     -- Constructors are added later
+              , dataSort       = Inf
+              , dataAbstr      = ConcreteDef
+              , dataMutual     = Nothing
+              , dataPathCons   = []
+              }
     Just{}  -> __IMPOSSIBLE__
     Nothing -> __IMPOSSIBLE__ -- typeError $ NoSuchBuiltinName b
+
+
+builtinKindOfName :: String -> TCM (Maybe KindOfName)
+builtinKindOfName b =
+               case find ((b ==) . builtinName) coreBuiltins of
+                    Nothing -> return Nothing
+                    Just d -> return . Just $
+                      case builtinDesc d of
+                        BuiltinDataCons{}  -> ConName
+                        BuiltinData{}      -> DefName
+                        BuiltinPrim{}      -> DefName
+                        BuiltinPostulate{} -> DefName
+                        BuiltinUnknown{}   -> DefName
