@@ -9,15 +9,19 @@ import Prelude hiding ( (<>), mapM, null )
 import Prelude hiding ( mapM, null )
 #endif
 
-import Control.Arrow ((***))
+import Control.Arrow ((***),first,second)
 import Control.Monad.State hiding (forM, mapM)
 import Control.Monad.Reader hiding (forM, mapM)
 
 import Data.Function
+import qualified Data.List as List
 import Data.Maybe
 import Data.Traversable (Traversable, traverse, forM, mapM)
 import qualified Data.Map as Map
 import qualified Data.Set as Set
+import Data.Semigroup (Semigroup((<>)))
+
+import Agda.Interaction.Options
 
 import Agda.Syntax.Common
 import qualified Agda.Syntax.Concrete as C
@@ -36,28 +40,32 @@ import Agda.Syntax.Info
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
 import qualified Agda.TypeChecking.Monad.Benchmark as Bench
+import Agda.TypeChecking.Warnings ( warning )
 
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Conversion
+import Agda.TypeChecking.Coverage.SplitTree
 import Agda.TypeChecking.Inlining
 import Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Patterns.Abstract (expandPatternSynonyms)
-import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.Pretty hiding ((<>))
+import qualified Agda.TypeChecking.Pretty as Pr
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Free
-import Agda.TypeChecking.CheckInternal (checkType, inferSort)
+import Agda.TypeChecking.CheckInternal
 import Agda.TypeChecking.With
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Injectivity
 import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.SizedTypes.Solve
 import Agda.TypeChecking.RecordPatterns
+import Agda.TypeChecking.Records
 import Agda.TypeChecking.CompiledClause (CompiledClauses'(..), hasProjectionPatterns)
 import Agda.TypeChecking.CompiledClause.Compile
 import Agda.TypeChecking.Primitive hiding (Nat)
 
-import Agda.TypeChecking.Rules.Term                ( checkExpr, inferExpr, inferExprForWith, checkDontExpandLast, checkTelescope, catchIlltypedPatternBlockedOnMeta )
+import Agda.TypeChecking.Rules.Term
 import Agda.TypeChecking.Rules.LHS                 ( checkLeftHandSide, LHSResult(..), bindAsPatterns )
 import Agda.TypeChecking.Rules.LHS.Problem         ( AsBinding(..) )
 import {-# SOURCE #-} Agda.TypeChecking.Rules.Decl ( checkDecls )
@@ -65,6 +73,7 @@ import {-# SOURCE #-} Agda.TypeChecking.Rules.Decl ( checkDecls )
 import Agda.Utils.Except ( MonadError(catchError, throwError) )
 import Agda.Utils.Functor
 import Agda.Utils.Lens
+import Agda.Utils.List
 import Agda.Utils.Maybe ( whenNothing )
 import Agda.Utils.Monad
 import Agda.Utils.Null
@@ -87,7 +96,7 @@ checkFunDef delayed i name cs = do
         info  <- flip setRelevance defaultArgInfo <$> relOfConst name
         case isAlias cs t of
           Just (e, mc, x) ->
-            traceCall (CheckFunDefCall (getRange i) (qnameName name) cs) $ do
+            traceCall (CheckFunDefCall (getRange i) name cs) $ do
               -- Andreas, 2012-11-22: if the alias is in an abstract block
               -- it has been frozen.  We unfreeze it to enable type inference.
               -- See issue 729.
@@ -100,7 +109,7 @@ checkFunDef delayed i name cs = do
         when (ismacro || Info.defMacro i == MacroDef) $ checkMacroType t
     `catchIlltypedPatternBlockedOnMeta` \ (err, x) -> do
         reportSDoc "tc.def" 20 $ vcat $
-          [ text "checking function definition got stuck on meta: " <+> text (show x) ]
+          [ "checking function definition got stuck on meta: " <+> text (show x) ]
         addConstraint $ CheckFunDef delayed i name cs
 
 checkMacroType :: Type -> TCM ()
@@ -112,7 +121,7 @@ checkMacroType t = do
       resType = abstract (telFromList (drop (length telList - 1) telList)) tr
   expectedType <- el primAgdaTerm --> el (primAgdaTCM <#> primLevelZero <@> primUnit)
   equalType resType expectedType
-    `catchError` \ _ -> typeError . GenericDocError =<< sep [ text "Result type of a macro must be"
+    `catchError` \ _ -> typeError . GenericDocError =<< sep [ "Result type of a macro must be"
                                                             , nest 2 $ prettyTCM expectedType ]
 
 -- | A single clause without arguments and without type signature is an alias.
@@ -132,24 +141,18 @@ isAlias cs t =
 
 -- | Check a trivial definition of the form @f = e@
 checkAlias :: Type -> ArgInfo -> Delayed -> Info.DefInfo -> QName -> A.Expr -> Maybe C.Expr -> TCM ()
-checkAlias t' ai delayed i name e mc = atClause name 0 (A.RHS e mc) $ do
-  reportSDoc "tc.def.alias" 10 $ text "checkAlias" <+> vcat
-    [ text (prettyShow name) <+> colon  <+> prettyTCM t'
+checkAlias t ai delayed i name e mc = atClause name 0 (A.RHS e mc) $ do
+  reportSDoc "tc.def.alias" 10 $ "checkAlias" <+> vcat
+    [ text (prettyShow name) <+> colon  <+> prettyTCM t
     , text (prettyShow name) <+> equals <+> prettyTCM e
     ]
 
-{-
-  -- Infer the type of the rhs
-  (v, t) <- applyRelevanceToContext (getRelevance ai) $
-                                    inferOrCheck e (Just t')
-  -- v <- coerce v t t'
--}
+  -- Infer the type of the rhs.
+  -- Andreas, 2018-06-09, issue #2170.
+  -- The context will only be resurrected if we have --irrelevant-projections.
+  v <- applyModalityToContextFunBody ai $ checkDontExpandLast CmpLeq e t
 
-  -- Infer the type of the rhs
-  v <- applyRelevanceToContext (getRelevance ai) $ checkDontExpandLast e t'
-  let t = t'
-
-  reportSDoc "tc.def.alias" 20 $ text "checkAlias: finished checking"
+  reportSDoc "tc.def.alias" 20 $ "checkAlias: finished checking"
 
   solveSizeConstraints DontDefaultToInfty
 
@@ -178,6 +181,7 @@ checkAlias t' ai delayed i name e mc = atClause name 0 (A.RHS e mc) $ do
                           , clauseUnreachable = Just False
                           } ]
                       , funCompiled = Just $ Done [] $ bodyMod v
+                      , funSplitTree = Just $ SplittingDone 0
                       , funDelayed  = delayed
                       , funAbstr    = Info.defAbstract i
                       }
@@ -187,7 +191,7 @@ checkAlias t' ai delayed i name e mc = atClause name 0 (A.RHS e mc) $ do
   when (Info.defInstance i == InstanceDef) $ do
     addTypedInstance name t
 
-  reportSDoc "tc.def.alias" 20 $ text "checkAlias: leaving"
+  reportSDoc "tc.def.alias" 20 $ "checkAlias: leaving"
 
 
 -- | Type check a definition by pattern matching.
@@ -218,20 +222,20 @@ checkFunDefS :: Type             -- ^ the type we expect the function to have
              -> TCM ()
 checkFunDefS t ai delayed extlam with i name withSub cs = do
 
-    traceCall (CheckFunDefCall (getRange i) (qnameName name) cs) $ do   -- TODO!! (qnameName)
+    traceCall (CheckFunDefCall (getRange i) name cs) $ do
         reportSDoc "tc.def.fun" 10 $
-          sep [ text "checking body of" <+> prettyTCM name
-              , nest 2 $ text ":" <+> prettyTCM t
-              , nest 2 $ text "full type:" <+> (prettyTCM . defType =<< getConstInfo name)
+          sep [ "checking body of" <+> prettyTCM name
+              , nest 2 $ ":" <+> prettyTCM t
+              , nest 2 $ "full type:" <+> (prettyTCM . defType =<< getConstInfo name)
               ]
 
         reportSDoc "tc.def.fun" 70 $
-          sep $ [ text "clauses:" ] ++ map (nest 2 . text . show . A.deepUnscope) cs
+          sep $ [ "clauses:" ] ++ map (nest 2 . text . show . A.deepUnscope) cs
 
         cs <- return $ map A.lhsToSpine cs
 
         reportSDoc "tc.def.fun" 70 $
-          sep $ [ text "spine clauses:" ] ++ map (nest 2 . text . show . A.deepUnscope) cs
+          sep $ [ "spine clauses:" ] ++ map (nest 2 . text . show . A.deepUnscope) cs
 
         -- Ensure that all clauses have the same number of trailing hidden patterns
         -- This is necessary since trailing implicits are no longer eagerly inserted.
@@ -244,7 +248,7 @@ checkFunDefS t ai delayed extlam with i name withSub cs = do
         cs <- traceCall NoHighlighting $ do -- To avoid flicker.
           forM (zip cs [0..]) $ \ (c, clauseNo) -> do
             atClause name clauseNo (A.clauseRHS c) $ do
-              c <- applyRelevanceToContext (getRelevance ai) $ do
+              (c,b) <- applyModalityToContextFunBody ai $ do
                 checkClause t withSub c
               -- Andreas, 2013-11-23 do not solve size constraints here yet
               -- in case we are checking the body of an extended lambda.
@@ -255,24 +259,61 @@ checkFunDefS t ai delayed extlam with i name withSub cs = do
               -- Andreas, 2013-10-27 add clause as soon it is type-checked
               -- TODO: instantiateFull?
               inTopContext $ addClauses name [c]
-              return c
+              return (c,b)
+
+        (cs, CPC isOneIxs) <- return $ (second mconcat . unzip) cs
+
+        let isSystem = not . null $ isOneIxs
+
+        canBeSystem <- do
+          -- allow VarP and ConP i0/i1 fallThrough = yes, DotP
+          let pss = map namedClausePats cs
+              allowed p = case p of
+                VarP{} -> True
+                -- pattern inserted by splitPartial
+                ConP _ cpi [] | conPFallThrough cpi -> True
+                DotP{} -> True
+                _      -> False
+          return $! all (allowed . namedArg) (concat pss)
+        when isSystem $ unless canBeSystem $
+          typeError $ GenericError "no pattern matching or path copatterns in systems!"
+
 
         reportSDoc "tc.def.fun" 70 $ inTopContext $ do
-          sep $ [ text "checked clauses:" ] ++ map (nest 2 . text . show) cs
+          sep $ [ "checked clauses:" ] ++ map (nest 2 . text . show) cs
 
         -- After checking, remove the clauses again.
         -- (Otherwise, @checkInjectivity@ loops for issue 801).
         modifyFunClauses name (const [])
 
         reportSDoc "tc.cc" 25 $ inTopContext $ do
-          sep [ text "clauses before injectivity test"
+          sep [ "clauses before injectivity test"
               , nest 2 $ prettyTCM $ map (QNamed name) cs  -- broken, reify (QNamed n cl) expect cl to live at top level
               ]
         reportSDoc "tc.cc" 60 $ inTopContext $ do
-          sep [ text "raw clauses: "
+          sep [ "raw clauses: "
               , nest 2 $ sep $ map (text . show . QNamed name) cs
               ]
 
+
+        -- Systems have their own coverage and "coherence" check, we
+        -- also add an absurd clause for the cases not needed.
+        (cs,sys) <- if not isSystem then return (cs, Nothing) else do
+                 fullType <- flip abstract t <$> getContextTelescope
+                 sys <- inTopContext $ checkSystemCoverage name isOneIxs fullType cs
+                 tel <- getContextTelescope
+                 let c = Clause
+                       { clauseFullRange = noRange
+                       , clauseLHSRange  = noRange
+                       , clauseTel = tel
+                       , namedClausePats = teleNamedArgs tel
+                       , clauseBody = Nothing
+                       , clauseType = Just (defaultArg t)
+                       , clauseCatchall = False
+                       , clauseUnreachable = Just False
+                       }
+                 return (cs ++ [c], Just sys)
+
         -- Annotate the clauses with which arguments are actually used.
         cs <- instantiateFull {- =<< mapM rebindClause -} cs
         -- Andreas, 2010-11-12
@@ -290,31 +331,44 @@ checkFunDefS t ai delayed extlam with i name withSub cs = do
           checkInjectivity name cs
 
         reportSDoc "tc.cc" 15 $ inTopContext $ do
-          sep [ text "clauses before compilation"
+          sep [ "clauses before compilation"
               , nest 2 $ sep $ map (prettyTCM . QNamed name) cs
               ]
 
+        reportSDoc "tc.cc.raw" 65 $ do
+          sep [ "clauses before compilation"
+              , nest 2 $ sep $ map (text . show) cs
+              ]
+
         -- add clauses for the coverage checker (needs to reduce)
         inTopContext $ addClauses name cs
 
+        reportSDoc "tc.cc.type" 60 $ "  type   : " <+> (text . prettyShow) t
+        reportSDoc "tc.cc.type" 60 $ "  context: " <+> (text . prettyShow =<< getContextTelescope)
+
         fullType <- flip telePi t <$> getContextTelescope
 
+        reportSLn  "tc.cc.type" 80 $ show fullType
+
         -- Coverage check and compile the clauses
-        cc <- Bench.billTo [Bench.Coverage] $
-          inTopContext $ compileClauses (Just (name, fullType)) cs
+        (mst, cc) <- Bench.billTo [Bench.Coverage] $
+          inTopContext $ compileClauses (if isSystem then Nothing else (Just (name, fullType)))
+                                        cs
 
         -- Clause compilation runs the coverage checker, which might add
         -- some extra clauses.
         cs <- defClauses <$> getConstInfo name
 
         reportSDoc "tc.cc" 60 $ inTopContext $ do
-          sep [ text "compiled clauses of" <+> prettyTCM name
+          sep [ "compiled clauses of" <+> prettyTCM name
               , nest 2 $ pretty cc
               ]
 
         -- The macro tag might be on the type signature
         ismacro <- isMacro . theDef <$> getConstInfo name
 
+        covering <- funCovering . theDef <$> getConstInfo name
+
         -- Add the definition
         inTopContext $ addConstant name =<< do
           -- If there was a pragma for this definition, we can set the
@@ -324,17 +378,19 @@ checkFunDefS t ai delayed extlam with i name withSub cs = do
              emptyFunction
              { funClauses        = cs
              , funCompiled       = Just cc
+             , funSplitTree      = mst
              , funDelayed        = delayed
              , funInv            = inv
              , funAbstr          = Info.defAbstract i
-             , funExtLam         = extlam
+             , funExtLam         = (\ e -> e { extLamSys = sys }) <$> extlam
              , funWith           = with
              , funCopatternLHS   = hasProjectionPatterns cc
+             , funCovering       = covering
              }
           useTerPragma $ defaultDefn ai name fullType defn
 
         reportSDoc "tc.def.fun" 10 $ do
-          sep [ text "added " <+> prettyTCM name <+> text ":"
+          sep [ "added " <+> prettyTCM name <+> ":"
               , nest 2 $ prettyTCM . defType =<< getConstInfo name
               ]
 
@@ -342,7 +398,7 @@ checkFunDefS t ai delayed extlam with i name withSub cs = do
 --   which comes from a possible termination pragma.
 useTerPragma :: Definition -> TCM Definition
 useTerPragma def@Defn{ defName = name, theDef = fun@Function{}} = do
-  tc <- asks envTerminationCheck
+  tc <- asksTC envTerminationCheck
   let terminates = case tc of
         NonTerminating -> Just False
         Terminating    -> Just True
@@ -394,24 +450,146 @@ data WithFunctionProblem
     , wfClauses    :: [A.Clause]           -- ^ The given clauses for the with function
     }
 
+checkSystemCoverage
+  :: QName
+  -> [Int]
+  -> Type
+  -> [Clause]
+  -> TCM System
+checkSystemCoverage f [n] t cs = do
+  reportSDoc "tc.sys.cover" 10 $ text (show (n , length cs)) <+> prettyTCM t
+  TelV gamma t <- telViewUpTo n t
+  addContext gamma $ do
+  TelV (ExtendTel a _) _ <- telViewUpTo 1 t
+  a <- reduce $ unEl $ unDom a
+
+  case a of
+    Def q [Apply phi] -> do
+      [iz,io] <- mapM getBuiltinName' [builtinIZero, builtinIOne]
+      ineg <- primINeg
+      imin <- primIMin
+      imax <- primIMax
+      i0 <- primIZero
+      i1 <- primIOne
+      let
+        isDir (ConP q _ []) | Just (conName q) == iz = Just False
+        isDir (ConP q _ []) | Just (conName q) == io = Just True
+        isDir _ = Nothing
+
+        collectDirs :: [Int] -> [DeBruijnPattern] -> [(Int,Bool)]
+        collectDirs [] [] = []
+        collectDirs (i : is) (p : ps) | Just d <- isDir p = (i,d) : collectDirs is ps
+                                      | otherwise         = collectDirs is ps
+        collectDirs _ _ = __IMPOSSIBLE__
+
+        dir :: (Int,Bool) -> Term
+        dir (i,False) = ineg `apply` [argN $ var i]
+        dir (i,True) = var i
+
+        -- andI and orI have cases for singletons to improve error messages.
+        andI, orI :: [Term] -> Term
+        andI [] = i1
+        andI [t] = t
+        andI (t:ts) = (\ x -> imin `apply` [argN t, argN x]) $ andI ts
+
+        orI [] = i0
+        orI [t] = t
+        orI (t:ts) = imax `apply` [argN t, argN (orI ts)]
+
+      let
+        pats = map (take n . map (namedThing . unArg) . namedClausePats) cs
+        alphas :: [[(Int,Bool)]] -- the face maps corresponding to each clause
+        alphas = map (collectDirs (downFrom n)) pats
+        phis :: [Term] -- the φ terms for each clause (i.e. the alphas as terms)
+        phis = map andI $ map (map dir) alphas
+        psi = orI $ phis
+        pcs = zip phis cs
+        boolToI True = i1
+        boolToI False = i0
+
+      reportSDoc "tc.sys.cover" 20 $ fsep $ map prettyTCM pats
+      interval <- elInf primInterval
+      reportSDoc "tc.sys.cover" 10 $ "equalTerm " <+> prettyTCM (unArg phi) <+> prettyTCM psi
+      equalTerm interval (unArg phi) psi
+
+      forM_ (init $ init $ List.tails pcs) $ \ ((phi1,cl1):pcs') -> do
+        forM_ pcs' $ \ (phi2,cl2) -> do
+          phi12 <- reduce (imin `apply` [argN phi1, argN phi2])
+          forallFaceMaps phi12 (\ _ _ -> __IMPOSSIBLE__) $ \ sigma -> do
+            let args = sigma `applySubst` teleArgs gamma
+                t' = sigma `applySubst` t
+                fromReduced (YesReduction _ x) = x
+                fromReduced (NoReduction x) = ignoreBlocking x
+                body cl = do
+                  let extra = length (drop n $ namedClausePats cl)
+                  TelV delta _ <- telViewUpTo extra t'
+                  fmap (abstract delta) $ addContext delta $ do
+                    fmap fromReduced $ runReduceM $
+                      appDef' (Def f []) [cl] [] (map notReduced $ raise (size delta) args ++ teleArgs delta)
+            v1 <- body cl1
+            v2 <- body cl2
+            equalTerm t' v1 v2
+
+      sys <- forM (zip alphas cs) $ \ (alpha,cl) -> do
+
+            let
+                -- Δ = Γ_α , Δ'α
+                delta = clauseTel cl
+                -- Δ ⊢ b
+                Just b = clauseBody cl
+                -- Δ ⊢ ps : Γ , o : [φ] , Δ'
+                -- we assume that there's no pattern matching other
+                -- than from the system
+                ps = namedClausePats cl
+                extra = length (drop (size gamma + 1) ps)
+                -- size Δ'α = size Δ' = extra
+                -- Γ , α ⊢ u
+                takeLast n xs = drop (length xs - n) xs
+                weak [] = idS
+                weak (i:is) = weak is `composeS` liftS i (raiseS 1)
+                tel = telFromList (takeLast extra (telToList delta))
+                u = abstract tel (liftS extra (weak $ List.sort $ map fst alpha) `applySubst` b)
+            return (map (first var) alpha,u)
+
+      reportSDoc "tc.sys.cover.sys" 20 $ fsep $ prettyTCM gamma : map prettyTCM sys
+      reportSDoc "tc.sys.cover.sys" 40 $ fsep $ (text . show) gamma : map (text . show) sys
+      return (System gamma sys) -- gamma uses names from the type, not the patterns, could we do better?
+    _ -> __IMPOSSIBLE__
+checkSystemCoverage _ _ t cs = __IMPOSSIBLE__
+
+
+-- * Info that is needed after all clauses have been processed.
+data ClausesPostChecks = CPC
+    { cpcPartialSplits :: [Int]
+      -- ^ Which argument indexes have a partial split.
+    }
+
+instance Semigroup ClausesPostChecks where
+  (<>) (CPC xs) (CPC xs') = CPC (List.nub $ mappend xs xs')
+
+instance Monoid ClausesPostChecks where
+  mempty  = CPC []
+  mappend = (<>)
+
 -- | Type check a function clause.
 
 checkClause
   :: Type          -- ^ Type of function defined by this clause.
   -> Maybe Substitution  -- ^ Module parameter substitution arising from with-abstraction.
   -> A.SpineClause -- ^ Clause.
-  -> TCM Clause    -- ^ Type-checked clause.
+  -> TCM (Clause,ClausesPostChecks)  -- ^ Type-checked clause
 
 checkClause t withSub c@(A.Clause (A.SpineLHS i x aps) strippedPats rhs0 wh catchall) = do
-    reportSDoc "tc.lhs.top" 30 $ text "Checking clause" $$ prettyA c
+    reportSDoc "tc.lhs.top" 30 $ "Checking clause" $$ prettyA c
     unlessNull (trailingWithPatterns aps) $ \ withPats -> do
       typeError $ UnexpectedWithPatterns $ map namedArg withPats
     traceCall (CheckClause t c) $ do
       aps <- expandPatternSynonyms aps
       cxtNames <- reverse . map (fst . unDom) <$> getContext
       when (not $ null strippedPats) $ reportSDoc "tc.lhs.top" 50 $
-        text "strippedPats:" <+> vcat [ prettyA p <+> text "=" <+> prettyTCM v <+> text ":" <+> prettyTCM a | A.ProblemEq p v a <- strippedPats ]
-      checkLeftHandSide (CheckPatternShadowing c) (Just x) aps t withSub strippedPats $ \ lhsResult@(LHSResult npars delta ps absurdPat trhs patSubst asb) -> do
+        "strippedPats:" <+> vcat [ prettyA p <+> "=" <+> prettyTCM v <+> ":" <+> prettyTCM a | A.ProblemEq p v a <- strippedPats ]
+      closed_t <- flip abstract t <$> getContextTelescope
+      checkLeftHandSide (CheckPatternShadowing c) (Just x) aps t withSub strippedPats $ \ lhsResult@(LHSResult npars delta ps absurdPat trhs patSubst asb psplit) -> do
         -- Note that we might now be in irrelevant context,
         -- in case checkLeftHandSide walked over an irrelevant projection pattern.
 
@@ -445,27 +623,51 @@ checkClause t withSub c@(A.Clause (A.SpineLHS i x aps) strippedPats rhs0 wh catc
         -- Note that the with function doesn't necessarily share any part of
         -- the context with the parent (but withSub will take you from parent
         -- to child).
+
         inTopContext $ Bench.billTo [Bench.Typing, Bench.With] $ checkWithFunction cxtNames with
 
+        whenM (optDoubleCheck <$> pragmaOptions) $ case body of
+          Just v  -> do
+            reportSDoc "tc.lhs.top" 30 $ vcat
+              [ "double checking rhs"
+              , nest 2 (prettyTCM v <+> " : " <+> prettyTCM (unArg trhs))
+              ]
+            noConstraints $ dontAssignMetas $ checkInternal v $ unArg trhs
+          Nothing -> return ()
+
         reportSDoc "tc.lhs.top" 10 $ vcat
-          [ text "Clause before translation:"
+          [ "Clause before translation:"
           , nest 2 $ vcat
-            [ text "delta =" <+> do escapeContext (size delta) $ prettyTCM delta
-            , text "ps    =" <+> do P.fsep <$> prettyTCMPatterns ps
-            , text "body  =" <+> maybe (text "_|_") prettyTCM body
+            [ "delta =" <+> do escapeContext (size delta) $ prettyTCM delta
+            , "ps    =" <+> do P.fsep <$> prettyTCMPatterns ps
+            , "body  =" <+> maybe "_|_" prettyTCM body
+            , "type  =" <+> prettyTCM t
             ]
           ]
 
         reportSDoc "tc.lhs.top" 60 $ escapeContext (size delta) $ vcat
-          [ text "Clause before translation (raw):"
+          [ "Clause before translation (raw):"
           , nest 2 $ vcat
-            [ text "ps    =" <+> text (show ps)
-            , text "body  =" <+> text (show body)
+            [ "ps    =" <+> text (show ps)
+            , "body  =" <+> text (show body)
+            , "type  =" <+> text (show t)
             ]
           ]
 
+        -- check naturality wrt the interval.
+        let
+          iApplyVars :: [NamedArg DeBruijnPattern] -> [(Int, (Term,Term))]
+          iApplyVars ps = flip concatMap (map namedArg ps) $ \case
+                             IApplyP _ t u x -> [(dbPatVarIndex x,(t,u))]
+                             VarP{} -> []
+                             ProjP{}-> []
+                             LitP{} -> []
+                             DotP{} -> []
+                             DefP _ _ ps -> iApplyVars ps
+                             ConP _ _ ps -> iApplyVars ps
+
         -- compute body modification for irrelevant definitions, see issue 610
-        rel <- asks envRelevance
+        rel <- asksTC envRelevance
         let bodyMod body = case rel of
               Irrelevant -> dontCare <$> body
               _          -> body
@@ -474,7 +676,7 @@ checkClause t withSub c@(A.Clause (A.SpineLHS i x aps) strippedPats rhs0 wh catc
         -- treat them as catchalls.
         let catchall' = catchall || isNothing body
 
-        return $
+        return $ (, CPC psplit)
           Clause { clauseLHSRange  = getRange i
                  , clauseFullRange = getRange c
                  , clauseTel       = killRange delta
@@ -496,16 +698,20 @@ checkRHS
   -> A.RHS                   -- ^ Rhs to check.
   -> TCM (Maybe Term, WithFunctionProblem)
                                               -- Note: the as-bindings are already bound (in checkClause)
-checkRHS i x aps t lhsResult@(LHSResult _ delta ps absurdPat trhs _ _asb) rhs0 = handleRHS rhs0
+checkRHS i x aps t lhsResult@(LHSResult _ delta ps absurdPat trhs _ _asb _) rhs0 = handleRHS rhs0
   where
   handleRHS rhs =
     case rhs of
 
       -- Case: ordinary RHS
       A.RHS e _ -> Bench.billTo [Bench.Typing, Bench.CheckRHS] $ do
-        when absurdPat $ typeError $ AbsurdPatternRequiresNoRHS aps
-        v <- checkExpr e $ unArg trhs
-        return (Just v, NoWithFunction)
+        -- If there is an absurd pattern, we do not need a RHS. If we have
+        -- one we complain, ignore it and return the same @(Nothing, NoWithFunction)@
+        -- as the case dealing with @A.AbsurdRHS@.
+        mv <- if absurdPat
+              then Nothing <$ setCurrentRange rhs (warning $ AbsurdPatternRequiresNoRHS ps)
+              else Just <$> checkExpr e (unArg trhs)
+        return (mv, NoWithFunction)
 
       -- Case: no RHS
       A.AbsurdRHS -> do
@@ -524,16 +730,16 @@ checkRHS i x aps t lhsResult@(LHSResult _ delta ps absurdPat trhs _ _asb) rhs0 =
         -- a futile rewrite with a reflexive equation.
         -- Thus, we restore the state in this case,
         -- unless the rewrite expression contains questionmarks.
-        st <- get
+        st <- getTC
         let recurse = do
-             st' <- get
+             st' <- getTC
              -- Comparing the whole stInteractionPoints maps is a bit
              -- wasteful, but we assume
              -- 1. rewriting with a reflexive equality to happen rarely,
              -- 2. especially with ?-holes in the rewrite expression
              -- 3. and a large overall number of ?s.
              let sameIP = (==) `on` (^.stInteractionPoints)
-             when (sameIP st st') $ put st
+             when (sameIP st st') $ putTC st
              handleRHS $ A.RewriteRHS qes strippedPats rhs wh
 
         -- Get value and type of rewrite-expression.
@@ -555,7 +761,7 @@ checkRHS i x aps t lhsResult@(LHSResult _ delta ps absurdPat trhs _ _asb) rhs0 =
             return (eqt, El s dom, unArg a, unArg b)
             -- Note: the sort _s of the equality need not be the sort of the type @dom@!
           OtherType{} -> typeError . GenericDocError =<< do
-            text "Cannot rewrite by equation of type" <+> prettyTCM t'
+            "Cannot rewrite by equation of type" <+> prettyTCM t'
 
         -- Get the name of builtin REFL.
 
@@ -579,7 +785,7 @@ checkRHS i x aps t lhsResult@(LHSResult _ delta ps absurdPat trhs _ _asb) rhs0 =
         -- Process 'rewrite' clause like a suitable 'with' clause.
 
         -- The REFL constructor might have an argument
-        let reflPat  = A.ConP (ConPatInfo ConOCon patNoRange False) (unambiguous $ conName reflCon) $
+        let reflPat  = A.ConP (ConPatInfo ConOCon patNoRange ConPatEager) (unambiguous $ conName reflCon) $
               maybeToList $ fmap (\ ai -> Arg ai $ unnamed $ A.WildP patNoRange) reflInfo
 
         -- Andreas, 2015-12-25  Issue #1740:
@@ -602,21 +808,21 @@ checkRHS i x aps t lhsResult@(LHSResult _ delta ps absurdPat trhs _ _asb) rhs0 =
             cl = A.Clause (A.LHS i $ insertPatternsLHSCore pats $ A.LHSHead x $ killRange aps)
                    strippedPats rhs'' outerWhere False
         reportSDoc "tc.rewrite" 60 $ vcat
-          [ text "rewrite"
-          , text "  rhs' = " <> (text . show) rhs'
+          [ "rewrite"
+          , "  rhs' = " Pr.<> (text . show) rhs'
           ]
         checkWithRHS x qname t lhsResult [withExpr] [withType] [cl]
 
       -- Case: @with@
       A.WithRHS aux es cs -> do
         reportSDoc "tc.with.top" 15 $ vcat
-          [ text "TC.Rules.Def.checkclause reached A.WithRHS"
+          [ "TC.Rules.Def.checkclause reached A.WithRHS"
           , sep $ prettyA aux : map (parens . prettyA) es
           ]
         reportSDoc "tc.with.top" 20 $ do
           nfv <- getCurrentModuleFreeVars
           m   <- currentModule
-          sep [ text "with function module:" <+>
+          sep [ "with function module:" <+>
                 prettyList (map prettyTCM $ mnameToList m)
               ,  text $ "free variables: " ++ show nfv
               ]
@@ -641,7 +847,7 @@ checkWithRHS
   -> [A.Clause]              -- ^ With-clauses to check.
   -> TCM (Maybe Term, WithFunctionProblem)
                                 -- Note: as-bindings already bound (in checkClause)
-checkWithRHS x aux t (LHSResult npars delta ps _absurdPat trhs _ _asb) vs0 as cs = Bench.billTo [Bench.Typing, Bench.With] $ do
+checkWithRHS x aux t (LHSResult npars delta ps _absurdPat trhs _ _asb _) vs0 as cs = Bench.billTo [Bench.Typing, Bench.With] $ do
         let withArgs = withArguments vs0 as
             perm = fromMaybe __IMPOSSIBLE__ $ dbPatPerm ps
         (vs, as)  <- normalise (vs0, as)
@@ -649,12 +855,12 @@ checkWithRHS x aux t (LHSResult npars delta ps _absurdPat trhs _ _asb) vs0 as cs
         -- Andreas, 2012-09-17: for printing delta,
         -- we should remove it from the context first
         reportSDoc "tc.with.top" 25 $ escapeContext (size delta) $ vcat
-          [ text "delta  =" <+> prettyTCM delta
+          [ "delta  =" <+> prettyTCM delta
           ]
         reportSDoc "tc.with.top" 25 $ vcat
-          [ text "vs     =" <+> prettyTCM vs
-          , text "as     =" <+> prettyTCM as
-          , text "perm   =" <+> text (show perm)
+          [ "vs     =" <+> prettyTCM vs
+          , "as     =" <+> prettyTCM as
+          , "perm   =" <+> text (show perm)
           ]
 
         -- Split the telescope into the part needed to type the with arguments
@@ -668,12 +874,12 @@ checkWithRHS x aux t (LHSResult npars delta ps _absurdPat trhs _ _asb) vs0 as cs
         -- Andreas, 2012-09-17: for printing delta,
         -- we should remove it from the context first
         reportSDoc "tc.with.top" 25 $ escapeContext (size delta) $ vcat
-          [ text "delta1 =" <+> prettyTCM delta1
-          , text "delta2 =" <+> addContext delta1 (prettyTCM delta2)
+          [ "delta1 =" <+> prettyTCM delta1
+          , "delta2 =" <+> addContext delta1 (prettyTCM delta2)
           ]
         reportSDoc "tc.with.top" 25 $ vcat
-          [ text "perm'  =" <+> text (show perm')
-          , text "fPerm  =" <+> text (show finalPerm)
+          [ "perm'  =" <+> text (show perm')
+          , "fPerm  =" <+> text (show finalPerm)
           ]
 
         -- Create the body of the original function
@@ -690,25 +896,25 @@ checkWithRHS x aux t (LHSResult npars delta ps _absurdPat trhs _ _asb) vs0 as cs
             v    = Def aux $ map Apply $ us0 ++ us1 ++ map defaultArg withArgs ++ us2
         -- Andreas, 2013-02-26 add with-name to signature for printing purposes
         addConstant aux =<< do
-          useTerPragma $ defaultDefn defaultArgInfo aux typeDontCare emptyFunction
+          useTerPragma $ defaultDefn defaultArgInfo aux __DUMMY_TYPE__ emptyFunction
 
         -- Andreas, 2013-02-26 separate msgs to see which goes wrong
         reportSDoc "tc.with.top" 20 $
-          text "    with arguments" <+> do escapeContext (size delta) $ addContext delta1 $ prettyList (map prettyTCM vs)
+          "    with arguments" <+> do escapeContext (size delta) $ addContext delta1 $ prettyList (map prettyTCM vs)
         reportSDoc "tc.with.top" 20 $
-          text "             types" <+> do escapeContext (size delta) $ addContext delta1 $ prettyList (map prettyTCM as)
+          "             types" <+> do escapeContext (size delta) $ addContext delta1 $ prettyList (map prettyTCM as)
         reportSDoc "tc.with.top" 20 $
-          text "with function call" <+> prettyTCM v
+          "with function call" <+> prettyTCM v
         reportSDoc "tc.with.top" 20 $
-          text "           context" <+> (prettyTCM =<< getContextTelescope)
+          "           context" <+> (prettyTCM =<< getContextTelescope)
         reportSDoc "tc.with.top" 20 $
-          text "             delta" <+> do escapeContext (size delta) $ prettyTCM delta
+          "             delta" <+> do escapeContext (size delta) $ prettyTCM delta
         reportSDoc "tc.with.top" 20 $
-          text "            delta1" <+> do escapeContext (size delta) $ prettyTCM delta1
+          "            delta1" <+> do escapeContext (size delta) $ prettyTCM delta1
         reportSDoc "tc.with.top" 20 $
-          text "            delta2" <+> do escapeContext (size delta) $ addContext delta1 $ prettyTCM delta2
+          "            delta2" <+> do escapeContext (size delta) $ addContext delta1 $ prettyTCM delta2
         reportSDoc "tc.with.top" 20 $
-          text "              body" <+> prettyTCM v
+          "              body" <+> prettyTCM v
 
         return (Just v, WithFunction x aux t delta delta1 delta2 vs as t' ps npars perm' perm finalPerm cs)
 
@@ -722,19 +928,19 @@ checkWithFunction cxtNames (WithFunction f aux t delta delta1 delta2 vs as b qs
       withSub = liftS (size delta2) (wkS (countWithArgs as) idS) `composeS` renaming __IMPOSSIBLE__ (reverseP perm')
 
   reportSDoc "tc.with.top" 10 $ vcat
-    [ text "checkWithFunction"
+    [ "checkWithFunction"
     , nest 2 $ vcat
-      [ text "delta1 =" <+> prettyTCM delta1
-      , text "delta2 =" <+> addContext delta1 (prettyTCM delta2)
-      , text "t      =" <+> prettyTCM t
-      , text "as     =" <+> addContext delta1 (prettyTCM as)
-      , text "vs     =" <+> do addContext delta1 $ prettyTCM vs
-      , text "b      =" <+> do addContext delta1 $ addContext delta2 $ prettyTCM b
-      , text "qs     =" <+> do addContext delta $ prettyTCMPatternList qs
-      , text "perm'  =" <+> text (show perm')
-      , text "perm   =" <+> text (show perm)
-      , text "fperm  =" <+> text (show finalPerm)
-      , text "withSub=" <+> text (show withSub)
+      [ "delta1 =" <+> prettyTCM delta1
+      , "delta2 =" <+> addContext delta1 (prettyTCM delta2)
+      , "t      =" <+> prettyTCM t
+      , "as     =" <+> addContext delta1 (prettyTCM as)
+      , "vs     =" <+> do addContext delta1 $ prettyTCM vs
+      , "b      =" <+> do addContext delta1 $ addContext delta2 $ prettyTCM b
+      , "qs     =" <+> do addContext delta $ prettyTCMPatternList qs
+      , "perm'  =" <+> text (show perm')
+      , "perm   =" <+> text (show perm)
+      , "fperm  =" <+> text (show finalPerm)
+      , "withSub=" <+> text (show withSub)
       ]
     ]
 
@@ -746,20 +952,15 @@ checkWithFunction cxtNames (WithFunction f aux t delta delta1 delta2 vs as b qs
                              -- We normalise to get rid of Def's coming
                              -- from module applications.
   (withFunType, n) <- withFunctionType delta1 vs as delta2 b
-  reportSDoc "tc.with.type" 10 $ sep [ text "with-function type:", nest 2 $ prettyTCM withFunType ]
-  reportSDoc "tc.with.type" 50 $ sep [ text "with-function type:", nest 2 $ pretty withFunType ]
+  reportSDoc "tc.with.type" 10 $ sep [ "with-function type:", nest 2 $ prettyTCM withFunType ]
+  reportSDoc "tc.with.type" 50 $ sep [ "with-function type:", nest 2 $ pretty withFunType ]
 
   -- Andreas, 2013-10-21
   -- Check generated type directly in internal syntax.
-  setCurrentRange cs
-    (traceCall NoHighlighting $   -- To avoid flicker.
-     checkType withFunType)
-    `catchError` \err -> case err of
-      TypeError s e -> do
-        put s
-        wt <- reify withFunType
-        enterClosure e $ traceCall (CheckWithFunctionType wt) . typeError
-      err           -> throwError err
+  setCurrentRange cs $
+    traceCall NoHighlighting $   -- To avoid flicker.
+    traceCall (CheckWithFunctionType withFunType) $
+    checkType withFunType
 
   -- With display forms are closed
   df <- safeInTopContext $ makeOpen =<< withDisplayForm f aux delta1 delta2 n qs perm' perm
@@ -768,7 +969,7 @@ checkWithFunction cxtNames (WithFunction f aux t delta delta1 delta2 vs as b qs
 
   case dget df of
     Display n ts dt ->
-      reportSDoc "tc.with.top" 20 $ text "Display" <+> fsep
+      reportSDoc "tc.with.top" 20 $ "Display" <+> fsep
         [ text (show n)
         , prettyList $ map prettyTCM ts
         , prettyTCM dt
@@ -779,9 +980,9 @@ checkWithFunction cxtNames (WithFunction f aux t delta delta1 delta2 vs as b qs
   -- solveSizeConstraints -- Andreas, 2012-10-16 does not seem necessary
 
   reportSDoc "tc.with.top" 10 $ sep
-    [ text "added with function" <+> (prettyTCM aux) <+> text "of type"
+    [ "added with function" <+> (prettyTCM aux) <+> "of type"
     , nest 2 $ prettyTCM withFunType
-    , nest 2 $ text "-|" <+> (prettyTCM =<< getContextTelescope)
+    , nest 2 $ "-|" <+> (prettyTCM =<< getContextTelescope)
     ]
   reportSDoc "tc.with.top" 70 $ vcat
     [ nest 2 $ text $ "raw with func. type = " ++ show withFunType
@@ -812,7 +1013,7 @@ checkWhere wh@(A.WhereDecls whmod ds) ret = do
       [] -> ret
       [A.ScopedDecl scope ds] -> withScope_ scope $ loop ds
       [A.Section _ m tel ds]  -> newSection m tel $ do
-          local (\ e -> e { envCheckingWhere = True }) $ do
+          localTC (\ e -> e { envCheckingWhere = True }) $ do
             checkDecls ds
             ret
       _ -> __IMPOSSIBLE__
@@ -845,22 +1046,22 @@ checkWhere wh@(A.WhereDecls whmod ds) ret = do
 
 -- | Enter a new section during type-checking.
 
-newSection :: ModuleName -> A.Telescope -> TCM a -> TCM a
-newSection m tel cont = do
+newSection :: ModuleName -> A.GeneralizeTelescope -> TCM a -> TCM a
+newSection m gtel@(A.GeneralizeTel _ tel) cont = do
   reportSDoc "tc.section" 10 $
-    text "checking section" <+> prettyTCM m <+> fsep (map prettyAs tel)
+    "checking section" <+> prettyTCM m <+> fsep (map prettyA tel)
 
-  checkTelescope tel $ \ tel' -> do
+  checkGeneralizeTelescope gtel $ \ _ tel' -> do
     reportSDoc "tc.section" 10 $
-      text "adding section:" <+> prettyTCM m <+> text (show (size tel'))
+      "adding section:" <+> prettyTCM m <+> text (show (size tel'))
 
     addSection m
 
     reportSDoc "tc.section" 10 $ inTopContext $
-      nest 4 $ text "actual tele:" <+> do prettyTCM =<< lookupSection m
+      nest 4 $ "actual tele:" <+> do prettyTCM =<< lookupSection m
 
     withCurrentModule m cont
 
 -- | Set the current clause number.
 atClause :: QName -> Int -> A.RHS -> TCM a -> TCM a
-atClause name i rhs = local $ \ e -> e { envClause = IPClause name i rhs }
+atClause name i rhs = localTC $ \ e -> e { envClause = IPClause name i rhs }
