@@ -41,7 +41,7 @@ implicitP info = Arg (setOrigin Inserted info) $ unnamed $ A.WildP $ PatRange $
 insertImplicitPatterns :: ExpandHidden -> [NamedArg A.Pattern] ->
                           Telescope -> TCM [NamedArg A.Pattern]
 insertImplicitPatterns exh ps tel =
-  insertImplicitPatternsT exh ps (telePi tel typeDontCare)
+  insertImplicitPatternsT exh ps (telePi tel __DUMMY_TYPE__)
 
 -- | Insert trailing SizeLt patterns, if any.
 insertImplicitSizeLtPatterns :: Type -> TCM [NamedArg A.Pattern]
@@ -55,10 +55,10 @@ insertImplicitSizeLtPatterns t = do
 
   -- Search for the last SizeLt type among the hidden arguments.
   TelV tel _ <- telView t
-  let ts = reverse $ takeWhile (not . visible) $ telToList tel
-  keep <- reverse <$> dropWhileM (not <.> isSizeLt . snd . unDom) ts
+  let ts = takeWhile (not . visible) $ telToList tel
+  keep <- dropWhileEndM (not <.> isSizeLt . snd . unDom) ts
   -- Insert implicit patterns upto (including) the last SizeLt type.
-  return [ implicitP ai | Dom ai _ <- keep ]
+  return $ map (implicitP . domInfo) keep
 
 -- | Insert implicit patterns in a list of patterns.
 --   Even if 'DontExpandLast', trailing SIZELT patterns are inserted.
@@ -68,36 +68,41 @@ insertImplicitPatternsT DontExpandLast [] a = insertImplicitSizeLtPatterns a
 insertImplicitPatternsT exh            ps a = do
   TelV tel b <- telViewUpTo' (-1) (not . visible) a
   reportSDoc "tc.lhs.imp" 20 $
-    vcat [ text "insertImplicitPatternsT"
-         , nest 2 $ text "ps  = " <+> do
+    vcat [ "insertImplicitPatternsT"
+         , nest 2 $ "ps  = " <+> do
              brackets $ fsep $ punctuate comma $ map prettyA ps
-         , nest 2 $ text "tel = " <+> prettyTCM tel
-         , nest 2 $ text "b   = " <+> addContext tel (prettyTCM b)
+         , nest 2 $ "tel = " <+> prettyTCM tel
+         , nest 2 $ "b   = " <+> addContext tel (prettyTCM b)
+         ]
+  reportSDoc "tc.lhs.imp" 70 $
+    vcat [ "insertImplicitPatternsT"
+         , nest 2 $ "ps  = " <+> (text . show) ps
+         , nest 2 $ "tel = " <+> (text . show) tel
+         , nest 2 $ "b   = " <+> (text . show) b
          ]
   case ps of
     [] -> insImp dummy tel
-    p : ps -> do
+    p : _ -> do
       -- Andreas, 2015-05-11.
       -- If p is a projection pattern, make it visible for the purpose of
       -- calling insImp / insertImplicit, to get correct behavior.
-      let p' = applyWhen (isJust $ A.maybePostfixProjP p) (setHiding NotHidden) p
+      let p' = applyWhen (isJust $ A.isProjP p) (setHiding NotHidden) p
       hs <- insImp p' tel
-      case hs of
-        [] -> do
-          a <- reduce a
-          case unEl a of
-            Pi arg b -> do
-              (p :) <$> insertImplicitPatternsT exh ps (absBody b)
-            _ -> return (p : ps)
-        hs -> insertImplicitPatternsT exh (hs ++ p : ps) (telePi tel b)
+      -- Continue with implicit patterns inserted before @p@.
+      -- The list @hs ++ ps@ cannot be empty.
+      let ps0@(~(p1 : ps1)) = hs ++ ps
+      reduce a >>= piOrPath >>= \case
+        -- If @a@ is a function (or path) type, continue inserting after @p1@.
+        Left (_, b) -> (p1 :) <$> insertImplicitPatternsT exh ps1 (absBody b)
+        -- Otherwise, we are done.
+        Right{}     -> return ps0
   where
     dummy = defaultNamedArg (A.VarP __IMPOSSIBLE__)
 
     insImp p EmptyTel = return []
-    insImp p tel = case insertImplicit p $ map (argFromDom . fmap fst) $ telToList tel of
+    insImp p tel = case insertImplicit p $ telToList tel of
       BadImplicits   -> typeError WrongHidingInLHS
       NoSuchName x   -> typeError WrongHidingInLHS
       ImpInsert n    -> return $ map implicitArg n
-      NoInsertNeeded -> return []
 
     implicitArg h = implicitP $ setHiding h $ defaultArgInfo
