@@ -216,6 +216,7 @@ data LHSState a = LHSState
     --   Can be 'Irrelevant' to indicate that we came by
     --   an irrelevant projection and, hence, the rhs must
     --   be type-checked in irrelevant mode.
+  , _lhsPartialSplit :: ![Maybe Int] -- ^ have we splitted with a PartialFocus?
   }
 
 lhsTel :: Lens' Telescope (LHSState a)
@@ -230,6 +231,9 @@ lhsProblem f p = f (_lhsProblem p) <&> \x -> p {_lhsProblem = x}
 lhsTarget :: Lens' (Arg Type) (LHSState a)
 lhsTarget f p = f (_lhsTarget p) <&> \x -> p {_lhsTarget = x}
 
+lhsPartialSplit :: Lens' [Maybe Int] (LHSState a)
+lhsPartialSplit f p = f (_lhsPartialSplit p) <&> \x -> p {_lhsPartialSplit = x}
+
 instance Subst Term (Problem a) where
   applySubst rho (Problem eqs rps cont) = Problem (applySubst rho eqs) rps cont
 
@@ -244,31 +248,31 @@ instance Subst Term AbsurdPattern where
 
 instance PrettyTCM ProblemEq where
   prettyTCM (ProblemEq p v a) = sep
-    [ prettyA p <+> text "="
-    , nest 2 $ prettyTCM v <+> text ":"
+    [ prettyA p <+> "="
+    , nest 2 $ prettyTCM v <+> ":"
     , nest 2 $ prettyTCM a
     ]
 
 instance PrettyTCM AsBinding where
   prettyTCM (AsB x v a) =
-    sep [ prettyTCM x P.<> text "@" P.<> parens (prettyTCM v)
-        , nest 2 $ text ":" <+> prettyTCM a
+    sep [ prettyTCM x P.<> "@" P.<> parens (prettyTCM v)
+        , nest 2 $ ":" <+> prettyTCM a
         ]
 
 instance PrettyTCM DotPattern where
   prettyTCM (Dot e v a) = sep
-    [ prettyA e <+> text "="
-    , nest 2 $ prettyTCM v <+> text ":"
+    [ prettyA e <+> "="
+    , nest 2 $ prettyTCM v <+> ":"
     , nest 2 $ prettyTCM a
     ]
 
 instance PrettyTCM AbsurdPattern where
-  prettyTCM (Absurd r a) = text "() :" <+> prettyTCM a
+  prettyTCM (Absurd r a) = "() :" <+> prettyTCM a
 
 instance PP.Pretty AsBinding where
   pretty (AsB x v a) =
-    PP.pretty x PP.<+> PP.text "=" PP.<+>
-      PP.hang (PP.pretty v PP.<+> PP.text ":") 2 (PP.pretty a)
+    PP.pretty x PP.<+> "=" PP.<+>
+      PP.hang (PP.pretty v PP.<+> ":") 2 (PP.pretty a)
 
 instance InstantiateFull AsBinding where
   instantiateFull' (AsB x v a) = AsB x <$> instantiateFull' v <*> instantiateFull' a
