@@ -1,3 +1,4 @@
+{-# LANGUAGE BangPatterns  #-}
 {-# LANGUAGE CPP           #-}
 
 module Agda.TypeChecking.Rules.LHS.ProblemRest where
@@ -22,7 +23,7 @@ import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Rules.LHS.Problem
 import Agda.TypeChecking.Rules.LHS.Implicit
 
-import Agda.Utils.Functor (($>))
+import Agda.Utils.Functor
 import Agda.Utils.List
 import Agda.Utils.Size
 import Agda.Utils.Permutation
@@ -41,17 +42,19 @@ useNamesFromPattern ps tel = telFromList (zipWith ren ps telList ++ telRemaining
   where
     telList = telToList tel
     telRemaining = drop (length ps) telList -- telescope entries beyond patterns
-    ren (Arg ai (Named nm p)) dom@(Dom info (y, a)) =
+    ren (Arg ai (Named nm p)) dom@Dom{ unDom = (y, a) } =
       case p of
         -- Andreas, 2017-10-12, issue #2803, also preserve user-written hidden names.
         -- However, not if the argument is named, because then the name in the telescope
         -- is significant for implicit insertion.
-        A.VarP (A.BindName x) | not (isNoName x)
-                 , visible info || (getOrigin ai == UserWritten && nm == Nothing) ->
-          Dom info (nameToArgName x, a)
+        A.VarP (A.BindName x)
+          | not (isNoName x)
+          , visible dom || (getOrigin ai == UserWritten && nm == Nothing) ->
+          dom{ unDom = (nameToArgName x, a) }
+        A.AbsurdP{} | visible dom -> dom{ unDom = (stringToArgName "()", a) }
         A.PatternSynP{} -> __IMPOSSIBLE__  -- ensure there are no syns left
         -- Andreas, 2016-05-10, issue 1848: if context variable has no name, call it "x"
-        _ | visible info && isNoName y -> Dom info (stringToArgName "x", a)
+        _ | visible dom && isNoName y -> dom{ unDom = (stringToArgName "x", a) }
           | otherwise                  -> dom
 
 useOriginFrom :: (LensOrigin a, LensOrigin b) => [a] -> [b] -> [a]
@@ -93,44 +96,44 @@ initLHSState delta eqs ps a ret = do
   let problem = Problem eqs ps ret
       qs0     = teleNamedArgs delta
 
-  updateProblemRest $ LHSState delta qs0 problem (defaultArg a)
+  updateProblemRest $ LHSState delta qs0 problem (defaultArg a) []
 
 -- | Try to move patterns from the problem rest into the problem.
 --   Possible if type of problem rest has been updated to a function type.
 updateProblemRest :: LHSState a -> TCM (LHSState a)
-updateProblemRest st@(LHSState tel0 qs0 p@(Problem oldEqs ps ret) a) = do
+updateProblemRest st@(LHSState tel0 qs0 p@(Problem oldEqs ps ret) a psplit) = do
   ps <- addContext tel0 $ insertImplicitPatternsT ExpandLast ps $ unArg a
   reportSDoc "tc.lhs.imp" 20 $
-    text "insertImplicitPatternsT returned" <+> fsep (map prettyA ps)
+    "insertImplicitPatternsT returned" <+> fsep (map prettyA ps)
   -- (Issue 734: Do only the necessary telView to preserve clause types as much as possible.)
-  let m = length $ takeWhile (isNothing . A.maybePostfixProjP) ps
-  TelV gamma b <- telViewUpTo m $ unArg a
+  let m = length $ takeWhile (isNothing . A.isProjP) ps
+  (TelV gamma b, boundary) <- telViewUpToPathBoundaryP m $ unArg a
   forM_ (zip ps (telToList gamma)) $ \(p, a) ->
     unless (sameHiding p a) $ typeError WrongHidingInLHS
   let tel1      = useNamesFromPattern ps gamma
       n         = size tel1
       (ps1,ps2) = splitAt n ps
       tel       = telFromList $ telToList tel0 ++ telToList tel1
-      qs1       = teleNamedArgs tel1
+      qs1       = telePatterns tel1 boundary
       newEqs    = zipWith3 ProblemEq
                     (map namedArg ps1)
                     (map (patternToTerm . namedArg) qs1)
                     (flattenTel tel1 `useOriginFrom` ps1)
       tau       = raiseS n
   reportSDoc "tc.lhs.problem" 10 $ addContext tel0 $ vcat
-    [ text "checking lhs -- updated split problem:"
+    [ "checking lhs -- updated split problem:"
     , nest 2 $ vcat
-      [ text "ps    =" <+> fsep (map prettyA ps)
-      , text "a     =" <+> prettyTCM a
-      , text "tel1  =" <+> prettyTCM tel1
-      , text "ps1   =" <+> fsep (map prettyA ps1)
-      , text "ps2   =" <+> fsep (map prettyA ps2)
-      , text "b     =" <+> addContext tel1 (prettyTCM b)
+      [ "ps    =" <+> fsep (map prettyA ps)
+      , "a     =" <+> prettyTCM a
+      , "tel1  =" <+> prettyTCM tel1
+      , "ps1   =" <+> fsep (map prettyA ps1)
+      , "ps2   =" <+> fsep (map prettyA ps2)
+      , "b     =" <+> addContext tel1 (prettyTCM b)
       ]
     ]
   reportSDoc "tc.lhs.problem" 60 $ addContext tel0 $ vcat
     [ nest 2 $ vcat
-      [ text "qs1    =" <+> fsep (map pretty qs1)
+      [ "qs1    =" <+> fsep (map pretty qs1)
       ]
     ]
   return $ LHSState
@@ -142,4 +145,5 @@ updateProblemRest st@(LHSState tel0 qs0 p@(Problem oldEqs ps ret) a) = do
                    , _problemCont     = ret
                    }
     , _lhsTarget  = a $> b
+    , _lhsPartialSplit = psplit
     }
