@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE NondecreasingIndentation   #-}
 
 module Agda.TypeChecking.Rules.Record where
 
@@ -17,8 +18,13 @@ import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Pattern
 import Agda.Syntax.Position
 import qualified Agda.Syntax.Info as Info
+import Agda.Syntax.Scope.Monad (freshAbstractQName)
+import Agda.Syntax.Fixity
 
 import Agda.TypeChecking.Monad
+import Agda.TypeChecking.Monad.Builtin
+import Agda.TypeChecking.Names
+import Agda.TypeChecking.Primitive
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Reduce
@@ -29,10 +35,12 @@ import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.CompiledClause (hasProjectionPatterns)
 import Agda.TypeChecking.CompiledClause.Compile
 
-import Agda.TypeChecking.Rules.Data ( bindParameters, fitsIn, forceSort)
+import Agda.TypeChecking.Rules.Data ( getGeneralizedParameters, bindGeneralizedParameters, bindParameters, fitsIn, forceSort,
+                                      defineCompData, defineTranspOrHCompForFields )
 import Agda.TypeChecking.Rules.Term ( isType_ )
 import {-# SOURCE #-} Agda.TypeChecking.Rules.Decl (checkDecl)
 
+import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import Agda.Utils.Permutation
@@ -62,34 +70,46 @@ import Agda.Utils.Impossible
 checkRecDef
   :: Info.DefInfo              -- ^ Position and other info.
   -> QName                     -- ^ Record type identifier.
+  -> UniverseCheck             -- ^ Check universes?
   -> Maybe (Ranged Induction)  -- ^ Optional: (co)inductive declaration.
   -> Maybe HasEta              -- ^ Optional: user specified eta/no-eta
   -> Maybe QName               -- ^ Optional: constructor name.
-  -> [A.LamBinding]            -- ^ Record parameters.
+  -> A.DataDefParams           -- ^ Record parameters.
   -> A.Expr                    -- ^ Approximate type of constructor (@fields@ -> Set).
                                --   Does not include record parameters.
   -> [A.Field]                 -- ^ Field signatures.
   -> TCM ()
-checkRecDef i name ind eta con ps contel fields =
-  traceCall (CheckRecDef (getRange name) (qnameName name) ps fields) $ do
+checkRecDef i name uc ind eta con (A.DataDefParams gpars ps) contel fields =
+  traceCall (CheckRecDef (getRange name) name ps fields) $ do
     reportSDoc "tc.rec" 10 $ vcat
-      [ text "checking record def" <+> prettyTCM name
-      , nest 2 $ text "ps ="     <+> prettyList (map prettyAs ps)
-      , nest 2 $ text "contel =" <+> prettyA contel
-      , nest 2 $ text "fields =" <+> prettyA (map Constr fields)
+      [ "checking record def" <+> prettyTCM name
+      , nest 2 $ "ps ="     <+> prettyList (map prettyA ps)
+      , nest 2 $ "contel =" <+> prettyA contel
+      , nest 2 $ "fields =" <+> prettyA (map Constr fields)
       ]
     -- get type of record
-    t <- instantiateFull =<< typeOfConst name
-    bindParameters ps t $ \tel t0 -> do
+    def <- instantiateDef =<< getConstInfo name
+    t   <- instantiateFull $ defType def
+    let npars =
+          case theDef def of
+            DataOrRecSig n -> n
+            _              -> __IMPOSSIBLE__
+
+    parNames <- getGeneralizedParameters gpars name
+
+    bindGeneralizedParameters parNames t $ \ gtel t0 ->
+     bindParameters (npars - length parNames) ps t0 $ \ ptel t0 -> do
+
+      let tel = abstract gtel ptel
 
       -- Generate type of constructor from field telescope @contel@,
       -- which is the approximate constructor type (target missing).
 
       -- Check and evaluate field types.
-      reportSDoc "tc.rec" 15 $ text "checking fields"
+      reportSDoc "tc.rec" 15 $ "checking fields"
       contype <- workOnTypes $ instantiateFull =<< isType_ contel
       reportSDoc "tc.rec" 20 $ vcat
-        [ text "contype = " <+> prettyTCM contype ]
+        [ "contype = " <+> prettyTCM contype ]
 
       -- compute the field telescope (does not include record parameters)
       let TelV ftel _ = telView' contype
@@ -101,9 +121,15 @@ checkRecDef i name ind eta con ps contel fields =
       unless (null idxTel) $ typeError $ ShouldBeASort t0
       s <- forceSort s
 
+      -- needed for impredicative Prop (not implemented yet)
+      -- ftel <- return $
+      --   if s == Prop
+      --   then telFromList $ map (setRelevance Irrelevant) $ telToList ftel
+      --   else ftel
+
       reportSDoc "tc.rec" 20 $ do
         gamma <- getContextTelescope  -- the record params (incl. module params)
-        text "gamma = " <+> inTopContext (prettyTCM gamma)
+        "gamma = " <+> inTopContext (prettyTCM gamma)
 
       -- record type (name applied to parameters)
       rect <- El s . Def name . map Apply <$> getContextArgs
@@ -123,7 +149,7 @@ checkRecDef i name ind eta con ps contel fields =
           return (False, c, i)
 
       -- Add record type to signature.
-      reportSDoc "tc.rec" 15 $ text "adding record type to signature"
+      reportSDoc "tc.rec" 15 $ "adding record type to signature"
 
       etaenabled <- etaEnabled
 
@@ -143,7 +169,7 @@ checkRecDef i name ind eta con ps contel fields =
           -- We should turn it off until it is proven to be safe.
           haveEta      = maybe (Inferred NoEta) Specified eta
           -- haveEta      = maybe (Inferred $ conInduction == Inductive && etaenabled) Specified eta
-          con = ConHead conName conInduction $ map unArg fs
+          con = ConHead conName conInduction fs
 
           -- A record is irrelevant if all of its fields are.
           -- In this case, the associated module parameter will be irrelevant.
@@ -158,11 +184,11 @@ checkRecDef i name ind eta con ps contel fields =
       -- Disallow coinductive records with eta-equality
       when (conInduction == CoInductive && theEtaEquality haveEta == YesEta) $ do
         typeError . GenericDocError =<< do
-          sep [ text "Agda doesn't like coinductive records with eta-equality."
-              , text "If you must, use pragma"
-              , text "{-# ETA" <+> prettyTCM name <+> text "#-}"
+          sep [ "Agda doesn't like coinductive records with eta-equality."
+              , "If you must, use pragma"
+              , "{-# ETA" <+> prettyTCM name <+> "#-}"
               ]
-      reportSDoc "tc.rec" 30 $ text "record constructor is " <+> prettyTCM con
+      reportSDoc "tc.rec" 30 $ "record constructor is " <+> prettyTCM con
 
       -- Add the record definition.
 
@@ -190,6 +216,7 @@ checkRecDef i name ind eta con ps contel fields =
                   -- in case the record turns out to be recursive.
               -- Determined by positivity checker:
               , recMutual         = Nothing
+              , recComp           = emptyCompKit -- filled in later
               }
 
         -- Add record constructor to signature
@@ -202,6 +229,7 @@ checkRecDef i name ind eta con ps contel fields =
               , conData   = name
               , conAbstr  = Info.defAbstract conInfo
               , conInd    = conInduction
+              , conComp   = (emptyCompKit, Nothing) -- filled in later
               , conForced = []
               , conErased = []
               }
@@ -211,7 +239,7 @@ checkRecDef i name ind eta con ps contel fields =
         addNamedInstance conName name
 
       -- Check that the fields fit inside the sort
-      _ <- fitsIn [] contype s
+      _ <- fitsIn uc [] contype s
 
       {- Andreas, 2011-04-27 WRONG because field types are checked again
          and then non-stricts should not yet be irrelevant
@@ -222,19 +250,19 @@ checkRecDef i name ind eta con ps contel fields =
 
 {- Andreas, 2013-09-13 DEBUGGING the debug printout
       reportSDoc "tc.rec" 80 $ sep
-        [ text "current module record telescope"
+        [ "current module record telescope"
         , nest 2 $ (prettyTCM =<< getContextTelescope)
         ]
       reportSDoc "tc.rec" 80 $ sep
-        [ text "current module record telescope"
+        [ "current module record telescope"
         , nest 2 $ (text . show =<< getContextTelescope)
         ]
       reportSDoc "tc.rec" 80 $ sep
-        [ text "current module record telescope"
+        [ "current module record telescope"
         , nest 2 $ (inTopContext . prettyTCM =<< getContextTelescope)
         ]
       reportSDoc "tc.rec" 80 $ sep
-        [ text "current module record telescope"
+        [ "current module record telescope"
         , nest 2 $ do
            tel <- getContextTelescope
            text (show tel) $+$ do
@@ -243,13 +271,13 @@ checkRecDef i name ind eta con ps contel fields =
                telA <- reify tel
                text (show telA) $+$ do
                ctx <- getContextTelescope
-               text "should be empty:" <+> prettyTCM ctx
+               "should be empty:" <+> prettyTCM ctx
         ]
 -}
 
       let info = setRelevance recordRelevance defaultArgInfo
-          addRecordVar = addContext ("", Dom info rect)
-          -- the record variable has the empty name by intention, see issue 208
+          addRecordVar =
+            addRecordNameContext (setArgInfo info $ defaultDom rect)
 
       let m = qnameToMName name  -- Name of record module.
 
@@ -262,14 +290,14 @@ checkRecDef i name ind eta con ps contel fields =
         -- Add the record section.
 
         reportSDoc "tc.rec.def" 10 $ sep
-          [ text "record section:"
+          [ "record section:"
           , nest 2 $ sep
             [ prettyTCM m <+> (inTopContext . prettyTCM =<< getContextTelescope)
             , fsep $ punctuate comma $ map (return . P.pretty . getName) fields
             ]
           ]
         reportSDoc "tc.rec.def" 15 $ nest 2 $ vcat
-          [ text "field tel =" <+> escapeContext 1 (prettyTCM ftel)
+          [ "field tel =" <+> escapeContext 1 (prettyTCM ftel)
           ]
         addSection m
 
@@ -290,8 +318,168 @@ checkRecDef i name ind eta con ps contel fields =
           setModuleCheckpoint m
           checkRecordProjections m name hasNamedCon con tel' (raise 1 ftel) fields
 
+
+      -- we define composition here so that the projections are already in the signature.
+      escapeContext npars $ do
+        addCompositionForRecord name con tel fs ftel rect
+
       return ()
 
+
+addCompositionForRecord
+  :: QName      -- datatype name
+               -> ConHead
+               -> Telescope   -- Γ parameters
+               -> [Arg QName] -- projection names
+               -> Telescope   -- Γ ⊢ Φ field types
+               -> Type        -- Γ ⊢ T target type
+               -> TCM ()
+addCompositionForRecord name con tel fs ftel rect = do
+  compWays <- do
+    cxt <- getContextTelescope
+    escapeContext (size cxt) $
+      if null fs then Left . (,Just []) <$> defineCompData name con (abstract cxt tel) [] ftel rect []
+                 else Right <$>
+                      ifM (return (any (== Irrelevant) $ map getRelevance fs) `and2M` do not . optIrrelevantProjections <$> pragmaOptions)
+                          (return emptyCompKit)
+                          (defineCompKitR name (abstract cxt tel) ftel fs rect)
+  case compWays of
+    Right kit -> do
+      modifySignature $ updateDefinition name $ updateTheDef $ \ d ->
+        case d of
+          r@Record{} -> r { recComp = kit }
+          _          -> __IMPOSSIBLE__
+    Left y -> do
+      modifySignature $ updateDefinition (conName con) $ updateTheDef $ \ d ->
+        case d of
+          r@Constructor{} -> r { conComp = y }
+          _          -> __IMPOSSIBLE__
+
+defineCompKitR ::
+    QName          -- ^ some name, e.g. record name
+  -> Telescope   -- ^ param types Δ
+  -> Telescope   -- ^ fields' types Δ ⊢ Φ
+  -> [Arg QName] -- ^ fields' names
+  -> Type        -- ^ record type Δ ⊢ T
+  -> TCM CompKit
+defineCompKitR name params fsT fns rect = do
+  required <- mapM getTerm'
+        [ builtinInterval
+        , builtinIZero
+        , builtinIOne
+        , builtinIMin
+        , builtinIMax
+        , builtinINeg
+        , builtinPOr
+        , builtinItIsOne
+        ]
+  reportSDoc "tc.rec.cxt" 30 $ prettyTCM params
+  reportSDoc "tc.rec.cxt" 30 $ prettyTCM fsT
+  reportSDoc "tc.rec.cxt" 30 $ pretty rect
+  if not $ all isJust required then return $ emptyCompKit else do
+    transp <- whenDefined [builtinTrans]              (defineTranspOrHCompR DoTransp name params fsT fns rect)
+    hcomp  <- whenDefined [builtinTrans,builtinHComp] (defineTranspOrHCompR DoHComp name params fsT fns rect)
+    return $ CompKit
+      { nameOfTransp = transp
+      , nameOfHComp  = hcomp
+      }
+  where
+    whenDefined xs m = do
+      xs <- mapM getTerm' xs
+      if all isJust xs then m else return Nothing
+
+
+defineTranspOrHCompR ::
+  TranspOrHComp
+  -> QName       -- ^ some name, e.g. record name
+  -> Telescope   -- ^ param types Δ
+  -> Telescope   -- ^ fields' types Δ ⊢ Φ
+  -> [Arg QName] -- ^ fields' names
+  -> Type        -- ^ record type Δ ⊢ T
+  -> TCM (Maybe QName)
+defineTranspOrHCompR cmd name params fsT fns rect = do
+  let project = (\ t fn -> t `applyE` [Proj ProjSystem fn])
+  stuff <- fmap fst <$> defineTranspOrHCompForFields cmd Nothing project name params fsT fns rect
+
+  caseMaybe stuff (return Nothing) $ \ (theName, gamma, rtype, clause_types, bodies) -> do
+  -- phi = 1 clause
+  c' <- do
+           io <- primIOne
+           Just io_name <- getBuiltinName' builtinIOne
+           one <- primItIsOne
+           tInterval <- elInf primInterval
+           let
+              (ix,rhs) =
+                case cmd of
+                  -- TranspRArgs = phi : I, a0 : ..
+                  -- Γ = Δ^I , CompRArgs
+                  -- pats = ... | phi = i1
+                  -- body = a0
+                  DoTransp -> (1,Var 0 [])
+                  -- HCompRArgs = phi : I, u : .., a0 : ..
+                  -- Γ = Δ, CompRArgs
+                  -- pats = ... | phi = i1
+                  -- body = u i1 itIsOne
+                  DoHComp  -> (2,Var 1 [] `apply` [argN io, setRelevance Irrelevant $ argN one])
+
+              p = ConP (ConHead io_name Inductive [])
+                       (noConPatternInfo { conPType = Just (Arg defaultArgInfo tInterval)
+                                         , conPFallThrough = True })
+                         []
+
+              -- gamma, rtype
+
+              s = singletonS ix p
+
+              pats :: [NamedArg DeBruijnPattern]
+              pats = s `applySubst` teleNamedArgs gamma
+
+              t :: Type
+              t = s `applyPatSubst` rtype
+
+              gamma' :: Telescope
+              gamma' = unflattenTel (ns0 ++ ns1) $ s `applyPatSubst` (g0 ++ g1)
+               where
+                (g0,_:g1) = splitAt (size gamma - 1 - ix) $ flattenTel gamma
+                (ns0,_:ns1) = splitAt (size gamma - 1 - ix) $ teleNames gamma
+
+              c = Clause { clauseTel       = gamma'
+                         , clauseType      = Just $ argN t
+                         , namedClausePats = pats
+                         , clauseFullRange = noRange
+                         , clauseLHSRange  = noRange
+                         , clauseCatchall  = False
+                         , clauseBody      = Just $ rhs
+                         , clauseUnreachable = Just False
+                         }
+           reportSDoc "trans.rec.face" 17 $ text $ show c
+           return c
+  cs <- flip mapM (zip3 fns clause_types bodies) $ \ (fname, clause_ty, body) -> do
+          let
+              pats = teleNamedArgs gamma ++ [defaultNamedArg $ ProjP ProjSystem $ unArg fname]
+              c = Clause { clauseTel       = gamma
+                         , clauseType      = Just $ argN (unDom clause_ty)
+                         , namedClausePats = pats
+                         , clauseFullRange = noRange
+                         , clauseLHSRange  = noRange
+                         , clauseCatchall  = False
+                         , clauseBody      = Just body
+                         , clauseUnreachable = Just False
+                         }
+          reportSDoc "trans.rec" 17 $ text $ show c
+          reportSDoc "trans.rec" 16 $ text "type =" <+> text (show (clauseType c))
+          reportSDoc "trans.rec" 15 $ prettyTCM $ abstract gamma (unDom clause_ty)
+          reportSDoc "trans.rec" 10 $ text "body =" <+> prettyTCM (abstract gamma body)
+          return c
+  addClauses theName $ c' : cs
+  reportSDoc "trans.rec" 15 $ text $ "compiling clauses for " ++ show theName
+  (mst, cc) <- inTopContext (compileClauses Nothing cs)
+  whenJust mst $ setSplitTree theName
+  setCompiledClauses theName cc
+  reportSDoc "trans.rec" 15 $ text $ "compiled"
+  return $ Just theName
+
+
 {-| @checkRecordProjections m r q tel ftel fs@.
 
     [@m@    ]  name of the generated module
@@ -321,21 +509,21 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
       setScope scope >> checkProjs ftel1 ftel2 (fs' ++ fs)
 
     -- Case: projection.
-    checkProjs ftel1 (ExtendTel (Dom ai t) ftel2) (A.Field info x _ : fs) =
+    checkProjs ftel1 (ExtendTel (dom@Dom{domInfo = ai,unDom = t}) ftel2) (A.Field info x _ : fs) =
       traceCall (CheckProjection (getRange info) x t) $ do
       -- Andreas, 2012-06-07:
       -- Issue 387: It is wrong to just type check field types again
       -- because then meta variables are created again.
       -- Instead, we take the field type t from the field telescope.
       reportSDoc "tc.rec.proj" 5 $ sep
-        [ text "checking projection" <+> prettyTCM x
+        [ "checking projection" <+> prettyTCM x
         , nest 2 $ vcat
-          [ text "top   =" <+> (inTopContext . prettyTCM =<< getContextTelescope)
-          , text "tel   =" <+> (inTopContext . prettyTCM $ tel)
-          , text "ftel1 =" <+> prettyTCM ftel1
-          , text "t     =" <+> prettyTCM t
-          , text "ftel2 =" <+> addContext ftel1 (underAbstraction_ ftel2 prettyTCM)
-          , text "abstr =" <+> (text . show) (Info.defAbstract info)
+          [ "top   =" <+> (inTopContext . prettyTCM =<< getContextTelescope)
+          , "tel   =" <+> (inTopContext . prettyTCM $ tel)
+          , "ftel1 =" <+> prettyTCM ftel1
+          , "t     =" <+> prettyTCM t
+          , "ftel2 =" <+> addContext ftel1 (underAbstraction_ ftel2 prettyTCM)
+          , "abstr =" <+> (text . show) (Info.defAbstract info)
           ]
         ]
 
@@ -363,20 +551,23 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
           projcall o = Var 0 [Proj o projname]
           rel      = getRelevance ai
           -- the recursive call
-          recurse  = checkProjs (abstract ftel1 $ ExtendTel (Dom ai t)
+          recurse  = checkProjs (abstract ftel1 $ ExtendTel dom
                                  $ Abs (nameToArgName $ qnameName projname) EmptyTel)
                                 (ftel2 `absApp` projcall ProjSystem) fs
 
-      reportSDoc "tc.rec.proj" 25 $ nest 2 $ text "finalt=" <+> do
+      reportSDoc "tc.rec.proj" 25 $ nest 2 $ "finalt=" <+> do
         inTopContext $ prettyTCM finalt
 
-      -- Andreas, 2012-02-20 do not add irrelevant projections if
-      -- disabled by --no-irrelevant-projections
-      ifM (return (rel == Irrelevant) `and2M` do not . optIrrelevantProjections <$> pragmaOptions) recurse $ do
-
+      -- -- Andreas, 2012-02-20 do not add irrelevant projections if
+      -- -- disabled by --no-irrelevant-projections
+      -- ifM (return (rel == Irrelevant) `and2M` do not . optIrrelevantProjections <$> pragmaOptions) recurse $ do
+      -- Andreas, 2018-06-09 issue #2170
+      -- Always create irrelevant projections (because the scope checker accepts irrelevant fields).
+      -- If --no-irrelevant-projections, then their use should be disallowed by the type checker for expressions.
+      do
         reportSDoc "tc.rec.proj" 10 $ sep
-          [ text "adding projection"
-          , nest 2 $ prettyTCM projname <+> text ":" <+> inTopContext (prettyTCM finalt)
+          [ "adding projection"
+          , nest 2 $ prettyTCM projname <+> ":" <+> inTopContext (prettyTCM finalt)
           ]
 
         -- The body should be
@@ -402,10 +593,13 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
             (_ptel,[rt]) = splitAt (size tel - 1) telList
             cpo    = if hasNamedCon then PatOCon else PatORec
             cpi    = ConPatternInfo { conPRecord = Just cpo
+                                    , conPFallThrough = False
                                     , conPType   = Just $ argFromDom $ fmap snd rt
                                     , conPLazy   = True }
             conp   = defaultArg $ ConP con cpi $
-                     [ Arg ai' $ unnamed $ varP "x" | Dom ai' _ <- telToList ftel ]
+                     [ Arg ai' $ unnamed $ varP ("x" :: String)
+                     | Dom{domInfo = ai'} <- telToList ftel
+                     ]
             body   = Just $ bodyMod $ var (size ftel2)
             cltel  = ftel
             clause = Clause { clauseLHSRange  = getRange info
@@ -426,20 +620,20 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
               -- index of the record argument (in the type),
               -- start counting with 1:
               , projIndex    = size tel -- which is @size ptel + 1@
-              , projLams     = ProjLams $ map (\ (Dom ai (x,_)) -> Arg ai x) telList
+              , projLams     = ProjLams $ map (argFromDom . fmap fst) telList
               }
 
         reportSDoc "tc.rec.proj" 80 $ sep
-          [ text "adding projection"
+          [ "adding projection"
           , nest 2 $ prettyTCM projname <+> text (show clause)
           ]
         reportSDoc "tc.rec.proj" 70 $ sep
-          [ text "adding projection"
-          , nest 2 $ prettyTCM projname <+> text (show (clausePats clause)) <+> text "=" <+>
-                       inTopContext (addContext ftel (maybe (text "_|_") prettyTCM (clauseBody clause)))
+          [ "adding projection"
+          , nest 2 $ prettyTCM projname <+> text (show (clausePats clause)) <+> "=" <+>
+                       inTopContext (addContext ftel (maybe "_|_" prettyTCM (clauseBody clause)))
           ]
         reportSDoc "tc.rec.proj" 10 $ sep
-          [ text "adding projection"
+          [ "adding projection"
           , nest 2 $ prettyTCM (QNamed projname clause)
           ]
 
@@ -447,10 +641,10 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
               -- projection functions are defined. Record pattern
               -- translation is defined in terms of projection
               -- functions.
-        cc <- compileClauses Nothing [clause]
+        (mst , cc) <- compileClauses Nothing [clause]
 
         reportSDoc "tc.cc" 60 $ do
-          sep [ text "compiled clauses of " <+> prettyTCM projname
+          sep [ "compiled clauses of " <+> prettyTCM projname
               , nest 2 $ text (show cc)
               ]
 
@@ -460,6 +654,7 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
               emptyFunction
                 { funClauses        = [clause]
                 , funCompiled       = Just cc
+                , funSplitTree      = mst
                 , funProjection     = Just projection
                 , funMutual         = Just []  -- Projections are not mutually recursive with anything
                 , funTerminates     = Just True
@@ -467,8 +662,9 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
                 })
               { defArgOccurrences = [StrictPos] }
           computePolarity [projname]
-          when (Info.defInstance info == InstanceDef) $
-            addTypedInstance projname finalt
+
+        when (Info.defInstance info == InstanceDef) $
+          addTypedInstance projname t
 
         recurse
 
