@@ -6,7 +6,9 @@
 
 module Agda.TypeChecking.Serialise.Base where
 
+import Control.Exception (evaluate)
 import Control.Monad
+import Control.Monad.Catch (catchAll)
 import Control.Monad.Reader
 import Control.Monad.State.Strict (StateT, gets)
 
@@ -20,6 +22,7 @@ import Data.Int (Int32)
 import Data.Maybe
 import qualified Data.Binary as B
 import qualified Data.Binary.Get as B
+import Data.Text.Lazy (Text)
 import Data.Typeable ( cast, Typeable, typeOf, TypeRep )
 
 import Agda.Syntax.Common (NameId)
@@ -79,7 +82,7 @@ data Dict = Dict
   -- Dictionaries which are serialized:
   { nodeD        :: !(HashTable Node    Int32)    -- ^ Written to interface file.
   , stringD      :: !(HashTable String  Int32)    -- ^ Written to interface file.
-  , bstringD     :: !(HashTable L.ByteString Int32) -- ^ Written to interface file.
+  , textD        :: !(HashTable Text    Int32)    -- ^ Written to interface file.
   , integerD     :: !(HashTable Integer Int32)    -- ^ Written to interface file.
   , doubleD      :: !(HashTable Double  Int32)    -- ^ Written to interface file.
   -- Dicitionaries which are not serialized, but provide
@@ -92,7 +95,7 @@ data Dict = Dict
   -- Fresh UIDs and reuse statistics:
   , nodeC        :: !(IORef FreshAndReuse)  -- counters for fresh indexes
   , stringC      :: !(IORef FreshAndReuse)
-  , bstringC     :: !(IORef FreshAndReuse)
+  , textC        :: !(IORef FreshAndReuse)
   , integerC     :: !(IORef FreshAndReuse)
   , doubleC      :: !(IORef FreshAndReuse)
   , termC        :: !(IORef FreshAndReuse)
@@ -141,7 +144,7 @@ type Memo = HashTable (Int32, TypeRep) U    -- (node index, type rep)
 data St = St
   { nodeE     :: !(Array Int32 Node)     -- ^ Obtained from interface file.
   , stringE   :: !(Array Int32 String)   -- ^ Obtained from interface file.
-  , bstringE  :: !(Array Int32 L.ByteString) -- ^ Obtained from interface file.
+  , textE     :: !(Array Int32 Text)     -- ^ Obtained from interface file.
   , integerE  :: !(Array Int32 Integer)  -- ^ Obtained from interface file.
   , doubleE   :: !(Array Int32 Double)   -- ^ Obtained from interface file.
   , nodeMemo  :: !Memo
@@ -179,6 +182,14 @@ class Typeable a => EmbPrj a where
     tickICode a
     icod_ a
 
+  -- Simple enumeration types can be (de)serialized using (from/to)Enum.
+
+  default value :: (Enum a) => Int32 -> R a
+  value i = liftIO (evaluate (toEnum (fromIntegral i))) `catchAll` const malformed
+
+  default icod_ :: (Enum a) => a -> S Int32
+  icod_ = return . fromIntegral . fromEnum
+
 -- | Increase entry for @a@ in 'stats'.
 tickICode :: forall a. Typeable a => a -> S ()
 tickICode _ = whenM (asks collectStats) $ do
@@ -350,7 +361,6 @@ vcase valu = \ix -> do
           liftIO $ H.insert memo (ix, aTyp) (U v)
           return v
 
-
 -- | @icodeArgs proxy (a1, ..., an)@ maps @icode@ over @a1@, ..., @an@
 --   and returns the corresponding list of @Int32@.
 
