@@ -13,9 +13,10 @@ import Agda.TypeChecking.Serialise.Instances.Internal ()
 import Agda.TypeChecking.Serialise.Instances.Errors ()
 
 instance EmbPrj Interface where
-  icod_ (Interface a b c d e f g h i j k l m n) = icodeN' Interface a b c d e f g h i j k l m n
+  icod_ (Interface a b c d e f g h i j k l m n o p q) =
+    icodeN' Interface a b c d e f g h i j k l m n o p q
 
   value = vcase valu where
-    valu [a, b, c, d, e, f, g, h, i, j, k, l, m, n] =
-      valuN Interface a b c d e f g h i j k l m n
+    valu [a, b, c, d, e, f, g, h, i, j, k, l, m, n, o, p, q] =
+      valuN Interface a b c d e f g h i j k l m n o p q
     valu _ = malformed
