@@ -25,6 +25,7 @@ instance EmbPrj HP.NameKind where
   icod_ HP.Record          = icodeN 8 ()
   icod_ HP.Argument        = icodeN 9 ()
   icod_ HP.Macro           = icodeN 10 ()
+  icod_ HP.Generalizable   = icodeN 11 ()
 
   value = vcase valu where
     valu []      = valuN HP.Bound
@@ -38,27 +39,32 @@ instance EmbPrj HP.NameKind where
     valu [8]     = valuN HP.Record
     valu [9]     = valuN HP.Argument
     valu [10]    = valuN HP.Macro
+    valu [11]    = valuN HP.Generalizable
     valu _       = malformed
 
 instance EmbPrj HP.Aspect where
-  icod_ HP.Comment       = icodeN 0 ()
-  icod_ HP.Option        = icodeN 1 ()
-  icod_ HP.Keyword       = icodeN 2 ()
-  icod_ HP.String        = icodeN 3 ()
-  icod_ HP.Number        = icodeN 4 ()
+  icod_ HP.Comment        = icodeN 0 ()
+  icod_ HP.Keyword       = icodeN 1 ()
+  icod_ HP.String        = icodeN 2 ()
+  icod_ HP.Number        = icodeN 3 ()
   icod_ HP.Symbol        = icodeN' HP.Symbol
-  icod_ HP.PrimitiveType = icodeN 6 ()
-  icod_ (HP.Name mk b)   = icodeN 7 HP.Name mk b
+  icod_ HP.PrimitiveType = icodeN 4 ()
+  icod_ (HP.Name mk b)   = icodeN 5 HP.Name mk b
+  icod_ HP.Pragma        = icodeN 6 ()
+  icod_ HP.Background    = icodeN 7 ()
+  icod_ HP.Markup        = icodeN 8 ()
 
   value = vcase valu where
     valu [0]        = valuN HP.Comment
-    valu [1]        = valuN HP.Option
-    valu [2]        = valuN HP.Keyword
-    valu [3]        = valuN HP.String
-    valu [4]        = valuN HP.Number
+    valu [1]        = valuN HP.Keyword
+    valu [2]        = valuN HP.String
+    valu [3]        = valuN HP.Number
     valu []         = valuN HP.Symbol
-    valu [6]        = valuN HP.PrimitiveType
-    valu [7, mk, b] = valuN HP.Name mk b
+    valu [4]        = valuN HP.PrimitiveType
+    valu [5, mk, b] = valuN HP.Name mk b
+    valu [6]        = valuN HP.Pragma
+    valu [7]        = valuN HP.Background
+    valu [8]        = valuN HP.Markup
     valu _          = malformed
 
 instance EmbPrj HP.OtherAspect where
@@ -70,7 +76,7 @@ instance EmbPrj HP.OtherAspect where
   icod_ HP.TypeChecks          = icodeN 5 ()
   icod_ HP.UnsolvedConstraint  = icodeN 6 ()
   icod_ HP.PositivityProblem   = icodeN 7 ()
-  icod_ HP.ReachabilityProblem = icodeN 8 ()
+  icod_ HP.Deadcode            = icodeN 8 ()
   icod_ HP.CoverageProblem     = icodeN 9 ()
   icod_ HP.CatchallClause      = icodeN 10 ()
 
@@ -83,7 +89,7 @@ instance EmbPrj HP.OtherAspect where
     valu [5] = valuN HP.TypeChecks
     valu [6] = valuN HP.UnsolvedConstraint
     valu [7] = valuN HP.PositivityProblem
-    valu [8] = valuN HP.ReachabilityProblem
+    valu [8] = valuN HP.Deadcode
     valu [9] = valuN HP.CoverageProblem
     valu [10] = valuN HP.CatchallClause
     valu _   = malformed
