@@ -56,6 +56,7 @@ import Control.Monad.Trans.Maybe
 import Control.Monad.Reader (asks)
 
 import Data.Foldable (Foldable, foldMap, forM_)
+import qualified Data.Foldable as Fold
 import Data.Function
 import qualified Data.List as List
 import Data.Monoid
@@ -92,14 +93,16 @@ import Agda.Utils.Except ( MonadError(catchError) )
 import Agda.Utils.Function
 import Agda.Utils.Functor
 import Agda.Utils.Lens
+import Agda.Utils.NonemptyList as NeList
 
-import qualified Agda.Utils.List as List hiding ( uncons )
+import qualified Agda.Utils.List as List
 
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import Agda.Utils.Pretty (Pretty, prettyShow)
 import qualified Agda.Utils.Pretty as P
+import Agda.Utils.Singleton
 import Agda.Utils.Size
 import Agda.Utils.Tuple
 import qualified Agda.Utils.VarSet as VarSet
@@ -130,7 +133,7 @@ solveSizeConstraints flag =  do
       ] ++ map prettyTCM cs0
   let -- Error for giving up
       cannotSolve = typeError . GenericDocError =<<
-        vcat (text "Cannot solve size constraints" : map prettyTCM cs0)
+        vcat ("Cannot solve size constraints" : map prettyTCM cs0)
 
   -- 2. Cluster the constraints by common size metas.
 
@@ -143,12 +146,12 @@ solveSizeConstraints flag =  do
     -- @allMetas@ does not reduce or instantiate;
     -- this is why we require the size constraints to be normalised.
     return (cl, map metaId . Set.toList $
-      sizeMetaSet `Set.intersection` Set.fromList (allMetas c))
+      sizeMetaSet `Set.intersection` allMetas singleton c)
 
   -- Now, some constraints may have no metas (clcs), the others have at least one (othercs).
-  let classify :: (a, [b]) -> Either a (a, (b,[b]))
+  let classify :: (a, [b]) -> Either a (a, NonemptyList b)
       classify (cl, [])     = Left  cl
-      classify (cl, (x:xs)) = Right (cl, (x,xs))
+      classify (cl, (x:xs)) = Right (cl, x :! xs)
   let (clcs, othercs) = List.mapEither classify cms
 
   -- We cluster the constraints by their metas.
@@ -163,19 +166,18 @@ solveSizeConstraints flag =  do
   -- Solve the clusters.
 
   constrainedMetas <- Set.unions <$> do
-    forM  (ccs) $ \ (cs :: [CC]) -> do
-      when (null cs) __IMPOSSIBLE__
+    forM  (ccs) $ \ (cs :: NonemptyList CC) -> do
 
       -- Convert each constraint in the cluster to the largest context.
       -- (Keep fingers crossed).
 
-      enterClosure (List.maximumBy (compare `on` (length . envContext . clEnv)) cs) $ \ _ -> do
+      enterClosure (Fold.maximumBy (compare `on` (length . envContext . clEnv)) cs) $ \ _ -> do
         -- Get all constraints that can be cast to the longest context.
         cs' :: [TCM.Constraint] <- catMaybes <$> do
-          mapM (runMaybeT . castConstraintToCurrentContext) cs
+          mapM (runMaybeT . castConstraintToCurrentContext) $ toList cs
 
         reportSDoc "tc.size.solve" 20 $ vcat $
-          [ text "converted size constraints to context: " <+> do
+          [ "converted size constraints to context: " <+> do
               tel <- getContextTelescope
               inTopContext $ prettyTCM tel
           ] ++ map (nest 2 . prettyTCM) cs'
@@ -192,7 +194,7 @@ solveSizeConstraints flag =  do
     --       for cs0 $ \ Closure{ clValue = ValueCmp _ _ u v } ->
     --         allMetas u ++ allMetas v
 
-    -- Set the unconstrained size metas to ∞.
+    -- Set the unconstrained, open size metas to ∞.
     ms <- S.getSizeMetas False -- do not get interaction metas
     unless (null ms) $ do
       inf <- primSizeInf
@@ -200,8 +202,8 @@ solveSizeConstraints flag =  do
         unless (m `Set.member` constrainedMetas) $ do
         unlessM (isFrozen m) $ do
         reportSDoc "tc.size.solve" 20 $
-          text "solution " <+> prettyTCM (MetaV m []) <+>
-          text " := "      <+> prettyTCM inf
+          "solution " <+> prettyTCM (MetaV m []) <+>
+          " := "      <+> prettyTCM inf
         assignMeta 0 m t (List.downFrom $ size tel) inf
 
   -- -- Double check.
@@ -267,17 +269,17 @@ castConstraintToCurrentContext' cl = do
   sigma <- liftTCM $ getModuleParameterSub modN
 
   -- Debug printing.
-  reportSDoc "tc.constr.cast" 40 $ text "casting constraint" $$ do
+  reportSDoc "tc.constr.cast" 40 $ "casting constraint" $$ do
     tel <- getContextTelescope
     inTopContext $ nest 2 $ vcat $
-      [ text "current module                = " <+> prettyTCM modM
-      , text "current module telescope      = " <+> prettyTCM gamma1
-      , text "current context               = " <+> prettyTCM tel
-      , text "constraint module             = " <+> prettyTCM modN
-      , text "constraint module telescope   = " <+> prettyTCM delta1
-      , text "constraint context            = " <+> (prettyTCM =<< enterClosure cl (const $ getContextTelescope))
-      , text "constraint                    = " <+> enterClosure cl prettyTCM
-      , text "module parameter substitution = " <+> prettyTCM sigma
+      [ "current module                = " <+> prettyTCM modM
+      , "current module telescope      = " <+> prettyTCM gamma1
+      , "current context               = " <+> prettyTCM tel
+      , "constraint module             = " <+> prettyTCM modN
+      , "constraint module telescope   = " <+> prettyTCM delta1
+      , "constraint context            = " <+> (prettyTCM =<< enterClosure cl (const $ getContextTelescope))
+      , "constraint                    = " <+> enterClosure cl prettyTCM
+      , "module parameter substitution = " <+> prettyTCM sigma
       ]
 
   -- If gamma2 < 0, we must be in the wrong context.
@@ -321,12 +323,12 @@ castConstraintToCurrentContext :: Closure TCM.Constraint -> MaybeT TCM TCM.Const
 castConstraintToCurrentContext cl = do
   -- The checkpoint of the contraint
   let cp = envCurrentCheckpoint $ clEnv cl
-  sigma <- caseMaybeM (view $ eCheckpoints . key cp)
+  sigma <- caseMaybeM (viewTC $ eCheckpoints . key cp)
           (do
             -- We are not in a descendant of the constraint checkpoint.
             -- Here be dragons!!
-            gamma <- asks envContext -- The target context
-            let findInGamma (Dom _ (x, t)) =
+            gamma <- asksTC envContext -- The target context
+            let findInGamma (Dom {unDom = (x, t)}) =
                   -- match by name (hazardous)
                   -- This is one of the seven deadly sins (not respecting alpha).
                   List.findIndex ((x ==) . fst . unDom) gamma
@@ -340,9 +342,8 @@ castConstraintToCurrentContext cl = do
             guard $ getAll $ runFree (All . (`VarSet.member` coveredVars)) IgnoreAll (clValue cl)
             -- Turn cand into a substitution.
             -- Since we ignored the free variables in sorts, we better patch up
-            -- the substitution with some dummy term (Sort Prop) rather than __IMPOSSIBLE__.
-            let dummy = Sort Prop
-            return $ parallelS $ map (maybe dummy var) cand
+            -- the substitution with some dummy term rather than __IMPOSSIBLE__.
+            return $ parallelS $ map (maybe __DUMMY_TERM__ var) cand
           ) return -- Phew, we've got the checkpoint! All is well.
   -- Apply substitution to constraint and pray that the Gods are merciful on us.
   return $ applySubst sigma (clValue cl)
@@ -364,19 +365,19 @@ solveSizeConstraints_ flag cs0 = do
     forM ccs $ \ (c0, HypSizeConstraint cxt hids hs sc) -> do
       case simplify1 (\ sc -> return [sc]) sc of
         Left _ -> typeError . GenericDocError =<< do
-          text "Contradictory size constraint" <+> prettyTCM c0
+          "Contradictory size constraint" <+> prettyTCM c0
         Right cs -> return $ (c0,) . HypSizeConstraint cxt hids hs <$> cs
 
   -- Cluster constraints according to the meta variables they mention.
   -- @csNoM@ are the constraints that do not mention any meta.
   let (csNoM, csMs) = (`List.partitionMaybe` ccs') $ \ p@(c0, c) ->
-        fmap (p,) $ List.uncons $ map (metaId . sizeMetaId) $ Set.toList $ flexs c
+        fmap (p,) $ NeList.fromList $ map (metaId . sizeMetaId) $ Set.toList $ flexs c
   -- @css@ are the clusters of constraints.
-      css :: [[(CC,HypSizeConstraint)]]
+      css :: [NonemptyList (CC,HypSizeConstraint)]
       css = cluster' csMs
 
   -- Check that the closed constraints are valid.
-  solveCluster flag csNoM
+  whenJust (NeList.fromList csNoM) $ solveCluster flag
 
   -- Now, process the clusters.
   forM_ css $ solveCluster flag
@@ -385,16 +386,16 @@ solveSizeConstraints_ flag cs0 = do
 
 -- | Solve a cluster of constraints sharing some metas.
 --
-solveCluster :: DefaultToInfty -> [(CC,HypSizeConstraint)] -> TCM ()
-solveCluster flag [] = return ()
+solveCluster :: DefaultToInfty -> NonemptyList (CC,HypSizeConstraint) -> TCM ()
 solveCluster flag ccs = do
-  let cs = map snd ccs
+  let cs = fmap snd ccs
+  let prettyCs   = map prettyTCM $ NeList.toList cs
   let err reason = typeError . GenericDocError =<< do
         vcat $
-          [ text $ "Cannot solve size constraints" ] ++ map prettyTCM cs ++
+          [ text $ "Cannot solve size constraints" ] ++ prettyCs ++
           [ text $ "Reason: " ++ reason ]
   reportSDoc "tc.size.solve" 20 $ vcat $
-    [ text "Solving constraint cluster" ] ++ map prettyTCM cs
+    [ "Solving constraint cluster" ] ++ prettyCs
   -- Find the super context of all contexts.
 {-
   -- We use the @'ctxId'@s.
@@ -409,7 +410,7 @@ solveCluster flag ccs = do
           _          -> a
       res = foldl' max (Right ci) cis'
       noContext ((c,is),(c',is')) = typeError . GenericDocError =<< vcat
-        [ text "Cannot solve size constraints; the following constraints have no common typing context:"
+        [ "Cannot solve size constraints; the following constraints have no common typing context:"
         , prettyTCM c
         , prettyTCM c'
         ]
@@ -417,7 +418,7 @@ solveCluster flag ccs = do
 -}
   -- We rely on the fact that contexts are only extended...
   -- Just take the longest context.
-  let HypSizeConstraint gamma hids hs _ = List.maximumBy (compare `on` (length . sizeContext)) cs
+  let HypSizeConstraint gamma hids hs _ = Fold.maximumBy (compare `on` (length . sizeContext)) cs
   -- Length of longest context.
   let n = size gamma
 
@@ -426,11 +427,11 @@ solveCluster flag ccs = do
   -- Canonicalize the constraints.
   -- This is unsound in the presence of hypotheses.
       csC :: [SizeConstraint]
-      csC = applyWhen (null hs) (mapMaybe canonicalizeSizeConstraint) csL
+      csC = applyWhen (null hs) (mapMaybe canonicalizeSizeConstraint) $ NeList.toList csL
   reportSDoc "tc.size.solve" 30 $ vcat $
-    [ text "Size hypotheses" ] ++
+    [ "Size hypotheses" ] ++
     map (prettyTCM . HypSizeConstraint gamma hids hs) hs ++
-    [ text "Canonicalized constraints" ] ++
+    [ "Canonicalized constraints" ] ++
     map (prettyTCM . HypSizeConstraint gamma hids hs) csC
 
   -- -- ALT:
@@ -461,8 +462,8 @@ solveCluster flag ccs = do
   --     csC = for cs $ \ (HypSizeConstraint cxt _ _ c) -> sub (size cxt) c
 
   -- reportSDoc "tc.size.solve" 30 $ vcat $
-  --   [ text "Size hypotheses" ]           ++ map prettyC hs ++
-  --   [ text "Canonicalized constraints" ] ++ map prettyC csC
+  --   [ "Size hypotheses" ]           ++ map prettyC hs ++
+  --   [ "Canonicalized constraints" ] ++ map prettyC csC
 
   -- Convert size metas to flexible vars.
   let metas :: [SizeMeta]
@@ -479,7 +480,7 @@ solveCluster flag ccs = do
   -- --    g :: Size.Graph NamedRigid Int Label
   -- g <- either err return $ constraintGraph csF hg
   -- reportSDoc "tc.size.solve" 40 $ vcat $
-  --   [ text "Constraint graph"
+  --   [ "Constraint graph"
   --   , text (show g)
   --   ]
 
@@ -508,12 +509,12 @@ solveCluster flag ccs = do
     t <- getMetaType x
     reportSDoc "tc.size.solve" 20 $ inTopContext $ modifyContext (const gamma) $ do
       let args = map (Apply . defaultArg . var) xs
-      text "solution " <+> prettyTCM (MetaV x args) <+> text " := " <+> prettyTCM u
+      "solution " <+> prettyTCM (MetaV x args) <+> " := " <+> prettyTCM u
     reportSDoc "tc.size.solve" 60 $ vcat
       [ text $ "  xs = " ++ show xs
       , text $ "  u  = " ++ show u
       ]
-    ifM (isFrozen x) (return Set.empty) $ do
+    ifM (isFrozen x `or2M` (not <$> asksTC envAssignMetas)) (return Set.empty) $ do
       assignMeta n x t xs u
       return $ Set.singleton x
     -- WRONG:
@@ -532,7 +533,7 @@ solveCluster flag ccs = do
   let noIP = Set.null $ Set.intersection ims ms
 
   unless (null ms) $ reportSDoc "tc.size.solve" 30 $ fsep $
-    [ text "cluster did not solve these size metas: " ] ++ map prettyTCM (Set.toList ms)
+    [ "cluster did not solve these size metas: " ] ++ map prettyTCM (Set.toList ms)
 
   solvedAll <- do
     -- If no metas are left, we have solved this cluster completely.
@@ -548,12 +549,12 @@ solveCluster flag ccs = do
         -- If one variable is frozen, we cannot set it (and hence not all) to ∞
         let no = do
               reportSDoc "tc.size.solve" 30 $
-                prettyTCM (MetaV m []) <+> text "is frozen, cannot set it to ∞"
+                prettyTCM (MetaV m []) <+> "is frozen, cannot set it to ∞"
               return False
-        ifM (isFrozen m `or2M` do not <$> asks envAssignMetas) no $ {-else-} do
+        ifM (isFrozen m `or2M` do not <$> asksTC envAssignMetas) no $ {-else-} do
           reportSDoc "tc.size.solve" 20 $
-            text "solution " <+> prettyTCM (MetaV m []) <+>
-            text " := "      <+> prettyTCM inf
+            "solution " <+> prettyTCM (MetaV m []) <+>
+            " := "      <+> prettyTCM inf
           t <- jMetaType . mvJudgement <$> lookupMeta m
           TelV tel core <- telView t
           unlessM (isJust <$> isSizeType core) __IMPOSSIBLE__
@@ -562,10 +563,10 @@ solveCluster flag ccs = do
 
   -- Double check.
   when solvedAll $ do
-    let cs0 = map fst ccs
+    let cs0 = map fst $ toList ccs
         -- Error for giving up
         cannotSolve = typeError . GenericDocError =<<
-          vcat (text "Cannot solve size constraints" : map prettyTCM cs0)
+          vcat ("Cannot solve size constraints" : map prettyTCM cs0)
     flip catchError (const cannotSolve) $
       noConstraints $
         forM_ cs0 $ \ cl -> enterClosure cl solveConstraint
@@ -739,7 +740,7 @@ instance PrettyTCM HypSizeConstraint where
       -- text ("[#cxt=" ++ show (size cxt) ++ "]") <+> do
       prettyList (map prettyTCM cxtNames) <+> do
       applyUnless (null hs)
-       (((hcat $ punctuate (text ", ") $ map prettyTCM hs) <+> text "|-") <+>)
+       (((hcat $ punctuate ", " $ map prettyTCM hs) <+> "|-") <+>)
        (prettyTCM c)
 
 -- | Turn a constraint over de Bruijn indices into a size constraint.
@@ -750,7 +751,7 @@ computeSizeConstraint c = do
     case clValue c of
       ValueCmp CmpLeq _ u v -> do
         reportSDoc "tc.size.solve" 50 $ sep $
-          [ text "converting size constraint"
+          [ "converting size constraint"
           , prettyTCM c
           ]
         ma <- sizeExpr u
@@ -770,7 +771,7 @@ sizeExpr :: Term -> TCM (Maybe DBSizeExpr)
 sizeExpr u = do
   u <- reduce u -- Andreas, 2009-02-09.
                 -- This is necessary to surface the solutions of metavariables.
-  reportSDoc "tc.conv.size" 60 $ text "sizeExpr:" <+> prettyTCM u
+  reportSDoc "tc.conv.size" 60 $ "sizeExpr:" <+> prettyTCM u
   s <- sizeView u
   case s of
     SizeInf     -> return $ Just Infty
@@ -783,6 +784,7 @@ sizeExpr u = do
       _           -> return Nothing
   where
     isVar (Proj{})  = Nothing
+    isVar (IApply _ _ v) = isVar (Apply (defaultArg v))
     isVar (Apply v) = case unArg v of
       Var i [] -> Just i
       _        -> Nothing
