@@ -5,9 +5,9 @@ module Agda.TypeChecking.Telescope where
 import Prelude hiding (null)
 
 import Control.Applicative hiding (empty)
-import Control.Monad (unless, guard)
+import Control.Monad
 
-import Data.Foldable (forM_)
+import Data.Foldable (forM_, find)
 import Data.IntSet (IntSet)
 import qualified Data.IntSet as IntSet
 import qualified Data.List as List
@@ -18,10 +18,12 @@ import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Pattern
 import Agda.Syntax.Position
 
+import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Free
+import Agda.TypeChecking.Warnings
 
 import Agda.Utils.Functor
 import Agda.Utils.List
@@ -36,10 +38,12 @@ import qualified Agda.Utils.VarSet as VarSet
 import Agda.Utils.Impossible
 
 -- | Flatten telescope: (Γ : Tel) -> [Type Γ]
-flattenTel :: Telescope -> [Dom Type]
+flattenTel :: Subst t a => Tele (Dom a) -> [Dom a]
 flattenTel EmptyTel          = []
 flattenTel (ExtendTel a tel) = raise (size tel + 1) a : flattenTel (absBody tel)
 
+{-# SPECIALIZE flattenTel :: Telescope -> [Dom Type] #-}
+
 -- | Order a flattened telescope in the correct dependeny order: Γ ->
 --   Permutation (Γ -> Γ~)
 --
@@ -75,22 +79,22 @@ teleNames = map (fst . unDom) . telToList
 teleArgNames :: Telescope -> [Arg ArgName]
 teleArgNames = map (argFromDom . fmap fst) . telToList
 
-teleArgs :: (DeBruijn a) => Telescope -> [Arg a]
+teleArgs :: (DeBruijn a) => Tele (Dom t) -> [Arg a]
 teleArgs tel =
   [ Arg info (deBruijnVar i)
-  | (i, Dom info (n,_)) <- zip (downFrom $ size l) l ]
+  | (i, Dom {domInfo = info, unDom = (n,_)}) <- zip (downFrom $ size l) l ]
   where l = telToList tel
 
 withNamedArgsFromTel :: [a] -> Telescope -> [NamedArg a]
 xs `withNamedArgsFromTel` tel =
   [ Arg info (Named (Just $ Ranged noRange $ argNameToString name) x)
-  | (x, Dom info (name,_)) <- zip xs l ]
+  | (x, Dom {domInfo = info, unDom = (name,_)}) <- zip xs l ]
   where l = telToList tel
 
 teleNamedArgs :: (DeBruijn a) => Telescope -> [NamedArg a]
 teleNamedArgs tel =
-  [ Arg info (Named (Just $ Ranged noRange $ argNameToString name) (deBruijnVar i))
-  | (i, Dom info (name,_)) <- zip (downFrom $ size l) l ]
+  [ fmap (deBruijnVar i <$) $ namedArgFromDom dom
+  | (i, dom) <- zip (downFrom $ size l) l ]
   where l = telToList tel
 
 -- | A variant of `teleNamedArgs` which takes the argument names (and the argument info)
@@ -100,7 +104,7 @@ teleNamedArgs tel =
 tele2NamedArgs :: (DeBruijn a) => Telescope -> Telescope -> [NamedArg a]
 tele2NamedArgs tel0 tel =
   [ Arg info (Named (Just $ Ranged noRange $ argNameToString argName) (debruijnNamedVar varName i))
-  | (i, Dom info (argName,_), Dom _ (varName,_)) <- zip3 (downFrom $ size l) l0 l ]
+  | (i, Dom{domInfo = info, unDom = (argName,_)}, Dom{unDom = (varName,_)}) <- zip3 (downFrom $ size l) l0 l ]
   where
   l  = telToList tel
   l0 = telToList tel0
@@ -149,7 +153,7 @@ varDependencies tel = allDependencies IntSet.empty
     ts = flattenTel tel
 
     directDependencies :: Int -> IntSet
-    directDependencies i = allFreeVars $ ts !! (n-1-i)
+    directDependencies i = allFreeVars $ indexWithDefault __IMPOSSIBLE__ ts (n-1-i)
 
     allDependencies :: IntSet -> IntSet -> IntSet
     allDependencies =
@@ -210,7 +214,8 @@ splitTelescopeExact is tel = guard ok $> SplitTel tel1 tel2 perm
     checkDependencies soFar []     = True
     checkDependencies soFar (j:js) = ok && checkDependencies (IntSet.insert j soFar) js
       where
-        fv' = allFreeVars $ ts0 !! (n-1-j)
+        fv' = allFreeVars $  -- newline because of CPP
+                indexWithDefault __IMPOSSIBLE__ ts0 (n-1-j)
         fv  = fv' `IntSet.intersection` IntSet.fromAscList [ 0 .. n-1 ]
         ok  = fv `IntSet.isSubsetOf` soFar
 
@@ -227,6 +232,17 @@ splitTelescopeExact is tel = guard ok $> SplitTel tel1 tel2 perm
     m     = size is
     (tel1, tel2) = telFromList -*- telFromList $ splitAt m $ telToList tel'
 
+instantiateTelescopeN
+  :: Telescope    -- ^ ⊢ Γ
+  -> [(Int,Term)] -- ^ Γ ⊢ var k_i : A_i ascending order, Γ ⊢ u_i : A_i
+  -> Maybe (Telescope,    -- ⊢ Γ'
+            Substitution) -- Γ' ⊢ σ : Γ
+instantiateTelescopeN tel []         = return (tel, IdS)
+instantiateTelescopeN tel ((k,t):xs) = do
+  (tel', sigma, _) <- instantiateTelescope tel k t
+  (tel'', sigma')  <- instantiateTelescopeN tel' (map (subtract 1 -*- applyPatSubst sigma) xs)
+  return (tel'', applyPatSubst sigma sigma')
+
 -- | Try to instantiate one variable in the telescope (given by its de Bruijn
 --   level) with the given value, returning the new telescope and a
 --   substitution to the old one. Returns Nothing if the given value depends
@@ -280,7 +296,7 @@ expandTelescopeVar gamma k delta c = (tel', rho)
     (ts1,a:ts2) = fromMaybe __IMPOSSIBLE__ $
                     splitExactlyAt k $ telToList gamma
 
-    cpi         = ConPatternInfo
+    cpi         = noConPatternInfo
       { conPRecord = Just PatOSystem
       , conPType   = Just $ snd <$> argFromDom a
       , conPLazy   = True
@@ -317,49 +333,227 @@ telViewUpTo' n p t = do
   where
     absV a x (TelV tel t) = TelV (ExtendTel a (Abs x tel)) t
 
+telViewPath :: Type -> TCM TelView
+telViewPath = telViewUpToPath (-1)
+
+-- | @telViewUpToPath n t@ takes off $t$
+--   the first @n@ (or arbitrary many if @n < 0@) function domains or Path types.
+telViewUpToPath :: Int -> Type -> TCM TelView
+telViewUpToPath 0 t = return $ TelV EmptyTel t
+telViewUpToPath n t = do
+  vt <- pathViewAsPi $ t
+  case vt of
+    Left (a,b)     -> absV a (absName b) <$> telViewUpToPath (n - 1) (absBody b)
+    Right (El _ t) | Pi a b <- t
+                   -> absV a (absName b) <$> telViewUpToPath (n - 1) (absBody b)
+    Right t        -> return $ TelV EmptyTel t
+  where
+    absV a x (TelV tel t) = TelV (ExtendTel a (Abs x tel)) t
+
+-- | [[ (i,(x,y)) ]] = [(i=0) -> x, (i=1) -> y]
+type Boundary = Boundary' (Term,Term)
+type Boundary' a = [(Term,a)]
+
+-- | Like @telViewUpToPath@ but also returns the @Boundary@ expected
+-- by the Path types encountered. The boundary terms live in the
+-- telescope given by the @TelView@.
+-- Each point of the boundary has the type of the codomain of the Path type it got taken from, see @fullBoundary@.
+telViewUpToPathBoundary' :: Int -> Type -> TCM (TelView,Boundary)
+telViewUpToPathBoundary' 0 t = return $ (TelV EmptyTel t,[])
+telViewUpToPathBoundary' n t = do
+  vt <- pathViewAsPi' $ t
+  case vt of
+    Left ((a,b),xy) -> addEndPoints xy . absV a (absName b) <$> telViewUpToPathBoundary' (n - 1) (absBody b)
+    Right (El _ t) | Pi a b <- t
+                   -> absV a (absName b) <$> telViewUpToPathBoundary' (n - 1) (absBody b)
+    Right t        -> return $ (TelV EmptyTel t,[])
+  where
+    absV a x (TelV tel t, cs) = (TelV (ExtendTel a (Abs x tel)) t, cs)
+    addEndPoints xy (telv@(TelV tel _),cs) = (telv, (var $ size tel - 1, xyInTel):cs)
+      where
+       xyInTel = raise (size tel) xy
+
+
+fullBoundary :: Telescope -> Boundary -> Boundary
+fullBoundary tel bs =
+      -- tel = Γ
+      -- ΔΓ ⊢ b
+      -- Δ ⊢ a = PiPath Γ bs b
+      -- Δ.Γ ⊢ T is the codomain of the PathP at variable i
+      -- Δ.Γ ⊢ i : I
+      -- Δ.Γ ⊢ [ (i=0) -> t_i; (i=1) -> u_i ] : T
+      -- Δ.Γ | PiPath Γ bs A ⊢ teleElims tel bs : b
+   let es = teleElims tel bs
+       l  = size tel
+   in map (\ (t@(Var i []), xy) -> (t, xy `applyE` (drop (l - i) es))) bs
+
+-- | @(TelV Γ b, [(i,t_i,u_i)]) <- telViewUpToPathBoundary n a@
+--  Input:  Δ ⊢ a
+--  Output: ΔΓ ⊢ b
+--          ΔΓ ⊢ i : I
+--          ΔΓ ⊢ [ (i=0) -> t_i; (i=1) -> u_i ] : b
+telViewUpToPathBoundary :: Int -> Type -> TCM (TelView,Boundary)
+telViewUpToPathBoundary i a = do
+   (telv@(TelV tel b), bs) <- telViewUpToPathBoundary' i a
+   return $ (telv, fullBoundary tel bs)
+
+-- | @(TelV Γ b, [(i,t_i,u_i)]) <- telViewUpToPathBoundaryP n a@
+--  Input:  Δ ⊢ a
+--  Output: Δ.Γ ⊢ b
+--          Δ.Γ ⊢ T is the codomain of the PathP at variable i
+--          Δ.Γ ⊢ i : I
+--          Δ.Γ ⊢ [ (i=0) -> t_i; (i=1) -> u_i ] : T
+-- Useful to reconstruct IApplyP patterns after teleNamedArgs Γ.
+telViewUpToPathBoundaryP :: Int -> Type -> TCM (TelView,Boundary)
+telViewUpToPathBoundaryP = telViewUpToPathBoundary'
+
+telViewPathBoundaryP :: Type -> TCM (TelView,Boundary)
+telViewPathBoundaryP = telViewUpToPathBoundaryP (-1)
+
+
+-- | @teleElimsB args bs = es@
+--  Input:  Δ.Γ ⊢ args : Γ
+--          Δ.Γ ⊢ T is the codomain of the PathP at variable i
+--          Δ.Γ ⊢ i : I
+--          Δ.Γ ⊢ bs = [ (i=0) -> t_i; (i=1) -> u_i ] : T
+--  Output: Δ.Γ | PiPath Γ bs A ⊢ es : A
+teleElims :: DeBruijn a => Telescope -> Boundary' (a,a) -> [Elim' a]
+teleElims tel [] = map Apply $ teleArgs tel
+teleElims tel boundary = recurse (teleArgs tel)
+  where
+    recurse = fmap updateArg
+    matchVar x =
+      snd <$> flip find boundary (\case
+        (Var i [],_) -> i == x
+        _            -> __IMPOSSIBLE__)
+    updateArg a@(Arg info p) =
+      case deBruijnView p of
+        Just i | Just (t,u) <- matchVar i -> IApply t u p
+        _                                 -> Apply a
+
+pathViewAsPi :: Type -> TCM (Either (Dom Type, Abs Type) Type)
+pathViewAsPi t = either (Left . fst) Right <$> pathViewAsPi' t
+
+pathViewAsPi' :: Type -> TCM (Either ((Dom Type, Abs Type), (Term,Term)) Type)
+pathViewAsPi' t = do
+  pathViewAsPi'whnf <*> reduce t
+
+pathViewAsPi'whnf :: TCM (Type -> Either ((Dom Type, Abs Type), (Term,Term)) Type)
+pathViewAsPi'whnf = do
+  view <- pathView'
+  minterval  <- getBuiltin' builtinInterval
+  return $ \ t -> case view t of
+    PathType s l p a x y | Just interval <- minterval ->
+      let name | Lam _ (Abs n _) <- unArg a = n
+               | otherwise = "i"
+          i = El Inf interval
+      in
+        Left $ ((defaultDom $ i, Abs name $ El (raise 1 s) $ raise 1 (unArg a) `apply` [defaultArg $ var 0]), (unArg x, unArg y))
+
+    _    -> Right t
+
+-- | returns Left (a,b) in case the type is @Pi a b@ or @PathP b _ _@
+--   assumes the type is in whnf.
+piOrPath :: Type -> TCM (Either (Dom Type, Abs Type) Type)
+piOrPath t = do
+  t <- pathViewAsPi'whnf <*> pure t
+  case t of
+    Left (p,_) -> return $ Left p
+    Right (El _ (Pi a b)) -> return $ Left (a,b)
+    Right t -> return $ Right t
+
+telView'UpToPath :: Int -> Type -> TCM TelView
+telView'UpToPath 0 t = return $ TelV EmptyTel t
+telView'UpToPath n t = do
+  vt <- pathViewAsPi'whnf <*> pure t
+  case vt of
+    Left ((a,b),_)     -> absV a (absName b) <$> telViewUpToPath (n - 1) (absBody b)
+    Right (El _ t) | Pi a b <- t
+                   -> absV a (absName b) <$> telViewUpToPath (n - 1) (absBody b)
+    Right t        -> return $ TelV EmptyTel t
+  where
+    absV a x (TelV tel t) = TelV (ExtendTel a (Abs x tel)) t
+
+telView'Path :: Type -> TCM TelView
+telView'Path = telView'UpToPath (-1)
+
+isPath :: Type -> TCM (Maybe (Dom Type, Abs Type))
+isPath t = either Just (const Nothing) <$> pathViewAsPi t
+
+telePatterns :: (DeBruijn a, DeBruijn (Pattern' a)) =>
+                 Telescope -> Boundary -> [NamedArg (Pattern' a)]
+telePatterns = telePatterns' teleNamedArgs
+
+telePatterns' :: (DeBruijn a, DeBruijn (Pattern' a)) =>
+                (forall a. (DeBruijn a) => Telescope -> [NamedArg a]) -> Telescope -> Boundary -> [NamedArg (Pattern' a)]
+telePatterns' f tel [] = f tel
+telePatterns' f tel boundary = recurse $ f tel
+  where
+    recurse = (fmap . fmap . fmap) updateVar
+    matchVar x =
+      snd <$> flip find boundary (\case
+        (Var i [],_) -> i == x
+        _            -> __IMPOSSIBLE__)
+    o = PatOSystem
+    updateVar x =
+      case deBruijnView x of
+        Just i | Just (t,u) <- matchVar i -> IApplyP o t u x
+        _                           -> VarP o x
+
 -- | Decomposing a function type.
 
-mustBePi :: MonadTCM tcm => Type -> tcm (Dom Type, Abs Type)
+mustBePi :: MonadReduce m => Type -> m (Dom Type, Abs Type)
 mustBePi t = ifNotPiType t __IMPOSSIBLE__ $ \ a b -> return (a,b)
 
 -- | If the given type is a @Pi@, pass its parts to the first continuation.
 --   If not (or blocked), pass the reduced type to the second continuation.
-ifPi :: MonadTCM tcm => Term -> (Dom Type -> Abs Type -> tcm a) -> (Term -> tcm a) -> tcm a
+ifPi :: MonadReduce m => Term -> (Dom Type -> Abs Type -> m a) -> (Term -> m a) -> m a
 ifPi t yes no = do
-  t <- liftTCM $ reduce t
+  t <- reduce t
   case t of
     Pi a b -> yes a b
     _      -> no t
 
 -- | If the given type is a @Pi@, pass its parts to the first continuation.
 --   If not (or blocked), pass the reduced type to the second continuation.
-ifPiType :: MonadTCM tcm => Type -> (Dom Type -> Abs Type -> tcm a) -> (Type -> tcm a) -> tcm a
+ifPiType :: MonadReduce m => Type -> (Dom Type -> Abs Type -> m a) -> (Type -> m a) -> m a
 ifPiType (El s t) yes no = ifPi t yes (no . El s)
 
 -- | If the given type is blocked or not a @Pi@, pass it reduced to the first continuation.
 --   If it is a @Pi@, pass its parts to the second continuation.
-ifNotPi :: MonadTCM tcm => Term -> (Term -> tcm a) -> (Dom Type -> Abs Type -> tcm a) -> tcm a
+ifNotPi :: MonadReduce m => Term -> (Term -> m a) -> (Dom Type -> Abs Type -> m a) -> m a
 ifNotPi = flip . ifPi
 
 -- | If the given type is blocked or not a @Pi@, pass it reduced to the first continuation.
 --   If it is a @Pi@, pass its parts to the second continuation.
-ifNotPiType :: MonadTCM tcm => Type -> (Type -> tcm a) -> (Dom Type -> Abs Type -> tcm a) -> tcm a
+ifNotPiType :: MonadReduce m => Type -> (Type -> m a) -> (Dom Type -> Abs Type -> m a) -> m a
 ifNotPiType = flip . ifPiType
 
--- | A safe variant of piApply.
+ifNotPiOrPathType :: (MonadReduce tcm, MonadTCM tcm) => Type -> (Type -> tcm a) -> (Dom Type -> Abs Type -> tcm a) -> tcm a
+ifNotPiOrPathType t no yes = do
+  ifPiType t yes (\ t -> either (uncurry yes . fst) (const $ no t) =<< (liftTCM pathViewAsPi'whnf <*> pure t))
 
-piApplyM :: Type -> Args -> TCM Type
-piApplyM t []           = return t
-piApplyM t (arg : args) = do
-  (_, b) <- mustBePi t
-  absApp b (unArg arg) `piApplyM` args
 
-piApply1 :: MonadTCM tcm => Type -> Term -> tcm Type
-piApply1 t v = do
-  (_, b) <- mustBePi t
-  return $ absApp b v
+-- | A safe variant of 'piApply'.
+
+class PiApplyM a where
+  piApplyM :: MonadReduce m => Type -> a -> m Type
+
+instance PiApplyM Term where
+  piApplyM t v = ifNotPiType t __IMPOSSIBLE__ {-else-} $ \ _ b -> return $ absApp b v
+
+instance PiApplyM a => PiApplyM (Arg a) where
+  piApplyM t = piApplyM t . unArg
+
+instance PiApplyM a => PiApplyM (Named n a) where
+  piApplyM t = piApplyM t . namedThing
+
+instance PiApplyM a => PiApplyM [a] where
+  piApplyM t = foldl (\ mt v -> mt >>= (`piApplyM` v)) (return t)
+
 
 -- | Compute type arity
+
 typeArity :: Type -> TCM Nat
 typeArity t = do
   TelV tel _ <- telView t
@@ -372,36 +566,41 @@ typeArity t = do
 data OutputTypeName
   = OutputTypeName QName
   | OutputTypeVar
+  | OutputTypeVisiblePi
   | OutputTypeNameNotYetKnown
   | NoOutputTypeName
 
--- | Strips all Pi's and return the head definition name, if possible.
-getOutputTypeName :: Type -> TCM OutputTypeName
+-- | Strips all hidden and instance Pi's and return the argument
+--   telescope and head definition name, if possible.
+getOutputTypeName :: Type -> TCM (Telescope, OutputTypeName)
 getOutputTypeName t = do
-  TelV tel t' <- telView t
-  ifBlocked (unEl t') (\ _ _ -> return OutputTypeNameNotYetKnown) $ \ _ v ->
+  TelV tel t' <- telViewUpTo' (-1) notVisible t
+  ifBlocked (unEl t') (\ _ _ -> return (tel , OutputTypeNameNotYetKnown)) $ \ _ v ->
     case v of
       -- Possible base types:
-      Def n _  -> return $ OutputTypeName n
-      Sort{}   -> return NoOutputTypeName
-      Var n _  -> return OutputTypeVar
+      Def n _  -> return (tel , OutputTypeName n)
+      Sort{}   -> return (tel , NoOutputTypeName)
+      Var n _  -> return (tel , OutputTypeVar)
+      Pi{}     -> return (tel , OutputTypeVisiblePi)
       -- Not base types:
       Con{}    -> __IMPOSSIBLE__
       Lam{}    -> __IMPOSSIBLE__
       Lit{}    -> __IMPOSSIBLE__
       Level{}  -> __IMPOSSIBLE__
       MetaV{}  -> __IMPOSSIBLE__
-      Pi{}     -> __IMPOSSIBLE__
       DontCare{} -> __IMPOSSIBLE__
+      Dummy s    -> __IMPOSSIBLE_VERBOSE__ s
 
+-- | Register the definition with the given type as an instance
 addTypedInstance :: QName -> Type -> TCM ()
 addTypedInstance x t = do
-  n <- getOutputTypeName t
+  (tel , n) <- getOutputTypeName t
   case n of
     OutputTypeName n -> addNamedInstance x n
     OutputTypeNameNotYetKnown -> addUnknownInstance x
-    NoOutputTypeName -> typeError $ WrongInstanceDeclaration
-    OutputTypeVar -> typeError $ WrongInstanceDeclaration
+    NoOutputTypeName -> warning $ WrongInstanceDeclaration
+    OutputTypeVar -> warning $ WrongInstanceDeclaration
+    OutputTypeVisiblePi -> warning $ InstanceWithExplicitArg x
 
 resolveUnknownInstanceDefs :: TCM ()
 resolveUnknownInstanceDefs = do
