@@ -222,11 +222,11 @@ buildWithFunction cxtNames f aux t delta qs npars withSub perm n1 n cs = mapM bu
             where
             fromWithP (A.WithP _ p) = p
             fromWithP _ = __IMPOSSIBLE__
-      reportSDoc "tc.with" 50 $ text "inheritedPats:" <+> vcat [ prettyA p <+> text "=" <+> prettyTCM v <+> text ":" <+> prettyTCM a
+      reportSDoc "tc.with" 50 $ "inheritedPats:" <+> vcat [ prettyA p <+> "=" <+> prettyTCM v <+> ":" <+> prettyTCM a
                                                                | A.ProblemEq p v a <- inheritedPats ]
       (strippedPats, ps') <- stripWithClausePatterns cxtNames f aux t delta qs npars perm ps
-      reportSDoc "tc.with" 50 $ hang (text "strippedPats:") 2 $
-                                  vcat [ prettyA p <+> text "==" <+> prettyTCM v <+> (text ":" <+> prettyTCM t)
+      reportSDoc "tc.with" 50 $ hang "strippedPats:" 2 $
+                                  vcat [ prettyA p <+> "==" <+> prettyTCM v <+> (":" <+> prettyTCM t)
                                        | A.ProblemEq p v t <- strippedPats ]
       rhs <- buildRHS strippedPats rhs
       let (ps1, ps2) = splitAt n1 ps'
@@ -234,7 +234,7 @@ buildWithFunction cxtNames f aux t delta qs npars withSub perm n1 n cs = mapM bu
                      (inheritedPats ++ strippedPats)
                      rhs wh catchall
       reportSDoc "tc.with" 20 $ vcat
-        [ text "buildWithClause returns" <+> prettyA result
+        [ "buildWithClause returns" <+> prettyA result
         ]
       return result
 
@@ -289,7 +289,7 @@ Example:
 
   test : (s : Nat × Stream Nat) (t : Stream Nat) → SEq (delay s) t → SEq t (delay s)
   ~force (test (a     , as) t p) with force t
-  ~force (test (suc n , as) t p) | b , bs = {!!}
+  ~force (test (suc n , as) t p) | b , bs = ?
 @
 
 With function:
@@ -339,27 +339,27 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
   -- instantiations from qs, so we make sure
   -- that t is the top-level type of the parent function and add patterns for
   -- the module parameters to ps before stripping.
-  let paramPat i _ = A.VarP $ A.BindName (cxtNames !! i)
+  let paramPat i _ = A.VarP $ A.BindName $ indexWithDefault __IMPOSSIBLE__ cxtNames i
       ps' = zipWith (fmap . fmap . paramPat) [0..] (take npars qs) ++ ps
   psi <- insertImplicitPatternsT ExpandLast ps' t
   reportSDoc "tc.with.strip" 10 $ vcat
-    [ text "stripping patterns"
-    , nest 2 $ text "t   = " <+> prettyTCM t
-    , nest 2 $ text "ps  = " <+> fsep (punctuate comma $ map prettyA ps)
-    , nest 2 $ text "ps' = " <+> fsep (punctuate comma $ map prettyA ps')
-    , nest 2 $ text "psi = " <+> fsep (punctuate comma $ map prettyA psi)
-    , nest 2 $ text "qs  = " <+> fsep (punctuate comma $ map (prettyTCM . namedArg) qs)
-    , nest 2 $ text "perm= " <+> text (show perm)
+    [ "stripping patterns"
+    , nest 2 $ "t   = " <+> prettyTCM t
+    , nest 2 $ "ps  = " <+> fsep (punctuate comma $ map prettyA ps)
+    , nest 2 $ "ps' = " <+> fsep (punctuate comma $ map prettyA ps')
+    , nest 2 $ "psi = " <+> fsep (punctuate comma $ map prettyA psi)
+    , nest 2 $ "qs  = " <+> fsep (punctuate comma $ map (prettyTCM . namedArg) qs)
+    , nest 2 $ "perm= " <+> text (show perm)
     ]
 
   -- Andreas, 2015-11-09 Issue 1710: self starts with parent-function, not with-function!
   (ps', strippedPats) <- runWriterT $ strip (Def parent []) t psi qs
   reportSDoc "tc.with.strip" 50 $ nest 2 $
-    text "strippedPats:" <+> vcat [ prettyA p <+> text "=" <+> prettyTCM v <+> text ":" <+> prettyTCM a | A.ProblemEq p v a <- strippedPats ]
+    "strippedPats:" <+> vcat [ prettyA p <+> "=" <+> prettyTCM v <+> ":" <+> prettyTCM a | A.ProblemEq p v a <- strippedPats ]
   let psp = permute perm ps'
   reportSDoc "tc.with.strip" 10 $ vcat
-    [ nest 2 $ text "ps' = " <+> fsep (punctuate comma $ map prettyA ps')
-    , nest 2 $ text "psp = " <+> fsep (punctuate comma $ map prettyA $ psp)
+    [ nest 2 $ "ps' = " <+> fsep (punctuate comma $ map prettyA ps')
+    , nest 2 $ "psp = " <+> fsep (punctuate comma $ map prettyA $ psp)
     ]
   return (strippedPats, psp)
   where
@@ -377,10 +377,10 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
     -- Case: out of with-clause patterns.
     strip self t [] qs@(_ : _) = do
       reportSDoc "tc.with.strip" 15 $ vcat
-        [ text "strip (out of A.Patterns)"
-        , nest 2 $ text "qs  =" <+> fsep (punctuate comma $ map (prettyTCM . namedArg) qs)
-        , nest 2 $ text "self=" <+> prettyTCM self
-        , nest 2 $ text "t   =" <+> prettyTCM t
+        [ "strip (out of A.Patterns)"
+        , nest 2 $ "qs  =" <+> fsep (punctuate comma $ map (prettyTCM . namedArg) qs)
+        , nest 2 $ "self=" <+> prettyTCM self
+        , nest 2 $ "t   =" <+> prettyTCM t
         ]
       -- Andreas, 2015-06-11, issue 1551:
       -- As the type t develops, we need to insert more implicit patterns,
@@ -412,15 +412,15 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
     strip self t ps0@(p0 : ps) qs0@(q : qs) = do
       p <- liftTCM $ (traverse . traverse) expandLitPattern p0
       reportSDoc "tc.with.strip" 15 $ vcat
-        [ text "strip"
-        , nest 2 $ text "ps0 =" <+> fsep (punctuate comma $ map prettyA ps0)
-        , nest 2 $ text "exp =" <+> prettyA p
-        , nest 2 $ text "qs0 =" <+> fsep (punctuate comma $ map (prettyTCM . namedArg) qs0)
-        , nest 2 $ text "self=" <+> prettyTCM self
-        , nest 2 $ text "t   =" <+> prettyTCM t
+        [ "strip"
+        , nest 2 $ "ps0 =" <+> fsep (punctuate comma $ map prettyA ps0)
+        , nest 2 $ "exp =" <+> prettyA p
+        , nest 2 $ "qs0 =" <+> fsep (punctuate comma $ map (prettyTCM . namedArg) qs0)
+        , nest 2 $ "self=" <+> prettyTCM self
+        , nest 2 $ "t   =" <+> prettyTCM t
         ]
       case namedArg q of
-        ProjP o d -> case A.maybePostfixProjP p of
+        ProjP o d -> case A.isProjP p of
           Just (o', AmbQ ds) -> do
             -- Andreas, 2016-12-28, issue #2360:
             -- We disambiguate the projection in the with clause
@@ -446,11 +446,14 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
         -- we can strip the dot from the with clause.
         VarP PatODot x | A.DotP _ u <- namedArg p
                        , A.Var y <- unScope u ->
-          (setNamedArg p (A.VarP (A.BindName y)) :) <$>
+          (setNamedArg p (A.VarP $ A.BindName y) :) <$>
             recurse (var (dbPatVarIndex x))
 
         VarP _ x  -> (p :) <$> recurse (var (dbPatVarIndex x))
 
+        IApplyP{}  -> typeError $ GenericError $ "with clauses not supported in the presence of Path patterns" -- TODO maybe we can support them now?
+        DefP{}  -> typeError $ GenericError $ "with clauses not supported in the presence of hcomp patterns" -- TODO this should actually be impossible
+
         DotP o v  -> do
           (a, _) <- mustBePi t
           tell [ProblemEq (namedArg p) v a]
@@ -458,7 +461,7 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
 
         q'@(ConP c ci qs') -> do
          reportSDoc "tc.with.strip" 60 $
-           text "parent pattern is constructor " <+> prettyTCM c
+           "parent pattern is constructor " <+> prettyTCM c
          (a, b) <- mustBePi t
          -- The type of the current pattern is a datatype.
          Def d es <- liftTCM $ normalise (unEl $ unDom a)
@@ -510,7 +513,7 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
 
           p@(A.PatternSynP pi' c' ps') -> do
              reportSDoc "impossible" 10 $
-               text "stripWithClausePatterns: encountered pattern synonym " <+> prettyA p
+               "stripWithClausePatterns: encountered pattern synonym " <+> prettyA p
              __IMPOSSIBLE__
 
           p -> do
@@ -524,21 +527,25 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
 
           p@(A.PatternSynP pi' c' [ps']) -> do
              reportSDoc "impossible" 10 $
-               text "stripWithClausePatterns: encountered pattern synonym " <+> prettyA p
+               "stripWithClausePatterns: encountered pattern synonym " <+> prettyA p
              __IMPOSSIBLE__
 
           _ -> mismatch
       where
         recurse v = do
-          t' <- piApply1 t v
+          caseMaybeM (liftTCM $ isPath t) (return ()) $ \ _ ->
+            typeError $ GenericError $
+              "With-clauses currently not supported under Path abstraction."
+
+          t' <- piApplyM t v
           strip (self `apply1` v) t' ps qs
 
         mismatch = addContext delta $ typeError $
           WithClausePatternMismatch (namedArg p0) q
         mismatchOrigin o o' = addContext delta . typeError . GenericDocError =<< fsep
-          [ text "With clause pattern"
+          [ "With clause pattern"
           , prettyA p0
-          , text "is not an instance of its parent pattern"
+          , "is not an instance of its parent pattern"
           , P.fsep <$> prettyTCMPatterns [q]
           , text $ "since the parent pattern is " ++ prettyProjOrigin o ++
                    " and the with clause pattern is " ++ prettyProjOrigin o'
@@ -575,25 +582,27 @@ stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
           Defn {defType = ct, theDef = Constructor{conPars = np}}  <- getConInfo c
           -- Compute the argument telescope for the constructor
           let ct' = ct `piApply` take np us
-          TelV tel' _ <- liftTCM $ telView ct'
+          TelV tel' _ <- liftTCM $ telViewPath ct'
+          -- (TelV tel' _, _boundary) <- liftTCM $ telViewPathBoundaryP ct'
 
           reportSDoc "tc.with.strip" 20 $
-            vcat [ text "ct  = " <+> prettyTCM ct
-                 , text "ct' = " <+> prettyTCM ct'
-                 , text "np  = " <+> text (show np)
-                 , text "us  = " <+> prettyList (map prettyTCM us)
-                 , text "us' = " <+> prettyList (map prettyTCM $ take np us)
+            vcat [ "ct  = " <+> prettyTCM ct
+                 , "ct' = " <+> prettyTCM ct'
+                 , "np  = " <+> text (show np)
+                 , "us  = " <+> prettyList (map prettyTCM us)
+                 , "us' = " <+> prettyList (map prettyTCM $ take np us)
                  ]
 
+          -- TODO Andrea: preserve IApplyP patterns in v, see _boundary?
           -- Compute the new type
           let v  = Con c ci [ Apply $ Arg info (var i) | (i, Arg info _) <- zip (downFrom $ size qs') qs' ]
               t' = tel' `abstract` absApp (raise (size tel') b) v
               self' = tel' `abstract` apply1 (raise (size tel') self) v  -- Issue 1546
 
           reportSDoc "tc.with.strip" 15 $ sep
-            [ text "inserting implicit"
+            [ "inserting implicit"
             , nest 2 $ prettyList $ map prettyA (ps' ++ ps)
-            , nest 2 $ text ":" <+> prettyTCM t'
+            , nest 2 $ ":" <+> prettyTCM t'
             ]
 
           -- Insert implicit patterns (just for the constructor arguments)
@@ -674,25 +683,25 @@ withDisplayForm f aux delta1 delta2 n qs perm@(Perm m _) lhsPerm = do
                  . flip (foldr addContext) (for [1..n] $ \ i -> "w" ++ show i)
                  . addContext delta2
   reportSDoc "tc.with.display" 20 $ vcat
-    [ text "withDisplayForm"
+    [ "withDisplayForm"
     , nest 2 $ vcat
-      [ text "f      =" <+> text (prettyShow f)
-      , text "aux    =" <+> text (prettyShow aux)
-      , text "delta1 =" <+> prettyTCM delta1
-      , text "delta2 =" <+> do addContext delta1 $ prettyTCM delta2
-      , text "n      =" <+> text (show n)
-      , text "perm   =" <+> text (show perm)
-      , text "top    =" <+> do addFullCtx $ prettyTCM topArgs
-      , text "qs     =" <+> prettyList (map pretty qs)
-      , text "qsToTm =" <+> prettyTCM tqs0 -- ctx would be permuted form of delta1 ++ delta2
-      , text "ys     =" <+> text (show ys)
-      , text "rho    =" <+> text (prettyShow rho)
-      , text "qs[rho]=" <+> do addFullCtx $ prettyTCM tqs
-      , text "dt     =" <+> do addFullCtx $ prettyTCM dt
+      [ "f      =" <+> text (prettyShow f)
+      , "aux    =" <+> text (prettyShow aux)
+      , "delta1 =" <+> prettyTCM delta1
+      , "delta2 =" <+> do addContext delta1 $ prettyTCM delta2
+      , "n      =" <+> text (show n)
+      , "perm   =" <+> text (show perm)
+      , "top    =" <+> do addFullCtx $ prettyTCM topArgs
+      , "qs     =" <+> prettyList (map pretty qs)
+      , "qsToTm =" <+> prettyTCM tqs0 -- ctx would be permuted form of delta1 ++ delta2
+      , "ys     =" <+> text (show ys)
+      , "rho    =" <+> text (prettyShow rho)
+      , "qs[rho]=" <+> do addFullCtx $ prettyTCM tqs
+      , "dt     =" <+> do addFullCtx $ prettyTCM dt
       ]
     ]
   reportSDoc "tc.with.display" 70 $ nest 2 $ vcat
-      [ text "raw    =" <+> text (show display)
+      [ "raw    =" <+> text (show display)
       ]
 
   return display
@@ -719,6 +728,7 @@ patsToElims = map $ toElim . fmap namedThing
 
     toTerm :: DeBruijnPattern -> DisplayTerm
     toTerm p = case p of
+      IApplyP o _ _ x -> DTerm $ var $ dbPatVarIndex x -- TODO, should be an Elim' DisplayTerm ?
       ProjP _ d   -> DDef d [] -- WRONG. TODO: convert spine to non-spine ... DDef d . defaultArg
       VarP PatODot x -> DDot  $ var $ dbPatVarIndex x
       VarP o x      -> DTerm  $ var $ dbPatVarIndex x
@@ -726,3 +736,4 @@ patsToElims = map $ toElim . fmap namedThing
       DotP o t    -> DDot   $ t
       ConP c cpi ps -> DCon c (fromConPatternInfo cpi) $ toTerms ps
       LitP l      -> DTerm  $ Lit l
+      DefP o q ps -> DDef q $ map Apply $ toTerms ps
