@@ -6,9 +6,8 @@ module Agda.Utils.AssocList where
 
 import Prelude hiding (lookup)
 
-import Control.Applicative
-
 import qualified Data.List as List
+import Data.Function
 
 import Agda.Utils.Tuple
 
@@ -43,6 +42,13 @@ insert k v = ((k,v) :)
 update :: Eq k => k -> v -> AssocList k v -> AssocList k v
 update k v = updateAt k $ const v
 
+-- | O(n).
+--   Delete a binding.
+--   The key must be in the domain of the finite map.
+--   Otherwise, an internal error is raised.
+delete :: Eq k => k -> AssocList k v -> AssocList k v
+delete k = List.deleteBy ((==) `on` fst) (k, __IMPOSSIBLE__)
+
 -- | O(n).
 --   Update the value at a key with a certain function.
 --   The key must be in the domain of the finite map.
