@@ -89,7 +89,7 @@ import qualified Data.List as List
 import qualified Data.Map.Strict as Map
 import Data.Map.Strict (Map)
 import qualified Data.Maybe as Maybe
-import Data.Maybe (maybeToList, fromMaybe, catMaybes)
+import Data.Maybe (maybeToList, fromMaybe)
 import qualified Data.Set as Set
 import Data.Set (Set)
 import qualified Data.Tree as Tree
@@ -171,7 +171,7 @@ data Edge n e = Edge
 
 instance (Pretty n, Pretty e) => Pretty (Edge n e) where
   pretty (Edge s t e) =
-    pretty s <+> text "--(" <> pretty e <> text ")-->" <+> pretty t
+    pretty s <+> "--(" <> pretty e <> ")-->" <+> pretty t
 
 ------------------------------------------------------------------------
 -- Queries
