@@ -1,3 +1,5 @@
+{-# LANGUAGE ScopedTypeVariables #-}
+
 -- | Utility functions for lists.
 
 module Agda.Utils.List where
@@ -17,11 +19,20 @@ import qualified Agda.Utils.Bag as Bag
 
 import Agda.Utils.Tuple
 
+-- | Append a single element at the end.
+--   Time: O(length); use only on small lists.
+snoc :: [a] -> a -> [a]
+snoc xs x = xs ++ [x]
+
 -- | Case distinction for lists, with list first.
 --   Cf. 'Agda.Utils.Null.ifNull'.
 caseList :: [a] -> b -> (a -> [a] -> b) -> b
-caseList []     n c = n
-caseList (x:xs) n c = c x xs
+caseList xs n c = listCase n c xs
+
+-- | Case distinction for lists, with list first.
+--   Cf. 'Agda.Utils.Null.ifNull'.
+caseListM :: Monad m => m [a] -> m b -> (a -> [a] -> m b) -> m b
+caseListM mxs n c = listCase n c =<< mxs
 
 -- | Case distinction for lists, with list last.
 listCase :: b -> (a -> [a] -> b) -> [a] -> b
@@ -32,13 +43,21 @@ listCase n c (x:xs) = c x xs
 headMaybe :: [a] -> Maybe a
 headMaybe = listToMaybe
 
--- | Head function (safe). Returns a value on empty lists.
+-- | Head function (safe). Returns a default value on empty lists.
 --
 -- > headWithDefault 42 []      = 42
 -- > headWithDefault 42 [1,2,3] = 1
 headWithDefault :: a -> [a] -> a
 headWithDefault def = fromMaybe def . headMaybe
 
+-- | Tail function (safe).
+tailMaybe :: [a] -> Maybe [a]
+tailMaybe = fmap snd . uncons
+
+-- | Tail function (safe).  Returns a default list on empty lists.
+tailWithDefault :: [a] -> [a] -> [a]
+tailWithDefault def = fromMaybe def . tailMaybe
+
 -- | Last element (safe).
 lastMaybe :: [a] -> Maybe a
 lastMaybe [] = Nothing
@@ -52,6 +71,17 @@ last2 (x : y : xs) = Just $ loop x y xs
   loop x y (z:xs) = loop y z xs
 last2 _ = Nothing
 
+-- | Drop from the end of a list.
+--   @dropEnd n = reverse . drop n . reverse@
+--   (Forces the whole list even for @n==0@.)
+dropEnd :: forall a. Int -> [a] -> [a]
+dropEnd n = snd . foldr f (n, [])
+  where
+  f :: a -> (Int, [a]) -> (Int, [a])
+  f x (n, xs)
+    | n <= 0    = (0, x:xs)
+    | otherwise = (n-1, xs)
+
 -- | Opposite of cons @(:)@, safe.
 uncons :: [a] -> Maybe (a, [a])
 uncons []     = Nothing
@@ -68,12 +98,28 @@ initLast (a:as) = Just $ loop a as where
   loop a []      = ([], a)
   loop a (b : bs) = mapFst (a:) $ loop b bs
 
+-- | init, safe.
+initMaybe :: [a] -> Maybe [a]
+initMaybe = fmap fst . initLast
+
 -- | Lookup function (partially safe).
 (!!!) :: [a] -> Int -> Maybe a
 []       !!! _         = Nothing
 (x : _)  !!! 0         = Just x
 (_ : xs) !!! n         = xs !!! (n - 1)
 
+-- | Lookup function with default value for index out of range.
+--   The name is chosen akin to 'Data.List.genericIndex'.
+indexWithDefault :: a -> [a] -> Int -> a
+indexWithDefault a []       _ = a
+indexWithDefault a (x : _)  0 = x
+indexWithDefault a (_ : xs) n = indexWithDefault a xs (n - 1)
+
+-- | Find an element satisfying a predicate and return it with its index.
+--   TODO: more efficient implementation!?
+findWithIndex :: (a -> Bool) -> [a] -> Maybe (a, Int)
+findWithIndex p as = headMaybe $ filter (p . fst) $ zip as [0..]
+
 -- | downFrom n = [n-1,..1,0]
 downFrom :: Integral a => a -> [a]
 downFrom n | n <= 0     = []
@@ -108,6 +154,19 @@ deal f a ~(bs,cs) = case f a of
   Left  b -> (b:bs, cs)
   Right c -> (bs, c:cs)
 
+-- | Split off the largest suffix whose elements satisfy a predicate.
+--
+--   @spanEnd p xs = (ys, zs)@
+--   where @xs = ys ++ zs@
+--   and @all p zs@
+--   and @maybe True (not . p) (lastMaybe yz)@.
+spanEnd :: forall a. (a -> Bool) -> [a] -> ([a], [a])
+spanEnd p = snd . foldr f (True, ([], []))
+  where
+  f :: a -> (Bool, ([a], [a])) -> (Bool, ([a], [a]))
+  f x (b', (xs, ys)) = (b, if b then (xs, x:ys) else (x:xs, ys))
+    where b = b' && p x
+
 -- | A generalized version of @takeWhile@.
 --   (Cf. @mapMaybe@ vs. @filter@).
 takeWhileJust :: (a -> Maybe b) -> [a] -> [b]
@@ -174,21 +233,39 @@ stripPrefixBy eq = loop
     | eq p r    = loop pat rest
     | otherwise = Nothing
 
--- | Result of 'preOrSuffix'.
-data PreOrSuffix a
-  = IsPrefix a [a] -- ^ First list is prefix of second.
-  | IsSuffix a [a] -- ^ First list is suffix of second.
-  | IsBothfix      -- ^ The lists are equal.
-  | IsNofix        -- ^ The lists are incomparable.
+-- | @stripSuffix suf xs = Just pre@ iff @xs = pre ++ suf@.
+stripSuffix :: Eq a => Suffix a -> [a] -> Maybe (Prefix a)
+stripSuffix [] = Just
+stripSuffix s  = stripReversedSuffix (reverse s)
 
--- | Compare lists with respect to prefix partial order.
-preOrSuffix :: Eq a => [a] -> [a] -> PreOrSuffix a
-preOrSuffix []     []     = IsBothfix
-preOrSuffix []     (b:bs) = IsPrefix b bs
-preOrSuffix (a:as) []     = IsSuffix a as
-preOrSuffix (a:as) (b:bs)
-  | a == b    = preOrSuffix as bs
-  | otherwise = IsNofix
+type ReversedSuffix a = [a]
+
+-- | @stripReversedSuffix rsuf xs = Just pre@ iff @xs = pre ++ reverse suf@.
+stripReversedSuffix :: forall a. Eq a => ReversedSuffix a -> [a] -> Maybe (Prefix a)
+stripReversedSuffix rs = final . foldr step (SSSStrip rs)
+  where
+  -- Step of the automaton (reading input from right to left).
+  step :: a -> StrSufSt a -> StrSufSt a
+  step x = \case
+    SSSMismatch   -> SSSMismatch
+    SSSResult xs  -> SSSResult (x:xs)
+    SSSStrip []   -> SSSResult [x]
+    SSSStrip (y:ys)
+      | x == y    -> SSSStrip ys
+      | otherwise -> SSSMismatch
+
+  -- Output of the automaton.
+  final :: StrSufSt a -> Maybe (Prefix a)
+  final = \case
+    SSSResult xs -> Just xs
+    SSSStrip []  -> Just []
+    _            -> Nothing  -- We have not stripped the whole suffix or encountered a mismatch.
+
+-- | Internal state for stripping suffix.
+data StrSufSt a
+  = SSSMismatch                 -- ^ Error.
+  | SSSStrip (ReversedSuffix a) -- ^ "Negative string" to remove from end. List may be empty.
+  | SSSResult [a]               -- ^ "Positive string" (result). Non-empty list.
 
 -- | Split a list into sublists. Generalisation of the prelude function
 --   @words@.
