@@ -8,8 +8,14 @@ module Agda.Utils.Monad
     )
     where
 
-import Prelude             hiding (concat)
-import Control.Monad       hiding (mapM, forM)
+import Prelude       hiding (concat)
+import Control.Monad hiding (mapM, forM)
+
+#if __GLASGOW_HASKELL__ >= 800
+import qualified Control.Monad.Fail as Fail
+#endif
+
+import Control.Monad.Identity ( Identity )
 import Control.Monad.State
 import Control.Monad.Writer
 import Data.Traversable as Trav hiding (for, sequence)
@@ -23,10 +29,20 @@ import Agda.Utils.Except
   )
 
 import Agda.Utils.List
+import Agda.Utils.Null (ifNotNullM)
 
 #include "undefined.h"
 import Agda.Utils.Impossible
 
+---------------------------------------------------------------------------
+
+#if __GLASGOW_HASKELL__ >= 800
+instance Fail.MonadFail Identity where
+  fail = error
+#endif
+
+---------------------------------------------------------------------------
+
 -- | Binary bind.
 (==<<) :: Monad m => (a -> b -> m c) -> (m a, m b) -> m c
 k ==<< (ma, mb) = ma >>= \ a -> k a =<< mb
@@ -107,6 +123,14 @@ mapM' f = Fold.foldl (\ mb a -> liftM2 mappend mb (f a)) (return mempty)
 forM' :: (Foldable t, Monad m, Monoid b) => t a -> (a -> m b) -> m b
 forM' = flip mapM'
 
+-- Variations of Traversable
+
+mapMM :: (Traversable t, Monad m) => (a -> m b) -> m (t a) -> m (t b)
+mapMM f mxs = Trav.mapM f =<< mxs
+
+forMM :: (Traversable t, Monad m) => m (t a) -> (a -> m b) -> m (t b)
+forMM = flip mapMM
+
 -- Continuation monad -----------------------------------------------------
 
 -- Andreas, 2017-04-11, issue #2543
@@ -135,6 +159,13 @@ dropWhileM :: Monad m => (a -> m Bool) -> [a] -> m [a]
 dropWhileM p []       = return []
 dropWhileM p (x : xs) = ifM (p x) (dropWhileM p xs) (return (x : xs))
 
+-- | A monadic version of @'dropWhileEnd' :: (a -> Bool) -> [a] -> m [a]@.
+--   Effects happen starting at the end of the list until @p@ becomes false.
+dropWhileEndM :: Monad m => (a -> m Bool) -> [a] -> m [a]
+dropWhileEndM p []       = return []
+dropWhileEndM p (x : xs) = ifNotNullM (dropWhileEndM p xs) (return . (x:)) $ {-else-}
+  ifM (p x) (return []) (return [x])
+
 -- | A ``monadic'' version of @'partition' :: (a -> Bool) -> [a] -> ([a],[a])
 partitionM :: (Functor m, Applicative m) => (a -> m Bool) -> [a] -> m ([a],[a])
 partitionM f [] =
