@@ -9,7 +9,13 @@ module Agda.Utils.Pretty
 
 import Data.Int ( Int32 )
 import Data.Data (Data(..))
+import qualified Data.Map as Map
 
+#if __GLASGOW_HASKELL__ < 800
+import qualified Data.Semigroup as S
+#endif
+
+import qualified Text.PrettyPrint as P
 import Text.PrettyPrint hiding (TextDetails(Str), empty)
 
 import Agda.Utils.NonemptyList
@@ -17,6 +23,13 @@ import Agda.Utils.NonemptyList
 #include "undefined.h"
 import Agda.Utils.Impossible
 
+#if __GLASGOW_HASKELL__ < 800
+-- ASR (2019-03-27). I used @(<>)@ from the monoid instance because I
+-- cannot hide this instance.
+instance S.Semigroup Doc where
+  (<>) = (<>)
+#endif
+
 -- * Pretty class
 
 -- | While 'Show' is for rendering data in Haskell syntax,
@@ -54,9 +67,12 @@ instance Pretty Char where
 instance Pretty Doc where
   pretty = id
 
+instance Pretty () where
+  pretty _ = P.empty
+
 instance Pretty a => Pretty (Maybe a) where
-  prettyPrec p Nothing  = text "Nothing"
-  prettyPrec p (Just x) = mparens (p > 0) $ text "Just" <+> prettyPrec 10 x
+  prettyPrec p Nothing  = "Nothing"
+  prettyPrec p (Just x) = mparens (p > 0) $ "Just" <+> prettyPrec 10 x
 
 instance Pretty a => Pretty [a] where
   pretty = prettyList
