@@ -31,6 +31,7 @@ quote s = "\"" ++ concatMap escape s ++ "\""
            | c `elem` escapeChars = ['\\', c]
            | otherwise            = [c]
 
+  escapeChars :: String
   escapeChars = "\"\\"
 
 -- | Turns the string into a Haskell string literal, avoiding escape
@@ -102,7 +103,7 @@ ltrim = dropWhile isSpace
 
 -- | Remove trailing whitespace.
 rtrim :: String -> String
-rtrim = reverse . ltrim . reverse
+rtrim = List.dropWhileEnd isSpace
 
 -- | Remove leading and trailing whitesapce.
 trim :: String -> String
