@@ -88,16 +88,3 @@ addSuffix s NoSuffix      = s
 addSuffix s (Prime n)     = s ++ replicate n '\''
 addSuffix s (Index i)     = s ++ show i
 addSuffix s (Subscript i) = s ++ map toSubscriptDigit (show i)
-
--- | Add first available @Suffix@ to a name.
-
-nameVariant
-  :: (String -> Bool) -- ^ Is the given name already taken?
-  -> String           -- ^ Name of which we want an available variant.
-  -> String           -- ^ Name extended by suffix that is not taken already.
-nameVariant taken x
-  | taken x   = addSuffix x $ trampoline step $ Subscript 1
-  | otherwise = x
-  where
-    -- if the current suffix is taken, repeat with next suffix, else done
-    step s = if taken (addSuffix x s) then Right (nextSuffix s) else Left s
