@@ -2,10 +2,6 @@
 {-# LANGUAGE BangPatterns #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 
-#if __GLASGOW_HASKELL__ <= 708
-{-# LANGUAGE DeriveDataTypeable #-}
-#endif
-
 -- | Strict tries (based on "Data.Map.Strict" and "Agda.Utils.Maybe.Strict").
 
 module Agda.Utils.Trie
@@ -15,6 +11,7 @@ module Agda.Utils.Trie
   , toList, toAscList, toListOrderedBy
   , lookup, member, lookupPath, lookupTrie
   , mapSubTries, filter
+  , valueAt
   ) where
 
 import Prelude hiding (null, lookup, filter)
@@ -30,12 +27,9 @@ import qualified Data.Map.Strict as Map
 import Data.Monoid
 import qualified Data.List as List
 
-#if __GLASGOW_HASKELL__ <= 708
-import Data.Typeable ( Typeable )
-#endif
-
 import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Null
+import Agda.Utils.Lens
 
 -- | Finite map from @[k]@ to @v@.
 --
@@ -45,9 +39,6 @@ data Trie k v = Trie !(Strict.Maybe v) !(Map k (Trie k v))
            , Eq
            , Functor
            , Foldable
-#if __GLASGOW_HASKELL__ <= 708
-           , Typeable
-#endif
            )
 
 -- | Empty trie.
@@ -173,3 +164,9 @@ filter p (Trie mv ts) = Trie mv' (Map.filter (not . null) $ filter p <$> ts)
         Strict.Just v | p v -> mv
         _                   -> Strict.Nothing
 
+-- | Key lens.
+valueAt :: Ord k => [k] -> Lens' (Maybe v) (Trie k v)
+valueAt path f t = f (lookup path t) <&> \ case
+  Nothing -> delete path t
+  Just v  -> insert path v t
+
